//
// GPSMapEdit
// (c) Konstantin Galichsky (kg@geopainting.com), 2003-2009
// ALAN and ALAN Map 500 are registered trademarks of ALAN Group.
//
// ALAN Map 500 (OS 3.x) TRL tracks export.
//

# include "StdAfx.h"
# include "AlanTrl3.h"
# include "Save.h"
# include "Map.h"
# include "Status.h"
# include "SaveFileDlg.h"
# include "FileWriter.h"
# include "fast_floor.h"
# include "Options.h"
# include "Globals.h"

class CAlanTrl3CFExport : public CMapSaver {
  public:
	virtual bool CanSave (const map_t & _map) {return true;}
	virtual SaveStatus_t Save (HANDLE _hFile, const char * _strFileName, const map_t & _map);

  private:
	SaveStatus_t _Save (HANDLE _hFile, const map_t & _map);
};

DECLARE_WPTRK_EXPORT (CAlanTrl3CFExport, "alan-trl-3cf", "trl", "ALAN Map 500 OS 3.x CF tracks (*.trl)");

SaveStatus_t CAlanTrl3CFExport::Save (HANDLE _hFile, const char * _strFileName, const map_t & _map) {
	if (_hFile) {
		return _Save (_hFile, _map);
	} else {
		//
		// Prepare file name.
		//
		char strFileName [MAX_PATH + 1];
		::strncpy (strFileName, _strFileName, MAX_PATH);
		strFileName [MAX_PATH] = '\0';
		if (! SaveFileDlg ("trl", "ALAN Map 500 tracks (*.trl)\0*.trl\0", strFileName))
			return ssCanceled;

		//
		// Open the file.
		//
		const CFileForWrite hFile (strFileName);
		if (! hFile.IsValid ())
			return Error ("Can't create file:\n %s", strFileName);

		return _Save (hFile, _map);
	}
}

SaveStatus_t CAlanTrl3CFExport::_Save (HANDLE _hFile, const map_t & _map) {
	// Get time zone shift.
	TIME_ZONE_INFORMATION tzi;
	if (::GetTimeZoneInformation (& tzi) == TIME_ZONE_ID_DAYLIGHT)
		tzi.Bias += tzi.DaylightBias;
	const double dtTimeShift = float (tzi.Bias)/(24*60);

	CTrl3CFTrackLog * const pLogs = new CTrl3CFTrackLog [c_cTrl3CFMaxTracks];
	CArrayHolder<CTrl3CFTrackLog> pLogs_Holder (pLogs);
	::memset (pLogs, 0, sizeof (CTrl3CFTrackLog)*c_cTrl3CFMaxTracks);

	CTrl3CFHdr hdr;
	::memset (& hdr, 0, sizeof (hdr));
	hdr.dwSignature = c_dwTrl3Signature;

	//
	// Write the file.
	//
	SetProgressStatus (0);

	size_t cOrdinal = 0;
	bool bLostPoints = false;

	size_t cRgn = 0;
	const size_t cRgns = _map.GetRgnCount ();
	for (size_t cKind = 0; cKind < rkMax; ++ cKind) {
		const layer_t & layer = _map.layers [cKind];
		const rgns_t & rgns = layer.rgns;
		for (rgns_t::const_iterator iRgn = rgns.begin (); iRgn != rgns.end (); ++ iRgn, ++ cRgn) {
			const rgn_t & rgn = * iRgn;
			if (rgn.IsDeleted () || ! rgn.IsTrack ())
				continue;

			// Save track.
			for (size_t cElement = 0; cElement < rgn.elements.size (); ++ cElement) {
				const rgn_element_t & element = rgn.elements [cElement];
				if (element.bDeleted)
					continue;
				const points_t & points = element.points;
				const size_t cPoints = points.size ();
				if (cPoints == 0)
					continue;

				if (cOrdinal >= c_cTrl3CFMaxTracks)
					break;

				SYSTEMTIME st0;
				bool bFirstTimeStamp = true;

				const track_info_t * const pTrkExtra = rgn.pExtra->wtr.track ();
				const size_t cPointsExtra = element.pExtra ? element.pExtra->TrackPoints.size () : 0;

				CTrl3CFTrackLog & log = pLogs [cOrdinal];
				for (size_t cPoint = 0; cPoint < cPoints; ++ cPoint) {
					if (cPoint >= c_cTrl3MaxPoints) {
						bLostPoints = true;
						break;
					}

					CTrl3Point & pt = log.points [cPoint];
					const point_t & point = points [cPoint];

					// Position.
					pt.lX =  fast_floor (point.x*36000 + .5f);
					pt.lY = -fast_floor (point.y*36000 + .5f);

					if (pTrkExtra && cPoint < cPointsExtra) {
						const trackpoint_info_t & PointExtra = element.pExtra->TrackPoints [cPoint];

						CTrl3Extra & extra = log.extra [cPoint];
						CTrl3Time  & time  = log.time  [cPoint];

						// Altitude.
						if (PointExtra.wFlags & trackpoint_info_t::defAltitude)
							extra.wAltitude = fast_floor (PointExtra.fAltitude/c_fTrl3CFAltitudeFactor + .5f);

						// Speed.
						if (PointExtra.wFlags & trackpoint_info_t::defSpeed)
							extra.wSpeed = fast_floor (PointExtra.fSpeed/c_fTrl3CFSpeedFactor + .5f);

						// Time.
						if (PointExtra.wFlags & trackpoint_info_t::defTime) {
							SYSTEMTIME st;
							::VariantTimeToSystemTime (PointExtra.dtTime - dtTimeShift, & st);
							time.hour = st.wHour;
							time.min  = st.wMinute;
							time.sec  = st.wSecond;

							if (bFirstTimeStamp) {
								bFirstTimeStamp = false;
								st0 = st;
							}
						}

						// DOP.
						if (PointExtra.wFlags & trackpoint_info_t::defDOP)
							time.dop = fast_floor (PointExtra.fDOP/c_fTrlDOPFactor + .5f);
					}
				}

				CTrl3Track & track = hdr.tracks [cOrdinal];

				track.wPoints    = cPoints;
				track.wNextPoint = cPoints;

				// Name.
				::strncpy (track.strName, rgn.strLabel.c_str (), sizeof (track.strName));
				track.strName [sizeof (track.strName) - 1] = '\0';

				// Comment.
				::memset (& track.strComment, ' ', sizeof (track.strComment));
				::strncpy (track.strComment, rgn.strComment.c_str (), rgn.strComment.size () < sizeof (track.strComment) ? rgn.strComment.size () : sizeof (track.strComment));

				track.dwOccupied = 1;
				track.dwShow     = (pTrkExtra->wFlags & track_info_t::defHideInGPS) ? 0 : 1;
				track.dwFill     = 0;

				// Date.
				if (! bFirstTimeStamp) {
					track.btYear  = st0.wYear - 2000;// + 208;
					track.btMonth = st0.wMonth;
					track.btDay   = st0.wDay;
				}

				++ cOrdinal;
			}

			SetProgressStatus (fast_floor (100.f*cRgn/cRgns + .5f));
		}
	}
	hdr.wTracks = cOrdinal;

	if (cOrdinal > c_cTrl3CFMaxTracks)
		Warning ("Only first %d tracks are saved.\n(NOTE:this is the TRL format limitation).", c_cTrl3CFMaxTracks);
	if (bLostPoints)
		Warning ("Just 2500 points per track are saved.\n(NOTE:this is the TRL format limitation).");

	CFileWriter fw (_hFile);
	fw.WriteBin (pLogs, c_cTrl3CFMaxTracks*sizeof (CTrl3CFTrackLog));
	fw.WriteBin (hdr);
	fw.WriteBin (300, 0);

	fw.FlushBuffer ();

	SetProgressStatus (100);

	return ssOK;
}
