//
// GPSMapEdit
// (c) Konstantin Galichsky (kg@geopainting.com), 2002-2009
//
// 'Attachments' dialog.
//

# include "StdAfx.h"
# include "Resource.h"
# include "Attachments.h"
# include "Map.h"
# include "MapView.h"
# include "NewMap.h"
# include "Load.h"
# include "UndoRedo.h"
# include "BackupAction.h"
# include "Scale.h"
# include "Position.h"
# include "Selection.h"
# include "DeleteAttachment.h"
# include "CombinedAction.h"
# include "ShowHideAttachment.h"
# include "Dialog.h"
# include "ListView.h"
# include "ParseHelpers.h"
# include "Globals.h"

static
void _LocateAttachment (const file_ref_t * _pFR) {
	ClearSelPoints ();
	UnselectObjects ();

	rect_t rect;
	{
		CCritSec::CLock lock (g_cs);

		for (size_t cKind = 0; cKind < rkMax; ++ cKind) {
			layer_t & layer = g_map.layers [cKind];
			rgns_t & rgns = layer.rgns;
			const size_t cRgns = layer.cRgns;

			for (rgns_t::iterator iRgn = rgns.begin (); iRgn != rgns.end (); ++ iRgn) {
				const rgn_t & rgn = * iRgn;
				if (rgn.IsDeleted () || rgn.pFileRef != _pFR)
					continue;

				rect.Extend (rgn.rectBound);
			
				find_info_t fi;
				fi.pLayer           = & layer;
				fi.iRgn             = iRgn;
				fi.cElementOrOrigin = -1;
				g_Selection.push_back (fi);
			}
		}

		for (renderers_t::iterator ri = g_map.renderers.begin (); ri != g_map.renderers.end (); ++ ri) {
			renderer_base_t * const pRenderer = & ri->data ();
			if (pRenderer != _pFR)
				continue;

			rect.Extend (pRenderer->rect);

			find_info_t fi;
			fi.pLayer    = NULL;
			fi.pRenderer = pRenderer;
			g_Selection.push_back (fi);

			break;
		}
	}

	if (! rect.Invalid ()) {
		if (rect.Width () < 2e-3f) {
			rect.x0 -= 1e-3f;
			rect.x1 += 1e-3f;
		}
		if (rect.Height () < 2e-3f) {
			rect.y0 -= 1e-3f;
			rect.y1 += 1e-3f;
		}

		SetScale (GetRectScale (rect), false);
		SetToCenter (point_t (rect.MeanX (), rect.MeanY ()));
	}
	Refresh ();
}

enum {
	catType = 0,
	catTrks = 1,
	catWpts = 2,
	catRts  = 3,
	catShow = 4,
	catFile = 5,
};

class CAttachments : public CDialog {
  public:
	CAttachments () : CDialog (IDD_ATTACHMENTS) {}

	static CAttachments * g_pAttachments;

	void Reset ();

	// Override.
	virtual bool OnInit (HWND _hwndFocus);
	virtual void OnShowModeless () {Reset ();}
	virtual void OnDestroy ();
	virtual void OnCommand (WORD _wCmd, WORD _wNotifyCode, HWND _hwndCtl);
	virtual bool OnNotify (int _idCtrl, NMHDR * _pInfo);

  private:
	CListView m_Att;

	void _FillAttachments ();
	file_ref_t * _GetFileRef (size_t _cItem = -1);
	void _UpdateButtons      (size_t _cItem = -1);

	static int CALLBACK _CompareItems (LPARAM _lItem1, LPARAM _lItem2, LPARAM _lData);
};

CAttachments * CAttachments::g_pAttachments;

void CAttachments::Reset () {
	_FillAttachments ();
	_UpdateButtons ();
}

int CALLBACK CAttachments::_CompareItems (LPARAM _lItem1, LPARAM _lItem2, LPARAM _lData) {
	const CAttachments * const pThis = reinterpret_cast<const CAttachments *> (_lData);
	const size_t cSortingColumn = pThis->m_Att.GetSortingColumn ();

	char str1 [64], str2 [64];
	ListView_GetItemText (pThis->m_Att, _lItem1, cSortingColumn, str1, 64); 
	ListView_GetItemText (pThis->m_Att, _lItem2, cSortingColumn, str2, 64);

	if (cSortingColumn == catTrks || cSortingColumn == catWpts || cSortingColumn == catRts) {
		// Compare as integers.
		return ParseDecU (str1, str1 + 64) - ParseDecU (str2, str2 + 64);
	} else
		// Compare as strings.
		return ::strcmp (str1, str2);
}

struct fr_data_t {
	size_t cWaypoints;
	size_t cTracks;
	size_t cRoutes;
	bool bSelected;
	fr_data_t () : cWaypoints (0), cTracks (0), cRoutes (0), bSelected (false) {}
};

void CAttachments::_FillAttachments () {
	CCritSec::CLock lock (g_cs);

	// Prepare statistics data.
	typedef std::map<const file_ref_t *, fr_data_t> fr2data_t;
	fr2data_t fr2data;

	for (size_t cKind = 0; cKind < rkMax; ++ cKind) {
		layer_t & layer = g_map.layers [cKind];
		rgns_t & rgns = layer.rgns;
		const size_t cRgns = layer.cRgns;

		for (rgns_t::const_iterator iRgn = rgns.begin (); iRgn != rgns.end (); ++ iRgn) {
			const rgn_t & rgn = * iRgn;
			if (rgn.IsDeleted () || ! rgn.IsWTR ())
				continue;

			fr_data_t & data = fr2data [rgn.pFileRef];
			if (rgn.IsWaypoint ())
				++ data.cWaypoints;
			else if (rgn.IsTrack ())
				++ data.cTracks;
			else if (rgn.IsRoute ())
				++ data.cRoutes;
		}
	}

	// Save selection status for each item.
	const size_t cItems = ListView_GetItemCount (m_Att);
	size_t cItem;
	for (cItem = 0; cItem < cItems; ++ cItem) {
		if (! IsItemSelected (m_Att, cItem))
			continue;

		file_ref_t * const pFR = _GetFileRef (cItem);
		if (! pFR)
			continue;

		fr_data_t & data = fr2data [pFR];
		data.bSelected = true;
	}

	// Save selection.
	const size_t cSel = ListView_GetSelectionMark (m_Att);
	const LPARAM lParamSel = cSel == -1 ? -1 : GetLParam (m_Att, cSel);

	ListView_DeleteAllItems (m_Att);

	// Tracks/waypoints/routes.
	cItem = 0;
	for (file_refs_t::const_iterator ia = g_map.file_refs.begin (); ia != g_map.file_refs.end (); ++ ia) {
		if (ia->bDeleted || ! ia->bAttachment)
			continue;

		const fr_data_t & data = fr2data [& * ia];

		InsertItem     (m_Att, cItem, reinterpret_cast<LPARAM> (& * ia), "vector");
		SetSubItemText (m_Att, cItem, catTrks, "%d", data.cTracks);
		SetSubItemText (m_Att, cItem, catWpts, "%d", data.cWaypoints);
		SetSubItemText (m_Att, cItem, catRts,  "%d", data.cRoutes);
		SetSubItemText (m_Att, cItem, catShow, ia->bVisible ? "shown" : "hidden");
		SetSubItemText (m_Att, cItem, catFile, "%s", const_cast<char *> (ia->strFileName.c_str ()));

		if (data.bSelected)
			SelectItem (m_Att, cItem);

		++ cItem;
	}

	// Raster maps.
	for (renderers_t::const_iterator ri = g_map.renderers.begin (); ri != g_map.renderers.end (); ++ ri) {
		const renderer_base_t * const pRenderer = & ri->data ();
		if (pRenderer->bDeleted || pRenderer->bOverlay)
			continue;

		InsertItem     (m_Att, cItem, reinterpret_cast<LPARAM> (pRenderer), "raster");
		SetSubItemText (m_Att, cItem, catTrks, "0");
		SetSubItemText (m_Att, cItem, catWpts, "0");
		SetSubItemText (m_Att, cItem, catRts,  "0");
		SetSubItemText (m_Att, cItem, catShow, pRenderer->bVisible ? "shown" : "hidden");
		SetSubItemText (m_Att, cItem, catFile, "%s", const_cast<char *> (pRenderer->strFileName.c_str ()));

		const fr_data_t & data = fr2data [pRenderer];
		if (data.bSelected)
			SelectItem (m_Att, cItem);

		++ cItem;
	}

	m_Att.Sort ();

	// Restore selection.
	if (lParamSel != -1)
		m_Att.SelectByLParam (lParamSel);
}

file_ref_t * CAttachments::_GetFileRef (size_t _cItem) {
	const size_t cSel = _cItem == -1 ? ListView_GetSelectionMark (m_Att) : _cItem;
	if (cSel == -1)
		return NULL;

	file_ref_t * const pFR = reinterpret_cast<file_ref_t *> (GetLParam (m_Att, cSel));
	assert (pFR);
	if (pFR->bDeleted)
		return NULL;

	return pFR;
}

void CAttachments::_UpdateButtons (size_t _cItem) {
	const size_t cItems = ListView_GetItemCount (m_Att);
	const size_t cSelItems = ListView_GetSelectedCount (m_Att);
	if (cSelItems <= 1) {
		if (_cItem == -1)
			_cItem = ListView_GetSelectionMark (m_Att);

		file_ref_t * const pFR = _GetFileRef (_cItem);
		EnableDlgItem (IDC_LOCATE_ATTACHMENT, pFR != NULL);
		EnableDlgItem (IDC_SHOW_ATTACHMENT,   pFR != NULL && ! pFR->bVisible);
		EnableDlgItem (IDC_HIDE_ATTACHMENT,   pFR != NULL &&   pFR->bVisible);
		EnableDlgItem (IDC_REMOVE_ATTACHMENT, pFR != NULL);
	} else {
		size_t cHidden = 0;
		size_t cShown = 0;

		for (size_t cItem = 0; cItem < cItems; ++ cItem) {
			if (! IsItemSelected (m_Att, cItem))
				continue;

			file_ref_t * const pFR = _GetFileRef (cItem);
			if (! pFR)
				continue;

			if (pFR->bVisible)
				++ cShown;
			else
				++ cHidden;
		}

		EnableDlgItem (IDC_LOCATE_ATTACHMENT, cSelItems == 1);
		EnableDlgItem (IDC_SHOW_ATTACHMENT,   cHidden   != 0);
		EnableDlgItem (IDC_HIDE_ATTACHMENT,   cShown    != 0);
		EnableDlgItem (IDC_REMOVE_ATTACHMENT, cSelItems != 0);
	}

	EnableDlgItem (IDC_REMOVE_ALL_ATTACHMENTS, cItems != 0);
}

bool CAttachments::OnInit (HWND _hwndFocus) {
	// Attachments.
	m_Att.Init (m_hWnd, IDC_MAP_ATTACHMENTS, "Attachments", -1);
	m_Att.SetExtStyle (LVS_EX_FULLROWSELECT | LVS_EX_HEADERDRAGDROP);

	AddColumn (m_Att, catType, "Type");
	AddColumn (m_Att, catTrks, "Trks");
	AddColumn (m_Att, catWpts, "Wpts");
	AddColumn (m_Att, catRts,  "Rts");
	AddColumn (m_Att, catShow, "Show");
	AddColumn (m_Att, catFile, "File path");

	m_Att.SetSorting (_CompareItems, reinterpret_cast<LPARAM> (this));
	_FillAttachments ();
	SetSelection (m_Att, 0);

	AutoSizeColumns (m_Att);
	m_Att.LoadHeaders ();

	// Buttons.
	_UpdateButtons ();

	return true;
}

void CAttachments::OnDestroy () {
	m_Att.SaveHeaders ();
	m_Att.Destroy ();
}

void CAttachments::OnCommand (WORD _wCmd, WORD _wNotifyCode, HWND _hwndCtl) {
	switch (_wCmd) {
		case IDC_SHOW_ATTACHMENT:
		case IDC_HIDE_ATTACHMENT: {
			const bool bShow = _wCmd == IDC_SHOW_ATTACHMENT;

			const size_t cSelItems = ListView_GetSelectedCount (m_Att);
			if (cSelItems <= 1) {
				file_ref_t * const pFR = _GetFileRef ();
				if (pFR)
					ShowAttachment (pFR, bShow, IsRenderer (g_map, * pFR));
			} else {
				const size_t cItems = ListView_GetItemCount (m_Att);
				for (size_t cItem = 0; cItem < cItems; ++ cItem) {
					if (! IsItemSelected (m_Att, cItem))
						continue;

					file_ref_t * const pFR = _GetFileRef (cItem);
					if (! pFR)
						continue;

					ShowAttachment (pFR, bShow, IsRenderer (g_map, * pFR));
				}
			}

			Reset ();
			SetFocus (bShow ? IDC_HIDE_ATTACHMENT : IDC_SHOW_ATTACHMENT);

			Refresh ();

			return;
		}

		case IDC_ADD_ATTACHMENT:
			CFileTypeInfo::ShowOpenFile (false);
			return;

		case IDC_REMOVE_ATTACHMENT: {
			const size_t cSel = ListView_GetSelectionMark (m_Att);

			const size_t cSelItems = ListView_GetSelectedCount (m_Att);
			if (cSelItems <= 1) {
				file_ref_t * const pFR = _GetFileRef ();
				if (pFR)
					DeleteAttachment (pFR, IsRenderer (g_map, * pFR));
			} else {
				const size_t cItems = ListView_GetItemCount (m_Att);
				for (size_t cItem = 0; cItem < cItems; ++ cItem) {
					if (! IsItemSelected (m_Att, cItem))
						continue;

					file_ref_t * const pFR = _GetFileRef (cItem);
					if (! pFR)
						continue;

					DeleteAttachment (pFR, IsRenderer (g_map, * pFR));
				}
			}

			_FillAttachments ();
			if (cSel != -1)
				SetSelection (m_Att, cSel);

			_UpdateButtons ();

			Refresh ();

			return;
		}

		case IDC_REMOVE_ALL_ATTACHMENTS: {
			const int nDepth = BeginCombinedAction ();

			const size_t cItems = ListView_GetItemCount (m_Att);
			for (size_t cItem = 0; cItem < cItems; ++ cItem) {
				file_ref_t * const pFR = _GetFileRef (cItem);
				if (! pFR)
					continue;

				DeleteAttachment (pFR, IsRenderer (g_map, * pFR));
			}

			EndCombinedAction (nDepth, "Detach All");

			Reset ();
			Refresh ();

			return;
		}

		case IDC_LOCATE_ATTACHMENT: {
			const size_t cSelItems = ListView_GetSelectedCount (m_Att);
			if (cSelItems != 1)
				return;

			file_ref_t * const pFR = _GetFileRef ();
			if (pFR) {
				const bool bWasVisible = pFR->bVisible;
				if (! bWasVisible)
					ShowAttachment (pFR, true, IsRenderer (g_map, * pFR));
				_LocateAttachment (pFR);

				if (! bWasVisible)
					Reset ();
			}

			return;
		}

		case IDC_REFRESH:
			Reset ();
			return;

		case IDCANCEL:
			Hide ();
			return;
	}
}

bool CAttachments::OnNotify (int _idCtrl, NMHDR * _pInfo) {
	switch (_pInfo->code) {
		case NM_RCLICK:
			m_Att.PopupMenu_OnHeaderRClick (_pInfo);
			break;

		case LVN_COLUMNCLICK: {
			// Save selection.
			const size_t cSel = ListView_GetSelectionMark (m_Att);
			const LPARAM lParamSel = cSel == -1 ? -1 : GetLParam (m_Att, cSel);

			m_Att.Sort_OnColumnClick (_pInfo);

			// Restore selection.
			if (lParamSel != -1)
				m_Att.SelectByLParam (lParamSel);

			break;
		}

		case LVN_ITEMCHANGED: {
			const NMLISTVIEW * const pInfo2 = reinterpret_cast<const NMLISTVIEW *> (_pInfo);
			if (pInfo2->uNewState & LVIS_SELECTED)
				_UpdateButtons (pInfo2->iItem);

			break;
		}
	}

	return false;
}

static
void _ResetAttachments () {
	if (CAttachments::g_pAttachments)
		CAttachments::g_pAttachments->Reset ();
}

INVOKE_AFTER_REMOVE_MAP (_ResetAttachments);
INVOKE_ON_BACKUP_MAP    (_ResetAttachments);
INVOKE_BEFORE_GC        (_ResetAttachments);

void ShowAttachments () {
	ShowModeless (CAttachments::g_pAttachments, g_hWnd);
}