//
// GPSMapEdit
// (c) Konstantin Galichsky (kg@geopainting.com), 2009
//
// Bookmark types & attributes.
//

# include "StdAfx.h"
# include "BookmarkTypes.h"
# include "Bookmark.h"
# include "Log.h"
# include "ParseHelpers.h"
# include "File.h"
# include "Wrappers.h"

bool ValidateBookmarkAttribute (BmAttributeKind _CurrentAttributeKind, const char * _strValue, size_t _cLen) {
	switch (_CurrentAttributeKind) {
		case bakString:
		case bakMultiString:
		case bakTime:
		case bakDate:
			return true;

		case bakReal: {
			if (_cLen == 0)
				return true;
			const char * p;
			float f;
			return ParseReal (_strValue, _strValue + _cLen, & p, f) && p == _strValue + _cLen;
		}

		case bakInteger: {
			if (_cLen == 0)
				return true;
			const char * p;
			unsigned long ul;
			return ParseDecU (_strValue, _strValue + _cLen, & p, ul) && p == _strValue + _cLen;
		}

		case bakBoolean:
			return _cLen == 1 && (_strValue [0] == '0' || _strValue [0] == '1');

		default:
			assert (false);
			return true;
	}
}

///////////////////////////////////////////////////////////////

static bm_types_t g_bm_types;
static bool g_bBookmarkTypesLoaded;

struct bm_info_t {
	CCPtr<CIcon> pIcon;
	string_t strTypeName;
	string_t strSubTypeName;
	const bm_attributes_t * pAttributes;
};
typedef std::map<string_t, bm_info_t> type2info_t;
static type2info_t g_type2info;

static
bool _LoadBookmarkTypes (bm_types_t & _Types);

static
void _InitBookmarkTypes () {
	if (g_bBookmarkTypesLoaded)
		return;
	g_bBookmarkTypesLoaded = true;

	_LoadBookmarkTypes (g_bm_types);

	const size_t cTypes = g_bm_types.size ();
	for (size_t cType = 0; cType < cTypes; ++ cType) {
		const bm_type_t & t = g_bm_types [cType];
		const bm_subtypes_t & vecSubTypes = t.vecSubTypes;

		const size_t cSubTypes = vecSubTypes.size ();
		for (size_t cSubType = 0; cSubType < cSubTypes; ++ cSubType) {
			const bm_subtype_t & st = vecSubTypes [cSubType];

			string_t strType (t.strId);
			strType.append (1, ':');
			strType.append (st.strId);

			bm_info_t & bmi = g_type2info [strType];
			bmi.pIcon          = st.pIcon;
			bmi.strTypeName    = t.strName;
			bmi.strSubTypeName = st.strName;
			bmi.pAttributes    = & st.vecAttributes;
		}
	}
}

bm_types_t & BookmarkTypes () {
	_InitBookmarkTypes ();
	return g_bm_types;
}

CCPtr<CIcon> GetBookmarkIcon (const string_t & _strType, const string_t & _strSubType) {
	_InitBookmarkTypes ();

	string_t strType (_strType);
	strType.append (1, ':');
	strType.append (_strSubType);

	const type2info_t::const_iterator i = g_type2info.find (strType);
	if (i == g_type2info.end ()) {
		if (_strType == "comment" && _strSubType == "info")
			return NULL;
		return GetBookmarkIcon ("comment", "info");
	}
	return i->second.pIcon;
}

bool GetBookmarkDisplayType (const bookmark_t & _bm, string_t & _strType, string_t & _strSubType) {
	_InitBookmarkTypes ();

	string_t strType (_bm.strType);
	strType.append (1, ':');
	strType.append (_bm.strSubType);

	const type2info_t::const_iterator i = g_type2info.find (strType);
	if (i == g_type2info.end ()) {
		_strType   .append (_bm.strType);
		_strSubType.append (_bm.strSubType);
		return false;
	} else {
		_strType   .append (i->second.strTypeName);
		_strSubType.append (i->second.strSubTypeName);
		return true;
	}
}

const bm_attributes_t * GetBookmarkTypeAttributes (const bookmark_t & _bm) {
	_InitBookmarkTypes ();

	string_t strType (_bm.strType);
	strType.append (1, ':');
	strType.append (_bm.strSubType);

	const type2info_t::const_iterator i = g_type2info.find (strType);
	if (i == g_type2info.end ())
		return NULL;
	return i->second.pAttributes;
}

const bm_attribute_t * FindBookmarkTypeAttribute (const bm_attributes_t & _attributes, const char * _strId) {
	const size_t cAttributes = _attributes.size ();
	for (size_t c = 0; c < cAttributes; ++ c)
		if (_attributes [c].strId == _strId)
			return & _attributes [c];
	return NULL;
}

//////////////////////////////////////

static
bool _ParseQuotedString (const char ** _pp, const char * _pEnd, string_t & _str) {
	if (** _pp == '\"') {
		++ * _pp;
		const char * const pStart = * _pp;
		while (** _pp != '\"' && * _pp < _pEnd)
			++ * _pp;
		_str.assign (pStart, * _pp);
		if (* _pp < _pEnd)
			++ * _pp;
	} else {
		const char * const pStart = * _pp;
		OmitNonSpaces (_pp, _pEnd);
		_str.assign (pStart, * _pp);
	}

	return true;
}

static
BmAttributeKind _ParseBookmarkKind (const char ** _pp, const char * _pEnd) {
	const char * const pKind = * _pp;
	OmitNonSpaces (_pp, _pEnd);

	if (HasPrefix (pKind, * _pp, "string", 6))
		return bakString;
	else if (HasPrefix (pKind, * _pp, "multistring", 11))
		return bakMultiString;
	else if (HasPrefix (pKind, * _pp, "number", 6))
		return bakReal;
	else if (HasPrefix (pKind, * _pp, "integer", 7))
		return bakInteger;
	else if (HasPrefix (pKind, * _pp, "boolean", 7))
		return bakBoolean;
	else if (HasPrefix (pKind, * _pp, "time", 4))
		return bakTime;
	else if (HasPrefix (pKind, * _pp, "date", 4))
		return bakDate;
	else
		return bakString;
}

static
bool _LoadBookmarkTypesFile (const char * _strPath, const char * _strFileName, bm_types_t & _Types) {
	ReportText ("Parsing '%s'...", _strFileName);

	CFileForRead hFile (_strFileName);
	if (! hFile.IsValid ()) {
		ReportWarning ("Failed to open file.");
		return false;
	}

	const size_t cFileSize = ::GetFileSize (hFile, NULL);
	CWinHandle hMapping = ::CreateFileMapping (hFile, NULL, PAGE_READONLY, 0, 0, NULL);
	if (! hMapping.IsValid ())
		return false;
	CMapViewOfFileForReading mvf (hMapping);
	const char * const pFileData = reinterpret_cast<const char *> (mvf.Get ());
	const char * const pEnd = pFileData + cFileSize;

	enum {modeOut, modeInAttributes, modeInSubtypes} mode;
	bm_type_t * pType = NULL;

	typedef std::map<string_t, bm_attribute_t> id2attribute_t;
	id2attribute_t id2attribute;

	CCPtr<CIcon> pDefaultIcon;

	size_t cFileLine = 1;
	const char * pLine = pFileData;
	do {
		const char * const pNextLine = FindNextLine (pLine, pEnd);

		if (pLine [0] != ';' && pLine [0] != '\n' && pLine [0] != '\r') {
			if (cFileLine == 1) {
				if (! HasPrefix (pLine, pNextLine, ".bookmark_types 001", 19))
					return false;
			} else {
				if (pLine [0] == '.') {
					const char * p = pLine + 1;

					if (HasPrefix (p, pNextLine, "type", 4)) {
						mode = modeOut;

						p += 4;
						OmitSpaces (& p, pNextLine);

						// Id.
						const char * const pId = p;
						OmitNonSpaces (& p, pNextLine);
						string_t strId (pId, p);

						OmitSpaces (& p, pNextLine);

						// Name.
						string_t strName;
						_ParseQuotedString (& p, pNextLine, strName);

						_Types.push_back (bm_type_t ());
						pType = & _Types.back ();
						pType->strId  .swap (strId);
						pType->strName.swap (strName);

						id2attribute.clear ();
					} else if (HasPrefix (p, pNextLine, "default_heading", 15)) {
						p += 7;
						OmitSpaces (& p, pNextLine);

						const char * const pVal = p;
						OmitNonSpaces (& p, pNextLine);
						string_t strVal (pVal, p);

						if (strVal == "true" && pType)
							pType->bHasHeadingByDefault = true;
					} else if (HasPrefix (p, pNextLine, "has_heading", 11)) {
						p += 11;
						OmitSpaces (& p, pNextLine);

						pType->bHasHeadingIsConst   = true;
						pType->bHasHeadingByDefault = true;
					} else if (HasPrefix (p, pNextLine, "no_heading", 10)) {
						p += 10;
						OmitSpaces (& p, pNextLine);

						pType->bHasHeadingIsConst   = true;
						pType->bHasHeadingByDefault = false;
					} else if (HasPrefix (p, pNextLine, "attributes", 10)) {
						mode = modeInAttributes;

						p += 10;
						OmitSpaces (& p, pNextLine);

						assert (id2attribute.empty ());
					} else if (HasPrefix (p, pNextLine, "subtypes", 8)) {
						mode = modeInSubtypes;

						p += 8;
						OmitSpaces (& p, pNextLine);
					} else {
						ReportWarning ("Unexpected definition in line %d.", cFileLine);
					}
				} else {
					const char * p = pLine;
					OmitSpaces (& p, pNextLine);

					if (mode == modeInAttributes) {
						// Id.
						const char * const pId = p;
						OmitNonSpaces (& p, pNextLine);
						string_t strId (pId, p);

						OmitSpaces (& p, pNextLine);

						// Kind.
						const BmAttributeKind Kind = _ParseBookmarkKind (& p, pNextLine);
						OmitSpaces (& p, pNextLine);

						// Name.
						string_t strName;
						_ParseQuotedString (& p, pNextLine, strName);

						bm_attribute_t ta;
						ta.Kind = Kind;
						ta.strId  .swap (strId);
						ta.strName.swap (strName);
						id2attribute [ta.strId] = ta;
					} else if (mode == modeInSubtypes) {
						// Id.
						const char * const pId = p;
						OmitNonSpaces (& p, pNextLine);
						string_t strId (pId, p);

						OmitSpaces (& p, pNextLine);

						// Name.
						string_t strName;
						_ParseQuotedString (& p, pNextLine, strName);

						OmitSpaces (& p, pNextLine);

						// Icon.
						string_t strIcon;
						_ParseQuotedString (& p, pNextLine, strIcon);
						CCPtr<CIcon> pIcon;
						string_t strIconFullPath = _strPath;
						strIconFullPath.append (strIcon);
						if (! DoesFileExist (strIconFullPath.c_str ())) {
							ReportWarning ("Icon file '%s' is not found.", strIconFullPath.c_str ());

							if (! pDefaultIcon) {
								string_t strDefaultIcon = _strPath;
								strDefaultIcon.append ("default.ico");
								pDefaultIcon = new CIcon (strDefaultIcon.c_str ());
							}
							pIcon = pDefaultIcon;
						} else
							pIcon = new CIcon (strIconFullPath.c_str ());

						OmitSpaces (& p, pNextLine);

						// Attributes.
						bm_attributes_t vecAttributes;
						if (* p == '{') {
							++ p;
							OmitSpaces (& p, pNextLine);
							while (* p != '}' && p < pNextLine) {
								const char * pAttributeId = p;
								OmitNonSpaces (& p, pNextLine);
								if (p [-1] == '}')
									-- p;
								bool bObligatory = false;
								bool bUseAsLabel = false;
								while (* pAttributeId == '*' || * pAttributeId == '@') {
									if (* pAttributeId == '*')
										bObligatory = true;
									if (* pAttributeId == '@')
										bUseAsLabel = true;
									++ pAttributeId;
								}
								string_t strAttributeId (pAttributeId, p);

								const id2attribute_t::iterator i = id2attribute.find (strAttributeId);
								if (i == id2attribute.end ())
									ReportWarning ("Incorrect attribute ID '%s' for type '%s'.", strAttributeId.c_str (), strId.c_str ());
								else {
									vecAttributes.push_back (i->second);
									if (bObligatory)
										vecAttributes.back ().bObligatory = true;
									if (bUseAsLabel)
										vecAttributes.back ().bUseAsLabel = true;
								}

								OmitSpaces (& p, pNextLine);
							}
						}

						pType->vecSubTypes.push_back (bm_subtype_t ());
						bm_subtype_t & st = pType->vecSubTypes.back ();
						st.strId  .swap (strId);
						st.strName.swap (strName);
//						st.strIcon.swap (strIconFullPath);
						st.pIcon = pIcon;
						st.vecAttributes.swap (vecAttributes);
					} else {
						ReportWarning ("Unexpected data in line %d.", cFileLine);
					}
				}
			}
		}

		++ cFileLine;

		pLine = pNextLine;
	} while (pLine < pEnd);

	return true;
}

static
bool _LoadBookmarkTypes (bm_types_t & _Types) {
	_Types.clear ();

	// Get path to files.
	char strPath0 [MAX_PATH];
	::GetModuleFileName (NULL, strPath0, MAX_PATH);
	char * const pFileName = ::PathFindFileName (strPath0);
	if (pFileName != NULL)
		* pFileName = '\0';
	char * const pPathEnd = strPath0 + ::strlen (strPath0);
	::strcpy (pPathEnd, "_BookmarkTypes\\");
	if (! DoesFileExist (strPath0)) {
		::strcpy (pPathEnd, "..\\_BookmarkTypes\\");
		if (! DoesFileExist (strPath0))
			return false;
	}
	char strPath [MAX_PATH];
	::PathCanonicalize (strPath, strPath0);

	bool bRet = false;

	char strFilter [MAX_PATH];
	::strcpy (strFilter, strPath);
	::strcat (strFilter, "*.dat");
	WIN32_FIND_DATA ff;
	const HANDLE hFind = ::FindFirstFile (strFilter, & ff);
	if (hFind == INVALID_HANDLE_VALUE)
		return false;
	do {
		// Get DAT file name.
		char strFileName [MAX_PATH];
		::strcpy (strFileName, strPath);
		::strcat (strFileName, ff.cFileName);

		const bool bOK = _LoadBookmarkTypesFile (strPath, strFileName, _Types);
		if (bOK)
			bRet = true;
	} while (::FindNextFile (hFind, & ff));

	return bRet;
}
