//
// GPSMapEdit
// (c) Konstantin Galichsky (kg@geopainting.com), 2009
//
// 'List of Bookmarks' dialog.
//

# include "StdAfx.h"
# include "Resource.h"
# include "BookmarksList.h"
# include "Bookmark.h"
# include "BookmarkTypes.h"
# include "Map.h"
# include "MapView.h"
# include "NewMap.h"
# include "UndoRedo.h"
# include "BackupAction.h"
# include "ShowRgnOnMap.h"
# include "Scale.h"
# include "Position.h"
# include "Selection.h"
# include "DeleteObject.h"
# include "CombinedAction.h"
# include "Dialog.h"
# include "ListView.h"
# include "ParseHelpers.h"
# include "PrintUnits.h"
# include "fast_floor.h"
# include "Globals.h"

enum {
	cbmType       = 0,
	cbmSubType    = 1,
	cbmProcessed  = 2,
	cbmPosition   = 3,
	cbmHeading    = 4,
	cbmDate       = 5,
	cbmAttrValid  = 6,
	cbmAttributes = 7,
	cbmFile       = 8,
	cbmFilePos    = 9,
};

class CBookmarks : public CDialog {
  public:
	CBookmarks () : CDialog (IDD_BOOKMARKS_LIST) {}

	static CBookmarks * g_pBookmarks;

	void Reset ();

	// Override.
	virtual bool OnInit (HWND _hwndFocus);
	virtual void OnShowModeless () {Reset ();}
	virtual void OnDestroy ();
	virtual void OnCommand (WORD _wCmd, WORD _wNotifyCode, HWND _hwndCtl);
	virtual bool OnNotify (int _idCtrl, NMHDR * _pInfo);

  private:
	CListView m_List;
	void _FillBookmarks ();
	rgn_t * _GetBookmark (size_t _cItem = -1);
	void _UpdateButtons  (size_t _cItem = -1);
	void _UpdateTotal ();

	void _GetSubItemText (const bookmark_t & _bm, size_t _cCol, char * _strBuf, size_t _cBufSize) const;

	typedef std::vector<rgn_t *> vecRgns_t;
	vecRgns_t m_vecRgns;

	struct LessBookmarks {
		LessBookmarks (const CBookmarks * _pBookmarks, bool _bAscending, size_t _cCol) :
			m_pBookmarks (_pBookmarks),
			m_bAscending (_bAscending),
			m_cCol (_cCol)
		{}

		bool operator () (const rgn_t * _pRgn1, const rgn_t * _pRgn2);

	  private:
		const CBookmarks * const m_pBookmarks;
		const bool m_bAscending;
		const size_t m_cCol;
	};
	friend struct LessBookmarks;

	TIME_ZONE_INFORMATION m_tzi;
};

CBookmarks * CBookmarks::g_pBookmarks;

bool CBookmarks::LessBookmarks::operator () (const rgn_t * _pRgn1, const rgn_t * _pRgn2) {
	assert (_pRgn1 && _pRgn1->IsBookmark ());
	assert (_pRgn2 && _pRgn2->IsBookmark ());
	const bookmark_t & bm1 = * _pRgn1->pExtra->pBookmark;
	const bookmark_t & bm2 = * _pRgn2->pExtra->pBookmark;

	float fCmp = 0;
	switch (m_cCol) {
		case cbmType:      fCmp = ::lstrcmp (bm1.strType   .c_str (), bm2.strType   .c_str ()); break;
		case cbmSubType:   fCmp = ::lstrcmp (bm1.strSubType.c_str (), bm2.strSubType.c_str ()); break;
		case cbmProcessed: fCmp = int (bm1.bProcessed) - int (bm2.bProcessed); break;
		case cbmPosition:  fCmp = bm1.ptPosFixed.y == bm2.ptPosFixed.y ? bm1.ptPosFixed.x - bm2.ptPosFixed.x : bm1.ptPosFixed.y - bm2.ptPosFixed.y; break;
		case cbmHeading:   fCmp = (bm1.bHasHeading ? bm1.fHeading : 0) - (bm2.bHasHeading ? bm2.fHeading : 0); break;
		case cbmDate:      fCmp = bm1.dt - bm2.dt; break;
		case cbmFile:      fCmp = ::lstrcmp (bm1.strFileName.c_str (), bm2.strFileName.c_str ()); break;
		case cbmFilePos:   fCmp = (bm1.strFileName.empty () ? 0 : bm1.dwFilePos) - (bm2.strFileName.empty () ? 0 : bm2.dwFilePos); break;
		case cbmAttrValid: fCmp = int (bm1.AreAllAttributesValid ()) - int (bm2.AreAllAttributesValid ());

		case cbmAttributes:
		default: {
			char strText1 [64], strText2 [64];
			m_pBookmarks->_GetSubItemText (bm1, m_cCol, strText1, ARRAY_SIZE (strText1));
			m_pBookmarks->_GetSubItemText (bm2, m_cCol, strText2, ARRAY_SIZE (strText2));
			fCmp = ::lstrcmp (strText1, strText2);
		}
	}

	return m_bAscending ? fCmp < 0 : fCmp > 0;
}

void CBookmarks::Reset () {
	_FillBookmarks ();
	_UpdateButtons ();
}

void CBookmarks::_FillBookmarks () {
	CCritSec::CLock lock (g_cs);

	CWaiter w;

	rgn_t * const pRgnSel = _GetBookmark ();

	// Save selection status for each item.
	typedef std::set<rgn_t *> setRgns_t;
	setRgns_t setRgnsSelected;
	const size_t cItemsOld = ListView_GetItemCount (m_List);
	size_t cItem;
	for (cItem = 0; cItem < cItemsOld; ++ cItem) {
		if (! IsItemSelected (m_List, cItem))
			continue;

		rgn_t * const pRgn = _GetBookmark (cItem);
		if (! pRgn || pRgn->IsImmutable ())
			continue;

		setRgnsSelected.insert (pRgn);
	}

	// Update bookmarks list.
	m_vecRgns.clear ();

	layer_t & layer = g_map.layers [rkPoint];
	rgns_t & rgns = layer.rgns;
	rgns_t::iterator iRgn;
	for (iRgn = rgns.begin (); iRgn != rgns.end (); ++ iRgn) {
		rgn_t & rgn = * iRgn;
		if (rgn.IsImmutable () || ! rgn.IsBookmark ())
			continue;

		m_vecRgns.push_back (& rgn);
	}

	// Sort items.
	if (m_List.GetSortingColumn () != -1)
		std::sort (m_vecRgns.begin (), m_vecRgns.end (), LessBookmarks (this, m_List.IsSortingAscending (), m_List.GetSortingColumn ()));

	// Fill the list view.
	const size_t cRgns = m_vecRgns.size ();
	ListView_SetItemCount (m_List, cRgns);

	// Restore selection.
	for (cItem = 0; cItem < cRgns; ++ cItem) {
		rgn_t * const pRgn = m_vecRgns [cItem];
		SelectItem (m_List, cItem, setRgnsSelected.find (pRgn) != setRgnsSelected.end ());

		if (pRgn == pRgnSel) {
			SetSelection           (m_List, cItem);
			ListView_EnsureVisible (m_List, cItem, FALSE);
		}
	}

	_UpdateTotal ();
}

rgn_t * CBookmarks::_GetBookmark (size_t _cItem) {
	const size_t cSel = _cItem == -1 ? ListView_GetSelectionMark (m_List) : _cItem;
	if (cSel == -1)
		return NULL;

	assert (cSel < m_vecRgns.size ());
	if (cSel >= m_vecRgns.size ())
		return NULL;
	rgn_t * const pRgn = m_vecRgns [cSel];
	assert (pRgn);
	if (pRgn == NULL || pRgn->IsDeleted ())
		return NULL;

	return pRgn;
}

void CBookmarks::_UpdateButtons (size_t _cItem) {
	const size_t cSelItems = ListView_GetSelectedCount (m_List);
	EnableDlgItem (IDC_LOCATE, cSelItems == 1);
	EnableDlgItem (IDC_DELETE, cSelItems != 0);
	EnableDlgItem (IDC_EDIT,   cSelItems == 1);
}

void CBookmarks::_UpdateTotal () {
	const size_t cRgns = m_vecRgns.size ();
	SetDlgItemTextF (IDC_TOTAL, IDS_TOTAL, cRgns);
}

bool CBookmarks::OnInit (HWND _hwndFocus) {
	// Get time zone bias + daylight saving time bias.
	if (::GetTimeZoneInformation (& m_tzi) == TIME_ZONE_ID_DAYLIGHT)
		m_tzi.Bias += m_tzi.DaylightBias;

	// Bookmarks.
	m_List.Init (m_hWnd, IDC_LIST, "Bookmarks", -1);
	m_List.SetExtStyle (LVS_EX_FULLROWSELECT | LVS_EX_HEADERDRAGDROP);

	AddColumn (m_List, cbmType,       "Type");
	AddColumn (m_List, cbmSubType,    "Subtype");
	AddColumn (m_List, cbmProcessed,  "Processed");
	AddColumn (m_List, cbmPosition,   "Position");
	AddColumn (m_List, cbmHeading,    "Heading",     true);
	AddColumn (m_List, cbmDate,       "Date");
	AddColumn (m_List, cbmAttrValid,  "Valid attr.");
	AddColumn (m_List, cbmAttributes, "Attributes");
	AddColumn (m_List, cbmFile,       "File base");
	AddColumn (m_List, cbmFilePos,    "Pos in file", true);

	m_List.Sort ();

	_FillBookmarks ();
	SetSelection (m_List, 0);

	AutoSizeColumns (m_List);
	m_List.LoadHeaders ();

	// Buttons.
	_UpdateButtons ();

	return true;
}

void CBookmarks::OnDestroy () {
	m_List.SaveHeaders ();
	m_List.Destroy ();
}

void CBookmarks::OnCommand (WORD _wCmd, WORD _wNotifyCode, HWND _hwndCtl) {
	switch (_wCmd) {
		case IDC_EDIT: {
			const size_t cSelItems = ListView_GetSelectedCount (m_List);
			if (cSelItems <= 1) {
				rgn_t * const pRgn = _GetBookmark ();
				if (pRgn) {
					if (ShowBookmarkProperties (m_hWnd, * pRgn))
						_FillBookmarks ();
				}
			}
			break;
		}

		case IDC_DELETE: {
			const size_t cSel = ListView_GetSelectionMark (m_List);

			const size_t cSelItems = ListView_GetSelectedCount (m_List);
			if (cSelItems <= 1) {
				rgn_t * const pRgn = _GetBookmark ();
				if (pRgn)
					DeleteObject (pRgn);
			} else {
				const size_t cItems = ListView_GetItemCount (m_List);
				for (size_t cItem = 0; cItem < cItems; ++ cItem) {
					if (! IsItemSelected (m_List, cItem))
						continue;

					rgn_t * const pRgn = _GetBookmark (cItem);
					if (! pRgn)
						continue;

					DeleteObject (pRgn);
				}
			}

			_FillBookmarks ();
			if (cSel != -1)
				SetSelection (m_List, cSel);

			_UpdateButtons ();

			Refresh ();

			return;
		}

		case IDC_LOCATE: {
			const size_t cSelItems = ListView_GetSelectedCount (m_List);
			if (cSelItems != 1)
				return;

			const rgn_t * const pRgn = _GetBookmark ();
			if (pRgn)
				ShowRgnOnMap (* pRgn);

			return;
		}

		case IDC_REFRESH:
			Reset ();
			return;

		case IDCANCEL:
			Hide ();
			return;
	}
}

bool CBookmarks::OnNotify (int _idCtrl, NMHDR * _pInfo) {
	switch (_pInfo->code) {
		case NM_RCLICK:
			if (_idCtrl != IDC_LIST)
				break;
			m_List.PopupMenu_OnHeaderRClick (_pInfo);
			break;

		case NM_DBLCLK:
			if (_idCtrl != IDC_LIST)
				break;

			SendCommand (IDC_EDIT);
			break;

		case LVN_COLUMNCLICK:
			if (_idCtrl != IDC_LIST)
				break;

			m_List.Sort_OnColumnClick (_pInfo);
			_FillBookmarks ();

			break;

		case LVN_ITEMCHANGED: {
			if (_idCtrl != IDC_LIST)
				break;

			const NMLISTVIEW * const pInfo2 = reinterpret_cast<const NMLISTVIEW *> (_pInfo);
			if (pInfo2->uNewState & LVIS_SELECTED)
				_UpdateButtons (pInfo2->iItem);

			break;
		}

		case LVN_GETDISPINFO: {
			if (_idCtrl != IDC_LIST)
				break;

			const NMLVDISPINFO * const pInfo2 = reinterpret_cast<const NMLVDISPINFO *> (_pInfo);
			assert ((size_t) pInfo2->item.iItem < m_vecRgns.size ());
			if ((size_t) pInfo2->item.iItem >= m_vecRgns.size ())
				break;
			const rgn_t * const pRgn = m_vecRgns [pInfo2->item.iItem];
			if (pRgn == NULL)
				break;
			if (! pRgn->IsBookmark ())
				break;
			const bookmark_t & bm = * pRgn->pExtra->pBookmark;

			if (pInfo2->item.mask & LVIF_TEXT)
				_GetSubItemText (bm, pInfo2->item.iSubItem, pInfo2->item.pszText, pInfo2->item.cchTextMax);

			break;
		}
	}

	return false;
}

void CBookmarks::_GetSubItemText (const bookmark_t & _bm, size_t _cCol, char * _strBuf, size_t _cBufSize) const {
	switch (_cCol) {
		default:
			assert (false);
			_strBuf [0] = '\0';
			break;

		case cbmType: {
			string_t strType, strSubType;
			GetBookmarkDisplayType (_bm, strType, strSubType);
			::_tcsncpy  (_strBuf, strType.c_str (), _cBufSize);
			break;
		}
		case cbmSubType: {
			string_t strType, strSubType;
			GetBookmarkDisplayType (_bm, strType, strSubType);
			::_tcsncpy  (_strBuf, strSubType.c_str (), _cBufSize);
			break;
		}

		case cbmProcessed:
			::_tcsncpy  (_strBuf, _bm.bProcessed ? "yes" : "no", _cBufSize);
			break;

		case cbmPosition: {
			char strPos [32];
			PrintPosition (strPos, _bm.ptPosFixed);
			::_tcsncpy  (_strBuf, strPos, _cBufSize);
			break;
		}

		case cbmHeading: {
			if (_bm.bHasHeading) {
				char strHeading [16];
				PrintAzimuth (strHeading, fast_floor (_bm.fHeading + .5f));
				::_tcsncpy  (_strBuf, strHeading, _cBufSize);
			} else
				::_tcsncpy  (_strBuf, "-", _cBufSize);
			break;
		}

		case cbmDate: {
			if (_bm.dt == 0)
				::_tcsncpy  (_strBuf, "-", _cBufSize);
			else {
				// Convert to local time.
				const DATE dtLocalTime = _bm.dt - float (m_tzi.Bias)/(24*60);
				SYSTEMTIME st;
				::VariantTimeToSystemTime (dtLocalTime, & st);

				char strDate [32];
				if (! ::GetDateFormat (LOCALE_USER_DEFAULT, DATE_SHORTDATE, & st, NULL, strDate, sizeof (strDate)/sizeof (strDate [0])))
					strDate [0] = '\0';

				::_tcsncpy  (_strBuf, strDate, _cBufSize);
			}
			break;
		}

		case cbmAttrValid: {
			const bool bAttrValid = _bm.AreAllAttributesValid ();
			::_tcsncpy  (_strBuf, bAttrValid ? "yes" : "no", _cBufSize);
			break;
		}
		case cbmAttributes: {
			string_t strAttributes;
			for (bookmark_t::k2v_t::const_iterator i = _bm.attributes.begin (); i != _bm.attributes.end (); ++ i) {
				const string_t & strKey   = i->first;
				const string_t & strValue = i->second;
				if (strValue.empty ())
					continue;

				if (! strAttributes.empty ())
					strAttributes.append ("; ");
	//			strAttributes.append (strKey);
	//			strAttributes.append ("=");
				strAttributes.append (strValue);
			}
			::_tcsncpy  (_strBuf, strAttributes.c_str (), _cBufSize);
			break;
		}

		case cbmFile: {
			if (_bm.strFileName.empty ())
				::_tcsncpy  (_strBuf, "-", _cBufSize);
			else
				::_tcsncpy  (_strBuf, _bm.strFileName.c_str (), _cBufSize);
			break;
		}

		case cbmFilePos: {
			if (_bm.strFileName.empty ())
				::_tcsncpy  (_strBuf, "-", _cBufSize);
			else
				::sprintf  (_strBuf, "%d", _bm.dwFilePos);
			break;
		}
	}
}

static
void _ResetBookmarks () {
	if (CBookmarks::g_pBookmarks)
		CBookmarks::g_pBookmarks->Reset ();
}

INVOKE_AFTER_REMOVE_MAP (_ResetBookmarks);
INVOKE_ON_BACKUP_MAP    (_ResetBookmarks);
INVOKE_BEFORE_GC        (_ResetBookmarks);

void ShowBookmarks () {
	ShowModeless (CBookmarks::g_pBookmarks, g_hWnd);
}