//
// GPSMapEdit
// (c) Konstantin Galichsky (kg@geopainting.com), 2008-2009
//
// COM object model: the statistics info.
//

# include "StdAfx.h"
# include "COMStatistics.h"
# include "COMBase.h"
# include "Map.h"
# include "Types.h"

struct ut_data2_t {
	const type_desc_t * pType;
	RGN_KIND            Kind; // NOTE: this field is required for unknown types.
	string_t            strTypeName;

	size_t cCount;
	size_t cCountWithAddress;
	size_t cCountWithContact;
	double fLengthInKm;
	double fAreaInKm2;
	size_t cOccupiedSizeInB;

	ut_data2_t () :
		pType (NULL), Kind (rkMax),
		cCount (0), cCountWithAddress (0), cCountWithContact (0),
		fLengthInKm (0), fAreaInKm2 (0),
		cOccupiedSizeInB (0)
	{}
};

static
void operator += (ut_data2_t & _1, const ut_data2_t & _2) {
	_1.cCount            += _2.cCount;
	_1.cCountWithAddress += _2.cCountWithAddress;
	_1.cCountWithContact += _2.cCountWithContact;
	_1.fLengthInKm       += _2.fLengthInKm;
	_1.fAreaInKm2        += _2.fAreaInKm2;
	_1.cOccupiedSizeInB  += _2.cOccupiedSizeInB;
}

static
void _CountRgn (ut_data2_t & _ts, const rgn_t & _rgn, float _fLengthInKm, float _fAreaInKm2) {
	++ _ts.cCount;

	if (_rgn.pExtra) {
		if (! _rgn.pExtra->strNumber.empty () && ! _rgn.pExtra->strStreetDesc.empty () && _rgn.pExtra->pCity != NULL && ! _rgn.pExtra->pCity->IsDeleted ())
			++ _ts.cCountWithAddress;

		if (
			! _rgn.pExtra->strPhone      .empty () ||
			! _rgn.pExtra->strFax        .empty () ||
			! _rgn.pExtra->strEmail      .empty () ||
			! _rgn.pExtra->strWebPage    .empty () ||
			! _rgn.pExtra->strDescription.empty ()
		)
			++ _ts.cCountWithContact;
	}

	_ts.fLengthInKm += _fLengthInKm;
	_ts.fAreaInKm2  += _fAreaInKm2;

	if (_rgn.cEndPos > _rgn.cBeginPos)
		// NOTE: _rgn.cEndPos may be incorrect for damaged files (w/o correct end).
		_ts.cOccupiedSizeInB += _rgn.cEndPos - _rgn.cBeginPos;
}

/////////////////////////////////////////////////////////////////

class CComStatistics : public CComDualInterface<IStatistics, & LIBID_GPSMapEdit> {
  public:
	// IStatistics.
	STDMETHOD (get_CountOfTypes) (long * _ret);
	STDMETHOD (GetStatForType) (long _lIndex, StatPropertyForType _sp, VARIANT * _ret);
	STDMETHOD (GetStatForMap) (StatPropertyForMap _sp, double * _ret);

	//
	// The statistics data to report.
	//

	typedef std::vector<ut_data2_t> type_stats_t;
	type_stats_t m_type_stats;

	size_t cObjects;
	size_t cPoints;
	size_t cPolylines;
	size_t cPolygons;
	size_t cMaxNumberOfElementsPerObject;
	size_t cMaxNumberOfVerticesPerElement;
	size_t cVertices;
	size_t cRoutingNodes;
	size_t cRoutingLinks;
	size_t cExternalNodes;
	size_t cTurnRestrictions;
	float fPolylinesLengthInKm;
	float fRoadsLengthInKm;
	float fPolygonsAreaInKm2;
};

STDMETHODIMP CComStatistics::get_CountOfTypes (long * _ret) {
	* _ret = m_type_stats.size ();
	return S_OK;
}

STDMETHODIMP CComStatistics::GetStatForType (long _lIndex, StatPropertyForType _sp, VARIANT * _ret) {
	::VariantClear (_ret);

	if (_lIndex < 0 || _lIndex >= (int) m_type_stats.size ())
		return E_INVALIDARG;

	CComVariant vRet;
	USES_CONVERSION;

	const ut_data2_t & ts = m_type_stats [_lIndex];
	switch (_sp) {
		case spKind:              if (ts.pType) vRet = (long) ts.Kind;           break;
		case spType:              if (ts.pType) vRet = (long) ts.pType->Type;    break;
		case spTypeNameEnglish:   vRet = ts.pType ? A2OLE (ts.pType->strTypeName) : A2OLE (ts.strTypeName.c_str ()); break;
		case spTypeNameLocalized: if (ts.pType) vRet = A2OLE (ts.pType->strTypeNameLocalized); break;

		case spCount:             vRet = (long) ts.cCount;            break;
		case spCountWithAddress:  vRet = (long) ts.cCountWithAddress; break;
		case spCountWithContact:  vRet = (long) ts.cCountWithContact; break;
		case spLengthInKm:        vRet = ts.fLengthInKm;              break;
		case spAreaInKm2:         vRet = ts.fAreaInKm2;               break;
		case spOccupiedSizeInB:   vRet = (long) ts.cOccupiedSizeInB;  break;

		default:
			return E_INVALIDARG;
	}

	vRet.Detach (_ret);
	return S_OK;
}

STDMETHODIMP CComStatistics::GetStatForMap (StatPropertyForMap _sp, double * _ret) {
	switch (_sp) {
		case spObjects:                       * _ret = cObjects;                       break;
		case spPoints:                        * _ret = cPoints;                        break;
		case spPolylines:                     * _ret = cPolylines;                     break;
		case spPolygons:                      * _ret = cPolygons;                      break;
		case spMaxNumberOfElementsPerObject:  * _ret = cMaxNumberOfElementsPerObject;  break;
		case spMaxNumberOfVerticesPerElement: * _ret = cMaxNumberOfVerticesPerElement; break;
		case spVertices:                      * _ret = cVertices;                      break;
		case spRoutingNodes:                  * _ret = cRoutingNodes;                  break;
		case spRoutingLinks:                  * _ret = cRoutingLinks;                  break;
		case spExternalNodes:                 * _ret = cExternalNodes;                 break;
		case spTurnRestrictions:              * _ret = cTurnRestrictions;              break;
		case spPolylinesLengthInKm:           * _ret = fPolylinesLengthInKm;           break;
		case spRoadsLengthInKm:               * _ret = fRoadsLengthInKm;               break;
		case spPolygonsAreaInKm2:             * _ret = fPolygonsAreaInKm2;             break;

		default:
			* _ret = 0;
			return E_INVALIDARG;
	}

	return S_OK;
}

HRESULT GetStatistics (long _lZoomLevel, IStatistics ** _pRet) {
	CCritSec::CLock lock (g_cs);

	const size_t cTypes = g_map.pTypeSet->Types ().size ();
	const type_desc_t * const pUnknownTD = g_map.pTypeSet->GetUnknownTD ();

	size_t cPoints = 0, cPolylines = 0, cPolygons = 0;
	size_t cMaxNodes = 0, cMaxElements = 0, cTotalNodes = 0;
	size_t cRGLinks = 0;

	size_t cLabels   = 0, cLabelsSize   = 0;
	size_t cComments = 0, cCommentsSize = 0;
	size_t cRgnsSize = 0;
	size_t cElementsTotal = 0, cElementsSize = 0;
	size_t cOriginsSize   = 0;
	size_t cPointsSize    = 0;
	size_t cPointsSize2   = 0;
	
	// The array of counters being parallel to types list.
	ut_data2_t * const pTypeStats = reinterpret_cast<ut_data2_t *> (alloca (sizeof (ut_data2_t)*cTypes));
	::memset (pTypeStats, 0, sizeof (ut_data2_t)*cTypes);

	typedef std::map<DWORD, ut_data2_t> mapTypes_t;
	mapTypes_t mapUnknownTypes;
	mapTypes_t mapSpecificTypes;

	size_t cRGNodes          = 0;
	size_t cExternalNodes    = 0;
	size_t cTurnRestrictions = 0;

	float fPolylinesLengthInKm = 0;
	float fRoadsLengthInKm     = 0;
	float fPolygonsAreaInKm2   = 0;

	ut_data2_t utdRoads;
	ut_data2_t utdCities;
	ut_data2_t utdWater;

	//
	// Count statistics.
	//

	{
		// Count types.
		for (size_t cKind = 0; cKind < rkMax; ++ cKind) {
			const rgns_t & rgns = g_map.layers [cKind].rgns;
			for (rgns_t::const_iterator iRgn = rgns.begin (); iRgn != rgns.end (); ++ iRgn) {
				const rgn_t & rgn = * iRgn;
				if (rgn.IsDeleted () || rgn.IsAttachment () || rgn.IsNotMapObject ())
					continue;

				bool bHasElements = false;

				double fLength = 0;
				double fArea   = 0;

				// Elements.
				const size_t cElements = rgn.elements.size ();
				for (size_t cElement = 0; cElement < cElements; ++ cElement) {
					const rgn_element_t & element = rgn.elements [cElement];
					if (element.bDeleted)
						continue;

					if (
						_lZoomLevel == -1 ||
						element.cLevel == _lZoomLevel ||
						(rgn.uchEndLevel != 0 && element.cLevel <= _lZoomLevel && _lZoomLevel <= rgn.uchEndLevel)
					) {
						bHasElements = true;

						++ cElementsTotal;
						cElementsSize += sizeof (element);
						cPointsSize  += element.points.capacity ()*sizeof (point_t);
						cPointsSize2 += element.points.size ()*sizeof (point_t);

						if (cKind == rkPolyline)
							fLength += element.fLengthOrArea;
						else if (cKind == rkPolygon) {
							if (element.bHole)
								fArea -= element.fLengthOrArea;
							else
								fArea += element.fLengthOrArea;
						}

						const size_t cNodes = element.points.size ();
						cTotalNodes += cNodes;

						if (cNodes > cMaxNodes)
							cMaxNodes = cNodes;

						if (element.pExtra) {
							const point_extras_t & PointExtras = element.pExtra->PointExtras;
							const size_t cNodes2 = PointExtras.size ();
							size_t cRGNodesInElement = 0;
							for (size_t cNode = 0; cNode < cNodes2 && cNode < cNodes; ++ cNode) {
								const rgnode_t * const pNode = PointExtras [cNode].GetRGNode ();
								if (pNode != NULL)
									++ cRGNodesInElement;
							}
							if (cRGNodesInElement != 0)
								cRGLinks += cRGNodesInElement - 1;
						}
					}
				}

				// Origins.
				const size_t cOrigins = rgn.origins.size ();
				for (size_t cOrigin = 0; cOrigin < cOrigins; ++ cOrigin) {
					const rgn_origin_t & origin = rgn.origins [cOrigin];
					if (origin.bDeleted)
						continue;

					if (
						_lZoomLevel == -1 ||
						origin.cLevel == _lZoomLevel ||
						(rgn.uchEndLevel != 0 && origin.cLevel <= _lZoomLevel && _lZoomLevel <= rgn.uchEndLevel)
					) {
						bHasElements = true;

						++ cTotalNodes;

						cOriginsSize += sizeof (rgn_origin_t);
					}
				}

				if (! bHasElements)
					continue;

				if (cElements > cMaxElements)
					cMaxElements = cElements;

				// Count labels.
				if (! rgn.strLabel.empty ()) {
					++ cLabels;
					cLabelsSize += rgn.strLabel.size ();
				}

				// Count comments.
				if (! rgn.strComment.empty ()) {
					++ cComments;
					cCommentsSize += rgn.strComment.size ();
				}

				cRgnsSize += sizeof (rgn);

				if (rgn.IsPolyline ())
					++ cPolylines;
				else if (rgn.IsPolygon ())
					++ cPolygons;
				else
					++ cPoints;

				fLength *= 40e3/360;
				fArea   *= 40e3*40e3/(360*360)*g_map.fCosY;

				fPolylinesLengthInKm += fLength;
				fPolygonsAreaInKm2   += fArea;

				if (rgn.IsRoad ())
					fRoadsLengthInKm += fLength;

				// Count types.
				const type_desc_t * const pTD = rgn.GetTypeDesc ();
				assert (pTD);
				if (pTD && ! pTD->bUnknown) {
					if (pTD->Type == rgn.GetType ()) {
						_CountRgn (pTypeStats [pTD->cOrdinal], rgn, fLength, fArea);
					} else {
						const DWORD dwType = rgn.GetType () + (DWORD (rgn.GetKind ()) << 24);
						mapTypes_t::iterator i = mapSpecificTypes.find (dwType);
						if (i == mapSpecificTypes.end ())
							i = mapSpecificTypes.insert (mapTypes_t::value_type (dwType, ut_data2_t ())).first;
						_CountRgn (i->second, rgn, fLength, fArea);
					}
				} else {
					const DWORD dwType = rgn.GetType () + (DWORD (rgn.GetKind ()) << 24);
					mapTypes_t::iterator i = mapUnknownTypes.find (dwType);
					if (i == mapUnknownTypes.end ())
						i = mapUnknownTypes.insert (mapTypes_t::value_type (dwType, ut_data2_t ())).first;
					_CountRgn (i->second, rgn, fLength, fArea);
				}

				if (rgn.IsRoad ())
					_CountRgn (utdRoads, rgn, fLength, fArea);

				if (pTD && pTD->Category == catCities)
					_CountRgn (utdCities, rgn, fLength, fArea);

				if (pTD && pTD->Category == catWater)
					_CountRgn (utdWater, rgn, fLength, fArea);
			}
		}

		// Count RG nodes.
		g_map.CountRGNodes ();

		for (rgnodes_t::const_iterator in = g_map.rgnodes.begin (); in != g_map.rgnodes.end (); ++ in) {
			const rgnode_t & node = * in;
			if (! node.IsUsed ())
				continue;

			++ cRGNodes;

			if (node.bIsExternal)
				++ cExternalNodes;
		}

		// Count turn restictions.
		for (rgrestrictions_t::const_iterator ir = g_map.rgrestrictions.begin (); ir != g_map.rgrestrictions.end (); ++ ir) {
			const rgrestriction_t & restriction = ir->second;
			if (restriction.IsUsed ())
				++ cTurnRestrictions;
		}
	}

	ut_data2_t utdKind [rkMax];

	//
	// Output results.
	//

	CComStatistics * const pStatistics = new CComStatistics;

	// For each type.
	// NOTE: omit 0-th type ("unknown").
	size_t cType = 1;
	CTypeSet::types_t::const_iterator it = g_map.pTypeSet->Types ().begin ();
	for (++ it; it != g_map.pTypeSet->Types ().end () && cType < cTypes; ++ it, ++ cType) {
		if (pTypeStats [cType].cCount == 0)
			continue;
		const type_desc_t & td = * it;

		utdKind [td.Kind] += pTypeStats [cType];

		pStatistics->m_type_stats.push_back (ut_data2_t ());
		ut_data2_t & ts = pStatistics->m_type_stats.back ();
		ts       = pTypeStats [cType];
		ts.pType = & td;
		ts.Kind  = static_cast<RGN_KIND> (td.Kind);
	}

	// Specific types.
	mapTypes_t::iterator i;
	for (i = mapSpecificTypes.begin (); i != mapSpecificTypes.end (); ++ i) {
		const DWORD dwType = i->first & 0xFFFFFF;
		const BYTE Kind    = i->first >> 24;
		assert (Kind < rkMax);

		const type_desc_t * pTD = g_map.pTypeSet->GetTypeDesc (dwType, Kind);
		if (pTD == NULL)
			pTD = pUnknownTD;

		utdKind [Kind] += i->second;

		pStatistics->m_type_stats.push_back (ut_data2_t ());
		ut_data2_t & ts = pStatistics->m_type_stats.back ();
		ts       = i->second;
		ts.pType = pTD;
		ts.Kind  = static_cast<RGN_KIND> (Kind);
	}

	// Unknown types.
	for (i = mapUnknownTypes.begin (); i != mapUnknownTypes.end (); ++ i) {
		const DWORD dwType = i->first & 0xFFFFFF;
		const BYTE Kind    = i->first >> 24;
		assert (Kind < rkMax);

		utdKind [Kind] += i->second;

		pStatistics->m_type_stats.push_back (ut_data2_t ());
		ut_data2_t & ts = pStatistics->m_type_stats.back ();
		ts       = i->second;
		ts.pType = pUnknownTD;
		ts.Kind  = static_cast<RGN_KIND> (Kind);
	}

	// Totals.
	ut_data2_t utdAll;
	utdAll += utdKind [rkPoint];
	utdAll += utdKind [rkPolyline];
	utdAll += utdKind [rkPolygon];

	utdAll              .strTypeName = "Total";
	utdKind [rkPoint]   .strTypeName = "All points";
	utdKind [rkPolyline].strTypeName = "All polylines";
	utdKind [rkPolygon] .strTypeName = "All polygons";
	utdRoads            .strTypeName = "All polyline roads";
	utdCities           .strTypeName = "All cities";
	utdWater            .strTypeName = "All hydrographic";

	pStatistics->m_type_stats.insert (pStatistics->m_type_stats.begin (), 7, ut_data2_t ());
	pStatistics->m_type_stats [0] = utdAll;
	pStatistics->m_type_stats [1] = utdKind [rkPoint];
	pStatistics->m_type_stats [2] = utdKind [rkPolyline];
	pStatistics->m_type_stats [3] = utdKind [rkPolygon];
	pStatistics->m_type_stats [4] = utdRoads;
	pStatistics->m_type_stats [5] = utdCities;
	pStatistics->m_type_stats [6] = utdWater;

	pStatistics->cPoints    = cPoints;
	pStatistics->cPolygons  = cPolygons;
	pStatistics->cPolylines = cPolylines;
	pStatistics->cObjects   = cPoints + cPolygons + cPolylines;

	pStatistics->cMaxNumberOfVerticesPerElement = cMaxNodes;
	pStatistics->cMaxNumberOfElementsPerObject  = cMaxElements;

	pStatistics->cRoutingNodes     = cRGNodes;
	pStatistics->cRoutingLinks     = cRGLinks;
	pStatistics->cExternalNodes    = cExternalNodes;
	pStatistics->cTurnRestrictions = cTurnRestrictions;

	pStatistics->fPolylinesLengthInKm = fPolylinesLengthInKm;
	pStatistics->fRoadsLengthInKm     = fRoadsLengthInKm;
	pStatistics->fPolygonsAreaInKm2   = fPolygonsAreaInKm2;

	return pStatistics->QueryInterface (__uuidof (IStatistics), reinterpret_cast<void **> (_pRet));
}
