//
// GPSMapEdit
// (c) Konstantin Galichsky (kg@geopainting.com), 2007-2009
// (c) Nikita Karnauhov (jinx@ac-sw.com), 2007
//
// Clean up external nodes.
//

# include "StdAfx.h"
# include "CleanUpExternalNodes.h"
# include "SetRGNodeRestriction.h"
# include "SetRGNodeExternal.h"
# include "CombinedAction.h"
# include "Types.h"
# include "UndoRedo.h"
# include "MapView.h"

void _MakeUTurnRestriction (map_t & _map, rgn_t * _pRgn, rgnode_t * _pRGNode, bool _bFirstPoint) {
	// Check existing u-turn restriction to avoid duplication.
	typedef rgrestrictions_t::const_iterator I;
	const std::pair<I, I> bounds = _map.rgrestrictions.equal_range (_pRGNode);
	for (I iRestriction = bounds.first; iRestriction != bounds.second; ++ iRestriction) {
		const rgrestriction_t & restr = iRestriction->second;

		if (restr.pFromRoad == _pRgn && restr.pFromRoad == restr.pToRoad)
			return;
	}

	rgrestriction_t restriction;
	restriction.pFromRoad      = _pRgn;
	restriction.pToRoad        = _pRgn;
	restriction.pNode          = _pRGNode;
	restriction.bFromDirection = ! _bFirstPoint ? 1 : 0;
	restriction.bToDirection   = _bFirstPoint   ? 1 : 0;

	CAddRGNodeRestrictionAction * const pAddRGNodeAction = new CAddRGNodeRestrictionAction (restriction);
	PerformEditAction (pAddRGNodeAction);
}

void _CleanUpExternalNodes (map_t & _map, size_t _cMinNodeCount) {
	CCritSec::CLock lock (g_cs);

	const int nDepth = BeginCombinedAction ();

	// Ensure that node counters are correct.
	_map.CountRGNodes (true);

	const size_t cLevels = _map.Levels.size ();

	rgns_t & rgns = _map.layers [rkPolyline].rgns;
	for (rgns_t::iterator iRgn = rgns.begin (); iRgn != rgns.end (); ++ iRgn) {
		rgn_t & rgn = * iRgn;
		if (rgn.IsDeleted () || rgn.IsNotMapObject ())
			continue;

		const bool bRPolyline = rgn.HasRGNodes () && g_map.pTypeSet->pfIsRoad (rgn);
		if (! bRPolyline)
			continue;
		if ((rgn.dwFlags & rgn_t::fDirIndicator) != 0)
			continue;
		if (rgn.pExtra && rgn.pExtra->bOneWay)
			continue;

		rgn_elements_t & elements = rgn.elements;
		const size_t cElements = elements.size ();
		for (size_t cElement = 0; cElement < cElements; ++ cElement) {
			rgn_element_t & element = elements [cElement];
			if (element.bDeleted)
				continue;
			if (element.pExtra == NULL)
				continue;

			point_extras_t & PointExtras = element.pExtra->PointExtras;
			const size_t cPoints = PointExtras.size ();
			const points_t & points = element.points;
			for (size_t cPoint = 0; cPoint < cPoints; ++ cPoint) {
				rgnode_t * const pRGNode = PointExtras [cPoint].GetRGNode ();
				if (pRGNode == NULL)
					continue;

				if (pRGNode->bIsExternal && pRGNode->btPolylines > _cMinNodeCount)
					_MakeUTurnRestriction (_map, & rgn, pRGNode, cPoint == 0);
			}
		}
	}

	for (rgnodes_t::iterator iNode = _map.rgnodes.begin (); iNode != _map.rgnodes.end (); ++ iNode) {
		rgnode_t & node = * iNode;
		if (! node.IsUsed ())
			continue;

		if (node.bIsExternal && node.btPolylines > _cMinNodeCount)
			SetRGNodeExternal (node, false);
	}

//	_map.CountRGNodes (true);

	EndCombinedAction (nDepth, "clean up external nodes");

	Refresh ();
}

/*
class CCleanUpExternalNodesAction : public CEditAction {
  public:
	CCleanUpExternalNodesAction (map_t & _map, size_t _cMinNodeCount) :
		m_map (_map),
		m_cMinNodeCount (_cMinNodeCount)
	{}

	virtual void GetName (string_t & _str) const {_str.append ("clean up external nodes");}
	virtual cu_t CanUndo () const {return cuYes;}
	virtual void Undo ();
	virtual void Redo ();

	virtual bool Do ();

  private:
	void _MakeUTurnRestriction (rgn_t * _pRgn, rgnode_t * _pRGNode);

	map_t & m_map;
	size_t m_cMinNodeCount;

	typedef std::list<rgnode_t *> nodesCleared_t;
	nodesCleared_t m_nodesCleared;
	typedef std::list<CAddRGNodeRestrictionAction *> restrictionsAdded_t;
	restrictionsAdded_t m_restrictionsAdded;
};

void CCleanUpExternalNodesAction::_MakeUTurnRestriction (rgn_t * _pRgn, rgnode_t * _pRGNode) {
	typedef rgrestrictions_t::iterator I;
	std::pair<I, I> bounds = m_map.rgrestrictions.equal_range (_pRGNode);

	for (I iRestriction = bounds.first; iRestriction != bounds.second; ++ iRestriction) {
		rgrestriction_t & restr = iRestriction->second;

		if (restr.pFromRoad == _pRgn && restr.pFromRoad == restr.pToRoad)
			return;
	}

	rgrestriction_t restriction;
	restriction.pFromRoad      = _pRgn;
	restriction.pToRoad        = _pRgn;
	restriction.pNode          = _pRGNode;
	restriction.bFromDirection = true;
	restriction.bToDirection   = false;

	CAddRGNodeRestrictionAction * const pAddRGNodeAction = new CAddRGNodeRestrictionAction (restriction);

	PerformEditAction (pAddRGNodeAction);
	m_restrictionsAdded.push_back(pAddRGNodeAction);
}

bool CCleanUpExternalNodesAction::Do () {
	CCritSec::CLock lock (g_cs);

	// Ensure that node counters are correct.
	m_map.CountRGNodes (true);

	const size_t cLevels = m_map.Levels.size ();
	for (int nKind = 0; nKind < rkMax; ++ nKind) {
		rgns_t & rgns = m_map.layers [nKind].rgns;
		if (nKind == rkPoint || nKind == rkPolygon)
			continue;

		for (rgns_t::iterator iRgn = rgns.begin (); iRgn != rgns.end (); ++ iRgn) {
			rgn_t & rgn = * iRgn;
			if (rgn.IsDeleted () || rgn.IsNotMapObject ())
				continue;

			const bool bRPolyline = nKind == rkPolyline && rgn.HasRGNodes () && g_map.pTypeSet->pfIsRoad (rgn);
			if (! bRPolyline)
				continue;
			if ((rgn.dwFlags & rgn_t::fDirIndicator) != 0)
				continue;
			if (rgn.pExtra) {
				if (rgn.pExtra->bOneWay)
					continue;
			}

			rgn_elements_t & elements = rgn.elements;
			const size_t cElements = elements.size ();

			for (size_t cElement = 0; cElement < cElements; ++ cElement) {
				rgn_element_t & element = elements [cElement];
				if (element.bDeleted)
					continue;

				if (element.pExtra) {
					// Nodes.
					point_extras_t & PointExtras = element.pExtra->PointExtras;
					const size_t cPoints = PointExtras.size ();
					const points_t & points = element.points;
					for (size_t cPoint = 0; cPoint < cPoints; ++ cPoint) {
						rgnode_t * const pRGNode = PointExtras [cPoint].GetRGNode ();
						if (pRGNode == NULL)
							continue;

						if (pRGNode->bIsExternal && pRGNode->btPolylines > m_cMinNodeCount)
							_MakeUTurnRestriction (&rgn, pRGNode);
					}
				}
			}
		}
	}

	for (rgnodes_t::iterator iNode = m_map.rgnodes.begin (); iNode != m_map.rgnodes.end (); ++ iNode) {
		rgnode_t & node = * iNode;
		if (! node.IsUsed ())
			continue;

		if (node.bIsExternal && node.btPolylines > m_cMinNodeCount) {
			node.bIsExternal = false;

			m_nodesCleared.push_back (& node);
		}
	}

	Refresh ();

	return true;
}

void CCleanUpExternalNodesAction::Undo () {
	CCritSec::CLock lock (g_cs);

	for (nodesCleared_t::iterator i = m_nodesCleared.begin (); i != m_nodesCleared.end (); ++ i) {
		rgnode_t * const pNode = * i;
		pNode->bIsExternal = true;
	}
	for (restrictionsAdded_t::iterator i2 = m_restrictionsAdded.begin (); i2 != m_restrictionsAdded.end (); ++ i2) {
		CAddRGNodeRestrictionAction * const pAction = * i2;
		pAction->Undo ();
	}
}

void CCleanUpExternalNodesAction::Redo () {
	CCritSec::CLock lock (g_cs);

	for (nodesCleared_t::iterator i = m_nodesCleared.begin (); i != m_nodesCleared.end (); ++ i) {
		rgnode_t * const pNode = * i;
		pNode->bIsExternal = false;
	}
	for (restrictionsAdded_t::iterator i2 = m_restrictionsAdded.begin (); i2 != m_restrictionsAdded.end (); ++ i2) {
		CAddRGNodeRestrictionAction * const pAction = * i2;
		pAction->Redo ();
	}
}*/

/////////////////////////////////////////

void ClearExternalNodes () {
	_CleanUpExternalNodes (g_map, 0);
//	PerformEditAction (new CCleanUpExternalNodesAction (g_map, 0));
}

void CleanUpExternalNodes () {
	_CleanUpExternalNodes (g_map, 1);
//	PerformEditAction (new CCleanUpExternalNodesAction (g_map, 1));
}

void RemoveExternalNodesInsideSelection () {
	CCritSec::CLock lock (g_cs);

	if (g_Selection.size () != 1 || ! g_Selection [0].pLayer || ! g_Selection [0].iRgn->IsPolygon ())
		return;
	const rgn_t & selRgn = * g_Selection [0].iRgn;

	rgns_t & rgns = g_map.layers [rkPolyline].rgns;
	for (rgns_t::iterator iRgn = rgns.begin (); iRgn != rgns.end (); ++ iRgn) {
		rgn_t & rgn = * iRgn;
		if (rgn.IsDeleted () || rgn.IsNotMapObject ())
			continue;

		const bool bRPolyline = rgn.HasRGNodes () && g_map.pTypeSet->pfIsRoad (rgn);
		if (! bRPolyline)
			continue;

		rgn_elements_t & elements = rgn.elements;
		const size_t cElements = elements.size ();
		for (size_t cElement = 0; cElement < cElements; ++ cElement) {
			rgn_element_t & element = elements [cElement];
			if (element.bDeleted)
				continue;
			if (element.pExtra == NULL)
				continue;

			point_extras_t & PointExtras = element.pExtra->PointExtras;
			const size_t cPoints = PointExtras.size ();
			const points_t & points = element.points;
			for (size_t cPoint = 0; cPoint < cPoints; ++ cPoint) {
				rgnode_t * const pRGNode = PointExtras [cPoint].GetRGNode ();
				if (pRGNode == NULL)
					continue;
				static const double c_fE = 1e-6;

				if (pRGNode->bIsExternal && ! pRGNode->bIsTrimNode)
					if (selRgn.Contains (points [cPoint], c_fE))
						SetRGNodeExternal (* pRGNode, false);
			}
		}
	}
}
