//
// GPSMapEdit
// (c) Konstantin Galichsky (kg@geopainting.com), 2002-2009
// (c) Vladimir Kudinov (boba@nm.ru), 2003
//
// Convert objects' labels.
//

# include "StdAfx.h"
# include "Resource.h"
# include "ConvertLabels.h"
# include "UndoRedo.h"
# include "Map.h"
# include "MapView.h"
# include "Load.h"
# include "Status.h"
# include "Label.h"
# include "Types.h"
# include "Transliteration.h"
# include "Settings.h"
# include "ParseHelpers.h"
# include "PrintUnits.h"
# include "fast_floor.h"
# include "Globals.h"

///////////////////////////////////////////////////////

// This method returns a pointer to the first character of the label text 
// skipping all the special ~[0x...] codes.
static
char * _GetLabelText (string_t & label) {
	size_t index;
	if (label.size () >= 4 && ::_strnicmp (label.c_str (), "~[0x", 4) == 0) {
		if (label.size () > 6 && label [6] == ']')
			index = 7;
		else if (label.size () > 8 && label [8] == ']')
			index = 9;
		else
			index = 0;
	} else
		index = 0;
	return & label [index]; // HACK: we return the pointer to a character in the string's internal array 
}

///////////////////////////////////////////////////////

enum ConvertLabelAction_t {
	claUpperCase     = 0,
	claLowerCase     = 1,
	claCapitalize    = 2,
	claCapitalizeAll = 3,
	claTransliterate = 4,
	claMeters2Feet   = 5,
	claFeet2Meters   = 6
};

// Settings' names.
const char SETTING_NAME_ACTION         [] = "ConvertLabels_Action";
const char SETTING_ALL_ATTRIBUTES      [] = "ConvertLabels_AllAttributes";
//const char SETTING_NAME_TRANSLITERATOR [] = "ConvertLabels_Transliterator";

struct CConvertLabelsOperationSettings {
	ConvertLabelAction_t action;
	bool bAllAttributes;

	const CTransliterator * pTransliterator;

	CConvertLabelsOperationSettings () :
		action (claUpperCase),
		bAllAttributes (false),
		pTransliterator (NULL)
	{}

	void LoadSettings () {
		action = static_cast<ConvertLabelAction_t> (GetSetting (SETTING_NAME_ACTION, (DWORD) claUpperCase));
		bAllAttributes = GetSetting (SETTING_ALL_ATTRIBUTES, false);
		pTransliterator = NULL;
	}

	void SaveSettings () {
		SaveSetting (SETTING_NAME_ACTION,    (DWORD) action);
		SaveSetting (SETTING_ALL_ATTRIBUTES, (DWORD) bAllAttributes);
	}
};

class CConvertLabelsAction : public CEditAction {
  public:
	CConvertLabelsAction (const CConvertLabelsOperationSettings & _opSettings) :
		m_opSettings (_opSettings)
	{}
	virtual void GetName (string_t & _str) const;
	virtual cu_t CanUndo () const {return cuBackup;}
	virtual bool Do ();
	virtual void Undo () {}
	virtual void Redo () {}

  private:
	CConvertLabelsOperationSettings m_opSettings;

	void ApplyToLabel (string_t & _label, WORD _wCodePage, rgn_t * _pRgn);
};

void CConvertLabelsAction::GetName (string_t & _str) const {
	switch (m_opSettings.action) {
		case claUpperCase:     _str.append ("upper case"); break;
		case claLowerCase:     _str.append ("lower case"); break;
		case claCapitalize:    _str.append ("capitalize"); break;
		case claCapitalizeAll: _str.append ("capitalize all words"); break;
		case claMeters2Feet:   _str.append ("meters to feet"); break;
		case claFeet2Meters:   _str.append ("feet to meters"); break;
		case claTransliterate: _str.append ("transliterate labels"); break;

		default: assert (false);
	}
}

bool CConvertLabelsAction::Do () {
	CCritSec::CLock lock (g_cs);

	const WORD wCodePage = g_map.Header.wCodePage;

	const bool bAllStrings = m_opSettings.bAllAttributes && m_opSettings.action != claMeters2Feet && m_opSettings.action != claFeet2Meters;

	// Apply to labels of all objects.
	for (size_t cKind = 0; cKind < rkMax; ++ cKind) {
		rgns_t & rgns = g_map.layers [cKind].rgns;

		for (rgns_t::iterator iRgn = rgns.begin (); iRgn != rgns.end (); ++ iRgn) {
			rgn_t & rgn = * iRgn;
			if (rgn.IsImmutable () || rgn.IsNotMapObject ())
				continue;

			string_t & label = rgn.strLabel;
			if (! label.empty ())
				ApplyToLabel (label, wCodePage, & rgn);

			if (bAllStrings) {
				if (rgn.pExtra && ! rgn.pExtra->strStreetDesc.empty ())
					ApplyToLabel (rgn.pExtra->strStreetDesc, wCodePage, NULL);
				if (rgn.pExtra && ! rgn.pExtra->strDescription.empty ())
					ApplyToLabel (rgn.pExtra->strDescription, wCodePage, NULL);
			}
		}
	}

	if (bAllStrings) {
		// Apply to cities.
		for (cities_t::iterator ic = g_map.Cities.begin (); ic != g_map.Cities.end (); ++ ic) {
			city_t & city = * ic;
			if (city.IsDeleted ())
				continue;

			ApplyToLabel (city.strName, wCodePage, NULL);
		}

		// Apply to regions.
		for (regions_t::iterator ir = g_map.Regions.begin (); ir != g_map.Regions.end (); ++ ir) {
			region_t & region = * ir;
			if (region.IsDeleted ())
				continue;

			ApplyToLabel (region.strName, wCodePage, NULL);
		}

		// Apply to countries.
		for (countries_t::iterator icc = g_map.Countries.begin (); icc != g_map.Countries.end (); ++ icc) {
			country_t & country = * icc;
			if (country.IsDeleted ())
				continue;

			ApplyToLabel (country.strName, wCodePage, NULL);
		}
	}

	return true;
}

static
void _ConvertElevationUnits (float _k, rgn_t & _rgn) {
	if (g_map.pTypeSet->pfIsHeight (_rgn)) {
		float fElevation = 0;
		const char * p;
		if (! ParseReal (_rgn.strLabel.c_str (), _rgn.strLabel.c_str () + _rgn.strLabel.size (), & p, fElevation))
			return;
		fElevation *= _k;

		char strElevation [20];
		PrintElevation (strElevation, NULL, fElevation, g_ElevationUnit, true);

		_rgn.strLabel = strElevation;
	} else if (g_map.pTypeSet->pfIsFeature (_rgn)) {
		const char * pPrefixEnd = ::strstr (_rgn.strLabel.c_str (), "~[0x1f]");
		if (pPrefixEnd == NULL)
			return;
		const char * pElevation = pPrefixEnd + 7;

		float fElevation = 0;
		const char * p;
		if (! ParseReal (pElevation, _rgn.strLabel.c_str () + _rgn.strLabel.size (), & p, fElevation))
			return;
		fElevation *= _k;

		char strElevation [20];
		PrintElevation (strElevation, NULL, fElevation, g_ElevationUnit, true);

		string_t strPrefix (_rgn.strLabel.c_str (), pElevation);
		_rgn.strLabel.swap (strPrefix);
		_rgn.strLabel.append (strElevation);
	}
}

void CConvertLabelsAction::ApplyToLabel (string_t & _strLabel, WORD _wCodePage, rgn_t * _pRgn) {
	char * strLabelTextFirstChar = _GetLabelText (_strLabel);
	char * strLabelTextLastChar  = & _strLabel [_strLabel.length ()];

	switch (m_opSettings.action) {
		case claUpperCase:
//			LCMapString ?
			::CharUpperBuff (strLabelTextFirstChar, strLabelTextLastChar - strLabelTextFirstChar);
			break;

		case claLowerCase:
			::CharLowerBuff (strLabelTextFirstChar, strLabelTextLastChar - strLabelTextFirstChar);
			break;

		case claCapitalize:	{
			// First, set all the text to lower case.
			::CharLowerBuff (strLabelTextFirstChar, strLabelTextLastChar - strLabelTextFirstChar);

			// Find first letter, skipping punctuation marks, whitespace and digits.
			for (char * pLetter = strLabelTextFirstChar; pLetter < strLabelTextLastChar; pLetter = ::CharNextExA (_wCodePage, pLetter, 0)) {
				if (::IsCharAlpha (* pLetter)) {
					::CharUpperBuff (pLetter, 1);
					break;
				}
			}
			break;
		}

		case claCapitalizeAll: {
			// First, set all the text to lower case.
			::CharLowerBuff (strLabelTextFirstChar, strLabelTextLastChar - strLabelTextFirstChar);

			bool bSpace = true;
			for (char * pLetter = strLabelTextFirstChar; pLetter < strLabelTextLastChar; pLetter = ::CharNextExA (_wCodePage, pLetter, 0)) {
				if (::IsCharAlphaNumeric (* pLetter)) {
					if (bSpace) {
						::CharUpperBuff (pLetter, 1);
						bSpace = false;
					}
				} else
					bSpace = true;
			}
			break;
		}

		case claMeters2Feet:
			if (_pRgn)
				_ConvertElevationUnits (c_feet_per_m,   * _pRgn);
			break;
		case claFeet2Meters:
			if (_pRgn)
				_ConvertElevationUnits (1/c_feet_per_m, * _pRgn);
			break;

		case claTransliterate: {
			if (m_opSettings.pTransliterator == NULL)
				break;

			// Copy any symbols prior label text.
			string_t strTransliterated (& _strLabel [0], strLabelTextFirstChar);
			// Transliterate the symbols that consist the label text
			m_opSettings.pTransliterator->Do (strLabelTextFirstChar, strLabelTextLastChar, strTransliterated);

			_strLabel.swap (strTransliterated);

			break;
		}

		default: assert (false);
	}
}

///////////////////////////

# include "Dialog.h"

class CConvertLabelsDialog : public CDialog {
  public:
	CConvertLabelsDialog (CConvertLabelsOperationSettings & _OpSettings) :
		CDialog (IDD_CONVERT_LABELS),
		m_OpSettings (_OpSettings)
	{}

	// Override.
	virtual bool OnInit (HWND _hwndFocus);
	virtual void OnCommand (WORD _wCmd, WORD _wNotifyCode, HWND _hwndCtl);

  private:
	CConvertLabelsOperationSettings & m_OpSettings;

	HWND m_hTransliteration;
};

bool CConvertLabelsDialog::OnInit (HWND _hwndFocus) {
	// Prepare transliterator.
	g_Transliteration.Load ();
	const size_t cTransliteratorsNum = g_Transliteration.GetNumOfTransliterators ();

	//
	// Prepare action radiobuttons.
	//
	// NOTE: it is assumed that IDs for controls occupy continuous ranges and 
	// are sorted in same order as respective values in enums.
	UINT uiRadioButton = IDC_LABELS_TRANSLITERATE;
	switch (m_OpSettings.action) {
		case claUpperCase:     uiRadioButton = IDC_LABELS_TO_UPPER_CASE;    break;
		case claLowerCase:     uiRadioButton = IDC_LABELS_TO_LOWER_CASE;    break;
		case claCapitalize:    uiRadioButton = IDC_LABELS_CAPITALIZE_FIRST; break;
		case claCapitalizeAll: uiRadioButton = IDC_LABELS_CAPITALIZE_ALL;   break;
		case claTransliterate: uiRadioButton = IDC_LABELS_TRANSLITERATE;    break;
		case claMeters2Feet:   uiRadioButton = IDC_LABELS_M2FT;             break;
		case claFeet2Meters:   uiRadioButton = IDC_LABELS_FT2M;             break;
		default: assert (false);
	}
	::CheckRadioButton (m_hWnd, uiRadioButton, uiRadioButton, uiRadioButton);
	if (cTransliteratorsNum == 0)
		EnableDlgItem (IDC_LABELS_TRANSLITERATE, false);

	// Prepare elevation units translators.
	if ((g_map.pTypeSet->dwFeatures & CTypeSet::tsfHasElevationUnits) == 0) {
		EnableDlgItem (IDC_LABELS_M2FT, false);
		EnableDlgItem (IDC_LABELS_FT2M, false);
	}

	//
	// Prepare 'Transliterate' field.
	//
	m_hTransliteration = GetDlgItem (IDC_LABELS_CHOOSE_TRANSLITERATION);

	EnableDlgItem (
		IDC_LABELS_CHOOSE_TRANSLITERATION,
		m_OpSettings.action == claTransliterate && cTransliteratorsNum != 0
	);

	for (size_t cTrl = 0; cTrl < cTransliteratorsNum; ++ cTrl) {
		const CTransliterator * const pTransliterator = g_Transliteration.GetTransliterator (cTrl);

		::SendMessage (m_hTransliteration, CB_INSERTSTRING, cTrl, (LPARAM) pTransliterator->GetName ().c_str ());

		const WORD wLangId = pTransliterator->GetLangId ();
		if (wLangId == ::GetUserDefaultLangID ()) {
			::SendMessage (m_hTransliteration, CB_SETCURSEL, cTrl, 0);
			m_OpSettings.pTransliterator = pTransliterator;
		}
	}

	// Apply to all attributes.
	CheckDlgButton (IDC_ALL_ATTRIBUTES, m_OpSettings.bAllAttributes);

	return true;
}

void CConvertLabelsDialog::OnCommand (WORD _wCmd, WORD _wNotifyCode, HWND _hwndCtl) {
	const WORD controlID = _wCmd;

	// Disable transliteration drop-down list when transliteration is not selected
	// NOTE: it is assumed that controls IDs occupy a continuous range from
	// IDC_LABELS_TO_UPPER_CASE to IDC_LABELS_TRANSLITERATE.
	if (
		(IDC_LABELS_TO_UPPER_CASE <= controlID && controlID <= IDC_LABELS_TRANSLITERATE) ||
		controlID == IDC_LABELS_M2FT ||
		controlID == IDC_LABELS_FT2M
	)
		EnableDlgItem (IDC_LABELS_CHOOSE_TRANSLITERATION, controlID == IDC_LABELS_TRANSLITERATE);

	switch (controlID) {
		// Transliteration selection combo-box.
		case IDC_LABELS_CHOOSE_TRANSLITERATION:
			if (_wNotifyCode == CBN_SELCHANGE) {
				const size_t cSelIndex = ::SendMessage (m_hTransliteration, CB_GETCURSEL, 0, 0);
				if (cSelIndex < g_Transliteration.GetNumOfTransliterators ())
					m_OpSettings.pTransliterator = g_Transliteration.GetTransliterator (cSelIndex);
			}
			return;

		// Action selection radio-buttons.
		case IDC_LABELS_TO_UPPER_CASE:    m_OpSettings.action = claUpperCase;     return;
		case IDC_LABELS_TO_LOWER_CASE:    m_OpSettings.action = claLowerCase;     return;
		case IDC_LABELS_CAPITALIZE_FIRST: m_OpSettings.action = claCapitalize;    return;
		case IDC_LABELS_CAPITALIZE_ALL:   m_OpSettings.action = claCapitalizeAll; return;
		case IDC_LABELS_M2FT:             m_OpSettings.action = claMeters2Feet;   return;
		case IDC_LABELS_FT2M:             m_OpSettings.action = claFeet2Meters;   return;
		case IDC_LABELS_TRANSLITERATE:    m_OpSettings.action = claTransliterate; return;

		case IDC_ALL_ATTRIBUTES:
			m_OpSettings.bAllAttributes = IsDlgButtonChecked (IDC_ALL_ATTRIBUTES);
			return;

		case IDOK:     EndDialog (0); return;
		case IDCANCEL: EndDialog (1); return;
	}
}

void ConvertLabels (HWND _hWnd) {
	CConvertLabelsOperationSettings opSettings;
	// Load from settings stored in registry.
	opSettings.LoadSettings ();

	CConvertLabelsDialog cld (opSettings);
	if (cld.ShowModal (_hWnd) == 1)
		// Dialog was canceled.
		return;

	// On OK save choosen settings into the registry.
	opSettings.SaveSettings ();

	PerformEditAction (new CConvertLabelsAction (opSettings));

	Refresh ();
}
