//
// GPSMapEdit
// (c) Konstantin Galichsky (kg@geopainting.com), 2008-2009
//
// DVR binary files parser.
//

# include "StdAfx.h"
# include "Map.h"
# include "Types.h"
# include "Load.h"
# include "Log.h"
# include "Status.h"
# include "File.h"
# include "FileWriter.h"
# include "ParseHelpers.h"
# include "DrawHelpers.h"
# include "fast_floor.h"
# include "Wrappers.h"
# include "Globals.h"
# include "Version.h"

# define USE_GDIPLUS
# ifdef USE_GDIPLUS
	# ifndef min
	# define min __min
	# endif
	# ifndef max
	# define max __max
	# endif
	# include <GdiPlus.h>
	# pragma comment (lib, "GdiPlus.lib")
	# pragma comment (linker, "/delayload:GdiPlus.dll")
	static bool g_bG;
# endif // USE_GDIPLUS

const DWORD c_dwDvrFrameMarker = 0x56801234;

# ifdef _MSC_VER
#	pragma pack (push, 1)
# endif

struct CDvrFrameHdr {
	DWORD dwStartMarker;

	WORD  wFramesRate;
	WORD  w0;

	WORD  wPrevFrameSizeBy256;
	WORD  w1;
	WORD  w21;
	WORD  w22;
	WORD  wFrameSizeBy256;
	WORD  w2;
	WORD  w11;
	WORD  w12;

	BYTE  btSecond;
	BYTE  btMinute;
	BYTE  btHour;
	BYTE  btDay;
	BYTE  btMonth;
	BYTE  btYear;
	BYTE  bt1 [2];

	DWORD dwHdrAndJpgSize;

	WORD  w00;
	WORD  w01;
	WORD  w02;
	WORD  w03;
	WORD  w04;
	WORD  wFramePairIdx;
	WORD  w06;
	WORD  wSomeIdx0;
	WORD  wSomeIdx1;
	WORD  wIdxInSecond;
	WORD  wFrameIdx;
	WORD  wSomeIdx3;
/*
	WORD  wSizeOfPrev4Frames;
	WORD  wSizeOfPrev2Frames;
	WORD  wSizeOfCurrent4Frames;
	WORD  wSizeOfCurrent2Frames;
*/
	WORD w32 [18];

	char  strFirmware [8];

	DWORD dwF0;
	DWORD dwF1;
	DWORD dwF2;

	union {
		BYTE btFlags;
		struct {
			BYTE bTurnRightSignal : 1;
			BYTE bTurnLeftSignal  : 1;
			BYTE bReverse         : 1;
			BYTE bSafetyAlarm     : 1;
			BYTE bDimensionLights : 1;
			BYTE bBrake           : 1;

			BYTE bSpeedInKMH      : 1;
		};
	};
	BYTE  bt2;

	union {
		BYTE bt3;
		struct {
			BYTE btHeading : 3;
			BYTE bLonWest  : 1;
			BYTE bLatNorth : 1;
		};
	};
	BYTE  bt4;

	WORD  wSpeed;
	BYTE  btDoors;
	BYTE  bt5;

	DWORD dwLongitude;
	DWORD dwLatitude;

	BYTE  btVehicleKind;
	BYTE  bt6;
	WORD  wVehicleID;
};

enum {
	DVR_WAV_LEN_max = 2560,
	DVR_WAV_LEN_min = 2048
};

# ifdef _MSC_VER
#	pragma pack (pop)
# endif

static
int _GetDvrFPS (const CDvrFrameHdr * _pFrameHdr) {
	switch (_pFrameHdr->wFramesRate) {
		case 0x08: return  1;
		case 0x10: return  2;
		case 0x18: return  4;
		case 0x20: return  4;
		case 0x40: return  8;
		case 0x30: return  8;
		case 0x60: return 15;
		case 0x80: return 12;
		case 0xC8: return 25;
		default:   return -1;
	}
}

static
float _ParseDvrPosition (DWORD _dw) {
	const WORD wHi = _dw >> 16;
	const WORD wLo = _dw & 0xFFFF;
	const int nDeg = fast_floor (wHi*0.01f);
	const int nMin = wHi - nDeg*100;
	const float fSec = wLo*0.01;
	return nDeg + (nMin + fSec/60.f)/60.f;
}

static
bool _ParseDvrHdr (const CDvrFrameHdr * _pFrameHdr, point_t & _pt) {
	if (_pFrameHdr->dwLongitude == 0 || _pFrameHdr->dwLatitude == 0)
		return false;

	_pt.x = _ParseDvrPosition (_pFrameHdr->dwLongitude);
	_pt.y = _ParseDvrPosition (_pFrameHdr->dwLatitude);

	if (_pFrameHdr->bLonWest != 0)
		_pt.x = -_pt.x;
	if (_pFrameHdr->bLatNorth == 0)
		_pt.y = -_pt.y;

	return true;
}

static
void _ParseDvrHdr (const CDvrFrameHdr * _pFrameHdr, trackpoint_info_t & _tpi, bool _bAddMilliseconds = false) {
	SYSTEMTIME st;
	st.wMilliseconds = 0;
	st.wSecond = _pFrameHdr->btSecond;
	st.wMinute = _pFrameHdr->btMinute;
	st.wHour   = _pFrameHdr->btHour;
	st.wDay    = _pFrameHdr->btDay;
	st.wMonth  = _pFrameHdr->btMonth;
	st.wYear   = _pFrameHdr->btYear + 2000;

	DATE dt;
	::SystemTimeToVariantTime (& st, & dt);

	_tpi.wFlags |= trackpoint_info_t::defTime;
	_tpi.dtTime = dt;
	if (_bAddMilliseconds) {
		_tpi.dtTime += _pFrameHdr->wIdxInSecond/(50.*24*60*60);
		_tpi.wFlags |= trackpoint_info_t::defTimeMilliseconds;
	}

	_tpi.wFlags |= trackpoint_info_t::defSpeed;
	_tpi.fSpeed = _pFrameHdr->bSpeedInKMH != 0 ? _pFrameHdr->wSpeed : _pFrameHdr->wSpeed*c_m_per_smile/1000;

	_tpi.wFlags |= trackpoint_info_t::defHeading;
	static const int s_nHeadings [] = {90, 270, 180, 0};
	_tpi.fHeading = _pFrameHdr->btHeading == 0 || _pFrameHdr->btHeading > 4 ? 0 : s_nHeadings [_pFrameHdr->btHeading - 1];
}

///////////////////////////////////////////////////////////////////////

const size_t c_cDvrChannels = 4;
const int    c_nDvrTopStrip = 5; // Amount of top scan lines of the image to be hidden.

class CDvrVideoRenderer : public CTrackVideoRendererBase {
  public:
	typedef std::map<ULONGLONG, float> pos2heading_t;

	CDvrVideoRenderer (const char * _strFileName, size_t _cFPS, pos2heading_t & _pos2heading, DWORD _dwVehicleID);
	~CDvrVideoRenderer ();

	// Override.
	virtual size_t GetChannels () const {return c_cDvrChannels;}
	virtual size_t GetWidth  (size_t _cChannel) const {return m_cWidth [_cChannel];}
	virtual size_t GetHeight (size_t _cChannel) const {
		const size_t cHeight = m_cHeight [_cChannel];
		if (cHeight == 0)
			return 0;
		return m_cWidth [_cChannel] == 720 ? 2*(cHeight - c_nDvrTopStrip) : (cHeight - c_nDvrTopStrip);
	}
	virtual size_t GetFPS () const {return m_cFPS;}
	virtual void GetVehicleID (string_t & _strId) const;
	virtual bool GetVehicleInfo (point_t & _pt, trackpoint_info_t & _tpi) const;
	virtual bool IsAtStart () const;
	virtual bool IsAtEnd   () const;
	virtual double GetProgress () const;
	virtual void SetProgress (double _fPart);
	virtual DWORD GetProgressPrec () const;
	virtual void SetProgressPrec (DWORD _fProgressPrec);
	virtual void GoForward  (size_t _cTimes, bool _bFast);
	virtual void GoBackward (size_t _cTimes);
	virtual void GoFrame (size_t _cTrackPointVideoData0, size_t _cTrackPointVideoData1);

	virtual void SetSpeed (float _fSpeed);
	virtual void Mute (bool _bMute);
	virtual double GetSoundLevel ();

	virtual bool Render (HDC _hDC, int _nLeft, int _nTop, int _nWidth, int _nHeight, size_t _cChannel) const;
	virtual bool CaptureDIBitmap (size_t _cChannel, void * _pBits, const BITMAPINFO * _pBMI) const;

  private:
	const DWORD m_dwVehicleID;

	CFileForRead m_hFile;
	ULONGLONG m_ullFileSize;

	int m_nLogPixelX, m_nLogPixelY;

	// Frames per second.
	const size_t m_cFPS;

	float m_fSpeed;
	bool m_bMute;

	mutable CMemDC m_MemDC_current;
	mutable CMemDC m_MemDC [c_cDvrChannels];
	mutable size_t m_cWidth_current, m_cHeight_current;
	mutable size_t m_cWidth [c_cDvrChannels], m_cHeight [c_cDvrChannels];
	size_t _ParseImageImpl () const;
	size_t _ParseImage () const;
	void _ClearImages ();

	struct cache_item_t {
		bool bUsed;
		BYTE btTime;
		ULONGLONG ullOffset;
		size_t cChannel;
		CMemDC MemDC;
		size_t cWidth, cHeight;
	};
	enum {c_CacheItems = 8};
	mutable cache_item_t Cache [c_CacheItems];
	mutable size_t m_btCacheTime;
	cache_item_t * _FindCacheItem (ULONGLONG _ullOffset) const;
	cache_item_t * _FindFreeOrLRUCacheItem () const;

	bool m_bAtEnd;
	ULONGLONG m_ullOffset;
	CDvrFrameHdr m_FrameHdr;
	mutable std::vector<BYTE> m_JpegData;

	void _SetFrame (ULONGLONG _ullFileOffset, bool _bScanForward);
	void _GoForward1 ();
	void _GoBackward1 ();
	void _UpdateAllChannels (bool _bScanForward);

	TIME_ZONE_INFORMATION m_tzi;

	pos2heading_t m_pos2heading;

	////////////////////////////

	HWAVEOUT m_hWaveOut;
	DWORD m_dwAvgBytesPerSec;

	mutable DWORD m_dwPrevChunkTime;
	mutable bool m_bPrevChunkDefined;

	struct wav_chunk_t {
		WAVEHDR wh;
		bool bBusy;
		char baWAV [DVR_WAV_LEN_max];
	};
	enum {CHUNKS = 4};
	mutable wav_chunk_t m_WavChunks [CHUNKS];
	size_t _GetFreeWavChunk () const;

	double m_fPrevEnergy;
};

CDvrVideoRenderer::CDvrVideoRenderer (const char * _strFileName, size_t _cFPS, pos2heading_t & _pos2heading, DWORD _dwVehicleID) :
	CTrackVideoRendererBase (_strFileName),
	m_dwVehicleID (_dwVehicleID),
	m_hFile (_strFileName),
	m_ullOffset (0),
	m_bAtEnd (false),
	m_cFPS (_cFPS),
	m_cWidth_current (0), m_cHeight_current (0),
	m_hWaveOut (NULL),
	m_fSpeed (0),
	m_fPrevEnergy (0),
	m_bMute (false)
{
	m_pos2heading.swap (_pos2heading);

	::memset (m_cWidth,  0, sizeof (m_cWidth));
	::memset (m_cHeight, 0, sizeof (m_cHeight));

	ULARGE_INTEGER uli;
	uli.LowPart = ::GetFileSize (m_hFile, & uli.HighPart);
	m_ullFileSize = uli.QuadPart;

	m_nLogPixelX = ::GetDeviceCaps (CClientDC (NULL), LOGPIXELSX);
	m_nLogPixelY = ::GetDeviceCaps (CClientDC (NULL), LOGPIXELSY);

	// Get local time info.
	if (::GetTimeZoneInformation (& m_tzi) == TIME_ZONE_ID_DAYLIGHT)
		m_tzi.Bias += m_tzi.DaylightBias;

	// Clear cache.
	m_btCacheTime = 0;
	for (size_t c = 0; c < c_CacheItems; ++ c) {
		cache_item_t & ci = Cache [c];
		ci.bUsed    = false;
		ci.cChannel = -1;
		ci.cWidth   = 0;
		ci.cHeight  = 0;
	}

	// Prepare to play sound.
	WAVEFORMATEX wf;
	wf.wFormatTag      = WAVE_FORMAT_PCM; 
	wf.nChannels       = 2;
	wf.nSamplesPerSec  = 8000;
	wf.wBitsPerSample  = 16;
	wf.nBlockAlign     = wf.nChannels*(wf.wBitsPerSample/8);
	wf.nAvgBytesPerSec = wf.nSamplesPerSec*wf.nChannels*(wf.wBitsPerSample/8);
	wf.cbSize          = 0;
	const MMRESULT mr = ::waveOutOpen (& m_hWaveOut, WAVE_MAPPER, & wf, 0, 0, CALLBACK_NULL);
	if (mr != MMSYSERR_NOERROR) {
# ifdef _DEBUG
		char strError [64];
		::waveOutGetErrorText (mr, strError, ARRAY_SIZE (strError));
		ATLTRACE ("Failed to open WAV device: %s (%d)\n", strError, mr);
# endif // _DEBUG
	}
	m_dwAvgBytesPerSec = wf.nAvgBytesPerSec;

	::memset (m_WavChunks, 0, sizeof (m_WavChunks));
	for (size_t cChunk = 0; cChunk < CHUNKS; ++ cChunk) {
		WAVEHDR & wh = m_WavChunks [cChunk].wh;
		wh.lpData         = m_WavChunks [cChunk].baWAV;
		wh.dwBufferLength = DVR_WAV_LEN_max;
		wh.dwFlags        = WHDR_DONE;
		wh.dwLoops        = 1;
		wh.dwUser         = cChunk;

		::waveOutPrepareHeader (m_hWaveOut, & wh, sizeof (wh));
	}

	m_bPrevChunkDefined = false;
}

CDvrVideoRenderer::~CDvrVideoRenderer () {
	if (m_hWaveOut) {
		::waveOutReset (m_hWaveOut);

		for (size_t c = 0; c < CHUNKS; ++ c) {
			wav_chunk_t & chunk = m_WavChunks [c];
			if ((chunk.wh.dwFlags & (WHDR_DONE | WHDR_PREPARED)) != 0)
				::waveOutUnprepareHeader (m_hWaveOut, & chunk.wh, sizeof (chunk.wh));
		}

		::waveOutClose (m_hWaveOut);
	}
}

size_t CDvrVideoRenderer::_GetFreeWavChunk () const {
	for (size_t c = 0; c < CHUNKS; ++ c) {
		const wav_chunk_t & chunk = m_WavChunks [c];
		if (chunk.bBusy == false || (chunk.wh.dwFlags & WHDR_DONE) != 0)
			return c;
	}

	return -1;
}

void CDvrVideoRenderer::_UpdateAllChannels (bool _bScanForward) {
	const ULONGLONG ullOffsetSaved = m_ullOffset;

	bool bChannel [c_cDvrChannels];
	::memset (bChannel, 0, sizeof (bChannel));
	size_t cChannels = 0;

	for (size_t cAttempts = 0; cAttempts < c_cDvrChannels + 2; ++ cAttempts) {
		const size_t cChannel = _ParseImage ();
		if (cChannel != -1) {
			if (bChannel [cChannel] == false) {
				++ cChannels;
				bChannel [cChannel] = true;

				m_MemDC [cChannel].Swap (m_MemDC_current);
				std::swap (m_cWidth  [cChannel], m_cWidth_current);
				std::swap (m_cHeight [cChannel], m_cHeight_current);
			}
		}

		if (cChannels == c_cDvrChannels)
			break;
		if (cAttempts == 1 && cChannels == 1)
			// NOTE: this is single channel stream.
			break;

		if (_bScanForward)
			_GoForward1 ();
		else
			_GoBackward1 ();
	}

	_SetFrame (ullOffsetSaved, true);
}

static
void _SetFilePointer (HANDLE _hFile, ULONGLONG _ullOffset) {
	LONG lHi = _ullOffset >> 32;
	::SetFilePointer (_hFile, DWORD (_ullOffset), & lHi, FILE_BEGIN);
}

void CDvrVideoRenderer::_SetFrame (ULONGLONG _ullFileOffset, bool _bScanForward) {
	m_ullOffset = _ullFileOffset & 0xFFFFFFFFFFFFFF00;

	// Read/scan frame header.
	while (0 <= m_ullOffset && m_ullOffset < m_ullFileSize) {
		_SetFilePointer (m_hFile, m_ullOffset);

		DWORD dwRead = 0;
		::ReadFile (m_hFile, & m_FrameHdr, sizeof (m_FrameHdr), & dwRead, NULL);
		if (m_FrameHdr.dwStartMarker == c_dwDvrFrameMarker)
			break;

		if (_bScanForward)
			m_ullOffset += 256;
		else
			m_ullOffset -= 256;
	}
	m_bAtEnd = m_ullOffset >= m_ullFileSize - 512;

	m_fPrevEnergy = 0;
}

void CDvrVideoRenderer::GetVehicleID (string_t & _strId) const {
	char strVehicleID [16];
	::sprintf (strVehicleID, "%d", m_dwVehicleID);
	_strId.append (strVehicleID);
}

bool CDvrVideoRenderer::GetVehicleInfo (point_t & _pt, trackpoint_info_t & _tpi) const {
	_ParseDvrHdr (& m_FrameHdr, _tpi, true);
	_tpi.dtTime += float (m_tzi.Bias)/(24*60);

	// Fix heading.
	const pos2heading_t::const_iterator ih = m_pos2heading.upper_bound (m_ullOffset);
	if (ih != m_pos2heading.end ()) {
		_tpi.fHeading = ih->second;
		_tpi.wFlags |= trackpoint_info_t::defHeading;
	}

	return _ParseDvrHdr (& m_FrameHdr, _pt);
}

void CDvrVideoRenderer::SetSpeed (float _fSpeed) {
	m_fSpeed = _fSpeed;
}

bool CDvrVideoRenderer::IsAtStart () const {
	return m_ullOffset == 0;
}

bool CDvrVideoRenderer::IsAtEnd () const {
	return m_bAtEnd;
}

double CDvrVideoRenderer::GetProgress () const {
	return double ((LONGLONG) m_ullOffset)/(LONGLONG) m_ullFileSize;
}

void CDvrVideoRenderer::SetProgress (double _fPart) {
	_ClearImages ();

	const ULONGLONG ullOffset = m_ullFileSize <= 0xFFFFFFFF ? fast_floor (_fPart*(long) m_ullFileSize + .5f) : _fPart*(LONGLONG) m_ullFileSize;
	_SetFrame (ullOffset, true);
	_UpdateAllChannels (true);
}

DWORD CDvrVideoRenderer::GetProgressPrec () const {
	return m_ullOffset >> 8;
}

void CDvrVideoRenderer::SetProgressPrec (DWORD _fProgressPrec) {
	_ClearImages ();

	const ULONGLONG ullOffset = ((LONGLONG) _fProgressPrec) << 8;
	_SetFrame (ullOffset, true);
	_UpdateAllChannels (true);

}

void CDvrVideoRenderer::_GoForward1 () {
	const size_t cFrameSize = m_FrameHdr.wFrameSizeBy256*512;
	_SetFrame (m_ullOffset + cFrameSize, true);
}

void CDvrVideoRenderer::GoForward (size_t _cTimes, bool _bFast) {
	for (size_t c = 0; c < _cTimes; ++ c)
		_GoForward1 ();
	if (! _bFast)
		_UpdateAllChannels (true);
}

void CDvrVideoRenderer::_GoBackward1 () {
	size_t cFrameSize = m_FrameHdr.wPrevFrameSizeBy256*512;
	if (cFrameSize == 0)
		cFrameSize = 512;
	_SetFrame (cFrameSize < m_ullOffset ? m_ullOffset - cFrameSize : 0, false);
}

void CDvrVideoRenderer::GoBackward (size_t _cTimes) {
	for (size_t c = 0; c < _cTimes; ++ c)
		_GoBackward1 ();
	_UpdateAllChannels (false);
}

void CDvrVideoRenderer::GoFrame (size_t _cTrackPointVideoData0, size_t _cTrackPointVideoData1) {
	_ClearImages ();

	ULARGE_INTEGER uli;
	uli.LowPart  = _cTrackPointVideoData0;
	uli.HighPart = _cTrackPointVideoData1;
	_SetFrame (uli.QuadPart, true);
	_UpdateAllChannels (true);
}

bool CDvrVideoRenderer::CaptureDIBitmap (size_t _cChannel, void * _pBits, const BITMAPINFO * _pBMI) const {
	if (_cChannel >= c_cDvrChannels)
		return false;

	const int nWidth  = m_cWidth  [_cChannel];
	const int nHeight = m_cHeight [_cChannel] - c_nDvrTopStrip;
	if (nWidth <= 0 || nHeight <= 0)
		return false;

	const int nBmpHeight = _pBMI->bmiHeader.biHeight;
	const int nBmpWidth  = _pBMI->bmiHeader.biWidth;
	const int nBmpScanWidth = Align4 (3*nBmpWidth);

	HBITMAP hbmp = m_MemDC [_cChannel].GetBitmap ();
	BITMAP bmp;
	::GetObject (hbmp, sizeof (BITMAP), & bmp);

	// NOTE: GetDIBits() does not work if height of DIB is not equal to one of DDB.
	BITMAPINFO bmi = * _pBMI;
	bmi.bmiHeader.biHeight = m_cHeight [_cChannel];
	if (::GetDIBits (m_MemDC [_cChannel], m_MemDC [_cChannel].GetBitmap (), 0, nHeight, _pBits, const_cast<BITMAPINFO *> (& bmi), DIB_RGB_COLORS) != nHeight)
		return false;

	const bool bDoubleH = m_cWidth [_cChannel] == 720;
	if (bDoubleH) {
		BYTE * const pBits = reinterpret_cast<BYTE *> (_pBits);
		for (int y = nHeight - 1; y >= 0; -- y) {
			::memcpy (pBits + nBmpScanWidth*(2*y + 0), pBits + nBmpScanWidth*y, nBmpScanWidth);
			::memcpy (pBits + nBmpScanWidth*(2*y + 1), pBits + nBmpScanWidth*y, nBmpScanWidth);
		}
	}

	return true;
}

bool CDvrVideoRenderer::Render (HDC _hDC, int _nLeft, int _nTop, int _nWidth, int _nHeight, size_t _cChannel) const {
	if (_cChannel >= c_cDvrChannels)
		return false;

	const int nWidth  = m_cWidth  [_cChannel];
	const int nHeight = m_cHeight [_cChannel] - c_nDvrTopStrip;
	if (nWidth <= 0 || nHeight <= 0)
		return false;

	const bool bDoubleH = m_cWidth [_cChannel] == 720;
	const int nHeight2 = bDoubleH ? nHeight*2 : nHeight;

	const int ws = __min (nWidth  *_nWidth/nWidth, nWidth  *_nHeight/nHeight2);
	const int hs = __min (nHeight2*_nWidth/nWidth, nHeight2*_nHeight/nHeight2);

	const bool bCompress = ws < nWidth || hs < nHeight2;
	::SetStretchBltMode (_hDC, bCompress ? HALFTONE : COLORONCOLOR);

	::StretchBlt (_hDC, _nLeft + (_nWidth - ws)/2, _nTop + (_nHeight - hs)/2, ws, hs, m_MemDC [_cChannel], 0, c_nDvrTopStrip, nWidth, nHeight, SRCCOPY);

	const int c_nStatusLevel = _nHeight - 15;

	// Show door status.
//	::SelectObject (_hDC, g_hFont);
	::SetTextAlign (_hDC, TA_TOP | TA_LEFT);
	::SetROP2 (_hDC, R2_COPYPEN);
	if (m_FrameHdr.btDoors != 0)
		DrawText2 (_hDC, 5, c_nStatusLevel, "Door is open", c_colBlack, c_colWhite);

	// Show turn and safety alarm lights.
	if (m_FrameHdr.bTurnLeftSignal != 0 || m_FrameHdr.bTurnRightSignal != 0 || m_FrameHdr.bSafetyAlarm != 0) {
		::SelectObject (_hDC, (::GetTickCount () & 0x200) == 0 ? g_hbrGreen : g_hbrYellow);
		::SelectObject (_hDC, g_hpenBlackThin);
		if (m_FrameHdr.bTurnLeftSignal  != 0 || m_FrameHdr.bSafetyAlarm != 0)
			DrawArrow (_hDC, 15,           c_nStatusLevel - 10, 180, 20, 5, 10);
		if (m_FrameHdr.bTurnRightSignal != 0 || m_FrameHdr.bSafetyAlarm != 0)
			DrawArrow (_hDC, _nWidth - 15, c_nStatusLevel - 10, 0,   20, 5, 10);
	}

	// Show brake light.
	if (m_FrameHdr.bBrake != 0) {
		::SelectObject (_hDC, g_hbrRed);
		::SelectObject (_hDC, g_hpenBlackThin);
		::Rectangle (_hDC, _nWidth/2 - 20, c_nStatusLevel - 10, _nWidth/2 + 20, c_nStatusLevel);
	}

	return true;
}

void CDvrVideoRenderer::_ClearImages () {
	for (size_t cChannel = 0; cChannel < c_cDvrChannels; ++ cChannel) {
		m_cWidth  [cChannel] = 0;
		m_cHeight [cChannel] = 0;
		m_MemDC [cChannel].Resize (CClientDC (NULL), 0, 0);
	}
}

size_t CDvrVideoRenderer::_ParseImageImpl () const {
	const size_t cFrameSize = m_FrameHdr.wFrameSizeBy256*512;

	// Play sound, if any.
	if (cFrameSize - m_FrameHdr.dwHdrAndJpgSize >= DVR_WAV_LEN_min && m_hWaveOut && ! m_bMute) {
		const size_t cChunk = _GetFreeWavChunk ();
		if (cChunk != -1) {
			wav_chunk_t & chunk = m_WavChunks [cChunk];

			// Determine the size of wave chunk.
			size_t cWavSize = (cFrameSize - m_FrameHdr.dwHdrAndJpgSize)/512*512;
			if (cWavSize > DVR_WAV_LEN_max)
				cWavSize = DVR_WAV_LEN_max;
			assert (cWavSize <= DVR_WAV_LEN_max);
			assert (cWavSize >= DVR_WAV_LEN_min);

			// Estimate required playback rate.
			float fPlaybackRate = m_fSpeed;
/*			{
				::timeBeginPeriod (1);
				const DWORD dwTime = ::timeGetTime ();
				::timeEndPeriod (1);

				if (m_bPrevChunkDefined) {
					const DWORD dwElapsed = dwTime - m_dwPrevChunkTime;
					if (dwElapsed != 0)
						fPlaybackRate = fast_floor (float (m_dwAvgBytesPerSec)*dwElapsed/(1000*cWavSize) + .5f);
				}
				m_dwPrevChunkTime = dwTime;
				m_bPrevChunkDefined = true;
			}*/

			if (1 <= fPlaybackRate && fPlaybackRate < 1.5f) {
				// Read chunk data.
				_SetFilePointer (m_hFile, m_ullOffset + cFrameSize - cWavSize);
				DWORD dwRead = 0;
				::memset (chunk.baWAV, 0, sizeof (chunk.baWAV));
				::ReadFile (m_hFile, (void *) chunk.baWAV, cWavSize, & dwRead, NULL);

				// Schedule next wave chunk.
				WAVEHDR & wh = chunk.wh;
				wh.dwBufferLength = cWavSize;
				const MMRESULT mr = ::waveOutWrite (m_hWaveOut, & wh, sizeof (wh));
				if (mr == MMSYSERR_NOERROR)
					chunk.bBusy = true;
				else
					ATLTRACE ("::waveOutWrite() failed. Error code is %d\n", mr);
			}
		}

# ifdef _DEBUG
		static DWORD g_dwPrevTC;

		::timeBeginPeriod (1);
		DWORD dwTC = ::timeGetTime ();

		char buf [128];
		::sprintf (buf, "dt=%d ms  len=%d frame=%d chunk=%d.\n", dwTC - g_dwPrevTC, cFrameSize - m_FrameHdr.dwHdrAndJpgSize, m_FrameHdr.wFrameIdx, cChunk);
//		::OutputDebugString (buf);

		g_dwPrevTC = dwTC;
# endif // _DEBUG
	}

	// Read JPEG data.
	if (m_FrameHdr.dwHdrAndJpgSize <= 512) {
		m_JpegData.clear ();
		return -1;
	}
	const size_t cJpgSize = m_FrameHdr.dwHdrAndJpgSize - 512;

	if (cJpgSize > m_JpegData.size ())
		m_JpegData.resize (cJpgSize);

	_SetFilePointer (m_hFile, m_ullOffset + 512);
	DWORD dwRead = 0;
	if (! m_JpegData.empty ())
		::ReadFile (m_hFile, & * m_JpegData.begin (), cJpgSize, & dwRead, NULL);

	// Parse JPEG data.
	const size_t cBytes = m_JpegData.size ();
	if (cBytes < 2*sizeof (WORD))
		return -1;
	assert (! m_JpegData.empty ());
	const BYTE * const pJpg = & * m_JpegData.begin ();
	const WORD wJpgStartMarker = * reinterpret_cast<const WORD *> (pJpg);
	const WORD wJpgEndMarker   = * reinterpret_cast<const WORD *> (pJpg + cBytes - sizeof (WORD));
	if (wJpgStartMarker != 0xD8FF || wJpgEndMarker != 0xD9FF)
		return -1;

	CGlobalPtr hGlobal (cBytes);
	void * const pLock = hGlobal.Lock ();

	::memcpy (pLock, pJpg, cBytes);

	// Make IStream for image file.
	CComPtr<IStream> pStream;
	HRESULT hr = ::CreateStreamOnHGlobal (hGlobal.Detach (), TRUE, & pStream);
	if (FAILED (hr))
		return -1;

# ifdef USE_GDIPLUS
	if (g_bG == false) {
		g_bG = true;
		Gdiplus::GdiplusStartupInput gdiplusStartupInput;
		ULONG_PTR gdiplusToken;
		Gdiplus::GdiplusStartup (& gdiplusToken, & gdiplusStartupInput, NULL);
	}

	Gdiplus::Image im (pStream, FALSE);
	const int nWidth  = im.GetWidth  ();
	const int nHeight = im.GetHeight ();
# else // ! USE_GDIPLUS
	// Load from IStream.
	CComPtr<IPicture> pPicture;
	hr = ::OleLoadPicture (pStream, cBytes, FALSE, IID_IPicture, reinterpret_cast<void **> (& pPicture));
	if (FAILED (hr))
		return -1;

	OLE_XSIZE_HIMETRIC hmWidth;
	pPicture->get_Width (& hmWidth);
	const int nWidth  = ::MulDiv (hmWidth  + 1, m_nLogPixelX, 2540);

	OLE_YSIZE_HIMETRIC hmHeight;
	pPicture->get_Height (& hmHeight);
	const int nHeight = ::MulDiv (hmHeight + 1, m_nLogPixelY, 2540);
# endif // ! USE_GDIPLUS

	if ((int) m_cWidth_current != nWidth || (int) m_cHeight_current != nHeight)
		m_MemDC_current.Resize (CClientDC (NULL), nWidth, nHeight);

	m_cWidth_current  = nWidth;
	m_cHeight_current = nHeight;

# ifdef USE_GDIPLUS
	Gdiplus::Graphics g (m_MemDC_current);
	if (g.DrawImage (& im, 0, 0, nWidth, nHeight) != Gdiplus::Ok)
		return -1;
# else // ! USE_GDIPLUS
	hr = pPicture->Render (m_MemDC_current, 0, nHeight, nWidth, -nHeight, 0, 0, hmWidth, hmHeight, NULL);
	if (FAILED (hr))
		return -1;
# endif // ! USE_GDIPLUS

	// Get channel #.
	const int N = nWidth == 720 ? 2 : 1;
	const int F = 0;
	const COLORREF col1 = ::GetPixel (m_MemDC_current, N*(F*144 + 96      + 8), 2);
	const COLORREF col2 = ::GetPixel (m_MemDC_current, N*(F*144 + 96 + 16 + 8), 2);
	const int cc1 = GetRValue (col1) >= 128 ? 1 : 0;
	const int cc2 = GetRValue (col2) >= 128 ? 1 : 0;
	const size_t cChannel = (cc1 << 1) + cc2;
	if (cChannel >= c_cDvrChannels)
		return -1;

# ifdef _DEBUG
	// Show frame #.
	char strFrameIdx [32];
	::sprintf (strFrameIdx, "#%d, ch%d", m_FrameHdr.wFrameIdx, cChannel);
	DrawText2 (m_MemDC_current, 5, 5, strFrameIdx, c_colBlack, c_colWhite);
# endif // _DEBUG

	return cChannel;
}

size_t CDvrVideoRenderer::_ParseImage () const {
	cache_item_t * const pCI = _FindCacheItem (m_ullOffset);
	if (pCI != NULL) {
		if (m_cWidth_current != pCI->cWidth || m_cHeight_current != pCI->cHeight)
			m_MemDC_current.Resize (CClientDC (NULL), pCI->cWidth, pCI->cHeight);
		::BitBlt (m_MemDC_current, 0, 0, pCI->cWidth, pCI->cHeight, pCI->MemDC, 0, 0, SRCCOPY);
		m_cWidth_current  = pCI->cWidth;
		m_cHeight_current = pCI->cHeight;
		return pCI->cChannel;
	} else {
		const size_t cChannel = _ParseImageImpl ();
		if (cChannel == -1) {
			if (m_cWidth_current == 0 || m_cHeight_current == 0)
				return -1;
			const RECT rect = {0, 0, m_cWidth_current, m_cHeight_current};
			::FillRect (m_MemDC_current, & rect, g_hbrRed);
		}

		cache_item_t * const pCI = _FindFreeOrLRUCacheItem ();
		pCI->btTime    = m_btCacheTime ++;
		pCI->bUsed     = true;
		pCI->ullOffset = m_ullOffset;
		pCI->cChannel  = cChannel;
		if (pCI->cWidth != m_cWidth_current || pCI->cHeight != m_cHeight_current)
			pCI->MemDC.Resize (CClientDC (NULL), m_cWidth_current, m_cHeight_current);
		::BitBlt (pCI->MemDC, 0, 0, m_cWidth_current, m_cHeight_current, m_MemDC_current, 0, 0, SRCCOPY);
		pCI->cWidth    = m_cWidth_current;
		pCI->cHeight   = m_cHeight_current;

		return cChannel;
	}
}

CDvrVideoRenderer::cache_item_t * CDvrVideoRenderer::_FindCacheItem (ULONGLONG _ullOffset) const {
	for (size_t c = 0; c < c_CacheItems; ++ c) {
		cache_item_t & ci = Cache [c];
		if (! ci.bUsed)
			continue;
		if (ci.ullOffset == _ullOffset)
			return & ci;
	}
	return NULL;
}

CDvrVideoRenderer::cache_item_t * CDvrVideoRenderer::_FindFreeOrLRUCacheItem () const {
	cache_item_t * pCI = NULL;
	BYTE btMostTimeDiff = 0;
	for (size_t c = 0; c < c_CacheItems; ++ c) {
		cache_item_t & ci = Cache [c];
		if (! ci.bUsed)
			return & ci;

		const BYTE btTimeDiff = m_btCacheTime - ci.btTime;
		if (btMostTimeDiff <= btTimeDiff) {
			btMostTimeDiff = btTimeDiff;
			pCI = & ci;
		}
	}
	return pCI;
}

void CDvrVideoRenderer::Mute (bool _bMute) {
	m_bMute = _bMute;
}

double CDvrVideoRenderer::GetSoundLevel () {
	const size_t cFrameSize = m_FrameHdr.wFrameSizeBy256*512;

	// Play sound, if any.
	if (cFrameSize - m_FrameHdr.dwHdrAndJpgSize >= DVR_WAV_LEN_min) {
		size_t cWavSize = (cFrameSize - m_FrameHdr.dwHdrAndJpgSize)/512*512;
		if (cWavSize > DVR_WAV_LEN_max)
			cWavSize = DVR_WAV_LEN_max;
		assert (cWavSize <= DVR_WAV_LEN_max);
		assert (cWavSize >= DVR_WAV_LEN_min);

		// Read chunk data.
		_SetFilePointer (m_hFile, m_ullOffset + cFrameSize - cWavSize);

		short * const pChunk = reinterpret_cast<short *> (alloca (cWavSize));
		::memset (pChunk, 0, cWavSize);
		DWORD dwRead = 0;
		::ReadFile (m_hFile, pChunk, cWavSize, & dwRead, NULL);

		// Calculate average energy.
		double fEnergy = 0;
		for (size_t c = 0; c < cWavSize/2; ++ c)
			fEnergy += float (pChunk [c])*pChunk [c];
		fEnergy /= cWavSize/2;

		m_fPrevEnergy = fEnergy;

		return fEnergy;
	} else {
		// NOTE: every second frame has no wave data. Let's return the previous result.
		double fEnergy = m_fPrevEnergy;
		m_fPrevEnergy = 0;
		return fEnergy;
	}
}

///////////////////////////////////////////////////////////////////////

# ifdef _MSC_VER
#	pragma pack (push, 1)
# endif

struct CDvrDigestRecord {
	ULONGLONG ullFrameHdrOffset;
	float x, y;
	DATE dtTime;
	float fSpeed;
	DWORD dwReserved0;
	DWORD dwReserved1;
};

# ifdef _MSC_VER
#	pragma pack (pop)
# endif

///////////////////////////////////////////////////////////////////////

class CDvrLoader : public CMapLoader {
  public:
	// Override.
	virtual bool IsValid (const char * _pData, size_t _cFileSize) const;
	virtual LoadStatus_t Load (const char * _pData, size_t _cFileSize, const char * _strFileName, map_t * _pMap);
};

# if defined (_DEBUG) || defined (GME_AS)
DECLARE_WPTRK_LOADER (CDvrLoader, "j16-dvr", "dvr", "J16 DVR movie/track (*.dvr)", ftiTracks | ftiMapFileByParts | ftiLoadOver4GB);
# endif // _DEBUG

bool CDvrLoader::IsValid (const char * _pData, size_t _cFileSize) const {
	if (_cFileSize < sizeof (DWORD))
		return false;

	const size_t cScanSize = __min (_cFileSize, 0x10000);
	for (size_t c = 0; c < cScanSize; c += 256) {
		const DWORD dw = * reinterpret_cast<const DWORD *> (_pData + c);
		if (dw == c_dwDvrFrameMarker)
			return true;
	}
	return false;
}

LoadStatus_t CDvrLoader::Load (const char * _pData, size_t _cFileSize, const char * _strFileName, map_t * _pMap) {
	SetProgressStatus (0);

	const ULONGLONG ullFileSize = GetFileSize ();

	// Report vehicle ID.
	DWORD dwVehicleID = 0;
	{
		const CDvrFrameHdr * pFrameHdr = NULL;
		ULONGLONG ullOffset = 0;
		while (ullOffset < ullFileSize) {
			const ULONGLONG ullHdrEndOffset = __min (ullOffset + 512, ullFileSize);
			pFrameHdr = reinterpret_cast<const CDvrFrameHdr *> (GetFilePart (ullOffset, ullHdrEndOffset));
			if (pFrameHdr == NULL)
				return lsFailed;
			if (pFrameHdr->dwStartMarker == c_dwDvrFrameMarker)
				break;

			ullOffset += 256;
		}

		if (pFrameHdr) {
			dwVehicleID = pFrameHdr->wVehicleID;
			ReportText ("* Vehicle ID:\t\t%d.", pFrameHdr->wVehicleID);
		}
	}

	// Get time zone.
	TIME_ZONE_INFORMATION tzi;
	if (::GetTimeZoneInformation (& tzi) == TIME_ZONE_ID_DAYLIGHT)
		tzi.Bias += tzi.DaylightBias;

	size_t cDeclaredFPS = 0;
	size_t cActualFPS = 0;

	std::vector<CDvrDigestRecord> Digest;
	std::vector<DWORD> Tracks;

	// Prepare name of the digest.
	char strDigestName [MAX_PATH];
	::strcpy (strDigestName, _strFileName);
	::strcat (strDigestName, ".digest");

	bool bCreateDigest = true;
	if (DoesFileExist (strDigestName)) {
		// Read the digest file.
		DATE dt0 = 0;
		{
			CFileForRead hDigest (strDigestName);
			if (! hDigest.IsValid ()) {
				ReportWarning ("Failed to read digest file.");
				return lsFailed;
			}

			DWORD dwBytesRead = 0;
			::ReadFile (hDigest, & dt0, sizeof (dt0), & dwBytesRead, NULL);

			DWORD dwRealFPS = 0;
			::ReadFile (hDigest, & dwRealFPS, sizeof (dwRealFPS), & dwBytesRead, NULL);
			cActualFPS = dwRealFPS;

			DWORD dwTracks = 0;
			::ReadFile (hDigest, & dwTracks, sizeof (dwTracks), & dwBytesRead, NULL);
			Tracks.resize (dwTracks);
			if (dwTracks != 0)
				::ReadFile (hDigest, & * Tracks.begin (), dwTracks*sizeof (DWORD), & dwBytesRead, NULL);

			DWORD dwDigestRecords = 0;
			::ReadFile (hDigest, & dwDigestRecords, sizeof (dwDigestRecords), & dwBytesRead, NULL);
			Digest.resize (dwDigestRecords);
			if (dwDigestRecords != 0)
				::ReadFile (hDigest, & * Digest.begin (), Digest.size ()*sizeof (CDvrDigestRecord), & dwBytesRead, NULL);
		}

		// Scan for first frame.
		const CDvrFrameHdr * pFrameHdr = NULL;
		ULONGLONG ullOffset = 0;
		while (ullOffset < ullFileSize) {
			const ULONGLONG ullHdrEndOffset = __min (ullOffset + 512, ullFileSize);
			pFrameHdr = reinterpret_cast<const CDvrFrameHdr *> (GetFilePart (ullOffset, ullHdrEndOffset));
			if (pFrameHdr == NULL)
				return lsFailed;
			if (pFrameHdr->dwStartMarker == c_dwDvrFrameMarker)
				break;

			ullOffset += 256;
		}

		// Verify the digest.
		trackpoint_info_t point_extra;
		_ParseDvrHdr (pFrameHdr, point_extra);
		cDeclaredFPS = _GetDvrFPS (pFrameHdr);

		if (point_extra.dtTime != dt0) {
			::DeleteFile (strDigestName);

			Digest.clear ();
			Tracks.clear ();
			cDeclaredFPS = 0;
			cActualFPS = 0;

			bCreateDigest = true;
		} else
			bCreateDigest = false;
	}

	if (bCreateDigest) {
		//
		// Parse original DVR file.
		//

		DATE dt0 = 0;
		bool bTimeDefined = false;

		size_t cFrame = 0;
		size_t cFirstFrameInTrack = 0;
		size_t cLastFrameInTrack = 0;

		float fCosY = 1;

		WORD wPrevFrameIdx = 0;

		ULONGLONG ullOffset = 0;
		do {
			if (g_bStop)
				return lsInterrupted;

			// Scan for next frame.
			const CDvrFrameHdr * pFrameHdr = NULL;
			ullOffset &= 0xFFFFFFFFFFFFFF00;
			bool bScanned = false;
			while (ullOffset < ullFileSize) {
				const ULONGLONG ullHdrEndOffset = __min (ullOffset + 512, ullFileSize);
				pFrameHdr = reinterpret_cast<const CDvrFrameHdr *> (GetFilePart (ullOffset, ullHdrEndOffset));
				if (pFrameHdr == NULL)
					return lsFailed;
				if (pFrameHdr->dwStartMarker == c_dwDvrFrameMarker)
					break;

				if (! bScanned) {
					bScanned = true;
					ReportWarning ("Incorrect frame at offset %08x.%08x.", DWORD (ullOffset >> 32), (DWORD) ullOffset);
				}

				ullOffset += 256;
			}
			if (ullOffset >= ullFileSize)
				break;

			if (! bTimeDefined) {
				// Save start time.
				trackpoint_info_t point_extra;
				_ParseDvrHdr (pFrameHdr, point_extra);

				dt0 = point_extra.dtTime;
				bTimeDefined = true;
			}

			const bool bNewTrack = pFrameHdr->wFrameIdx == 1 && wPrevFrameIdx != 0;
			wPrevFrameIdx = pFrameHdr->wFrameIdx;

			// Mark next track, if any.
			if (bNewTrack || bScanned) {
				const size_t cDigestRecords = Digest.size ();
				if (cDigestRecords != 0 && (Tracks.empty () || Tracks.back () != cDigestRecords))
					Tracks.push_back (cDigestRecords);

				if (cActualFPS == 0 && cDigestRecords != 0) {
					const DATE dtStart = Digest [0                 ].dtTime;
					const DATE dtEnd   = Digest [cDigestRecords - 1].dtTime;
					const double fDurationInS = (dtEnd - dtStart)*24*60*60;
					if (fDurationInS > 0)
						cActualFPS = fast_floor_d ((cLastFrameInTrack - cFirstFrameInTrack)/fDurationInS + .5f);
				}
			}

			if (cDeclaredFPS == 0)
				cDeclaredFPS = _GetDvrFPS (pFrameHdr);

			point_t pt;
			if (_ParseDvrHdr (pFrameHdr, pt)) {
				if (Digest.empty ())
					fCosY = ::cos (pt.y*c_PI/180);

				if (Digest.empty () || pt.x != Digest.back ().x || pt.y != Digest.back ().y) {
					trackpoint_info_t point_extra;
					_ParseDvrHdr (pFrameHdr, point_extra);

					bool bPosOK = true;
					if (! Digest.empty ()) {
						// Check for jumping to incredibly large distance.
						const CDvrDigestRecord & dr_prev = Digest.back ();
						float fDiffInS = ::fabs (point_extra.dtTime - dr_prev.dtTime)*24*60*60;
						if (fDiffInS == 0)
							fDiffInS = 1;
						const float d = ::fabs (pt.x - dr_prev.x)*fCosY + ::fabs (pt.y - dr_prev.y);
						const float fDInM = d*40e6/360;
						const float fSpeed = fDInM/fDiffInS;
						if (fSpeed > 1000)
							bPosOK = false;
					}

					if (bPosOK) {
						if (cActualFPS == 0 && ! Digest.empty ()) {
							const float fDiffInS = ::fabs (point_extra.dtTime - Digest.back ().dtTime)*24*60*60;
							if (fDiffInS > 3) {
								const DATE dtStart = Digest [0                 ].dtTime;
								const DATE dtEnd   = Digest [Digest.size () - 1].dtTime;
								const double fDurationInS = (dtEnd - dtStart)*24*60*60;
								if (fDurationInS > 0)
									cActualFPS = fast_floor_d ((cLastFrameInTrack - cFirstFrameInTrack)/fDurationInS + .5f);
							}
						}

						if (Digest.empty ())
							cFirstFrameInTrack = cFrame;
						cLastFrameInTrack = cFrame;

						Digest.push_back (CDvrDigestRecord ());
						CDvrDigestRecord & dr = Digest.back ();
						::memset (& dr, 0, sizeof (CDvrDigestRecord));
						dr.ullFrameHdrOffset = ullOffset;
						dr.x                 = pt.x;
						dr.y                 = pt.y;
						dr.dtTime            = point_extra.dtTime;
						dr.fSpeed            = point_extra.fSpeed;
					}
				}
			}

			const size_t cFrameSize = pFrameHdr->wFrameSizeBy256*512;
//			assert (cFrameSize < 65536);
			ullOffset += cFrameSize;

			++ cFrame;

			SetProgressStatus (50*ullOffset/ullFileSize);
		} while (ullOffset < ullFileSize);

		// Complete last track.
		{
			const size_t cDigestRecords = Digest.size ();
			if (cDigestRecords != 0 && (Tracks.empty () || Tracks.back () != cDigestRecords))
				Tracks.push_back (cDigestRecords);

			if (cActualFPS == 0 && cDigestRecords != 0) {
				const DATE dtStart = Digest [0                 ].dtTime;
				const DATE dtEnd   = Digest [cDigestRecords - 1].dtTime;
				const double fDurationInS = (dtEnd - dtStart)*24*60*60;
				if (fDurationInS > 0)
					cActualFPS = fast_floor_d ((cLastFrameInTrack - cFirstFrameInTrack)/fDurationInS + .5f);
			}
		}

		//
		// Create the digest file.
		//
		CFileForWrite hDigest (strDigestName);
		if (! hDigest.IsValid ()) {
			ReportWarning ("Failed to create digest file.");
		} else {
			CFileWriter fw (hDigest);
			fw.WriteBin (dt0);
			fw.WriteBin ((DWORD) cActualFPS);
			fw.WriteBin ((DWORD) Tracks.size ());
			if (! Tracks.empty ())
				fw.WriteBin (& * Tracks.begin (), Tracks.size ()*sizeof (DWORD));
			fw.WriteBin ((DWORD) Digest.size ());
			if (! Digest.empty ())
				fw.WriteBin (& * Digest.begin (), Digest.size ()*sizeof (CDvrDigestRecord));
		}
	}

	// Restore headings.
	CDvrVideoRenderer::pos2heading_t pos2heading;
	const size_t cDigestRecords = Digest.size ();
	for (size_t c = 1; c + 1 < cDigestRecords; ++ c) {
		const CDvrDigestRecord & dr0 = Digest [c - 1];
		const CDvrDigestRecord & dr1 = Digest [c];
		const CDvrDigestRecord & dr2 = Digest [c + 1];

		if (dr2.x == dr0.x && dr2.y - dr0.y)
			continue;
		float fHeading = 90 - 180.f/c_PI*::atan2 (dr2.y - dr0.y, (dr2.x - dr0.x)*::cos (Digest [0].y*c_PI/180.f));
		if (fHeading < 0)
			fHeading += 360;
		pos2heading [dr1.ullFrameHdrOffset] = fHeading;
	}

	SetProgressStatus (50);

	ReportText ("* Declared frame rate:\t%d fps", cDeclaredFPS);
	ReportText ("* Actual frame rate:\t%d fps",   cActualFPS);

	file_refs_t file_refs;
	file_refs.push_back (file_ref_t ());
	file_refs.back ().strFileName = _strFileName;
	file_refs.back ().bAttachment = true;
	file_refs.back ().pTrackVideo = new CDvrVideoRenderer (_strFileName, cActualFPS ? cActualFPS : cDeclaredFPS, pos2heading, dwVehicleID);

	// Prepare base file name.
	const char * const pExt  = ::PathFindExtension (_strFileName);
	const char * const pName = ::PathFindFileName  (_strFileName);
	const string_t strBaseTrackName (pName, pExt);

	const size_t cTracks = Tracks.size ();
	for (size_t cTrack = 0; cTrack < cTracks; ++ cTrack) {
		const size_t cBegin = cTrack == 0 ? 0 : Tracks [cTrack - 1];
		const size_t cEnd = Tracks [cTrack];

		const size_t cPoints = cEnd - cBegin;
		if (cPoints == 0)
			continue;

		points_t points (cPoints);
		trackpoints_t points_extra (cPoints);
		size_t c;
		for (c = 0; c < cPoints; ++ c) {
			const CDvrDigestRecord & dr = Digest [cBegin + c];

			point_t & pt = points [c];
			pt.x = dr.x;
			pt.y = dr.y;

			trackpoint_info_t & point_extra = points_extra [c];
			point_extra.dtTime      = dr.dtTime + float (tzi.Bias)/(24*60);
			point_extra.fSpeed      = dr.fSpeed;
			point_extra.cVideoData0 = dr.ullFrameHdrOffset;
			point_extra.cVideoData1 = dr.ullFrameHdrOffset >> 32;
			point_extra.wFlags      = trackpoint_info_t::defTime | trackpoint_info_t::defSpeed | trackpoint_info_t::defVideo;
		}

		// Search for 'chinese' steps in the track.
		std::vector<size_t> steps;
		for (c = 1; c + 1 < cPoints; ++ c) {
			const point_t & pt0 = points [c - 1];
			const point_t & pt1 = points [c];
			const point_t & pt2 = points [c + 1];

			bool bIsStep = false;
			if (pt1.x == pt0.x && pt1.y == pt2.y)
				bIsStep = true;
			if (pt1.y == pt0.y && pt1.x == pt2.x)
				bIsStep = true;

			if (bIsStep) {
				steps.push_back (c);
				++ c;
			}
		}

		// Remove the steps.
		const size_t cSteps = steps.size ();
		if (cSteps != 0) {
			for (size_t c = 0; c < cSteps; ++ c) {
				const size_t cStep = steps [cSteps - 1 - c];
				points      .erase (points      .begin () + cStep);
				points_extra.erase (points_extra.begin () + cStep);
			}
		}


		//
		// Create track.
		//

		rgn_t & rgn = _pMap->CreateRgn (rkPolyline);
		rgn.SetTrackType ();

		rgn.pFileRef  = & file_refs.back ();
		rgn.cBeginPos = Digest [cBegin].ullFrameHdrOffset;
		rgn.cEndPos   = cEnd < Digest.size () ? Digest [cEnd].ullFrameHdrOffset : _cFileSize;
		rgn.cFileLine = -1;
	//	rgn.cOrdinal = m_cOrdinal ++;

		if (cTrack == 0)
			rgn.strLabel = strBaseTrackName;
		else {
			char strLabel [MAX_PATH];
			::sprintf (strLabel, "%s#%02d", strBaseTrackName.c_str (), cTrack);
			rgn.strLabel = strLabel;
		}

		rgn.elements.push_back (rgn_element_t ());
		rgn_element_t & element = rgn.elements.back ();
		element.cFileOffset = rgn.cBeginPos;

		//
		element.cLevel = 0;
		element.points.swap (points);

		// Extra properties.
		track_info_t * const pTrackExtra = rgn.pExtra->wtr.CreateTrack ();
		element.pExtra->TrackPoints.swap (points_extra);
	}

	_pMap->file_refs.swap (file_refs);

	// Suppress warning about empty Zoom etc.
	_pMap->bNoHeader  = true;
	_pMap->cHeaderLen = 0;

	return lsOK;
}
