//
// GPSMapEdit
// (c) Konstantin Galichsky (kg@geopainting.com), 2007-2008
//
// Highlight routing nodes by color.
//

# include "StdAfx.h"
# include "Resource.h"
# include "DrawNodes.h"
# include "Map.h"
# include "Colors.h"
# include "Globals.h"

static CColors * g_pNodesColors;

static const COLORREF g_colNodes [] = {
	c_colBlack,
	c_colRed,
	c_colRed,
	c_colBlue,
	c_colWhite,
	c_colYellow,
	c_colGreen,
	c_colMagenta
};

void InitDrawNodes () {
	assert (g_pNodesColors == NULL);
	g_pNodesColors = new CColors;
	g_pNodesColors->LoadFromSetting ("RoutingNodesColors", g_colNodes, ARRAY_SIZE (g_colNodes));
}

void DestroyDrawNodes () {
	delete g_pNodesColors;
	g_pNodesColors = NULL;
}

bool AdjustColorsForRoutingNodes (HWND _hWnd) {
	assert (g_pNodesColors != NULL);

	const size_t cColors = g_pNodesColors->GetColors ();
	const char ** const strColorNames = (const char **) alloca (sizeof (char *)*cColors);

	fstring str [8];
	assert (cColors == ARRAY_SIZE (str));
	str [0].LoadFromResource (IDS_RNC_NO_RESTRICTIONS);
    str [1].LoadFromResource (IDS_RNC_HAS_RESTRICTIONS);
    str [2].LoadFromResource (IDS_RNC_MISPOSITION);
    str [3].LoadFromResource (IDS_RNC_CARDINALITY_1);
    str [4].LoadFromResource (IDS_RNC_CARDINALITY_2);
    str [5].LoadFromResource (IDS_RNC_CARDINALITY_3);
    str [6].LoadFromResource (IDS_RNC_CARDINALITY_4);
    str [7].LoadFromResource (IDS_RNC_CARDINALITY_5);

	for (size_t c = 0; c < __min (cColors, ARRAY_SIZE (str)); ++ c)
		strColorNames [c] = str [c].c_str ();

	if (g_pNodesColors->Adjust (_hWnd, g_colNodes, strColorNames, cColors)) {
		g_pNodesColors->SaveToSetting ("RoutingNodesColors");
		return true;
	}
	return false;
}

HBRUSH NodeCardinalityToBrush (size_t _cCardinality) {
	assert (g_pNodesColors != NULL);

	return g_pNodesColors->GetBrush (__min (_cCardinality, 5) + 2);
}

HPEN NodeContourPen (bool _bHasRestrictions) {
	return g_pNodesColors->GetPen (_bHasRestrictions ? 1 : 0);
}

/////////////////////////////////////////////////////

# include "Tool.h"
# include "Position.h"
# include "DrawHelpers.h"

static
long _Point2Long (int _x, int _y) {
	return (_x & 0xFFFF) + (_y << 16);
}

struct point_info_t {
	long l;
	const rgn_t * pRgn;

	bool operator < (const point_info_t & _other) const {return l < _other.l;}
};

struct cell_t {
	std::vector<point_info_t> pts;
	bool bSorted;
};
typedef std::vector<cell_t> cells_t;
static cells_t g_cells;

enum {
	CELL_SIZE  = 20,
	CELL_LIMIT = 256
};

void ResetDrawnNodes () {
	const size_t cW = (g_lWidth  + CELL_SIZE - 1)/CELL_SIZE + 1;
	const size_t cH = (g_lHeight + CELL_SIZE - 1)/CELL_SIZE + 1;
	const size_t cN = cW*cH;

	g_cells.resize (cN);

	for (size_t c = 0; c < cN; ++ c) {
		cell_t & cell = g_cells [c];
		cell.pts.clear ();
		cell.bSorted = false;
	}
}

void SaveDrawnNodes (const POINT * _pPoints, size_t _cPoints, const rgn_t * _pRgn) {
	const Tool_t Tool = CTool::GetCurrentToolId ();
	if (Tool != toolEditNodes && Tool != toolCreate)
		return;

	const size_t cW = (g_lWidth + CELL_SIZE - 1)/CELL_SIZE + 1;
	const size_t cN = g_cells.size ();

	for (size_t c = 0; c < _cPoints; ++ c) {
		const POINT & p = _pPoints [c];
		if (p.x < 0 || p.x >= g_lWidth || p.y < 0 || p.y >= g_lHeight)
			continue;
		const size_t cX = (p.x + CELL_SIZE - 1)/CELL_SIZE;
		const size_t cY = (p.y + CELL_SIZE - 1)/CELL_SIZE;
		const size_t cIdx = cX + cY*cW;
		assert (cIdx < cN);
		if (cIdx >= cN)
			continue;

		cell_t & cell = g_cells [cIdx];
		if (cell.pts.size () >= CELL_LIMIT)
			continue;
		cell.pts.push_back (point_info_t ());
		point_info_t & pi = cell.pts.back ();
		pi.l    = (p.x & 0xFFFF) + (p.y << 16);
		pi.pRgn = _pRgn;
	}
}

void HighlightNearestDrawnNodes (HDC _hDC, int _x, int _y, int _d, const rgn_t * _pExceptRgn) {
	::SelectObject (_hDC, g_hpenGrayThin);
	::SetROP2      (_hDC, R2_XORPEN);

	const int nW = (g_lWidth  + CELL_SIZE - 1)/CELL_SIZE + 1;
	const int nH = (g_lHeight + CELL_SIZE - 1)/CELL_SIZE + 1;

	const int nX0 = __max (0,      (_x - _d + CELL_SIZE - 1)/CELL_SIZE);
	const int nX1 = __min (nW - 1, (_x + _d + CELL_SIZE - 1)/CELL_SIZE);
	const int nY0 = __max (0,      (_y - _d + CELL_SIZE - 1)/CELL_SIZE);
	const int nY1 = __min (nH - 1, (_y + _d + CELL_SIZE - 1)/CELL_SIZE);

	for (int nY = nY0; nY <= nY1; ++ nY) {
		for (int nX = nX0; nX <= nX1; ++ nX) {
			const size_t cIdx = nX + nY*nW;
			assert (cIdx < (size_t) nW*nH);
			if (cIdx >= (size_t) nW*nH)
				continue;
			cell_t & cell = g_cells [cIdx];

			if (! cell.bSorted) {
				std::sort (cell.pts.begin (), cell.pts.end ());
				cell.bSorted = true;
			}

			long lPrev = -1;

			const size_t cPoints = cell.pts.size ();
			for (size_t c = 0; c < cPoints; ++ c) {
				const point_info_t & pi = cell.pts [c];
				if (pi.pRgn == _pExceptRgn)
					continue;
				// NOTE: avoid drawing points with the same position.
				if (pi.l == lPrev)
					continue;
				lPrev = pi.l;

				const int x = pi.l & 0xFFFF;
				const int y = pi.l >> 16;

				const long dist2 = (x - _x)*(x - _x) + (y - _y)*(y - _y);
				if (dist2 > _d*_d)
					continue;

				DrawCross (_hDC, x, y, pi.pRgn->IsPolyline () ? 3 : 2);
			}
		}
	}
}

