//
// GPSMapEdit
// (c) Konstantin Galichsky (kg@geopainting.com), 2002-2009
//
// Dialog for editing rectangle coordinates.
//

# include "StdAfx.h"
# include "Resource.h"
# include "EditRect.h"
# include "PtAlgo.h"
# include "EditHelpers.h"
# include "PrintUnits.h"
# include "ParsePosition.h"
# include "Dialog.h"
# include "Clipboard.h"
# include "Globals.h"

class CEditRect : public CDialog {
  public:
	CEditRect (rect_t & _rect) : CDialog (IDD_EDIT_RECT), m_rect (_rect), m_hwndNextCV (NULL) {}

	// Override.
	virtual bool OnInit (HWND _hwndFocus);
	virtual void OnDestroy ();
	virtual void OnCommand (WORD _wCmd, WORD _wNotifyCode, HWND _hwndCtl);
	virtual INT_PTR OnMessage (UINT _uMsg, WPARAM _wParam, LPARAM _lParam);

  private:
	rect_t & m_rect;
	HWND m_hwndNextCV;

	void _UpdateData ();

	void _Copy () const;
	void _Paste ();
	void _UpdatePaste ();

	static bool _ParseRect (const char * _strRect, const char * _strEnd, rect_t & _rect);
	static bool _DoPaste (rect_t & _rect);
};

void CEditRect::_UpdateData () {
	char buf [20];
	PrintLon (buf, m_rect.x0);
	SetDlgItemText (IDC_LEFT,   buf);
	PrintLon (buf, m_rect.x1);
	SetDlgItemText (IDC_RIGHT,  buf);
	PrintLat (buf, m_rect.y0);
	SetDlgItemText (IDC_BOTTOM, buf);
	PrintLat (buf, m_rect.y1);
	SetDlgItemText (IDC_TOP,    buf);
}

bool CEditRect::OnInit (HWND _hwndFocus) {
	_UpdateData ();

	EnableDlgItem (IDOK, false);

	_UpdatePaste ();
	m_hwndNextCV = ::SetClipboardViewer (m_hWnd);

	return true;
}

void CEditRect::OnDestroy () {
	::ChangeClipboardChain (m_hWnd, m_hwndNextCV);
}

void CEditRect::OnCommand (WORD _wCmd, WORD _wNotifyCode, HWND _hwndCtl) {
	switch (_wCmd) {
		case IDC_LEFT:
		case IDC_TOP:
		case IDC_RIGHT:
		case IDC_BOTTOM: {
			if (_wNotifyCode == EN_CHANGE) {
				if (
					ValidateAsArc (GetDlgItem (IDC_LEFT))  &&
					ValidateAsArc (GetDlgItem (IDC_TOP))   &&
					ValidateAsArc (GetDlgItem (IDC_RIGHT)) &&
					ValidateAsArc (GetDlgItem (IDC_BOTTOM)) &&
					GetTextAsArc (GetDlgItem (IDC_LEFT))   < GetTextAsArc (GetDlgItem (IDC_RIGHT)) &&
					GetTextAsArc (GetDlgItem (IDC_BOTTOM)) < GetTextAsArc (GetDlgItem (IDC_TOP))
				) {
					EnableDlgItem (IDOK, true);
				} else
					EnableDlgItem (IDOK, false);
			}
			break;
		}

		case IDC_COPY:
			_Copy ();
			break;

		case IDC_PASTE:
			_Paste ();
			break;

		case IDOK: {
			m_rect.x0 = GetTextAsArc (GetDlgItem (IDC_LEFT));
			m_rect.x1 = GetTextAsArc (GetDlgItem (IDC_RIGHT));
			m_rect.y0 = GetTextAsArc (GetDlgItem (IDC_BOTTOM));
			m_rect.y1 = GetTextAsArc (GetDlgItem (IDC_TOP));

			EndDialog (0);

			break;
		}

		case IDCANCEL:
			EndDialog (1);
			break;
	}
}

INT_PTR CEditRect::OnMessage (UINT _uMsg, WPARAM _wParam, LPARAM _lParam) {
	switch (_uMsg) {
		case WM_CHANGECBCHAIN: {
			const HWND hwndRemoved = reinterpret_cast<HWND> (_wParam);
			const HWND hwndNext    = reinterpret_cast<HWND> (_lParam);
			if (hwndRemoved == m_hwndNextCV)
				m_hwndNextCV = hwndNext;
			if (m_hwndNextCV != NULL)
				::SendMessage (m_hwndNextCV, WM_CHANGECBCHAIN, _wParam, _lParam);
			break;
		}

		case WM_DRAWCLIPBOARD:
			_UpdatePaste ();
			if (m_hwndNextCV != NULL)
				::SendMessage (m_hwndNextCV, WM_DRAWCLIPBOARD, _wParam, _lParam);
			break;
	}

	return CDialog::OnMessage (_uMsg, _wParam, _lParam);
}

void CEditRect::_Copy () const {
	CClipboard cb;
	::EmptyClipboard ();

	enum {BUF_SIZE = 20};

	const HGLOBAL hGlobal = ::GlobalAlloc (GMEM_MOVEABLE, 4*BUF_SIZE);
	char * const pData = reinterpret_cast<char *> (::GlobalLock (hGlobal));
	if (! pData)
		return;

	char bufX0 [BUF_SIZE], bufX1 [BUF_SIZE], bufY0 [BUF_SIZE], bufY1 [BUF_SIZE];
	PrintLon (bufX0, m_rect.x0);
	PrintLon (bufX1, m_rect.x1);
	PrintLat (bufY0, m_rect.y0);
	PrintLat (bufY1, m_rect.y1);

	::strcpy (pData, bufY0);
	::strcat (pData, " ");
	::strcat (pData, bufX0);
	::strcat (pData, " ");
	::strcat (pData, bufY1);
	::strcat (pData, " ");
	::strcat (pData, bufX1);

	::GlobalUnlock (hGlobal);
	::SetClipboardData (CF_TEXT, hGlobal);
}

void CEditRect::_Paste () {
	_DoPaste (m_rect);
	_UpdateData ();
	EnableDlgItem (IDOK, true);
}

bool CEditRect::_ParseRect (const char * _strRect, const char * _strEnd, rect_t & _rect) {
	const char * p = _strRect;
	point_t pt0;
	if (! ParsePosition (pt0, p, _strEnd, & p))
		return false;
	point_t pt1;
	if (! ParsePosition (pt1, p, _strEnd, & p))
		return false;

	_rect = rect_t ();
	_rect.Extend (pt0);
	_rect.Extend (pt1);

	return true;
}

bool CEditRect::_DoPaste (rect_t & _rect) {
	CClipboard cb;

	const HGLOBAL hGlobal = ::GetClipboardData (CF_TEXT);
	if (! hGlobal)
		return false;
	const char * const pData = reinterpret_cast<char *> (::GlobalLock (hGlobal));
	const size_t cSize = ::GlobalSize (hGlobal);

	const bool bOK = _ParseRect (pData, pData + cSize, _rect);

	::GlobalUnlock (hGlobal);

	return bOK;
}

void CEditRect::_UpdatePaste () {
	if (::IsClipboardFormatAvailable (CF_TEXT)) {
		rect_t rect;
		const bool bOK = _DoPaste (rect);
		EnableDlgItem (IDC_PASTE, bOK);
	} else
		EnableDlgItem (IDC_PASTE, false);
}

bool EditRect (HWND _hWnd, rect_t & _rect) {
	CEditRect erd (_rect);
	return erd.ShowModal (_hWnd) == 0;
}