//
// GPSMapEdit
// (c) Konstantin Galichsky (kg@geopainting.com), 2008-2009
//
// 'File | Export' support.
//

# include "StdAfx.h"
# include "Resource.h"
# include "FileTypeInfo.h"
# include "Map.h"
# include "Save.h"
# include "Load.h"
# include "Menu.h"
# include "Log.h"
# include "Status.h"
# include "SaveFileDlg.h"
# include "Threads.h"
# include "EHFilter.h"
# include "Wrappers.h"
# include "Globals.h"

static
SaveStatus_t _ExecuteSaveFileSafely2 (CMapSaver * _pSaver, HANDLE _hFile, const char * _strFileName, const map_t & _map) {
	try {
		return _pSaver->Save (_hFile, _strFileName, _map);
	} catch (const COutOfMemory & _oom) {
		_oom.ReportToMessageLog ();
		return ssFailed;
	} catch (const std::exception & _e) {
		ReportText ("Unhandled exception: %s.", _e.what ());
		return ssFailed;
	}
}

SaveStatus_t ExecuteSaveFileSafely (CMapSaver * _pSaver, HANDLE _hFile, const char * _strFileName, const map_t & _map) {
	__try {
		return _ExecuteSaveFileSafely2 (_pSaver, _hFile, _strFileName, _map);
	} __except (EHFilter (_exception_info ())) {
		return ssFailed;
	}
}

# if defined (_DEBUG)
# define _SYNC_EXPORT // Enable one-thread exporting in debug build.
# endif

static HANDLE g_hExportThread;

struct export_data_t {
	const CFileTypeInfo * pFTI;
	string_t strFileName;
	bool bOpenFile;
};

static
DWORD WINAPI _ExportMapProc (LPVOID _pParams) {
	const export_data_t * const pParams = reinterpret_cast<export_data_t *> (_pParams);
	const CFileTypeInfo * const pFTI = pParams->pFTI;
	const string_t strFileName = pParams->strFileName;
	const bool bOpenFile = pParams->bOpenFile;
	delete pParams;

	CCom com;

	pFTI->_ExportFileImpl (strFileName.c_str (), bOpenFile);

	g_bLoading = false;
	UpdateMenu ();

	return 0;
}

void WaitExportThread () {
# if ! defined (_SYNC_EXPORT)
	if (g_hExportThread == NULL)
		return;

	WaitThreadForStop (g_hExportThread);

	::CloseHandle (g_hExportThread);
	g_hExportThread = NULL;
# endif // _SYNC_EXPORT
}

void StopExportThread () {
	g_bStop = true;

	WaitExportThread ();
}

void CFileTypeInfo::ExportFile (const char * _strFileName) const {
	char strFileName [MAX_PATH + 1] = "";
	const bool bShowSaveAs = (m_dwFlags & ftiShowSaveAs) != 0;
	if (bShowSaveAs) {
		if (_strFileName && _strFileName [0] != '\0') {
			::strcpy (strFileName, _strFileName);
		} else {
			// Append additional '\0' to the file filter string.
			char * const strFilter = (char *) ::alloca (sizeof (char)*(m_strFilter.size () + 2));
			::memcpy (strFilter, m_strFilter.c_str (), sizeof (char)*(m_strFilter.size () + 1));
			strFilter [m_strFilter.size () + 1] = '\0';

			::strncpy (strFileName, g_strFileName, MAX_PATH);
			strFileName [MAX_PATH] = '\0';

			assert (m_vecExts.size () == 1);
			if (! SaveFileDlg (m_vecExts [0].c_str (), strFilter, strFileName, "Export"))
				return;
		}
	}
	const bool bCopyMap = (m_dwFlags & ftiSaveCopyOfMap) != 0;

	// Wait for previous parsing ends.
	if (g_hExportThread)
		StopExportThread ();

	g_bStop = false;

	g_bLoading = true;
	UpdateMenu ();

	export_data_t * const pParams = new export_data_t;
	pParams->pFTI        = this;
	pParams->strFileName = strFileName;
	pParams->bOpenFile   = bShowSaveAs; 

# if defined (_SYNC_EXPORT)
	_ExportMapProc (pParams);
# else
	if (bCopyMap) {
		// Perform the exporting in parallel thread.
		DWORD dwThreadId = 0;
		assert (! g_hExportThread);
		g_hExportThread = ::CreateThread (NULL, 0, & _ExportMapProc, pParams, 0, & dwThreadId);
		if (! g_hExportThread) {
			ErrorMsgBox ("Internal error.");
			return;
		}
		::SetThreadPriority (g_hExportThread, THREAD_PRIORITY_BELOW_NORMAL);
		// NOTE: the thread's handle 'g_hExportThread' is not closed until StopExportThread() or next exporting.
	} else {
		_ExportMapProc (pParams);
	}
# endif
}

void CFileTypeInfo::_ExportFileImpl (const char * _strFileName, bool _bOpenFile) const {
	CMapSaver * const pSaver = m_pfCreateSaver ();
	const CObjectHolder<CMapSaver> pSaver_Holder (pSaver);

	//
	// Prepare map to export.
	//

	SetStatus (GetStr (IDS_STATUS_PREPARE_EXPORT));

	// Prepare map copy, if required.
	const bool bCopyMap = (m_dwFlags & ftiSaveCopyOfMap) != 0;
	map_t * pMap;
	if (bCopyMap) {
		pMap = new map_t;

		CCritSec::CLock lock (g_cs);
		* pMap = g_map;
	} else
		pMap = & g_map;
	CObjectHolder<map_t> pMap_Holder (bCopyMap ? pMap : NULL);

	CCritSec::CLock lock (g_cs, ! bCopyMap);

	// Check if map is savable.
	SetProgressStatus (0);
	if (! pSaver->PrepareToSave (* pMap)) {
		SetStatusReady ();
		return;
	}

	ClearLog ();
	if (_strFileName != NULL && _strFileName [0] != '\0')
		ReportText ("Exporting map to file '%s'...", _strFileName);

	CWaiter w;

	const DWORD dwStartTime = ::GetTickCount ();

	// Open the file, if required.
	HANDLE hFile = NULL;
	if (_bOpenFile && _strFileName != NULL && _strFileName [0] != '\0') {
		CFileForWrite hFile2 (_strFileName);
		if (! hFile2.IsValid ()) {
			ErrorMsgBox ("Can't create file:\n %s", _strFileName);
			return;
		}

		hFile = hFile2.Detach ();
	}
	CWinHandle hFile_Handler (hFile);

	//
	// Export.
	//

	SetStatus (GetStr (IDS_STATUS_EXPORTING));
	SetProgressStatus (0);
	const SaveStatus_t ss = ExecuteSaveFileSafely (pSaver, hFile, hFile ? _strFileName : g_strFileName, * pMap);

	//
	// Report about results.
	//

	if (ss == ssFailed) {
		ErrorMsgBox ("Export has failed.");
		SetStatusReady ();

		if (_strFileName != NULL && _strFileName [0] != '\0')
			::DeleteFile (_strFileName);

		ReportText ("");
		ReportText ("ERROR: Export has failed.");
		ShowMessageLog ();

		return;
	}

	const DWORD dwElapsedTime = ::GetTickCount () - dwStartTime;
	ReportText ("Export time %0.2f s", .001*dwElapsedTime);

	SetStatusReady ();

	if (ss == ssOK && dwElapsedTime >= 20000)
		::PlaySound ("tada.wav", NULL, SND_ASYNC);
}
