//
// GPSMapEdit
// (c) Konstantin Galichsky (kg@geopainting.com), 2008
//
// Extract all elements from object.
//

# include "StdAfx.h"
# include "UndoRedo.h"
# include "ExtractElement.h"
# include "Label.h"
# include "Globals.h"

class CExtractAllElementsAction : public CEditAction {
  public:
	CExtractAllElementsAction (rgn_t & _rgn) : m_rgn (_rgn) {}

	virtual void GetName (string_t & _str) const;
	virtual cu_t CanUndo () const {return cuYes;}
	virtual bool Do ();
	virtual void Undo ();
	virtual void Redo ();

  private:
	rgn_t & m_rgn;

	std::vector<rgn_t *> m_NewRgns;
	rgn_t * _CreateNewRgn ();

	void _ReplaceRoadInTurnRestrictions (rgn_t * _pNewRgn, const rgn_t * _pRoadOld, rgn_t * _pRoadNew);
};

void CExtractAllElementsAction::GetName (string_t & _str) const {
	CCritSec::CLock lock (g_cs);

	_str.append ("extract element [");
	GetHintText (m_rgn, _str);
	_str.append ("]");
}

rgn_t * CExtractAllElementsAction::_CreateNewRgn () {
	rgn_t * const pNewRgn = & g_map.CreateRgn (m_rgn.GetKind ());

	pNewRgn->strLabel    = m_rgn.strLabel;
	pNewRgn->strComment  = m_rgn.strComment;
	pNewRgn->uchEndLevel = m_rgn.uchEndLevel;
	if (m_rgn.pExtra)
		pNewRgn->pExtra = m_rgn.pExtra;

	pNewRgn->SetType (m_rgn.GetTypeDesc (), m_rgn.GetType (), m_rgn.GetKind ());

	if (m_rgn.dwFlags & rgn_t::fDirIndicator)
		pNewRgn->dwFlags |= rgn_t::fDirIndicator;
	if (m_rgn.dwFlags & rgn_t::fDontFind)
		pNewRgn->dwFlags |= rgn_t::fDontFind;
	if (m_rgn.dwFlags & rgn_t::fModified)
		pNewRgn->dwFlags |= rgn_t::fModified;

	return pNewRgn;
}

bool CExtractAllElementsAction::Do () {
	CCritSec::CLock lock (g_cs);

	assert (m_NewRgns.empty ());

	if (m_rgn.IsPolyline () || m_rgn.IsPolygon ()) {
		// Check if at least two elements rest.
		const size_t cNonDeleted = m_rgn.GetElements ();
		if (cNonDeleted < 2)
			return true;

		// Get element to extract.
		rgn_elements_t & elements = m_rgn.elements;
		const size_t cElements = elements.size ();
		for (size_t cElement = 0; cElement < cElements; ++ cElement) {
			rgn_element_t & element = elements [cElement];
			if (element.bDeleted)
				continue;

			// Create new object.
			rgn_t * const pNewRgn = _CreateNewRgn ();
			m_NewRgns.push_back (pNewRgn);

			// Copy the element to the new object.
			pNewRgn->elements.push_back (rgn_element_t ());
			rgn_element_t & elementNew = pNewRgn->elements.back ();
			elementNew.cFileOffset = -1;
			elementNew.cLevel      = element.cLevel;
			elementNew.points      = element.points;
			if (element.pExtra) {
				elementNew.pExtra  = element.pExtra;

				pNewRgn->AddRefRGNodes ();

				_ReplaceRoadInTurnRestrictions (pNewRgn, & m_rgn, pNewRgn);
			}

			pNewRgn->RefreshAttributes (g_map);
		}
	} else {
		// Check if at least two origins rest.
		const size_t cNonDeleted = m_rgn.GetOrigins ();
		if (cNonDeleted < 2)
			return true;

		// Get origin to extract.
		rgn_origins_t & origins = m_rgn.origins;
		const size_t cOrigins = origins.size ();
		for (size_t cOrigin = 0; cOrigin < cOrigins; ++ cOrigin) {
			rgn_origin_t & origin = origins [cOrigin];
			if (origin.bDeleted)
				continue;

			// Create new object.
			rgn_t * const pNewRgn = _CreateNewRgn ();
			m_NewRgns.push_back (pNewRgn);

			// Copy the origin to the new object.
			pNewRgn->origins.push_back (rgn_origin_t ());
			rgn_origin_t & originNew = pNewRgn->origins.back ();
			originNew.cFileOffset = -1;
			originNew.cLevel      = origin.cLevel;
			originNew.point       = origin.point;

			pNewRgn->RefreshAttributes (g_map);
		}
	}

	m_rgn.Delete (g_map, true);

	return true;
}

void CExtractAllElementsAction::Undo () {
	CCritSec::CLock lock (g_cs);

	m_rgn.Undelete (g_map, true);

	const size_t cNewRgns = m_NewRgns.size ();
	for (size_t c = 0; c < cNewRgns; ++ c) {
		rgn_t * const pNewRgn = m_NewRgns [c];
		pNewRgn->Delete (g_map, true);
		_ReplaceRoadInTurnRestrictions (pNewRgn, pNewRgn, & m_rgn);
	}
}

void CExtractAllElementsAction::Redo () {
	CCritSec::CLock lock (g_cs);

	const size_t cNewRgns = m_NewRgns.size ();
	for (size_t c = 0; c < cNewRgns; ++ c) {
		rgn_t * const pNewRgn = m_NewRgns [c];
		pNewRgn->Undelete (g_map, true);
		_ReplaceRoadInTurnRestrictions (pNewRgn, & m_rgn, pNewRgn);
	}

	m_rgn.Delete (g_map, true);
}

void CExtractAllElementsAction::_ReplaceRoadInTurnRestrictions (rgn_t * _pNewRgn, const rgn_t * _pRoadOld, rgn_t * _pRoadNew) {
	assert (_pNewRgn && ! _pNewRgn->elements.empty ());
	point_extras_t & PointExtras = _pNewRgn->elements [0].pExtra->PointExtras;

	const size_t cPoints = _pNewRgn->elements [0].points.size ();

	const size_t cPointExtras = PointExtras.size ();
	for (size_t c = 0; c < cPointExtras; ++ c) {
		rgnode_t * const pNode = PointExtras [c].GetRGNode ();
		if (! pNode)
			continue;

		typedef std::pair<rgrestrictions_t::iterator, rgrestrictions_t::iterator> rr_t;
		const rr_t rr = g_map.rgrestrictions.equal_range (pNode);
		for (rgrestrictions_t::iterator ir = rr.first; ir != rr.second; ++ ir) {
			rgrestriction_t & r = ir->second;

			if (c == 0) {
				if (r.pFromRoad == _pRoadOld && r.bFromDirection == 0)
					r.pFromRoad = _pRoadNew;
				if (r.pToRoad   == _pRoadOld && r.bToDirection   != 0)
					r.pToRoad   = _pRoadNew;
			} else if (c == cPoints - 1) {
				if (r.pFromRoad == _pRoadOld && r.bFromDirection != 0)
					r.pFromRoad = _pRoadNew;
				if (r.pToRoad   == _pRoadOld && r.bToDirection   == 0)
					r.pToRoad   = _pRoadNew;
			} else {
				if (r.pFromRoad == _pRoadOld)
					r.pFromRoad = _pRoadNew;
				if (r.pToRoad   == _pRoadOld)
					r.pToRoad   = _pRoadNew;
			}
		}
	}
}

/////////////////////////////////////////////////////////////////

void ExtractAllElements (rgn_t & _rgn) {
	PerformEditAction (new CExtractAllElementsAction (_rgn));
}
