//
// GPSMapEdit
// (c) Konstantin Galichsky (kg@geopainting.com), 2008-2009
//
// Filtering for 'jitters' in tracks.
//

# include "StdAfx.h"
# include "Map.h"
# include "FilterTrackJitters.h"
# include "Globals.h"

static
size_t _FilterTrackJitters_BySpeed (points_t & _points, trackpoints_t & _TrackPoints) {
	// Estimate average delta t.
	float ave_dt = 0;
	size_t cN = 0;

	const size_t cPoints = __min (_points.size (), _TrackPoints.size ());
	for (size_t c = 1; c < cPoints; ++ c) {
		if (
			(_TrackPoints [c - 1].wFlags & trackpoint_info_t::defTime) == 0 ||
			(_TrackPoints [c    ].wFlags & trackpoint_info_t::defTime) == 0
		)
			continue;
		const float dt = (_TrackPoints [c].dtTime - _TrackPoints [c - 1].dtTime)*24*60*60;
		if (dt < 0)
			continue;
		if (cN >= 2 && dt > ave_dt/cN*10)
			continue;

		ave_dt += dt;
		++ cN;
	}
	if (cN == 0)
		return 0;
	ave_dt /= cN;

	size_t cJitters = 0;

	bool bRepeat;
	do {
		bRepeat = false;

		const size_t cPoints = __min (_points.size (), _TrackPoints.size ());

		if (cPoints == 0)
			return 0;
		const float fCosY = ::cos (_points [0].y*c_PI/180.f);

		for (size_t c = 1; c + 1 < cPoints; ++ c) {
			const point_t & pt0 = _points [c - 1];
			const point_t & pt1 = _points [c];
			const point_t & pt2 = _points [c + 2 < cPoints ? c + 2 : c + 1];
			const float d1 = (::fabs ((pt1.x - pt0.x)*fCosY) + ::fabs (pt1.y - pt0.y))*40e6/180;
			const float d2 = (::fabs ((pt2.x - pt0.x)*fCosY) + ::fabs (pt2.y - pt0.y))*40e6/180;
			if (d1 < d2*100)
				continue;

			float dt;
			if (
				(_TrackPoints [c - 1].wFlags & trackpoint_info_t::defTime) == 0 ||
				(_TrackPoints [c    ].wFlags & trackpoint_info_t::defTime) == 0
			)
				dt = ave_dt;
			else {
				dt = (_TrackPoints [c].dtTime - _TrackPoints [c - 1].dtTime)*24*60*60;
				if (dt < 0)
					dt = ave_dt;
				else if (dt > ave_dt*100000) {
					// Protect from double jitters.
					if ((_TrackPoints [c + 2].wFlags & trackpoint_info_t::defTime) != 0) {
						const float dt2 = (_TrackPoints [c + 2].dtTime - _TrackPoints [c - 1].dtTime)*24*60*60;
						if (dt2 > 0 && dt2 < dt*10)
							dt = ave_dt;
					}
				}
			}

			const float v = d1/dt;
			if (v > 1500/3.6f) {
				_points     .erase (_points     .begin () + c);
				_TrackPoints.erase (_TrackPoints.begin () + c);
				++ cJitters;
				bRepeat = true;
				break;
			}
		}
	} while (bRepeat);

	return cJitters;
}

static
size_t _FilterTrackJitters_ByLeg (points_t & _points, trackpoints_t & _TrackPoints) {
	size_t cDeletedPoints = 0;

	for (size_t cAttempt = 0; cAttempt < 100; ++ cAttempt) {
		bool bModified = false;

		size_t cPoints = _points.size ();
		if (cPoints == 0)
			break;
		if (_TrackPoints.size () != cPoints)
			break;

		const float fCosY = ::cos (_points [0].y*c_PI/180);

		for (size_t c = 0; c < cPoints; ++ c) {
			const point_t & pt     = _points [c];
			const point_t & ptPrev = _points [c == 0           ? 0           : c - 1];
			const point_t & ptNext = _points [c == cPoints - 1 ? cPoints - 1 : c + 1];

			//const trackpoint_info_t & point_extra = TrackPoints [c];

			size_t cDelete = -1;

			const float d = ::fabs (pt.x - ptPrev.x)*fCosY + ::fabs (pt.y - ptPrev.y);
			const float fDInM = d*40e6/360;
			if (fDInM > 10000) {
				if (c == 1) {
					cDelete = 0;
				} else if (c == cPoints - 1) {
					cDelete = cPoints - 1;
				} else {
					const float d_2 = ::fabs (ptNext.x - ptPrev.x)*fCosY + ::fabs (ptNext.y - ptPrev.y);
					const float fD2InM = d_2*40e6/360;
					if (fD2InM*1000 < fDInM)
						cDelete = c;
				}
			}

			if (cDelete != -1) {
				_points     .erase (_points     .begin () + cDelete);
				_TrackPoints.erase (_TrackPoints.begin () + cDelete);
				-- cPoints;
				c += 2;

				++ cDeletedPoints;
				bModified = true;
			}

			assert (cPoints == _points.size ());
		}

		if (! bModified)
			break;
	}

	return cDeletedPoints;
}

size_t FilterTrackJitters (points_t & _points, trackpoints_t & _TrackPoints) {
	const size_t cJitters_BySpeed = _FilterTrackJitters_BySpeed (_points, _TrackPoints);
	const size_t cJitters_ByLeg   = _FilterTrackJitters_ByLeg   (_points, _TrackPoints);
	return cJitters_BySpeed + cJitters_ByLeg;
}

size_t FilterTrackJitters (rgn_t & _rgn) {
	assert (_rgn.IsTrack ());
	assert (_rgn.pExtra);
	if (! _rgn.pExtra)
		return 0;

	track_info_t * const pTrack = _rgn.pExtra->wtr.track ();
	assert (pTrack);
	if (pTrack == NULL)
		return 0;

	size_t cDeletedPoints = 0;

	const size_t cElements = _rgn.elements.size ();
	for (size_t cElement = 0; cElement < cElements; ++ cElement) {
		rgn_element_t & element = _rgn.elements [cElement];
		if (element.bDeleted || ! element.pExtra)
			continue;

		cDeletedPoints += FilterTrackJitters (element.points, element.pExtra->TrackPoints);
	}

	return cDeletedPoints;
}
