//
// GPSMapEdit
// (c) Konstantin Galichsky (kg@geopainting.com), 2002-2009
//
// GPS panel window.
//

# include "StdAfx.h"
# include "Resource.h"
# include "GPSPanel.h"
# include "Positioning.h"
# include "Tracking.h"
# include "PtAlgo.h"
# include "Map.h"
# include "Position.h"
# include "Settings.h"
# include "MapView.h"
# include "PrintUnits.h"
# include "fast_floor.h"
# include "DrawHelpers.h"
# include "Dialog.h"
# include "Static.h"
# include "Button.h"
# include "Globals.h"

void _UpdateGPSPanel (HWND _hWnd);

class CGPSPanel : public CDialog {
  public:
	CGPSPanel () : CDialog (IDD_GPS_TRACKING) {}

	static CGPSPanel * s_pGPSPanel;

	// Override.
	virtual bool OnInit (HWND _hwndFocus);
	virtual void OnClose ();
	virtual	void OnPaint (const PAINTSTRUCT & _ps) const;
	virtual void OnCommand (WORD _wCmd, WORD _wNotifyCode, HWND _hwndCtl);
	virtual void OnMove (int _x, int _y);

  private:
	HWND m_hCompass;
};

CGPSPanel * CGPSPanel::s_pGPSPanel;

bool CGPSPanel::OnInit (HWND _hwndFocus) {
	// 'Keep visible' field.
	CheckDlgButton (IDC_KEEP_VISIBLE, GetSetting ("GPSPanel_KeepVisible", true) ? BST_CHECKED : BST_UNCHECKED);

	// 'Record track' field.
	SetButtonIcon (m_hWnd, IDC_RECORD_TRACK, IDI_RECORD_TRACK);
	const bool bRecordTrack = GetSetting ("GPSPanel_RecordTrack", false);
	CheckDlgButton (IDC_RECORD_TRACK, bRecordTrack ? BST_CHECKED : BST_UNCHECKED);
//	if (bRecordTrack)
//		StartTracking ();

	// Set bookmark.
	SetButtonIcon (m_hWnd, ID_CREATE_BOOKMARK_BY_GPS, IDI_SET_BOOKMARK);

	// Compass.
	m_hCompass = GetDlgItem (IDC_COMPASS);
	::ShowWindow (m_hCompass, SW_HIDE);

	_UpdateGPSPanel (m_hWnd);

	return true;
}

void CGPSPanel::OnClose () {
	::SendMessage (g_hWnd, WM_COMMAND, ID_GPS_TRACKING, 0);
	StopTracking ();
}

void CGPSPanel::OnPaint (const PAINTSTRUCT & _ps) const {
	const HDC hDC = _ps.hdc;

	//
	// Draw compass.
	//

	WINDOWPLACEMENT wp = {sizeof (WINDOWPLACEMENT)};
	::GetWindowPlacement (m_hCompass, & wp);

	const int x0 = (wp.rcNormalPosition.left + wp.rcNormalPosition.right)/2;
	const int y0 = (wp.rcNormalPosition.top + wp.rcNormalPosition.bottom)/2;
	const int r  = (wp.rcNormalPosition.right - wp.rcNormalPosition.left)/2;

	::SelectObject (hDC, ::GetStockObject (BLACK_PEN));
	::SelectObject (hDC, ::GetStockObject (WHITE_BRUSH));
	::Ellipse (hDC, x0 - r, y0 - r, x0 + r, y0 + r);

	const PositionInfo_t pi = GetPositionInfo ();

	if (pi.dwDef & PositionInfo_t::defHeading) {
		const float fA = - pi.fHeading*c_PI/180.f;
		DrawCompassArrow (hDC, x0, y0, fA, (r*2)/3);
	}
}

void CGPSPanel::OnMove (int _x, int _y) {
	RECT rect;
	::GetWindowRect (m_hWnd, & rect);
	POINT p = {rect.left, rect.top};
	::ScreenToClient (g_hMapWnd, & p);
	if (p.x < 0)
		p.x = 0;
	if (p.y < 0)
		p.y = 0;

	SaveSetting ("GPSPanelX", DWORD (p.x));
	SaveSetting ("GPSPanelY", DWORD (p.y));
}

void CGPSPanel::OnCommand (WORD _wCmd, WORD _wNotifyCode, HWND _hwndCtl) {
	switch (_wCmd) {
		case IDC_KEEP_VISIBLE:
			SaveSetting ("GPSPanel_KeepVisible", IsDlgButtonChecked (IDC_KEEP_VISIBLE));
			break;

		case IDC_RECORD_TRACK: {
			const bool bRecordTrack = IsDlgButtonChecked (IDC_RECORD_TRACK);
			SaveSetting ("GPSPanel_RecordTrack", bRecordTrack);
			if (bRecordTrack)
				StartTracking ();
			else
				StopTracking ();
			break;
		}

		case IDC_SHOW_TRACKS: {
			char strTracksFolder [MAX_PATH];
			GetTracksFolder (strTracksFolder);
			::ShellExecute (m_hWnd, NULL, strTracksFolder, NULL, NULL, SW_SHOWNORMAL);
			break;
		}

		case ID_CREATE_BOOKMARK_BY_GPS:
			::SendMessage (g_hWnd, WM_COMMAND, ID_CREATE_BOOKMARK_BY_GPS, 0);
			break;
	}
}

bool InitGPSPanel () {
	assert (g_hMapWnd);

	ShowModeless (CGPSPanel::s_pGPSPanel, g_hMapWnd);
	if (CGPSPanel::s_pGPSPanel == NULL)
		return false;
	::ShowWindow   (* CGPSPanel::s_pGPSPanel, SW_HIDE);
	::SetWindowPos (* CGPSPanel::s_pGPSPanel, NULL, GetSetting ("GPSPanelX", (DWORD) 0), GetSetting ("GPSPanelY", (DWORD) 0), 0, 0, SWP_NOZORDER | SWP_NOSIZE);

	return true;
}

void ShowGPSPanel () {
	if (CGPSPanel::s_pGPSPanel == NULL)
		return;
	::ShowWindow (* CGPSPanel::s_pGPSPanel, SW_SHOW);
	_UpdateGPSPanel (* CGPSPanel::s_pGPSPanel);
//	::SetTimer (g_hwndGPSPanel, 10, 1000, NULL);

	if (GetSetting ("GPSPanel_RecordTrack", false))
		StartTracking ();
}

void UpdateGPSPanel () {
	if (CGPSPanel::s_pGPSPanel == NULL)
		return;
	_UpdateGPSPanel (* CGPSPanel::s_pGPSPanel);
}

void HideGPSPanel () {
	::ShowWindow (* CGPSPanel::s_pGPSPanel, SW_HIDE);
//	::KillTimer (* CGPSPanel::s_pGPSPanel, 10);

	StopTracking ();
}

static
void _UpdateGPSPanel (HWND _hWnd) {
	string_t strStatus;
	GetPositioningStatus (strStatus);
	::SetDlgItemText (_hWnd, IDC_COM_STATUS, strStatus.c_str ());

	const PositionInfo_t pi = GetPositionInfo ();

	// Fix.
	if (pi.dwDef & (PositionInfo_t::defFix | PositionInfo_t::defSatellites)) {
		const char * strFix = NULL;
		switch (pi.Fix) {
			case PositionInfo_t::fixInvalid:    strFix = "Invalid";    break;
			case PositionInfo_t::fixUnknown:    strFix = "Unknown";    break;
			case PositionInfo_t::fix2D:         strFix = "2D";         break;
			case PositionInfo_t::fix3D:         strFix = "3D";         break;
			case PositionInfo_t::fix2D_diff:    strFix = "2D diff";    break;
			case PositionInfo_t::fix3D_diff:    strFix = "3d diff";    break;
			case PositionInfo_t::fixPrec:       strFix = "Prec";       break;
			case PositionInfo_t::fixRTK:        strFix = "RTK";        break;
			case PositionInfo_t::fixFRTK:       strFix = "Float RTK";  break;
			case PositionInfo_t::fixEstimated:  strFix = "Estimated";  break;
			case PositionInfo_t::fixManual:     strFix = "Manual";     break;
			case PositionInfo_t::fixSimulation: strFix = "Simulation"; break;
			default:                            strFix = "Unknown";    break;
		}
		::SetDlgItemText (_hWnd, IDC_GPS_FIX, strFix);
	} else
		::SetDlgItemText (_hWnd, IDC_GPS_FIX, "-");

	// Satellites.
	if (pi.dwDef & PositionInfo_t::defSatellites)
		::SetDlgItemInt (_hWnd, IDC_SATELLITES, pi.cSatellites, FALSE);
	else
		::SetDlgItemText (_hWnd, IDC_SATELLITES, "-");

	// Lat/Lon.
	if (pi.dwDef & PositionInfo_t::defXY) {
		char strXY [64];
		PrintPosition (strXY, point_t (pi.fX, pi.fY));

		if (pi.dwDef & PositionInfo_t::defXYPrec) {
			char strPrec [32];
			char strPrecUnits [16];
			PrintDistance (strPrec, strPrecUnits, pi.fXYPrec);

			SetDlgItemTextF (_hWnd, IDC_POSITION, "%s  %s %s", strXY, strPrec, strPrecUnits);
		} else
			SetDlgItemTextF (_hWnd, IDC_POSITION, strXY);
	} else
		::SetDlgItemText (_hWnd, IDC_POSITION, "-");

	// Speed.
	if (pi.dwDef & PositionInfo_t::defSpeed) {
		char strSpeed [32];
		char strSpeedUnits [16];
		PrintSpeed (strSpeed, strSpeedUnits, pi.fSpeed);

		SetDlgItemTextF (_hWnd, IDC_SPEED, "%s %s", strSpeed, strSpeedUnits);
	} else
		::SetDlgItemText (_hWnd, IDC_SPEED, "-");

	// Altitude.
	if (pi.dwDef & PositionInfo_t::defAltitude) {
		char strAltitude [32];
		char strAltitudeUnits [16];
		PrintElevation (strAltitude, strAltitudeUnits, pi.fAltitude);

		if (pi.dwDef & PositionInfo_t::defAltitudePrec) {
			char strAltitudePrec [32];
			PrintElevation (strAltitudePrec, NULL, pi.fAltitudePrec);

			SetDlgItemTextF (_hWnd, IDC_ALTITUDE, "%s  %s %s", strAltitude, strAltitudePrec, strAltitudeUnits);
		} else
			SetDlgItemTextF (_hWnd, IDC_ALTITUDE, "%s %s", strAltitude, strAltitudeUnits);
	} else
		::SetDlgItemText (_hWnd, IDC_ALTITUDE, "-");

	// Redraw position cursor.
	if (pi.dwDef & PositionInfo_t::defXY) {
		static float s_fPrevX = -200, s_fPrevY = -200;
		if (s_fPrevX != pi.fX || s_fPrevY != pi.fY) {
			s_fPrevX = pi.fX;
			s_fPrevY = pi.fY;

			Redraw ();
		}

		// Scroll map to keep position visible, if need.
		if (::IsDlgButtonChecked (_hWnd, IDC_KEEP_VISIBLE) == BST_CHECKED) {
			bool bValid = false;
			rect_t rectVisibleCenter;
			{
				CCritSec::CLock lock (g_cs);

				bValid = ! g_map.rectBound.Invalid ();
				if (bValid) {
					rect_t rectVisible;
					GetVisibleRect (rectVisible);

					rectVisibleCenter.x0 = rectVisible.x0 + rectVisible.Width ()*.15f;
					rectVisibleCenter.x1 = rectVisible.x0 + rectVisible.Width ()*.85f;
					rectVisibleCenter.y0 = rectVisible.y0 + rectVisible.Height ()*.15f;
					rectVisibleCenter.y1 = rectVisible.y0 + rectVisible.Height ()*.85f;
				}
			}
			if (bValid) {
				const point_t ptGPS (pi.fX, pi.fY);
				if (! rectVisibleCenter.Contains (ptGPS)) {
					SetToCenter (ptGPS);
					Refresh ();
				}
			}
		}
	}

	// Update compass.
	::InvalidateRect (_hWnd, NULL, TRUE);
	::UpdateWindow (_hWnd);
}
