# ifndef __GarminImg_H_
# define __GarminImg_H_

//
// (c) Konstantin Galichsky (kg@geopainting.com), 2002-2009
// Garmin and MapSource are registered trademarks or trademarks of Garmin Ltd. or one of its subsidiaries.
//

# ifdef _MSC_VER
#	pragma pack (push, 1)
# endif

	// 24-bit integer representation of coordinate: unit is 360/2^24 deg.
typedef       BYTE       B3_t [3];
typedef const BYTE const_B3_t [3];

inline
DWORD ConvertB3ToDWORD (const_B3_t _data) {
    return (DWORD (_data [2]) << 16) + (DWORD (_data [1]) << 8) + DWORD (_data [0]);
}

//
// NOTES:
//
// Garmin IMG file contains low-level 'sub-files' (which are like files in terms
// of Garmin's embedded file system).
//
// The sub-files have 8-digit (hex or dec) identifier and 3-character extension (type).
// The types are:
//   TRE - map structure (zoom levels, object groups, copyrights)
//   RGN - map objects
//   LBL - labels and extra properties of map objects
// The optional types:
//   GMP - index of subfiles (NT format)
//   NET - objects' extra properties, list of streets and cross-roads (for lock on road)
//   NOD - road graph (for auto-routing)
//   MDR - binds all map sheets (for global search)
//   SRT - sort table (used with MDR)
//   DEM - terrain shading
//

extern const char c_strImgSignature [];

	//
	// FAT block: contain index information about sub-files' blocks.
	// NOTE: Always 512-byte.
	//
struct CFileIndexBlock {
	// 0 - for last dummy block (deleted ?), 1 - for ordinary blocks.
	BYTE btValid;

	//
	// Sub-file name.
	//
	char strName [8]; // I02034A6
	char strType [3]; // RGN, TRE, LBL, NET, NOD etc

	//
	// Exact sub-file length (in bytes).
	//
	DWORD dwLength;

	BYTE btZero010;

	//
	// Ordinal of the index block for the given sub-file.
	// NOTE: large files are described by several index blocks.
	//
	BYTE btIndex;

	BYTE btZero012 [14];

	//
	// The 'FAT': index of block numbers.
	//
	WORD wStartBlock;
	WORD dataBlockNumbers [239]; // 0xFFFF means no block
};

	//
	// The header of IMG file.
	// NOTE: 3*512 bytes (except FAT).
	//
struct CImgHeader {
	BYTE btZero00 [10];

	//
	// The map update date.
	//
	BYTE btUpdateMonth;
	BYTE btUpdateYear;     // Last two digits: e.g. 99 for 1999, 03 for 2003 etc

	BYTE btZero0C [3];

	// Minus sum of all bytes except this.
	// NOTE: as a result, the sum of all bytes in the file should be zero.
	BYTE btCheckSum;       

	// The signature.
	char strSignature [7]; // c_strImgSignature = "DSKIMG\0"

	BYTE bt017;            // 0x02

	WORD wSectors;         // cluster/512
	WORD wHeads;           // 0x10
	WORD wCylinders;       // 0x20

	WORD wZero01e;
	BYTE btZero20 [16 + 9];

	WORD wYear;
	BYTE wMonth;
	BYTE btDay;
	BYTE btHour;
	BYTE btMin;
	BYTE btSec;

	BYTE btFAT0BlockStart; // 0x2

	char strGarmin [8];    // "GARMIN\0\0"

	// The display name.
	// NOTE: complemented by strName2 defined below.
	char strName [20];     // "<mapname>"

	WORD wHeads2;          // 0x0010 = 'heads' ?
	WORD wSectors2;        // 0x0004 =(+18)=(+1c4) - 'sectors' ?

	//
	// Block size (may be 512, 1024, 2048 etc).
	//
	BYTE btBlockSizePowerOf2_1; // 9
	BYTE btBlockSizePowerOf2_2; // 0=512, 1=1024, 2=2048, 4=8192 etc
	WORD GetBlockSize () const {
		return 1 << (btBlockSizePowerOf2_1 + btBlockSizePowerOf2_2);
	}

	WORD wBlockSize;       // 0800=sectors*heads*cylinders/2^btBlockSizePowerOf2_2

	char strName2 [31];    // 30 spaces \0

	BYTE btZero0084 [314];

// 0x1be:
	// The "partition table".
	BYTE btBoot;           // 0x00
	BYTE btStartHead;      // 0x00
	BYTE btStartSector;    // 0x01
	BYTE btStartCylinder;  // 0x00
	BYTE btSystemType;     // 0x00

	BYTE btEndHead;        // =wHeads-1
	BYTE btEndSector;      // =wSectors
	BYTE btEndCylinder;    // =wCylinders-1

	DWORD dwRelSector;     // 0x00
	DWORD dwClusterSize;   // 0x800*2^btBlockSizePowerOf2_2

	BYTE btZero01ce [48];

	WORD wSignature;       // 0xAA55

	BYTE btZero0200 [512];

// 0x400:
	BYTE btHasFAT;         // 01=FAT, 00=NT
	BYTE btSpaces [11];    // 20 20 20 20 20 20 20 20 20 20 20

	//
	// The offset of first sub-file in the IMG file.
	// NOTE: the value may be used to determine the FAT size.
	//
	DWORD dwDataOffset;

//0410  03000000000000000000000000000000
//0420  0000 0100 0200 0300 -0400 0500		Layouts table (6 * 200 = C00)
//      FFFFFFFF					(Block length=200 , siehe IMG+61)
	BYTE bt0410 [512 - 16];

// 0x600:

	//
	// The FAT.
	//
	// NOTE: the real number of index blocks may differ: typicaly at least
	// 3 blocks	for TRE, LBL and RGN plus 1 'last' dummy block.
	CFileIndexBlock FAT [1];
};

	//
	// The common header for all sub-files.
	//
struct CSubFileHeader {
	// Header length (in bytes).
	// NOTE: check it before access to last fields of headers.
	WORD wHeaderLength;

	// The signature.
	char strSignature [10]; // 'GARMIN RGN', 'GARMIN TRE', 'GARMIN LBL', 'GARMIN NET' rtc

	BYTE bt1;     // 0x01
	BYTE btVer;   // 0x0 for most maps, 0x80 for locked TRE.

	// Sub-file creation date/time.
	WORD wYear;
	BYTE btMonth;
	BYTE btDay;
	BYTE btHour;
	BYTE btMin;
	BYTE btSec;
// 15:
};

	///////////////////////////////////////////////////////////
	//
	// Header of GMP sub-file.
	//
struct CGmpHeader : public CSubFileHeader {
	DWORD dwZero015;

	DWORD dwTreOffset;
	DWORD dwRgnOffset;
	DWORD dwLblOffset;
	DWORD dwNetOffset;
	DWORD dwNodOffset;

	DWORD dwZero02d;
	DWORD dwZero031;
};


	///////////////////////////////////////////////////////////
	//
	// Header of TRE sub-file.
	//
struct CTreHeader : public CSubFileHeader {
	// The map boundaries.
	B3_t b3MapNorth;
	B3_t b3MapEast;
	B3_t b3MapSouth;
	B3_t b3MapWest;

	// Zoom levels definitions (CTreLevel).
	DWORD dwLevelsOffset;
	DWORD dwLevelsLength;

	// Object groups (CTreGroup).
	DWORD dwGroupsOffset;
	DWORD dwGroupsLength;

	// Copyright strings (stored in LBL).
	DWORD dwCopyrightsOffset;
	DWORD dwCopyrightsLength;
	WORD  wCopyrightsRecSize;

	WORD  wZero003b;
	WORD  wZero003d;

	// Display flags.
	BYTE  bDetailed           : 1, // 0001=R&R 0001=World 1101=MG-Austr. 0000=overview map
	      bTransparent        : 1,
	      bStreetBeforeNumber : 1,
	      bZipBeforeCity      : 1,
	      btUnknown0          : 4;
	BYTE  btDrawPriority      : 5, // 19=R&R 05=World 1E=MG-Austr. 00=overview map
	      bTransparent2       : 1,
	      btUnknown1          : 2;
	WORD  wZero0041;
	BYTE  btUnknown10;    // 43: 01=R&R 01=World 01=MG-Austr. 01=overview map
	BYTE  btUnknown11;    // 44: 04=R&R 01=World 03=MG-Austr. 01=overview map
	BYTE  btUnknown12;    // 45: 0d=R&R 05=World 0E=MG-Austr. 04=overview map
	BYTE  btZero046;
	WORD  w047;           // 47: 0x0001
	BYTE  btZero049;      // 49: 0

	// Types of polylines.
	DWORD dwPolylineInfoOffset;
	DWORD dwPolylineInfoLength;
	WORD  wPolylineInfoRecSize;

	WORD  w054;
	WORD  wZero056;

	// Types of polygons.
	DWORD dwPolygonInfoOffset;
	DWORD dwPolygonInfoLength;
	WORD  wPolygonInfoRecSize;

	WORD  w062;
	WORD  wZero064;

	// Types of POI and points.
	DWORD dwPointInfoOffset;
	DWORD dwPointInfoLength;
	WORD  wPointInfoRecSize;

	WORD  w070;
	WORD  wZero072;
// 74:

	// ID of the map.
	// NOTE: TRE header for WorldMap-99 does not contain this field!
	DWORD dwMapID;
// 78:
};

struct CTreHeader2 : public CTreHeader {
	DWORD dwZero078;

	// Object groups V2 (CTreGroup2).
	DWORD dwGroups2Offset;
	DWORD dwGroups2Length;
	WORD  wGroup2RecSize; // 13

	DWORD dw086;          // 0x487

	DWORD dwTre8Offset;   // Order: polyline, polygon, POI; each sorted by type (1 type 1 levels 1 subtype)
	DWORD dwTre8Length;
	WORD  wTre8RecSize;   // 3

	WORD  wPolylines2TypesNum;
	WORD  wPolygons2TypesNum;
	WORD  wPoints2TypesNum;
// 9A:
};

	// TRE1: zoom level info.
	// NOTE: come in sorted order, less detailed level is first.
struct CTreZoomLevel {
	BYTE btZoom   : 4, // Zoom level for MapSource.
	     unknown  : 3, // Always 0?
	     bExtPrev : 1; // 1 means no own data but the data from previous level are applied
	BYTE btBits;           // Number of bits per coordinate.
	WORD wGroups;          // Number of TRE groups for this zoom level.
};

	// TRE2: object group info.
	// NOTE: 16 bytes for all zoom levels except for the most detailed (stored last)
	// which has 14-byte size.
struct CTreGroup {
	DWORD cRgnOffset : 28,
	      btKinds    : 4; // 0x1 - has points, 0x2 - has indexed points,
	                      // 0x4 - has polylines, 0x8 - has polygons.

	// Reference point for all objects in the group.
	// NOTE: all coordinates are defined as difference from this point.
	B3_t b3X0;
	B3_t b3Y0;

// 0a:
	// The group area is rectangle with the ref point in the center, the size is
	// (wWidth*2 + 1) x (wHeight*2 + 1) in variable grid units for this zoom level.
	WORD wWidth   : 15,
	     bEndMark : 1;    // 1 for last group in chain referred by 'parent' group
	WORD wHeight  : 16;

// 0e:
	// 1-based index of the first group in chain for the next zoom level.
	// NOTE: omited for last (i.e most detailed 0-th) zoom level.
	WORD wNextLevelIndex;
};

	// TRE3: copyright strings in LBL sub-file.
struct CTreCopyright {
	B3_t b3Offset;
};

	// TRE4: types of polylines.
struct CTrePolylineInfo {
	BYTE btType;
	BYTE btLevel;
};

	// TRE5: types of polygons.
struct CTrePolygonInfo {
	BYTE btType;
	BYTE btLevel;
};

	// TRE6: types of points.
struct CTrePointInfo {
	BYTE btType;
	BYTE btLevel;
	BYTE btSubType;
};

struct CTreGroup2_size13 {
	DWORD dwPolygonsOffset;
	DWORD dwPolylinesOffset;
	DWORD dwPointsOffset;
	BYTE btObjects;
};

struct CTreGroup2_size17 {
	DWORD dwSize;           // ??
	BYTE  btUnknown;        // ??
	DWORD dwPolygonsOffset;
	DWORD dwPolylinesOffset;
	DWORD dwPointsOffset;
};

	///////////////////////////////////////////////////////////
	//
	// Header of RGN sub-file.
	//
struct CRgnHeader : public CSubFileHeader {
	DWORD dwDataOffset; // NOTE: just equal to wHeaderLength
	DWORD dwDataLength;
// 1d:
};

struct CRgnHeader2 : public CRgnHeader {
	DWORD dwPolygons2Offset;
	DWORD dwPolygons2Length;

	DWORD dwZero025;
	DWORD dwZero029;
	DWORD dwZero02D;
	DWORD dwZero031;
	DWORD dwZero035;

	DWORD dwPolylines2Offset;
	DWORD dwPolylines2Length;

	DWORD dwZero041;
	DWORD dwZero045;
	DWORD dwZero049; // ?? 0x00000001
	DWORD dwZero04D; // ?? 0x00000001
	DWORD dwZero051;

	DWORD dwPoints2Offset;
	DWORD dwPoints2Length;

	DWORD dwZero05D;
	DWORD dwZero061;
	DWORD dwZero065; // ?? 0x20000001 or 0x00000001
	DWORD dwZero069; // ?? 0x00000003
	DWORD dwZero06D;
	DWORD dwZero071;
	DWORD dwZero075;
	DWORD dwZero079; // ?? 0x00000001
};

struct CRgnPOIData {
	DWORD btType        : 8,
	      b3LabelOffset : 22,// bExtra=0 - label in LBL1, bExtra=1 - offset in LBL6
	      bExtra        : 1, 
	      bHasSubType   : 1; // NOTE: sometimes 0.

	// NOTE: relative to TRE group.
	WORD  b2X;
	WORD  b2Y;

	// NOTE: available if bHasSubType=1.
	BYTE  btSubType;
};

struct CRgnPointData {
	DWORD btType        : 8,
	      b3LabelOffset : 22,
	      bExtra        : 1, // Always 0 ?
	      bHasSubType   : 1; // Always 0 ?

	// NOTE: relative to TRE group.
	WORD  b2X;
	WORD  b2Y;
};

struct CRgnPolylineData {
	DWORD btType         : 6,
	      bDirIndicator  : 1,
	      bTwoBytesInLen : 1,
	      b3LabelOffset  : 22,
	      bExtraBit      : 1, // 1=one more bit per point in bitstream
	      bInNet         : 1; // 1=b3LabelOffset is offset in NET sub-file

	// NOTE: relative to TRE group.
	WORD  b2X0;
	WORD  b2Y0;

	union {
		BYTE btBitstreamLen; // if bTwoBytesInLen=0
		WORD wBitstreamLen;  // if bTwoBytesInLen=1
	};
};

struct CRgnPolygonData {
	DWORD btType         : 7,
	      bTwoBytesInLen : 1,
	      b3LabelOffset  : 22,
	      bUnknown       : 1, // ??
	      bInNet         : 1; // ??

	// NOTE: relative to TRE group.
	WORD  b2X0;
	WORD  b2Y0;

	union {
		BYTE btBitstreamLen; // if bTwoBytesInLen=0
		WORD wBitstreamLen;  // if bTwoBytesInLen=1
	};
};

struct CRgnPOI2Data {
	BYTE  btType;
	BYTE  btSubType     : 5,
	      bHasLabel     : 1,
	      bUnknown0     : 1,
	      bHasExtraByte : 1; // 1 if btExtraByte presented

	// NOTE: relative to TRE group.
	WORD  b2X;
	WORD  b2Y;

	DWORD b3LabelOffset : 22,// bExtra=0 - label in LBL1, bExtra=1 - offset in LBL6
	      bExtra        : 1, 
	      bUnknown1     : 1,
	      btExtraByte   : 8; // Missed if bHasExtraByte==0.
};

struct CRgnPoly2Data {
	BYTE  btType;
	BYTE  btSubType     : 5,
	      bHasLabel     : 1,
	      bUnknown0     : 1,
	      bUnknown1     : 1; // This bit is sometimes presented in BlueChart, it shows some additional data.

	// NOTE: relative to TRE group.
	WORD  b2X0;
	WORD  b2Y0;
};

//////////////////////////////////////////////////////////

union CLblPOIExtraDef {
	BYTE bt;
	struct {
		BYTE bHasNumber         : 1,
		     bHasStreet         : 1,
		     bHasCity           : 1,
		     bHasZip            : 1,
		     bHasPhone          : 1,
		     bHasExit           : 1,
		     bHasTidePrediction : 1,
		     bNumberIsInPlace   : 1; // 0x1 - number is stored in-place as 11-base, 0x0 - number is 3-byte ref to LBL
	};
};

	///////////////////////////////////////////////////////////
	//
	// Header of LBL sub-file.
	//
struct CLblHeader : public CSubFileHeader {
	//
	// Reference to the labels bitstream.
	//
	DWORD dwDataOffset;
	DWORD dwDataLength;
	BYTE btLblOffsetMultiplierPowerOf2;
	BYTE btLblCoding;        // 6 by 6-bit coding, 9 for 8-bit SBCS, 10 for MBCS.

	//
	// NOTE: empty entries have dwLengthX=0.
	//

	// Countries list.
	DWORD dwCountriesOffset;
	DWORD dwCountriesLength;
	WORD  wCountriesRecSize; // 3

	DWORD dwZero1;

	// Regions/Provinces/States list.
	DWORD dwRegionsOffset;
	DWORD dwRegionsLength;
	WORD  wRegionsRecSize;   // 5

	DWORD dwZero2;

	// Cities list.
	DWORD dwCitiesOffset;
	DWORD dwCitiesLength;
	WORD  wCitiesRecSize;    // 5

	WORD  wZero3;
	WORD  w4;                // 8000 R&R+MG Europe, 0000 Overview map

	// Index of POIs sorted by types and name.
	DWORD dwPOIIndexOffset;
	DWORD dwPOIIndexLength;
	WORD  wPOIIndexRecSize;  // 4

	DWORD dwZero5;

	// POI extra properties.
	DWORD dwPOIExtraOffset;
	DWORD dwPOIExtraLength;
	BYTE  btPOIExtraOffsetMultiplierPowerOf2;
	CLblPOIExtraDef POIExtraDef;

	WORD  wZero6;
	BYTE  btZero6;

	// Index of POI types.
	DWORD dwPOITypeIndexOffset;
	DWORD dwPOITypeIndexLength;
	WORD  wPOITypeIndexRecSize; // 4

	DWORD dwZero7;

	// Zip codes.
	DWORD dwZipsOffset;
	DWORD dwZipsLength;
	WORD  wZipsRecSize;      // 3

	DWORD dwZero8;

	//
	// NOTE: the following fields (LBL9 - LBL12) are not empty if highway exits with facilities are presented.
	//

	// Highways.
	DWORD dwHighwaysOffset;
	DWORD dwHighwaysLength;
	WORD  wHighwaysRecSize;  // 6

	DWORD dwZero9;

	// Exits.
	DWORD dwExitsOffset;
	DWORD dwExitsLength;
	WORD  wExitsRecSize;     // 5

	WORD  w10;               // 0, 1, 5 ??
	WORD  wZero10;

	// Highway extra properties.
	DWORD dwHighwayExtraOffset;
	DWORD dwHighwayExtraLength;
	WORD  wHighwayExtraRecSize; // 3

	DWORD dwZero11;

	//
	// NOTE: the fields below are optional (e.g. not defined in WorldMap-99).
	// Header length is to be checked before access.
	//

// 0xAA:
	WORD  wCodePage;
	WORD  w12;
	WORD  w12a;

// 0xB0:
	// Reference to some info string (plain 8-bit coding). E.g. "American", "Western European Sort".
	DWORD dwSortingOffset;
	DWORD dwSortingLength;

// 0xB8:
	DWORD dwOffset14;
	DWORD dwLength14;
	WORD  wRecSize14;

	WORD  wZero14;

// 0xC4:
	DWORD dwTidePredictionsOffset;
	DWORD dwTidePredictionsLength;
	WORD  wTidePredictionsRecSize;

	WORD  wZero15;
};

	// LBL2.
struct CLblCountry {
	B3_t b3NameOffset;  // In LBL
};

	// LBL3.
struct CLblRegion {
	WORD wCountryIndex; // 1-based
	B3_t b3NameOffset;  // In LBL
};

	// LBL4.
struct CLblCity {
	union {
		// bPoint=1
		struct {
			BYTE btPointIndex;      // 1-based
			WORD wGroupIndex;       // 1-based
		};
		// bPoint=0
		B3_t     b3LabelOffset;
	};
	WORD wRegionIndex : 14, // 1-based, 0 is special value
	     bNoRegion    : 1,  // 0 - wRegionIndex is region index, 1 - wRegionIndex is country index
	     bPoint       : 1;
};

	// LBL5.
struct CLblPOIIndexItem {
	BYTE btPOIIndex;    // 1-based index of POI inside group.
	WORD wGroupIndex;   // 1-based index of group containing POI.
	BYTE btPOISubType;  // Subtype of referenced-POI.
};

	// LBL6 for POI postal address and phone.
	// NOTE: the size is variable.
struct CLblPOIExtra {
	DWORD b3LabelOffset      : 22,
	      bUnknown           : 1, // always 0 ?
	      bHasFlags          : 1, // bHasFlags=0 means that bits from btFlags should be used.
	      // The following 8 bits are available if bHasFlags=1.
	      btFlags            : 8;
};

	// LBL6 for highway exits.
	// NOTE: the size is variable.
struct CLblExitExtra {
	B3_t b3LabelOffset;

//  BYTE btFlags;         // Defined only if 23-th bit of b3LabelOffset is 1.

//	B3_t b3Street;        // NOTE: just 0-21 bits
	                      // 22-nd means 'overnight parking'
						  // 23-rd means that facilities are defined

//	WORD wHighwayIndex;   // ? 1 or 2 byte, depending on the amount of highways (LBL9)
//	WORD wFirstFacility;  // ? 1 or 2 byte, depending on the amount of facilities (LBL10)
};

	// LBL7.
struct CLblPOITypeIndexItem {
	BYTE btType;          // POI type.
	B3_t b3StartIndex;     // 1-based index of starting position in array of CLblPOIIndexItem.
};

struct CLblZip {
	B3_t b3CodeOffset;    // In LBL.
};

	// LBL9.
struct CLblHighway {
	B3_t b3NameOffset;    // In LBL
	WORD wExtraIndex;     // 1-based index in LBL11.
	BYTE btZero;          // ??
};

	// LBL10.
struct CLblExitFacility {
	DWORD b3LabelOffset : 22, // In LBL
	      bUnknown      : 1,
	      bLast         : 1,  // =1 for last facility for the given exit

	// 0x00 - Truck Stop/24-hour Diesel Fuel With Restaurant
	// 0x01 - Truck Fuel/Diesel Fuel With Large Vehicle Clearance
	// 0x02 - Gas/Automobile Fuel
	// 0x03 - Food/Restaurant
	// 0x04 - Lodging/Hotel/Motel - Call 1-866-394-8768 For Reservations
	// 0x05 - Auto service/Vehicle Repair and Service
	// 0x06 - Auto service/Diesel Engine Service
	// 0x07 - Auto service/Commercial Vehicle Wash
	// 0x08 - Camp/Campground and RV Service
	// 0x09 - Hospital/Medical Facilities
	// 0x0a - Store/Automated Teller Machines
	// 0x0b - Park/Forest, Park, Preserve, or Lake
	// 0x0c - Point Of Interest/Useful Services, Sites, or Attractions
	// 0x0d - Fast Food
	      btType        : 4,

	      bUnknown2     : 1,

	// 0x0 - north
	// 0x1 - south
	// 0x2 - east
	// 0x3 - west
	// 0x4 - inner side
	// 0x5 - outer side
	// 0x6 - both sides
	// 0x7 - not defined
	      btDirection   : 3;

	// 0x01 - Truck/RV Parking
	// 0x02 - Convenience Story
	// 0x04 - Diesel Fuel
	// 0x08 - Car Wash
	// 0x10 - Liquid Propane
	// 0x20 - Truck Scales
	// 0x40 - Open 24 Hours
	// 0x80 - not used
	BYTE btFacilities;
};

	// LBL11.
struct CLblHighwayExtra {
	BYTE btZero;        // 0
	WORD wRegionIndex;  // 1-based.
};


	///////////////////////////////////////////////////////////
	//
	// Header of NET sub-file.
	//
struct CNetHeader : public CSubFileHeader {
	DWORD dwDataOffset; // NOTE: just equal to wHeaderLength
	DWORD dwDataLength;
	BYTE  btDataOffsetMultiplierPowerOf2;

	// NET2 - segmented roads offsets.
	DWORD dw2Offset;
	DWORD dw2Length;
	BYTE  bt2MultiplierPowerOf2;

	// NET3 - sorted roads offsets.
	DWORD dwSortedRoadsOffset;
	DWORD dwSortedRoadsLength;
	WORD  wSortedRoadsRecSize;  // 0x3

	DWORD dwUnknown;  // 0
	BYTE  btUnknown0; // 0x1
	BYTE  btUnknown1;
// 37:

	//
	// NOTE: the fields below are optional (e.g. not defined in WorldMap-99).
	// Header length is to be checked before access.
	//
	DWORD dwUnknown2; // 0
// 3b:

	DWORD dwUnknown3; // 0x1
	DWORD dwUnknown4; // 0
};

	// NET1.
struct CNetData {
	DWORD b3LabelOffset : 22,
	      bSegmented    : 1, // Pointer to segmented roads follows
	      bLastLabel    : 1; // At most 4 labels.
};

struct CNetDataFlags {
	BYTE b0              : 1, // ?
	     bOneWay         : 1,
	     bLockOnRoad     : 1, // ?
	     b3              : 1, // ?
	     bHasAddress     : 1,
	     bAddrStartRight : 1,
	     bHasNodInfo     : 1,
	     bMajorHW        : 1; // ?
};

struct CNetAddressFlags {
	BYTE btUnknown    : 2,
	     btZipsFmt    : 2,
	     btCitiesFmt  : 2,
	     btNumbersFmt : 2;
};

	// NET2.
struct CNet2 {
	// TODO: research.
};

	// NET3.
struct CNetSorted {
	B3_t b3NetDataOffset; // Bits 22-23 define label number (0-3).
};


	///////////////////////////////////////////////////////////
	//
	// Header of NOD sub-file.
	//
struct CNodHeader : public CSubFileHeader {
	// The routing graph: nodes and groups of nodes.
	DWORD dwNodesOffset;
	DWORD dwNodesLength;
	BYTE  btNodesFlag0;              // ?? 0x01, 0x03, 0x07, 0x27; lower bit is always set
	BYTE  bLeftSideTraffic : 1,
		  btNodesFlags1    : 7;      // ?? 0 or 1
	WORD  wNodesExtra;               // ?? 0
	BYTE  btGroupMultiplierPowerOf2; // 6 in most cases, somewhere 4
	BYTE  btNodesMultiplierPowerOf2; // 0 or 1
	BYTE  btLinkRefSize;             // 0x4 or 0x5==sizeof(CNodLinkRef)
	BYTE  btNodes0;                  // ?? 0

	// Extra routing data for links.
	DWORD dwLinksOffset;
	DWORD dwLinksLength;
	DWORD dwLinksExtra;              // 0

	// External nodes.
	DWORD dwExternalNodesOffset;
	DWORD dwExternalNodesLength;
	BYTE  btExternalNodesRecSize;    // Always 9

	BYTE  bt3;                       // 0
	DWORD dw3;                       // 0
// 3f:
};

	// NOD1, per node.
struct CNodNode {
	// To be aligned and multiplied by (1<<CNodHeader::GroupMultiplierPowerOf2);
	BYTE btGroupOffset;

	BYTE fUnknown012          : 3, // ??
	     bIsExternal          : 1,
	     bHasTurnRestrictions : 1,
		 bLargePosOffset      : 1,
		 bHasLinks            : 1,
		 bUnknown7            : 1;
};

	// NOD1, per neighbour node.
struct CNodLinkHdr {
	WORD fDestClass  : 3,
	     fCurveLen   : 3,
	     bForward    : 1,
	     bNewHeading : 1,

	     fNode       : 6, // If bFarLink==1, this is index in group's table of nodes.
		                  // If bFarLink==0, this is high byte of relative offset to neighbour node.
	     bFarNode    : 1, // Link is to different group.
	     bLast       : 1;
};

	// NOD1, per node group.
struct CNodGroup {
	BYTE fRestrictionsBytesInSize : 2,
	     bUnknown2                : 1,
		 bUnknown3                : 1, // ?? 0, sometimes =1
		 bUnknown4567             : 4;

	// Center point for all nodes in the group.
	B3_t b3Lon0;
	B3_t b3Lat0;

	BYTE btLinks;    // Number of CNodLinkRef records.
	BYTE btNodes;    // Number of 3-byte offsets to nodes in other groups.
};

	// NOD1, record in the link table of group.
struct CNodLinkRef {
	DWORD b3NetLabelOffset : 22,
	      bNoDelivery      : 1,
	      bNoEmergency     : 1,
	      fSpeedLimit      : 3,
	      bOneWay          : 1,
		  fRouteClass      : 3,
		  bToll            : 1;

	// NOTE: optional byte, presented only if CNodHeader::btLinkRefSize==5.
	BYTE  bNoCar           : 1,
	      bNoBus           : 1,
		  bNoTaxi          : 1,
		  bUnknown3        : 1,  // ??
		  bNoPedestrian    : 1,
	      bNoBicycle       : 1,
		  bNoTruck         : 1,
		  bUnknown7        : 1;  // ??
};

	// NOD1, header of record in the turn restrictions table of group.
struct CNodTurnRestriction {
	BYTE btHdr;         // ?? 0x05 or 0x15 or 0x25
	BYTE bTimeData : 1, // Time-depending restriction.
	     bUnknown1 : 1, // ??
		 bUnknown2 : 1, // ??
	     bVehicles : 1, // Additional vehicle 1-byte bitmask follows, ?? probably defines kinds of vehicles.
		 bUnknown4 : 1, // ??
		 fLinks    : 3; // [fLinks+1] nodes follows (2 bytes per node), then [fLinks] link indices (1 byte per link)
	BYTE btExtra;       // ?? always 00 - may be size of extra data?
};

	// NOD2.
struct CNodLink {
	DWORD bUnknown0    : 1, // ?? 1
	      fSpeedLimit  : 3,
	      fRouteClass  : 3,
	      bUnknown7    : 1, // ?? 0
	      b3NodeOffset : 24;

	WORD wNodes;
};

	// NOD3.
struct CNodExternal {
	B3_t b3Lon;
	B3_t b3Lat;
	B3_t b3NodeOffset;
};


	///////////////////////////////////////////////////////////
	//
	// Header of MDR sub-file.
	//
struct CMdrHeader : public CSubFileHeader {
	// TODO: research.
};


	///////////////////////////////////////////////////////////
	//
	// Header of SRT sub-file.
	//
struct CSrtHeader : public CSubFileHeader {
	// TODO: research.
};

# ifdef _MSC_VER
#	pragma pack (pop)
# endif


	// Parse 6-bit code to 8-bit ASCII one.
void Append6bitSymbol (int & _state, BYTE _bt, char _buf [128], size_t & _pos);

const float c_fGarminUnit = 180.f/(1UL << 23);

float ConvertLongToDeg (long _l);
float ConvertB3ToDeg (const_B3_t _data);
long ConvertB2ToGU (size_t _cShiftBits, long _2bRelative, long _3bBase);

	// Parse compressed coordinates bitstream (RGN).
# include "PtAlgo.h"
void GetPointsFromBitstream (
	const BYTE * _pBitStream, size_t _cBytes, size_t _cShiftBits, BYTE _btBitDef, bool _bExtraBit,
	float _fMapMeanX,
	long _lX0, long _lY0, points_t & _points, bool * _pbExtras = NULL, bool _bMarine = false
);

# endif // __GarminImg_H_
