//
// GPSMapEdit
// (c) Konstantin Galichsky (kg@geopainting.com), 2002-2009
// Garmin and MapSource are registered trademarks or trademarks of Garmin Ltd. or one of its subsidiaries.
//
// Garmin IMG parser.
//

# include "StdAfx.h"
# include "GarminImg.h"
# include "GarminImgContainer.h"
# include "Map.h"
# include "Bookmark.h"
# include "Types.h"
# include "GarminTypes.h"
# include "GenerateRGNodes.h"
# include "Load.h"
# include "Log.h"
# include "Status.h"
# include "ParseHelpers.h"
# include "PrintUnits.h"
# include "fast_floor.h"
# include "Wrappers.h"
# include "Globals.h"

struct less_point {
	bool operator () (const point_t & _pt1, const point_t & _pt2) const {
		if (_pt1.x == _pt2.x)
			return _pt1.y < _pt2.y;
		return _pt1.x < _pt2.x;
	}
};

static
double _Distance2 (const point_t & _pt1, const point_t & _pt2, float _fCosY2) {
	return double (_pt1.x - _pt2.x)*(_pt1.x - _pt2.x)*_fCosY2 + double (_pt1.y - _pt2.y)*(_pt1.y - _pt2.y);
}

static
size_t _FindNearestPoint (const point_t & _pt, const points_t & _points, float _fCosY2) {
	const size_t cPoints = _points.size ();
	if (cPoints == 0)
		return -1;

	size_t cRet = -1;
	double d2_min = 0;
	for (size_t c = 0; c < cPoints; ++ c) {
		const point_t & pt = _points [c];
		const double d2 = _Distance2 (pt, _pt, _fCosY2);
		if (d2_min > d2 || c == 0) {
			d2_min = d2;
			cRet = c;
		}
	}
	return cRet;
}

typedef std::multimap<DWORD, rgn_t *> id2mrgn_t;
typedef std::pair<id2mrgn_t::const_iterator, id2mrgn_t::const_iterator> id2mrgn_range_t;

static
bool _FindNearestPoint (const point_t & _pt, DWORD _dwRoadID, const id2mrgn_t & _roadid2rgn, float _fCosY2, point_t & _ptNearest, const point_t * _pExceptPoint = NULL) {
	const id2mrgn_range_t range = _roadid2rgn.equal_range (_dwRoadID);
	assert (range.first != range.second);
	if (range.first == range.second)
		return false;

	point_t ptRet;
	bool bRet = false;
	double d2_min = 0;
	for (id2mrgn_t::const_iterator ir = range.first; ir != range.second; ++ ir) {
		rgn_t * const pRgn = ir->second;

		assert (! pRgn->elements.empty ());
		const points_t & points = pRgn->elements [0].points;
		const size_t cPoints = points.size ();
		for (size_t c = 0; c < cPoints; ++ c) {
			const point_t & pt = points [c];
			if (_pExceptPoint && * _pExceptPoint == pt)
				continue;
			const double d2 = _Distance2 (pt, _pt, _fCosY2);
			if (d2_min > d2 || bRet == false) {
				d2_min = d2;
				_ptNearest = pt;
				bRet = true;
			}
		}
	}
	return bRet;
}

static
bool _FindCrossingPoint (const point_t & _pt0, DWORD _dwRoadID1, DWORD _dwRoadID2, const id2mrgn_t & _roadid2rgn, float _fCosY2, point_t & _pt) {
	const id2mrgn_range_t range1 = _roadid2rgn.equal_range (_dwRoadID1);
	assert (range1.first != range1.second);
	if (range1.first == range1.second)
		return false;

	const id2mrgn_range_t range2 = _roadid2rgn.equal_range (_dwRoadID2);
	assert (range2.first != range2.second);
	if (range2.first == range2.second)
		return false;

	bool bRet = false;
	double d2_min = 0;
	for (id2mrgn_t::const_iterator ir1 = range1.first; ir1 != range1.second; ++ ir1) {
		rgn_t * const pRgn1 = ir1->second;

		assert (! pRgn1->elements.empty ());
		const points_t & points1 = pRgn1->elements [0].points;
		const size_t cPoints1 = points1.size ();
		for (size_t c1 = 0; c1 < cPoints1; ++ c1) {
			const point_t & pt1 = points1 [c1];

			for (id2mrgn_t::const_iterator ir2 = range2.first; ir2 != range2.second; ++ ir2) {
				rgn_t * const pRgn2 = ir2->second;

				assert (! pRgn2->elements.empty ());
				const points_t & points2 = pRgn2->elements [0].points;
				const size_t cPoints2 = points2.size ();
				for (size_t c2 = 0; c2 < cPoints2; ++ c2) {
					const point_t & pt2 = points2 [c2];
					if (! (pt2 == pt1))
						continue;

					const double d2 = _Distance2 (pt2, _pt0, _fCosY2);
					if (d2_min > d2 || bRet == false) {
						d2_min = d2;
						_pt = pt2;
						bRet = true;
					}
				}
			}
		}
	}
	return bRet;
}

//# define assert(_x)

# ifdef _DEBUG
void _PrintBytes (const char * _strComment, const BYTE * _pBytes, size_t _cBytes) {
	const size_t cCommentLen = _strComment != NULL ? ::strlen (_strComment) : 0;
	char * const strBuf = (char *) alloca (cCommentLen + _cBytes*3 + 1);

	if (_strComment != NULL)
		::strcpy (strBuf, _strComment);

	for (size_t c = 0; c < _cBytes; ++ c) {
		const int hi = _pBytes [c] >> 4;
		const int lo = _pBytes [c] & 0xF;

		strBuf [cCommentLen + c*3]     = ' ';
		strBuf [cCommentLen + c*3 + 1] = hi <= 9 ? hi + '0' : hi - 10 + 'A';
		strBuf [cCommentLen + c*3 + 2] = lo <= 9 ? lo + '0' : lo - 10 + 'A';
	}
	strBuf [cCommentLen + _cBytes*3] = '\0';

	::OutputDebugString (strBuf);
	::OutputDebugString ("\r\n");
}

void _PrintBits (const char * _strComment, const BYTE * _pBytes, size_t _cBytes) {
	::OutputDebugString (_strComment);
	::OutputDebugString (" ");

	for (int nByte = _cBytes - 1; nByte >= 0; -- nByte) {
		const BYTE bt = _pBytes [nByte];
		for (int nBit = 7; nBit >= 0; -- nBit) {
			::OutputDebugString (((bt >> nBit) & 0x1) != 0 ? "1" : "0");
		}
	}

	::OutputDebugString ("\r\n");
}
# endif // _DEBUG

# include "GarminImgLoader2.inl"

///////////////////////////////

struct ComparePoints {
	bool operator () (const point_t & _1, const point_t & _2) const {
		return _1.x != _2.x ? _1.x > _2.x : _1.y > _2.y;
	}
};

///////////////////////////////

struct img_submap_t {
	size_t cRgnOffsetHdr, cRgnOffset;
	size_t cTreOffsetHdr, cTreOffset;
	size_t cLblOffsetHdr, cLblOffset;
	size_t cNetOffsetHdr, cNetOffset;
	size_t cNodOffsetHdr, cNodOffset;
	size_t cGmpOffset;

	string_t strSections;
	string_t strName;
	size_t   cSize;

	img_submap_t () {::memset (this, 0, sizeof (img_submap_t));}
};

struct CGroup {
	size_t cRgnOffset;
	BYTE   btKinds;
	long   lX0, lY0;
	size_t cLevel;
	size_t cShiftBits;

	size_t cPOIsOffset,      cPOIsLength;
	size_t cPointsOffset,    cPointsLength;
	size_t cPolylinesOffset, cPolylinesLength;
	size_t cPolygonsOffset,  cPolygonsLength;

	size_t cPoints2Offset;
	size_t cPolylines2Offset;
	size_t cPolygons2Offset;
};

/////////////////////////////////////////////////////////

class CGarminImgLoader : public CMapLoader {
  public:
	// Override.
	virtual bool IsValid (const char * _pData, size_t _cFileSize) const;
	virtual LoadStatus_t Load (const char * _pData, size_t _cFileSize, const char * _strFileName, map_t * _pMap);

	LoadStatus_t _Load (const BYTE * _pData, size_t _cFileSize, const char * _strFileName, map_t * _pMap);
	LoadStatus_t _LoadSubmap (const string_t & _strSubfileName, const img_submap_t & _sm, size_t _cProgressSize, size_t _cProgressStart);

  private:
	rect_t m_rectMapBound;
	rect_t m_rectMapBoundE;

	rgn_t & _AddPoint (size_t _cLevel, DWORD _dwType, const point_t & _pt,                const char * _strLabel,                      size_t _cBeginPos, size_t _cSize);
	rgn_t & _AddPoly  (size_t _cLevel, DWORD _dwType, RGN_KIND _Kind, points_t & _points, const char * _strLabel, bool _bDirIndicator, size_t _cBeginPos, size_t _cSize);

	void _ParsePoly2  (const BYTE * _p, const BYTE * _pEnd, const CGroup & _group, bool _bPolygon);
	bool _ParseNumber (const BYTE * _p, size_t & _cExtraOffset, bool _bNumberIsInPlace, string_t & _strNumber);

	map_t * m_pMap;
	const BYTE * m_pStart, * m_pEnd;

	bool m_b150;

	// The counter of objects.
//	size_t m_cOrdinal;

	file_refs_t m_file_refs;

	float m_fMapMeanX;

	typedef std::map<size_t, string_t> labels_t;
	labels_t m_labels;
	size_t m_cLblOffsetMultiplierPowerOf2;

	void _ReportPAIWarning (const char * _strFmt, const point_t & _pt, const BYTE * _pExtra);

	///////////////////////////

	struct node_info_t {
		DWORD dwNodeID;
		DWORD dwRoadID;
		point_t ptOriginal;
		rgnode_t * pNode;
	};
	typedef std::map<point_t, node_info_t, less_point> pt2node_t;
	pt2node_t m_pt2node;

	void _AssociateNodeAndPosition (DWORD _dwNodeOffset, DWORD _dwRoadID, const point_t & _ptOriginal, const point_t & _ptFixed);

	struct turn_restriction_t {
		DWORD dwNode [3];
		DWORD dwRoadFrom, dwRoadTo;
		BYTE  btVehiclesMask;
		time_periods_t TimePeriods;
	};
	typedef std::set<DWORD> setIDs_t;
	bool _InsertNode         (DWORD _dwNodeOffset, setIDs_t & _setNodesVisited, size_t _cDepth);
	void _FindNeighbourNodes (DWORD _dwNodeOffset, setIDs_t & _setNodesVisited, size_t _cDepth);

	const BYTE * m_pNodNodes;
	DWORD m_dwNodNodesLength;
	int m_nNodesGroupMultiplierPowerOf2;
	int m_nNodesMultiplierPowerOf2;

	size_t m_cLinkRefSize;

	DWORD _GetNodeGroupOffset (DWORD _dwNodeOffset) const;
	bool _GetNodePosition (DWORD _dwNodeOffset, point_t & _pt) const;
	bool _IsValidNode (DWORD _dwNodeOffset) const;
};

class CGarminRgnLoader : public CGarminImgLoader {
  public:
	// Override.
	virtual bool IsValid (const char * _pData, size_t _cFileSize) const {
		return CGarminImgLoader::IsValid (_pData + 0x3C, _cFileSize - 0x3C);
	}
	virtual LoadStatus_t Load (const char * _pData, size_t _cFileSize, const char * _strMapFileName, map_t * _pMap) {
		return CGarminImgLoader::Load (_pData + 0x3C, _cFileSize - 0x3C, _strMapFileName, _pMap);
	}
};

DECLARE_MAP_LOADER (CGarminImgLoader, "garmin-img", "img", "Garmin MapSource map (*.img)");
DECLARE_MAP_LOADER (CGarminRgnLoader, "garmin-rgn", "rgn", "Garmin MapSource POI map (*.rgn)");

bool CGarminImgLoader::IsValid (const char * _pData, size_t _cFileSize) const {
	if (_cFileSize < 0x200)
		return false;

	// Decode the signature.
	const BYTE btXORMask = _pData [0x10] ^ BYTE ('D');
	char strSignature [6];
	for (size_t c = 0; c < sizeof (strSignature); ++ c)
		strSignature [c] = _pData [0x10 + c] ^ btXORMask;

	// Check the signature.
	if (::memcmp (strSignature, c_strImgSignature, sizeof (strSignature)) == 0)
		return true;

	return false;
}

# define MAXLEVELSCOUNT 9

rgn_t & CGarminImgLoader::_AddPoint (
	size_t _cLevel, DWORD _dwType, const point_t & _pt, const char * _strLabel,
	size_t _cBeginPos, size_t _cSize
) {
	assert (_cLevel < MAXLEVELSCOUNT);

	rgn_t & rgn = m_pMap->CreateRgn (rkPoint);

	// Initialize the rgn.
	rgn.cBeginPos = _cBeginPos;
	rgn.cEndPos   = _cBeginPos + _cSize;
	rgn.cFileLine = -1;
//	rgn.cOrdinal  = m_cOrdinal ++;
	rgn.pFileRef  = & m_file_refs.back ();

	rgn.SetType (m_pMap->pTypeSet, _dwType, rkPoint);

	if (_strLabel && _strLabel [0] != '\0')
		rgn.strLabel = _strLabel;

	rgn.origins.resize (1);
	rgn_origin_t & origin = rgn.origins [0];
	origin.cFileOffset = _cBeginPos;
	origin.cLevel      = _cLevel;
	origin.point       = _pt;

	return rgn;
}

rgn_t & CGarminImgLoader::_AddPoly (
	size_t _cLevel, DWORD _dwType, RGN_KIND _Kind, points_t & _points, const char * _strLabel, bool _bDirIndicator,
	size_t _cBeginPos, size_t _cSize
) {
	assert (_cLevel < MAXLEVELSCOUNT);

	rgn_t & rgn = m_pMap->CreateRgn (_Kind);

	// Initialize the rgn.
	rgn.cBeginPos = _cBeginPos;
	rgn.cEndPos   = _cBeginPos + _cSize;
	rgn.cFileLine = -1;
//	rgn.cOrdinal  = m_cOrdinal ++;
	rgn.pFileRef  = & m_file_refs.back ();

	rgn.SetType (m_pMap->pTypeSet, _dwType, _Kind);

	if (_strLabel && _strLabel [0] != '\0')
		rgn.strLabel = _strLabel;

	if (_bDirIndicator)
		rgn.dwFlags |= rgn_t::fDirIndicator;

	rgn.elements.resize (1);
	rgn_element_t & element = rgn.elements [0];
	element.cFileOffset = _cBeginPos;
	element.cLevel      = _cLevel;
	element.bHole       = false;
	element.points.swap (_points);

	return rgn;
}

void CGarminImgLoader::_ParsePoly2 (const BYTE * _p, const BYTE * _pEnd, const CGroup & _group, bool _bPolygon) {
	const BYTE * p = _p;
	while (p < _pEnd) {
		const CRgnPoly2Data * const pPolyData2 = reinterpret_cast<const CRgnPoly2Data *> (p);
		size_t cDataLen = sizeof (CRgnPoly2Data);
		p += sizeof (CRgnPoly2Data);

		// Type.
		const DWORD dwType = 0x10000 + (WORD (pPolyData2->btType) << 8) + (pPolyData2->btSubType);
		// First point.
		const long lX0 = ConvertB2ToGU (_group.cShiftBits, pPolyData2->b2X0, _group.lX0);
		const long lY0 = ConvertB2ToGU (_group.cShiftBits, pPolyData2->b2Y0, _group.lY0);
		// Zoom level.
		const size_t cLevel = _group.cLevel;

		size_t cBitStreamSize; 
		if ((* p & 0x1) == 0) {
			cBitStreamSize = ((* reinterpret_cast<const WORD *> (p)) >> 2) - 1;
			p        += 2;
			cDataLen += 2;
		} else {
			cBitStreamSize = ((* p) >> 1) - 1;
			++ p;
			++ cDataLen;
		}
		if (cBitStreamSize > 0xFFFF)
			break;

		const BYTE btBitDef = * p;
		++ p;
		++ cDataLen;

		const BYTE * const pBitStream = p;
		p        += cBitStreamSize;
		cDataLen += cBitStreamSize;
		if (p > _pEnd)
			break;
		points_t points;
		GetPointsFromBitstream (pBitStream, cBitStreamSize, _group.cShiftBits, btBitDef, false, m_fMapMeanX, lX0, lY0, points, NULL, true);

		if (! points.empty ()) {
			const bool bPosOK = m_rectMapBoundE.Contains (points [0]) && m_rectMapBoundE.Contains (points [points.size () - 1]);
			assert (bPosOK);
			if (! bPosOK) {
				ReportWarning ("Unexpected format of points bitstream. Some polygons/polylines will be ignored.");
				break;
			}
		}

		// Label.
		const char * strLabel = NULL;
		if (pPolyData2->bHasLabel) {
			const DWORD dwLabel = (* reinterpret_cast<const DWORD *> (p));
			const DWORD dwLabelOffset = dwLabel & 0x3FFFFF;
//			const bool bExtra         = dwLabel & 0x400000;
			p        += 3;
			cDataLen += 3;

//			if (bExtra) {
//				// TODO: research.
//				assert (false);
//			} else
				strLabel = dwLabelOffset ? m_labels [dwLabelOffset << m_cLblOffsetMultiplierPowerOf2].c_str () : NULL;
		}

		rgn_t & rgn = _AddPoly (
			cLevel, dwType, _bPolygon ? rkPolygon : rkPolyline, points, strLabel, false,
			reinterpret_cast<const BYTE *> (pPolyData2) - m_pStart, cDataLen
		);

		assert (pPolyData2->bUnknown0 == 0);
		if (pPolyData2->bUnknown1) {
			// NOTE: this bit is sometimes presented in BlueChart, it shows some additional data.
			break;
		}
	}
}

bool CGarminImgLoader::_ParseNumber (const BYTE * _p, size_t & _cExtraOffset, bool _bNumberIsInPlace, string_t & _strNumber) {
	if ((_p [0] & 0x80) == 0 && ! _bNumberIsInPlace) {
		const DWORD dwNumberOffset = ((DWORD (_p [0]) << 16) + (DWORD (_p [2]) << 8) + _p [1]) & 0x3FFFFF;
		_cExtraOffset += 3;

		const labels_t::iterator i = m_labels.find (dwNumberOffset << m_cLblOffsetMultiplierPowerOf2);
		if (i == m_labels.end ())
			return false;

		_strNumber = i->second;
	} else {
		for (size_t c = 0; c < 10; ++ c) {
			++ _cExtraOffset;
			if (_p [c] == 0xf8)
				// Omit this really empty byte =last+(-)+(-).
				break;

			const bool bLast = (_bNumberIsInPlace ? true : c > 0) && (_p [c] & 0x80) != 0;
			const BYTE bt = _p [c] & 0x7F;
			const size_t c1 = bt / 11;
			const size_t c2 = bt % 11;
			_strNumber.append (1, c1 == 10 ? '-' : '0' + c1);
			if (! bLast || c2 != 10)
				_strNumber.append (1, c2 == 10 ? '-' : '0' + c2);
			if (bLast)
				break;
		}

		const size_t cLen = _strNumber.size ();
		if (cLen > 1 && _strNumber [cLen - 1] == '-') {
			string_t str (_strNumber.c_str (), cLen - 1);
			_strNumber.swap (str);
		}
	}

	return true;
}

void CGarminImgLoader::_ReportPAIWarning (const char * _strFmt, const point_t & _pt, const BYTE * _pExtra) {
	char strPosition [32];
	PrintPosition (strPosition, _pt);
	ReportWarning (_pExtra - m_pStart, _strFmt, strPosition);
}

LoadStatus_t CGarminImgLoader::Load (const char * _pData, size_t _cFileSize, const char * _strMapFileName, map_t * _pMap) {
	const BYTE btXORMask = _pData [0x10] ^ BYTE ('D');
	if (btXORMask) {
		ReportText ("* XOR:    \t%d", btXORMask);

		//
		// Decode.
		//

		BYTE * const pData2 = new BYTE [_cFileSize];
		CByteArrayHolder bah (pData2);
		::memcpy (pData2, _pData, _cFileSize);
		for (size_t c = 0; c < _cFileSize; ++ c)
			pData2 [c] ^= btXORMask;

		m_b150 = btXORMask == 150;

		return _Load (pData2, _cFileSize, _strMapFileName, _pMap);
	}

	m_b150 = false;
	return _Load (reinterpret_cast<const BYTE *> (_pData), _cFileSize, _strMapFileName, _pMap);
}

LoadStatus_t CGarminImgLoader::_Load (const BYTE * _pData, size_t _cFileSize, const char * _strMapFileName, map_t * _pMap) {
	m_pMap   = _pMap;
	m_pStart = _pData;
	m_pEnd   = _pData + _cFileSize;

	m_file_refs.clear ();
	m_file_refs.push_back (file_ref_t ());
	m_file_refs.back ().strFileName = _strMapFileName;
	m_file_refs.back ().bAttachment = false;

//	m_cOrdinal = 0;

	SetProgressStatus (0);

	// Check the integrity.
//	BYTE btCheckSum = 0;
//	for (size_t c = 0; c < _cFileSize; ++ c)
//		btCheckSum += _pData [c];
//	if (btCheckSum != 0)
//		ReportWarning ("Checksum is incorrect.");


	///////////////////////////////////////////////////////////
	//
	// Parse file header.
	//

	const CImgHeader * const pImgHeader = reinterpret_cast<const CImgHeader *> (_pData);
	m_pMap->pTypeSet = & g_tsGarmin;
	m_pMap->Header.ElevationUnits = unFeet;

	// Name.
	m_pMap->Header.strName.assign (
		pImgHeader->strName,
		GetLineLen (
			pImgHeader->strName,
			pImgHeader->strName + sizeof (pImgHeader->strName)
		)
	);
	m_pMap->Header.strName.append (
		pImgHeader->strName2,
		GetLineLen (
			pImgHeader->strName2,
			pImgHeader->strName2 + sizeof (pImgHeader->strName2) - 1
		)
	);

	// Transparency.
	// NOTE: map is assumed transparent until polygon 0x4b ('Background') will be found.
	m_pMap->Header.Transparent = map_header_t::trY;

	// Block size.
	const size_t cBlockSize = pImgHeader->GetBlockSize ();
	ReportText ("* Block size:\t%d", cBlockSize);

	// Update date.
	ReportText (
		"* Updated:\t%02d/%02d",
		pImgHeader->btUpdateYear > 90 ? 1900 + pImgHeader->btUpdateYear : 2000 + pImgHeader->btUpdateYear,
		pImgHeader->btUpdateMonth
	);

	//
	// Scan header sections.
	//

//	if (pImgHeader->btHasFAT == 0)
//		ReportWarning (0x400, "The map is probably in NT format which is not supported. This file may be failed to read.");

	typedef std::map<string_t, img_submap_t> name2submap_t;
	name2submap_t name2submap;

	char strSubFileName [8];
	strSubFileName [0] = 0;

	// Validate FAT.
	size_t cDataOffset = pImgHeader->dwDataOffset;
	if (cDataOffset == 0) {
		// NOTE: in new IMG files, this field is 0.
		ReportWarning ("No data offset field. This file may be failed to read.");

		// Let us restore its value.
		// First, omit empty blocks.
		cDataOffset = sizeof (CImgHeader) - sizeof (pImgHeader->FAT);
		size_t cIdxBlock;
		for (cIdxBlock = 0; cDataOffset < _cFileSize; ++ cIdxBlock) {
			const CFileIndexBlock & fib = pImgHeader->FAT [cIdxBlock];
			if (fib.btValid != 0)
				break;

			cDataOffset += sizeof (CFileIndexBlock);
		}

		// Second, count non-empty blocks.
		for (; cDataOffset < _cFileSize; ++ cIdxBlock) {
			const CFileIndexBlock & fib = pImgHeader->FAT [cIdxBlock];
			if (fib.btValid != 1)
				break;

			cDataOffset += sizeof (CFileIndexBlock);
		}
	}
	if (
		cDataOffset < sizeof (CImgHeader) || 
		cDataOffset >= _cFileSize
	) {
		ReportWarning ("The file header contains incorrect data offset field.");
		return lsFailed;
	}

	size_t cMpsOffset = 0;
	size_t cMpsLength = 0;
	std::vector<string_t> submap_ids;

	// Explore FAT.
	assert (sizeof (CImgHeader) - sizeof (pImgHeader->FAT) == 0x600);
	assert (sizeof (CFileIndexBlock) == 0x200);
	const size_t cIdxBlocks = (cDataOffset - (sizeof (CImgHeader) - sizeof (pImgHeader->FAT)))/sizeof (CFileIndexBlock);
	for (size_t cIdxBlock = 0; cIdxBlock < cIdxBlocks; ++ cIdxBlock) {
		const CFileIndexBlock & fib = pImgHeader->FAT [cIdxBlock];
		if (fib.btValid != 1 || fib.btIndex > 0)
			continue;

		if (
			::memcmp (fib.strName, "        ", sizeof (fib.strName)) == 0 &&
			::memcmp (fib.strType, "   ",      sizeof (fib.strType)) == 0
		)
			continue;
		const string_t strSubFileName (fib.strName, sizeof (fib.strName));

		if (name2submap.find (strSubFileName) == name2submap.end ())
			submap_ids.push_back (strSubFileName);
		img_submap_t & submap = name2submap [strSubFileName];

		if (! submap.strSections.empty ())
			submap.strSections.append (";");
		submap.strSections.append (fib.strType, sizeof (fib.strType));
		submap.cSize += fib.dwLength;

		// TODO: check FAT 'defragmentation'.

		if (fib.wStartBlock*cBlockSize + fib.dwLength > _cFileSize)
			ReportWarning ("The file is damaged: header refers to data out of file end.");

		if (::memcmp        (fib.strType, "RGN", sizeof (fib.strType)) == 0) {
			submap.cRgnOffset = fib.wStartBlock*cBlockSize;
			submap.cRgnOffsetHdr = submap.cRgnOffset;
		} else if (::memcmp (fib.strType, "TRE", sizeof (fib.strType)) == 0) {
			submap.cTreOffset = fib.wStartBlock*cBlockSize;
			submap.cTreOffsetHdr = submap.cTreOffset;
		} else if (::memcmp (fib.strType, "LBL", sizeof (fib.strType)) == 0) {
			submap.cLblOffset = fib.wStartBlock*cBlockSize;
			submap.cLblOffsetHdr = submap.cLblOffset;
		} else if (::memcmp (fib.strType, "NET", sizeof (fib.strType)) == 0) {
			submap.cNetOffset = fib.wStartBlock*cBlockSize;
			submap.cNetOffsetHdr = submap.cNetOffset;
		} else if (::memcmp (fib.strType, "NOD", sizeof (fib.strType)) == 0) {
			submap.cNodOffset = fib.wStartBlock*cBlockSize;
			submap.cNodOffsetHdr = submap.cNodOffset;
		} else if (::memcmp (fib.strType, "GMP", sizeof (fib.strType)) == 0) {
			ReportWarning ("The map %s is in NT format which is not supported.", strSubFileName);
# if ! defined (_DEBUG)
			return lsFailed;
# else // _DEBUG
			const size_t cGmpOffset = fib.wStartBlock*cBlockSize;
			const CGmpHeader * const pGmpHeader = reinterpret_cast<const CGmpHeader *> (m_pStart + cGmpOffset);
			if (::memcmp (pGmpHeader->strSignature, "GARMIN GMP", sizeof (pGmpHeader->strSignature)) != 0) {
				ReportWarning (cGmpOffset, "Unexpected format: GMP header.");
			} else {
				submap.cGmpOffset = cGmpOffset;

				submap.cTreOffsetHdr = cGmpOffset + pGmpHeader->dwTreOffset;
				submap.cRgnOffsetHdr = cGmpOffset + pGmpHeader->dwRgnOffset;
				submap.cLblOffsetHdr = cGmpOffset + pGmpHeader->dwLblOffset;
				submap.cNetOffsetHdr = cGmpOffset + pGmpHeader->dwNetOffset;
				submap.cNodOffsetHdr = cGmpOffset + pGmpHeader->dwNodOffset;

				submap.cTreOffset = cGmpOffset;
				submap.cRgnOffset = cGmpOffset;
				submap.cLblOffset = cGmpOffset;
				submap.cNetOffset = cGmpOffset;
				submap.cNodOffset = cGmpOffset;
			}
# endif // _DEBUG
		} else if (
			::memcmp (fib.strName, "MAPSOURC", sizeof (fib.strName)) == 0 &&
			::memcmp (fib.strType, "MPS",      sizeof (fib.strType)) == 0
		) {
			cMpsOffset = fib.wStartBlock*cBlockSize;
			cMpsLength = fib.dwLength;
		} else {
			const string_t strSubFileType (fib.strType, sizeof (fib.strType));
			ReportText ("* Defined: \t%s.%s", strSubFileName.c_str (), strSubFileType.c_str ());
		}
	}

	// Parse MAPSOURC.MPS (the directory of submaps).
	if (cMpsLength != 0) {
		const BYTE * p = _pData + cMpsOffset;
		const BYTE * const pEnd = p + cMpsLength;
		size_t cMpsRecord = 0;
		while (p < pEnd) {
			if (* p != 'L') // NOTE: last record is marked as 'V'.
				break;
			++ p;

			const WORD wLen = * reinterpret_cast<const WORD *> (p);
			p += sizeof (WORD);

			// Parse unknown fields.
			const BYTE * p2 = p;
			const DWORD dw0 = * reinterpret_cast<const DWORD *> (p2);
			p2 += sizeof (DWORD);
			const WORD w1 = * reinterpret_cast<const WORD *> (p2);
			p2 += sizeof (WORD);
			const WORD w2 = * reinterpret_cast<const WORD *> (p2);
			p2 += sizeof (WORD);

			const char * const strName0 = reinterpret_cast<const char *> (p2);
			const size_t cName0Len = ::strlen (strName0);
			p2 += cName0Len + 1;

			const char * const strName1 = reinterpret_cast<const char *> (p2);
			const size_t cName1Len = ::strlen (strName1);
			p2 += cName1Len + 1;

			const char * const strName2 = reinterpret_cast<const char *> (p2);
			const size_t cName2Len = ::strlen (strName2);
			p2 += cName2Len + 1;

			const WORD w3 = * reinterpret_cast<const WORD *> (p2);
			p2 += sizeof (WORD);
			const WORD w4 = * reinterpret_cast<const WORD *> (p2);
			p2 += sizeof (WORD);
			const DWORD dw5 = * reinterpret_cast<const DWORD *> (p2);
			p2 += sizeof (DWORD);

			// Join all string fields.
/*			string_t strName (strName0, cName0Len);
			if (cName1Len) {
				if (! strName.empty ())
					strName.append ("; ");
				strName.append (strName1, cName1Len);
			}
			if (cName2Len) {
				if (! strName.empty ())
					strName.append ("; ");
				strName.append (strName2, cName2Len);
			}*/
			string_t strName (strName1, cName1Len);

			// Save name of submap.
			const string_t & strId = submap_ids [cMpsRecord];
			const name2submap_t::iterator ns = name2submap.find (strId);
			if (ns != name2submap.end ()) {
				img_submap_t & submap = ns->second;
				submap.strName.swap (strName);
			}

//ATLTRACE ("%08x %04x %04x %04x %04x %08x %s\n", dw0, w1, w2, w3, w4, dw5, strName.c_str ());

			p += wLen;
			assert (p2 == p);

			++ cMpsRecord;
		}
	}

	// Count submaps.
	size_t cSubMaps = 0;
	name2submap_t::const_iterator ns;
	for (ns = name2submap.begin (); ns != name2submap.end (); ++ ns) {
		const string_t & strSubName = ns->first;
		const img_submap_t & submap = ns->second;

		if (submap.cRgnOffset != 0 && submap.cTreOffset != 0 && submap.cLblOffset != 0)
			++ cSubMaps;
	}

	// Ask user which submaps are to load.
	CGarminImgContainerItems items;
	size_t cSelectedSubMaps = name2submap.size ();
	if (cSubMaps > 1) {
		for (name2submap_t::const_iterator ns = name2submap.begin (); ns != name2submap.end (); ++ ns) {
			const string_t & strSubName = ns->first;
			const img_submap_t & submap = ns->second;

			CGarminImgContainerItem item;
			item.bSelected   = true;
			item.strID       = strSubName;
			item.strSections = submap.strSections;
			item.cSize       = submap.cSize;
			if (submap.cRgnOffset == 0 || submap.cTreOffset == 0 || submap.cLblOffset == 0) {
				item.strDescription = "(Not a map, just special Garmin data)";
				item.bSelected      = false;
			} else {
				item.strDescription = submap.strName;
			}

			items [strSubName] = item;
		}

		if (! Silent ()) {
			if (! SelectGarminImgContainerItems (items)) {
				ReportWarning ("Cancelled by user.");
				return lsFailed;
			}
		}

		// Count selected submaps.
		cSelectedSubMaps = 0;
		string_t strSelectedName;
		for (CGarminImgContainerItems::const_iterator si = items.begin (); si != items.end (); ++ si) {
			const CGarminImgContainerItem & item = si->second;
			if (item.bSelected) {
				++ cSelectedSubMaps;
				if (cSelectedSubMaps == 1)
					strSelectedName = item.strDescription;
			}
		}

		// Fix map name.
		if (! strSelectedName.empty ())
			m_pMap->Header.strName.swap (strSelectedName);
	}

	// Load submaps.
	LoadStatus_t ls = lsFailed;
	size_t cSubmap = 0;
	bool bHasMapData = false;
	for (ns = name2submap.begin (); ns != name2submap.end (); ++ ns) {
		const string_t & strSubName = ns->first;
		const img_submap_t & submap = ns->second;

		// Omit non-selected subfiles.
		if (! items.empty ()) {
			const CGarminImgContainerItems::const_iterator si = items.find (strSubName);
			if (si == items.end ())
				continue;
			if (! si->second.bSelected)
				continue;
		}

		// Omit incomplete submaps.
		if (submap.cRgnOffset == 0 || submap.cTreOffset == 0 || submap.cLblOffset == 0) {
			ReportWarning (0x600, "Sub-map %s is not complete.", strSubName.c_str ());
			continue;
		}

		if (m_pMap->Header.strID.empty ())
			m_pMap->Header.strID = strSubFileName;

		bHasMapData = true;
		ReportText ("*** Loading sub-map: '%s' ...", strSubName.c_str ());

		if (submap.cNetOffset != 0)
			ReportText ("* Defined: \tNET");
		if (submap.cNodOffset != 0)
			ReportText ("* Defined: \tNOD");

		map_t map;
		map.cHeaderLen            = cDataOffset;
		map.bNoHeader             = false;
		map.pTypeSet              = & g_tsGarmin;
		map.Header.ElevationUnits = unFeet;
		map.Header.Transparent    = map_header_t::trY;
		map.Header.strName        = m_pMap->Header.strName;
		map.Header.strID          = strSubName;

		m_pMap = & map;
		ls = _LoadSubmap (strSubName, submap, 50*(cSubmap + 1)/cSelectedSubMaps - 50*cSubmap/cSelectedSubMaps, 50*cSubmap/cSelectedSubMaps);
		_pMap->Join (map);

		++ cSubmap;
	}

	if (! bHasMapData) {
		ReportWarning (0x600, "The file does not contain any map data.");
		return lsFailed;
	}

	SetProgressStatus (50);

	_pMap->file_refs.swap (m_file_refs);

	return ls;
}

LoadStatus_t CGarminImgLoader::_LoadSubmap (const string_t & _strSubfileName, const img_submap_t & _sm, size_t _cProgressSize, size_t _cProgressStart) {
	///////////////////////////////////////////////////////////
	//
	// Parse TRE sub-file.
	//

	const CTreHeader * const pTreHeader = reinterpret_cast<const CTreHeader *> (m_pStart + _sm.cTreOffsetHdr);
	if (::memcmp (pTreHeader->strSignature, "GARMIN TRE", sizeof (pTreHeader->strSignature)) != 0) {
		ReportWarning (_sm.cTreOffset, "Unexpected format: TRE header is not found.");
		return lsFailed;
	}
	if (pTreHeader->wHeaderLength < 0x74)
        ReportWarning ("Unexpected size of TRE header: %d byte(s).", pTreHeader->wHeaderLength);

	// Get map's bounding rectangle.
	m_rectMapBound.x0 = ConvertB3ToDeg (pTreHeader->b3MapWest);
	m_rectMapBound.x1 = ConvertB3ToDeg (pTreHeader->b3MapEast);
	m_rectMapBound.y1 = ConvertB3ToDeg (pTreHeader->b3MapNorth);
	m_rectMapBound.y0 = ConvertB3ToDeg (pTreHeader->b3MapSouth);
	if (m_rectMapBound.x1 == -180.f)
		m_rectMapBound.x1 = 180.f;
	assert (m_rectMapBound.x0 < m_rectMapBound.x1);
	assert (m_rectMapBound.y0 < m_rectMapBound.y1);

	m_rectMapBoundE.x0 = m_rectMapBound.x0 - m_rectMapBound.Width ()/4;
	m_rectMapBoundE.x1 = m_rectMapBound.x1 + m_rectMapBound.Width ()/4;
	m_rectMapBoundE.y0 = m_rectMapBound.y0 - m_rectMapBound.Height ()/4;
	m_rectMapBoundE.y1 = m_rectMapBound.y1 + m_rectMapBound.Height ()/4;

	m_fMapMeanX = m_rectMapBound.MeanX ();

	// Map generator description.
	// NOTE: the string is stored right after the header and before TRE data.
	const char * const strGenerator = reinterpret_cast<const char *> (m_pStart + _sm.cTreOffset + pTreHeader->wHeaderLength);
	if (* strGenerator)
		ReportText ("* Generated:\t%s", strGenerator);

	//
	// Parse zoom levels.
	//
	const size_t cLevels = pTreHeader->dwLevelsLength/sizeof (CTreZoomLevel);
	m_pMap->Levels.resize (cLevels);

	const bool bLocked = (pTreHeader->btVer & 0x80) != 0;
	const CTreZoomLevel * levels = NULL;
	if (bLocked) {
		# include "GarminImgLoader.inl"
			ReportText ("*********************************************************************************");
			ReportWarning ("The map is locked. (NOTE: such maps are not supported by intention).");
			ReportText ("*********************************************************************************");
			return lsFailed;
		}
	} else {
		if (pTreHeader->btVer != 0)
			ReportWarning ("Unexpected format of TRE header: %d", pTreHeader->btVer);

		levels = reinterpret_cast<const CTreZoomLevel *> (m_pStart + _sm.cTreOffset + pTreHeader->dwLevelsOffset);
	}
	CArrayHolder<CTreZoomLevel> levels_holder (bLocked ? levels : NULL);

	size_t cLevel;
	for (cLevel = 0; cLevel < cLevels; ++ cLevel) {
		m_pMap->Levels [cLevel].cZoom = levels [cLevels - 1 - cLevel].btZoom;
		m_pMap->Levels [cLevel].SetBasicScale (levels [cLevels - 1 - cLevel].btBits);
	}

	//
	// Parse groups.
	//
    const BYTE * const pGroupsData = m_pStart + _sm.cTreOffset + pTreHeader->dwGroupsOffset;

	// Count groups except 0-th zoom level.
	size_t cGroupsN = 0;
	for (cLevel = 1; cLevel < cLevels; ++ cLevel)
		cGroupsN += levels [cLevels - 1 - cLevel].wGroups;
	const size_t cGroups0 = levels [cLevels - 1].wGroups;

	const size_t cGroups = cGroups0 + cGroupsN;
	CGroup * const groups = new CGroup [cGroups];
	CByteArrayHolder bah (reinterpret_cast<BYTE *> (groups));

	float fTreSize2 = 0;
	size_t cTreGroups = 0;

	// Counter of bytes.
	size_t i;
	// Counter of groups.
	size_t k = 0;
    // Full-length records (all except 0-th zoom level).
    for (i = 0; i < cGroupsN*sizeof (CTreGroup); i += sizeof (CTreGroup)) {
		assert (k < cGroups);

		const CTreGroup * const pTreGroup = reinterpret_cast<const CTreGroup *> (pGroupsData + i);
		groups [k].cRgnOffset = pTreGroup->cRgnOffset;
		groups [k].btKinds    = pTreGroup->btKinds;
		groups [k].lX0        = ConvertB3ToDWORD (pTreGroup->b3X0);
		groups [k].lY0        = ConvertB3ToDWORD (pTreGroup->b3Y0);

		groups [k].cPoints2Offset    = -1;
		groups [k].cPolylines2Offset = -1;
		groups [k].cPolygons2Offset  = -1;

        ++ k;
    }
	assert (k == cGroupsN);
    // Short 14-byte records (only for 0-th zoom level).
	assert (i == cGroupsN*sizeof (CTreGroup));
    for (; i + 14 < pTreHeader->dwGroupsLength; i += 14) {
		assert (k < cGroups);
		if (k >= cGroups) {
			ReportWarning ("Unexpected data in TRE header.");
			return lsFailed;
		}

		const CTreGroup * const pTreGroup = reinterpret_cast<const CTreGroup *> (pGroupsData + i);
		groups [k].cRgnOffset = pTreGroup->cRgnOffset;
		groups [k].btKinds    = pTreGroup->btKinds;
		groups [k].lX0        = ConvertB3ToDWORD (pTreGroup->b3X0);
		groups [k].lY0        = ConvertB3ToDWORD (pTreGroup->b3Y0);
		groups [k].cLevel     = 0;

		groups [k].cPoints2Offset    = -1;
		groups [k].cPolylines2Offset = -1;
		groups [k].cPolygons2Offset  = -1;

		fTreSize2 += pTreGroup->wWidth*pTreGroup->wWidth + pTreGroup->wHeight*pTreGroup->wHeight;
		cTreGroups += 2;

        ++ k;
    }
	assert (k == cGroups);
    // NOTE: just 4 bytes rest: expected btKind == 0.
	assert (i + 4 == pTreHeader->dwGroupsLength);
	const CTreGroup * const pTreGroup = reinterpret_cast<const CTreGroup *> (pGroupsData + i);
	if (i + 4 < pTreHeader->dwGroupsLength || pTreGroup->btKinds != 0)
		ReportWarning ("Unexpected format: last TRE group is not empty.");

    k = 0;
    for (cLevel = cLevels; cLevel >= 1; -- cLevel) {
		const BYTE btBits = levels [cLevels - cLevel].btBits;
		const size_t cg   = levels [cLevels - cLevel].wGroups;

        for (size_t c = 0; c < cg; ++ c) {
			assert (k < cGroups);
			groups [k].cLevel     = cLevel - 1;
            groups [k].cShiftBits = 24 - btBits;
            ++ k;
        }
    }
	assert (k == cGroups);

	// Groups data for 'marine' objects.
	if (pTreHeader->wHeaderLength >= 0x9A) {
		const CTreHeader2 * const pTreHeader2 = reinterpret_cast<const CTreHeader2 *> (pTreHeader);

		if (pTreHeader2->dwGroups2Length != 0) {
			const BYTE * const pGroupsData2 = m_pStart + _sm.cTreOffset + pTreHeader2->dwGroups2Offset;
			k = 0;
			for (const BYTE * p = pGroupsData2; p < pGroupsData2 + pTreHeader2->dwGroups2Length; p += pTreHeader2->wGroup2RecSize) {
				if (pTreHeader2->wGroup2RecSize <= 13) {
					const CTreGroup2_size13 * const pGroup2 = reinterpret_cast<const CTreGroup2_size13 *> (p);
					if (k < cGroups) {
						groups [k].cPoints2Offset    = pGroup2->dwPointsOffset;
						groups [k].cPolylines2Offset = pGroup2->dwPolylinesOffset;
						groups [k].cPolygons2Offset  = pGroup2->dwPolygonsOffset;
					}
				} else if (pTreHeader2->wGroup2RecSize >= 17) {
					const CTreGroup2_size17 * const pGroup2 = reinterpret_cast<const CTreGroup2_size17 *> (p);
					if (k < cGroups && pGroup2->dwSize != 0) {
						//groups [k].cPoints2Offset    = pGroup2->dwPointsOffset;
						//groups [k].cPolylines2Offset = pGroup2->dwPolylinesOffset;
						//groups [k].cPolygons2Offset  = pGroup2->dwPolygonsOffset;
/*
00 00 00 00 . 01 . 00 00 00 00 . 5A 00 00 00 . 5E 00 00 00
00 00 00 00 . 01 . 00 00 00 00 . 5A 00 00 00 . 5E 00 00 00
00 00 00 00 . 01 . 00 00 00 00 . 5A 00 00 00 . 5E 00 00 00
00 00 00 00 . 00 . 00 00 00 00 . 5A 00 00 00 . 5E 00 00 00
F8 01 00 00 . 01 . 00 00 00 00 . 5A 00 00 00 . 5E 00 00 00
8F 04 00 00 . 00 . 00 00 00 00 . 5A 00 00 00 . 5E 00 00 00
87 06 00 00 . 01 . 00 00 00 00 . 5A 00 00 00 . 5E 00 00 00
BA 08 00 00 . 00 . 00 00 00 00 . 5A 00 00 00 . 5E 00 00 00
1E 09 00 00 . 00 . 00 00 00 00 . 5A 00 00 00 . 5E 00 00 00

16 0B 00 00 . 01 . 27 00 00 00 . 5A 00 00 00 . 7C 00 00 00
49 0D 00 00 . 00 . 54 00 00 00 . 5E 00 00 00 . BF 00 00 00
AD 0D 00 00 . 00 . 5A 00 00 00 . 5E 00 00 00 . F2 00 00 00
*/
					}
				}
				++ k;
			}
		}
	}
	k = 0;

	// TreSize.
	m_pMap->Header.cTreSize = cTreGroups != 0 ? fast_floor (::sqrt (fTreSize2/cTreGroups) + .5f) : 511;

	// Address order.
	m_pMap->Header.bPOINumberFirst = pTreHeader->bStreetBeforeNumber == 0;
	m_pMap->Header.bPOIZipFirst    = pTreHeader->bZipBeforeCity      != 0;

	// TRE3 (copyright strings in LBL).
	assert (pTreHeader->wCopyrightsRecSize == 3);
	const size_t cCopyrights = pTreHeader->dwCopyrightsLength/sizeof (CTreCopyright);
	const CTreCopyright * const pCopyrights = reinterpret_cast<const CTreCopyright *> (m_pStart + _sm.cTreOffset + pTreHeader->dwCopyrightsOffset);

	///////////////////////////////////////////////////////////
	//
	// Parse LBL sub-file.
	//

	const CLblHeader * const pLblHeader = reinterpret_cast<const CLblHeader *> (m_pStart + _sm.cLblOffsetHdr);
	if (::memcmp (pLblHeader->strSignature, "GARMIN LBL", sizeof (pLblHeader->strSignature)) != 0) {
		ReportWarning (_sm.cLblOffset, "Unexpected format: LBL header is not found.");
		return lsFailed;
	}

	// Optional fields.
	if (pLblHeader->wHeaderLength > 0xB0 && pLblHeader->dwSortingLength != 0) {
		const char * const strSorting = reinterpret_cast<const char *> (m_pStart + _sm.cLblOffset + pLblHeader->dwSortingOffset);
		ReportText ("* Sorting:   \t%s", strSorting);
	}

	// Code page.
	if (pLblHeader->wHeaderLength >= 0xAC)
		m_pMap->Header.wCodePage = pLblHeader->wCodePage;

	// Label coding.
	switch (pLblHeader->btLblCoding) {
		case 6:
			m_pMap->Header.CodeSchema = bpsAscii7;
			m_pMap->Header.wCodePage = 0; // NOTE: some Garmin maps mistakenly have non-ASCII7 code page.
			ReportText ("* Text coding:\t6 bit");
			break;

		case 9:
			m_pMap->Header.CodeSchema = bpsSBCS;
			ReportText ("* Text coding:\t8 bit");
			break;

		case 10:
			m_pMap->Header.CodeSchema = bpsMBCS;
			ReportText ("* Text coding:\t16 bit");
			break;

		default:
			ReportWarning ("Text coding %d-bit per symbol is not supported.", pLblHeader->btLblCoding);
			return lsFailed;
	}

	// Parse labels.
	const BYTE * const pLblData = reinterpret_cast<const BYTE *> (m_pStart + _sm.cLblOffset + pLblHeader->dwDataOffset);
	const size_t cLblDataLength = pLblHeader->dwDataLength;

	m_cLblOffsetMultiplierPowerOf2 = pLblHeader->btLblOffsetMultiplierPowerOf2;

	for (size_t cOffset = 0; cOffset < cLblDataLength;) {
		const size_t cLabelOffset = cOffset;

		// For each label.
		int state = 0;
		char strLabel [128];
		size_t pos = 0;
		do {
			SetProgressStatus (_cProgressStart + (20*_cProgressSize/100)*cOffset/cLblDataLength);
			if (g_bStop)
				return lsInterrupted;

			if (pLblHeader->btLblCoding == 6) {
				// Convert 3 bytes into 4 symbols (6 bits each).
				const BYTE bt1 =  (pLblData [cOffset + 0]) >> 2;
				const BYTE bt2 = ((pLblData [cOffset + 0] & 0x03) << 4) + ((pLblData [cOffset + 1]) >> 4);
				const BYTE bt3 = ((pLblData [cOffset + 1] & 0x0F) << 2) + ((pLblData [cOffset + 2]) >> 6);
				const BYTE bt4 =  (pLblData [cOffset + 2] & 0x3F);

				if (bt1 > 0x2F) {
					++ cOffset;
					break;
				}
				Append6bitSymbol (state, bt1, strLabel, pos);
				if (bt2 > 0x2F) {
					++ cOffset;
					break;
				}
				Append6bitSymbol (state, bt2, strLabel, pos);
				if (bt3 > 0x2F) {
					cOffset += 2;
					break;
				}
				Append6bitSymbol (state, bt3, strLabel, pos);
				if (bt4 > 0x2F) {
					cOffset += 3;
					break;
				}
				Append6bitSymbol (state, bt4, strLabel, pos);

				cOffset += 3;
			} else if (pLblHeader->btLblCoding == 9 || pLblHeader->btLblCoding == 10) {
				const BYTE bt = pLblData [cOffset];
				if (bt == 0) {
					++ cOffset;
					break;
				}

				if (pos < sizeof (strLabel) - 7 + 1) {
					switch (bt) {
						// Interstate highway
						case 0x01: ::memcpy (strLabel + pos, "~[0x01]", 7); pos += 7; break;
						// Highway - shield
						case 0x02: ::memcpy (strLabel + pos, "~[0x02]", 7); pos += 7; break;
						// Highway - round symbol
						case 0x03: ::memcpy (strLabel + pos, "~[0x03]", 7); pos += 7; break;
						// Highway - small symbol - blue/red (E-highway).
						case 0x04: ::memcpy (strLabel + pos, "~[0x04]", 7); pos += 7; break;
						// Highway - big symbol (D-highway).
						case 0x05: ::memcpy (strLabel + pos, "~[0x05]", 7); pos += 7; break;
						// Highway - small symbol - white (Internal highway).
						case 0x06: ::memcpy (strLabel + pos, "~[0x06]", 7); pos += 7; break;

						// Only the second part is shown; with the cursor you see the whole word.
						case 0x1b: ::memcpy (strLabel + pos, "~[0x1b]", 7); pos += 7; break;
						// Only the first part is shown; with the cursor you see the whole word.
						case 0x1c: ::memcpy (strLabel + pos, "~[0x1c]", 7); pos += 7; break;
						// In areas type 0x4A = <map_name>1D<number_of_detail_map>
						// In LBL2/LBL3 = <long_name>1D<short_name>
						case 0x1d: ::memcpy (strLabel + pos, "~[0x1d]", 7); pos += 7; break;
						// Only the second part is shown; with the cursor you see the whole word, separated by a blank.
						case 0x1e: ::memcpy (strLabel + pos, "~[0x1e]", 7); pos += 7; break;
						// Only the first part is shown; with the cursor you see the whole word, separated by a blank.
						case 0x1f: ::memcpy (strLabel + pos, "~[0x1f]", 7); pos += 7; break;

						default:
							assert (bt >= 32);
							strLabel [pos ++] = bt;
					}
				}
				++ cOffset;
			}
		} while (cOffset < cLblDataLength);

		// Align next string offset.
		if (pLblHeader->btLblCoding == 6 && pLblHeader->btLblOffsetMultiplierPowerOf2 == 1) {
			if (cOffset & 0x1)
				++ cOffset;
		}

		// Save the offset -> label association.
		m_labels [cLabelOffset].assign (strLabel, pos);
	}
	SetProgressStatus (_cProgressStart + (20*_cProgressSize/100));

	// Copyright info (referred by TRE).
	for (size_t cc = 0; cc < cCopyrights; ++ cc) {
		const string_t & strCopyright = m_labels [ConvertB3ToDWORD (pCopyrights [cc].b3Offset) << m_cLblOffsetMultiplierPowerOf2];

		// Replace "~[0x1f]" to (c).
		char * const str = reinterpret_cast<char *> (alloca ((strCopyright.size () + 1)*sizeof (char)));
		size_t c = 0;
		for (const char * p = strCopyright.c_str (); * p != '\0'; ++ p)
			if (::strncmp (p, "~[0x1f]", 7) == 0) {
				str [c ++] = '\xA9';
				p += 6;
			} else
				str [c ++] = * p;
		str [c] = '\0';

		if (m_pMap->Header.strCopyright.empty ())
			m_pMap->Header.strCopyright = str;
		else {
			m_pMap->Header.strCopyright.append (1, '|');
			m_pMap->Header.strCopyright.append (str);
		}

		if (cc == 0)
			ReportText ("* Copyright info:\t%s", str);
		else
			ReportText ("* \t             \t%s", str);
	}

	// Countries.
	fvector<country_t *> vecCountries;
	size_t cCountries = 0;
	if (pLblHeader->dwCountriesLength) {
		assert (pLblHeader->wCountriesRecSize >= sizeof (CLblCountry));
		cCountries = pLblHeader->dwCountriesLength/pLblHeader->wCountriesRecSize;
		const BYTE * const pCountries = m_pStart + _sm.cLblOffset + pLblHeader->dwCountriesOffset;

		vecCountries.resize (cCountries);
		for (size_t c = 0; c < cCountries; ++ c) {
			m_pMap->Countries.push_back (country_t ());
			country_t & country = m_pMap->Countries.back ();
			vecCountries [c] = & country;

			const CLblCountry * const pCountry = reinterpret_cast<const CLblCountry *> (pCountries + c*pLblHeader->wCountriesRecSize);

			// Name.
			const DWORD dwNameOffset = ConvertB3ToDWORD (pCountry->b3NameOffset);
			const string_t & strName = m_labels [dwNameOffset << m_cLblOffsetMultiplierPowerOf2];

			const char c_strDelimiter [] = "~[0x1d]";
			const char * const pDelimeter = ::strstr (strName.c_str (), c_strDelimiter);
			if (pDelimeter == NULL) {
				country.strName.assign (strName);
			} else {
				const char * const pAbbreviation = pDelimeter + sizeof (c_strDelimiter)/sizeof (c_strDelimiter [0]) - 1;
				country.strName.assign (strName.c_str (), pDelimeter - strName.c_str ());
				country.strAbbrevName.assign (pAbbreviation);
			}
		}
	}

	// Regions.
	fvector<region_t *> vecRegions;
	size_t cRegions = 0;
	if (pLblHeader->dwRegionsLength) {
		assert (pLblHeader->wRegionsRecSize >= sizeof (CLblRegion));
		cRegions = pLblHeader->dwRegionsLength/pLblHeader->wRegionsRecSize;
		const BYTE * const pRegions = m_pStart + _sm.cLblOffset + pLblHeader->dwRegionsOffset;

		vecRegions.resize (cRegions);
		for (size_t c = 0; c < cRegions; ++ c) {
			m_pMap->Regions.push_back (region_t ());
			region_t & region = m_pMap->Regions.back ();
			vecRegions [c] = & region;

			const CLblRegion * const pRegion = reinterpret_cast<const CLblRegion *> (pRegions + c*pLblHeader->wRegionsRecSize);

			// Name.
			const DWORD dwNameOffset = ConvertB3ToDWORD (pRegion->b3NameOffset);
			const string_t & strName = m_labels [dwNameOffset << m_cLblOffsetMultiplierPowerOf2];

			const char c_strDelimiter [] = "~[0x1d]";
			const char * const pDelimeter = ::strstr (strName.c_str (), c_strDelimiter);
			if (pDelimeter == NULL) {
				region.strName.assign (strName);
			} else {
				const char * const pAbbreviation = pDelimeter + sizeof (c_strDelimiter)/sizeof (c_strDelimiter [0]) - 1;
				region.strName.assign (strName.c_str (), pDelimeter - strName.c_str ());
				region.strAbbrevName.assign (pAbbreviation);
			}

			// Country.
			const size_t cCountry = pRegion->wCountryIndex - 1;
			if (cCountry < cCountries)
				region.pCountry = vecCountries [cCountry];
			else {
				assert (false);
				region.bDeleted = true;
			}
		}
	}

	typedef std::map<country_t *, region_t *> Country2Region_t;
	Country2Region_t DummyRegions;

	// Cities.
	size_t cCities = 0;
	fvector<city_t *> vecCities;
	const BYTE * pCities = NULL;
	if (pLblHeader->dwCitiesLength) {
		assert (pLblHeader->wCitiesRecSize >= sizeof (CLblCity));
		cCities = pLblHeader->dwCitiesLength/pLblHeader->wCitiesRecSize;
		pCities = m_pStart + _sm.cLblOffset + pLblHeader->dwCitiesOffset;

		vecCities.resize (cCities);
		for (size_t c = 0; c < cCities; ++ c) {
			SetProgressStatus (_cProgressStart + (20*_cProgressSize/100) + (10*_cProgressSize/100)*c/cCities);

			const CLblCity * const pCity = reinterpret_cast<const CLblCity *> (pCities + c*pLblHeader->wCitiesRecSize);

//			assert (pCities [c].wRegionIndex != 0);
			if (pCity->wRegionIndex == 0) { // ??
				vecCities [c] = NULL;
				continue;
			}

			m_pMap->Cities.push_back (city_t ());
			city_t & city = m_pMap->Cities.back ();
			vecCities [c] = & city;

			// Name.
			// NOTE: if pCities [c].bPoint=1, then name of city will be determined after TRE group parsing.
			if (pCity->bPoint == 0) {
				const DWORD dwLabelOffset = ConvertB3ToDWORD (pCity->b3LabelOffset);
				assert (dwLabelOffset);
				if (dwLabelOffset)
					city.strName = m_labels [dwLabelOffset << m_cLblOffsetMultiplierPowerOf2];
			}

			// Region.
			if (pCity->bNoRegion == 0) {
				const size_t cRegion = pCity->wRegionIndex - 1;
				if (cRegion >= cRegions) {
					city.bDeleted = true;
					ReportWarning ("Unexpected index of region for findable city.");
					continue;
				}

				city.pRegion = vecRegions [cRegion];
			} else {
				const size_t cCountry = pCity->wRegionIndex - 1;
				assert (cCountry < vecCountries.size ());
				if (cCountry >= vecCountries.size ()) {
					city.bDeleted = true;
					ReportWarning ("Unexpected index of country for findable city.");
					continue;
				}
				country_t * const pCountry = vecCountries [cCountry];

				// Get 'dummy' region associated with the country.
				region_t * pDummyRegion = NULL;
				const Country2Region_t::iterator dri = DummyRegions.find (pCountry);
				if (dri == DummyRegions.end ()) {
					m_pMap->Regions.push_back (region_t ());
					region_t & region = m_pMap->Regions.back ();
					DummyRegions [pCountry] = & region;

					region.pCountry = pCountry;

					pDummyRegion = & region;
				} else
					pDummyRegion = dri->second;

				city.pRegion = pDummyRegion;
			}
		}
	}

	// Zips.
	fvector<zip_t *> vecZips;
	size_t cZips = 0;
	if (pLblHeader->dwZipsLength) {
		assert (pLblHeader->wZipsRecSize >= sizeof (CLblZip));
		cZips = pLblHeader->dwZipsLength/pLblHeader->wZipsRecSize;
		const BYTE * const pZips = m_pStart + _sm.cLblOffset + pLblHeader->dwZipsOffset;

		vecZips.resize (cZips);
		for (size_t c = 0; c < cZips; ++ c) {
			m_pMap->Zips.push_back (zip_t ());
			zip_t & zip = m_pMap->Zips.back ();
			vecZips [c] = & zip;

			const CLblZip * const pZip = reinterpret_cast<const CLblZip *> (pZips + c*pLblHeader->wZipsRecSize);

			// Name.
			const DWORD dwCodeOffset = ConvertB3ToDWORD (pZip->b3CodeOffset);
			zip.strName.assign (m_labels [dwCodeOffset << m_cLblOffsetMultiplierPowerOf2]);
		}
	}

	// POI extra properties.
	const BYTE * pPOIExtraData = NULL;
	if (pLblHeader->dwPOIExtraLength)
		pPOIExtraData = m_pStart + _sm.cLblOffset + pLblHeader->dwPOIExtraOffset;

	// Highways.
	size_t cHighways = 0;
	fvector<highway_t *> vecHighways;
	if (pLblHeader->dwHighwaysLength) {
		assert (pLblHeader->wHighwaysRecSize >= sizeof (CLblHighway));
		cHighways = pLblHeader->dwHighwaysLength/pLblHeader->wHighwaysRecSize;
		const BYTE * const pHighways = m_pStart + _sm.cLblOffset + pLblHeader->dwHighwaysOffset;

		const BYTE * pHighwayExtras = NULL;
		size_t cHighwayExtras = 0;
		if (pLblHeader->dwHighwayExtraLength) {
			assert (pLblHeader->wHighwayExtraRecSize >= sizeof (CLblHighwayExtra));
			cHighwayExtras = pLblHeader->dwHighwayExtraLength/pLblHeader->wHighwayExtraRecSize;
			pHighwayExtras = m_pStart + _sm.cLblOffset + pLblHeader->dwHighwayExtraOffset;
		}

		vecHighways.resize (cHighways);
		for (size_t c = 0; c < cHighways; ++ c) {
			m_pMap->Highways.push_back (highway_t ());
			highway_t & highway = m_pMap->Highways.back ();
			vecHighways [c] = & highway;

			const CLblHighway * const pHighway = reinterpret_cast<const CLblHighway *> (pHighways + c*pLblHeader->wHighwaysRecSize);

			// Name.
			const DWORD dwNameOffset = ConvertB3ToDWORD (pHighway->b3NameOffset) & 0x3FFFFF;
			if (dwNameOffset) {
				highway.strName.assign (m_labels [dwNameOffset << m_cLblOffsetMultiplierPowerOf2]);
				assert (! highway.strName.empty ());
			} else {
				highway.bDeleted = true;
				continue;
			}

			// Region.
			const size_t cExtra = pHighway->wExtraIndex - 1;
			if (cExtra >= cHighwayExtras) {
				highway.bDeleted = true;
				continue;
			}
			const CLblHighwayExtra * const pHighwayExtra = reinterpret_cast<const CLblHighwayExtra *> (pHighwayExtras + cExtra*pLblHeader->wHighwayExtraRecSize);
			const size_t cRegion = pHighwayExtra->wRegionIndex - 1;
			if (cRegion < cRegions)
				highway.pRegion = vecRegions [cRegion];
			else
				highway.bDeleted = true;
		}
	}

	// Exit facilities.
	const CLblExitFacility * pExitFacilities = NULL;
	size_t cExitFacilities = 0;
	if (pLblHeader->dwExitsLength) {
		assert (pLblHeader->wExitsRecSize >= sizeof (CLblExitFacility));
		cExitFacilities = pLblHeader->dwExitsLength/pLblHeader->wExitsRecSize;
		pExitFacilities = reinterpret_cast<const CLblExitFacility *> (m_pStart + _sm.cLblOffset + pLblHeader->dwExitsOffset);
	}

	SetProgressStatus (_cProgressStart + (30*_cProgressSize/100));

	///////////////////////////////////////////////////////////
	//
	// Parse NET sub-file.
	//

	const CNetHeader * pNetHeader = NULL;
	const BYTE * pNetData = NULL;
	if (_sm.cNetOffset) {
		pNetHeader = reinterpret_cast<const CNetHeader *> (m_pStart + _sm.cNetOffsetHdr);
		if (::memcmp (pNetHeader->strSignature, "GARMIN NET", sizeof (pNetHeader->strSignature)) != 0) {
			ReportWarning (_sm.cNetOffset, "Unexpected format: NET header is not found.");
			return lsOK;
		}
//		if (pNetHeader->wHeaderLength < 0x37)
//			ReportWarning ("Unexpected size of NET header: 0x%04x", pNetHeader->wHeaderLength);

		// NOTE: size is pNetHeader->dwDataLength.
		pNetData = m_pStart + _sm.cNetOffset + pNetHeader->dwDataOffset;
	}

	///////////////////////////////////////////////////////////
	//
	// Parse NOD sub-file.
	//

	m_pNodNodes = NULL;
	m_dwNodNodesLength = 0;
	const BYTE * pNodLinks = NULL;
	DWORD dwNodLinksLength = 0;
	const BYTE * pNodExternalNodes = NULL;
	size_t cExternalNodesLen = 0, cExternalNodeSize = 0;
	m_nNodesGroupMultiplierPowerOf2 = 0;
	m_nNodesMultiplierPowerOf2      = 0;
	m_cLinkRefSize = sizeof (CNodLinkRef);
	if (_sm.cNodOffset) {
		const CNodHeader * const pNodHeader = reinterpret_cast<const CNodHeader *> (m_pStart + _sm.cNodOffsetHdr);
		if (::memcmp (pNodHeader->strSignature, "GARMIN NOD", sizeof (pNodHeader->strSignature)) != 0) {
			ReportWarning (_sm.cNodOffset, "Unexpected format: NOD header is not found.");
			return lsOK;
		}
		if (pNodHeader->wHeaderLength < 0x3e)
			ReportWarning ("Unexpected size of NOD header: 0x%04x", pNodHeader->wHeaderLength);

		m_pNodNodes                     = m_pStart + _sm.cNodOffset + pNodHeader->dwNodesOffset;
		m_dwNodNodesLength              = pNodHeader->dwNodesLength;
		m_nNodesGroupMultiplierPowerOf2 = pNodHeader->btGroupMultiplierPowerOf2;
		m_nNodesMultiplierPowerOf2      = pNodHeader->btNodesMultiplierPowerOf2;
		m_cLinkRefSize                  = pNodHeader->btLinkRefSize;

		if (pNodHeader->bLeftSideTraffic)
			m_pMap->Header.bLeftSideTraffic = true;

		pNodLinks                       = m_pStart + _sm.cNodOffset + pNodHeader->dwLinksOffset;
		dwNodLinksLength                = pNodHeader->dwLinksLength;

		pNodExternalNodes               = m_pStart + _sm.cNodOffset + pNodHeader->dwExternalNodesOffset;
		cExternalNodesLen               = pNodHeader->dwExternalNodesLength;
		cExternalNodeSize               = pNodHeader->btExternalNodesRecSize;
	}

	///////////////////////////////////////////////////////////
	//
	// Parse RGN sub-file.
	//

	const CRgnHeader * const pRgnHeader = reinterpret_cast<const CRgnHeader *> (m_pStart + _sm.cRgnOffsetHdr);
	if (::memcmp (pRgnHeader->strSignature, "GARMIN RGN", sizeof (pRgnHeader->strSignature)) != 0) {
		ReportWarning (_sm.cRgnOffset, "Unexpected format: RGN header is not found.");
		return lsFailed;
	}
    if (pRgnHeader->wHeaderLength < 0x1D)
		ReportWarning ("Unexpected size of RGN header: 0x%04x", pRgnHeader->wHeaderLength);

    const BYTE * const pRgnData = m_pStart + _sm.cRgnOffset + pRgnHeader->dwDataOffset;
	const size_t cRgnDataLength = pRgnHeader->dwDataLength;

	// Prepare offsets of RGN1.
	for (k = 0; k < cGroups; ++ k) {
		if (g_bStop)
			return lsInterrupted;

		CGroup & group = groups [k];

        group.cPOIsOffset      = -1;
		group.cPOIsLength      = -1;
		group.cPointsOffset    = -1;
		group.cPointsLength    = -1;
		group.cPolylinesOffset = -1;
		group.cPolylinesLength = -1;
		group.cPolygonsOffset  = -1;
		group.cPolygonsLength  = -1;

        if (group.btKinds == 0)    // Empty group.
			continue;

		const bool bHasPOIs      = (group.btKinds & 0x1) != 0;
		const bool bHasPoints    = (group.btKinds & 0x2) != 0;
		const bool bHasPolylines = (group.btKinds & 0x4) != 0;
        const bool bHasPolygons  = (group.btKinds & 0x8) != 0;

        const size_t cAddKinds = bHasPOIs + bHasPoints + bHasPolylines + bHasPolygons - 1;
		assert (group.cRgnOffset < cRgnDataLength);
		if (group.cRgnOffset >= cRgnDataLength) {
			ReportWarning ("Unexpected data offset in TRE group.");
			return lsFailed;
		}
		const WORD * pOffsets = reinterpret_cast<const WORD *> (pRgnData + group.cRgnOffset);

        if (bHasPOIs)
            group.cPOIsOffset          = group.cRgnOffset + cAddKinds*sizeof (WORD);

        if (bHasPoints) {
            if (! bHasPOIs)
                group.cPointsOffset    = group.cRgnOffset + cAddKinds*sizeof (WORD);
            else
                group.cPointsOffset    = group.cRgnOffset + * (pOffsets ++);
        }

        if (bHasPolylines) {
            if (! bHasPOIs && ! bHasPoints)
                group.cPolylinesOffset = group.cRgnOffset + cAddKinds*sizeof (WORD);
            else
                group.cPolylinesOffset = group.cRgnOffset + * (pOffsets ++);
        }

        if (bHasPolygons) {
            if (cAddKinds == 0)
                group.cPolygonsOffset  = group.cRgnOffset + cAddKinds*sizeof (WORD);
            else
                group.cPolygonsOffset  = group.cRgnOffset + * pOffsets;
        }
	}

	// Prepare lengths of RGN1.
    for (k = 0; k < cGroups; ++ k) {
		if (g_bStop)
			return lsInterrupted;

		CGroup & group = groups [k];
        if (group.btKinds == 0)    // Empty group.
			continue;

		// POI.
        if (group.cPOIsOffset != -1) {
            if (group.cPointsOffset != -1)         // points follow
                group.cPOIsLength = group.cPointsOffset - group.cPOIsOffset;
            else if (group.cPolylinesOffset != -1) // polylines follow
                group.cPOIsLength = group.cPolylinesOffset - group.cPOIsOffset;
            else if (group.cPolygonsOffset != -1)  // polygons follow
                group.cPOIsLength = group.cPolygonsOffset - group.cPOIsOffset;
            else if (k + 1 < cGroups)
				group.cPOIsLength = groups [k + 1].cRgnOffset - group.cPOIsOffset;
			else
				group.cPOIsLength = cRgnDataLength - group.cPOIsOffset;
        }
		// Point.
        if (group.cPointsOffset != -1) {
            if (group.cPolylinesOffset != -1)     // polylines follow
                group.cPointsLength = group.cPolylinesOffset - group.cPointsOffset;
            else if (group.cPolygonsOffset != -1) // polygons follow
                group.cPointsLength = group.cPolygonsOffset - group.cPointsOffset;
            else if (k + 1 < cGroups)
				group.cPointsLength = groups [k + 1].cRgnOffset - group.cPointsOffset;
			else
				group.cPointsLength = cRgnDataLength - group.cPointsOffset;
        }
		// Polyline.
        if (group.cPolylinesOffset != -1) {
            if (group.cPolygonsOffset != -1)      // polygons follow
                group.cPolylinesLength = group.cPolygonsOffset - group.cPolylinesOffset;
            else if (k + 1 < cGroups)
				group.cPolylinesLength = groups [k + 1].cRgnOffset - group.cPolylinesOffset;
			else
				group.cPolylinesLength = cRgnDataLength - group.cPolylinesOffset;
		}
		// Polygon.
        if (group.cPolygonsOffset != -1) {
			if (k + 1 < cGroups)
				group.cPolygonsLength = groups [k + 1].cRgnOffset - group.cPolygonsOffset;
			else
				group.cPolygonsLength = cRgnDataLength - group.cPolygonsOffset;
		}
	}
	SetProgressStatus (_cProgressStart + 30*_cProgressSize/100);

	// 'Marine' Objects.
	const bool bHasObjectsV2 = pRgnHeader->wHeaderLength >= 0x7D;
	const CRgnHeader2 * const pRgnHeader2 = bHasObjectsV2 ? reinterpret_cast<const CRgnHeader2 *> (pRgnHeader) : NULL;

	//
	// Complete names of cities.
	//
	typedef std::map<const CRgnPointData *, city_t *> point2city_t;
	point2city_t point2city;
	for (size_t c = 0; c < cCities; ++ c) {
		if (vecCities [c] == NULL)
			continue;
		city_t & city = * vecCities [c];

		const CLblCity * const pCity = reinterpret_cast<const CLblCity *> (pCities + c*pLblHeader->wCitiesRecSize);

		if (pCity->bPoint == 0) {
			// NOTE: the name is already determined.
			assert (! city.strName.empty ());
			continue;
		}

		const size_t cGroup = pCity->wGroupIndex - 1;
		assert (cGroup < cGroups);
		if (cGroup < cGroups) {
			const CGroup & group = groups [cGroup];

			if (_sm.cGmpOffset != 0)
				// TODO: research.
				continue;

			assert (group.btKinds & 0x2);
			if ((group.btKinds & 0x2) == 0)
				continue;

			const CRgnPointData * pPointData = reinterpret_cast<const CRgnPointData *> (pRgnData + group.cPointsOffset);
			pPointData += pCity->btPointIndex - 1;
			assert (reinterpret_cast<const BYTE *> (pPointData) + sizeof (CRgnPointData) <= pRgnData + group.cPointsOffset + group.cPointsLength);

			assert (pPointData->b3LabelOffset != 0);
			city.strName = m_labels [pPointData->b3LabelOffset << m_cLblOffsetMultiplierPowerOf2];
			assert (! city.strName.empty ());

			point2city.insert (point2city_t::value_type (pPointData, & city));
		} else {
			ReportWarning ("Unexpected index of group.");
		}
	}

	//
	// Parse POIs.
	//
	for (k = 0; k < cGroups; ++ k) {
		SetProgressStatus (_cProgressStart + (30*_cProgressSize/100) + (10*_cProgressSize/100)*k/cGroups);
		if (g_bStop)
			return lsInterrupted;

		const CGroup & group = groups [k];
        if (group.cPOIsOffset == -1)
			continue;

		assert (group.cPOIsLength != -1);
        for (i = group.cPOIsOffset; i < group.cPOIsOffset + group.cPOIsLength; i += sizeof (CRgnPOIData)) {
			const CRgnPOIData * const pPOIData = reinterpret_cast<const CRgnPOIData *> (pRgnData + i);
			if (reinterpret_cast<const BYTE *> (pPOIData + sizeof (CRgnPOIData)) > m_pEnd)
				break;

			// Label.
			const char * strLabel;
			if (pPOIData->bExtra && pPOIExtraData) {
				const CLblPOIExtra * const pPOIExtra = reinterpret_cast<const CLblPOIExtra *> (pPOIExtraData + (pPOIData->b3LabelOffset << pLblHeader->btPOIExtraOffsetMultiplierPowerOf2));
				if (reinterpret_cast<const BYTE *> (pPOIExtra) + sizeof (CLblPOIExtra) >= m_pEnd)
					break;
				strLabel = pPOIExtra->b3LabelOffset ? m_labels [pPOIExtra->b3LabelOffset << m_cLblOffsetMultiplierPowerOf2].c_str () : NULL;
			} else
				strLabel = pPOIData->b3LabelOffset  ? m_labels [pPOIData->b3LabelOffset  << m_cLblOffsetMultiplierPowerOf2].c_str () : NULL;
            // Position.
			point_t pt;
            pt.x = c_fGarminUnit*ConvertB2ToGU (group.cShiftBits, pPOIData->b2X, group.lX0);
            pt.y = c_fGarminUnit*ConvertB2ToGU (group.cShiftBits, pPOIData->b2Y, group.lY0);
            // Type.
            const USHORT ushType = (WORD (pPOIData->btType) << 8) + (pPOIData->bHasSubType ? pPOIData->btSubType : 0);
            // Zoom level.
			const size_t cLevel = group.cLevel;

			if (pPOIData->bHasSubType == 0)
				-- i;

			assert (m_rectMapBoundE.Contains (pt));
			if (! m_rectMapBoundE.Contains (pt)) {
				ReportWarning ("Unexpected position of POI. It will be ignored.");
				continue;
			}

			rgn_t & rgn = _AddPoint (
				cLevel, ushType, pt, strLabel,
				reinterpret_cast<const BYTE *> (pPOIData) - m_pStart, sizeof (CRgnPOIData) - (pPOIData->bHasSubType == 0 ? 1 : 0)
			);

			//
			// Extra properties.
			//
			if (pPOIData->bExtra && pPOIExtraData) {
				const CLblPOIExtra * const pPOIExtra = reinterpret_cast<const CLblPOIExtra *> (pPOIExtraData + (pPOIData->b3LabelOffset << pLblHeader->btPOIExtraOffsetMultiplierPowerOf2));
				if (reinterpret_cast<const BYTE *> (pPOIExtra) + sizeof (CLblPOIExtra) >= m_pEnd)
					break;

				size_t cExtraOffset = 3;
				const BYTE * const pExtra = reinterpret_cast<const BYTE *> (pPOIExtra);

				assert (pPOIExtra->bUnknown == 0);

				if (0x20 <= pPOIData->btType && pPOIData->btType <= 0x27) {
					//
					// Exit with facilities.
					//
					cExtraOffset = 3;

					// ??.
					if (pPOIExtra->bHasFlags) {
						// TODO: research.
						const BYTE btFlags = pExtra [cExtraOffset];
//						assert (btFlags == 0xA0);

						++ cExtraOffset;
					}

					// Street.
					const DWORD dwStreetOffset_ = ConvertB3ToDWORD (pExtra + cExtraOffset);
					const DWORD dwStreetOffset = dwStreetOffset_ & 0x3FFFFF;
					if (dwStreetOffset)
						rgn.pExtra->strStreetDesc = m_labels [dwStreetOffset << m_cLblOffsetMultiplierPowerOf2];

					const bool bHasFacilities = (dwStreetOffset_ & 0x800000) != 0;

					// Overnight parking.
					const bool bOvernightParking = (dwStreetOffset_ & 0x400000) != 0;
					if (bOvernightParking)
						rgn.pExtra->strExtra.append ("OvernightParking=1\r\n");

					cExtraOffset += 3;

					// Highway.
					size_t cHighwayIndex;
					if (cHighways > 255) {
						cHighwayIndex = * reinterpret_cast<const WORD *> (pExtra + cExtraOffset);
						cExtraOffset += 2;
					} else {
						cHighwayIndex = * (pExtra + cExtraOffset);
						++ cExtraOffset;
					}
					if (cHighwayIndex != 0) {
						-- cHighwayIndex;
						assert (cHighwayIndex < cHighways);
						if (cHighwayIndex < cHighways)
							rgn.pExtra->pHighway = vecHighways [cHighwayIndex];
					}

					if (bHasFacilities) {
						// Facilities.
						size_t cFirstFacility;
						if (cExitFacilities > 255) {
							cFirstFacility = * reinterpret_cast<const WORD *> (pExtra + cExtraOffset);
							cExtraOffset += 2;
						} else {
							cFirstFacility = * (pExtra + cExtraOffset);
							++ cExtraOffset;
						}
						-- cFirstFacility;
						assert (cFirstFacility < cExitFacilities);
						if (cFirstFacility < cExitFacilities) {
							const CLblExitFacility * pFacility = pExitFacilities + cFirstFacility;
							size_t cExitCount = 1;
							while (true) {
								// Facility type.
								char buf [32];
								::sprintf (buf, "Exit%d=0x%x,", cExitCount, pFacility->btType);
								rgn.pExtra->strExtra.append (buf);

								// Facility direction.
								switch (pFacility->btDirection) {
									case 0x0: rgn.pExtra->strExtra.append ("N,"); break;
									case 0x1: rgn.pExtra->strExtra.append ("S,"); break;
									case 0x2: rgn.pExtra->strExtra.append ("E,"); break;
									case 0x3: rgn.pExtra->strExtra.append ("W,"); break;
									case 0x4: rgn.pExtra->strExtra.append ("I,"); break;
									case 0x5: rgn.pExtra->strExtra.append ("O,"); break;
									case 0x6: rgn.pExtra->strExtra.append ("B,"); break;
									case 0x7: rgn.pExtra->strExtra.append (","); break;
									default: assert (false);
								}

								// Facilities.
								::sprintf (buf, "0x%x,", pFacility->btFacilities);
								rgn.pExtra->strExtra.append (buf);

								// Description.
								const char * strDescription = pFacility->b3LabelOffset ? m_labels [pFacility->b3LabelOffset << m_cLblOffsetMultiplierPowerOf2].c_str () : NULL;
								if (strDescription)
									rgn.pExtra->strExtra.append (strDescription);

								rgn.pExtra->strExtra.append ("\r\n");

								assert (pFacility->bUnknown == 0);
								assert (pFacility->bUnknown2 == 0);

								if (pFacility->bLast)
									break;
								++ pFacility;
								++ cExitCount;
							}
						}
					}
				} else if (0x29 <= pPOIData->btType && pPOIData->btType <= 0x30) {
					//
					// Searchable POIs.
					//

					// Bit mask of defined fields.
					CLblPOIExtraDef POIExtraDef;
					POIExtraDef.bt = 0;
					if (pPOIExtra->bHasFlags) {
						POIExtraDef.bt = pLblHeader->POIExtraDef.bt;

						BYTE btGlobalBit = 0x1;
						BYTE btLocalBit  = 0x1;
						for (size_t cBit = 0; cBit < 8; ++ cBit) {
							if ((pLblHeader->POIExtraDef.bt & btGlobalBit) != 0) {
								if ((pPOIExtra->btFlags & btLocalBit) == 0)
									POIExtraDef.bt ^= btGlobalBit;
								btLocalBit <<= 1;
							}
							btGlobalBit <<= 1;
						}

						++ cExtraOffset;
					} else
						POIExtraDef.bt = pLblHeader->POIExtraDef.bt;

					// Number.
					if (POIExtraDef.bHasNumber) {
						const BYTE * const pNumber = pExtra + cExtraOffset;
						if (! _ParseNumber (pNumber, cExtraOffset, POIExtraDef.bNumberIsInPlace, rgn.pExtra->strNumber))
							_ReportPAIWarning ("Unknown format for house number in POI address at (%s). The number will be lost.", pt, pExtra);

						// TODO:
						//12- 3   12 # 3
						//12-03   12 # 3
						//12-13   12 APT 3
						//12-23   12 BLDG 3
						//12-33   12 DEPT 3
						//12-43   12 FL 3
						//12-53   12 RM 3
						//12-63   12 STE 3
						//12-73   12 UNIT 3
					}
					// Street.
					if (POIExtraDef.bHasStreet) {
						const DWORD dwStreetOffset = ConvertB3ToDWORD (pExtra + cExtraOffset) & 0x3FFFFF;
						const labels_t::iterator i = m_labels.find (dwStreetOffset << m_cLblOffsetMultiplierPowerOf2);
						if (i != m_labels.end ())
							rgn.pExtra->strStreetDesc = i->second;
						else
							_ReportPAIWarning ("Unexpected street in POI address at (%s). It will be lost.", pt, pExtra);

						cExtraOffset += 3;
					}

					// City.
					if (POIExtraDef.bHasCity) {
						size_t cCityIndex;
						if (cCities > 255) {
							cCityIndex = * reinterpret_cast<const WORD *> (pExtra + cExtraOffset);
							cExtraOffset += 2;
						} else {
							cCityIndex = * (pExtra + cExtraOffset);
							++ cExtraOffset;
						}
						-- cCityIndex;

						if (cCityIndex < cCities)
							rgn.pExtra->pCity = vecCities [cCityIndex];
						else
							_ReportPAIWarning ("Unexpected index of city in POI address at (%s). It will be lost.", pt, pExtra);
					}

					// Zip.
					if (POIExtraDef.bHasZip) {
						size_t cZipIndex;
						if (cZips > 255) {
							cZipIndex = * reinterpret_cast<const WORD *> (pExtra + cExtraOffset);
							cExtraOffset += 2;
						} else {
							cZipIndex = * (pExtra + cExtraOffset);
							++ cExtraOffset;
						}
						-- cZipIndex;

						if (cZipIndex < cZips)
							rgn.pExtra->pZip = vecZips [cZipIndex];
						else
							_ReportPAIWarning ("Unexpected index of zip code in POI address at (%s). It will be lost.", pt, pExtra);
					}

					// Phone.
					if (POIExtraDef.bHasPhone) {
						const BYTE * const pPhone = pExtra + cExtraOffset;
						if (! _ParseNumber (pPhone, cExtraOffset, false, rgn.pExtra->strPhone))
							_ReportPAIWarning ("Unknown format of phone number of POI at (%s). The number will be lost.", pt, pExtra);
					}

					// ??.
					if (POIExtraDef.bHasExit) {
//						assert (false);
						// TODO: research.
// 77 45 80 - BE - DD 47 07 - 19 - DE 00 - C3 57 3D 15 11 80 - ! 00 00 00 00 A4 26
// 02 ED F8 -      F4 A3 05 - 13 - 35 01 - C3 57 34 36 0C 94 - ! 00 00 00 00 93 6F
					}
				} else {
					// TODO: research.
//					assert (false); // failed for marine POIs base map.
				}
			}
		}
	}
	SetProgressStatus (_cProgressStart + (40*_cProgressSize/100));

	//
	// Parse points.
	//
	for (k = 0; k < cGroups; ++ k) {
		SetProgressStatus (_cProgressStart + (40*_cProgressSize/100) + (10*_cProgressSize/100)*k/cGroups);
		if (g_bStop)
			return lsInterrupted;

		const CGroup & group = groups [k];
		if (group.cPointsOffset == -1)
			continue;

		assert (group.cPointsLength != -1);
		for (i = group.cPointsOffset; i < group.cPointsOffset + group.cPointsLength; i += sizeof (CRgnPointData)) {
			const CRgnPointData * const pPointData = reinterpret_cast<const CRgnPointData *> (pRgnData + i);
			if (reinterpret_cast<const BYTE *> (pPointData) + sizeof (CRgnPointData) > m_pEnd)
				break;

			// Label.
			const char * const strLabel = pPointData->b3LabelOffset ? m_labels [pPointData->b3LabelOffset << m_cLblOffsetMultiplierPowerOf2].c_str () : NULL;
			// Position.
			point_t pt;
			pt.x = c_fGarminUnit*ConvertB2ToGU (group.cShiftBits, pPointData->b2X, group.lX0);
			pt.y = c_fGarminUnit*ConvertB2ToGU (group.cShiftBits, pPointData->b2Y, group.lY0);
            // Type.
			const USHORT ushType = pPointData->btType;
            // Zoom level.
			const size_t cLevel = group.cLevel;

			assert (pPointData->bHasSubType == 0);
			assert (pPointData->bExtra == 0);

			assert (m_rectMapBoundE.Contains (pt));
			if (! m_rectMapBoundE.Contains (pt)) {
				ReportWarning ("Unexpected position of POI. It will be ignored.");
				continue;
			}

			rgn_t & rgn = _AddPoint (
				cLevel, ushType << 8, pt, strLabel,
				reinterpret_cast<const BYTE *> (pPointData) - m_pStart, sizeof (CRgnPointData)
			);

			const point2city_t::const_iterator ip = point2city.find (pPointData);
			if (ip != point2city.end ()) {
				rgn.dwFlags |= rgn_t::fIndexedCity;
				rgn.pExtra->pCity = ip->second;
			}

			// Workaround for cgpsmapper.exe bug (incorrect m_rectMapBound in TRE header for re-saved WorldMap).
			if (! m_rectMapBound.Contains (pt)) {
				m_rectMapBound.Extend (pt);
				m_fMapMeanX = m_rectMapBound.MeanX ();
			}
		}
	}
	SetProgressStatus (_cProgressStart + (50*_cProgressSize/100));

	//
	// Parse polylines.
	//

	setIDs_t setNodes;

	typedef std::map<DWORD, rgn_t *> id2rgn_t;
	id2rgn_t nodeid2rgn;

	id2mrgn_t roadid2rgn;

	for (k = 0; k < cGroups; ++ k) {
		SetProgressStatus (_cProgressStart + (50*_cProgressSize/100) + (20*_cProgressSize/100)*k/cGroups);
		if (g_bStop)
			return lsInterrupted;

		if (groups [k].cPolylinesOffset == -1)
			continue;

		i = groups [k].cPolylinesOffset;
		assert (groups [k].cPolylinesLength != -1);
		const size_t cEnd = i + groups [k].cPolylinesLength;
		assert (cEnd <= cRgnDataLength);
		while (i + sizeof (CRgnPolylineData) - 1 <= cEnd) {
			const CRgnPolylineData * const pPolylineData = reinterpret_cast<const CRgnPolylineData *> (pRgnData + i);
			if (reinterpret_cast<const BYTE *> (pPolylineData) + sizeof (CRgnPolylineData) - 1 > m_pEnd)
				break;

			// Label.
			const char * strLabel;
			if (pPolylineData->bInNet && pNetData) {
				const CNetData * const pNet = reinterpret_cast<const CNetData *> (pNetData + (pPolylineData->b3LabelOffset << pNetHeader->btDataOffsetMultiplierPowerOf2));
				if (reinterpret_cast<const BYTE *> (pNet) + sizeof (CNetData) >= m_pEnd)
					break;
				strLabel = pNet->b3LabelOffset          ? m_labels [pNet->b3LabelOffset          << m_cLblOffsetMultiplierPowerOf2].c_str () : NULL;
			} else
				strLabel = pPolylineData->b3LabelOffset ? m_labels [pPolylineData->b3LabelOffset << m_cLblOffsetMultiplierPowerOf2].c_str () : NULL;
			// First point.
			const long lX0 = ConvertB2ToGU (groups [k].cShiftBits, pPolylineData->b2X0, groups [k].lX0);
			const long lY0 = ConvertB2ToGU (groups [k].cShiftBits, pPolylineData->b2Y0, groups [k].lY0);
			// Type.
			const USHORT ushType = pPolylineData->btType;
			// Zoom level.
			const size_t cLevel = groups [k].cLevel;

			size_t cBitStreamSize; 
			if (pPolylineData->bTwoBytesInLen == 0) {
				cBitStreamSize = pPolylineData->btBitstreamLen;
				i += sizeof (CRgnPolylineData) - 1;
			} else {
				cBitStreamSize = pPolylineData->wBitstreamLen;
				i += sizeof (CRgnPolylineData);
			}

			const BYTE btBitDef = pRgnData [i];
			++ i;
			const BYTE * const pBitStream = pRgnData + i;
			i += cBitStreamSize;
			if (i > cEnd)
				break;
			assert (pBitStream + cBitStreamSize <= pRgnData + cRgnDataLength);
			if (pBitStream + cBitStreamSize > pRgnData + cRgnDataLength) {
				ReportWarning ("Unknown format of polyline. It will be ignored.");
				break;
			}

			points_t points;
			bool baPointExtras [256];
			GetPointsFromBitstream (pBitStream, cBitStreamSize, groups [k].cShiftBits, btBitDef, pPolylineData->bExtraBit, m_fMapMeanX, lX0, lY0, points, baPointExtras);
			const size_t cPoints = points.size ();
			assert (cPoints != 0 && m_rectMapBoundE.Contains (points [0]));
			if (cPoints != 0 && ! m_rectMapBoundE.Contains (points [0])) {
				ReportWarning (pBitStream - m_pStart, "Unexpected format of polyline points. It will be ignored.");
				continue;
			}

			rgn_t & rgn = _AddPoly (
				cLevel, ushType, rkPolyline, points, strLabel, pPolylineData->bDirIndicator,
				reinterpret_cast<const BYTE *> (pPolylineData) - m_pStart, sizeof (CRgnPolylineData) - (pPolylineData->bTwoBytesInLen == 0 ? 1 : 0) + 1 + cBitStreamSize
			);

			// Read extra attributes from NET and NOD.
			if (pPolylineData->bInNet && pNetData) {
				const BYTE * p = pNetData + (pPolylineData->b3LabelOffset << pNetHeader->btDataOffsetMultiplierPowerOf2);

				assert (Garmin_IsRoad (ushType));

				// Parse secondary labels.
				while (true) {
					const CNetData * const pNet = reinterpret_cast<const CNetData *> (p);
					p += 3;
					if (pNet->bSegmented == 1) {
						// Offset to NET2
						p += 3;
					}
					if (pNet->bLastLabel == 1)
						break;
				}

				const CNetDataFlags * const pFlags = reinterpret_cast<const CNetDataFlags *> (p);
				++ p;
				if (pFlags->bLockOnRoad)
					m_pMap->Header.bMG = true;

				// ?? Length.
# ifdef _DEBUG
				const DWORD dwRoadLength = ConvertB3ToDWORD (p);
				const float fLengthInM = dwRoadLength*7.8369461/c_feet_per_m;
# endif // _DEBUG
				p += 3;

				// List of segments per each level.
				size_t cSegmentsTotal = 0;
				size_t cLevels2 = 0;
				while (p < m_pEnd) {
					const BYTE btSegmentsPerLevel = (* p) & 0x7F;
					const bool bLast = ((* p) & 0x80) != 0;
					++ p;

					cSegmentsTotal += btSegmentsPerLevel;
					++ cLevels2;

					if (bLast)
						break;
				}
				assert (cLevels2 <= cLevels);
				// List of segments.
				for (size_t c = 0; c < cSegmentsTotal; ++ c) {
					const DWORD dwRoadInfo = ConvertB3ToDWORD (p);
					p += 3;

					const DWORD dwIndex = (dwRoadInfo & 0xFF);
					const DWORD dwGroup = ((dwRoadInfo & 0xFFFF00) >> 8);
					assert (dwGroup <= cGroups);
				}

				if (pFlags->bHasAddress) {
					m_pMap->Header.bNumbering = true;
# ifdef _DEBUG
//					_PrintBytes (strLabel, p, 20);
# endif // _DEBUG

					const BYTE btBlocks = (* p) + 1;
					++ p;

					const CNetAddressFlags * const pPAFlags = reinterpret_cast<const CNetAddressFlags *> (p);
					p += sizeof (CNetAddressFlags);

					// Zip code data.
					switch (pPAFlags->btZipsFmt) {
						case 0: { // Var len.
							const BYTE btLen = * p;
							++ p;
							p += btLen;
							break;
						}
						case 1: // ??
//							assert (false);
							break;
						case 2: { // 1- or 2-byte index.
							WORD wZipIdx;
							if (cZips > 255) {
								wZipIdx = * reinterpret_cast<const WORD *> (p);
								p += sizeof (WORD);
							} else {
								wZipIdx = * p;
								p += sizeof (BYTE);
							}
							if (wZipIdx == 0) {
								// ??
							} else if (wZipIdx <= vecZips.size ())
								rgn.pExtra->pZip = vecZips [wZipIdx - 1];
							break;
						}
						case 3: // Nothing.
							break;
						default:
							assert (false);
							break;
					}

					// City data.
					switch (pPAFlags->btCitiesFmt) {
						case 0: { // Var len.
							const BYTE btLen = * p;
							++ p;
							p += btLen;
							break;
						}
						case 1: // ??
//							assert (false);
							break;
						case 2: { // 1- or 2-byte index.
							WORD wCityIdx;
							if (cCities > 255) {
								wCityIdx = * reinterpret_cast<const WORD *> (p);
								p += sizeof (WORD);
							} else {
								wCityIdx = * p;
								p += sizeof (BYTE);
							}
							if (wCityIdx == 0) {
								// ??
							} else if (wCityIdx <= vecCities.size ())
								rgn.pExtra->pCity = vecCities [wCityIdx - 1];
							break;
						}
						case 3: // Nothing.
							break;
						default:
							assert (false);
							break;
					}

					// House numbers data.
					switch (pPAFlags->btNumbersFmt) {
						case 0: { // Var len.
							const BYTE btLen = * p;
							++ p;

//							_ParseHouseNumbers (rgn, btBlocks, p, btLen);
							p += btLen;
							break;
						}
						case 1: // ??
//							assert (false);
							break;
						case 2: // Plain number??
//							assert (false);
							break;
						case 3: // Nothing.
							break;
						default:
							assert (false);
							break;
					}
				}

				if (pFlags->bHasNodInfo) {
					const BYTE btHdr = * p;
					++ p;

					if (pNodLinks != NULL && cLevel == 0) {
						m_pMap->Header.bRouting = true;

						DWORD dwLinkOffset = -1;
						if ((btHdr & 0x3) == 0x01) {
							dwLinkOffset = * reinterpret_cast<const WORD *> (p);
							p += 2;
						} else if ((btHdr & 0x3) == 0x02) {
							dwLinkOffset = ConvertB3ToDWORD (p);
							p += 3;
						} else 
							assert (false);

						assert (dwLinkOffset < dwNodLinksLength);
						if (dwLinkOffset < dwNodLinksLength) {
							const BYTE * pr = pNodLinks + dwLinkOffset;
							const CNodLink * const pNodLink = reinterpret_cast<const CNodLink *> (pr);
							pr += sizeof (CNodLink);

							// Routing attributes.
							rgn.pExtra->fSpeedLimit = pNodLink->fSpeedLimit;
							rgn.pExtra->fRouteClass = pNodLink->fRouteClass;
							rgn.pExtra->bOneWay     = pFlags->bOneWay;

							// Road ID.
							const DWORD dwRoadID = pPolylineData->b3LabelOffset;
							rgn.pExtra->dwRoadID = dwRoadID;
							// NOTE: NET1 offset is not unique for single road object, that's why multimap<> is used.
							roadid2rgn.insert (id2rgn_t::value_type (dwRoadID, & rgn));

							// Node.
							const DWORD dwNodeID = pNodLink->b3NodeOffset << m_nNodesMultiplierPowerOf2;
							assert (_IsValidNode (dwNodeID));
							if (_IsValidNode (dwNodeID)) {
								setNodes.insert (dwNodeID);
								nodeid2rgn [dwNodeID] = & rgn;
							}

							// Bitmask (?? of points with routing nodes).
							const size_t cNodes = pNodLink->wNodes;
	//						assert (cNodes <= cPoints);
							const size_t cNodeBytes = (cNodes + 7)/8;
							for (size_t c = 0; c < cNodeBytes; ++ c) {
								const BYTE bt = * pr;
								++ pr;

	/*							size_t cBits = 8;
								if (c + 1 == cNodeBytes)
									cBits = cNodes & 0x7;
								for (size_t cBit = 0; cBit < cBits; ++ cBit) {
									const bool bBit = (bt >> cBit) & 0x1;
								}*/
							}
						}
					}
				}
			}
		}
	}
	SetProgressStatus (_cProgressStart + (70*_cProgressSize/100));

	//
	// Parse polygons.
	//
	for (k = 0; k < cGroups; ++ k) {
		SetProgressStatus (_cProgressStart + (70*_cProgressSize/100) + (15*_cProgressSize/100)*k/cGroups);
		if (g_bStop)
			return lsInterrupted;

		if (groups [k].cPolygonsOffset == -1)
			continue;

		i = groups [k].cPolygonsOffset;
		assert (groups [k].cPolygonsLength != -1);
		const size_t cEnd = i + groups [k].cPolygonsLength;
		assert (cEnd <= cRgnDataLength);
		while (i + sizeof (CRgnPolygonData) - 1 <= cEnd) {
			const CRgnPolygonData * const pPolygonData = reinterpret_cast<const CRgnPolygonData *> (pRgnData + i);
			if (reinterpret_cast<const BYTE *> (pPolygonData) + sizeof (CRgnPolygonData) - 1 > m_pEnd)
				break;

			// Label.
			const char * strLabel;
			if (pPolygonData->bInNet && pNetData) {
				const CNetData * const pNet = reinterpret_cast<const CNetData *> (pNetData + (pPolygonData->b3LabelOffset << pNetHeader->btDataOffsetMultiplierPowerOf2));
				if (reinterpret_cast<const BYTE *> (pNet) + sizeof (CNetData) >= m_pEnd)
					break;
				strLabel = pNet->b3LabelOffset         ? m_labels [pNet->b3LabelOffset         << m_cLblOffsetMultiplierPowerOf2].c_str () : NULL;
			} else
				strLabel = pPolygonData->b3LabelOffset ? m_labels [pPolygonData->b3LabelOffset << m_cLblOffsetMultiplierPowerOf2].c_str () : NULL;
			// First point.
			const long lX0 = ConvertB2ToGU (groups [k].cShiftBits, pPolygonData->b2X0, groups [k].lX0);
			const long lY0 = ConvertB2ToGU (groups [k].cShiftBits, pPolygonData->b2Y0, groups [k].lY0);
			// Type.
			const USHORT ushType = pPolygonData->btType;
			// Zoom level.
			const size_t cLevel = groups [k].cLevel;

			assert (pPolygonData->bUnknown == 0);

			size_t cBitStreamSize; 
			if (pPolygonData->bTwoBytesInLen == 0) {
				cBitStreamSize = pPolygonData->btBitstreamLen;
				i += sizeof (CRgnPolygonData) - 1;
			} else {
				cBitStreamSize = pPolygonData->wBitstreamLen;
				i += sizeof (CRgnPolygonData);
			}

			const BYTE btBitDef = pRgnData [i];
			++ i;
			const BYTE * const pBitStream = pRgnData + i;
			i += cBitStreamSize;
			if (i > cEnd)
				break;
			assert (pBitStream + cBitStreamSize <= pRgnData + cRgnDataLength);
			if (pBitStream + cBitStreamSize > pRgnData + cRgnDataLength) {
				ReportWarning ("Unknown format of polygon. It will be ignored.");
				break;
			}

			points_t points;
			GetPointsFromBitstream (pBitStream, cBitStreamSize, groups [k].cShiftBits, btBitDef, false, m_fMapMeanX, lX0, lY0, points);
			assert (! points.empty () && m_rectMapBoundE.Contains (points [0]));
			if  (! points.empty () && ! m_rectMapBoundE.Contains (points [0])) {
				ReportWarning (pBitStream - m_pStart, "Unexpected format of polygon points. It will be ignored.");
				continue;
			}

			_AddPoly (
				cLevel, ushType, rkPolygon, points, strLabel, false,
				reinterpret_cast<const BYTE *> (pPolygonData) - m_pStart, sizeof (CRgnPolygonData) - (pPolygonData->bTwoBytesInLen == 0 ? 1 : 0) + 1 + cBitStreamSize
			);

			if (ushType == 0x4b) // Background
				m_pMap->Header.Transparent = map_header_t::trN;
		}
	}

	SetProgressStatus (_cProgressStart + (85*_cProgressSize/100));

	//
	// 'Marine' and 3-byte type objects.
	//
	if (pRgnHeader2) {
		const BYTE * const pPoints2Data    = m_pStart + _sm.cRgnOffset + pRgnHeader2->dwPoints2Offset;
		const BYTE * const pPolylines2Data = m_pStart + _sm.cRgnOffset + pRgnHeader2->dwPolylines2Offset;
		const BYTE * const pPolygons2Data  = m_pStart + _sm.cRgnOffset + pRgnHeader2->dwPolygons2Offset;

		for (k = 0; k < cGroups; ++ k) {
			SetProgressStatus (_cProgressStart + (85*_cProgressSize/100) + (15*_cProgressSize/100)*k/cGroups);
			if (g_bStop)
				return lsInterrupted;

			const CGroup & group = groups [k];

			// Points.
			if (group.cPoints2Offset != -1 && pRgnHeader2->dwPoints2Length != 0 && group.cPoints2Offset <= pRgnHeader2->dwPoints2Length) {
				const BYTE * p    = pPoints2Data + group.cPoints2Offset;
				const BYTE * pEnd = pPoints2Data + pRgnHeader2->dwPoints2Length;
				if (k + 1 < cGroups) {
					const CGroup & groupNext = groups [k + 1];
					if (groupNext.cPoints2Offset != -1)
						pEnd = pPoints2Data + groupNext.cPoints2Offset;
				}

				while (p < pEnd) {
					const CRgnPOI2Data * const pPOIData2 = reinterpret_cast<const CRgnPOI2Data *> (p);
					size_t cDataLen = 6;
					if (pPOIData2->bHasLabel)
						cDataLen += 3;
					if (pPOIData2->bHasExtraByte)
						++ cDataLen;

					// Type.
					const DWORD dwType = 0x10000 + (WORD (pPOIData2->btType) << 8) + (pPOIData2->btSubType);
					// Position.
					point_t pt;
					pt.x = c_fGarminUnit*ConvertB2ToGU (group.cShiftBits, pPOIData2->b2X, group.lX0);
					pt.y = c_fGarminUnit*ConvertB2ToGU (group.cShiftBits, pPOIData2->b2Y, group.lY0);
					// Zoom level.
					const size_t cLevel = group.cLevel;
					// Label.
					const char * strLabel = NULL;
					if (pPOIData2->bHasLabel) {
						if (pPOIData2->bExtra && pPOIExtraData) {
							const CLblPOIExtra * const pPOIExtra = reinterpret_cast<const CLblPOIExtra *> (pPOIExtraData + (pPOIData2->b3LabelOffset << pLblHeader->btPOIExtraOffsetMultiplierPowerOf2));
							if (reinterpret_cast<const BYTE *> (pPOIExtra) + sizeof (CLblPOIExtra) >= m_pEnd)
								break;
							strLabel = pPOIExtra->b3LabelOffset ? m_labels [pPOIExtra->b3LabelOffset << m_cLblOffsetMultiplierPowerOf2].c_str () : NULL;
						} else
							strLabel = pPOIData2->b3LabelOffset ? m_labels [pPOIData2->b3LabelOffset << m_cLblOffsetMultiplierPowerOf2].c_str () : NULL;
					}

					p += cDataLen;

					assert (m_rectMapBoundE.Contains (pt));
					if (! m_rectMapBoundE.Contains (pt)) {
						ReportWarning ("Unexpected position of POI. It will be ignored.");
						continue;
					}

					rgn_t & rgn = _AddPoint (cLevel, dwType, pt, strLabel, p - m_pStart, cDataLen);
				}
			}

			// Polylines.
			if (group.cPolylines2Offset != -1 && pRgnHeader2->dwPolylines2Length != 0 && group.cPolylines2Offset <= pRgnHeader2->dwPolylines2Length) {
				const BYTE * const pStart = pPolylines2Data + group.cPolylines2Offset;
				const BYTE *       pEnd   = pPolylines2Data + pRgnHeader2->dwPolylines2Length;
				if (k + 1 < cGroups) {
					const CGroup & groupNext = groups [k + 1];
					if (groupNext.cPolylines2Offset != -1)
						pEnd = pPolylines2Data + groupNext.cPolylines2Offset;
				}

				if (pStart < pEnd)
					_ParsePoly2 (pStart, pEnd, group, false);
			}

			// Polygons.
			if (group.cPolygons2Offset != -1 && pRgnHeader2->dwPolygons2Length != 0 && group.cPolygons2Offset <= pRgnHeader2->dwPolygons2Length) {
				const BYTE * const pStart = pPolygons2Data + group.cPolygons2Offset;
				const BYTE *       pEnd   = pPolygons2Data + pRgnHeader2->dwPolygons2Length;
				if (k + 1 < cGroups) {
					const CGroup & groupNext = groups [k + 1];
					if (groupNext.cPolygons2Offset != -1)
						pEnd = pPolygons2Data + groupNext.cPolygons2Offset;
				}

				if (pStart < pEnd)
					_ParsePoly2 (pStart, pEnd, group, true);
			}
		}
	}

	//
	// Routing graph nodes.
	//

/*	if (setNodes.empty () && m_pNodNodes) {
		// Find missed nodes by brute-force scanning.
		size_t cMissedNodes = 0, cFoundNodes = 0;
		for (DWORD dwNodeOffset = 0; dwNodeOffset < m_dwNodNodesLength - 0x100; dwNodeOffset += (1 << m_nNodesMultiplierPowerOf2)) {
			if (! _IsValidNode (dwNodeOffset))
				continue;

			const DWORD dwGroupOffset = _GetNodeGroupOffset (dwNodeOffset);
			const CNodGroup * const pNodGroup = reinterpret_cast<const CNodGroup *> (m_pNodNodes + dwGroupOffset);

			const BYTE * pn = m_pNodNodes + dwNodeOffset;
			const CNodNode * const pNodNode = reinterpret_cast<const CNodNode *> (pn);
			pn += sizeof (CNodNode);

			if (pNodNode->bHasLinks == 0)
				continue;

			// Position.
			if (pNodNode->bLargePosOffset)
				pn += 4;
			else
				pn += 3;

			const CNodLinkHdr * const pLinkHdr = reinterpret_cast<const CNodLinkHdr *> (pn);
			pn += sizeof (CNodLinkHdr);
			// Node.
			if (pLinkHdr->bFarNode == 1) {
				BYTE btFarNodeIdx = pLinkHdr->fNode;
				if (pLinkHdr->fNode == 0x3F) {
					btFarNodeIdx = * pn;
					++ pn;
				}
				if (btFarNodeIdx >= pNodGroup->btNodes)
					continue;
			} else
				++ pn;

			// Link.
			const BYTE btLinkIdx = * pn;
			++ pn;
			if (btLinkIdx >= pNodGroup->btLinks)
				continue;

			setNodes.insert (dwNodeOffset);
		}

		const size_t cNodes2 = setNodes.size ();
		ATLTRACE ("Brute-force scanned nodes: %d", cNodes2);
	}*/

	// Add external nodes to the set of nodes.
	if (cExternalNodesLen != 0 && m_dwNodNodesLength != 0) {
		assert ((cExternalNodesLen % cExternalNodeSize) == 0);
		const BYTE * p = pNodExternalNodes;
		while (p < pNodExternalNodes + cExternalNodesLen) {
			const CNodExternal * const pExternalNode = reinterpret_cast<const CNodExternal *> (p);
			p += cExternalNodeSize;

			const DWORD dwNodeOffset = ConvertB3ToDWORD (pExternalNode->b3NodeOffset) << m_nNodesMultiplierPowerOf2;
			assert (_IsValidNode (dwNodeOffset));
			setNodes.insert (dwNodeOffset);
		}
	}

	// Traverse the routing graph to fill the set of all nodes.
	setIDs_t setNodesVisited = setNodes;
	setIDs_t::iterator in;
	for (in = setNodes.begin (); in != setNodes.end (); ++ in) {
		const DWORD dwNodeOffset = * in;
		_FindNeighbourNodes (dwNodeOffset, setNodesVisited, 0);
	}

	typedef std::vector<turn_restriction_t> turn_restrictions_t;
	turn_restrictions_t turn_restrictions;

	// For each node group.
	setIDs_t setNodeGroupsVisited;
	for (in = setNodesVisited.begin (); in != setNodesVisited.end (); ++ in) {
		const DWORD dwNodeOffset = * in;
		assert (dwNodeOffset < m_dwNodNodesLength);
		if (dwNodeOffset >= m_dwNodNodesLength)
			continue;
		const CNodNode * const pNodNode = reinterpret_cast<const CNodNode *> (m_pNodNodes + dwNodeOffset);

		// Get common header of the nodes group.
		const DWORD dwGroupOffset = _GetNodeGroupOffset (dwNodeOffset);
		assert (dwGroupOffset < m_dwNodNodesLength);
		if (dwGroupOffset >= m_dwNodNodesLength)
			continue;

		// Exclude repeating visits of the same group.
		const std::pair<setIDs_t::iterator, bool> pr = setNodeGroupsVisited.insert (dwGroupOffset);
		if (pr.second == false)
			continue;

		const BYTE * pg = m_pNodNodes + dwGroupOffset;
		const CNodGroup * const pNodGroup = reinterpret_cast<const CNodGroup *> (pg);
		pg += sizeof (CNodGroup);

		// Verify group.
		const float x = ConvertB3ToDeg (pNodGroup->b3Lon0);
		      float y = ConvertB3ToDeg (pNodGroup->b3Lat0);
		if (y > 90)
			y -= 180;
		assert (m_rectMapBound.Contains (x, y));
		if (! m_rectMapBound.Contains (x, y))
			continue;

		// Links table.
		const BYTE * const pLinksTable = pg;
		pg += pNodGroup->btLinks*m_cLinkRefSize;
		for (size_t cLink = 0; cLink < pNodGroup->btLinks; ++ cLink) {
			const CNodLinkRef & link = * reinterpret_cast<const CNodLinkRef *> (pLinksTable + cLink*m_cLinkRefSize);

			const id2mrgn_range_t range = roadid2rgn.equal_range (link.b3NetLabelOffset);
			for (id2mrgn_t::const_iterator ir = range.first; ir != range.second; ++ ir) {
				rgn_t * const pRgn = ir->second;
				assert (pRgn->pExtra->dwRoadID == link.b3NetLabelOffset);

				pRgn->pExtra->bNoDelivery   = link.bNoDelivery;
				pRgn->pExtra->bNoEmergency  = link.bNoEmergency;
				pRgn->pExtra->bToll         = link.bToll;
				//assert (pRgn->pExtra->bOneWay     == link.bOneWay);
				pRgn->pExtra->bOneWay       = link.bOneWay;
				//assert (pRgn->pExtra->fSpeedLimit == link.fSpeedLimit);
				pRgn->pExtra->fSpeedLimit   = link.fSpeedLimit;
				//assert (pRgn->pExtra->fRouteClass == link.fRouteClass);
				pRgn->pExtra->fRouteClass   = link.fRouteClass;

				if (m_cLinkRefSize >= sizeof (CNodLinkRef)) {
					pRgn->pExtra->bNoCar        = link.bNoCar;
					pRgn->pExtra->bNoBus        = link.bNoBus;
					pRgn->pExtra->bNoTaxi       = link.bNoTaxi;
					pRgn->pExtra->bNoPedestrian = link.bNoPedestrian;
					pRgn->pExtra->bNoBicycle    = link.bNoBicycle;
					pRgn->pExtra->bNoTruck      = link.bNoTruck;
				}
			}
		}

		// Nodes table.
		const BYTE * const pNodesTable = pg;
		pg += 3*pNodGroup->btNodes;

		// Table of turn restrictions.
		if (pNodGroup->fRestrictionsBytesInSize != 0) {
			// Read size of the table.
			size_t cSize = 0;
			if (pNodGroup->fRestrictionsBytesInSize == 1)
				cSize = * pg;
			else if (pNodGroup->fRestrictionsBytesInSize >= 2)
				cSize = * reinterpret_cast<const WORD *> (pg);
			pg += pNodGroup->fRestrictionsBytesInSize;

			const BYTE * const pRestrictions = pg;
			const BYTE * const pRestrictionsEnd = __min (pRestrictions + cSize, m_pEnd);
			while (pg < pRestrictionsEnd) {
				const BYTE * const pRestriction = pg;

				// Turn restriction header.
				const CNodTurnRestriction * const pTR = reinterpret_cast<const CNodTurnRestriction *> (pg);
				pg += sizeof (CNodTurnRestriction);

				bool bUnknownFormat = false;
				if (
					(pTR->btHdr == 0x05 || pTR->btHdr == 0x15 || pTR->btHdr == 0x25) &&
					pTR->bUnknown4 == 0 &&
					pTR->fLinks >= 2 &&
					pTR->btExtra == 0
				) {
					// Kinds of vehicles.
					BYTE btVehiclesMask = 0xFF;
					if (pTR->bVehicles) {
						btVehiclesMask = ~(* pg);
						++ pg;
					}

					turn_restriction_t * pRestr = NULL;

					if (pTR->fLinks >= 3) {
						// Nodes.
						pg += (pTR->fLinks + 1)*sizeof (WORD);
						// Links.
						pg += pTR->fLinks;

						ReportWarning ("%d-node turn restriction found. Such restrictions are not supported and will be ignored.", pTR->fLinks + 1);
					} else if (pTR->fLinks == 2) {
						// Nodes.
						const WORD wNodeFrom   = * reinterpret_cast<const WORD *> (pg);
						pg += sizeof (WORD);
						const WORD wNodeMiddle = * reinterpret_cast<const WORD *> (pg);
						pg += sizeof (WORD);
						const WORD wNodeTo     = * reinterpret_cast<const WORD *> (pg);
						pg += sizeof (WORD);

						// Links.
						const BYTE btLinkIdxFrom = * pg;
						++ pg;
						const BYTE btLinkIdxTo   = * pg;
						++ pg;
						assert (btLinkIdxFrom < pNodGroup->btLinks);
						assert (btLinkIdxTo   < pNodGroup->btLinks);

						DWORD dwNodeOffsetFrom;
						if ((wNodeFrom & 0x8000) != 0)
							dwNodeOffsetFrom = dwGroupOffset - (wNodeFrom & 0x7FFF);
						else
							dwNodeOffsetFrom = ConvertB3ToDWORD (pNodesTable + 3*wNodeFrom) << m_nNodesMultiplierPowerOf2;

						assert (wNodeMiddle & 0x8000);
						const DWORD dwNodeOffsetMiddle = dwGroupOffset - (wNodeMiddle & 0x7FFF);

						DWORD dwNodeOffsetTo;
						if ((wNodeTo & 0x8000) != 0)
							dwNodeOffsetTo = dwGroupOffset - (wNodeTo & 0x7FFF);
						else
							dwNodeOffsetTo = ConvertB3ToDWORD (pNodesTable + 3*wNodeTo) << m_nNodesMultiplierPowerOf2;

						_InsertNode (dwNodeOffsetFrom,   setNodesVisited, 0);
						_InsertNode (dwNodeOffsetMiddle, setNodesVisited, 0);
						_InsertNode (dwNodeOffsetTo,     setNodesVisited, 0);

						if (
							btLinkIdxTo   < pNodGroup->btLinks &&
							btLinkIdxFrom < pNodGroup->btLinks
						) {
							turn_restriction_t tr;
							tr.dwNode [0]     = dwNodeOffsetFrom;
							tr.dwNode [1]     = dwNodeOffsetMiddle;
							tr.dwNode [2]     = dwNodeOffsetTo;
							tr.dwRoadFrom     = reinterpret_cast<const CNodLinkRef *> (pLinksTable + btLinkIdxFrom*m_cLinkRefSize)->b3NetLabelOffset;
							tr.dwRoadTo       = reinterpret_cast<const CNodLinkRef *> (pLinksTable + btLinkIdxTo  *m_cLinkRefSize)->b3NetLabelOffset;
							tr.btVehiclesMask = btVehiclesMask;
							turn_restrictions.push_back (tr);

							pRestr = & turn_restrictions.back ();
						}
					}

					time_periods_t TimePeriods;

					// Time limitations.
					if (pTR->bTimeData != 0) {
						while (pg < pRestrictionsEnd) {
							time_period_t TimePeriod;
							TimePeriod.btFromHour = 0;
							TimePeriod.btFromMin  = 0;
							TimePeriod.btToHour   = 24;
							TimePeriod.btToMin    = 0;

							const BYTE btTimeHdr1 = * pg;
							++ pg;
							assert (btTimeHdr1 == 0 || btTimeHdr1 == 0x40);
							if (btTimeHdr1 != 0 && btTimeHdr1 != 0x40) {
								bUnknownFormat = true;
								break;
							}

							const BYTE btTimeHdr2 = * pg;
							++ pg;
							const bool bDaysOfWeek = (btTimeHdr2 & 0x01) != 0;
							const bool bTimeFrom   = (btTimeHdr2 & 0x04) != 0;
							const bool bTimeTo     = (btTimeHdr2 & 0x08) != 0;
							const bool bLast       = (btTimeHdr2 & 0x80) != 0;
							assert ((btTimeHdr2 & 0x72) == 0);
							if ((btTimeHdr2 & 0x72) != 0) {
								bUnknownFormat = true;
								break;
							}

							BYTE btDaysOfWeek = 0x7F;
							if (bDaysOfWeek) {
								btDaysOfWeek = * pg;
								++ pg;
								assert ((btDaysOfWeek & 0x80) == 0);
							}
							TimePeriod.btDaysOfWeek = btDaysOfWeek;

							if (bTimeFrom) {
								const BYTE btHours0 = * pg;
								++ pg;
								const BYTE btHours = btHours0 & 0x7F;
								assert (btHours < 24);

								BYTE btMinutes = 0;
								if (btHours0 & 0x80) {
									btMinutes = * pg;
									++ pg;
								}
								assert (btMinutes < 60);

								TimePeriod.btFromHour = btHours;
								TimePeriod.btFromMin  = btMinutes;
								TimePeriod.btToHour   = btHours + 1;
								TimePeriod.btToMin    = btMinutes;
							}
							if (bTimeTo) {
								const BYTE btHours0 = * pg;
								++ pg;
								const BYTE btHours = btHours0 & 0x7F;
								assert (btHours < 24);

								BYTE btMinutes = 0;
								if (btHours0 & 0x80) {
									btMinutes = * pg;
									++ pg;
								}
								assert (btMinutes < 60);

								TimePeriod.btToHour = btHours;
								TimePeriod.btToMin  = btMinutes;
							}

							TimePeriods.push_back (TimePeriod);

							if (bLast)
								break;
						}
					}
					if (pRestr && ! TimePeriods.empty ())
						pRestr->TimePeriods.swap (TimePeriods);
				} else
					bUnknownFormat = true;
				if (bUnknownFormat) {
					ReportWarning (pRestriction - m_pStart, "Unknown turn restriction format. The restriction will be ignored.");
# ifdef _DEBUG
					_PrintBytes ("TR:", pRestriction, __min (64, pRestrictions + cSize - pRestriction));
# endif // _DEBUG
					//assert (false);
					break;
				}
			}
		}
	}

	//
	// Fix up positions of nodes in turn restrictions.
	//
	const size_t cTRs = turn_restrictions.size ();
	size_t cTR;
	for (cTR = 0; cTR < cTRs; ++ cTR) {
		const turn_restriction_t & tr = turn_restrictions [cTR];

		// Get 'middle' point.
		point_t ptMiddle0;
		if (! _GetNodePosition (tr.dwNode [1], ptMiddle0)) {
			ReportWarning ("Unexpected value of NodeID=%d in turn restriction.", tr.dwNode [1]);
			continue;
		}
		const float fCosY = ::cos (ptMiddle0.y*c_PI/180);

		point_t ptMiddle;
		if (! _FindCrossingPoint (ptMiddle0, tr.dwRoadFrom, tr.dwRoadTo, roadid2rgn, fCosY*fCosY, ptMiddle)) {
			ReportWarning ("Failed to locate NodeID=%d as crossroads of RoadID=%d and RoadID=%d.", tr.dwNode [1], tr.dwRoadFrom, tr.dwRoadTo);
			continue;
		}

		_AssociateNodeAndPosition (tr.dwNode [1], tr.dwRoadFrom, ptMiddle0, ptMiddle);

		// Get 'from' and 'to' points.
		point_t ptFrom0, ptTo0;
		if (! _GetNodePosition (tr.dwNode [0], ptFrom0)) {
			ReportWarning ("Unexpected value of NodeID=%d in turn restriction.", tr.dwNode [0]);
			continue;
		}
		if (! _GetNodePosition (tr.dwNode [2], ptTo0)) {
			ReportWarning ("Unexpected value of NodeID=%d in turn restriction.", tr.dwNode [2]);
			continue;
		}

		point_t ptFrom, ptTo;
		if (! _FindNearestPoint (ptFrom0, tr.dwRoadFrom, roadid2rgn, fCosY*fCosY, ptFrom, & ptMiddle)) {
			ReportWarning ("Failed to locate NodeID=%d along RoadID=%d.", tr.dwNode [0], tr.dwRoadFrom);
			continue;
		}
		if (! _FindNearestPoint (ptTo0,   tr.dwRoadTo,   roadid2rgn, fCosY*fCosY, ptTo,   & ptMiddle)) {
			ReportWarning ("Failed to locate NodeID=%d along RoadID=%d.", tr.dwNode [2], tr.dwRoadTo);
			continue;
		}

		_AssociateNodeAndPosition (tr.dwNode [0], tr.dwRoadFrom, ptFrom0, ptFrom);
		_AssociateNodeAndPosition (tr.dwNode [2], tr.dwRoadTo,   ptTo0,   ptTo);
	}

	const size_t cNodes = setNodesVisited.size ();
	ATLTRACE ("nodes found %d\n", cNodes);

	//
	// Fix up positions of external nodes.
	//
	size_t cCollisions = 0, cCoincidedNodes = 0;
	for (in = setNodesVisited.begin (); in != setNodesVisited.end (); ++ in) {
		const DWORD dwNodeOffset = * in;
		assert (dwNodeOffset < m_dwNodNodesLength);
		if (dwNodeOffset >= m_dwNodNodesLength)
			continue;
		const BYTE * pn = m_pNodNodes + dwNodeOffset;
		const CNodNode * const pNodNode = reinterpret_cast<const CNodNode *> (pn);
		pn += sizeof (CNodNode);

		// Only external nodes are to be associated with position.
		if (pNodNode->bIsExternal == 0)
			continue;

		// Get common header of the nodes group.
		const DWORD dwGroupOffset = _GetNodeGroupOffset (dwNodeOffset);
		assert (dwGroupOffset < m_dwNodNodesLength);
		if (dwGroupOffset >= m_dwNodNodesLength)
			continue;
		const CNodGroup * const pNodGroup = reinterpret_cast<const CNodGroup *> (m_pNodNodes + dwGroupOffset);
		// Links table.
		const BYTE * const pLinksTable = reinterpret_cast<const BYTE *> (pNodGroup + 1);

		// Parse position.
		if (pNodNode->bLargePosOffset)
			pn += 2*sizeof (SHORT);
		else
			pn += 3;

		point_t ptNodeOriginal;
		if (! _GetNodePosition (dwNodeOffset, ptNodeOriginal))
			continue;
//@		assert (m_rectMapBound.Contains (ptNodeOriginal));
		if (! m_rectMapBound.Contains (ptNodeOriginal))
			continue;

		const float fCosY = ::cos (ptNodeOriginal.y*c_PI/180);

		point_t ptNodeFixed = ptNodeOriginal;
		bool bPosIsFixed = false;

		// Links to neighbour nodes.
		// NOTE: here we get just the first link, if any.
		DWORD dwRoadID = -1;
		if (pNodNode->bHasLinks == 1 && pn < m_pNodNodes + dwGroupOffset) {
			const CNodLinkHdr * const pLinkHdr = reinterpret_cast<const CNodLinkHdr *> (pn);
			pn += sizeof (CNodLinkHdr);

			// Node.
			if (pLinkHdr->bFarNode == 1) {
				assert (pLinkHdr->fNode < pNodGroup->btNodes);
				if (pLinkHdr->fNode == 0x3F)
					++ pn;
			} else
				++ pn;

			// Link.
			const BYTE btLinkIdx = * pn;
			++ pn;
			assert (btLinkIdx < pNodGroup->btLinks);
			if (btLinkIdx < pNodGroup->btLinks) {
				const CNodLinkRef & link = * reinterpret_cast<const CNodLinkRef *> (pLinksTable + btLinkIdx*m_cLinkRefSize);
				dwRoadID = link.b3NetLabelOffset;

				// Get the first link passing through the node.
				if (_FindNearestPoint (ptNodeOriginal, dwRoadID, roadid2rgn, fCosY*fCosY, ptNodeFixed))
					bPosIsFixed = true;
			} else {
# ifdef _DEBUG
				ATLTRACE ("NodeID=%d GroupID=%d LinkIdx=%02x/%02x NodeU012=%d GroupU2=%d GroupU3=%d GroupU4567=%d\n", dwNodeOffset, dwGroupOffset, btLinkIdx, pNodGroup->btLinks, pNodNode->fUnknown012, pNodGroup->bUnknown2, pNodGroup->bUnknown3, pNodGroup->bUnknown4567);
				_PrintBytes ("LINK: ", (const BYTE *) pLinkHdr, 10);
# endif // _DEBUG
			}
		}

		if (! bPosIsFixed) {
			const id2rgn_t::const_iterator i = nodeid2rgn.find (dwNodeOffset);
			if (i != nodeid2rgn.end ()) {
				rgn_t * const pRgn = i->second;

				assert (! pRgn->elements.empty ());
				const points_t & points = pRgn->elements [0].points;
				assert (! points.empty ());
				const size_t cPointNearest = _FindNearestPoint (ptNodeOriginal, points, fCosY*fCosY);
				if (cPointNearest != -1) {
					// Fix up the node position using nearset point of the associated road.
					ptNodeFixed = points [cPointNearest];
					bPosIsFixed = true;
				}
			}
		}

		assert (bPosIsFixed);

		// Detect position collision.
		const pt2node_t::const_iterator ip = m_pt2node.find (ptNodeFixed);
		if (ip != m_pt2node.end ()) {
			const node_info_t & infoPrev = ip->second;

			if (infoPrev.dwNodeID != dwNodeOffset) {
				ATLTRACE ("Nodes %d and %d coincide.\n", infoPrev.dwNodeID, dwNodeOffset);
				++ cCollisions;

				const double d2_prev = _Distance2 (infoPrev.ptOriginal, ptNodeFixed, fCosY*fCosY);
				const double d2_new  = _Distance2 (ptNodeOriginal,      ptNodeFixed, fCosY*fCosY);
				if (d2_prev <= d2_new) {
					if (dwRoadID != -1) {
						const point_t ptExcept = ptNodeFixed;
						_FindNearestPoint (ptNodeOriginal, dwRoadID, roadid2rgn, fCosY*fCosY, ptNodeFixed, & ptExcept);
					}
				} else {
					const point_t ptExcept = ptNodeFixed;
					point_t ptNodeFixed_prev = ptNodeFixed;
					if (_FindNearestPoint (infoPrev.ptOriginal, infoPrev.dwRoadID, roadid2rgn, fCosY*fCosY, ptNodeFixed_prev, & ptExcept)) {
						const pt2node_t::const_iterator ip2 = m_pt2node.find (ptNodeFixed_prev);
						if (ip2 == m_pt2node.end ()) {
							m_pt2node.insert (pt2node_t::value_type (ptNodeFixed_prev, infoPrev));
						} else {
							const node_info_t & infoPrev2 = ip2->second;
							ReportWarning ("NodeID=%d is not identified.\n", infoPrev.dwNodeID);
							++ cCoincidedNodes;
						}
					}
				}
			}
		}

		// Associate the position of node and the node.
		_AssociateNodeAndPosition (dwNodeOffset, dwRoadID, ptNodeOriginal, ptNodeFixed);
	}

	if (cCoincidedNodes >= 5)
		ReportWarning ("Failed to identify %d node(s).", cCoincidedNodes);

	// Create routing nodes.
	typedef std::map<DWORD, rgnode_t *> offset2node_t;
	offset2node_t offset2node;

	if (! m_pt2node.empty ()) {
		layer_t & polylines = m_pMap->layers [rkPolyline];
		rgns_t::iterator iRgn;
		for (iRgn = polylines.rgns.begin (); iRgn != polylines.rgns.end (); ++ iRgn) {
			rgn_t & rgn = * iRgn;
			assert (rgn.IsMutable ());
			if (! m_pMap->pTypeSet->pfIsRoad (rgn))
				continue;

			const size_t cElements = rgn.elements.size ();
			for (size_t c = 0; c < cElements; ++ c) {
				rgn_element_t & element = rgn.elements [c];
				assert (! element.bDeleted);
				if (element.cLevel != 0)
					continue;

				points_t & points = element.points;
				RemoveRepeatingPoints (points, false);
				const size_t cPoints = points.size ();
				for (size_t cPoint = 0; cPoint < cPoints; ++ cPoint) {
					const point_t & pt = points [cPoint];

					// NOTE: this is required for GenerateRGNodes().
					m_pMap->rectBound.Extend (pt);

					const pt2node_t::iterator i = m_pt2node.find (pt);
					if (i == m_pt2node.end ())
						continue;

					node_info_t & info = i->second;
					if (info.pNode == NULL) {
						info.pNode = & m_pMap->CreateNode (info.dwNodeID);
						offset2node [info.dwNodeID] = info.pNode;
					}
					if (! element.pExtra || element.pExtra->PointExtras.size () < cPoints)
						element.pExtra->PointExtras.resize (cPoints);
					element.pExtra->PointExtras [cPoint].SetRGNode (info.pNode, cPoint, cPoints);
				}
			}
		}
	}

# ifdef _DEBUG
// Search for orphan nodes.
/*const size_t cPts = m_pt2node.size ();
for (pt2node_t::const_iterator ii = m_pt2node.begin (); ii != m_pt2node.end (); ++ ii) {
	const node_info_t & info = ii->second;
	if (info.pNode != NULL)
		continue;
	const point_t & pt = ii->first;

	rgn_t & rgn = _AddPoint (0, 0, pt, NULL, 0, 0);
	rgn.pExtra->pBookmark = new bookmark_t;
	rgn.pExtra->pBookmark->FixPos (rgn);
	rgn.pExtra->pBookmark->strType = "orphan";
}*/
# endif // _DEBUG

	// Create turn restrictions.
	const offset2node_t::const_iterator inEnd = offset2node.end ();
	for (cTR = 0; cTR < cTRs; ++ cTR) {
		const turn_restriction_t & tr = turn_restrictions [cTR];

		bool bNodeNotFound = false;
		offset2node_t::const_iterator in [3];
		for (size_t c = 0; c < 3; ++ c) {
			in [c] = offset2node.find (tr.dwNode [c]);
			if (in [c] == inEnd) {
				char strPos [32] = "??";
				point_t ptMiddle;
				if (_GetNodePosition (tr.dwNode [1], ptMiddle))
					PrintPosition (strPos, ptMiddle);

				ReportWarning ("Unexpected NodeID=%d in turn restriction at (%s). The restriction will be ignored.", tr.dwNode [c], strPos);
				bNodeNotFound = true;
				break;
			}
		}
		if (bNodeNotFound)
			continue;
		rgnode_t * const pNodeFrom   = in [0]->second;
		rgnode_t * const pNodeMiddle = in [1]->second;
		rgnode_t * const pNodeTo     = in [2]->second;

		bool bRoadsFound = false;
		const id2mrgn_range_t rangeFrom = roadid2rgn.equal_range (tr.dwRoadFrom);
		const id2mrgn_range_t rangeTo   = roadid2rgn.equal_range (tr.dwRoadTo);
		for (id2mrgn_t::const_iterator irFrom = rangeFrom.first; irFrom != rangeFrom.second; ++ irFrom) {
			rgn_t * const pRoadFrom = irFrom->second;
			for (id2mrgn_t::const_iterator irTo = rangeTo.first; irTo != rangeTo.second; ++ irTo) {
				rgn_t * const pRoadTo = irTo->second;

				bool bFromDirection = false, bToDirection = false;
				if (
					pRoadFrom->GetDirection (pNodeFrom  ->dwNodeID, pNodeMiddle->dwNodeID, bFromDirection) &&
					pRoadTo  ->GetDirection (pNodeMiddle->dwNodeID, pNodeTo    ->dwNodeID, bToDirection)
				) {
					rgrestriction_t & restriction = AddRGNodeRestriction (* m_pMap, pNodeMiddle);
					restriction.pFromRoad      = pRoadFrom;
					restriction.pToRoad        = pRoadTo;
					restriction.bFromDirection = bFromDirection;
					restriction.bToDirection   = bToDirection;
					restriction.btVehicleKinds = tr.btVehiclesMask;
					restriction.TimePeriods    = tr.TimePeriods;

					bRoadsFound = true;
					break;
				}
			}
			if (bRoadsFound)
				break;
		}
		if (! bRoadsFound) {
			char strPos [32] = "??";
			point_t ptMiddle;
			if (_GetNodePosition (tr.dwNode [1], ptMiddle))
				PrintPosition (strPos, ptMiddle);

			ReportWarning (
				"Unexpected turn restriction at (%s): [node %d]->(road %d)->[node %d]->(road %d)->[node %d]",
				strPos,
				tr.dwNode [0], tr.dwRoadFrom,
				tr.dwNode [1], tr.dwRoadTo,
				tr.dwNode [2]
			);
		}
	}

	if (m_dwNodNodesLength != 0) {
		assert (! m_pMap->rectBound.Invalid ());
		if (! m_pMap->rectBound.Invalid ()) {
			m_pMap->fCosY = ::cos (m_pMap->rectBound.MeanY ()*c_PI/180);

			// Create nodes (1) for dead ends (since Garmin IMG do not define them)
			// and also (2) for any lost crossroads.
			GenerateRGNodesInplace (m_pMap);
		}
	}

	////////////////////////////////////////////////////////////////////////
	//
	// External nodes (second pass).
	//
	if (cExternalNodesLen != 0 && m_dwNodNodesLength != 0) {
		const BYTE * p = pNodExternalNodes;
		while (p < pNodExternalNodes + cExternalNodesLen) {
			const CNodExternal * const pExternalNode = reinterpret_cast<const CNodExternal *> (p);
			p += cExternalNodeSize;

			const DWORD dwNodeID = ConvertB3ToDWORD (pExternalNode->b3NodeOffset) << m_nNodesMultiplierPowerOf2;
			const offset2node_t::const_iterator i = offset2node.find (dwNodeID);
			if (i != offset2node.end ()) {
				rgnode_t * const pRGNode = i->second;
				pRGNode->bIsExternal = true;
			} else {
				const float x = ConvertB3ToDeg (pExternalNode->b3Lon);
				      float y = ConvertB3ToDeg (pExternalNode->b3Lat);
				if (y > 90)
					y -= 180;

				char strPos [32];
				PrintPosition (strPos, point_t (x, y));
				ReportWarning ("External node %d lost at (%s).", dwNodeID, strPos);
# ifdef _DEBUG
				rgn_t & rgn = _AddPoint (0, 0, point_t (x, y), NULL, 0, 0);
				rgn.pExtra->pBookmark = new bookmark_t;
				rgn.pExtra->pBookmark->FixPos (rgn);
				rgn.pExtra->pBookmark->strType = "external_orphan";
				char strNodeID [10];
				::sprintf (strNodeID, "%d", dwNodeID);
				rgn.pExtra->pBookmark->attributes ["nodeid"] = strNodeID;

				setIDs_t::const_iterator ii = setNodesVisited.find (dwNodeID);
				assert (ii != setNodesVisited.end ());
# endif // _DEBUG
			}
		}
	}
	
# ifdef _DEBUG
	////////////////////////////////////////////////////////////////////////
	//
	// TODO: research.
	//

	if (pLblHeader->dwPOIIndexLength != 0) {
		assert (pLblHeader->wPOIIndexRecSize == sizeof (CLblPOIIndexItem));
		const size_t cPOIIndexSize = pLblHeader->dwPOIIndexLength/sizeof (CLblPOIIndexItem);
		const CLblPOIIndexItem * const pPOIIndex = reinterpret_cast<const CLblPOIIndexItem *> (m_pStart + _sm.cLblOffset + pLblHeader->dwPOIIndexOffset);

		for (size_t c = 0; c < cPOIIndexSize; ++ c) {
			const CLblPOIIndexItem & ii = pPOIIndex [c];

			assert (ii.btPOIIndex != 0);
			const size_t cPOIIndex = ii.btPOIIndex - 1;

			assert (ii.wGroupIndex != 0);
			const size_t cGroup = ii.wGroupIndex - 1;
			assert (cGroup < cGroups);
			const CGroup & group = groups [cGroup];

//			assert (ii.btPOISubType <= 0x10);

			const char * strLabel = NULL;
			USHORT ushType = 0;
			bool bFound = false;

			// Find N-th POI in the group.
			size_t cPOI = 0;
			assert (group.cPOIsLength != -1);
			for (i = group.cPOIsOffset; i < group.cPOIsOffset + group.cPOIsLength; i += sizeof (CRgnPOIData)) {
				const CRgnPOIData * const pPOIData = reinterpret_cast<const CRgnPOIData *> (pRgnData + i);
				if (pPOIData->bHasSubType == 0)
					-- i;

				if (cPOI < cPOIIndex) {
					++ cPOI;
					continue;
				}

				// Label.
				if (pPOIData->bExtra && pPOIExtraData) {
					const CLblPOIExtra * const pPOIExtra = reinterpret_cast<const CLblPOIExtra *> (pPOIExtraData + (pPOIData->b3LabelOffset << pLblHeader->btPOIExtraOffsetMultiplierPowerOf2));
					if (reinterpret_cast<const BYTE *> (pPOIExtra) + sizeof (CLblPOIExtra) >= m_pEnd)
						break;
					strLabel = pPOIExtra->b3LabelOffset ? m_labels [pPOIExtra->b3LabelOffset << m_cLblOffsetMultiplierPowerOf2].c_str () : NULL;
				} else
					strLabel = pPOIData->b3LabelOffset  ? m_labels [pPOIData->b3LabelOffset  << m_cLblOffsetMultiplierPowerOf2].c_str () : NULL;

	            // Type.
		        ushType = (WORD (pPOIData->btType) << 8) + (pPOIData->bHasSubType ? pPOIData->btSubType : 0);

				bFound = true;
				break;
			}

			assert (true);
		}
	}

	if (pLblHeader->dwPOITypeIndexLength != 0) {
		assert (pLblHeader->wPOITypeIndexRecSize == sizeof (CLblPOITypeIndexItem));
		const size_t cPOITypeIndexSize = pLblHeader->dwPOITypeIndexLength/sizeof (CLblPOITypeIndexItem);
		const CLblPOITypeIndexItem * const pPOITypeIndexItem = reinterpret_cast<const CLblPOITypeIndexItem *> (m_pStart + _sm.cLblOffset + pLblHeader->dwPOITypeIndexOffset);

		for (size_t c7 = 0; c7 < cPOITypeIndexSize; ++ c7) {
			const DWORD l = ConvertB3ToDWORD (pPOITypeIndexItem [c7].b3StartIndex);
		}
	}
# endif // _DEBUG

	return lsOK;
}

DWORD CGarminImgLoader::_GetNodeGroupOffset (DWORD _dwNodeOffset) const {
	const CNodNode * const pNodNode = reinterpret_cast<const CNodNode *> (m_pNodNodes + _dwNodeOffset);
	return (((_dwNodeOffset >> m_nNodesGroupMultiplierPowerOf2) + 1 + pNodNode->btGroupOffset) << m_nNodesGroupMultiplierPowerOf2);
}

bool CGarminImgLoader::_GetNodePosition (DWORD _dwNodeOffset, point_t & _pt) const {
	assert (_dwNodeOffset < m_dwNodNodesLength);
	if (_dwNodeOffset >= m_dwNodNodesLength)
		return false;

	const BYTE * pn = m_pNodNodes + _dwNodeOffset;
	const CNodNode * const pNodNode = reinterpret_cast<const CNodNode *> (pn);
	pn += sizeof (CNodNode);

	// Get common header of the nodes group.
	const DWORD dwGroupOffset = _GetNodeGroupOffset (_dwNodeOffset);
	assert (dwGroupOffset < m_dwNodNodesLength);
	if (dwGroupOffset >= m_dwNodNodesLength)
		return false;
	const CNodGroup * const pNodGroup = reinterpret_cast<const CNodGroup *> (m_pNodNodes + dwGroupOffset);
	const DWORD dwX0 = ConvertB3ToDWORD (pNodGroup->b3Lon0);
	const DWORD dwY0 = ConvertB3ToDWORD (pNodGroup->b3Lat0);

	// Parse position.
	short dx;
	short dy;
	if (pNodNode->bLargePosOffset) {
		dx = * reinterpret_cast<const SHORT *> (pn);
		pn += sizeof (SHORT);

		dy = * reinterpret_cast<const SHORT *> (pn);
		pn += sizeof (SHORT);
	} else {
		const DWORD dwPosOffset = ConvertB3ToDWORD (pn);
		pn += 3;

		const WORD wx = dwPosOffset & 0xfff;
		const WORD wy = ((dwPosOffset & 0xfff000) >> 12);

		dx = (wx & 0x800) != 0 ? - (0x800 - (wx & 0x7ff)) : wx;
		dy = (wy & 0x800) != 0 ? - (0x800 - (wy & 0x7ff)) : wy;
	}
	_pt.x = ConvertLongToDeg (dwX0 + dx);
	_pt.y = ConvertLongToDeg (dwY0 + dy);
	if (_pt.y > 90)
		_pt.y -= 180;
	return true;
}

bool CGarminImgLoader::_IsValidNode (DWORD _dwNodeOffset) const {
	if (_dwNodeOffset >= m_dwNodNodesLength)
		return false;

	const DWORD dwGroupOffset = _GetNodeGroupOffset (_dwNodeOffset);
	if (dwGroupOffset >= m_dwNodNodesLength)
		return false;

	point_t pt;
	if (! _GetNodePosition (_dwNodeOffset, pt))
		return false;
	if (! m_rectMapBoundE.Contains (pt))
		return false;

	return true;
}

void CGarminImgLoader::_AssociateNodeAndPosition (DWORD _dwNodeOffset, DWORD _dwRoadID, const point_t & _ptOriginal, const point_t & _ptFixed) {
	node_info_t info;
	info.dwNodeID   = _dwNodeOffset;
	info.ptOriginal = _ptOriginal;
	info.dwRoadID   = _dwRoadID;
	info.pNode      = NULL;
	const std::pair<pt2node_t::iterator, bool> pr = m_pt2node.insert (pt2node_t::value_type (_ptFixed, info));
}

bool CGarminImgLoader::_InsertNode (DWORD _dwNodeOffset, setIDs_t & _setNodesVisited, size_t _cDepth) {
	assert (_IsValidNode (_dwNodeOffset));
	const std::pair<setIDs_t::iterator, bool> pr = _setNodesVisited.insert (_dwNodeOffset);
	if (pr.second == true) {
		_FindNeighbourNodes (_dwNodeOffset, _setNodesVisited, _cDepth + 1);
		if (_cDepth > 30)
			return true;
	}
	return false;
}

void CGarminImgLoader::_FindNeighbourNodes (DWORD _dwNodeOffset, setIDs_t & _setNodesVisited, size_t _cDepth) {
	assert (_dwNodeOffset < m_dwNodNodesLength);
	if (_dwNodeOffset >= m_dwNodNodesLength)
		return;

	// Node header.
	const BYTE * pn = m_pNodNodes + _dwNodeOffset;
	const CNodNode * const pNodNode = reinterpret_cast<const CNodNode *> (pn);
	pn += sizeof (CNodNode);

	// Position.
	if (pNodNode->bLargePosOffset)
		pn += 4;
	else
		pn += 3;

	// Get common header of the nodes group.
	const DWORD dwGroupOffset = _GetNodeGroupOffset (_dwNodeOffset);
	assert (dwGroupOffset < m_dwNodNodesLength);
	if (dwGroupOffset >= m_dwNodNodesLength)
		return;
	const BYTE * pg = m_pNodNodes + dwGroupOffset;
	const CNodGroup * const pNodGroup = reinterpret_cast<const CNodGroup *> (pg);
	pg += sizeof (CNodGroup);

	// Verify group.
	const float x = ConvertB3ToDeg (pNodGroup->b3Lon0);
	      float y = ConvertB3ToDeg (pNodGroup->b3Lat0);
	if (y > 90)
		y -= 180;
//@	assert (m_rectMapBound.Contains (x, y));
	if (! m_rectMapBound.Contains (x, y))
		return;

	// Links table.
	const BYTE * const pLinksTable = pg;
	pg += pNodGroup->btLinks*m_cLinkRefSize;
	// Nodes table.
	const BYTE * const pNodesTable = pg;

	if (pNodGroup->btLinks == 0) {
		assert (pNodNode->bHasLinks == 0);
		return;
	}

	// Links to neighbour nodes.
	if (pNodNode->bHasLinks == 0)
		return;
	size_t cLink = 0;
	while (pn < m_pNodNodes + m_dwNodNodesLength) {
		const CNodLinkHdr * const pLinkHdr = reinterpret_cast<const CNodLinkHdr *> (pn);
		pn += sizeof (CNodLinkHdr);

		// Neighbour node.
		DWORD dwNeighbourNodeOffset = -1;
		if (pLinkHdr->bFarNode == 1) {
			BYTE btFarNodeIdx = pLinkHdr->fNode;
			if (pLinkHdr->fNode == 0x3F) {
				btFarNodeIdx = * pn;
				++ pn;
			}

			assert (btFarNodeIdx < pNodGroup->btNodes);
			if (btFarNodeIdx < pNodGroup->btNodes) {
				dwNeighbourNodeOffset = ConvertB3ToDWORD (pNodesTable + 3*btFarNodeIdx) << m_nNodesMultiplierPowerOf2;
				assert (dwNeighbourNodeOffset < m_dwNodNodesLength);
			} else
				return;
		} else {
			const BYTE bt = * pn;
			++ pn;

			const unsigned short w = WORD (bt) + 256*WORD (pLinkHdr->fNode);
			const int nNextNodeOffset = (w & 0x2000) != 0 ? -(0x4000 - w) : w;
//@			assert (LONG (_dwNodeOffset + nNextNodeOffset) >= 0);
//@			assert (_dwNodeOffset + nNextNodeOffset < dwGroupOffset);
			if (
				LONG (_dwNodeOffset + nNextNodeOffset) >= 0 &&
				_dwNodeOffset + nNextNodeOffset < dwGroupOffset
			) {
				dwNeighbourNodeOffset = _dwNodeOffset + nNextNodeOffset;
				assert (dwNeighbourNodeOffset < m_dwNodNodesLength);
			} else
				return;
		}
		if (dwNeighbourNodeOffset != -1 && dwNeighbourNodeOffset < m_dwNodNodesLength) {
			const bool bIsValidNode = _IsValidNode (dwNeighbourNodeOffset);
//@			assert (bIsValidNode);
			if (! bIsValidNode)
				return;
			if (_InsertNode (dwNeighbourNodeOffset, _setNodesVisited, _cDepth))
				return;
		}

		const bool bDirectLink = pLinkHdr->bNewHeading || cLink == 0;
		if (bDirectLink) {
			// Link.
			const BYTE btLinkIdx = * pn;
			++ pn;
//@			assert (btLinkIdx < pNodGroup->btLinks);
			if (btLinkIdx < pNodGroup->btLinks) {
				const CNodLinkRef & link = * reinterpret_cast <const CNodLinkRef *> (pLinksTable + btLinkIdx*m_cLinkRefSize);
			} else
				return;
		}

		// Curve length.
		if (pLinkHdr->fCurveLen)
//@
			break;
		if (pLinkHdr->fCurveLen == 0x7) {
			unsigned int length = 0;
			for (int i = 0; i < 3; ++ i) {
				const BYTE b = * pn;
				++ pn;

				if ((b & 0x80) == 0) {
					length <<= 8;
					length += b;
					break;
				}
				length <<= 6;
				length += (b & 0x3f);
			}
			const float fLengthInM = length*16/c_feet_per_m;
		}

		if (bDirectLink) {
			// Initial heading.
			const BYTE btInitialHeading = * pn;
			++ pn;
		}

//		if (pLinkHdr->bLast)
			break;
		++ cLink;
//@
		if (cLink == 2)
			break;
	}
}

/* Samples of time restrictions
40 8D  41 07    13
40 0D  3E 89 0F 0C.
40 8D  3E 8D 2D 0E
40 8D  3E 07    88 1E
40 0D  3E 07    87 2D.
40 8D  3E 06    09.
40 0D  3E 07    08.
40 8D  3E 07    09
40 8D  3E 0F    13
40 8D  3E 0C    0E
40 8D  3E 0C    0E
40 8D  3E 10    12
40 89  04 15
40 89  04 15
00 0C  07 09.
00 8C  10 12
00 8C  10 12
00 0C  07 09.
00 8C  10 12
00 04  14.
00 88  06
*/

/*
*2C        group
40        flags: has links
DF 8E CA  dx dy

80        A: new heading
3F F1     node offset=3FF1
4A        link idx
26        heading

18        A: curvelen=3
40        far node idx=0
00 27 38  curvelen?

80        A: new heading
5F        far node idx=1F
5D        link idx
13        heading

D4        A: curvelen=2, new heading
40        far node idx=0
BF 01 4B  curvelen?

*2C 40  E2 8E D3

80        A: new heading
00 0F     node offset=000F
4A        link idx
10        heading

08        A: curvelen=1
40        node idx=0
80        link idx
33 08     curvelen

*2C 40  E4 8E D1

80        A: new heading
3F D9     node offset=3FD9
4A        link idx
38        heading

08        A: curvelen=1
40        node idx=0
BF        link idx
F1 10     curvelen

*2B 40  E4 BE E3

C0
3F 9C
2D 14 AF

80        A: new heading
3F 47     node offset=3F47
36        link idx
24        heading

40        A: 
80
5E 20 03

*2B 40  E5 8E D4

80        A: new heading
00 C1 47 32 D4
40 3F 47 19
42 3E DB 37

80
BF CD 4A 08 08

*2B 40  3A 5F D0

C0 00 6B A1 1B 84
A0 E1 A3 21 A9

*2A 40  3A AF D1

80 00 84 9E 1B 74
A0 80 6F A0 20 A9

*2A 40  3C FF D2

C0 63 99 1B 84
A0 E2 9A 20 A9

*2A 40  3D FF E5

A0 3F A2 36
20 E8 D5
C0 BF 3E 89 22

*2A 40  40 5F 3C

80 3E AC 08 7E 3A
40 80 85 21 
*/