//
// GPSMapEdit
// (c) Konstantin Galichsky (kg@geopainting.com), 2008-2009
//
// Garmin skin parser.
//

# include "StdAfx.h"
# include "MapSkins.h"
# include "Log.h"
# include "GarminTypes.h"
# include "ParseHelpers.h"
# include "Globals.h"

class CGarminSkinParser {
  public:
	CGarminSkinParser (const char * _pStart, const char * _pEnd, bool _bDay) :
		m_pStart (_pStart),
		m_pEnd (_pEnd),
		m_bDay (_bDay),
		m_pMapSkin (NULL),
		m_GroupType (ST_NONE),
		m_dwType (0),
		m_ushSubType (-1),
		m_strTypeNameEnglish (NULL),
		m_strTypeNameLocalized (NULL),
		m_bAppendXpm (false),
		m_hBmp (NULL),
		m_hBmpMask (NULL),
		m_pBitmap (NULL),
		m_pMask (NULL),
		m_ulW_prev (0),
		m_ulH_prev (0),
		m_langidUser (PRIMARYLANGID (::GetUserDefaultLangID ())),
		m_hdcScreen (NULL),
		m_cIconW (::GetSystemMetrics (SM_CXICON)),
		m_cIconH (::GetSystemMetrics (SM_CYICON)),
		m_hdcIcon (CClientDC (NULL), ::GetSystemMetrics (SM_CXICON), ::GetSystemMetrics (SM_CYICON)),
		m_pTD_Unknown (g_tsGarmin.GetUnknownTD ()),
		m_cPoints (0),
		m_cPolylines (0),
		m_cPolygons (0)
	{}
	~CGarminSkinParser () {
		if (m_hBmp != NULL)
			::DeleteObject (m_hBmp);
		if (m_hBmpMask != NULL)
			::DeleteObject (m_hBmpMask);
	}

	bool Parse (CMapSkin * _pMapSkin);
	bool ParseLine (const char * _pLine, const char * _pNextLine);
	bool ParseType (const char * _pValue, const char * _pEnd, const char ** _ppEnd, unsigned long & _ulType);
	void ParseString (const char * _p, const char * _pNextLine, string_t & _str);

  private:
	const char * const m_pStart;
	const char * const m_pEnd;
	const bool m_bDay;

	size_t m_cLine;

	enum SectionType {
		ST_NONE, ST_ID, ST_DRAWORDER, ST_POINT, ST_LINE, ST_POLYGON
	} m_GroupType;

	DWORD  m_dwType;
	USHORT m_ushSubType;
	size_t m_cLineWidth;
	size_t m_cBorderWidth;
	const char * m_strTypeNameEnglish;
	const char * m_strTypeNameLocalized;
	bool m_bAppendXpm;
	string_t m_strXpm;
	size_t m_cXpmStart;

	CMapSkin * m_pMapSkin;

	COLORREF m_colors [4];
	DWORD m_dwColors [4];
	bool m_bColorDefined [4];
	BITMAPINFOHEADER m_bmi;
	DWORD * m_pBitmap;
	BYTE  * m_pMask;
	HBITMAP m_hBmp;
	HBITMAP m_hBmpMask;
	unsigned long m_ulW_prev, m_ulH_prev;
	bool m_bBmpDefined;
	bool ProcessXpm (bool _bMakeMask);

	bool ProcessPoint ();
	void ProcessPoint (type_desc_t * _pTD);
	bool ProcessPolyLine ();
	bool ProcessPolygon ();
	size_t m_cPoints;
	size_t m_cPolylines;
	size_t m_cPolygons;

	const LANGID m_langidUser;
	void SetTypeName (type_desc_t * _pTD);

	const type_desc_t * const m_pTD_Unknown;

	CClientDC m_hdcScreen;
	const size_t m_cIconW;
	const size_t m_cIconH;
	CMemDC m_hdcIcon;
};

static
BYTE _Parse2Hex (const char * _p) {
	const BYTE bt0 = _p [0] > '9' ? ::toupper (_p [0]) - 'A' + 10 : _p [0] - '0';
	const BYTE bt1 = _p [1] > '9' ? ::toupper (_p [1]) - 'A' + 10 : _p [1] - '0';
	return (bt0 << 4) + bt1;
}

struct xpm_node_t {
	size_t cColorIdx;
	bool bTransparent;
	DWORD  dwColor;
	xpm_node_t * pChildren;

	xpm_node_t () : cColorIdx (-1), bTransparent (false), dwColor (0), pChildren (NULL) {}
	~xpm_node_t () {delete [] pChildren;}
};

bool CGarminSkinParser::ProcessXpm (bool _bMakeMask) {
	m_bBmpDefined = false;
	::memset (m_colors,        0, sizeof (m_colors));
	::memset (m_dwColors,      0, sizeof (m_dwColors));
	::memset (m_bColorDefined, 0, sizeof (m_bColorDefined));

	assert (! m_strXpm.empty ());
	const char * p = m_strXpm.c_str ();
	const char * const pEnd = p + m_strXpm.size ();

	// Bitmap width.
	OmitSpaces (& p, pEnd);
	unsigned long ulW = 0;
	if (! ParseDecU (p, pEnd, & p, ulW)) {
		ReportWarning ("Decimal expected for width at line %d.", m_cXpmStart);
		return false;
	}

	// Bitmap height.
	OmitSpaces (& p, pEnd);
	unsigned long ulH = 0;
	if (! ParseDecU (p, pEnd, & p, ulH)) {
		ReportWarning ("Decimal expected for height at line %d.", m_cXpmStart);
		return false;
	}

	// Colors number.
	OmitSpaces (& p, pEnd);
	unsigned long ulColors = 0;
	if (! ParseDecU (p, pEnd, & p, ulColors)) {
		ReportWarning ("Decimal expected for colors number at line %d.", m_cXpmStart);
		return false;
	}

	// Characters per pixel.
	OmitSpaces (& p, pEnd);
	unsigned long ulPixelLen = 0;
	if (! ParseDecU (p, pEnd, & p, ulPixelLen)) {
		ReportWarning ("Decimal expected for characters per pixel at line %d.", m_cXpmStart);
		return false;
	}
	if (ulPixelLen == 0)
		ulPixelLen = 1;

	// Parse palette.
	xpm_node_t nodes [256];
	for (size_t cColorIdx = 0; cColorIdx < ulColors; ++ cColorIdx) {
		if (* p == '\n')
			++ p;

		const char * const pSymbol = p;
		const BYTE btChar = * p;
		p += ulPixelLen;

		xpm_node_t * pn = & nodes [btChar];
		for (const char * p2 = pSymbol + 1; p2 < p; ++ p2) {
			if (pn->pChildren == NULL)
				pn->pChildren = new xpm_node_t [256];
			pn = & (pn->pChildren [* p2]);
		}
		pn->cColorIdx = cColorIdx;

		OmitSpaces (& p, pEnd);
		if (* p != 'c')
			ReportWarning ("Delimiter 'c' expected for color at line %d.", m_cXpmStart);
		++ p;

		OmitSpaces (& p, pEnd);
		if (HasPrefix (p, pEnd, "None", 4)) {
			// Transparent.
			p += 4;

			pn->bTransparent = true;
		} else {
			// #RGB
			if (* p != '#') {
				ReportWarning ("Prefix '#' expected for color value at line %d.", m_cXpmStart);
				return false;
			}
			++ p;

			if (p + 5 >= pEnd) {
				ReportWarning ("Incorrect color value at line %d.", m_cXpmStart);
				return false;
			}
			for (size_t c = 0; c < 6; ++ c) {
				if (! ::isxdigit (p [c])) {
					ReportWarning ("Incorrect color code at line %d.", m_cXpmStart);
					return false;
				}
			}
			const BYTE r = _Parse2Hex (p);
			const BYTE g = _Parse2Hex (p + 2);
			const BYTE b = _Parse2Hex (p + 4);

			pn->dwColor = RGB (b, g, r);

			if (cColorIdx < ARRAY_SIZE (m_colors)) {
				m_colors   [cColorIdx] = RGB (r, g, b);
				m_dwColors [cColorIdx] = RGB (b, g, r);
				m_bColorDefined [cColorIdx] = true;
			}

			p += 6;
			OmitNonSpaces (& p, pEnd);
		}
	}

	if (ulW == 0 || ulH == 0)
		return true;

	if (m_hBmp == NULL || ulW != m_ulW_prev || ulH != m_ulH_prev) {
		m_ulW_prev = ulW;
		m_ulH_prev = ulH;
			
		if (m_hBmp != NULL) {
			::DeleteObject (m_hBmp);
			m_hBmp = NULL;
		}

		// Prepare buffer for bitmap.
		::memset (& m_bmi, 0, sizeof (m_bmi));
		m_bmi.biSize        = sizeof (m_bmi);
		m_bmi.biWidth       = ulW;
		m_bmi.biHeight      = - (int) ulH;
		m_bmi.biPlanes      = 1;
		m_bmi.biBitCount    = 32;
		m_bmi.biCompression = BI_RGB;
		m_bmi.biSizeImage   = 4*ulW*ulH;

		m_hBmp = ::CreateDIBSection (m_hdcScreen, (BITMAPINFO *) & m_bmi, DIB_RGB_COLORS, (void **) & m_pBitmap, NULL, 0);
	}
	::memset (m_pBitmap, 0, m_bmi.biSizeImage);
	m_bBmpDefined = true;

	if (_bMakeMask) {
		if (m_hBmpMask == NULL) {
			BITMAPINFO & bmiMask = * (BITMAPINFO *) alloca (sizeof (BITMAPINFOHEADER) + 2*sizeof (RGBQUAD));
			::memset (& bmiMask, 0, sizeof (bmiMask));
			bmiMask.bmiHeader.biSize         = sizeof (bmiMask);
			bmiMask.bmiHeader.biWidth        = m_cIconW;
			bmiMask.bmiHeader.biHeight       = - (int) m_cIconH;
			bmiMask.bmiHeader.biPlanes       = 1;
			bmiMask.bmiHeader.biBitCount     = 1;
			bmiMask.bmiHeader.biCompression  = BI_RGB;
			bmiMask.bmiHeader.biClrImportant = 2;
			bmiMask.bmiHeader.biClrUsed      = 2;
			::memset (& bmiMask.bmiColors [0], 0,    sizeof (bmiMask.bmiColors [0]));
			::memset (& bmiMask.bmiColors [1], 0xFF, sizeof (bmiMask.bmiColors [1]));
			bmiMask.bmiColors [1].rgbReserved = 0;

			m_hBmpMask = ::CreateDIBSection (m_hdcScreen, & bmiMask, DIB_PAL_COLORS, (void **) & m_pMask, NULL, 0);
		}
		::memset (m_pMask, 0x0, sizeof (BYTE)*m_cIconW*m_cIconH);
	}

	const COLORREF colBG = MapSkin_GetColor (spcBgColor, m_pMapSkin);
	const DWORD dwBG = RGB (GetBValue (colBG), GetGValue (colBG), GetRValue (colBG));

	// Parse bitmap.
	size_t cPixel = 0;
	while (p < pEnd && cPixel < ulW*ulH) {
		const BYTE btChar = * p;
		if (btChar == '\n') {
			++ p;
		} else {
			xpm_node_t * pn = & nodes [btChar];
			++ p;
			while (pn->pChildren) {
				pn = & (pn->pChildren [* p]);
				++ p;
			}

			if (! pn->bTransparent) {
				if (! m_bDay && ! _bMakeMask) {
					// Night colors for polygons and polylines.
					size_t cColorIdx = pn->cColorIdx;
					if (cColorIdx < 2)
						cColorIdx += 2;

					if (cColorIdx < ARRAY_SIZE (m_dwColors) && m_bColorDefined [cColorIdx])
						m_pBitmap [cPixel] = m_dwColors [cColorIdx];
					else
						m_pBitmap [cPixel] = pn->dwColor;
				} else
					m_pBitmap [cPixel] = pn->dwColor;

				if (_bMakeMask && m_pMask != NULL) {
					const int y = cPixel/ulW;
					const int x = cPixel - y*ulW;
					m_pMask [(m_cIconW*y + x)/8] |= (1 << (7 - (x & 0x7)));
				}
			}
			if (pn->bTransparent && ! _bMakeMask)
				m_pBitmap [cPixel] = dwBG;

			++ cPixel;
		}
	}
	if (p < pEnd)
		ReportWarning ("Unexpected bitmap size at line %d.", m_cXpmStart);

	return true;
}

void CGarminSkinParser::SetTypeName (type_desc_t * _pTD) {
	if (m_strTypeNameEnglish) {
		_pTD->strTypeName = m_strTypeNameEnglish;
		_pTD->SetTypeNameLocalized (NULL);
	}
	if (m_strTypeNameLocalized) {
		_pTD->SetTypeNameLocalized (m_strTypeNameLocalized);
		g_tsGarmin.bHasLocalizedTypes = true;
	}
}

bool CGarminSkinParser::ProcessPoint () {
	if (m_strXpm.empty ())
		// NOTE: this may take place for night mode if "DayXPM=" is defined while "NightXPM=" is not.
		return true;

	if (! ProcessXpm (true))
		return true;

	bool bApplyToSubTypes = false;
	DWORD dwType = 0;
	if (m_dwType > 0x1FF)
		dwType = m_dwType;
	else {
		if (m_ushSubType == (USHORT) -1) {
			dwType = (m_dwType << 8);
			bApplyToSubTypes = true;
		} else {
			dwType = (m_dwType << 8) + m_ushSubType;
		}
	}

	if (bApplyToSubTypes) {
		bool bApplied = false;

		for (USHORT ushSubType = 0; ushSubType <= 0x3F; ++ ushSubType) {
			dwType = (m_dwType << 8) + ushSubType;

			type_desc_t * const pTD = const_cast<type_desc_t *> (g_tsGarmin.GetTypeDesc (dwType, rkPoint));
			if (pTD == NULL || pTD == m_pTD_Unknown)
				continue;
				
			bApplied = true;
			ProcessPoint (pTD);
		}
		if (! bApplied) {
			dwType = (m_dwType << 8);
			type_desc_t & td = g_tsGarmin.AddCustomTypeDesc (dwType, rkPoint);
			ProcessPoint (& td);
		}
	} else {
		type_desc_t * const pTD = const_cast<type_desc_t *> (g_tsGarmin.GetTypeDesc (dwType, rkPoint));
		if (pTD != NULL && pTD != m_pTD_Unknown && pTD->Type == dwType) {
			ProcessPoint (pTD);
		} else {
			type_desc_t & td = g_tsGarmin.AddCustomTypeDesc (dwType, rkPoint);
			ProcessPoint (& td);
		}
	}

	return true;
}

void CGarminSkinParser::ProcessPoint (type_desc_t * _pTD) {
	assert (_pTD != NULL);

	RECT rect = {0, 0, m_cIconW, m_cIconH};
	::FillRect  (m_hdcIcon, & rect, g_hbrBlack);
	::SetDIBits (m_hdcIcon, m_hdcIcon.GetBitmap (), 0, ::labs (m_bmi.biHeight), m_pBitmap, (BITMAPINFO *) & m_bmi, DIB_RGB_COLORS);

	ICONINFO ii = {FALSE, m_bmi.biWidth/2, ::labs (m_bmi.biHeight)/2, m_hBmpMask, m_hdcIcon.GetBitmap ()};
	const HCURSOR hcur = ::CreateIconIndirect (& ii);
	_pTD->phCursor = & m_pMapSkin->AddIcon (hcur);
	_pTD->ResetCursorShift ();

	SetTypeName (_pTD);

	_pTD->DestroySecondaryAttributes ();

	++ m_cPoints;
}

bool CGarminSkinParser::ProcessPolyLine () {
	if (! ProcessXpm (false))
		return true;

	if (m_cBorderWidth == 0 && m_cLineWidth == 0 && ! m_bBmpDefined) {
		ReportWarning ("Unexpected dummy data for polyline type 0x%02x at line %d.", m_dwType, m_cLine);
		m_cLineWidth = 1;
	}

	const DWORD dwType = m_ushSubType == (USHORT) -1 ? m_dwType : (m_dwType << 8) + m_ushSubType;

	type_desc_t * pTD = const_cast<type_desc_t *> (g_tsGarmin.GetTypeDesc (dwType, rkPolyline));
	if (pTD == NULL || pTD == m_pTD_Unknown) {
		pTD = & g_tsGarmin.AddCustomTypeDesc (dwType, rkPolyline);
	} else
		pTD->DestroySecondaryAttributes ();

	if (m_cBorderWidth != 0 || m_cLineWidth != 0) {
		const bool bColLineDefined   = m_bColorDefined [m_bDay ? 0 : 2];
		const bool bColBorderDefined = m_bColorDefined [m_bDay ? 1 : 3];
		const COLORREF colLine   = m_colors [m_bDay ? 0 : 2];
		const COLORREF colBorder = m_colors [m_bDay ? 1 : 3];

		if (m_cBorderWidth == 0) {
			pTD->phPen  = bColLineDefined ? & m_pMapSkin->AddPen (::CreatePen (PS_SOLID, m_cLineWidth, colLine)) : & g_hpenNull;
			pTD->phPen2 = NULL;
		} else {
			pTD->phPen  = & m_pMapSkin->AddPen (::CreatePen (PS_SOLID, m_cLineWidth + m_cBorderWidth*2, colBorder));
			pTD->phPen2 = & m_pMapSkin->AddPen (::CreatePen (PS_SOLID, m_cLineWidth,                    colLine));
		}
	}

	if (m_bBmpDefined) {
		BYTE * const pPackedDIB = (BYTE *) ::alloca (sizeof (m_bmi) + m_bmi.biSizeImage);
		::memcpy (pPackedDIB, & m_bmi, sizeof (m_bmi));
		::memcpy (pPackedDIB + sizeof (m_bmi), m_pBitmap, m_bmi.biSizeImage);
		const LOGBRUSH lb = {BS_DIBPATTERNPT, DIB_RGB_COLORS, (DWORD) pPackedDIB};

		const HPEN hpen = ::ExtCreatePen (PS_GEOMETRIC | PS_ENDCAP_FLAT | PS_JOIN_ROUND, ::labs (m_bmi.biHeight), & lb, 0, NULL);
		pTD->phPen  = & m_pMapSkin->AddPen (hpen);
		pTD->phPen2 = NULL;
	}

	assert (pTD->phPen);

	SetTypeName (pTD);

	++ m_cPolylines;

	return true;
}

bool CGarminSkinParser::ProcessPolygon () {
	if (! ProcessXpm (false))
		return true;

	const DWORD dwType = m_ushSubType == (USHORT) -1 ? m_dwType : (m_dwType << 8) + m_ushSubType;

	type_desc_t * pTD = const_cast<type_desc_t *> (g_tsGarmin.GetTypeDesc (dwType, rkPolygon));
	if (pTD == NULL || pTD == m_pTD_Unknown) {
		pTD = & g_tsGarmin.AddCustomTypeDesc (dwType, rkPolygon);
	} else
		pTD->DestroySecondaryAttributes ();

	if (m_bBmpDefined) {
		const HBRUSH hbr = ::CreatePatternBrush (m_hBmp);
		pTD->phBrush = & m_pMapSkin->AddBrush (hbr);
	} else {
		const COLORREF colFill   = m_colors [m_bDay ? 0 : (m_bColorDefined [2] ? 2 : 0)];
		const COLORREF colBorder = m_colors [m_bDay ? 1 : (m_bColorDefined [3] ? 3 : 1)];
		const bool bBorderDefined = m_bColorDefined [m_bDay ? 1 : 3];

		const HBRUSH hbr = ::CreateSolidBrush (colFill);
		pTD->phBrush = & m_pMapSkin->AddBrush (hbr);

		if (bBorderDefined)
			pTD->phPen = & m_pMapSkin->AddPen (::CreatePen (PS_SOLID, 1, colBorder));

		if (dwType == 0x4b)
			m_pMapSkin->SetColor (spcBgColor, colFill);
	}

	SetTypeName (pTD);

	++ m_cPolygons;

	return true;
}

bool CGarminSkinParser::ParseType (const char * _pValue, const char * _pEnd, const char ** _ppEnd, unsigned long & _ulType) {
	const char * p = _pValue;

	// Omit spaces.
	while (p < _pEnd && * p == ' ')
		++ p;

	if (p + 1 < _pEnd && p [0] == '0' && p [1] == 'x')
		return ParseHex (p + 2, _pEnd, _ppEnd, _ulType);
	else
		return ParseDecU (p, _pEnd, _ppEnd, _ulType);
}

void CGarminSkinParser::ParseString (const char * _p, const char * _pNextLine, string_t & _str) {
	const char * pStringEnd = NULL;
	const char * p;
	for (p = _pNextLine - 1; p > _p + 1; -- p)
		if (* p == '\"') {
			pStringEnd = p;
			break;
		}
	if (! pStringEnd)
		pStringEnd = p + GetLineLen (_p, _pNextLine) - 1;

	if (! _str.empty ())
		_str.append ("\n");
	_str.append (_p + 1, pStringEnd);
}

bool CGarminSkinParser::Parse (CMapSkin * _pMapSkin) {
	m_pMapSkin = _pMapSkin;

	m_cLine = 1;

	const char * pLine = m_pStart;
	do {
		const char * pNextLine = FindNextLine (pLine, m_pEnd);

		if (! ParseLine (pLine, pNextLine))
			return false;

		pLine = pNextLine;

		// Count lines.
		++ m_cLine;
	} while (pLine < m_pEnd);

	ReportText ("Redefined: %d point(s), %d polyline(s), %d polygon(s)", m_cPoints, m_cPolylines, m_cPolygons);

	return true;
}

static
const char * _GetValue (const char * _pKey, const char * _pEnd) {
	if (_pKey >= _pEnd)
		return _pEnd;
	const char * const p = reinterpret_cast<const char *> (::memchr (_pKey, '=', _pEnd - _pKey));
	return p ? p + 1 : _pEnd;
}

bool CGarminSkinParser::ParseLine (const char * _p, const char * _pNextLine) {
	switch (_p [0]) {
		case ';':
		case '\n':
		case '\r':
			break;

		case '[': {
			if (HasPrefix (_p + 1, _pNextLine, "end]", 4)) {
				switch (m_GroupType) {
					case ST_POINT:   return ProcessPoint    ();
					case ST_LINE:    return ProcessPolyLine ();
					case ST_POLYGON: return ProcessPolygon  ();
					default: break;
				}
			} else if (HasPrefix (_p + 1, _pNextLine, "_point]", 7)) {
				m_GroupType = ST_POINT;
				m_dwType = -1;
				m_ushSubType = -1;
				m_strTypeNameEnglish = NULL;
				m_strTypeNameLocalized = NULL;
				m_strXpm.clear ();
			} else if (HasPrefix (_p + 1, _pNextLine, "_line]", 6)) {
				m_GroupType = ST_LINE;
				m_dwType = -1;
				m_ushSubType = -1;
				m_cLineWidth = 0;
				m_cBorderWidth = 0;
				m_strTypeNameEnglish = NULL;
				m_strTypeNameLocalized = NULL;
				m_strXpm.clear ();
			} else if (HasPrefix (_p + 1, _pNextLine, "_polygon]", 9)) {
				m_GroupType = ST_POLYGON;
				m_dwType = -1;
				m_ushSubType = -1;
				m_strTypeNameEnglish = NULL;
				m_strTypeNameLocalized = NULL;
				m_strXpm.clear ();
			} else if (HasPrefix (_p + 1, _pNextLine, "_draworder]", 11)) {
				m_GroupType = ST_DRAWORDER;
			} else if (HasPrefix (_p + 1, _pNextLine, "_id]", 4)) {
				m_GroupType = ST_ID;
			}
			break;
		}

		case '\"':
			if (m_bAppendXpm)
				ParseString (_p, _pNextLine, m_strXpm);
			break;

		default: {
			m_bAppendXpm = false;

			const char * const pKey = _p;
			const char * const pValue = _GetValue (_p, _pNextLine);

			if (HasPrefix (pKey, _pNextLine, "Type=", 5)) {
				unsigned long ul = 0;
				const char * pEnd = NULL;
				if (! ParseType (pValue, _pNextLine, & pEnd, ul)) {
					ReportWarning ("Can't parse type at line %d.", m_cLine);
					ul = 0;
				}
				m_dwType = ul;
			} else if (HasPrefix (pKey, _pNextLine, "SubType=", 8)) {
				unsigned long ul = 0;
				const char * pEnd = NULL;
				if (! ParseType (pValue, _pNextLine, & pEnd, ul)) {
					ReportWarning ("Can't parse subtype at line %d.", m_cLine);
					ul = 0;
				}
				m_ushSubType = ul;
			} else if (HasPrefix (pKey, _pNextLine, "String", 6)) {
				unsigned long ul = 0;
				const char * p = NULL;
				if (ParseType (pValue, _pNextLine, & p, ul) && (ul == 0x04 || ul == m_langidUser)) {
					OmitSpaces (& p, _pNextLine);
					if (* p == ',')
						++ p;
					const size_t cLen = GetLineLen (p, _pNextLine);
					const char * const strTypeName = m_pMapSkin->AddLabel (string_t (p, cLen));
					if (ul == 0x04)
						m_strTypeNameEnglish = strTypeName;
					else if (ul == m_langidUser && m_langidUser != 0x04)
						m_strTypeNameLocalized = strTypeName;
				}
			} else if (HasPrefix (pKey, _pNextLine, "Xpm=", 4)) {
				m_cXpmStart = m_cLine;
				m_strXpm.clear ();
				ParseString (pValue, _pNextLine, m_strXpm);
				m_bAppendXpm = true;
			} else if (HasPrefix (pKey, _pNextLine, "DayXpm=", 7)) {
//				if (m_bDay) {
					m_cXpmStart = m_cLine;
					m_strXpm.clear ();
					ParseString (pValue, _pNextLine, m_strXpm);
					m_bAppendXpm = true;
//				}
			} else if (HasPrefix (pKey, _pNextLine, "NightXpm=", 9)) {
				if (! m_bDay) {
					m_cXpmStart = m_cLine;
					m_strXpm.clear ();
					ParseString (pValue, _pNextLine, m_strXpm);
					m_bAppendXpm = true;
				}
			} else if (HasPrefix (pKey, _pNextLine, "LineWidth=", 10)) {
				unsigned long ul = 0;
				const char * p;
				if (! ParseDecU (pValue, _pNextLine, & p, ul))
					ReportWarning ("Decimal expected for 'LineWidth=' at line %d.", m_cLine);
				else
					m_cLineWidth = ul;
			} else if (HasPrefix (pKey, _pNextLine, "BorderWidth=", 12)) {
				unsigned long ul = 0;
				const char * p;
				if (! ParseDecU (pValue, _pNextLine, & p, ul))
					ReportWarning ("Decimal expected for 'BorderWidth=' at line %d.", m_cLine);
				else
					m_cBorderWidth = ul;
			}
			break;
		}
	}
	return true;
}

bool ParseSkin_Garmin (const char * _strFileName, const char * _strMode, CMapSkin * _pMapSkin) {
	CFileForRead hFile (_strFileName);
	if (! hFile.IsValid ())
		return false;

	const size_t cFileSize = ::GetFileSize (hFile, NULL);
	CWinHandle hMapping = ::CreateFileMapping (hFile, NULL, PAGE_READONLY, 0, 0, NULL);
	if (! hMapping.IsValid ())
		return false;
	const char * const pFileData = reinterpret_cast<const char *> (::MapViewOfFile (hMapping, FILE_MAP_READ, 0, 0, 0));

	CGarminSkinParser parser (pFileData, pFileData + cFileSize, ::stricmp (_strMode, "day") == 0);
	if (! parser.Parse (_pMapSkin)) {
		::UnmapViewOfFile (pFileData);
		return false;
	}

	::UnmapViewOfFile (pFileData);

	return true;
}

DECLARE_MAP_SKIN (ParseSkin_Garmin, "Garmin", "txt", "Garmin TYP source (*.txt)");