//
// GPSMapEdit
// (c) Konstantin Galichsky (kg@geopainting.com), 2008
//
// Helpers to connect Garmin GPSr over USB.
//

# include <StdAfx.h>
# include "GarminUSB.h"
# include "Positioning.h"
# include "Kernel.h"
# include "Wrappers.h"
# include "Globals.h"

# include <initguid.h>
DEFINE_GUID (GUID_DEVINTERFACE_GRMNUSB, 0x2c9c45c2L, 0x8e7d, 0x4c08, 0xa1, 0x2d, 0x81, 0x6b, 0xba, 0xe7, 0x22, 0xc0);

# define IOCTL_API_VERSION     CTL_CODE (FILE_DEVICE_UNKNOWN, 0x800, METHOD_BUFFERED, FILE_ANY_ACCESS)
# define IOCTL_ASYNC_IN        CTL_CODE (FILE_DEVICE_UNKNOWN, 0x850, METHOD_BUFFERED, FILE_ANY_ACCESS)
# define IOCTL_USB_PACKET_SIZE CTL_CODE (FILE_DEVICE_UNKNOWN, 0x851, METHOD_BUFFERED, FILE_ANY_ACCESS)

# define MAX_BUFFER_SIZE 4096
# define ASYNC_DATA_SIZE 64

enum {
	PT_TRANSPORT   = 0,
	PT_APPLICATION = 20,
};

enum {
	PID_DATA_AVAILABLE  = 2,
	PID_START_SESSION   = 5,
	PID_SESSION_STARTED = 6,

	PID_COMMAND_DATA    = 10,
	PID_PVT_DATA        = 51,
};

enum {
	Cmnd_Abort_Transfer = 0,  /* abort current transfer */
	Cmnd_Start_Pvt_Data = 49, /* start transmitting PVT data */
	Cmnd_Stop_Pvt_Data  = 50  /* stop transmitting PVT data */
};

# pragma pack(push, 1)
struct Packet_t {
    BYTE btPacketType;
    BYTE btReserved1;
    WORD wReserved2;
    WORD wPacketId;
    WORD wReserved3;
    DWORD dwDataSize;
    BYTE aData [1];
};
# pragma pack(pop)

class CGarminUSB {
  public:
	CGarminUSB (HANDLE _hUSB, HANDLE _hStopEvent) :
		m_hUSB (_hUSB), m_cUSBPacketSize (0),
		m_hStopEvent (_hStopEvent),
		m_bBulkyRead (false)
	{}

	HANDLE GetStopEvent () const {return m_hStopEvent;}

	void SetPacketSize ();
	DWORD GetApiVersion ();

	void SendPacket (bool _bAppLevel, WORD _wPacketId, size_t _cDataSize, const void * _pData);
	void SendPacket (const Packet_t & _Packet);
	Packet_t * GetPacket ();

  private:
	const CWinHandle m_hUSB;
	size_t m_cUSBPacketSize;
	bool   m_bBulkyRead;

	const HANDLE m_hStopEvent;
};

void CGarminUSB::SendPacket (bool _bAppLevel, WORD _wPacketId, size_t _cDataSize, const void * _pData) {
	const DWORD dwBytesToSend = sizeof (Packet_t) - sizeof (BYTE) + _cDataSize;
	BYTE * const pPacket_low = reinterpret_cast<BYTE *> (alloca (dwBytesToSend));

	Packet_t * const pPacket = reinterpret_cast<Packet_t *> (pPacket_low);
	pPacket->btPacketType = _bAppLevel ? PT_APPLICATION : PT_TRANSPORT;
	pPacket->btReserved1  = 0;
	pPacket->wReserved2   = 0;
	pPacket->wPacketId    = _wPacketId;
	pPacket->wReserved3   = 0;
	pPacket->dwDataSize   = _cDataSize;
	::memcpy (pPacket->aData, _pData, _cDataSize);

	SendPacket (* pPacket);
}

void CGarminUSB::SendPacket (const Packet_t & _Packet) {
	const DWORD dwBytesToWrite = sizeof (_Packet) - sizeof (_Packet.aData) + _Packet.dwDataSize;
	DWORD dwBytesWritten = 0;
	::WriteFile (m_hUSB, & _Packet, dwBytesToWrite, & dwBytesWritten, NULL);

	// If the packet size was an exact multiple of the USB packet
	// size, we must make a final write call with no data.
	if (dwBytesToWrite % m_cUSBPacketSize == 0)
		::WriteFile (m_hUSB, 0, 0, & dwBytesWritten, NULL);
}

// Gets a single packet.
// TODO: Since packets may come simultaneously through
// asynchrous reads and normal (ReadFile) reads, a full implementation
// may require a packet queue and multiple threads.
Packet_t * CGarminUSB::GetPacket () {
	while (true) {
		if (! m_bBulkyRead) {
			//
			// Normal mode.
			//

			Packet_t * pPacket = NULL;

			size_t cBufferSize = 0;
			BYTE * pBuffer = NULL;

			while (true) {
				// Read async data until the driver returns less than the
				// max async data size, which signifies the end of a packet.
				BYTE TempBuffer [ASYNC_DATA_SIZE];
				DWORD dwBytesReturned = 0;
				::DeviceIoControl (
					m_hUSB, IOCTL_ASYNC_IN,
					NULL, 0,
					TempBuffer, sizeof (TempBuffer), & dwBytesReturned,
					NULL
				);

				cBufferSize += ASYNC_DATA_SIZE;
				BYTE * const pNewBuffer = (BYTE *) ::malloc (cBufferSize);
				if (pBuffer)
					::memcpy (pNewBuffer, pBuffer, cBufferSize - ASYNC_DATA_SIZE);
				::memcpy (pNewBuffer + cBufferSize - ASYNC_DATA_SIZE, TempBuffer, ASYNC_DATA_SIZE);

				::free (pBuffer);

				pBuffer = pNewBuffer;

				if (dwBytesReturned != ASYNC_DATA_SIZE) {
					pPacket = (Packet_t *) pBuffer;
					break;
				}
			}

			if (pPacket->btPacketType == PT_TRANSPORT && pPacket->wPacketId == PID_DATA_AVAILABLE)
				m_bBulkyRead = true;

			return pPacket;
		} else {
			//
			// Bulky mode.
			//

			BYTE * const pNewBuffer = (BYTE *) ::malloc (MAX_BUFFER_SIZE);

			// TODO: A full implementation would keep reading (and queueing)
			// packets until the driver returns a 0 size buffer.
			DWORD dwBytesRead = 0;
			::ReadFile (m_hUSB, pNewBuffer, MAX_BUFFER_SIZE, & dwBytesRead, NULL);
    
			if (dwBytesRead == 0) {
				m_bBulkyRead = false;
				::free (pNewBuffer);
				continue;
			}

			return (Packet_t *) pNewBuffer;
		}
	}
}

void CGarminUSB::SetPacketSize () {
	// Get the USB packet size, which we need for sending packets.
	DWORD dwUSBPacketSize = 0;
	DWORD dwReturnedBytes = 0;
	::DeviceIoControl (
		m_hUSB, IOCTL_USB_PACKET_SIZE,
		NULL, 0,
		& dwUSBPacketSize, sizeof (dwUSBPacketSize), & dwReturnedBytes,
		NULL
	);

	m_cUSBPacketSize = dwUSBPacketSize;
}

DWORD CGarminUSB::GetApiVersion () {
	// Get the USB packet size, which we need for sending packets.
	DWORD dwApiVersion = 0;
	DWORD dwReturnedBytes = 0;
	::DeviceIoControl (
		m_hUSB, IOCTL_API_VERSION,
		NULL, 0,
		& dwApiVersion, sizeof (dwApiVersion), & dwReturnedBytes,
		NULL
	);

	return dwApiVersion;
}

static
DWORD WINAPI _UsbThreadProc (LPVOID _lpParameter);

HANDLE InitGarminUSB (HANDLE _hStopEvent) {
	CWinHandle hUSB = OpenDriverWithClass (GUID_DEVINTERFACE_GRMNUSB);
	if (! hUSB.IsValid ())
		return NULL;

	SetPositioningStatus ("Garmin over USB.");

	CGarminUSB * const pGU = new CGarminUSB (hUSB.Detach (), _hStopEvent);

	DWORD dwId = 0;
	return ::CreateThread (NULL, 0, _UsbThreadProc, pGU, 0, & dwId);
}
# pragma pack(push, 1)
typedef struct {
	double lat; /* latitude in radians */
	double lon; /* longitude in radians */
} Radian_Type;

struct D800_Pvt_Data_Type {
	float alt;        /* altitude above WGS 84 ellipsoid (meters) */
	float epe;        /* estimated position error, 2 sigma (meters) */
	float eph;        /* epe, but horizontal only (meters) */
	float epv;        /* epe, but vertical only (meters) */
	WORD  fix;        /* type of position fix */
	double tow;       /* time of week (seconds) */
	Radian_Type posn; /* latitude and longitude (radians) */
	float east;       /* velocity east (meters/second) */
	float north;      /* velocity north (meters/second) */
	float up;         /* velocity up (meters/second) */
	float msl_hght;   /* height of WGS 84 ellipsoid above MSL (meters) */
	WORD  leap_scnds; /* difference between GPS and UTC (seconds) */
	DWORD wn_days;    /* week number days */
};

# pragma pack(pop)

enum {
	fix_unusable = 0, /* failed integrity check */
	fix_invalid  = 1, /* invalid or unavailable */
	fix_2D       = 2, /* two dimensional */
	fix_3D       = 3, /* three dimensional */
	fix_2D_diff  = 4, /* two dimensional differential */
	fix_3D_diff  = 5  /* three dimensional differential */
};

static
void OnPVTData (const D800_Pvt_Data_Type * _pData) {
	if (_pData->fix == fix_unusable || _pData->fix == fix_invalid) {
		PushNoPositionInfo ();
		return;
	}

	PositionInfo_t pi;

	pi.SetTime (32508 /*1989 31 Dec*/ + 365 + _pData->wn_days + (_pData->tow - _pData->leap_scnds)/(24*60*60));

	pi.SetXY (_pData->posn.lon*180/c_PI, _pData->posn.lat*180/c_PI);
	pi.SetXYPrec (_pData->eph/2);

	pi.SetAltitude (_pData->alt + _pData->msl_hght);
	pi.SetAltitudePrec (_pData->epv/2);
	pi.SetGeoidHeight (- _pData->msl_hght);

	pi.SetHeading (::atan2 (_pData->east, _pData->north)*180/c_PI);
	pi.SetSpeed (::sqrt (_pData->east*_pData->east + _pData->north*_pData->north + _pData->up*_pData->up)*3.6f);
	pi.SetVertSpeed (_pData->up);

	switch (_pData->fix) {
		case fix_2D:      pi.SetFix (PositionInfo_t::fix2D);      break;
		case fix_3D:      pi.SetFix (PositionInfo_t::fix3D);      break;
		case fix_2D_diff: pi.SetFix (PositionInfo_t::fix2D_diff); break;
		case fix_3D_diff: pi.SetFix (PositionInfo_t::fix3D_diff); break;
		default:          pi.SetFix (PositionInfo_t::fixUnknown); break;
	}

	PushPositionInfo (pi);
}

static
DWORD WINAPI _UsbThreadProc (LPVOID _lpParameter) {
	CGarminUSB * const pGU = reinterpret_cast<CGarminUSB *> (_lpParameter);
	CObjectHolder<CGarminUSB> pGU_Holder = pGU;

	const HANDLE hStopEvent = pGU->GetStopEvent ();

	//
	// Start session.
	//

	pGU->SetPacketSize ();

	const DWORD dwApiVersion = pGU->GetApiVersion ();

	// Tell the device that we are starting a session.
	pGU->SendPacket (false, PID_START_SESSION, 0, NULL);

	// Wait until the device is ready to start the session.
	while (true) {
		Packet_t * const pPacket = pGU->GetPacket ();

		if (pPacket->btPacketType == PT_TRANSPORT && pPacket->wPacketId == PID_SESSION_STARTED) {
			::free (pPacket);
	        break;
		}

		::free (pPacket);

		if (::WaitForSingleObject (hStopEvent, 50) == WAIT_OBJECT_0)
			return 0;
    }

	//
	// Read PVT.
	//

	DWORD dwLastRequestTime = 0;
	while (true) {
		if (::GetTickCount () - dwLastRequestTime > 1000) {
			dwLastRequestTime = ::GetTickCount ();

			const WORD wCmdStartPvtData = Cmnd_Start_Pvt_Data;
			pGU->SendPacket (true, PID_COMMAND_DATA, 2, & wCmdStartPvtData);
		}

		Packet_t * const pPacket = pGU->GetPacket ();

		if (pPacket->btPacketType == PT_APPLICATION && pPacket->wPacketId == PID_PVT_DATA) {
			if (pPacket->dwDataSize >= sizeof (D800_Pvt_Data_Type))
				OnPVTData (reinterpret_cast<const D800_Pvt_Data_Type *> (pPacket->aData));
		}

		::free (pPacket);

		if (::WaitForSingleObject (hStopEvent, 50) == WAIT_OBJECT_0)
			break;
    }

	// Stop PVT sending.
	const WORD wCmdStopPvtData = Cmnd_Stop_Pvt_Data;
	pGU->SendPacket (true, PID_COMMAND_DATA, 2, & wCmdStopPvtData);

	return 0;
}
