//
// GPSMapEdit
// (c) Konstantin Galichsky (kg@geopainting.com), 2003-2009
// Holux is registered trademark of Holux Technology Inc.
//
// Holux TRK tracks export.
//

# include "StdAfx.h"
# include "HoluxTrk.h"
# include "Save.h"
# include "Map.h"
# include "Status.h"
# include "SaveFileDlg.h"
# include "FileWriter.h"
# include "fast_floor.h"
# include "Options.h"
# include "Globals.h"

class CHoluxTrkExport : public CMapSaver {
  public:
	virtual bool CanSave (const map_t & _map) {return true;}
	virtual SaveStatus_t Save (HANDLE _hFile, const char * _strFileName, const map_t & _map) {
		return _Save (_hFile, _strFileName, "trk", "Holux MapShow tracks (*.trk)\0*.trk\0", _map);
	}

  private:
	SaveStatus_t _Save (HANDLE _hFile, const char * _strFileName, const char * _strExt, const char * _strFilter, const map_t & _map);
	SaveStatus_t _Save (HANDLE _hFile, const map_t & _map);
};

DECLARE_WPTRK_EXPORT (CHoluxTrkExport, "holux-trk", "trk", "Holux MapShow tracks (*.trk)");

SaveStatus_t CHoluxTrkExport::_Save (HANDLE _hFile, const char * _strFileName, const char * _strExt, const char * _strFilter, const map_t & _map) {
	if (_hFile) {
		return _Save (_hFile, _map);
	} else {
		//
		// Prepare file name.
		//
		char strFileName [MAX_PATH + 1];
		::strncpy (strFileName, _strFileName, MAX_PATH);
		strFileName [MAX_PATH] = '\0';
		if (! SaveFileDlg (_strExt, _strFilter, strFileName))
			return ssCanceled;

		//
		// Open the file.
		//
		const CFileForWrite hFile (strFileName);
		if (! hFile.IsValid ())
			return Error ("Can't create file:\n %s", strFileName);

		return _Save (hFile, _map);
	}
}

SaveStatus_t CHoluxTrkExport::_Save (HANDLE _hFile, const map_t & _map) {
	CTrkFile trk;
	::memset (& trk, 0, sizeof (trk));

	trk.hdr.dwSignature = c_dwTrkSignature;

	//
	// Write the file.
	//
	SetProgressStatus (0);

	size_t cOrdinal = 0;
	bool bLostPoints = false;

	size_t cRgn = 0;
	const size_t cRgns = _map.GetRgnCount ();
	for (size_t cKind = 0; cKind < rkMax; ++ cKind) {
		const layer_t & layer = _map.layers [cKind];
		const rgns_t & rgns = layer.rgns;
		for (rgns_t::const_iterator iRgn = rgns.begin (); iRgn != rgns.end (); ++ iRgn, ++ cRgn) {
			const rgn_t & rgn = * iRgn;
			if (rgn.IsDeleted () || ! rgn.IsTrack ())
				continue;

			// Save track.
			for (size_t cElement = 0; cElement < rgn.elements.size (); ++ cElement) {
				const rgn_element_t & element = rgn.elements [cElement];
				if (element.bDeleted)
					continue;
				const points_t & points = element.points;
				const size_t cPoints = points.size ();
				if (cPoints == 0)
					continue;

				if (cOrdinal >= c_cTrkMaxTracks)
					break;

				CTrkTrackLog & log = trk.logs [cOrdinal];
				for (size_t cPoint = 0; cPoint < cPoints; ++ cPoint) {
					if (cPoint >= c_cTrkMaxPoints) {
						bLostPoints = true;
						break;
					}

					CTrkPoint & pt = log.points [cPoint];
					const point_t & point = points [cPoint];
					pt.lX =  fast_floor (point.x*36000 + .5f);
					pt.lY = -fast_floor (point.y*36000 + .5f);
				}

				CTrkTrack & track = trk.hdr.tracks [cOrdinal];

				track.wPoints = cPoints;
				track.wNextPoint = cPoints;

				// Name.
				::strncpy (track.strName, rgn.strLabel.c_str (), sizeof (track.strName));
				track.strName [sizeof (track.strName) - 1] = '\0';

				// Comment.
				::memset (& track.strComment, ' ', sizeof (track.strComment));
				::strncpy (track.strComment, rgn.strComment.c_str (), rgn.strComment.size () < sizeof (track.strComment) ? rgn.strComment.size () : sizeof (track.strComment));

				track.dwOccupied = 1;
				track.dwShow = 1;
				track.dwFill = 0;

				++ cOrdinal;
			}

			SetProgressStatus (fast_floor (100.f*cRgn/cRgns + .5f));
		}
	}
//	trk.hdr.wTracks = cOrdinal;

	if (cOrdinal >= c_cTrkMaxTracks)
		Warning ("Only first 4 tracks are saved.\n(NOTE:this is the TRK format limitation).");
	if (bLostPoints)
		Warning ("Some points are not saved.\n(NOTE:this is the TRK format limitation).");

	CFileWriter fw (_hFile);
	fw.WriteBin (trk);
	fw.FlushBuffer ();

	SetProgressStatus (100);

	return ssOK;
}
