//
// GPSMapEdit
// (c) Konstantin Galichsky (kg@geopainting.com), 2003-2007
// ALAN and ALAN Map 500 are registered trademarks of ALAN Group.
//
// ALAN Map 500 WRT (legacy) waypoints parser.
//

# include "StdAfx.h"
# include "HoluxWrt.h"
# include "Map.h"
# include "Types.h"
# include "Load.h"
# include "Log.h"
# include "Status.h"
# include "ParseHelpers.h"
# include "Wrappers.h"
# include "Globals.h"

class CAlanWrtLoader : public CMapLoader {
  public:
	// Override.
	virtual bool IsValid (const char * _pData, size_t _cFileSize) const;
	virtual LoadStatus_t Load (const char * _pData, size_t _cFileSize, const char * _strFileName, map_t * _pMap);
};

DECLARE_WPTRK_LOADER (CAlanWrtLoader, "alan-wrt", "wrt", "ALAN Map 500 legacy waypoints (*.wrt)", ftiWpts | ftiRoutes);

bool CAlanWrtLoader::IsValid (const char * _pData, size_t _cFileSize) const {
	const CWrtFile * const pWrt = reinterpret_cast<const CWrtFile *> (_pData);
	if (_cFileSize < sizeof (CWrtFile) - sizeof (pWrt->unused))
		return false;
	if (_cFileSize > 32768)
		return false;

	if (pWrt->WptHdr.dwSignature    != c_dwWrtWptSignature)
		return false;
//	if (pWrt->RoutesHdr.dwSignature != c_dwWrtRoutesSignature)
//		return false;

	return true;
}

LoadStatus_t CAlanWrtLoader::Load (const char * _pData, size_t _cFileSize, const char * _strFileName, map_t * _pMap) {
	const char * const pEnd = _pData + _cFileSize;

	file_refs_t file_refs;
	file_refs.push_back (file_ref_t ());
	file_refs.back ().strFileName = _strFileName;
	file_refs.back ().bAttachment = true;

	// Get time zone shift.
	TIME_ZONE_INFORMATION tzi;
	if (::GetTimeZoneInformation (& tzi) == TIME_ZONE_ID_DAYLIGHT)
		tzi.Bias += tzi.DaylightBias;
	const double dtTimeShift = float (tzi.Bias)/(24*60);

	SetProgressStatus (0);

	rgn_t * pWaypoints [c_cWrtMaxWaypoints];
	::memset (pWaypoints, 0, sizeof (pWaypoints));

	const CWrtFile * const pWrtFile = reinterpret_cast<const CWrtFile *> (_pData);

	for (size_t cWpt = 0; cWpt < pWrtFile->WptHdr.wCurrentWaypoints; ++ cWpt) {
		const WORD wIdx = pWrtFile->WptHdr.wIdx [cWpt];
		if (wIdx == 0xFFFF) {
			assert (pWrtFile->WptHdr.btUsed [cWpt] == 0x00);
			continue;
		}
		assert (pWrtFile->WptHdr.btUsed [wIdx] == 0xFF);
		const CWrtWaypoint & wpt = pWrtFile->Waypoints [wIdx];

		//
		// Create waypoint.
		//
		rgn_t & rgn = _pMap->CreateRgn (rkPoint);
		rgn.SetWaypointType ();

		rgn.pFileRef  = & file_refs.back ();
		rgn.cBeginPos = reinterpret_cast<const char *> (& wpt) - _pData;
		rgn.cEndPos   = pEnd - _pData;
		rgn.cFileLine = -1;

		// Name.
		rgn.strLabel.assign   (wpt.strName,    GetLineLen (wpt.strName,    wpt.strName    + sizeof (wpt.strName)));
		// Comment.
		rgn.strComment.assign (wpt.strComment, GetLineLen (wpt.strComment, wpt.strComment + sizeof (wpt.strComment)));

		// Position.
		rgn.origins.push_back (rgn_origin_t ());
		rgn_origin_t & origin = rgn.origins.back ();
		origin.cLevel      = 0;
		origin.cFileOffset = rgn.cBeginPos;
		origin.point.x =  wpt.lX/36000.f;
		origin.point.y = -wpt.lY/36000.f;

		// Extra.
		waypoint_info_t * const pExtra = rgn.pExtra->wtr.CreateWaypoint ();

		// Active.
		if (wpt.btActive)
			pExtra->SetShow (waypoint_info_t::showName);
		else
			pExtra->SetShow (waypoint_info_t::showHide);

		// Date and Time.
		const int nDay    =   wpt.dwDate       & 0x1F;
		const int nMonth  = ((wpt.dwDate >> 8) & 0xF);
		const int nYear   =  (wpt.dwDate >> 16);

		const int nSecond =  wpt.dwTime       % 60;
		const int nMinute = (wpt.dwTime/60)   % 60;
		const int nHour   = (wpt.dwTime/3600) % 60;

		SYSTEMTIME st = {nYear, nMonth, 0, nDay, nHour, nMinute, nSecond, 0};
		if (::SystemTimeToVariantTime (& st, & pExtra->dtTime)) {
			pExtra->dtTime += dtTimeShift;
			pExtra->wFlags |= waypoint_info_t::defTime;
		}

		pWaypoints [wIdx] = & rgn;

		SetProgressStatus (25*cWpt/pWrtFile->WptHdr.wCurrentWaypoints);
	}

	// Routes.
	if (pWrtFile->RoutesHdr.dwSignature == c_dwWrtRoutesSignature) {
		for (size_t cRt = 0; cRt < pWrtFile->RoutesHdr.wCurrentRoutes && cRt < c_cWrtMaxRoutes; ++ cRt) {
			const WORD wIdx = pWrtFile->RoutesHdr.wIdx [cRt];
			if (wIdx == 0xFFFF) {
				assert (pWrtFile->RoutesHdr.btUsed [cRt] == 0x00);
				continue;
			}
			assert (pWrtFile->RoutesHdr.btUsed [wIdx] == 0xFF);
			const CWrtRoute & rt = pWrtFile->Routes [wIdx];

			const size_t cPoints = __min (rt.wWaypoints, c_cWrtMaxWaypointsInRoute);

			points_t points (cPoints);
			routepoints_t rpoints (cPoints);

			for (size_t cPoint = 0; cPoint < cPoints; ++ cPoint) {
				const size_t cWptIdx = rt.wWptIdx [cPoint];
				const size_t cWptIdx2 = pWrtFile->WptHdr.wIdx [cWptIdx];
				const CWrtWaypoint & wpt = pWrtFile->Waypoints [cWptIdx2];

				points [cPoint].x =  wpt.lX/36000.f;
				points [cPoint].y = -wpt.lY/36000.f;

				rpoints [cPoint].pWaypoint = pWaypoints [cWptIdx2];
			}

			//
			// Create route.
			//
			rgn_t & rgn = _pMap->CreateRgn (rkPolyline);
			rgn.SetRouteType ();

			rgn.pFileRef  = & file_refs.back ();
			rgn.cBeginPos = reinterpret_cast<const char *> (& rt) - _pData;
			rgn.cEndPos   = reinterpret_cast<const char *> (pEnd) - _pData;
			rgn.cFileLine = -1;
		//	rgn.cOrdinal = m_cOrdinal ++;

			// Name.
			rgn.strLabel.assign   (rt.strName,    GetLineLen (rt.strName,    rt.strName    + sizeof (rt.strName)));
			// Comment.
			rgn.strComment.assign (rt.strComment, GetLineLen (rt.strComment, rt.strComment + sizeof (rt.strComment)));

			rgn.elements.push_back (rgn_element_t ());
			rgn_element_t & element = rgn.elements.back ();
			element.cFileOffset = rgn.cBeginPos;

			//
			element.cLevel = 0;
			element.points.swap (points);

			route_info_t * const pRouteExtra = rgn.pExtra->wtr.CreateRoute ();
			pRouteExtra->points.swap (rpoints);

			// Date and Time.
			const int nDay    =   rt.dwDate       & 0x1F;
			const int nMonth  = ((rt.dwDate >> 8) & 0xF);
			const int nYear   =  (rt.dwDate >> 16);

			const int nSecond =  rt.dwTime       % 60;
			const int nMinute = (rt.dwTime/60)   % 60;
			const int nHour   = (rt.dwTime/3600) % 60;

			SYSTEMTIME st = {nYear, nMonth, 0, nDay, nHour, nMinute, nSecond, 0};
			if (::SystemTimeToVariantTime (& st, & pRouteExtra->dtTime)) {
				pRouteExtra->dtTime += dtTimeShift;
				pRouteExtra->wFlags |= route_info_t::defTime;
			}

			SetProgressStatus (25 + 25*cRt/pWrtFile->RoutesHdr.wCurrentRoutes);
		}
	}

	_pMap->file_refs.swap (file_refs);

	// Suppress warning about empty Zoom etc.
	_pMap->bNoHeader = true;
	_pMap->cHeaderLen = sizeof (CWrtWaypointsHdr);

	SetProgressStatus (50);

	return lsOK;
}
