//
// GPSMapEdit
// (c) Konstantin Galichsky (kg@geopainting.com), 2002-2009
// (c) Eugene Kladov, 2009
//
// Huffman code builder utility.
//

# include "StdAfx.h"
# include "HuffmanCode.h"
# include "fast_log2.h"

void CHuffmanCodeTable::BuildFlat () {
	if (m_cNonEmptyCodes == 0)
		return;

	::memset (m_pcLen, 0, sizeof (* m_pcLen)*m_cN);

	DWORD dwMax = 0; 
	size_t c;
	for (c = 0; c < m_cN; ++ c)
		if (IsCodePresent (c))
			dwMax = __max (dwMax, c);

	const size_t cLen = fast_log2 (dwMax);

	for (c = 0; c < m_cN; ++ c) {
		if (IsCodePresent (c)) {
			m_pcLen   [c] = cLen;
			m_pdwCode [c] = c;
		}
	}
	m_bFlat = true;
}

struct coin_t {
	std::vector<int> symbols;

	float weight; // The sum of all weights in this coin_t.

	void operator += (const coin_t & _coin);
	bool operator <  (const coin_t & _coin) const {return weight <  _coin.weight;}
	bool operator == (const coin_t & _coin) const {return weight == _coin.weight;}
};

void coin_t::operator += (const coin_t & _coin) {
	const size_t cSymbols = _coin.symbols.size ();
	for (size_t c = 0; c < cSymbols; ++ c)
		symbols.push_back (_coin.symbols [c]);
	weight += _coin.weight;
}

// Reverse order of bits.
static 
DWORD _Reverse (DWORD _dwValue, DWORD _dwBitLen) {
	DWORD dwRet = 0;
	for (size_t c = 0; c < _dwBitLen; ++ c)
		dwRet += ((_dwValue & (1 << c)) >> c) << (_dwBitLen - c - 1);
	return dwRet;
}

bool CHuffmanCodeTable::_Build () {
	if (fast_log2 (m_cNonEmptyCodes) > m_cMaxBitLen)
		return false;

	// See http://en.wikipedia.org/wiki/Package-merge_algorithm
	// Package-merge algorithm to generate length-limited Huffman codes.
	float sum = 0.0f;
	size_t c;
	for (c = 0; c < m_cN; ++ c)
		sum += m_pcP [c];
	assert (sum > 0.0f);

	std::vector<float> p (m_cN); // Normalized frequencies, avoid overflow.
	for (c = 0; c < m_cN; ++ c)
		p [c] += m_pcP [c]/sum;

	std::vector<coin_t> prev_row; // The previous row of coins.
	std::vector<coin_t> coins;    // The coins of the currently calculated row.

	// Generate all coins.
	for (c = 0; c < m_cN; ++ c) {
		if (m_pcP [c] == 0)
			continue; 

		coin_t coin;
		coin.weight = p [c];
		coin.symbols.push_back (c);
		coins.push_back (coin);
	}
	std::sort (coins.begin (), coins.end ());

	for (size_t j = 1; j <= m_cMaxBitLen; ++ j) {
		prev_row = coins;
		coins.clear ();

		const size_t cPrevRowSize = prev_row.size ();
		for (size_t c = 0; c < cPrevRowSize - 1; c += 2) {
			coins.push_back (prev_row [c]);
			coins.back () += prev_row [c + 1];
		}
		if (j < m_cMaxBitLen) { // We start with 1 row.
			for (size_t c = 0; c < m_cN; ++ c) {
				if (m_pcP [c] == 0)
					continue; 
				coin_t coin;
				coin.weight = p [c];
				coin.symbols.push_back (c);
				coins.push_back (coin);
			}

			std::sort (coins.begin (), coins.end ());
		}
	}

	// Keep only lowest N-1.
	coins.resize (m_cNonEmptyCodes - 1);

	// Clear all lengths.
	for (c = 0; c < m_cN; ++ c)
		m_pcLen [c] = 0;

	const size_t cCoins = coins.size ();
	for (c = 0; c < cCoins; ++ c) {
		coin_t & coin = coins [c];
		const size_t cSymbols = coin.symbols.size ();
		for (size_t j = 0; j < cSymbols; ++ j)
			m_pcLen [coin.symbols [j]] ++;
	}

	std::vector<size_t> vBitLenCount (m_cMaxBitLen + 1, 0);
	std::vector<size_t> vCodes       (m_cMaxBitLen + 1, 0);
	// Generate Canonical Huffman codes (http://en.wikipedia.org/wiki/Canonical_Huffman_code)
	//
	// Pseudo code:
	// code = 0
	// while more symbols:
	//     print symbol, code
	//     code = code + 1
	//     if next bit length > current bit length:
	//         code = code << 1

	for (c = 0; c < m_cN; ++ c)
		if (m_pcP [c] != 0)
			vBitLenCount [m_pcLen [c]] ++;

	for (c = 1; c <= m_cMaxBitLen; ++ c)
		vCodes [c] = (vCodes [c - 1] + vBitLenCount [c - 1]) << 1; // Start from, follow the algorithm in pseudo code.

	for (c = 0; c < m_cN; ++ c)
		if (m_pcP [c] != 0) // Note: we've special left-to-right read order in BitStreamParser.
			m_pdwCode [c] = _Reverse (vCodes [m_pcLen [c]] ++, m_pcLen [c]);

	return true;
}

void CHuffmanCodeTable::Build () {
	m_bFlat = false;
	
	if (m_cNonEmptyCodes == 0)
		return;

	if (GetNonEmptyCodes () <= 10000 || m_bCanBeFlat == false) {
		if (_Build ())
			return;
	}

	BuildFlat ();
}
