//
// GPSMapEdit
// (c) Konstantin Galichsky (kg@geopainting.com), 2002-2009
//
// 'File|Load' command.
//

# include "StdAfx.h"
# include "Load.h"
# include "Types.h"
# include "Menu.h"
# include "Map.h"
# include "Scale.h"
# include "Levels.h"
# include "NewMap.h"
# include "Log.h"
# include "Status.h"
# include "Position.h"
# include "AddMap.h"
# include "RecentFiles.h"
# include "MapSkins.h"
# include "MapView.h"
# include "MiniMap.h"
# include "Tool.h"
# include "Keyboard.h"
# include "Threads.h"
# include "EHFilter.h"
# include "Globals.h"

ULONGLONG CMapLoader::GetFileSize () const {
	ULARGE_INTEGER uli;
	uli.LowPart  = m_dwFileSize;
	uli.HighPart = m_dwFileSizeHi;
	return uli.QuadPart;
}

const char * CMapLoader::InitMapFileByParts (HANDLE _hFileMapping, DWORD _dwFileSize, DWORD _dwFileSizeHi) {
	m_pFilePart          = NULL;
	m_hFileMapping       = _hFileMapping;
	m_dwFileSize         = _dwFileSize;
	m_dwFileSizeHi       = _dwFileSizeHi;
	m_ullFileOffsetBegin = 0;
	m_ullFileOffsetEnd   = 0;

	return GetFilePart (0, 1);
}

void CMapLoader::InitMapFileByPartsIdle (const char * _pData, DWORD _dwFileSize, DWORD _dwFileSizeHi) {
	m_pFilePart          = _pData;
	m_hFileMapping       = NULL;
	m_dwFileSize         = _dwFileSize;
	m_dwFileSizeHi       = _dwFileSizeHi;
	m_ullFileOffsetBegin = 0;
	m_ullFileOffsetEnd   = GetFileSize ();
}

const size_t c_cFilePartSize = 0x2000000; // =32M.

const char * CMapLoader::GetFilePart (ULONGLONG _ullBeginOffset, ULONGLONG _ullEndOffset) {
	const ULONGLONG ullFileSize = GetFileSize ();

	assert (_ullBeginOffset < _ullEndOffset);
	assert (_ullEndOffset <= ullFileSize);

	if (m_ullFileOffsetBegin > _ullBeginOffset || _ullEndOffset > m_ullFileOffsetEnd) {
		assert (m_hFileMapping != NULL);

		if (m_pFilePart)
			::UnmapViewOfFile (m_pFilePart);

		SYSTEM_INFO si;
		::GetSystemInfo (& si);
		m_ullFileOffsetBegin = (_ullBeginOffset/si.dwAllocationGranularity)*si.dwAllocationGranularity;

		m_ullFileOffsetEnd   = _ullBeginOffset + __max (_ullEndOffset - _ullBeginOffset, c_cFilePartSize);
		if (m_ullFileOffsetEnd > ullFileSize)
			m_ullFileOffsetEnd = ullFileSize;

		m_pFilePart = reinterpret_cast<char *> (::MapViewOfFile (m_hFileMapping, FILE_MAP_READ, 0, m_ullFileOffsetBegin, m_ullFileOffsetEnd - m_ullFileOffsetBegin));
		if (m_pFilePart == NULL) {
			ReportWarning ("Can't read file. The file probably does not fit virtual memory.");
			return NULL;
		}
	}

	return m_pFilePart + (_ullBeginOffset - m_ullFileOffsetBegin);
}

//////////////////////////////////////////////////////////////

# if defined (_DEBUG)
# define _SYNC_LOAD // Enable one-thread loading in debug build.
# endif

static
LoadStatus_t _ExecuteLoadMapSafely2 (map_t * _pMap, const char * _pData, DWORD _dwFileSize, const char * _strFileName, CMapLoader * _pLoader) {
# ifdef WIN32
	try {
		assert (_pMap);
		return _pMap->LoadMap (_pData, _dwFileSize, _strFileName, _pLoader);
	} catch (const COutOfMemory & _oom) {
		_oom.ReportToMessageLog ();
		return lsFailed;
	} catch (const std::exception & _e) {
		ReportText ("Unhandled exception: %s.", _e.what ());
		return lsFailed;
	}
# else
	return _pMap->LoadMap (_pData, _dwFileSize, _strFileName, _pLoader);
# endif // WIN32
}

static
LoadStatus_t _ExecuteLoadMapSafely (map_t * _pMap, const char * _pData, DWORD _dwFileSize, const char * _strFileName, CMapLoader * _pLoader) {
# ifdef WIN32
	__try {
		assert (_pMap);
		return _ExecuteLoadMapSafely2 (_pMap, _pData, _dwFileSize, _strFileName, _pLoader);
	} __except (EHFilter (_exception_info ())) {
		return lsFailed;
	}
# else
	return _pMap->LoadMap (_pData, _dwFileSize, _strFileName, _pLoader);
# endif // WIN32
}

# ifdef WIN32
LoadStatus_t CMapLoader::LoadFile (const char * _strFileName, bool _bSilent, map_t * _pMap, const CFileTypeInfo ** _ppSaverInfo, const CFileTypeInfo ** _ppLoaderInfo) {
	const CFileForRead hFile (_strFileName);
	if (! hFile.IsValid ()) {
		ReportWarning ("Can't open file '%s'.", _strFileName);
		return lsFailed;
	}

	DWORD dwFileSizeHi = 0;
	DWORD dwFileSize = ::GetFileSize (hFile, & dwFileSizeHi);

	const size_t cFileSize = dwFileSizeHi > 0 ? 0xFFFFFFFF : dwFileSize;
	// NOTE: some ECW/J2K files may be over 512MB. For huge files, allow to map only the first 1MB to verify header.
	const DWORD dwHdrSize = dwFileSizeHi > 0 ? 10000000 : __min (dwFileSize, 10000000);

	CWinHandle hFileMapping = ::CreateFileMapping (hFile, NULL, PAGE_READONLY, 0, 0, NULL);
	const char * pData = reinterpret_cast<char *> (::MapViewOfFile (hFileMapping, FILE_MAP_READ, 0, 0, dwHdrSize));
	if (! hFileMapping.IsValid () || ! pData) {
		ReportWarning ("Can't read file '%s'.", _strFileName);
		return lsFailed;
	}

	// Find loader.
	DWORD dwFlags = 0;
	CMapLoader * const pLoader = CFileTypeInfo::CreateLoader (pData, cFileSize, _strFileName, _ppSaverInfo, dwFlags, _ppLoaderInfo);
	if (! pLoader) {
		::UnmapViewOfFile (pData);

		ReportWarning ("Unknown file type.", _strFileName);
		return lsFailed;
	}

	if (dwFlags & ftiDontMapFile) {
		// NOTE: some ECW/J2K files may be huge while their native API does not use this memory-mapped file image at all.
		::UnmapViewOfFile (pData);
		pData = NULL;
	} else if (dwFlags & ftiMapFileByParts) {
		if (dwFileSizeHi > 0 && (dwFlags & ftiLoadOver4GB) == 0) {
			::UnmapViewOfFile (pData);
			ReportWarning ("File '%s' is too huge (> 4GB).", _strFileName);
			return lsFailed;
		}
		if (dwFileSizeHi > 0 || dwHdrSize != cFileSize) {
			::UnmapViewOfFile (pData);
			pData = pLoader->InitMapFileByParts (hFileMapping, dwFileSize, dwFileSizeHi);
		} else {
			pLoader->InitMapFileByPartsIdle (pData, dwFileSize, dwFileSizeHi);
		}
	} else {
		if (dwFileSizeHi > 0 || dwFileSize >= 0x80000000) {
			::UnmapViewOfFile (pData);
			ReportWarning ("File '%s' is too huge (> 2GB).", _strFileName);
			return lsFailed;
		}

		if (dwHdrSize != cFileSize) {
			::UnmapViewOfFile (pData);
			pData = reinterpret_cast<char *> (::MapViewOfFile (hFileMapping, FILE_MAP_READ, 0, 0, dwFileSize));
			if (! pData) {
				ReportWarning ("Can't read file '%s'. The file probably does not fit virtual memory.", _strFileName);
				return lsFailed;
			}
		}
	}

	pLoader->m_bSilent = _bSilent;

	const LoadStatus_t ls = _ExecuteLoadMapSafely (_pMap, pData, cFileSize, _strFileName, pLoader);

	if (dwFlags & ftiWTR)
		g_bHaveWaypointsOrTracks = true;
	if (dwFlags & ftiBookmarks)
		g_bHaveBookmarks = true;

	if ((dwFlags & ftiRenderer) == 0 || ls != lsOK)
		delete pLoader;
	else {
		CMapRenderer * const pMapRenderer = static_cast<CMapRenderer *> (pLoader);
		map_renderer_t * const pRenderer = new map_renderer_t (pMapRenderer);
		pRenderer->strFileName = _strFileName;

		_pMap->renderers.push_back (pRenderer);
		_pMap->bNoHeader = true;
		_pMap->rectBound = pRenderer->rect;
		_pMap->fCosY     = ::cos (pRenderer->rect.MeanY ()*c_PI/180.f);
	}

	if (pData)
		::UnmapViewOfFile (pData);

	return ls;
}
# endif // WIN32

///////////////////////////////////////////////

HANDLE g_hLoadThread;
volatile bool g_bLoading = false;
volatile bool g_bStop = false;
file_names_t g_last_file_names;

struct load_data_t {
	map_t *      pMap;
	file_names_t file_names;
	size_t       cFileType;
	bool         bRemoveOldMap;
	bool         bSilent;
};

DWORD WINAPI _LoadMapProc (LPVOID);

void LoadMap (const char * _strFileName, size_t _cFileType, bool _bRemoveOldMap, bool _bSilent, map_t & _map) {
	file_names_t file_names;
	file_names.push_back (_strFileName);

	LoadMap (file_names, _cFileType, _bRemoveOldMap, _bSilent, _map);
}

void LoadMap (const file_names_t & _file_names, size_t _cFileType, bool _bRemoveOldMap, bool _bSilent, map_t & _map) {
	// Wait for previous parsing ends.
	if (g_hLoadThread)
		StopLoadThread ();

	g_bStop = false;

	g_bLoading = true;
	UpdateMenu ();

	// Simulate removal of map if it is initially empty.
	// NOTE: 'Add' map(s) to empty one should be strongly equivalent of 'Open'.
	if (! _bRemoveOldMap) {
		CCritSec::CLock lock (g_cs);
		if (_map.empty ())
			_bRemoveOldMap = true;
	}

	// Remove old map.
	if (_bRemoveOldMap && & _map == & g_map) {
		const bool bCancelled = ! RemoveMap ();
		if (bCancelled) {
			g_bLoading = false;
			UpdateMenu ();
			return;
		}
	}

	if (& _map == & g_map)
		g_last_file_names.clear ();

	// Prepare input data for the thread proc.
	load_data_t * const pParams = new load_data_t;
	pParams->pMap          = & _map;
	pParams->cFileType     = _cFileType;
	pParams->bRemoveOldMap = _bRemoveOldMap;
	pParams->bSilent       = _bSilent;
	std::set<string_t> names;
	for (file_names_t::const_iterator fi = _file_names.begin (); fi != _file_names.end (); ++ fi) {
		// Make absolute file path.
		char strFullPathName [MAX_PATH + 1];
		char * p = NULL;
		::GetFullPathName (fi->c_str (), MAX_PATH, strFullPathName, & p);

		if (names.find (strFullPathName) != names.end ())
			// Omit duplicated name.
			continue;

		names.insert (strFullPathName);
		pParams->file_names.push_back (strFullPathName);
		if (& _map == & g_map)
			g_last_file_names.push_back (strFullPathName);
	}

# if defined (_SYNC_LOAD)
	_LoadMapProc (pParams);
# else
	// Perform the loading in parallel thread.
	DWORD dwThreadId = 0;
	assert (! g_hLoadThread);
	g_hLoadThread = ::CreateThread (NULL, 0, & _LoadMapProc, pParams, 0, & dwThreadId);
	if (! g_hLoadThread) {
		ErrorMsgBox ("Internal error.");
		return;
	}
	// NOTE: the thread's handle 'g_hLoadThread' is not closed until StopLoadThread() or next load.
# endif
}

static
DWORD WINAPI _LoadMapProc1 (LPVOID _pParams) {
	// Get input params.
	load_data_t * const pParams = reinterpret_cast<load_data_t *> (_pParams);

	map_t & _map = * pParams->pMap;

	load_data_t params;
	params.file_names.swap (pParams->file_names);
	params.pMap          = pParams->pMap;
	params.cFileType     = pParams->cFileType;
	params.bRemoveOldMap = pParams->bRemoveOldMap;
	params.bSilent       = pParams->bSilent;
	delete pParams;

	CWaiter w;

	ClearLog ();

	// Save view state.
//	const size_t cOrigLevel = GetVisibleLevel ();
	const float fOrigScale  = g_fScale;
	const float fOrigX      = g_X + _map.rectBound.x0;
	const float fOrigY      = g_Y + _map.rectBound.y0;

	const DWORD dwStartTime = ::GetTickCount ();

	bool bAtLeastOneFileIsLoaded = false;
	bool bShowMessageLog = false;
	for (file_names_t::const_iterator fi = params.file_names.begin (); fi != params.file_names.end (); ++ fi) {
		const char * strFileName = fi->c_str ();
		const bool bFirstFile = fi == params.file_names.begin ();

		if (! bFirstFile)
			ReportText ("");
		ReportText ("Loading data from file '%s'...", strFileName);

		g_bStop = false;

		map_t map;
		const CFileTypeInfo * pSaverInfo = NULL;
		const CFileTypeInfo * pLoaderInfo = NULL;
		const LoadStatus_t ls = CMapLoader::LoadFile (strFileName, params.bSilent, & map, & pSaverInfo, & pLoaderInfo);
		switch (ls) {
			case lsOK:
				break;

			case lsCanceled:
				ReportText ("Loading was canceled by user request.");
				continue;

			case lsInterrupted:
				ReportText ("Loading was interrupted by user request.");
				bShowMessageLog = true;
				goto l_end_of_load;

			default:
			case lsFailed:
				ReportText ("ERROR: loading has terminated.");
				bShowMessageLog = true;
				goto l_end_of_load;
		}

		if (params.bRemoveOldMap && ! params.bSilent && & _map == & g_map)
			AddRecentFile (strFileName);

		//
		// Actualize the new map.
		//

		if (bFirstFile) {
			if (params.bRemoveOldMap) {
				CCritSec::CLock lock (g_cs);
				_map.swap (map);

				if (& _map == & g_map) {
					::strcpy (g_strFileName, strFileName);
					g_pSaverInfo  = pSaverInfo;
					g_pLoaderInfo = pLoaderInfo;
				}
			} else {
				if (! CanAddMap (map, _map)) {
					ReportWarning ("Sorry, the map can't be merged with previously loaded data due to different type set.");
					bShowMessageLog = true;
				} else if (! AddMap (map, _map)) {
					if (bAtLeastOneFileIsLoaded)
						goto l_end_of_load;

					ResetStatus ();
					g_bLoading = false;
					UpdateMenu ();

					return 1;
				}
			}
		} else {
			if (CanAddMap (map, _map)) {
				CCritSec::CLock lock (g_cs);
				_map.Join (map);
			} else {
				ReportWarning ("Sorry, the map can't be merged with previously loaded data due to different type set.");
				bShowMessageLog = true;
			}
		}

		bAtLeastOneFileIsLoaded = true;
	}
l_end_of_load:
	g_bStop = false;

	//
	// Report about results.
	//

	// Report warnings count.
	if (GetWarningsCount ())
		ReportText ("%d warning(s).", GetWarningsCount ());

	// Report time of loading.
	ReportText ("Load time is %0.2f s.", .001*(::GetTickCount () - dwStartTime));

	// Report user interruption or load failure.
	if (bShowMessageLog) {
		w.Stop ();
		if (! params.bSilent)
			ShowMessageLog ();
	}
	if (! bAtLeastOneFileIsLoaded) {
		ResetStatus ();
		g_bLoading = false;
		UpdateMenu ();

		return 1;
	}

	//
	// Check if map is empty.
	//
	bool bRectBoundInvalid;
	{
		CCritSec::CLock lock (g_cs);
		bRectBoundInvalid = _map.rectBound.Invalid ();
	}
	if (bRectBoundInvalid) {
		ResetStatus ();
		g_bLoading = false;
		UpdateMenu ();

		ReportText ("ERROR: The map is empty.");
		if (! params.bSilent)
			ShowMessageLog ();

		return 1;
	}

	//
	// Set the view to the map's center.
	//

	// NOTE: enable update of status bar/progress.
	g_bLoading = false;

	if (& _map == & g_map) {
		// Make sure that all new loaded data will be visible.
		g_bShowAttachments = true;

		UpdateMapSkin ();

		if (g_bZoomToFullAfterAdd || params.bRemoveOldMap) {
			SetVisibleLevelAuto ();
			if (params.bRemoveOldMap) {
				const bool bOK = RestoreRecentMapPosition ();
				if (! bOK)
					SetFullMapScale ();
			} else
				SetFullMapScale ();
		} else {
			// Restore map scale & position.
	//		SetVisibleLevel (cOrigLevel);
			{
				CCritSec::CLock lock (g_cs);

				g_fScale = fOrigScale;
				g_X = fOrigX - _map.rectBound.x0;
				g_Y = fOrigY - _map.rectBound.y0;
			}

			RefreshScales ();
			Refresh ();
		}

		//
		// Render the mini-map.
		//
		RenderMiniMap ();
	} else {
		UpdateMapSkin ();
		Refresh ();
	}

	//
	// Update GUI elements.
	//

	// NOTE: the code below this should not Refresh() to avoid deadlock!
	g_bLoading = false;

	UpdateTitle ();
	MakeLevelsMenu ();
	UpdateMenu ();

	MapSkins_UpdateGUI ();

	// NOTE: previous Refresh() does not display ruler due to g_bLoading==false.
	// Make ruler to draw now.
	Redraw ();

	// Redraw mini-map.
	RedrawMiniMap ();

	CTool::_OnNewMap ();

	SetStatusReady ();

	return 0;
}

static
DWORD WINAPI _LoadMapProc (LPVOID _pParams) {
	CCom com;

	const DWORD dwRet = _LoadMapProc1 (_pParams);
	::PostMessage (g_hWnd, WM_USER_FILE_LOADED, dwRet, 0);
	g_bLoading = false;

	return dwRet;
}

DWORD WaitLoadThread () {
# if ! defined (_SYNC_LOAD)
	if (g_hLoadThread == NULL)
		return 0;

	WaitThreadForStop (g_hLoadThread);

	DWORD dwExitCode = 0;
	if (! ::GetExitCodeThread (g_hLoadThread, & dwExitCode))
		dwExitCode = 0;

	::CloseHandle (g_hLoadThread);
	g_hLoadThread = NULL;

	return dwExitCode;
# else
	return 0;
# endif // _SYNC_LOAD
}

void StopLoadThread () {
	g_bStop = true;

	WaitLoadThread ();
}
