//
// GPSMapEdit
// (c) Konstantin Galichsky (kg@geopainting.com), 2009
// (c) Fiz Stein (fzzz62@yahoo.com)
// Lowrance is a registered trademark of Lowrance Electronics, Inc.
//
// Lowrance LCM saver.
//

# include "StdAfx.h"
# include "Resource.h"
# include "Map.h"
# include "Save.h"
# include "LowranceTypes.h"
# include "GarminTypes.h"
# include "NavitelTypes.h"
# include "RussaTypes.h"
# include "LcmBuilder/LowranceAPI.h"
# include "ParseHelpers.h"
# include "Status.h"
# include "Log.h"
# include "Dialog.h"
# include "PtTrim.h"
# include "File.h"
# include "fast_floor.h"
# include "Wrappers.h"
# include "Globals.h"

# define RADtoDEG  57.295779513082322
# define DEGtoRAD  0.017453292519943296
# define PI        3.1415926535897932384626433832795
# define EARTH_R   6356752.3142

struct Merc {
    long Lon;
    long Lat;    
};

Merc Degree2Merc (double _lat, double _lon) {
	Merc T;
	T.Lat = fast_floor (EARTH_R*::log (::tan ((_lat*DEGtoRAD + PI/2)/2)) + .5f);
	T.Lon = fast_floor (EARTH_R*_lon*DEGtoRAD + .5f);
	return T;
}

void Merc2Degree (long _lLat, long _lLon, double * _pfLat, double * _pfLon) {
	* _pfLon = RADtoDEG*_lLon/EARTH_R;
	* _pfLat = RADtoDEG*(2*::atan (::exp (_lLat/EARTH_R)) - PI/2);
}

////////////////////////////////////////////////////////////////////////

# pragma comment (lib, "LcmBuilder/lcmBuilderDLL.lib")
# pragma comment (linker, "/delayload:lcmBuilderDll.dll")

class CLowranceSaver : public CMapSaver {
  public:
	virtual bool CanSave (const map_t & _map) {return _map.pTypeSet == & g_tsLowrance || _map.pTypeSet == & g_tsGarmin || _map.pTypeSet == & g_tsNavitel || _map.pTypeSet == & g_tsRussa;}
	virtual bool PrepareToSave (const map_t & _map);
	virtual SaveStatus_t Save (HANDLE _hFile, const char * _strFileName, const map_t & _map);
};

DECLARE_MAP_EXPORT2 (CLowranceSaver, "lowrance-lcm", "lcm", "Lowrance LCM map (*.lcm)", ftiSaveCopyOfMap);

bool CLowranceSaver::PrepareToSave (const map_t & _map) {
	if (! CheckVectorData (_map))
		return false;
	// TODO: check other limitations here.
	return true;
}

SaveStatus_t CLowranceSaver::Save (HANDLE _hFile, const char * _strFileName, const map_t & _map) {
	if (::strcmp (_map.pTypeSet->strID, "Lowrance") != 0)
		return Error ("Please switch the type set to \"Lowrance\" before exporting map.");

	// Check if DLL is installed.
	char strDLL [MAX_PATH];
	GetLocalFileName ("lcmBuilderDLL.dll", strDLL);
	if (! DoesFileExist (strDLL)) {
		CDialog ld (IDD_LCMBUILDER, true);
		ld.ShowModal (g_hWnd);

		return ssCanceled;
	}
	
	remove ("build.log");

//	SetStatus (GetStr (IDS_STATUS_EXPORTING));
	SetProgressStatus (0);
	
	bool flag = false;
	
	double dLastLat = 0, dLastLon = 0;
	
	double dLonLeft  = _map.rectBound.x0;
	double dLatLeft  = _map.rectBound.y0;
	double dLonRight = _map.rectBound.x1;
	double dLatRight = _map.rectBound.y1;
	
	Merc posLeft  = Degree2Merc (dLatLeft,  dLonLeft);
	Merc posRight = Degree2Merc (dLatRight, dLonRight);
	posLeft.Lat /= 0x3a980;
	posLeft.Lon /= 0x3a980;
	if (dLatLeft < 0)
		-- posLeft.Lat;
	if (dLonLeft < 0)
		-- posLeft.Lon;
	posLeft.Lat *= 0x3a980;
	posLeft.Lon *= 0x3a980;
	
	const long lNumFilesRows = (posRight.Lat - posLeft.Lat)/0x3a980 + 1;
	const long lNumFilesCols = (posRight.Lon - posLeft.Lon)/0x3a980 + 1;
    const long lNumFiles = lNumFilesRows*lNumFilesCols;
	
	long lFileIndex = 0;
	long lObjectIndex = -1, lObjectIndex2;
	long lType;
	double dValue, dColor;
	
	long lGridCountX = 6;
	long lGridCountY = 6;
	
	for (long lLatIndex = posLeft.Lat; lLatIndex < posRight.Lat; lLatIndex += 0x3a980) {
		for (long lLonIndex = posLeft.Lon; lLonIndex < posRight.Lon; lLonIndex += 0x3a980) {
			Merc2Degree (lLatIndex + 0x10, lLonIndex + 0x10, & dLatLeft, & dLonLeft);
			
			char strFileName [MAX_PATH];
			::sprintf (strFileName, "%s", _strFileName);
			char * const pPeriod = ::strstr (strFileName, ".");
			* pPeriod = '\0';
			::sprintf (strFileName, "%s_%02d.mp", strFileName, lFileIndex ++);
			
			char strShortFileName [512];
			::strcpy (strShortFileName, strFileName);
			while (::strstr (strShortFileName, "\\") != 0)
				::strcpy (strShortFileName, ::strstr (strShortFileName, "\\") + 1);
			
			SetStatus (strShortFileName);

			::NewLCMFile (strFileName);
			::Bounds (& dLatLeft, & dLonLeft);
	
			unsigned long mode = 0;
			::SetMasterMode (& mode);

			::GetBounds (& dLatLeft, & dLonLeft, & dLatRight, & dLonRight);
			::SetGridArray (& lGridCountX, & lGridCountY);
			
			for (long lGridX = 0; lGridX < lGridCountX; ++ lGridX) {
				for (long lGridY = 0; lGridY < lGridCountY; ++ lGridY) {
					long lGrid = lGridY*lGridCountX + lGridX;
					::SetGrid (& lGrid);
					
					rect_t trimRect;
					const long lLat1 = lLatIndex + lGridY*(0x3a980/lGridCountY);
					const long lLat2 = lLat1 + (0x3a980/lGridCountY);
					const long lLon1 = lLonIndex + lGridX*(0x3a980/lGridCountX);
					const long lLon2 = lLon1 + (0x3a980/lGridCountX);
					
					double x, y;
					Merc2Degree (lLat1, lLon1, & y, & x);
					trimRect.y0 = y;
					trimRect.x0 = x;
					
					Merc2Degree (lLat2, lLon2, & y, & x);
					trimRect.y1 = y;
					trimRect.x1 = x;

					SetProgressStatus (fast_floor ((lGridX*lGridCountY + lGridY)*100.f/(lGridCountX*lGridCountY) + .5f));

					//
					// Points.
					//			
					size_t cPoint = 0;
					const layer_t & point_layer = _map.layers [rkPoint];
					const size_t cPoints = point_layer.cRgns;
					const rgns_t & points = point_layer.rgns;
					
					float lNumPoints = points.size ();
					size_t cNumProcessed = 0;

					rgns_t::const_iterator iRgn;
					for (iRgn = points.begin (); iRgn != points.end (); ++ iRgn, ++ cPoint) {
						++ cNumProcessed;

						const rgn_t & rgn = * iRgn;
						if (rgn.IsImmutable () || rgn.IsNotMapObject ())
							continue;
						const USHORT rgn_ushType = rgn.GetType ();

						if (rgn.origins [0].cLevel > 0)
							continue;

						for (size_t cOrigin = 0; cOrigin < rgn.origins.size (); ++ cOrigin) {
							const rgn_origin_t & origin = rgn.origins [cOrigin];
							if (origin.bDeleted)
								continue;

							if (origin.point.x < trimRect.x0 || origin.point.x > trimRect.x1)
								continue;
							if (origin.point.y < trimRect.y0 || origin.point.y > trimRect.y1)
								continue;

							// City names.
							if (rgn_ushType >= 0x100 && rgn_ushType <= 0x600) {
								long lDataSet = 0;
								lType = rgn_ushType/0x100;
								double dLat = origin.point.y;
								double dLon = origin.point.x;
								::CityName (& lDataSet, (char *) rgn.strLabel.c_str (), & lType, & dLat, & dLon);
							}

							// Marine POIs: wrecks.
							if (rgn_ushType >= 0x1627 && rgn_ushType <= 0x162b) {
								long lDataSet  = 0;
								char * strName = (char *) rgn.strLabel.c_str ();
								long lLevel    = 0x63;
								lType          = 119 + ((rgn_ushType - 0x1627)/2);//120; //119, 120, 121
								double dLat    = origin.point.y;
								double dLon    = origin.point.x;

								::Wreck (& lDataSet, strName, & lLevel, & lType, & dLat, & dLon);
							}

							// Marine POIs: wrecks.
							if (rgn_ushType >= 0x1621 && rgn_ushType <= 0x1625) {
								long lDataSet  = 0;
								char * strName = (char *) rgn.strLabel.c_str ();
								long lLevel    = 0x63;
								lType          = 116 + ((rgn_ushType - 0x1621)/2);//116; //117, 118
								double dLat    = origin.point.y;
								double dLon    = origin.point.x;

								::Wreck (& lDataSet, strName, & lLevel, & lType, & dLat, & dLon);
							}

							// Marine POIs: navaids.
							if (rgn_ushType >= 0x1601 && rgn_ushType <= 0x1617) {
								long lDataSet  = 0;
								char * strName = (char *) rgn.strLabel.c_str ();
								long lLevel    = 0x63;
								lType          = 122 + ((rgn_ushType - 0x1601)/2);//122; //123..to...133 inclusive
								double dLat    = origin.point.y;
								double dLon    = origin.point.x;
								long lIndex    = 0;
								if (rgn_ushType > 0x1609)
									lIndex = 1;

								::Navaid (& lDataSet, strName, & lLevel, & lType, & lIndex, & dLat, & dLon);
							}

							// POIs.
							if (rgn_ushType >= 0x2401 && rgn_ushType <= 0x6d01) {
								structPOI newPOI;
								newPOI.dLatitude   = origin.point.y;
								newPOI.dLongitude  = origin.point.x;
								newPOI.lNumAddress = 0;
								newPOI.lPOIType    = rgn_ushType >> 8;
							//	if(rgn_ushType==0x6d01)
							//		newPOI.lPOIType=133;

								const bool bHasCity = rgn.pExtra && rgn.pExtra->pCity && ! rgn.pExtra->pCity->IsDeleted ();

								newPOI.sAddressPost = "";
								newPOI.sAddressPre = "";
								if (bHasCity) {
									newPOI.lNumAddress  = ::atol (rgn.pExtra->strNumber.c_str ());
									newPOI.sAddressPost = (char *) rgn.pExtra->strStreetDesc.c_str ();
									newPOI.sCity        = (char *) rgn.pExtra->pCity->strName.c_str ();
								} else
									newPOI.sCity = "NoCity";
								
								if (bHasCity)
									newPOI.sCountry = (char *) rgn.pExtra->pCity->pRegion->pCountry->strName.c_str ();
								else
									newPOI.sCountry = "NoCountry";
								
								newPOI.sCounty      = "";
								newPOI.sDescription = "My Description";
								newPOI.sName        = (char *) rgn.strLabel.c_str ();
								newPOI.sPhone       = (char *) rgn.pExtra->strPhone.c_str ();
								
								if (bHasCity)
									newPOI.sState = (char *) rgn.pExtra->pCity->pRegion->strName.c_str ();
								else
									newPOI.sState = "AZ";
								
								if (rgn.pExtra && rgn.pExtra->pZip && ! rgn.pExtra->pZip->IsDeleted ())
									newPOI.sZip = (char *) rgn.pExtra->pZip->strName.c_str ();
								else
									newPOI.sZip = "00000";
								
								long lDataSet = 0;
								::POI (& lDataSet, & newPOI);
							}

							if (rgn_ushType >= 0x7001 && rgn_ushType <= 0x7201) {
								structPOI newPOI;
								newPOI.dLatitude    = origin.point.y;
								newPOI.dLongitude   = origin.point.x;
								newPOI.lNumAddress  = 0;
								newPOI.lPOIType     = rgn_ushType >> 8;
								newPOI.sAddressPost = "";
								newPOI.sAddressPre  = "";

								const bool bHasCity = rgn.pExtra && rgn.pExtra->pCity && ! rgn.pExtra->pCity->IsDeleted ();
								if (bHasCity) {
									newPOI.lNumAddress  = ::atol  (rgn.pExtra->strNumber     .c_str ());
									newPOI.sAddressPost = (char *) rgn.pExtra->strStreetDesc .c_str ();
									newPOI.sCity        = (char *) rgn.pExtra->pCity->strName.c_str ();
								}
								else
									newPOI.sCity = "NoCity";

								if (bHasCity)
									newPOI.sCountry = (char *) rgn.pExtra->pCity->pRegion->pCountry->strName.c_str ();
								else
									newPOI.sCountry = "NoCountry";

								newPOI.sCounty      = "";
								newPOI.sDescription = "My Description";
								newPOI.sName        = (char *) rgn.strLabel.c_str ();
								newPOI.sPhone       = (char *) rgn.pExtra->strPhone.c_str ();

								if (bHasCity)
									newPOI.sState = (char *) rgn.pExtra->pCity->pRegion->strName.c_str ();
								else
									newPOI.sState = "AZ";

								if (rgn.pExtra && rgn.pExtra->pZip && ! rgn.pExtra->pZip->IsDeleted ())
									newPOI.sZip = (char *) rgn.pExtra->pZip->strName.c_str ();
								else
									newPOI.sZip = "00000";

								long lDataSet = 0;
								::TrailPOI (& lDataSet, & newPOI);
							}

							if (rgn_ushType == 0x1000 && ::strlen (rgn.strLabel.c_str ()) > 0) {
								lType = 43;
								double dLat = origin.point.y;
								double dLon = origin.point.x;
								long lDataSet = 0;
								::WaterName (& lDataSet, (char *) rgn.strLabel.c_str (), & lType, & dLat, & dLon);
							}

							if (rgn_ushType == 0x1010 && ::strlen (rgn.strLabel.c_str ()) > 0) {
								lType = 44;
								double dLat = origin.point.y;
								double dLon = origin.point.x;
								long lDataSet = 0;
								::WaterName (& lDataSet, (char *) rgn.strLabel.c_str (), & lType, & dLat, & dLon);
							}
						}
					} // points

					//
					// Polylines.
					//			
					size_t cPolyline = 0;
					const layer_t & polyline_layer = _map.layers [rkPolyline];
					const size_t cPolylines  = polyline_layer.cRgns;
					const rgns_t & polylines = polyline_layer.rgns;
					char * sValue;
					char * end;

					cNumProcessed = 0;
					lNumPoints = polylines.size ();

					for (iRgn = polylines.begin (); iRgn != polylines.end (); ++ iRgn, ++ cPolyline) {
						++ cNumProcessed;

						const rgn_t & rgn = * iRgn;
						if (rgn.IsImmutable () || rgn.IsNotMapObject ())
							continue;

						for (size_t cElement = 0; cElement < rgn.elements.size (); ++ cElement) {
							const rgn_element_t & element = rgn.elements [cElement];
							if (element.bDeleted)
								continue;

							points_t points = element.points;
							const bool bHasPointExtras = element.pExtra && ! element.pExtra->PointExtras.empty ();

							epoints_t epoints;
							epoints_t epoints2;
							Points2EPoints (element.points, element.pExtra ? & element.pExtra->PointExtras : NULL, epoints2);

							polyepoints_t parts;
							TrimPolylineOutsideRect (trimRect, epoints2, bHasPointExtras, parts, (rgn_t *) & rgn);
							if (parts.size () == 0)
								continue;

							for (polyepoints_t::iterator iPart = parts.begin (); iPart != parts.end (); ++ iPart) {
								epoints_t & points = * iPart;
								const size_t cPoints = points.size ();
								if (cPoints == 0)
									continue;

								long lEndLevel = rgn.uchEndLevel;
								if (lEndLevel < element.cLevel)
									lEndLevel = element.cLevel;

								for (long lDataSet = element.cLevel; lDataSet <= lEndLevel; ++ lDataSet) {
									if (lDataSet > 1)
										continue;

									lObjectIndex = -1;

									for (size_t cPoint = 0; cPoint < cPoints; ++ cPoint) {
										const epoint_t & point = points [cPoint];
										dLastLon = point.point.x;
										dLastLat = point.point.y;
										long lType2 = -1;
										long lLevel = 58;

										switch (rgn.GetType ()) {
										//roads are special case, no break's
										case 0x1:
											if (lType2 == -1)
												lType2 = ROADSYMBOL_121;
										case 0x3:
											if (lType2 == -1)
												lType2 = ROADSYMBOL_120;
										case 0x5:
											if (lType2 == -1)
												lType2 = ROADSYMBOL_119;

										case 0x10:
											if (lType2 == -1) {
												lLevel = 0x11;
												lType2 = ROADSYMBOL_122;
											}

										case 0x15:
											if (lType2 == -1) {
												lLevel = 0x11;
												lType2 = ROADSYMBOL_123;
											}

										case 0x17:
											if (lType2 == -1) {
												lLevel = 0x11;
												lType2 = ROADSYMBOL_124;
											}

											if (lObjectIndex == -1) {
												dLastLat = rgn.ptLabel.y;
												dLastLon = rgn.ptLabel.x;

												long lRotation = (int) rgn.chLabelAngle;

												const Merc p1 = Degree2Merc (dLastLat, dLastLon);
												const Merc p2 = Degree2Merc (dLatLeft, dLonLeft);
												const double dpLat = (p1.Lat - p2.Lat)/((double) 0x3a980)*lGridCountX;
												const double dpLon = (p1.Lon - p2.Lon)/((double) 0x3a980)*lGridCountY;

												long lGrid3 = dpLat;
												lGrid3 *= lGridCountX;
												lGrid3 += dpLon;
												::SetGrid (& lGrid3);

												if (lGrid == lGrid3)
													lObjectIndex = ::RoadSymbol (
														& lDataSet, (char *) rgn.strLabel.c_str (),
														& lType2, & lRotation, & lLevel, & dLastLat, & dLastLon
													);

												::SetGrid (& lGrid);
												lObjectIndex = ::Road (& lDataSet, (char *) rgn.strLabel.c_str (), &lType2);
												lObjectIndex2 = ::RoadSegment (& lDataSet, & lObjectIndex); 
											}

											dLastLon = point.point.x;
											dLastLat = point.point.y;

											::RoadPoint (& lDataSet, & lObjectIndex, & lObjectIndex2, & dLastLat, & dLastLon);
											break;

										case 0x30:
										case 0x32:
											lType = rgn.GetType ();
											if (lObjectIndex == -1)
												lObjectIndex = ::Stream (& lDataSet,  & lType, (char *) rgn.strLabel.c_str ());
											
											::StreamPoint (& lDataSet, & lObjectIndex, & dLastLat, & dLastLon);
											break;

										case 0x40:
										case 0x42:
											lType = rgn.GetType ();
											if (lObjectIndex == -1)
												lObjectIndex = ::Trail (& lDataSet,  & lType, (char *) rgn.strLabel.c_str ());
											
											::TrailPoint (& lDataSet, & lObjectIndex, & dLastLat, & dLastLon);
											break;

										case 0x78:
										case 0x7a:
											lType = rgn.GetType ();
											if (lObjectIndex == -1)
												lObjectIndex = ::Boundary (& lDataSet,  & lType, (char *) rgn.strLabel.c_str ());
										
											::BoundaryPoint (& lDataSet, & lObjectIndex, & lType, & dLastLat, & dLastLon);
											break;

										case 0x70:
											if (lObjectIndex == -1)
												lObjectIndex = ::Rail (& lDataSet);

											::RailPoint (& lDataSet, & lObjectIndex, & dLastLat, & dLastLon);
											break;

										case 0x50:
										case 0x51:
										case 0x52:
										case 0x53:
										case 0x54:
										case 0x55:
										case 0x56:
										case 0x57:
										case 0x58:
										case 0x59:
										case 0x5a:
										case 0x5b:
										case 0x5c:
										case 0x5d:
										case 0x5e:
											lType = rgn.GetType ();
											dValue = 120;
											dColor = 0;
											sValue =  (char *) rgn.strLabel.c_str ();
											dValue = ::fabs (::strtod ( sValue, & end));
											if (lObjectIndex == -1)
												lObjectIndex = ::Contour (& lDataSet, & lType, & dValue);

											::ContourPoint (& lDataSet, & lObjectIndex, & dLastLat, & dLastLon);
											break;

										case 0x60:
										case 0x62:
										case 0x64:
										case 0x66:
											lType = rgn.GetType ();
											dValue = 120;
											sValue = (char *) rgn.strLabel.c_str ();
											dValue = ::fabs (::strtod (sValue, & end));
											if (lObjectIndex == -1)
												lObjectIndex = ::Elevation (& lDataSet, & lType, & dValue);

											::ElevationPoint (& lDataSet, & lObjectIndex, & dLastLat, & dLastLon);
											break;
										}  //switch
									} //for cpoint
								} // for dataset
							} // for point
						} // for element
					} // for polyline

					//
					// Polygons.
					//			
					size_t cPolygon = 0;
					const layer_t & polygon_layer = _map.layers [rkPolygon];
					const size_t cPolygons = polygon_layer.cRgns;
					const rgns_t & polygons = polygon_layer.rgns;
					//	rgns_t::const_iterator iRgn;
				
					cNumProcessed = 0;
					lNumPoints = polygons.size ();
					long lGrid3;
					double dpPointLat, dpPointLon;

					for (iRgn = polygons.begin (); iRgn != polygons.end (); ++ iRgn, ++ cPolygon) {
						++ cNumProcessed;

						const rgn_t & rgn = * iRgn;
						if (rgn.IsImmutable ()  || rgn.IsNotMapObject ())
							continue;

						for (size_t cElement = 0; cElement < rgn.elements.size (); ++ cElement) {
							const rgn_element_t & element = rgn.elements [cElement];
							if (element.bDeleted)
								continue;

							points_t points = element.points;
							const bool bHasPointExtras = element.pExtra && ! element.pExtra->PointExtras.empty ();

							epoints_t epoints;
							epoints_t epoints2;
							Points2EPoints (element.points, element.pExtra ? & element.pExtra->PointExtras : NULL, epoints2);

							polyepoints_t parts;
							TrimPolygonOutsideRect (trimRect, epoints2, parts);
							if (parts.size () == 0)
								continue;

							long lEndLevel = rgn.uchEndLevel;
							if (lEndLevel < element.cLevel)
								lEndLevel = element.cLevel;
							
							for (long lDataSet = element.cLevel; lDataSet <= lEndLevel; ++ lDataSet) {
								if (lDataSet > 1)
									continue;
							
								for (polyepoints_t::iterator iPart = parts.begin (); iPart != parts.end (); ++ iPart) {
									lObjectIndex = -1;

			//						const points_t & points = element.points;
			//						const size_t cPoints = points.size ();
			//						if (cPoints == 0)
			//							continue;
							
									epoints_t & points = * iPart;
									const size_t cPoints = points.size ();
									if (cPoints == 0)
										continue;
								
									for (cPoint = 0; cPoint < cPoints; ++ cPoint) {
										const epoint_t & point = points [cPoint];
										dLastLon = point.point.x;
										dLastLat = point.point.y;

										switch (rgn.GetType ()) {
										case 0x1:
											if (lObjectIndex == -1) {
												lObjectIndex = ::UrbanArea (& lDataSet, (char *) rgn.strLabel.c_str ());
											}
										
											::UrbanAreaPoint (& lDataSet, & lObjectIndex, & dLastLat, & dLastLon);
											break;

										case 0x10:
											if (lObjectIndex == -1) {
												lObjectIndex = ::Lake (& lDataSet);
											}
											
											::LakePoint (& lDataSet, & lObjectIndex, & dLastLat, & dLastLon);
											break;

										case 0x12:
											if (lObjectIndex == -1) {
												lObjectIndex = ::Swamp (& lDataSet);
											}

											::SwampPoint (& lDataSet, & lObjectIndex, & dLastLat, & dLastLon);
											break;
																		
										case 0x15:
											if (lObjectIndex == -1) {
												lObjectIndex = ::Ocean (& lDataSet);
											}

											::OceanPoint (& lDataSet, & lObjectIndex, & dLastLat, & dLastLon);
											break;
										
										case 0x20: //0 "National Forest"
										case 0x25: //1 "National Park"
										case 0x2a: //2 "State Park"
										case 0x2f: //3 "City Park"
										case 0x34: //4 "Golf course"
										case 0x39: //5 "Cemetery"
											lType = (rgn.GetType ()- 0x20)/5;

											if (lType >= 3 && lObjectIndex == -1) {

												dpPointLat = rgn.ptLabel.y;
												dpPointLon = rgn.ptLabel.x;

												//								long lLevel = 58;
												long lRotation = (int) rgn.chLabelAngle;
												Merc p1 = Degree2Merc (dpPointLat, dpPointLon);
												Merc p2 = Degree2Merc (dLatLeft, dLonLeft);

												double dpLat = (p1.Lat - p2.Lat)/((double)0x3a980)*lGridCountX;
												double dpLon = (p1.Lon - p2.Lon)/((double)0x3a980)*lGridCountY;

												//gridY * countX + gridX
												lGrid3 = dpLat;
												lGrid3 *= lGridCountX;
												lGrid3 += dpLon;
												//SetGrid(&lGrid);

												long lNameType = CITYNAME_LEVEL_142;
												switch (rgn.GetType ()) {
													case 0x2f: //"City Park"
														lNameType = CITYNAME_LEVEL_142;
														break;
													case 0x34: //"Golf course"
														lNameType = CITYNAME_LEVEL_141;
														break;
													case 0x39: //"Cemetery"
														lNameType = CITYNAME_LEVEL_143;
														break;
												}
											
												if (::strlen (rgn.strLabel.c_str ()) > 0 && lGrid3 < 36) {
													::SetGrid (& lGrid3);
													::CityName (& lDataSet, (char *) rgn.strLabel.c_str (), & lNameType, & dpPointLat, & dpPointLon);
													::SetGrid (& lGrid);
												}
											}

											if (lObjectIndex == -1) {
												lObjectIndex = ::Park (& lDataSet, (char *) rgn.strLabel.c_str (), & lType);
											}

											::ParkPoint (& lDataSet, & lObjectIndex, & lType, & dLastLat, & dLastLon);
											break;
										}  //switch
									} //for cpoint
								} // for dataset
							} // for point
						} // for element
					} // for polygon
				}
			}

			SetStatus ("Writing...");
			SetProgressStatus (100);

			::BuildLCM ();
			::ClearLCM ();
		}
	}

	return ssOK;
}
