//
// GPSMapEdit
// (c) Konstantin Galichsky (kg@geopainting.com), 2009
// (c) Fiz Stein (fzzz62@yahoo.com)
// Lowrance is a registered trademark of Lowrance Electronics, Inc.
//
// The Lowrance(R) LCM-specific type set.
//

# include "StdAfx.h"
# include "LowranceTypes.h"
# include "GarminTypes.h"
# include "GarminRound.h"
# include "Map.h"
# include "Types.h"
# include "TypesDraw.h"
# include "TypesPOIDraw.h"
# include "Globals.h"

HPEN g_hpenLCMDepth1 [3];
HPEN g_hpenLCMDepth2 [3];
HPEN g_hpenLCMDepth3 [3];
HPEN g_hpenLCMDepth4 [3];
HPEN g_hpenLCMDepth5 [3];

const COLORREF c_colLCMDepth [5] = {
	RGB (0,  0,   0),
	RGB (0,  0,   255),
	RGB (0,  132, 132),
	RGB (0,  255, 255),
	RGB (255,255, 255)
};

//////////////////////////////////////////////////////

bool Lowrance_IsRoad (DWORD _dwType) {
	if (_dwType == 0x42 || _dwType == 0x43)
		return true;

	return Garmin_IsRoad (_dwType);
}

bool Lowrance_IsRoad (const rgn_t & _rgn) {
	return _rgn.IsPolyline () && Lowrance_IsRoad (_rgn.GetType ());
}

bool Lowrance_IsOneWayRoad (const rgn_t & _rgn) {
	return _rgn.pExtra && _rgn.pExtra->bOneWay;
}

bool Lowrance_HasAddress (DWORD _dwType) {
	const DWORD dwType = _dwType & 0xFF00;
	return (0x2400 <= dwType && dwType <= 0x7200) || (_dwType >= 0x100 && _dwType <= 0x600);
}

bool Lowrance_HasAddress (const rgn_t & _rgn) {
	return Lowrance_HasAddress (_rgn.GetType ());
}

bool Lowrance_HasCityAndZip (const rgn_t & _rgn) {
	return Lowrance_HasAddress (_rgn.GetType ());
}

bool Lowrance_IsCity (DWORD _dwType) {
	const DWORD dwType = _dwType & 0xFF00;
	return 0x0100 <= dwType && dwType <= 0x600;
}

bool Lowrance_IsCity (const rgn_t & _rgn) {
	return _rgn.IsPoint () && Lowrance_IsCity (_rgn.GetType ());
}

/////////////////////////////////////////////////////

struct POI_desc_t {
	BYTE btPointCode;
	BYTE btPOISubcodeMin;
	BYTE btPOISubcodeMax;
	BYTE btCategory;

	char * strTypeName;

	// The icon for point/POI. NOTE: can't be NULL; use g_hcurUnknown for default.
	HCURSOR *      phCursor;

	label_desc_t * pLabelDesc;
};

static POI_desc_t g_MarinePOIs [] = {
	{0x16, 0x01, 0,    catMarineNavaids, "Navaid, white",                  & g_hcurNavaidWhite},
	{0x16, 0x03, 0,    catMarineNavaids, "Navaid, red",                    & g_hcurNavaidRed},
	{0x16, 0x05, 0,    catMarineNavaids, "Navaid, green",                  & g_hcurNavaidGreen},
	{0x16, 0x07, 0,    catMarineNavaids, "Navaid, violet",                 & g_hcurNavaidViolet},

	{0x16, 0x11, 0,    catMarineNavaids, "Daybeacon, red triangle",        & g_hcurBeaconTriangle},
	{0x16, 0x13, 0,    catMarineNavaids, "Daybeacon, green square",        & g_hcurBeaconSquare},
	{0x16, 0x15, 0,    catMarineNavaids, "Daybeacon, white diamond",       & g_hcurBeaconDiamond},
	{0x16, 0x17, 0,    catMarineNavaids, "Unlit navaid, orange",           & g_hcurUNavaidOrange},

	{0x16, 0x21, 0,    catMarineDanger,  "Obstruction",			           & g_hcurObstruction},
	{0x16, 0x23, 0,    catMarineDanger,  "Rock, submerged at low water",   & g_hcurReef},
	{0x16, 0x25, 0,    catMarineDanger,  "Obstr., visible at high water",  & g_hcurObstructionAwash},
	{0x16, 0x27, 0,    catMarineDanger,  "Submerged wreck, dangerous",     & g_hcurWreckDangerous},
	{0x16, 0x29, 0,    catMarineDanger,  "Wreck",                          & g_hcurWreckVisible},
	{0x16, 0x2b, 0,    catMarineDanger,  "Submerged wreck, non-dangerous", & g_hcurWreckNonDangerous},
};

static POI_desc_t g_POIs [] = {
	{0x01, 0x00, 0,    catCities,        "Large city (over 10M)",          & g_hcurCityMega,   & g_BigCityLabel},
	{0x02, 0x00, 0,    catCities,        "Large city (5-10M)",             & g_hcurCityMega,   & g_BigCityLabel},
	{0x03, 0x00, 0,    catCities,        "Medium city (0.5-1M)",           & g_hcurCityMedium, & g_CityLabel},
	{0x04, 0x00, 0,    catCities,        "Medium city (100-200K)",         & g_hcurCitySmall,  & g_CityLabel},
	{0x05, 0x00, 0,    catCities,        "Small city (50-100K)",           & g_hcurCitySmall,  & g_CityLabel},
	{0x06, 0x00, 0,    catCities,        "Small city (20-50K)",            & g_hcurCitySmall,  & g_CityLabel},

	{0x10, 0x00, 0,    catLabels,        "Water Label",                    & g_hcurCitySmall,  & g_CityLabel},
	{0x10, 0x10, 0,    catLabels,        "Stream Label",                   & g_hcurCitySmall,  & g_CityLabel},

	//POIs 36-107
	{0x24, 0x01, 0, 	catServices,     "Post Office",                    & g_hcurPostOffice},
	{0x25, 0x01, 0, 	catReligion,     "Church",                         & g_hcurChurch},
	{0x26, 0x01, 0, 	catShopping,     "Convenience Store",              & g_hcurShopping},
	{0x27, 0x01, 0, 	catMarine,       "Fish Hatchery",                  & g_hcurFish},
	{0x28, 0x01, 0, 	catAuto,         "Auto Repair",                    & g_hcurAutoRepair},
	{0x29, 0x01, 0, 	catFood,         "Restaurant",                     & g_hcurDining},
	{0x2a, 0x01, 0, 	catShopping,     "Grocery Store",                  & g_hcurShopping},
	{0x2b, 0x01, 0, 	catEmrgGov,      "Police",                         & g_hcurPolice},
	{0x2c, 0x01, 0, 	catServices,     "Financial Services",             & g_hcurBank},
	{0x2d, 0x01, 0, 	catAuto,         "Car Wash",                       & g_hcurAutoWash},
	{0x2e, 0x01, 0, 	catShopping,     "Hardware Store",                 & g_hcurBusStation},
	{0x2f, 0x01, 0, 	catShopping,     "Department Store",               & g_hcurDepartmentStore},
	{0x30, 0x01, 0, 	catShopping,     "Pharmacy",                       & g_hcurPharmacy},
	{0x31, 0x01, 0,     catFood,         "Pizza",                          & g_hcurDining},
	{0x32, 0x01, 0, 	catEducation,    "University",                     & g_hcurCollege},
	{0x33, 0x01, 0, 	catTransport,    "Bus Line",                       & g_hcurBusStation},
	{0x34, 0x01, 0, 	catTransport,    "Airport",                        & g_hcurAirport},
	{0x35, 0x01, 0, 	catShopping,     "Auto Parts",                     & g_hcurCarDealer},
	{0x36, 0x01, 0, 	catAuto,         "Gas Station",                    & g_hcurPetrol},
	{0x37, 0x01, 0, 	catEmrgGov,      "Medical Service",                & g_hcurHospital},
	{0x38, 0x01, 0, 	catEmrgGov,      "Hospital",                       & g_hcurHospital},
	{0x39, 0x01, 0, 	catEntertainment,"Night Club",                     & g_hcurBar},
	{0x3a, 0x01, 0, 	catEntertainment,"Amusement",                      & g_hcurAmusementPark},
	{0x3b, 0x01, 0, 	catEntertainment,"Movie Theater",                  & g_hcurLeasure},
	{0x3c, 0x01, 0,     catFood,         "Fast Food",                      & g_hcurDining},
	{0x3d, 0x01, 0,     catFood,         "Asian Food",                     & g_hcurDining},
	{0x3e, 0x01, 0,     catFood,         "Sandwiches",                     & g_hcurDining},
	{0x3f, 0x01, 0,     catFood,         "Steak House",                    & g_hcurDining},
	{0x40, 0x01, 0, 	catEntertainment,"Bowling Center",                 & g_hcurBowling},
	{0x41, 0x01, 0, 	catEntertainment,"Skating Rink",                   & g_hcurIceSkate},
	{0x42, 0x01, 0, 	catAttractions,  "Museum",                         & g_hcurMuseum},
	{0x43, 0x01, 0, 	catTransport,    "Auto Rental",                    & g_hcurAutoRental},
	{0x44, 0x01, 0, 	catTransport,    "Taxi",                           & g_hcurAutoRental},
	{0x45, 0x01, 0, 	catLodging,      "Motel",                          & g_hcurHotel},
	{0x46, 0x01, 0, 	catAuto,         "Wrecker Service",                & g_hcurAutoRepair},
	{0x47, 0x01, 0,     catFood,         "Barbeque",                       & g_hcurDining},
	{0x48, 0x01, 0, 	catMarine,       "Fishing Tackle",                 & g_hcurFish},
	{0x49, 0x01, 0,     catFood,         "Deli",                           & g_hcurDining},
	{0x4a, 0x01, 0,     catFood,         "Mexican Food",                   & g_hcurDining},
	{0x4b, 0x01, 0, 	catEntertainment,"Golf",                           & g_hcurGolf},
	{0x4c, 0x01, 0, 	catShopping,     "General Merchandise",            & g_hcurShopping},
	{0x4d, 0x01, 0, 	catEntertainment,"Swimming Pool",                  & g_hcurSwimming},
	{0x4e, 0x01, 0, 	catShopping,     "Hunting Supply",                 & g_hcurBusStation},
	{0x4f, 0x01, 0, 	catMarine,       "Marina",                         & g_hcurWharf},
	{0x50, 0x01, 0, 	catAttractions,  "Historical Site",                & g_hcurGhostTown},
	{0x51, 0x01, 0, 	catLodging,      "Campground",                     & g_hcurCamping},
	{0x52, 0x01, 0,     catFood,         "Ice Cream Parlor",               & g_hcurDining},
	{0x53, 0x01, 0, 	catAttractions,  "Park",                           & g_hcurPark},
	{0x54, 0x01, 0, 	catLodging,      "Bed & Breakfast",                & g_hcurHotel},
	{0x55, 0x01, 0, 	catLodging,      "Resort",                         & g_hcurHotel},
	{0x56, 0x01, 0, 	catAuto,         "Truck Stop",                     & g_hcurTruckStop},
	{0x57, 0x01, 0, 	catMarine,       "Boat Repair",                    & g_hcurWharf},
	{0x58, 0x01, 0, 	catShopping,     "Retail Sports",                  & g_hcurSports},
	{0x59, 0x01, 0,     catFood,         "Sea Food",                       & g_hcurDining},
	{0x5a, 0x01, 0, 	catMarine,       "Boat Dealer",                    & g_hcurWharf},
	{0x5b, 0x01, 0, 	catShopping,     "Shopping Center",                & g_hcurShopping},
	{0x5c, 0x01, 0,     catFood,         "Italian Food",                   & g_hcurDining},
	{0x5d, 0x01, 0,     catFood,         "Banquet Room",                   & g_hcurDining},
	{0x5e, 0x01, 0, 	catTransport,    "Transportation (Other)",         & g_hcurBusStation},
	{0x5f, 0x01, 0,     catFood,         "Coffee Shop",                    & g_hcurDining},
	{0x60, 0x01, 0, 	catShopping,     "Bait Shop",                      & g_hcurWharf},
	{0x61, 0x01, 0, 	catAttractions,  "Tourist",                        & g_hcurBusStation},
	{0x62, 0x01, 0, 	catRecreation,   "Hunting Preserve",               & g_hcurBusStation},
	{0x63, 0x01, 0,     catReligion,     "Synagogue",                      & g_hcurChurch},
	{0x64, 0x01, 0, 	catEntertainment,"Theatre",                        & g_hcurTheatre},
	{0x65, 0x01, 0, 	catEntertainment,"Rec Center",                     & g_hcurSports},
	{0x66, 0x01, 0, 	catOthers,       "RailRoad",                       & g_hcurBusStation},
	{0x67, 0x01, 0, 	catMarine,       "Boat Rental",                    & g_hcurWharf},
	{0x68, 0x01, 0, 	catRecreation,   "Archery",                        & g_hcurBusStation},
	{0x69, 0x01, 0, 	catMarine,       "Marine (Other)",                 & g_hcurWharf},
	{0x6a, 0x01, 0, 	catServices,     "RV Repair",                      & g_hcurService},
	{0x6b, 0x01, 0, 	catAttractions,  "Zoo",                            & g_hcurZoo},
//	{0x6c, 0x01, 0, 	catLodging,      "Cabin",                          & g_hcurCabin},
	{0x6c, 0x01, 0, 	catRecreation,   "GeoCache",                       & g_hcurGeoCache},

	{0x70, 0x01, 0, 	catRecreation,   "Trail - Hiking",                 & g_hcurTrailHead},
	{0x72, 0x01, 0, 	catRecreation,   "Trail - Biking",                 & g_hcurTrailHead},
};

struct polyline_desc_t {
	BYTE btCode;
	BYTE btCategory;

	char * strTypeName;

	// The pens to draw polylines .
	HPEN *         phPen;
	// The pen for second pass. NOTE: NULL means not to draw 2nd pass.
	HPEN *         phPen2;

	label_desc_t * pLabelDesc;
};

static polyline_desc_t g_Polylines [] = {
	{0x1,  catRoads,      "Highway - Interstate",    & g_hpenHighway1,	    & g_hpenHighway2,     & g_RoadLabel},
	{0x3,  catRoads,      "Highway - US",            & g_hpenPrincipalHw,   NULL,                 & g_RoadLabel},
	{0x5,  catRoads,      "Highway - State",         & g_hpenPrincipalHw,   NULL,                 & g_RoadLabel},
	{0x10, catRoads,      "Road - Rural",            & g_hpenThinRoad,		NULL,                 & g_RoadLabel},
	{0x15, catRoads,      "Street - Major City",     & g_hpenThinRoad,      NULL,                 & g_RoadLabel},
	{0x17, catRoads,      "Street - Minor City",     & g_hpenThinRoad,      NULL,                 & g_RoadLabel},

	{0x30, catWater,      "Stream - Intermittent",   & g_hpenWater,         NULL,                 & g_LineRiverLabel},
	{0x32, catWater,      "Stream",                  & g_hpenWater,         NULL,                 & g_LineRiverLabel},

	{0x40, catRoads,      "Trail - Hiking",          & g_hpenTrail,         NULL,                 & g_RoadLabel},
	{0x42, catRoads,      "Trail - Bike",            & g_hpenTrail,         NULL,                 & g_RoadLabel},

	{0x50, catElevation,  "Contour - 0",             & g_hpenLCMDepth1 [0], NULL,                 & g_DepthLabel},
	{0x51, catElevation,  "Contour - 0",             & g_hpenLCMDepth2 [0], NULL,                 & g_DepthLabel},
	{0x52, catElevation,  "Contour - 0",             & g_hpenLCMDepth3 [0], NULL,                 & g_DepthLabel},
	{0x53, catElevation,  "Contour - 0",             & g_hpenLCMDepth4 [0], NULL,                 & g_DepthLabel},
	{0x54, catElevation,  "Contour - 0",             & g_hpenLCMDepth5 [0], NULL,                 & g_DepthLabel},

	{0x55, catElevation,  "Contour - 1",             & g_hpenLCMDepth1 [1], NULL,                 & g_DepthLabel},
	{0x56, catElevation,  "Contour - 1",             & g_hpenLCMDepth2 [1], NULL,                 & g_DepthLabel},
	{0x57, catElevation,  "Contour - 1",             & g_hpenLCMDepth3 [1], NULL,                 & g_DepthLabel},
	{0x58, catElevation,  "Contour - 1",             & g_hpenLCMDepth4 [1], NULL,                 & g_DepthLabel},
	{0x59, catElevation,  "Contour - 1",             & g_hpenLCMDepth5 [1], NULL,                 & g_DepthLabel},

	{0x5a, catElevation,  "Contour - 2",             & g_hpenLCMDepth1 [2], NULL,                 & g_DepthLabel},
	{0x5b, catElevation,  "Contour - 2",             & g_hpenLCMDepth2 [2], NULL,                 & g_DepthLabel},
	{0x5c, catElevation,  "Contour - 2",             & g_hpenLCMDepth3 [2], NULL,                 & g_DepthLabel},
	{0x5d, catElevation,  "Contour - 2",             & g_hpenLCMDepth4 [2], NULL,                 & g_DepthLabel},
	{0x5e, catElevation,  "Contour - 2",             & g_hpenLCMDepth5 [2], NULL,                 & g_DepthLabel},

	{0x60, catElevation,  "Elevation - 0",           & g_hpenHeight1,       NULL,                 & g_ContourLabel},
	{0x62, catElevation,  "Elevation - 1",           & g_hpenHeight1,       NULL,                 & g_ContourLabel},
	{0x64, catElevation,  "Elevation - 2",           & g_hpenHeight2,       NULL,                 & g_ContourLabel},
	{0x66, catElevation,  "Elevation - 3",           & g_hpenHeight3,       NULL,                 & g_ContourLabel},

	{0x70, catRoads,      "Railroad",                & g_hpenRailroad1,     & g_hpenRailroad2,    & g_DefLabel},

	{0x78, catBoundaries, "County/parish boundary",  & g_hpenBoundaryCnt1,  & g_hpenBoundaryCnt2, & g_BoundaryLabel},
	{0x7a, catBoundaries, "State/province boundary", & g_hpenBoundaryReg1,  & g_hpenBoundaryReg2, & g_BoundaryLabel},
	{0x7c, catBoundaries, "International boundary",  & g_hpenBoundaryInt1,  & g_hpenBoundaryInt2, & g_BoundaryLabel},

//	{0xff, "Grid Bounds",			& g_hpenThinRoad,     NULL,                 & g_RoadLabel},
};

struct polygon_desc_t {
	BYTE btCode;
	BYTE btCategory;

	char * strTypeName;

	// The brush to fill polygons.
	HBRUSH *       phBrush;
	// The pen to draw border.
	HPEN *         phPen;

	label_desc_t * pLabelDesc;
};

static polygon_desc_t g_Polygons [] = {
	{0x01, catCities,     "Urban Area",      & g_hbrLargeUrbanHousing, & g_hpenNull},
	{0x10, catWater,      "Lake/River",      & g_hbrWater,             & g_hpenCoast, & g_LakeLabel},
	{0x12, catWater,      "Swamp",           & g_hbrSwamp,             & g_hpenNull,  & g_LakeLabel},
//	{0x51, catWater,      "Wetland/swamp",   & g_hbrSwamp,             & g_hpenNull,  & g_LakeLabel},

	{0x15, catWater,      "Ocean",           & g_hbrWater,             & g_hpenCoast, & g_LakeLabel},
	{0x20, catRecreation, "National Forest", & g_hbrForest,            & g_hpenNull},
	{0x25, catRecreation, "National Park",   & g_hbrForest,            & g_hpenNull},
	{0x2a, catRecreation, "State Park",      & g_hbrForest,            & g_hpenNull},
	{0x2f, catRecreation, "City Park",       & g_hbrForest,            & g_hpenNull},

	{0x34, catSports,     "Golf course",     & g_hbrGolf,              & g_hpenNull,      },
	{0x39, catManMade,    "Cemetery",        & g_hbrCemetery,          & g_hpenBlackThin, }
};

static
void InitLowranceTypes (CTypeSet::types_t & _types) {
	// Init Lowrance-specific types.
	for (int i = 0; i < 3; ++ i){
		if (g_hpenLCMDepth1 [i]	== NULL) g_hpenLCMDepth1 [i] = ::CreatePen (PS_SOLID,	3 - i, c_colLCMDepth [0]);
		if (g_hpenLCMDepth2 [i]	== NULL) g_hpenLCMDepth2 [i] = ::CreatePen (PS_SOLID,	3 - i, c_colLCMDepth [1]);
		if (g_hpenLCMDepth3 [i]	== NULL) g_hpenLCMDepth3 [i] = ::CreatePen (PS_SOLID,	3 - i, c_colLCMDepth [2]);
		if (g_hpenLCMDepth4 [i]	== NULL) g_hpenLCMDepth4 [i] = ::CreatePen (PS_DOT,   	3 - i, c_colLCMDepth [3]);
		if (g_hpenLCMDepth5 [i]	== NULL) g_hpenLCMDepth5 [i] = ::CreatePen (PS_SOLID,	3 - i, c_colLCMDepth [4]);
	}

	_types.push_back (g_tdUnknown);

	// Prepare POIs.
	size_t cPOI;
	for (cPOI = 0; cPOI < ARRAY_SIZE (g_POIs); ++ cPOI) {
		const POI_desc_t & d = g_POIs [cPOI];

		_types.push_back (type_desc_t ());
		type_desc_t & td = _types.back ();

		td.Type           = (USHORT (d.btPointCode) << 8) + d.btPOISubcodeMin;
		td.Kind           = rkPoint;
		td.ulType2        = td.ulType + (d.btPOISubcodeMax == 0 ? 0 : d.btPOISubcodeMax - d.btPOISubcodeMin);
		td.Category       = static_cast<type_category_t> (d.btCategory);

		td.strTypeName    = d.strTypeName;
		td.phCursor       = d.phCursor;
		td.phBrush        = NULL;
		td.phPen          = NULL;
		td.phPen2         = NULL;

		const label_desc_t * const pLD = d.pLabelDesc ? d.pLabelDesc : & g_DefLabel;
		td.phFont         = pLD->phFont;
		td.colText        = pLD->colText;
		td.pfPrepareLabel = pLD->pfPrepareLabel;
		td.pfDrawLabel    = pLD->pfDrawLabel;
	}

	// Prepare marine POIs.
	for (cPOI = 0; cPOI < ARRAY_SIZE (g_MarinePOIs); ++ cPOI) {
		const POI_desc_t & d = g_MarinePOIs [cPOI];

		_types.push_back (type_desc_t ());
		type_desc_t & td = _types.back ();

		td.Type           = (USHORT (d.btPointCode) << 8) + d.btPOISubcodeMin;
		td.Kind           = rkPoint;
		td.ulType2        = td.ulType;
		td.Category       = static_cast<type_category_t> (d.btCategory);

		td.strTypeName    = d.strTypeName;
		td.phCursor       = d.phCursor;
		td.phBrush        = NULL;
		td.phPen          = NULL;
		td.phPen2         = NULL;

		const label_desc_t * const pLD = d.pLabelDesc ? d.pLabelDesc : & g_DefLabel;
		td.phFont         = pLD->phFont;
		td.colText        = pLD->colText;
		td.pfPrepareLabel = pLD->pfPrepareLabel;
		td.pfDrawLabel    = pLD->pfDrawLabel;
	}

	// Prepare polylines.
	for (size_t cPolyline = 0; cPolyline < ARRAY_SIZE (g_Polylines); ++ cPolyline) {
		const polyline_desc_t & d = g_Polylines [cPolyline];

		_types.push_back (type_desc_t ());
		type_desc_t & td = _types.back ();

		td.Type           = d.btCode;
		td.Kind           = rkPolyline;
		td.ulType2        = td.ulType;
		td.Category       = static_cast<type_category_t> (d.btCategory);

		td.strTypeName    = d.strTypeName;
		td.phCursor       = NULL;
		td.phBrush        = NULL;
		td.phPen          = d.phPen;
		td.phPen2         = d.phPen2;

		const label_desc_t * const pLD = d.pLabelDesc ? d.pLabelDesc : & g_DefLabel;
		td.phFont         = pLD->phFont;
		td.colText        = pLD->colText;
		td.pfPrepareLabel = pLD->pfPrepareLabel;
		td.pfDrawLabel    = pLD->pfDrawLabel;
	}

	// Prepare polygons.
	for (size_t cPolygon = 0; cPolygon < ARRAY_SIZE (g_Polygons); ++ cPolygon) {
		const polygon_desc_t & d = g_Polygons [cPolygon];

		_types.push_back (type_desc_t ());
		type_desc_t & td = _types.back ();

		td.Type           = d.btCode;
		td.Kind           = rkPolygon;
		td.ulType2        = td.ulType;
		td.Category       = static_cast<type_category_t> (d.btCategory);

		td.strTypeName    = d.strTypeName;
		td.phCursor       = NULL;
		td.phBrush        = d.phBrush;
		td.phPen          = d.phPen;
		td.phPen2         = NULL;

		const label_desc_t * const pLD = d.pLabelDesc ? d.pLabelDesc : & g_CenteredLabel;
		td.phFont         = pLD->phFont;
		td.colText        = pLD->colText;
		td.pfPrepareLabel = pLD->pfPrepareLabel;
		td.pfDrawLabel    = pLD->pfDrawLabel;
	}

}

static
void _InitLowranceTypes () {
	//
	// Prepare Lowrance-specific POI icons.
	//

	CTypeSet::types_t types_impl;
	InitLowranceTypes (types_impl);
	g_tsLowrance.SetTypes (types_impl);

	g_tsLowrance.strID            = "Lowrance";
	g_tsLowrance.strDisplayName   = "Lowrance";

	g_tsLowrance.pfGetLatGrid     = GarminLatGrid;
	g_tsLowrance.pfRoundPoint     = GarminRoundPoint;
	g_tsLowrance.pfRoundPoints    = GarminRoundPoints;
	g_tsLowrance.pfRoundRect      = GarminRoundRect;

	g_tsLowrance.pfIsCity         = Lowrance_IsCity;
	g_tsLowrance.pfIsHeight       = Garmin_IsHeight;
	g_tsLowrance.pfIsFeature      = Garmin_IsFeature;
	g_tsLowrance.pfIsRoad         = Lowrance_IsRoad;
	g_tsLowrance.pfIsUnpavedRoad  = RgnPredicateFalse;
	g_tsLowrance.pfIsRoundabout   = RgnPredicateFalse;
	g_tsLowrance.pfIsHwExit       = RgnPredicateFalse;
	g_tsLowrance.pfIsBackground   = RgnPredicateFalse;
	g_tsLowrance.pfIsOneWayRoad   = Lowrance_IsOneWayRoad;
	g_tsLowrance.pfHasAddress     = Lowrance_HasAddress;
	g_tsLowrance.pfHasCityAndZip  = Lowrance_HasCityAndZip;
	g_tsLowrance.pfHasPhone       = RgnPredicateFalse;
	g_tsLowrance.pfHasContact     = RgnPredicateFalse;
	g_tsLowrance.pfHasDescription = RgnPredicateFalse;

	g_tsLowrance.dwFeatures       = CTypeSet::tsfHasCopyright | CTypeSet::tsfHasElevationUnits;

	g_tsLowrance.cMaxBits         = 24;
}

static
void _DestroyLowranceTypes () {
	for (int i = 0; i < 3; ++ i) {
		::DeleteObject (g_hpenLCMDepth1 [i]);
		::DeleteObject (g_hpenLCMDepth2 [i]);
		::DeleteObject (g_hpenLCMDepth3 [i]);
		::DeleteObject (g_hpenLCMDepth4 [i]);
		::DeleteObject (g_hpenLCMDepth5 [i]);
	}
}

CTypeSet g_tsLowrance (_InitLowranceTypes, _DestroyLowranceTypes);

DWORD ConvertTypeSet_Garmin2Lowrance (RGN_KIND _Kind, DWORD _dwType) {
	switch (_Kind) {
		case rkPoint:
			switch (_dwType) {
				case 1:
				case 2:
				case 3:
				case 4:
					return 1;

				case 5:
				case 6:
				case 7:
					return 2;

				case 8:
				case 9:
				case 0x0a:
				case 0x0b:
					return 3;

				case 0x0c:
				case 0x0d:
					return 4;

				case 0x0e:
				case 0x0f:
					return 5;

				case 0x10:
				case 0x11:
					return 6;

				case 0x100:
				case 0x200:
				case 0x300:
				case 0x400:
					return 0x100;

				case 0x500:
				case 0x600:
				case 0x700:
					return 0x200;

				case 0x800:
				case 0x900:
					return 0x300;

				case 0xa00:
				case 0xb00:
					return 0x400;

				case 0xc00:
				case 0xd00:
				case 0xe00:
				case 0xf00:
					return 0x500;

				case 0x1000:
				case 0x1100:
					return 0x600;

				case 0x3002:
					return 0x3801;

				//marine NavAid  POIs
				case 0x160f:
				case 0x170f:
				case 0x180f:
				case 0x190f:
				case 0x1a0f:
				case 0x1b0f:
					return 0x1601; //"Navaid, white"

				case 0x1610:
				case 0x1710:
				case 0x1810:
				case 0x1910:
				case 0x1a10:
				case 0x1b10:
					return 0x1603; //"Navaid, red"

				case 0x1611:
				case 0x1711:
				case 0x1811:
				case 0x1911:
				case 0x1a11:
				case 0x1b11:
					return 0x1605; //"Navaid, green"

				case 0x1614:
				case 0x1714:
				case 0x1814:
				case 0x1914:
				case 0x1a14:
				case 0x1b14:
					return 0x1607; //"Navaid, violet"

				case 0x1604:
				case 0x1704:
				case 0x1804:
				case 0x1904:
				case 0x1a04:
				case 0x1b04:
					return 0x1611; //"Daybeacon, red triangle"

				case 0x1605:
				case 0x1705:
				case 0x1805:
				case 0x1905:
				case 0x1a05:
				case 0x1b05:
					return 0x1613; //"Daybeacon, green square"

				case 0x1606:
				case 0x1706:
				case 0x1806:
				case 0x1906:
				case 0x1a06:
				case 0x1b06:
					return 0x1615; //"Daybeacon, white diamond"

				case 0x1607:
				case 0x1707:
				case 0x1807:
				case 0x1907:
				case 0x1a07:
				case 0x1b07: //"Unlit navaid, white"
				case 0x160a:
				case 0x170a:
				case 0x180a:
				case 0x190a:
				case 0x1a0a:
				case 0x1b0a: //"Unlit navaid, black"
				case 0x160c:
				case 0x170c:
				case 0x180c:
				case 0x190c:
				case 0x1a0c:
				case 0x1b0c: //"Unlit navaid, orange"
					return 0x1617; //"Unlit navaid, orange"				

				//marine MarineDanger  POIs
				case 0x1c00:
					return 0x1621; //"Obstruction"

				case 0x1c01:
					return 0x1629; //"Wreck"

				case 0x1c02:
					return 0x1627; //"Submerged wreck, dangerous"

				case 0x1c03:
				case 0x1c07:
					return 0x162b; //"Submerged wreck, non-dangerous"

				case 0x1c05:
					return 0x1625; //"Obstr., visible at high water"

				case 0x1c0a:
					return 0x1623; //"Rock, submerged at low water"


				case 0x2f01:
					return 0x3601;

				case 0x2f05:
					return 0x2401;

				default:
					return 0;
			}
			
		case rkPolygon:
			switch (_dwType) {
				case 0x01:
				case 0x02:
					return 0x01;

				case 0x14:
				case 0x15:
				case 0x16:
				case 0x1e:
				case 0x1f:
					return 0x20;

				case 0x17:  // city park
					return 0x2f;

				case 0x18: // golf course
					return 0x34;

				case 0x1a: // cemetary
					return 0x39;

				case 0x28:
				case 0x29:
				case 0x32:
					return 0x15;

				case 0x3b:
				case 0x3c: // water
				case 0x3d:
				case 0x3e:
				case 0x3f:
				case 0x40:
				case 0x41:
				case 0x42:
				case 0x43:
				case 0x44:
				case 0x46:
				case 0x47:
				case 0x48:
				case 0x49:
					return 0x10;

				case 0x51:
				    return 0x12;

			}
			break;
			
		case rkPolyline:
			switch (_dwType) {
				case 0x01:
					return 0x01;
				case 0x02:
					return 0x03;
				case 0x03:
					return 0x05;
				case 0x04:
					return 0x10;
				case 0x05:
					return 0x15;
				case 0x06:
					return 0x17;


				case 0x14:
					return 0x70;
				case 0x16:
					return 0x42;
				case 0x18:
					return 0x32;
				case 0x1f:
					return 0x32;

				case 0x20:
					return 0x62;
				case 0x21:
					return 0x64;
				case 0x22:
					return 0x66;

				case 0x23:
					return 0x50;
				case 0x24:
					return 0x52;
				case 0x25:
					return 0x54;
				case 0x26:
					return 0x30;

				case 0x1c:
					return 0x7a;
				case 0x1d:
					return 0x78;
				case 0x1e:
					return 0x7c;
			}
	}
	return _dwType;
}

DECLARE_TYPESET_CONVERTOR (ConvertTypeSet_Garmin2Lowrance, & g_tsGarmin, & g_tsLowrance);

DWORD ConvertTypeSet_Lowrance2Garmin (RGN_KIND _Kind, DWORD _dwType) {
	switch (_Kind) {
		case rkPoint:
			switch (_dwType) {
				case 0xF001:
					return 0x2f08; // Ground transportation.
				case 0x3401:
					return 0x5902;
				case 0x3801:
					return 0x3002;
				case 0x3201:
					return 0x2c05;
				case 0x2401:
					return 0x2f05;

				case 0x1615:
					return 0x1b06;
				case 0x1617:
					return 0x1b07;
				case 0x1601:
					return 0x1b0f;
				case 0x1603:
					return 0x1b10;
				case 0x1605:
					return 0x1b11;
				case 0x1607:
					return 0x1b14;
				case 0x1629:
					return 0x1c01;
				case 0x1627:
					return 0x1c02;
				case 0x162b:
					return 0x1c07;
				case 0x1623:
					return 0x1c0a;
				case 0x3601:
					return 0x2f01;
			}
			break;

		case rkPolyline:
			switch (_dwType) {
				case 0x50:
				case 0x51:
					return 0x23;
				case 0x52:
					return 0x24;
				case 0x53:
					return 0x25;

				case 0x30:
					return 0x26;
				case 0x32:
					return 0x18;

				case 0x3f: // Tram line
				case 0x40: // Monorail
				case 0x41: // Cable car
					return 0x14; // Railroad

				case 0x42: // Earth road
					return 0x16; // Unpaved road
				case 0x43: // Tunnel
					return 0x05; // Collector road

				case 0x45: // District boundary
				case 0x46: // Fence
					return 0x1d; // County/parish boundary
				case 0x70:
					return 0x14;

				case 0x7a:
					return 0x1c;
				case 0x78:
					return 0x1d;
				case 0x7c:
					return 0x1e;

				case 0x01:
					return 0x01;
				case 0x03:
					return 0x02;
				case 0x05:
					return 0x03;
				case 0x10:
					return 0x04;
				case 0x15:
					return 0x05;
				case 0x17:
					return 0x06;

			}
			break;

		case rkPolygon:
			switch (_dwType) {
				case 0x39:
					return 0x1a;

				case 0x01:
					return 0x01;
				case 0x12:
					return 0x51;

				case 0x10: // water
					return 0x3c;
				case 0x15:
					return 0x28;

				case 0x25:
					return 0x15;
				case 0x2f:
					return 0x17;

				case 0x34: // golf course
					return 0x18;
				case 0x45: // cemetary
					return 0x1a;
			}
			break;
	}
	return _dwType;
}

DECLARE_TYPESET_CONVERTOR (ConvertTypeSet_Lowrance2Garmin, & g_tsLowrance, & g_tsGarmin);

//////////////////////////////////////////////////////////

# include "NavitelTypes.h"

DWORD ConvertTypeSet_Navitel2Lowrance (RGN_KIND _Kind, DWORD _dwType) {
	const DWORD dwGarminType = ConvertTypeSet_Navitel2Garmin (_Kind, _dwType);
	return ConvertTypeSet_Garmin2Lowrance (_Kind, dwGarminType);
}

DECLARE_TYPESET_CONVERTOR (ConvertTypeSet_Navitel2Lowrance, & g_tsNavitel, & g_tsLowrance);

# include "RussaTypes.h"

DWORD ConvertTypeSet_Russa2Lowrance (RGN_KIND _Kind, DWORD _dwType) {
	return ConvertTypeSet_Navitel2Lowrance (_Kind, _dwType);
}

DECLARE_TYPESET_CONVERTOR (ConvertTypeSet_Russa2Lowrance, & g_tsRussa, & g_tsLowrance);
