//
// GPSMapEdit
// (c) Konstantin Galichsky (kg@geopainting.com), 2008-2009
//
// Map skins.
//

# include "StdAfx.h"
# include "Resource.h"
# include "MapSkins.h"
# include "Map.h"
# include "Types.h"
# include "TypesPOIDraw.h"
# include "MapView.h"
# include "Log.h"
# include "Menu.h"
# include "Settings.h"
# include "Options.h"
# include "DrawHelpers.h"
# include "ParseHelpers.h"
# include "Str.h"
# include "Globals.h"

CMapSkin::CMapSkin () :
	m_hbmpPOIs (NULL),
	m_hbmpPrev (NULL),
	m_hdcMem (::CreateCompatibleDC (CClientDC (NULL)))
{
	::memset (m_bColorDefined, 0, sizeof (m_bColorDefined));
	::memset (m_colSetting,    0, sizeof (m_colSetting));
	::memset (m_hbrSetting,    0, sizeof (m_hbrSetting));
}

CMapSkin::~CMapSkin () {
	for (size_t c = 0; c < ARRAY_SIZE (m_hbrSetting); ++ c)
		if (m_hbrSetting [c] != NULL)
			::DeleteObject (m_hbrSetting [c]);

	for (pens_t::const_iterator ip = m_pens.begin (); ip != m_pens.end (); ++ ip) {
		const HPEN hpen = * ip;
		if (hpen)
			::DeleteObject (hpen);
	}

	for (brushes_t::const_iterator ib = m_brushes.begin (); ib != m_brushes.end (); ++ ib) {
		const HBRUSH hbr = * ib;
		if (hbr)
			::DeleteObject (hbr);
	}

	for (icons_t::const_iterator ii = m_icons.begin (); ii != m_icons.end (); ++ ii) {
		const HCURSOR hcur = * ii;
		if (hcur)
			::DestroyIcon (hcur);
	}

	if (m_hbmpPrev)
		::SelectObject (m_hdcMem, m_hbmpPrev);
	if (m_hbmpPOIs)
		::DeleteObject (m_hbmpPOIs);
}

void CMapSkin::SetPOIs (HBITMAP _hbmp, size_t _cW, size_t _cH) {
	m_hbmpPOIs = _hbmp;
	m_cPOIsW = _cW;
	m_cPOIsH = _cH;

	m_hbmpPrev = (HBITMAP) ::SelectObject (m_hdcMem, m_hbmpPOIs);
}

bool CMapSkin::DrawPOI (HDC _hDC, size_t _cIconIdx, int _x, int _y, bool _bCenter) const {
	if (_cIconIdx >= m_cPOIsW/m_cPOIsH)
		return false;

	if (_bCenter) {
		_x -= m_cPOIsH/2;
		_y -= m_cPOIsH/2;
	}

	const BLENDFUNCTION bf = {AC_SRC_OVER, 0, 0xFF, AC_SRC_ALPHA};
	return ::AlphaBlend (
		_hDC,     _x,                 _y, m_cPOIsH, m_cPOIsH,
		m_hdcMem, _cIconIdx*m_cPOIsH, 0,  m_cPOIsH, m_cPOIsH,
		bf
	) != FALSE;
}

///////////////////////////////////////////////////////

CMapSkinTypeInfo * CMapSkinTypeInfo::s_pFirst;

CMapSkinTypeInfo::CMapSkinTypeInfo (pfParseSkin_t _pfParseSkin, const char * _strTypeset, const char * _strExt, const char * _strFilter) :
	m_pfParseSkin (_pfParseSkin),
	m_strTypeset (_strTypeset),
	m_strExt (_strExt),
	m_strFilter (_strFilter)
{
	// Insert to the head of the list.
	m_pNext = CMapSkinTypeInfo::s_pFirst;
	CMapSkinTypeInfo::s_pFirst = this;
}

const CMapSkinTypeInfo * CMapSkinTypeInfo::FindByFileExt (const char * _strFileName) {
	const char * const pExt = ::PathFindExtension (_strFileName);

	CMapSkinTypeInfo::pfParseSkin_t pfParseSkin = NULL;
	if (pExt) {
		for (const CMapSkinTypeInfo * p = CMapSkinTypeInfo::s_pFirst; p; p = p->m_pNext) {
			if (::stricmp (p->m_strExt, pExt + 1) == 0)
				return p;
		}
	}
	return NULL;
}

///////////////////////////////////////////////////////

struct mapskin_info_t {
	string_t strTypeset;
	string_t strPath;
	string_t strName;
	string_t strMode;
	bool bLoaded;

	mapskin_info_t () : bLoaded (false) {}
};

typedef std::vector<mapskin_info_t> vecMapSkins_t;
static vecMapSkins_t g_vecMapSkins;

static
void _InitSkinsNames () {
	typedef std::map<string_t, size_t> mapNames2Count_t;
	mapNames2Count_t mapNames2Count;

	const size_t cMapSkins = g_vecMapSkins.size ();
	for (size_t c = 0; c < cMapSkins; ++ c) {
		mapskin_info_t & msi = g_vecMapSkins [c];

		const char * strFileName = ::PathFindFileName (msi.strPath.c_str ());
		char strName [MAX_PATH];
		::sprintf (strName, "%s [%s]", strFileName, msi.strMode.empty () ? "day" : msi.strMode.c_str ());

		const mapNames2Count_t::iterator i = mapNames2Count.find (strName);
		if (i == mapNames2Count.end ()) {
			mapNames2Count [strName] = 1;
		} else {
			size_t & cCount = i->second;
			::sprintf (strName + ::strlen (strName), " (%d)", cCount);
			++ cCount;
		}

		msi.strName = strName;
	}
}

static
void _LoadListOfMapSkins () {
	g_vecMapSkins.clear ();

	const size_t cMapSkins = GetSetting ("MapSkins", (DWORD) 0);
	g_vecMapSkins.reserve (cMapSkins);
	for (size_t c = 0; c < cMapSkins; ++ c) {
		char strKey [32];
		::sprintf (strKey, "MapSkin.%d", c);
		
		char strValue [256 + MAX_PATH];
		GetSetting (strKey, strValue, ARRAY_SIZE (strValue), "");

		mapskin_info_t msi;

		const char * p = strValue;
		const char * const pEnd = strValue + ::strlen (strValue);

		// Mode.
		const char * const pModeEnd = ::lstrchr (p, '*');
		if (pModeEnd == NULL)
			continue;
		msi.strMode.assign (p, pModeEnd);
		p = pModeEnd;

		if (p < pEnd && * p != '*')
			continue;
		++ p;

		// Typeset.
		const char * const pTypesetEnd = ::lstrchr (p, '*');
		if (pTypesetEnd == NULL)
			continue;
		msi.strTypeset.assign (p, pTypesetEnd);
		p = pTypesetEnd;

		if (p < pEnd && * p != '*')
			continue;
		++ p;

		// Path.
		msi.strPath.assign (p, pEnd);

		g_vecMapSkins.push_back (msi);
	}

	_InitSkinsNames ();
}

static
void _SaveListOfMapSkins () {
	const size_t cMapSkins = g_vecMapSkins.size ();
	SaveSetting ("MapSkins", (DWORD) cMapSkins);
	for (size_t c = 0; c < cMapSkins; ++ c) {
		const mapskin_info_t & msi = g_vecMapSkins [c];

		char strKey [32];
		::sprintf (strKey, "MapSkin.%d", c);
		char strValue [256 + MAX_PATH];
		::sprintf (strValue, "%s*%s*%s", msi.strMode.c_str (), msi.strTypeset.c_str (), msi.strPath.c_str ());

		SaveSetting (strKey, strValue);
	}
}

static
void _MarkMapSkinAsLoaded (const char * _strTypeset, size_t _cOrdinal) {
	const size_t cMapSkins = g_vecMapSkins.size ();
	for (size_t c = 0; c < cMapSkins; ++ c) {
		mapskin_info_t & msi = g_vecMapSkins [c];
		if (msi.strTypeset == _strTypeset)
			msi.bLoaded = false;
	}

	if (_cOrdinal >= cMapSkins)
		return;
	mapskin_info_t & msiToMark = g_vecMapSkins [_cOrdinal];
	assert (msiToMark.strTypeset == _strTypeset);
	msiToMark.bLoaded = true;
}

static
void _UnmarkAllMapSkinsAsLoaded () {
	const size_t cMapSkins = g_vecMapSkins.size ();
	for (size_t c = 0; c < cMapSkins; ++ c) {
		mapskin_info_t & msi = g_vecMapSkins [c];
		msi.bLoaded = false;
	}
}

static
size_t _GetLoadedMapSkin (const char * _strTypeset) {
	const size_t cMapSkins = g_vecMapSkins.size ();
	for (size_t c = 0; c < cMapSkins; ++ c) {
		const mapskin_info_t & msi = g_vecMapSkins [c];
		if (msi.bLoaded && msi.strTypeset == _strTypeset)
			return c;
	}

	return -1;
}

static
void _AddMapSkin (const char * _strPath) {
	const CMapSkinTypeInfo * const pMSTI = CMapSkinTypeInfo::FindByFileExt (_strPath);
	if (pMSTI == NULL)
		return;

	const size_t cDaySkinIdx = g_vecMapSkins.size ();

	mapskin_info_t msi;
	msi.strMode    = "day";
	msi.strPath    = _strPath;
	msi.strTypeset = pMSTI->m_strTypeset;
	g_vecMapSkins.push_back (msi);

	msi.strMode    = "night";
	g_vecMapSkins.push_back (msi);

	_InitSkinsNames ();
	_SaveListOfMapSkins ();

	// Is the same type set as current one?
	const char * strTypeset = NULL;
	{
		CCritSec::CLock lock (g_cs);
		if (g_map.pTypeSet && g_map.pTypeSet != & g_tsNull)
			strTypeset = g_map.pTypeSet->strID;
	}
	const bool bTheSameTypeset = strTypeset && ::stricmp (pMSTI->m_strTypeset, strTypeset) == 0;

	// Activate the skin.
	if (bTheSameTypeset) {
		SelectMapSkin (cDaySkinIdx, true);
	} else {
		const size_t cCurrentSkinIdx = GetSelectedMapSkin ();
		SelectMapSkin (cDaySkinIdx,     false);
		SelectMapSkin (cCurrentSkinIdx, false);
	}
}

static
void _RemoveMapSkin (size_t _cOrdinal) {
	const size_t cSel = GetSelectedMapSkin ();
	const bool bRefresh = _cOrdinal == cSel;

	assert (_cOrdinal < g_vecMapSkins.size ());
	if (_cOrdinal < g_vecMapSkins.size ())
		g_vecMapSkins.erase (g_vecMapSkins.begin () + _cOrdinal);

	_SaveListOfMapSkins ();

	// Make list of typesets.
	typedef std::set<string_t> typesets_t;
	typesets_t typesets;
	const size_t cMapSkins = g_vecMapSkins.size ();
	for (size_t c = 0; c < cMapSkins; ++ c) {
		const mapskin_info_t & msi = g_vecMapSkins [c];
		typesets.insert (msi.strTypeset);
	}

	// Update indecies of current map skins for each typeset.
	for (typesets_t::const_iterator i = typesets.begin (); i != typesets.end (); ++ i) {
		const string_t & strTypeset = * i;

		char strKey [32];
		::sprintf (strKey, "MapSkin.%s", strTypeset.c_str ());
		const size_t cSel = GetSetting (strKey, (DWORD) -1);
		if (cSel < _cOrdinal)
			continue;

		if (cSel == _cOrdinal) {
			SaveSetting (strKey, (DWORD) -1);
		} else {
			assert (cSel > _cOrdinal);
			SaveSetting (strKey, (DWORD) (cSel - 1));
		}
	}

	UpdateMapSkin ();

	if (bRefresh)
		Refresh ();
	MapSkins_UpdateGUI ();
}

static
void _ResetMapRefsToTypes () {
	CCritSec::CLock lock (g_cs);

	for (size_t cKind = 0; cKind < rkMax; ++ cKind) {
		rgns_t & rgns = g_map.layers [cKind].rgns;

		for (rgns_t::iterator iRgn = rgns.begin (); iRgn != rgns.end (); ++ iRgn) {
			rgn_t & rgn = * iRgn;
			if (rgn.IsNotMapObject ()) {
				if (rgn.IsWaypoint ())
					rgn.SetType (g_map.pTypeSet, rgn.GetType (), rgn.GetKind ());
				continue;
			}

			rgn.SetType (g_map.pTypeSet, rgn.GetType (), rgn.GetKind ());
		}
	}
}

///////////////////////////////////////////////////////

static CMapSkin * g_pSkin;

static
bool _LoadMapSkin (const char * _strFileName, const char * _strMode) {
	const CMapSkinTypeInfo * const pMSTI = CMapSkinTypeInfo::FindByFileExt (_strFileName);
	if (pMSTI == NULL)
		return false;

	CMapSkin * const pMapSkin = new CMapSkin;

//	ResetWarningsCounters ();
	if (! IsLogEmpty ())
		ReportText ("");
	const size_t cWarnings0 = GetWarningsCount ();
	ReportText ("Parsing skin file '%s', mode '%s'...", _strFileName, _strMode);
	const DWORD dwStartTime = ::GetTickCount ();

	if (! (* pMSTI->m_pfParseSkin) (_strFileName, _strMode, pMapSkin)) {
		ReportWarning ("Failed to read map skin '%s'.", _strFileName);
		ShowMessageLog ();

		if (GetWarningsCount () != cWarnings0)
			ShowMessageLog ();

		delete pMapSkin;

		return false;
	} else {
		ReportText ("Parsing time is %0.2f s", (::GetTickCount () - dwStartTime)*.001f);

		delete g_pSkin;
		g_pSkin = pMapSkin;

		return true;
	}
}

void InitMapSkins () {
	_LoadListOfMapSkins ();
}

void DestroyMapSkins () {
	delete g_pSkin;
	g_pSkin = NULL;
}

void UpdateMapSkin () {
	CCritSec::CLock lock (g_cs);

	const size_t cSkinOrdinal = GetSelectedMapSkin ();
	SelectMapSkin (cSkinOrdinal, false);
}

void SelectMapSkin (size_t _cOrdinal, bool _bRefresh) {
	CCritSec::CLock lock (g_cs);

	const char * strTypeset = NULL;
	if (_cOrdinal == -1) {
		delete g_pSkin;
		g_pSkin = NULL;

		{
			CCritSec::CLock lock (g_cs);
			if (g_map.pTypeSet && g_map.pTypeSet != & g_tsNull)
				strTypeset = g_map.pTypeSet->strID;
		}
/*		if (strTypeset != NULL) {
			if (_GetLoadedMapSkin (strTypeset) == -1)
				return;
		}*/

		// Set 'basic' skin.
		ResetTypes ();
		_UnmarkAllMapSkinsAsLoaded ();
	} else {
		const size_t cMapSkins = g_vecMapSkins.size ();
		if (_cOrdinal >= cMapSkins)
			return;
		const mapskin_info_t & msi = g_vecMapSkins [_cOrdinal];
		if (msi.bLoaded)
			return;

		strTypeset = msi.strTypeset.c_str ();

		const size_t cCurrentOrdinal = _GetLoadedMapSkin (strTypeset);
		if (cCurrentOrdinal == _cOrdinal)
			return;			

		ResetTypes ();
		_UnmarkAllMapSkinsAsLoaded ();

		if (! _LoadMapSkin (msi.strPath.c_str (), msi.strMode.c_str ())) {
			ResetTypes ();
			_UnmarkAllMapSkinsAsLoaded ();
			_ResetMapRefsToTypes ();

			char strKey [32];
			::sprintf (strKey, "MapSkin.%s", strTypeset);
			SaveSetting (strKey, (DWORD) -1);

			if (_bRefresh) {
				Refresh ();
				UpdateMenu ();
			}

			return;
		}
	}

	_MarkMapSkinAsLoaded (strTypeset, _cOrdinal);

	CompleteCustomTypes ();
	_ResetMapRefsToTypes ();

	if (strTypeset != NULL) {
		char strKey [32];
		::sprintf (strKey, "MapSkin.%s", strTypeset);
		SaveSetting (strKey, (DWORD) _cOrdinal);
	}

	if (_bRefresh) {
		Refresh ();
		UpdateMenu ();
	}
}

size_t GetSelectedMapSkin () {
	const char * strTypeset = NULL;
	{
		CCritSec::CLock lock (g_cs);
		if (g_map.pTypeSet)
			strTypeset = g_map.pTypeSet->strID;
	}
	if (strTypeset == NULL && g_map.pTypeSet != & g_tsNull)
		return -1;

	char strKey [32];
	::sprintf (strKey, "MapSkin.%s", strTypeset);
	const size_t cSel = GetSetting (strKey, (DWORD) -1);

	if (cSel >= g_vecMapSkins.size ())
		return -1;
	const mapskin_info_t & msi = g_vecMapSkins [cSel];

	return msi.strTypeset == strTypeset ? cSel : -1;
}

bool LoadMapSkin (const char * _strFileName, const char * _strMode) {
	ResetTypes ();
	_UnmarkAllMapSkinsAsLoaded ();

	const bool bRet = _LoadMapSkin (_strFileName, _strMode);
	if (! bRet) {
		ResetTypes ();
		_UnmarkAllMapSkinsAsLoaded ();
	} else {
		CompleteCustomTypes ();
	}

	_ResetMapRefsToTypes ();

	Refresh ();
	UpdateMenu ();

	return bRet;
}

//////////////////////////////////////////////////

void MapSkin_DrawPOI (HDC _hDC, size_t _cIconIdx, int _x, int _y, bool _bCenter) {
	if (g_pSkin && g_pSkin->DrawPOI (_hDC, _cIconIdx, _x, _y, _bCenter))
		return;

	POINT ptHS;
	GetCursorHotSpot (g_hcurUnknown, ptHS);
	::DrawIcon (_hDC, _x - ptHS.x, _y - ptHS.y, g_hcurUnknown);
}

COLORREF MapSkin_GetColor (CMapColor _mc, CMapSkin * _pMapSkin) {
	if (_pMapSkin == NULL)
		_pMapSkin = g_pSkin;

	if (_pMapSkin && _pMapSkin->IsColorDefined (_mc))
		return _pMapSkin->GetColor (_mc);

	switch (_mc) {
		case spcBgColor:
			return GetBackgroundColor ();

		case spcHWSign1Text:
		case spcHWSign2Text:
		case spcHWSign3Text:
			return c_colBlack;

		case spcHWSign1Bg:
		case spcHWSign2Bg:
		case spcHWSign3Bg:
			return c_colWhite;

		default:
			assert (false);
			return c_colBlack;
	}
}

HBRUSH MapSkin_GetBrush (CMapColor _mc) {
	if (g_pSkin && g_pSkin->IsColorDefined (_mc))
		return g_pSkin->GetBrush (_mc);

	switch (_mc) {
		case spcHWSign1Text:
		case spcHWSign2Text:
		case spcHWSign3Text:
			return g_hbrBlack;

		case spcHWSign1Bg:
		case spcHWSign2Bg:
		case spcHWSign3Bg:
			return g_hbrWhite;

		default:
			assert (false);
			return c_colBlack;
	}
}

//////////////////////////////////////////////////

static
bool _AddMapSkin (HWND _hWnd) {
	char strFileName [MAX_PATH] = "";

	static const char strAllSupportedFiles [] = "All supported files (";
	string_t strAllSupportedExts;

	size_t cFilterLen = 0;
	size_t cFilters = 0;
	const CMapSkinTypeInfo * p;
	for (p = CMapSkinTypeInfo::s_pFirst; p; p = p->m_pNext) {
		cFilterLen += ::strlen (p->m_strFilter) + 1;
		cFilterLen += ::strlen (p->m_strExt)    + 3;

		if (cFilters != 0)
			strAllSupportedExts.append (";");
		strAllSupportedExts.append ("*.");
		strAllSupportedExts.append (p->m_strExt);

		++ cFilters;
	}
	cFilterLen += ARRAY_SIZE (strAllSupportedFiles) - 1 + strAllSupportedExts.size ()*2 + 4;

	char * const strFilter = (char *) ::alloca (sizeof (char)*cFilterLen);
	char * pFilter = strFilter;

	::strcpy (pFilter, strAllSupportedFiles);
	pFilter += ARRAY_SIZE (strAllSupportedFiles) - 1;
	::strcpy (pFilter, strAllSupportedExts.c_str ());
	pFilter += strAllSupportedExts.size ();
	* pFilter = ')';
	++ pFilter;
	* pFilter = '\0';
	++ pFilter;
	::strcpy (pFilter, strAllSupportedExts.c_str ());
	pFilter += strAllSupportedExts.size ();
	* pFilter = '\0';
	++ pFilter;

	for (p = CMapSkinTypeInfo::s_pFirst; p; p = p->m_pNext) {
		::strcpy (pFilter, p->m_strFilter);
		pFilter += ::strlen (p->m_strFilter);
		* pFilter = '\0';
		++ pFilter;

		::strcpy (pFilter, "*.");
		pFilter += 2;
		::strcpy (pFilter, p->m_strExt);
		pFilter += ::strlen (p->m_strExt);
		* pFilter = '\0';
		++ pFilter;
	}
	* pFilter = '\0';
	++ pFilter;
	assert (pFilter == strFilter + cFilterLen);

	OPENFILENAME ofn;
	::memset (& ofn, 0, sizeof (ofn));
	ofn.lStructSize  = OPENFILENAME_SIZE_VERSION_400;
	ofn.lpstrFile    = strFileName;
	ofn.nMaxFile     = MAX_PATH;
	ofn.hInstance    = g_hResInst;
	ofn.hwndOwner    = _hWnd;
	ofn.lpstrFilter  = strFilter;
	ofn.nFilterIndex = 1;
	ofn.Flags        = OFN_EXPLORER | OFN_PATHMUSTEXIST | OFN_FILEMUSTEXIST | OFN_ENABLESIZING | OFN_HIDEREADONLY;
	if (::GetOpenFileName (& ofn)) {
		_AddMapSkin (strFileName);
		return true;
	}
	return false;
}

# include "Dialog.h"
# include "ListView.h"

enum {
	clTypeSet,
	clName,
	clPath,
};

class CEditMapSkins : public CDialog {
  public:
	CEditMapSkins () : CDialog (IDD_MAP_SKINS) {}

	// Override.
	virtual bool OnInit (HWND _hwndFocus) {
		// Map skins.
		m_List.Init (m_hWnd, IDC_SKINS, "ListOfMapSkins");
		m_List.SetExtStyle (LVS_EX_FULLROWSELECT | LVS_EX_HEADERDRAGDROP);
		AddColumn (m_List, clTypeSet, "Type set");
		AddColumn (m_List, clName,    "Name");
		AddColumn (m_List, clPath,    "Path");

		_UpdateList ();
		SetSelection (m_List, 0);

		AutoSizeColumns (m_List);
		m_List.LoadHeaders ();

		_UpdateButtons ();

		return true;
	}
	void OnDestroy () {
		m_List.SaveHeaders ();
		m_List.Destroy ();
	}

	virtual void OnCommand (WORD _wCmd, WORD _wNotifyCode, HWND _hwndCtl) {
		switch (_wCmd) {
			case IDC_ADD:
				if (_AddMapSkin (m_hWnd)) {
					_UpdateList ();
					MapSkins_UpdateGUI ();
				}
				break;

			case IDC_REMOVE: {
				const size_t cSel = ListView_GetSelectionMark (m_List);
				if (cSel == -1)
					break;
				const LPARAM lParamSel = GetLParam (m_List, cSel);

				_RemoveMapSkin (lParamSel);

				_UpdateList ();
				SetSelection (m_List, cSel);
				_UpdateButtons ();

				break;
			}

			case IDOK:
			case IDCANCEL:
				EndDialog (0);
				break;
		}
	}
	virtual bool OnNotify (int _idCtrl, NMHDR * _pInfo) {
		switch (_pInfo->code) {
			case NM_RCLICK:
				m_List.PopupMenu_OnHeaderRClick (_pInfo);
				break;

			case LVN_COLUMNCLICK: {
				// Save selection.
				const size_t cSel = ListView_GetSelectionMark (m_List);
				const LPARAM lParamSel = cSel == -1 ? -1 : GetLParam (m_List, cSel);

				m_List.Sort_OnColumnClick (_pInfo);

				// Restore selection.
				if (lParamSel != -1)
					m_List.SelectByLParam (lParamSel);

				break;
			}

			case LVN_ITEMCHANGED: {
				const NMLISTVIEW * const pInfo2 = reinterpret_cast<const NMLISTVIEW *> (_pInfo);
				if (
					(pInfo2->uOldState & LVIS_SELECTED) == 0 &&
					(pInfo2->uNewState & LVIS_SELECTED) != 0
				) {
					EnableDlgItem (IDC_REMOVE, pInfo2->iItem != -1);
				}

				break;
			}
		}

		return false;
	}

  private:
	CListView m_List;

	void _UpdateList ();
	void _UpdateButtons () {
		EnableDlgItem (IDC_REMOVE, ListView_GetSelectionMark (m_List) != -1);
	}
};

void CEditMapSkins::_UpdateList () {
	ListView_DeleteAllItems (m_List);

	size_t cItem = 0;
	const size_t cMapSkins = g_vecMapSkins.size ();
	for (size_t c = 0; c < cMapSkins; ++ c) {
		const mapskin_info_t & msi = g_vecMapSkins [c];

		InsertItem     (m_List, cItem, cItem,  "%s", msi.strTypeset.c_str ());
		SetSubItemText (m_List, cItem, clName, "%s", msi.strName.c_str ());
		SetSubItemText (m_List, cItem, clPath, "%s", msi.strPath.c_str ());

		++ cItem;
	}
}

bool MapSkins_EditList (HWND _hWnd) {
	CEditMapSkins ems;
	return ems.ShowModal (_hWnd) != 0;
}

////////////////////////////////////////////////

void MapSkins_InitMenu (HMENU _hmenu) {
	::AppendMenu (_hmenu, MF_STRING, ID_MAP_SKINS_0, "Basic");

	const char * strTypeset = NULL;
	{
		CCritSec::CLock lock (g_cs);
		if (g_map.pTypeSet)
			strTypeset = g_map.pTypeSet->strID;
	}

	if (strTypeset == NULL)
		return;

	const size_t cMapSkins = g_vecMapSkins.size ();
	for (size_t c = 0; c < cMapSkins; ++ c) {
		const mapskin_info_t & msi = g_vecMapSkins [c];

		if (c < 100) {
			const bool bEnable = ::strcmp (msi.strTypeset.c_str (), strTypeset) == 0;
			::AppendMenu (_hmenu, MF_STRING | (bEnable ? MF_ENABLED : MF_GRAYED), ID_MAP_SKINS_0 + 1 + c, msi.strName.c_str ());
		}
	}

	MapSkins_UpdateMenu (_hmenu);
}

void MapSkins_UpdateMenu (HMENU _hmenu) {
	const size_t cSelMapSkin = GetSelectedMapSkin ();
	::CheckMenuRadioItem (_hmenu, ID_MAP_SKINS_0, ID_MAP_SKINS_0 + 99, cSelMapSkin == -1 ? ID_MAP_SKINS_0 : ID_MAP_SKINS_0 + 1 + cSelMapSkin, MF_BYCOMMAND);
}

# include "Menu.h"

void MapSkins_UpdateGUI () {
	UpdateMapSkinsMenu ();
}