//
// GPSMapEdit
// (c) Konstantin Galichsky (kg@geopainting.com), 2002-2009
//
// Menu stuff.
//

# include "StdAfx.h"
# include "Resource.h"
# include "Menu.h"
# include "Settings.h"
# include "Toolbar.h"
# include "Map.h"
# include "MapView.h"
# include "MiniMap.h"
# include "Selection.h"
# include "Levels.h"
# include "Tool.h"
# include "ToolCreate.h"
# include "ToolTransform.h"
# include "Paste.h"

# include "GarminTypes.h"
# include "NavitelTypes.h"
# include "RussaTypes.h"

# include "UndoRedo.h"
# include "Log.h"
# include "RecentFiles.h"
# include "Favorites.h"
# include "MapSkins.h"
# include "Load.h"
# include "Positioning.h"
# include "Scale.h"
# include "CreatePOIByAddress.h"
# include "FindByAddress.h"

# include "Keyboard.h"

# include "Str.h"

# include "Globals.h"
# include "Version.h"

CMenuHolder::~CMenuHolder () {
	if (m_hMenu)
		::DestroyMenu (m_hMenu);
}

//////////////////////////////////////////////

bool IsSeparator (HMENU _hMenu, int _nPos) {
	const UINT uiState = ::GetMenuState (_hMenu, _nPos, MF_BYPOSITION);
	if ((uiState & MF_SEPARATOR) == 0)
		return false;
	return (uiState & MF_POPUP) == 0;
}

void DeleteDuplicatedMenuSeparators (HMENU _hMenu, bool _bRecursive, bool _bDeleteEmptySubMenus) {
	bool bFound;
	do {
		bFound = false;

		const int nItems = ::GetMenuItemCount (_hMenu);
		if (nItems == 0)
			return;

		for (int nPos = -1; nPos < nItems; ++ nPos) {
			bool bSeparator0 = nPos     == -1     || IsSeparator (_hMenu, nPos);
			bool bSeparator1 = nPos + 1 >= nItems || IsSeparator (_hMenu, nPos + 1);

			if (bSeparator0 && bSeparator1) {
				::DeleteMenu (_hMenu, nPos == -1 ? 0 : nPos, MF_BYPOSITION);
				bFound = true;
				break;
			}

			if (_bRecursive && nPos >= 0) {
				const HMENU hSubMenu = ::GetSubMenu (_hMenu, nPos);
				if (hSubMenu) {
					DeleteDuplicatedMenuSeparators (hSubMenu, true, _bDeleteEmptySubMenus);

					if (_bDeleteEmptySubMenus) {
						const int nSubItems = ::GetMenuItemCount (hSubMenu);
						if (nSubItems == 0)
							::DeleteMenu (_hMenu, nPos, MF_BYPOSITION);
					}
				}
			}
		}
	} while (bFound);
}

void NormalizeMenuText (char * _str) {
	// Remove \t and the built-in accelerator name.
	char * const pTab = lstrchr (_str, '\t');
	if (pTab)
		* pTab = '\0';
	// Remove all ampersands.
	while (true) {
		char * const pAmpersand = lstrchr (_str, '&');
		if (pAmpersand == NULL)
			break;

		::memmove (pAmpersand, pAmpersand + 1, ::strlen (pAmpersand + 1) + 1);
		_str = * pAmpersand == '&' ? pAmpersand + 1 : pAmpersand;
	}
}

///////////////////////////////////////////////////////////////

static
bool _FindParentMenu (HMENU _hmenuParent, size_t _cPosInParent, HMENU _hmenuThis, UINT _uiCmd, bool _bDelete) {
	const size_t cItems = ::GetMenuItemCount (_hmenuThis);
	for (size_t cPos = 0; cPos < cItems; ++ cPos) {
		const UINT uiCmd = ::GetMenuItemID (_hmenuThis, cPos);
		if (uiCmd == _uiCmd) {
			if (_hmenuParent != NULL) {
				if (_bDelete)
					::DeleteMenu     (_hmenuParent, _cPosInParent, MF_BYPOSITION);
				else
					::EnableMenuItem (_hmenuParent, _cPosInParent, MF_BYPOSITION | MF_GRAYED);
			}
			return true;
		}

		const HMENU hSubMenu = ::GetSubMenu (_hmenuThis, cPos);
		if (hSubMenu != NULL) {
			if (_FindParentMenu (_hmenuThis, cPos, hSubMenu, _uiCmd, _bDelete))
				return true;
		}
	}

	return false;
}

void DeleteParentMenu (HMENU _hMenu, UINT _uiCmd) {
	_FindParentMenu (NULL, 0, _hMenu, _uiCmd, true);
}

void DisableParentMenu (HMENU _hMenu, UINT _uiCmd) {
	_FindParentMenu (NULL, 0, _hMenu, _uiCmd, false);
}

///////////////////////////////////////////////////////////////

void _SetSubMenu (HMENU _hMenu, UINT _uiCmd, HMENU _hPopupMenu) {
	MENUITEMINFO mii;
	::ZeroMemory (& mii, sizeof (mii));
	mii.cbSize   = sizeof (mii);
	mii.fMask    = MIIM_SUBMENU;
	mii.hSubMenu = _hPopupMenu;
	::SetMenuItemInfo (_hMenu, _uiCmd, FALSE, & mii);
}

///////////////////////////////////////////////////////////////

static
void _EnumMenuItems (HMENU _hMenu, const char * _strMenuName, size_t & _cSubMenuIdx, menu_items_t & _items) {
	const size_t cItems = ::GetMenuItemCount (_hMenu);
	if (cItems == -1)
		return;

	for (size_t c = 0; c < cItems; ++ c) {
		if (IsSeparator (_hMenu, c))
			continue;

		// Prepare name of menu item.
		char strName [128];
		::GetMenuString (_hMenu, c, strName, ARRAY_SIZE (strName), MF_BYPOSITION);
		NormalizeMenuText (strName);

		char strMenuName [128];
		if (_strMenuName != NULL) {
			::strcpy (strMenuName, _strMenuName);
			::strcat (strMenuName, " | ");
			::strcat (strMenuName, strName);
		} else
			::strcpy (strMenuName, strName);

		const HMENU hSubMenu = ::GetSubMenu (_hMenu, c);
		if (hSubMenu == g_hmenuWTR || hSubMenu == g_hmenuBookmarks)
			continue;
		if (hSubMenu != NULL) {
			++ _cSubMenuIdx;
			_EnumMenuItems (hSubMenu, strMenuName, _cSubMenuIdx, _items);
			continue;
		}

		const UINT uiCmd = ::GetMenuItemID (_hMenu, c);

		_items.push_back (menu_item_t ());
		menu_item_t & mi = _items.back ();
		mi.uiCmd    = uiCmd;
		mi.strText  = strMenuName;
		mi.cSubMenu = _cSubMenuIdx;
	}
}

menu_items_t g_menu_items;

static
void _PrepareListOfMenuItems () {
	g_menu_items.clear ();

	size_t cSubMenuIdx = 0;
	_EnumMenuItems (g_MainMenu, NULL, cSubMenuIdx, g_menu_items);
}

///////////////////////////////////////////////////////////////

void _HideMenuItems (HMENU _hmenu);

const size_t c_cMaxLen = 50;

enum {
	// Main submenus.
    smFile         = 0,
	smEdit         = 1,
	smView         = 2,
	smFavorites    = 3,
	smTools        = 4,
	smHelp         = 5,

	// 'View' submenus.
	smViewScales   = 3,
	smViewLevels   = 6,
};

HMENU g_hmenuFile;
HMENU g_hmenuView;
HMENU g_hmenuViewLevels;
HMENU g_hmenuViewScales;
HMENU g_hmenuFavorites;
HMENU g_hmenuTools;
HMENU g_hmenuToolsCreateObject;
HMENU g_hToolsTransform;

void InitMenu () {
	assert (g_hWnd);

	const HMENU hMenu = ::GetMenu (g_hWnd);
	g_MainMenu.Init (hMenu);

	g_hmenuFile      = ::GetSubMenu (hMenu, smFile);
	g_hmenuView      = ::GetSubMenu (hMenu, smView);
	g_hmenuFavorites = ::GetSubMenu (hMenu, smFavorites);
	g_hmenuTools     = ::GetSubMenu (hMenu, smTools);

# ifndef _DEBUG
	// NOTE: not ready yet.
	::DeleteMenu (g_hmenuFile, ID_FILE_PAGE_SETUP,  MF_BYCOMMAND);
	::DeleteMenu (g_hmenuFile, ID_FILE_PRINT,       MF_BYCOMMAND);
	DeleteDuplicatedMenuSeparators (g_hmenuFile);

	::DeleteMenu (g_hmenuTools, ID_MERGE_ROADS,     MF_BYCOMMAND);
	::DeleteMenu (g_hmenuTools, ID_GENERATE_RG_END, MF_BYCOMMAND);
	::DeleteMenu (g_hmenuTools, ID_GENERALIZE_ADJACENT, MF_BYCOMMAND);
	::DeleteMenu (g_hmenuTools, ID_REMOVE_SELFINTERSECTIONS, MF_BYCOMMAND);
	DeleteDuplicatedMenuSeparators (g_hmenuTools);
# endif // _DEBUG

	// Init 'Import' and 'Export' menu.
	UpdateImportExportMenu ();

	InitRecentFiles ();

	// Init 'Scales' menu.
	g_hmenuViewScales = ::GetSubMenu (g_hmenuView, smViewScales);
	const size_t cStdScales = GetStdScales ();
	for (size_t cStdScale = 0; cStdScale < cStdScales; ++ cStdScale) {
		char strScaleText [64];
		const bool bText = GetStdScaleName (cStdScale, strScaleText);

		if (bText)
			::AppendMenu (g_hmenuViewScales, MF_STRING, ID_VIEW_SCALE_0 + cStdScale, strScaleText);
		else
			::AppendMenu (g_hmenuViewScales, MF_SEPARATOR, -1, NULL);
	}
	// Delete dummy placeholder.
	::DeleteMenu (g_hmenuViewScales, 0, MF_BYPOSITION);

	// Init 'Levels' menu.
	g_hmenuViewLevels = ::GetSubMenu (g_hmenuView, smViewLevels);
	MakeLevelsMenu ();

	// Init 'Map Skins' menu.
	UpdateMapSkinsMenu ();

	// Init 'Favorites' menu.
	InitFavorites ();

	// Init 'Tools' menu.
	g_hmenuToolsCreateObject = ::CreatePopupMenu ();
	InitToolsCreateObjectMenu (g_hmenuToolsCreateObject, false);
	_SetSubMenu (g_hmenuTools, ID_TOOL_CREATE_RGN, g_hmenuToolsCreateObject);

	g_hToolsTransform = ::CreatePopupMenu ();
	InitToolsTransformMenu (g_hToolsTransform, false);
	_SetSubMenu (g_hmenuTools, ID_TOOL_TRANSFORM, g_hToolsTransform);

	_PrepareListOfMenuItems ();
	_HideMenuItems (g_MainMenu);
	HideSelectedToolbarButtons ();

	g_MainMenu.UpdateODItems ();
}

void InitToolsCreateObjectMenu (HMENU _hMenu, bool _bHideMenuItems) {
	::AppendMenu (_hMenu, MF_STRING, ID_TOOLS_CREATE_0    + tcPoint,     GetStr (IDS_POINT_MENU));
	::AppendMenu (_hMenu, MF_STRING, ID_TOOLS_CREATE_0    + tcPolyline,  GetStr (IDS_POLYLINE_MENU));
	::AppendMenu (_hMenu, MF_STRING, ID_TOOLS_CREATE_0    + tcPolygon,   GetStr (IDS_POLYGON_MENU));
	::AppendMenu (_hMenu, MF_SEPARATOR, -1, NULL);
	::AppendMenu (_hMenu, MF_STRING, ID_TOOLS_CREATE_0    + tcPolylineV, GetStr (IDS_POLYLINE_V_MENU));
	::AppendMenu (_hMenu, MF_STRING, ID_TOOLS_CREATE_0    + tcPolygonV,  GetStr (IDS_POLYGON_V_MENU));
	::AppendMenu (_hMenu, MF_SEPARATOR, -1, NULL);
	::AppendMenu (_hMenu, MF_STRING, ID_TOOLS_CREATE_0    + tcStrip,     GetStr (IDS_STRIP_MENU));
	::AppendMenu (_hMenu, MF_STRING, ID_TOOLS_CREATE_0    + tcRect,      GetStr (IDS_RECT_MENU));
	::AppendMenu (_hMenu, MF_STRING, ID_TOOLS_CREATE_0    + tcDisk,      GetStr (IDS_DISK_MENU));
	::AppendMenu (_hMenu, MF_SEPARATOR, -1, NULL);
	::AppendMenu (_hMenu, MF_STRING, ID_TOOLS_CREATE_0    + tcCircle,    GetStr (IDS_CIRCLE_MENU));
	::AppendMenu (_hMenu, MF_SEPARATOR, -1, NULL);
	::AppendMenu (_hMenu, MF_STRING, ID_TOOLS_CREATE_0    + tcBookmark,  GetStr (IDS_BOOKMARK_MENU));
# if defined (GME_AS) || defined (_DEBUG)
	::AppendMenu (_hMenu, MF_STRING, ID_TOOLS_CREATE_0    + tcNote,      GetStr (IDS_NOTE_MENU));
# endif // GME_AS

	if (_bHideMenuItems)
		_HideMenuItems (_hMenu);
}

void InitToolsTransformMenu (HMENU _hMenu, bool _bHideMenuItems) {
	::AppendMenu (_hMenu, MF_STRING, ID_TOOLS_TRANSFORM_0 + 0,           GetStr (IDS_SHIFT_MENU));
	::AppendMenu (_hMenu, MF_STRING, ID_TOOLS_TRANSFORM_0 + 1,           GetStr (IDS_AFFINE_MENU));
	::AppendMenu (_hMenu, MF_STRING, ID_TOOLS_TRANSFORM_0 + 2,           GetStr (IDS_QUADRATIC_MENU));
	::AppendMenu (_hMenu, MF_STRING, ID_TOOLS_TRANSFORM_0 + 3,           GetStr (IDS_MTA_MENU));

	if (_bHideMenuItems)
		_HideMenuItems (_hMenu);
}

void CheckToolsCreateObjectMenu (HMENU _hMenu) {
	::CheckMenuRadioItem (_hMenu, ID_TOOLS_CREATE_0,    ID_TOOLS_CREATE_0 + tcMax, ID_TOOLS_CREATE_0 + GetCreateKind (), MF_BYCOMMAND);
}

void CheckToolsTransformMenu (HMENU _hMenu) {
	::CheckMenuRadioItem (_hMenu, ID_TOOLS_TRANSFORM_0, ID_TOOLS_TRANSFORM_0 + 3,  ID_TOOLS_TRANSFORM_0 + GetTransformationOrder (), MF_BYCOMMAND);
}

HMENU g_hmenuWTR;
HMENU g_hmenuBookmarks;

void UpdateImportExportMenu () {
	const HMENU hMenu = ::GetMenu (g_hWnd);

	// Init 'Import' menu.
	const HMENU hmenuImport = ::CreatePopupMenu ();
	CFileTypeInfo::InitImportMenu (hmenuImport);
	_SetSubMenu (hMenu, ID_FILE_IMPORT, hmenuImport);

	UpdateMenuAccelerators (hmenuImport);

	// Init 'Export' menu.
	const HMENU hmenuExport = ::CreatePopupMenu ();
	g_hmenuWTR = CFileTypeInfo::InitExportMenu (hmenuExport);
	_SetSubMenu (hMenu, ID_FILE_EXPORT, hmenuExport);

	UpdateMenuAccelerators (hmenuExport);
}

void UpdateMapSkinsMenu () {
	const HMENU hMenu = ::GetMenu (g_hWnd);

	// Init 'Map Skins' menu.
	const HMENU hmenuMapSkins = ::CreatePopupMenu ();
	MapSkins_InitMenu (hmenuMapSkins);
	_SetSubMenu (hMenu, ID_MAP_SKINS, hmenuMapSkins);

	UpdateMenuAccelerators (hmenuMapSkins);
}

void UpdateMenu () {
	const HMENU hMenu = ::GetMenu (g_hWnd);
	const HMENU hViewMenu = ::GetSubMenu (hMenu, smView);

	bool bEmptyMap, bNoMap;
	size_t cLevels;
	{
		CCritSec::CLock lock (g_cs);
		bEmptyMap = g_map.rectBound.Invalid ();
		bNoMap = g_map.empty ();
		cLevels = g_map.Levels.size ();
	}

	//
	// 'File' menu.
	//
	::EnableMenuItem (hMenu, ID_FILE_CLOSE,       MF_BYCOMMAND | (bNoMap           || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_FILE_OPEN,        MF_BYCOMMAND | (                    g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_FILE_ADD,         MF_BYCOMMAND | (                    g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_FILE_SAVE,        MF_BYCOMMAND | (! CanSave ()     || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_FILE_SAVE_AS,     MF_BYCOMMAND | (bNoMap           || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_FILE_IMPORT,      MF_BYCOMMAND | (                    g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_FILE_EXPORT,      MF_BYCOMMAND | (bNoMap           || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_FILE_PAGE_SETUP,  MF_BYCOMMAND | (bNoMap           || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_FILE_PRINT,       MF_BYCOMMAND | (bNoMap           || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_MAP_PROPERTIES,   MF_BYCOMMAND | (bEmptyMap        || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_ERROR_LOG,        MF_BYCOMMAND | (IsLogEmpty ()                  ? MF_GRAYED : MF_ENABLED));
	// Recent files.
	for (size_t cID = ID_FILE_RECENT_0; cID <= ID_FILE_RECENT_0 + 9; ++ cID) {
		const bool bIsFileAccessible = ::IsFileAccessible (cID - ID_FILE_RECENT_0);
		::EnableMenuItem (hMenu, cID,             MF_BYCOMMAND | (! bIsFileAccessible || g_bLoading ? MF_GRAYED : MF_ENABLED));
	}

	MENUITEMINFO mii = {sizeof (mii), MIIM_SUBMENU};
	::GetMenuItemInfo (hMenu, ID_FILE_EXPORT, FALSE, & mii);
	// Export | WTR to.
	::EnableMenuItem (mii.hSubMenu, 0, MF_BYPOSITION | (! g_bHaveWaypointsOrTracks || g_bLoading ? MF_GRAYED : MF_ENABLED));
	// Export | Bookmarks to.
	::EnableMenuItem (mii.hSubMenu, 1, MF_BYPOSITION | (! g_bHaveBookmarks         || g_bLoading ? MF_GRAYED : MF_ENABLED));

	// 'File | Export' menu.
	CFileTypeInfo::UpdateExportMenu (hMenu);

	//
	// 'Edit' menu.
	//

	// Undo.
	const bool bCanUndo = CanUndo () && ! g_bLoading;
	string_t strText = GetStr (IDS_UNDO_MENU);
	if (bCanUndo) {
		strText.append (" ");
		g_Undo.back ()->GetName (strText);

		char strTextCompact [c_cMaxLen + 32];
		::PathCompactPathEx (strTextCompact, strText.c_str (), c_cMaxLen, 0);
		strText = strTextCompact;
	}
	// Restore accelerator.
	char strUndo [128];
	::GetMenuString (hMenu, ID_EDIT_UNDO, strUndo, ARRAY_SIZE (strUndo), MF_BYCOMMAND);
	const char * const pUndoAcc = lstrchr (strUndo, '\t');
	if (pUndoAcc)
		strText.append (pUndoAcc);
	::ModifyMenu (hMenu, ID_EDIT_UNDO, MF_BYCOMMAND | MF_STRING | (bCanUndo ? MF_ENABLED : MF_GRAYED), ID_EDIT_UNDO, strText.c_str ());

	// Redo.
	const bool bCanRedo = CanRedo () && ! g_bLoading;
	strText = GetStr (IDS_REDO_MENU);
	if (bCanRedo) {
		strText.append (" ");
		g_Redo.back ()->GetName (strText);

		char strTextCompact [c_cMaxLen + 32];
		::PathCompactPathEx (strTextCompact, strText.c_str (), c_cMaxLen, 0);
		strText = strTextCompact;
	}
	// Restore accelerator.
	char strRedo [128];
	::GetMenuString (hMenu, ID_EDIT_REDO, strRedo, ARRAY_SIZE (strRedo), MF_BYCOMMAND);
	const char * const pRedoAcc = lstrchr (strRedo, '\t');
	if (pRedoAcc)
		strText.append (pRedoAcc);
	::ModifyMenu (hMenu, ID_EDIT_REDO, MF_BYCOMMAND | MF_STRING | (bCanRedo ? MF_ENABLED : MF_GRAYED), ID_EDIT_REDO, strText.c_str ());

	bool bCanCopy = false;
	if (! g_Selection.empty ()) {
		const size_t cSelItems = g_Selection.size ();
		for (size_t cSelItem = 0; cSelItem < cSelItems; ++ cSelItem) {
			const find_info_t & SelItem = g_Selection [cSelItem];
			if (SelItem.pLayer && ! SelItem.iRgn->IsDeleted ()) {
				if (! SelItem.iRgn->IsWTR ()) {
					bCanCopy = true;
					break;
				}
			}
		}
	}

	::EnableMenuItem (hMenu, ID_EDIT_CUT,               MF_BYCOMMAND | (! bCanCopy           || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_EDIT_COPY,              MF_BYCOMMAND | (! bCanCopy           || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_EDIT_PASTE,             MF_BYCOMMAND | (IsClipboardEmpty ()  || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_EDIT_DELETE,            MF_BYCOMMAND | (g_Selection.empty () || g_bLoading ? MF_GRAYED : MF_ENABLED));

	::EnableMenuItem (hMenu, ID_SELECT_ALL_OBJECTS,     MF_BYCOMMAND | (bEmptyMap            || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_SELECT_ALL_BOOKMARKS,   MF_BYCOMMAND | (bEmptyMap            || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_SELECT_ALL_NOTES,       MF_BYCOMMAND | (bEmptyMap            || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_SELECT_ALL_TRACKS,      MF_BYCOMMAND | (bEmptyMap            || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_SELECT_ALL_WPTS,        MF_BYCOMMAND | (bEmptyMap            || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_SELECT_ALL_ROUTES,      MF_BYCOMMAND | (bEmptyMap            || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_SELECT_ALL_RASTERS,     MF_BYCOMMAND | (bEmptyMap            || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_SELECT_ALL_ATTACHMENTS, MF_BYCOMMAND | (bEmptyMap            || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_SELECT_ALL_POINTS,      MF_BYCOMMAND | (bEmptyMap            || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_SELECT_ALL_POLYLINES,   MF_BYCOMMAND | (bEmptyMap            || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_SELECT_ALL_POLYGONS,    MF_BYCOMMAND | (bEmptyMap            || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_SELECT_ALL_ROADS,       MF_BYCOMMAND | (bEmptyMap            || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_CUSTOM_FILTER,          MF_BYCOMMAND | (bEmptyMap            || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_UNSELECT,               MF_BYCOMMAND | (g_Selection.empty () || g_bLoading ? MF_GRAYED : MF_ENABLED));

	::EnableMenuItem (hMenu, ID_FIND_NODEID,            MF_BYCOMMAND | (bEmptyMap            || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_FIND_ROADID,            MF_BYCOMMAND | (bEmptyMap            || g_bLoading ? MF_GRAYED : MF_ENABLED));

	bool bInvertSelectionDisabled = g_Selection.empty () || g_bLoading || IsInvertSelectionDisabled ();
	::EnableMenuItem (hMenu, ID_INVERT_SELECTION,       MF_BYCOMMAND | (bInvertSelectionDisabled ? MF_GRAYED : MF_ENABLED));

	//
	// 'View' menu.
	//
	::EnableMenuItem (hMenu, ID_ZOOM_IN,                MF_BYCOMMAND | (bEmptyMap            || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_ZOOM_OUT,               MF_BYCOMMAND | (bEmptyMap            || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_ZOOM_FULL,              MF_BYCOMMAND | (bEmptyMap            || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_ZOOM_TO_SELECTION,      MF_BYCOMMAND | (g_Selection.empty () || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_SHOW_MINI,              MF_BYCOMMAND | (bEmptyMap            || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_SHOW_SS,                MF_BYCOMMAND | (bEmptyMap            || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_SHOW_RULER,             MF_BYCOMMAND | (bEmptyMap            || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_SHOW_GRID,              MF_BYCOMMAND | (bEmptyMap            || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_SHOW_LABELS,            MF_BYCOMMAND | (bEmptyMap            || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_SHOW_POLYGON_CONTOURS,  MF_BYCOMMAND | (bEmptyMap            || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_TRANSPARENT_POLYGONS,   MF_BYCOMMAND | (bEmptyMap            || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_SHOW_MAP,               MF_BYCOMMAND | (bEmptyMap            || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_SHOW_COVERAGE,          MF_BYCOMMAND | (bEmptyMap            || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_SHOW_BOOKMARKS,         MF_BYCOMMAND | (bEmptyMap            || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_SHOW_ATTACHMENTS,       MF_BYCOMMAND | (bEmptyMap            || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_SHOW_RGNODES,           MF_BYCOMMAND | (bEmptyMap            || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_SHOW_ROAD_CLASSES,      MF_BYCOMMAND | (bEmptyMap            || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_HIGHLIGHT_ADDRESSES,    MF_BYCOMMAND | (bEmptyMap            || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_SHOW_NUMBERS,           MF_BYCOMMAND | (bEmptyMap            || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_SHOW_HLEVEL,            MF_BYCOMMAND | (bEmptyMap            || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_SHOW_GOOGLEMAPS,        MF_BYCOMMAND | (bEmptyMap            || g_bLoading ? MF_GRAYED : MF_ENABLED));

	::EnableMenuItem (hMenu, ID_GPS_TRACKING,           MF_BYCOMMAND | (false                              ? MF_GRAYED : MF_ENABLED));

	::EnableMenuItem (hMenu, ID_BOOKMARKS,              MF_BYCOMMAND | (bEmptyMap            || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_NOTE_DRAWINGS,          MF_BYCOMMAND | (bEmptyMap            || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_WTR,                    MF_BYCOMMAND | (bEmptyMap            || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_ATTACHMENTS,            MF_BYCOMMAND | (bEmptyMap            || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_TURN_RESTRICTIONS,      MF_BYCOMMAND | (bEmptyMap            || g_bLoading ? MF_GRAYED : MF_ENABLED));

	::EnableMenuItem (hMenu, ID_REFRESH,                MF_BYCOMMAND | (bEmptyMap            || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_GOTO_POSITION,          MF_BYCOMMAND | (bEmptyMap            || g_bLoading ? MF_GRAYED : MF_ENABLED));

	// 'Scale' sub-menu.
	::EnableMenuItem (hViewMenu, smViewScales,          MF_BYPOSITION | (bEmptyMap           || g_bLoading ? MF_GRAYED : MF_ENABLED));
	if (g_hScaleCB)
		::EnableWindow (g_hScaleCB, bEmptyMap || g_bLoading ? FALSE : TRUE);

	// 'Levels' sub-menu.
	::EnableMenuItem (hViewMenu, smViewLevels,          MF_BYPOSITION | (bEmptyMap           || g_bLoading ? MF_GRAYED : MF_ENABLED));
	UpdateLevelsMenu ();

	// 'Map Skins' sub-menu.
	MapSkins_UpdateMenu (hViewMenu);

	::CheckMenuItem (hMenu, ID_SHOW_MINI,               MF_BYCOMMAND | (::IsWindowVisible (g_hMiniMapWnd)       ? MF_CHECKED : MF_UNCHECKED));
	::CheckMenuItem (hMenu, ID_SHOW_SS,                 MF_BYCOMMAND | (::IsWindowVisible (g_hScaleSelectorWnd) ? MF_CHECKED : MF_UNCHECKED));
	::CheckMenuItem (hMenu, ID_SHOW_RULER,              MF_BYCOMMAND | (g_bShowRuler              ? MF_CHECKED : MF_UNCHECKED));
	::CheckMenuItem (hMenu, ID_SHOW_GRID,               MF_BYCOMMAND | (g_bShowGridLines          ? MF_CHECKED : MF_UNCHECKED));
	::CheckMenuItem (hMenu, ID_SHOW_LABELS,             MF_BYCOMMAND | (g_bShowLabels             ? MF_CHECKED : MF_UNCHECKED));
	::CheckMenuItem (hMenu, ID_SHOW_POLYGON_CONTOURS,   MF_BYCOMMAND | (g_bShowPolygonContours    ? MF_CHECKED : MF_UNCHECKED));
	::CheckMenuItem (hMenu, ID_TRANSPARENT_POLYGONS,    MF_BYCOMMAND | (g_bTransparentPolygons    ? MF_CHECKED : MF_UNCHECKED));
	::CheckMenuItem (hMenu, ID_SHOW_MAP,                MF_BYCOMMAND | (g_bShowMap                ? MF_CHECKED : MF_UNCHECKED));
	::CheckMenuItem (hMenu, ID_SHOW_COVERAGE,           MF_BYCOMMAND | (g_bShowCoverageArea       ? MF_CHECKED : MF_UNCHECKED));
	::CheckMenuItem (hMenu, ID_SHOW_BOOKMARKS,          MF_BYCOMMAND | (g_bShowBookmarksAndNotes  ? MF_CHECKED : MF_UNCHECKED));
	::CheckMenuItem (hMenu, ID_SHOW_ATTACHMENTS,        MF_BYCOMMAND | (g_bShowAttachments        ? MF_CHECKED : MF_UNCHECKED));
	::CheckMenuItem (hMenu, ID_SHOW_RGNODES,            MF_BYCOMMAND | (g_bShowRGNodes            ? MF_CHECKED : MF_UNCHECKED));
	::CheckMenuItem (hMenu, ID_SHOW_ROAD_CLASSES,       MF_BYCOMMAND | (g_bShowRoadClasses        ? MF_CHECKED : MF_UNCHECKED));
	::CheckMenuItem (hMenu, ID_HIGHLIGHT_ADDRESSES,     MF_BYCOMMAND | (g_bHighlightAddresses     ? MF_CHECKED : MF_UNCHECKED));
	::CheckMenuItem (hMenu, ID_SHOW_NUMBERS,            MF_BYCOMMAND | (g_bShowNumbers            ? MF_CHECKED : MF_UNCHECKED));
	::CheckMenuItem (hMenu, ID_SHOW_HLEVEL,             MF_BYCOMMAND | (g_bShowHLevel             ? MF_CHECKED : MF_UNCHECKED));
	::CheckMenuItem (hMenu, ID_SHOW_GOOGLEMAPS,         MF_BYCOMMAND | (g_bShowGoogleMaps         ? MF_CHECKED : MF_UNCHECKED));
	::CheckMenuItem (hMenu, ID_GPS_TRACKING,            MF_BYCOMMAND | (IsGPSUnitActive ()        ? MF_CHECKED : MF_UNCHECKED));

	//
	// 'Favorites' menu.
	//
	::EnableMenuItem (hMenu, ID_ADD_TO_FAVORITES,       MF_BYCOMMAND | (bEmptyMap   || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_REMOVE_FAVORITE,        MF_BYCOMMAND | (bEmptyMap   || g_bLoading ? MF_GRAYED : MF_ENABLED));
	UpdateFavoritesMenu_Enable ();

	//
	// 'Tools' menu.
	//
	static UINT c_uiToolCmdId [] = {
		ID_TOOL_DRAG,
		ID_TOOL_ZOOM,
		ID_TOOL_SELECT,
		ID_TOOL_ROTATE,
		ID_TOOL_EDIT_NODES,
		ID_TOOL_CREATE_RGN,
		ID_TOOL_TRIM,
		ID_TOOL_TRANSFORM,
		ID_TOOL_ROUTING,
		ID_TOOL_DISTANCE
	};
	for (size_t cTool = 0; cTool < ARRAY_SIZE (c_uiToolCmdId); ++ cTool) {
		::EnableMenuItem (hMenu, c_uiToolCmdId [cTool], MF_BYCOMMAND | (g_bLoading ? MF_GRAYED : MF_ENABLED));

		MENUITEMINFO mii;
		::memset (& mii, 0, sizeof (mii));
		mii.cbSize  = sizeof (mii);
		mii.fMask   = MIIM_FTYPE | MIIM_STATE;
		mii.fType   = MFT_RADIOCHECK;
		mii.fState &= ~MFS_CHECKED;
		::SetMenuItemInfo (hMenu, c_uiToolCmdId [cTool], FALSE, & mii);
	}

	UINT uiToolCmd = 0;
	switch (CTool::GetCurrentToolId ()) {
		default: assert (false);
		case toolDummy:
		case toolDrag:      uiToolCmd = ID_TOOL_DRAG;       break;
		case toolZoom:      uiToolCmd = ID_TOOL_ZOOM;       break;
		case toolSelect:    uiToolCmd = ID_TOOL_SELECT;     break;
		case toolRotate:    uiToolCmd = ID_TOOL_ROTATE;     break;
		case toolEditNodes: uiToolCmd = ID_TOOL_EDIT_NODES; break;
		case toolCreate:    uiToolCmd = ID_TOOL_CREATE_RGN; break;
		case toolTrim:      uiToolCmd = ID_TOOL_TRIM;       break;
		case toolTransform: uiToolCmd = ID_TOOL_TRANSFORM;  break;
		case toolRouting:   uiToolCmd = ID_TOOL_ROUTING;    break;
		case toolDistance:  uiToolCmd = ID_TOOL_DISTANCE;   break;
	}
	::CheckMenuRadioItem (hMenu, uiToolCmd, uiToolCmd, uiToolCmd, MF_BYCOMMAND);
	if (uiToolCmd == ID_TOOL_CREATE_RGN)
		CheckToolsCreateObjectMenu (hMenu);
	else if (uiToolCmd == ID_TOOL_TRANSFORM)
		CheckToolsTransformMenu (hMenu);

	// Repeat Last Context Menu Command.
	const bool bCanRepeat = CTool::IsLastCtxMenuCmdDefined ();
	strText = GetStr (IDS_REPEAT_1_MENU);
	if (bCanRepeat) {
		const char * strCmd = CTool::GetLastCtxMenuCmd ();
		if (strCmd != NULL && * strCmd != '\0') {
			strText = GetStr (IDS_REPEAT_2_MENU);
			strText.append (" [");
			strText.append (strCmd);
			strText.append ("]");

			char strTextCompact [c_cMaxLen + 32];
			::PathCompactPathEx (strTextCompact, strText.c_str (), c_cMaxLen, 0);
			strText = strTextCompact;
		}
	}
	// Restore accelerator.
	char strRepeat [128];
	::GetMenuString (hMenu, ID_REPEAT_LAST_CTX_COMMAND, strRepeat, ARRAY_SIZE (strRepeat), MF_BYCOMMAND);
	const char * const pRepeatAcc = lstrchr (strRepeat, '\t');
	if (pRepeatAcc)
		strText.append (pRepeatAcc);
	::ModifyMenu (hMenu, ID_REPEAT_LAST_CTX_COMMAND, MF_BYCOMMAND | MF_STRING | (! bEmptyMap && bCanRepeat ? MF_ENABLED : MF_GRAYED), ID_REPEAT_LAST_CTX_COMMAND, strText.c_str ());

	::EnableMenuItem (hMenu, ID_GENERALIZE_POLY,       MF_BYCOMMAND | (bEmptyMap || cLevels == 0 || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_GENERALIZE_ADJACENT,   MF_BYCOMMAND | (bEmptyMap || cLevels == 0 || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_JOIN_RGN_LEVELS,       MF_BYCOMMAND | (bEmptyMap || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_SPLIT_RGN_LEVELS,      MF_BYCOMMAND | (bEmptyMap || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_REMOVE_DUPLICATES,     MF_BYCOMMAND | (bEmptyMap || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_CONVERT_LABELS,        MF_BYCOMMAND | (bEmptyMap || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_REMOVE_HOLES,          MF_BYCOMMAND | (bEmptyMap || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_REMOVE_SELFINTERSECTIONS, MF_BYCOMMAND | (bEmptyMap || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_REMOVE_JITTERS,        MF_BYCOMMAND | (bEmptyMap || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_SLICE_MAP,             MF_BYCOMMAND | (bEmptyMap || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_SPLIT_MAP,             MF_BYCOMMAND | (bEmptyMap || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_VERIFY_MAP,            MF_BYCOMMAND | (bEmptyMap || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_GENERATE_RG,           MF_BYCOMMAND | (bEmptyMap || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_GENERATE_RG_END,       MF_BYCOMMAND | (bEmptyMap || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_REMOVE_RG,             MF_BYCOMMAND | (bEmptyMap || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_CLEAN_UP_EXT_NODES,    MF_BYCOMMAND | (bEmptyMap || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_CLEAR_EXT_NODES,       MF_BYCOMMAND | (bEmptyMap || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_FIX_ROUTE_PARAMS,      MF_BYCOMMAND | (bEmptyMap || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_MERGE_ROADS,           MF_BYCOMMAND | (bEmptyMap || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_EDIT_PAI,              MF_BYCOMMAND | (bEmptyMap || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_CREATE_POI_BY_ADDRESS, MF_BYCOMMAND | (bEmptyMap || g_bLoading || ! CreatePOIByAddressIsAvailable () ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_FIND_BY_ADDRESS,       MF_BYCOMMAND | (bEmptyMap || g_bLoading ? MF_GRAYED : MF_ENABLED));
	::EnableMenuItem (hMenu, ID_OPTIONS,               MF_BYCOMMAND | (false                   ? MF_GRAYED : MF_ENABLED));

	// Update the menu bar.
	g_MainMenu.UpdateODItems ();
	::DrawMenuBar (g_hWnd);

	// Update toolbar.
	const size_t cBtns = ::SendMessage (g_hToolbarWnd, TB_BUTTONCOUNT, 0, 0);
	for (size_t cBtn = 0; cBtn < cBtns; ++ cBtn) {
		TBBUTTON tbb;
		::SendMessage (g_hToolbarWnd, TB_GETBUTTON, cBtn, reinterpret_cast<LPARAM> (& tbb));

		const int nStyle = tbb.fsStyle & 0xF;
		if (nStyle == TBSTYLE_BUTTON || nStyle == TBSTYLE_GROUP)
			::SendMessage (g_hToolbarWnd, TB_PRESSBUTTON,  tbb.idCommand, MAKELONG (::GetMenuState (hMenu, tbb.idCommand, MF_BYCOMMAND) & MF_CHECKED ? TRUE : FALSE, 0));
		if (nStyle == TBSTYLE_CHECK)
			::SendMessage (g_hToolbarWnd, TB_CHECKBUTTON,  tbb.idCommand, MAKELONG (::GetMenuState (hMenu, tbb.idCommand, MF_BYCOMMAND) & MF_CHECKED ? TRUE : FALSE, 0));
		if (nStyle != TBSTYLE_SEP)
			::SendMessage (g_hToolbarWnd, TB_ENABLEBUTTON, tbb.idCommand, MAKELONG (::GetMenuState (hMenu, tbb.idCommand, MF_BYCOMMAND) & MF_GRAYED  ? FALSE : TRUE, 0));
	}
}

void PopupMenu (HMENU _hMenu) {
	POINT pt;
	::GetCursorPos (& pt);
	::TrackPopupMenu (
		_hMenu,
		TPM_LEFTALIGN | TPM_TOPALIGN | TPM_LEFTBUTTON,
		pt.x, pt.y,
		0, g_hWnd, NULL
	);
}

////////////////////////////////////////////////////////////////////////////

static
int _GetMenuItemCount (HMENU _hMenu) {
	size_t cCount = 0;
	for (size_t c = 0; true; ++ c) {
		MENUITEMINFO mii;
		::memset (& mii, 0, sizeof (mii));
		mii.cbSize = sizeof (mii);
		mii.fMask  = MIIM_FTYPE;
		if (! ::GetMenuItemInfo (_hMenu, c, TRUE, & mii))
			break;

		++ cCount;
	}

	return cCount;
}

CMenuBase::CMenuBase () :
	m_bToDestroy (false),
	m_bHorzMenu (true),
	m_hMenu (NULL),
	m_hwndToolbar (NULL),
	m_pMenuItems (NULL),
	m_cMenuItems (0),
	m_hilIcons (NULL)
{}

CMenuBase::~CMenuBase () {
	_DestroyMenuStrings ();

	if (m_hilIcons) {
		::ImageList_Destroy (m_hilIcons);
		m_hilIcons = NULL;
	}

	if (m_bToDestroy)
		::DestroyMenu (m_hMenu);
}

CPopupMenu::CPopupMenu () {
	assert (m_hMenu == NULL);
	m_hMenu = ::CreatePopupMenu ();
	m_bToDestroy = true;
	m_bHorzMenu = false;
}

CPopupMenu::CPopupMenu (HMENU _hMenu) {
	assert (m_hMenu == NULL);
	m_hMenu = _hMenu;
	m_bToDestroy = false;
	m_bHorzMenu = false;
}

int CMenuBase::ShowAsPopup (HWND _hwndParent, RECT & _rectExclude, bool _bRetCmd) const {
	TPMPARAMS tpm;
	tpm.cbSize    = sizeof (tpm);
	tpm.rcExclude = _rectExclude;
	return (int) ::TrackPopupMenuEx (
		m_hMenu,
		TPM_RIGHTALIGN | TPM_LEFTBUTTON | TPM_VERTICAL | (_bRetCmd ? TPM_RETURNCMD : 0),
		_rectExclude.right, _rectExclude.bottom,
		_hwndParent,
		& tpm
	);
}

void CMenuBase::_DestroyMenuStrings () {
	const size_t cMenuStrings = m_MenuStrings.size ();
	for (size_t c = 0; c < cMenuStrings; ++ c)
		delete [] m_MenuStrings [c];
	m_MenuStrings.clear ();
}

void CMenuBase::AppendItem (WORD _wCmdId, const char * _strItem) {
	::AppendMenu (m_hMenu, MF_STRING, _wCmdId, _strItem);
}

void CMenuBase::EnableItem (WORD _wCmdId, bool _bEnable) {
//	if (m_hwndToolbar)
//		::SendMessage (m_hwndToolbar, TB_SETSTATE, _wCmdId, _bEnable ? TBSTATE_ENABLED : 0);

	assert (m_hMenu);
	if (m_hMenu)
		::EnableMenuItem (m_hMenu, _wCmdId, MF_BYCOMMAND | (_bEnable ? MF_ENABLED : MF_GRAYED));
}

void CMenuBase::CheckItem (WORD _wCmdId, bool _bCheck) {
//	if (m_hwndToolbar)
//		::SendMessage (m_hwndToolbar, TB_SETSTATE, _wCmdId, TBSTATE_ENABLED | (_bCheck ? TBSTATE_CHECKED : 0));

	assert (m_hMenu);
	if (m_hMenu)
		::CheckMenuItem (m_hMenu, _wCmdId, MF_BYCOMMAND | (_bCheck ? MF_CHECKED : MF_UNCHECKED));
}

void CMenuBase::SetDefaultItem (WORD _wCmdId) {
	assert (m_hMenu);
	if (m_hMenu) {
		MENUITEMINFO mii;
		::memset (& mii, 0, sizeof (mii));
		mii.cbSize = sizeof (mii);
		mii.fMask  = MIIM_STATE;
		::GetMenuItemInfo (m_hMenu, _wCmdId, FALSE, & mii);
		mii.fState |= MFS_DEFAULT;
		::SetMenuItemInfo (m_hMenu, _wCmdId, FALSE, & mii);
	}
}

/*void CMenuBase::SetIcons (UINT _uiIconsBmp, const menuitem_data_t * _pMenuItems, size_t _cMenuItems) {
	m_pMenuItems = _pMenuItems;
	m_cMenuItems = _cMenuItems;

	assert (m_hilIcons == NULL);
	if (_uiIconsBmp != 0)
		m_hilIcons = LoadImageList (_uiIconsBmp, 16, c_colLtGray);

	if (GetToolbar () && m_hilIcons != NULL)
		::SendMessage (GetToolbar (), TB_SETIMAGELIST, 0, reinterpret_cast<LPARAM> (m_hilIcons));
}*/

size_t CMenuBase::_FindImageForCmd (WORD _wCmdId) const {
	if (! m_pMenuItems)
		return -1;

	for (size_t c = 0; c < m_cMenuItems; ++ c)
		if (m_pMenuItems [c].wCmdId == _wCmdId)
			return m_pMenuItems [c].wImage;

	return -1;
}

void CMenuBase::UpdateODItems () {
//	_MakeMenuItemsOD (m_hMenu, m_bHorzMenu);
}

void CMenuBase::_MakeMenuItemsOD (HMENU _hMenu, bool _bHorzMenu) {
	const size_t cItems = _GetMenuItemCount (_hMenu);
	if (cItems == 0)
		return;

	for (size_t c = 0; c < cItems; ++ c) {
		MENUITEMINFO mii;
		::memset (& mii, 0, sizeof (mii));
		mii.cbSize = sizeof (mii);
		mii.fMask  = MIIM_FTYPE | MIIM_SUBMENU;
		if (::GetMenuItemInfo (_hMenu, c, TRUE, & mii) == FALSE)
			continue;

		if (mii.hSubMenu != NULL) {
			// NOTE: apply to sub-menu recursively.
			_MakeMenuItemsOD (mii.hSubMenu, false);
			if (_bHorzMenu)
				continue;
		}

		if (mii.fType != MFT_STRING)
			continue;
					
		TCHAR strText [64] = _T ("");

		// Get menu item string.
		::memset (& mii, 0, sizeof (mii));
		mii.cbSize     = sizeof (mii);
		mii.fMask      = MIIM_TYPE;
		mii.fType      = MFT_STRING;
		mii.dwTypeData = strText;
		mii.cch        = sizeof (strText)/sizeof (strText [0]);
		if (::GetMenuItemInfo (_hMenu, c, TRUE, & mii) == FALSE)
			break;

		// Save menu item string to refer it via dwItemData.
		TCHAR * const strTextP = new TCHAR [mii.cch + 1];
		if (mii.cch)
			::_tcsncpy (strTextP, mii.dwTypeData, mii.cch);
		strTextP [mii.cch] = _T ('\0');
		m_MenuStrings.push_back (strTextP);

		// Set MFT_OWNERDRAW type.
		::memset (& mii, 0, sizeof (mii));
		mii.cbSize = sizeof (mii);
		mii.fMask  = MIIM_FTYPE | MIIM_DATA;
		mii.fType  = MFT_OWNERDRAW;
		mii.dwItemData = reinterpret_cast<DWORD> (strTextP);
		::SetMenuItemInfo (_hMenu, c, TRUE, & mii);
	}
}

const int c_x_icon0   = 3;
const int c_y_icon0   = 2;
const int c_x_icon1   = c_x_icon0 + 18;
const int c_y_icon1   = c_y_icon0 + 16;

const int c_x_text0   = c_x_icon1 + 3;
const int c_y_text0   = 2;

const int c_x_padding = 0;
//const int c_w_acc     = 64;

void CMenuBase::OnMenuMeasureItem (LPARAM _lParam) const {
	MEASUREITEMSTRUCT * const pMI = reinterpret_cast<MEASUREITEMSTRUCT *> (_lParam);

	// Get text.
	const TCHAR * const strText = reinterpret_cast<TCHAR *> (pMI->itemData);
	size_t cLen = ::_tcslen (strText);
	const TCHAR * const pTab = ::_tcschr (strText, _T ('\t'));
	if (pTab)
		cLen = pTab - strText;

	CClientDC hdcScreen (NULL);

	::SelectObject (hdcScreen, ::GetStockObject (SYSTEM_FONT));

	// Determine the text width.
	RECT rectText = {0, 0, 1, 1};
	::DrawText (hdcScreen, strText, cLen, & rectText, DT_LEFT | DT_TOP | DT_SINGLELINE | DT_CALCRECT);

	const int nH = ::GetSystemMetrics (SM_CYMENU);
	const int c_w_acc = nH*8/3;

	pMI->itemWidth = c_x_text0 + (rectText.right - rectText.left) + c_x_padding;
	if (pTab)
		pMI->itemWidth += c_w_acc;

	pMI->itemHeight = nH;
}

void CMenuBase::OnMenuDrawItem (LPARAM _lParam) const {
	const DRAWITEMSTRUCT * const pDI = reinterpret_cast<DRAWITEMSTRUCT *> (_lParam);

	const HDC hDC = pDI->hDC;
	const int x = pDI->rcItem.left;
	const int y = pDI->rcItem.top;
	const HMENU hmenu = reinterpret_cast<HMENU> (pDI->hwndItem);
	const size_t cItemID = pDI->itemID;

	const bool bSelected = (pDI->itemState & ODS_SELECTED) != 0;
	const bool bChecked  = (pDI->itemState & ODS_CHECKED ) != 0;
	const bool bGrayed   = (pDI->itemState & ODS_GRAYED  ) != 0;

	// Erase background.
	::FillRect (
		pDI->hDC, & pDI->rcItem,
		::GetSysColorBrush (bSelected ? COLOR_HIGHLIGHT : COLOR_MENU)
	);

	HBRUSH hbrushPrev = NULL;
	HPEN   hpenPrev   = NULL;

	// Detect radiobutton-style item.
	MENUITEMINFO mii;
	::memset (& mii, 0, sizeof (mii));
	mii.cbSize = sizeof (mii);
	mii.fMask  = MIIM_FTYPE;
	::GetMenuItemInfo (hmenu, cItemID, FALSE, & mii);
	const bool bRadioCheck = (mii.fType & MFT_RADIOCHECK) != 0;

	if (bRadioCheck) {
		if (bChecked) {
			// Draw option mark.
			hbrushPrev = (HBRUSH) ::SelectObject (hDC, ::GetStockObject (BLACK_BRUSH));
			hpenPrev   = (HPEN)   ::SelectObject (hDC, ::GetStockObject (BLACK_PEN));

			const int xc = x + (c_x_icon0 + c_x_icon1)/2;
			const int yc = (pDI->rcItem.top + pDI->rcItem.bottom)/2;
			const int r = 3;
			::Ellipse (
				hDC,
				xc - r,     yc - r,
				xc + r - 1, yc + r - 1
			);
		}
	} else {
		// Draw checkmark frame, if any.
		if (bChecked) {
			RECT rectCheck = {
				x + c_x_icon0, y + c_y_icon0,
				x + c_x_icon1, y + c_y_icon1
			};
			::DrawFrameControl (hDC, & rectCheck, DFC_BUTTON, DFCS_BUTTONCHECK | DFCS_CHECKED);
		} else {
			// Draw icon.
			HIMAGELIST hil = m_hilIcons;
			size_t cImageId = _FindImageForCmd (cItemID);
			if (cImageId == -1) {
				// Try to find icon in toolbar.
				hil = reinterpret_cast<HIMAGELIST> (::SendMessage (GetToolbar (), TB_GETIMAGELIST, 0, 0));
				cImageId = ::SendMessage (GetToolbar (), TB_GETBITMAP, cItemID, 0);
				if (cImageId == 0)
					cImageId = -1;
			}
			if (hil && cImageId != -1) {
				::ImageList_Draw (
					hil, cImageId, hDC,
					x + c_x_icon0,
					y + c_y_icon0,
					ILD_TRANSPARENT | (bGrayed ? ILD_SELECTED : 0)
				);
			}
		}
	}

	// Get text.
	const TCHAR * const strTextFull = reinterpret_cast<TCHAR *> (pDI->itemData);

	const size_t cLen = ::_tcslen (strTextFull);
	TCHAR * const strText = (TCHAR *) alloca ((cLen + 1)*sizeof (TCHAR));
	::_tcscpy (strText, strTextFull);

	// Truncate \t.
	TCHAR * pAccelerator = ::_tcschr (strText, _T ('\t'));
	if (pAccelerator) {
		* pAccelerator = _T ('\0');
		++ pAccelerator;
	}

	// Prepare to text drawing.
	const int nTextColor = bGrayed ? COLOR_GRAYTEXT : (bSelected ? COLOR_HIGHLIGHTTEXT : COLOR_MENUTEXT);
	const COLORREF clrPrevText = ::SetTextColor (hDC, ::GetSysColor (nTextColor));

	const int nBkModePrev = ::SetBkMode (hDC, TRANSPARENT);

	const int c_w_acc = ::GetSystemMetrics (SM_CYMENU)*8/3;

	::SelectObject (hDC, ::GetStockObject (SYSTEM_FONT));

	// Draw the text.
	RECT rectText = {
		x + c_x_text0,
		y + c_y_text0,
		pDI->rcItem.right - ((pAccelerator ? c_w_acc : 0) + c_x_padding),
		pDI->rcItem.bottom
	};
	::DrawText (hDC, strText, -1, & rectText, DT_LEFT | DT_TOP | DT_SINGLELINE);

/*{
	RECT rectText = {0, 0, 1, 1};
	::DrawText (hDC, strText, -1, & rectText, DT_LEFT | DT_TOP | DT_SINGLELINE | DT_CALCRECT);
	int w = rectText.right - rectText.left;
	assert (w <= pDI->rcItem.right - pDI->rcItem.left);
}*/

	if (pAccelerator) {
		RECT rectAccelerator = {
			pDI->rcItem.right - c_w_acc,
			y + c_y_text0,
			pDI->rcItem.right - c_x_padding,
			pDI->rcItem.bottom
		};
		::DrawText (hDC, pAccelerator, -1, & rectAccelerator, DT_LEFT | DT_TOP | DT_SINGLELINE);
	}

	// Restore the original HDC state.
	::SetTextColor (hDC, clrPrevText);
	::SetBkMode    (hDC, nBkModePrev);

	if (hbrushPrev)
		::SelectObject (hDC, hbrushPrev);
	if (hpenPrev)
		::SelectObject (hDC, hpenPrev);
}

////////////////////////////////////////////////

CMainMenu g_MainMenu;

void CMainMenu::Init (HMENU _hMenu) {
	m_hMenu = _hMenu;

	// Prepare menu icons.
	m_bHorzMenu = true;
	UpdateODItems ();
}

void CMainMenu::SetToolbar (HWND _hwndToolbar) {
	m_hwndToolbar = _hwndToolbar;
}

////////////////////////////////////////////////

# define KS_HIDDEN_MENU_ITEMS "HiddenMenuItems" GME_VER_SUFFIX
# define KS_SHOWN_MENU_ITEMS  "ShownMenuItems" GME_VER_SUFFIX

static const WORD s_wHiddenMenuItems_Default [] = {ID_JOIN_RGN_LEVELS, ID_SPLIT_RGN_LEVELS};

static
void _VectorToSet (const WORD * _pwItems, size_t _cItems, std::set<WORD> & _setItems) {
	for (size_t c = 0; c < _cItems; ++ c)
		_setItems.insert (_pwItems [c]);
}

void GetHiddenMenuItems (bool _bDefault, hidden_menu_items_t & _HiddenItems) {
	const WORD * pwHiddenMenuItems = s_wHiddenMenuItems_Default;
	size_t cHiddenItems = ARRAY_SIZE (s_wHiddenMenuItems_Default);

	size_t c;
	for (c = 0; c < cHiddenItems; ++ c)
		_HiddenItems.insert (pwHiddenMenuItems [c]);

	if (! _bDefault) {
		const size_t cSize = GetSettingLen (KS_HIDDEN_MENU_ITEMS);
		if (cSize != 0) {
			WORD * const pwHiddenMenuItems2 = (WORD *) alloca (cSize);
			GetSetting (KS_HIDDEN_MENU_ITEMS, pwHiddenMenuItems2, cSize);

			pwHiddenMenuItems = pwHiddenMenuItems2;
			cHiddenItems = cSize/sizeof (WORD);
		}
	}

	for (c = 0; c < cHiddenItems; ++ c)
		_HiddenItems.insert (pwHiddenMenuItems [c]);

	if (! _bDefault) {
		const size_t cSize = GetSettingLen (KS_SHOWN_MENU_ITEMS);
		if (cSize != 0) {
			WORD * const pwShownMenuItems = (WORD *) alloca (cSize);
			GetSetting (KS_SHOWN_MENU_ITEMS, pwShownMenuItems, cSize);

			const size_t cShownItems = cSize/sizeof (WORD);
			for (size_t c = 0; c < cShownItems; ++ c)
				_HiddenItems.erase (pwShownMenuItems [c]);
		}
	}
}

void _HideMenuItems (HMENU _hmenu) {
	hidden_menu_items_t HiddenItems;
	GetHiddenMenuItems (false, HiddenItems);

	for (hidden_menu_items_t::const_iterator i = HiddenItems.begin (); i != HiddenItems.end (); ++ i)
		::DeleteMenu (_hmenu, * i, MF_BYCOMMAND);

	DeleteDuplicatedMenuSeparators (_hmenu, true, true);
}

void SetHiddenMenuItems (const hidden_menu_items_t & _HiddenItems) {
	// Save list of hidden items.
	std::vector<WORD> wHiddenItems;
	for (hidden_menu_items_t::const_iterator i = _HiddenItems.begin (); i != _HiddenItems.end (); ++ i)
		wHiddenItems.push_back (* i);
	SaveSetting (KS_HIDDEN_MENU_ITEMS, wHiddenItems.empty () ? NULL : & * wHiddenItems.begin (), wHiddenItems.size ()*sizeof (WORD));

	// Save list of shown items.
	std::vector<WORD> wShownItems;
	for (size_t c = 0; c != ARRAY_SIZE (s_wHiddenMenuItems_Default); ++ c) {
		const WORD wID = s_wHiddenMenuItems_Default [c];
		if (_HiddenItems.find (wID) == _HiddenItems.end ())
			wShownItems.push_back (wID);
	}
	SaveSetting (KS_SHOWN_MENU_ITEMS, wShownItems.empty () ? NULL : & * wShownItems.begin (), wShownItems.size ()*sizeof (WORD));

	// Restore original menu.
	const HMENU hMainMenu = LoadMenu (IDM_MAIN);
	const HMENU hMainMenuOld = ::GetMenu (g_hWnd);
	::SetMenu (g_hWnd, hMainMenu);
	::DestroyMenu (hMainMenuOld);
	InitMenu ();

	UpdateMenuAccelerators (g_MainMenu);
}
