//
// GPSMapEdit
// (c) Konstantin Galichsky (kg@geopainting.com), 2002-2009
// Garmin and MapSource are registered trademarks or trademarks of Garmin Ltd. or one of its subsidiaries.
//
// Garmin MapSource MPS binary files parser.
//

# include "StdAfx.h"
# include "MpsFile.h"
# include "MapSourceHelpers.h"
# include "Map.h"
# include "Types.h"
# include "GarminTypes.h"
# include "RussaTypes.h"
# include "NaviTelTypes.h"
# include "Load.h"
# include "Log.h"
# include "Status.h"
# include "Globals.h"

class CMpsLoader : public CMapLoader {
  public:
	// Override.
	virtual bool IsValid (const char * _pData, size_t _cFileSize) const;
	virtual LoadStatus_t Load (const char * _pData, size_t _cFileSize, const char * _strFileName, map_t * _pMap);

  private:
	void ParseWaypoint (const CMpsWaypointHdr * _pWP, size_t _cOffset, size_t _cNextOffset);
	void ParseTrack    (const CMpsTrackHdr * _pTrack, size_t _cOffset, size_t _cNextOffset, size_t _cFileSize, bool _bRestore);

	map_t * m_pMap;

	//
	size_t m_cOrdinal;

	file_refs_t m_file_refs;
};

DECLARE_WPTRK_LOADER (CMpsLoader, "garmin-mps", "mps", "Garmin MapSource file (*.mps)", ftiWpts | ftiTracks | ftiMapFileByParts);

bool CMpsLoader::IsValid (const char * _pData, size_t _cFileSize) const {
	if (_cFileSize < 0x2d)
		return false;
	return ::memcmp (_pData, g_pMpsSignature, sizeof (g_pMpsSignature)) == 0;
}

void CMpsLoader::ParseWaypoint (const CMpsWaypointHdr * _pWP, size_t _cOffset, size_t _cNextOffset) {
	const size_t cNameLen = ::strlen (_pWP->strName);
	const CMpsWaypoint2 * const p2 = reinterpret_cast<const CMpsWaypoint2 *> (reinterpret_cast<const BYTE *> (_pWP->strName) + cNameLen + 1);

	const size_t cDescriptionLen = ::strlen (p2->strDescription);
	const CMpsWaypoint3 * const p3 = reinterpret_cast<const CMpsWaypoint3 *> (reinterpret_cast<const BYTE *> (p2->strDescription) + cDescriptionLen + 1);

	point_t point;
	point.x = p2->lX*180.f/(1UL << 31);
	point.y = p2->lY*180.f/(1UL << 31);

	// Find type.
	WORD wPOIType = GarminWPT2POIType (p3->dwType);
	if (wPOIType != 0 && m_pMap->pTypeSet == & g_tsNull) {
		// TODO: better solution is required!
		if (g_map.pTypeSet == & g_tsNull)
			m_pMap->pTypeSet = & g_tsGarmin;
		else if (g_map.pTypeSet == & g_tsGarmin || g_map.pTypeSet == & g_tsRussa || g_map.pTypeSet == & g_tsNavitel)
			m_pMap->pTypeSet = g_map.pTypeSet;
		else
			wPOIType = 0;
	}

	//
	// Add the waypoint to the map.
	//
	rgn_t & rgn = m_pMap->CreateRgn (rkPoint);
	rgn.SetWaypointType ();

	rgn.pFileRef = & m_file_refs.back ();
	rgn.cBeginPos = _cOffset;
	rgn.cEndPos   = _cNextOffset;
	rgn.cFileLine = -1;
//	rgn.cOrdinal = m_cOrdinal ++;

	if (cNameLen)
		rgn.strLabel.assign (_pWP->strName, cNameLen);
	if (cDescriptionLen)
		rgn.strComment.assign (p2->strDescription, cDescriptionLen);

	rgn.origins.push_back (rgn_origin_t ());
	rgn_origin_t & origin = rgn.origins.back ();
	origin.cLevel = 0;
	origin.point  = point;

	// Extra properties.
	waypoint_info_t * const pWptExtra = rgn.pExtra->wtr.CreateWaypoint ();

	// Altitude.
	if (p2->btHasAltitude) {
		pWptExtra->wFlags |= waypoint_info_t::defAltitude;
		pWptExtra->fAltitude = p2->dblAltitude;
	}

	// Proximity.
	if (p3->btHasProximity) {
		pWptExtra->wFlags |= waypoint_info_t::defProximity;
		pWptExtra->fProximity = p3->dblProximity;
	}

	// Depth.
	if (p3->btHasDepth) {
		pWptExtra->wFlags |= waypoint_info_t::defDepth;
		pWptExtra->fDepth = p3->dblDepth;
	}

	// Show.
	switch (p3->dwShow) {
		case 0: pWptExtra->SetShow (waypoint_info_t::showSymbol); break;
		default:
		case 1: pWptExtra->SetShow (waypoint_info_t::showName); break;
		case 2:	pWptExtra->SetShow (waypoint_info_t::showDescription); break;
	}

	// Garmin type.
	pWptExtra->wFlags |= waypoint_info_t::defGarminType;
	pWptExtra->wGarminType = p3->dwType;
}

void CMpsLoader::ParseTrack (const CMpsTrackHdr * _pTrack, size_t _cOffset, size_t _cNextOffset, size_t _cFileSize, bool _bRestore) {
	const size_t cNameLen = ::strlen (_pTrack->strName);

	const size_t cOffset2 = _cOffset + sizeof (CMpsRecordHdr) + cNameLen + 1;
	const size_t cOffsetPoints = cOffset2 + sizeof (CMpsTrack2) - sizeof (CMpsTrackPoint);

	const CMpsTrack2 * const p2 = reinterpret_cast<const CMpsTrack2 *> (reinterpret_cast<const BYTE *> (_pTrack->strName) + cNameLen + 1);
	DWORD dwPoints = p2->dwPoints;

	if (cOffsetPoints + dwPoints*sizeof (CMpsTrackPoint) != _cNextOffset) {
		if (_bRestore == false) {
			if (Silent () || YesNoMsgBox ("Track '%s' is probably damaged. Would you like to restore it?", _pTrack->strName)) {
				// Try to read track record up to the very end of file.
				const char * const pRecord = GetFilePart (_cOffset, _cFileSize);
				if (pRecord == NULL)
					return;
				const CMpsTrackHdr * const pTrack = reinterpret_cast<const CMpsTrackHdr *> (pRecord);
				ParseTrack (pTrack, _cOffset, _cFileSize, _cFileSize, true);
				return;
			}
		}

		// NOTE: the header defines incorrect number of points.
		ReportWarning ("Track '%s' is damaged. It will be restored.", _pTrack->strName);
		dwPoints = (_cNextOffset - cOffsetPoints)/sizeof (CMpsTrackPoint);
	}

	points_t points (dwPoints);
	trackpoints_t points_extra (dwPoints);

	const CMpsTrackPoint * pPoint = p2->points;
	for (size_t cPoint = 0; cPoint < dwPoints; ++ cPoint, ++ pPoint) {
		point_t & point = points [cPoint];
		point.x = pPoint->lX*180.f/(1UL << 31);
		point.y = pPoint->lY*180.f/(1UL << 31);

		assert (-180 <= point.x && point.x <= 180);
		assert (-90  <= point.y && point.y <= 90);

		trackpoint_info_t & point_extra = points_extra [cPoint];
		if (pPoint->btHasAltitude) {
			point_extra.wFlags |= trackpoint_info_t::defAltitude;
			point_extra.fAltitude = pPoint->dblAltitude;
		}
		if (pPoint->btHasTime) {
			point_extra.wFlags |= trackpoint_info_t::defTime;
			point_extra.dtTime = pPoint->dwDateTime/24./60./60. + 25569.;
		}
		if (pPoint->btHasDepth) {
			point_extra.wFlags |= trackpoint_info_t::defDepth;
			point_extra.fDepth = pPoint->dblDepth;
		}
	}

	//
	// Add the track to the map.
	//
	rgn_t & rgn = m_pMap->CreateRgn (rkPolyline);
	rgn.SetTrackType ();

	rgn.pFileRef  = & m_file_refs.back ();
	rgn.cBeginPos = _cOffset;
	rgn.cEndPos   = _cNextOffset;
	rgn.cFileLine = -1;
//	rgn.cOrdinal  = m_cOrdinal ++;

	if (cNameLen)
		rgn.strLabel.assign (_pTrack->strName, cNameLen);

	rgn.elements.push_back (rgn_element_t ());
	rgn_element_t & element = rgn.elements.back ();
	element.cFileOffset = _cOffset;

	//
	element.cLevel = 0;
	element.points.swap (points);

	track_info_t * const pTrackExtra = rgn.pExtra->wtr.CreateTrack ();
	element.pExtra->TrackPoints.swap (points_extra);

	if (p2->btDisplayOnGPS == 0)
		pTrackExtra->wFlags |= track_info_t::defHideInGPS;

	if (1 <= p2->dwColor && p2->dwColor < c_cGarminTrackColors) {
		pTrackExtra->wFlags |= track_info_t::defLineColor;
		pTrackExtra->colLine = g_GarminTrackColors [p2->dwColor];
	}
}

LoadStatus_t CMpsLoader::Load (const char * _pData, size_t _cFileSize, const char * _strMapFileName, map_t * _pMap) {
	m_pMap = _pMap;

	// Check the end of file for zeroes.
	if (_cFileSize < 0x2000000) { // <32M.
		const char * const pLastByte_tmp = GetFilePart (_cFileSize - 1, _cFileSize);
		if (* pLastByte_tmp == 0) {
			size_t cZeroes;
			for (cZeroes = 1; cZeroes < _cFileSize; ++ cZeroes) {
				const char * const pByte_tmp = GetFilePart (_cFileSize - 1 - cZeroes, _cFileSize - cZeroes);
				if (* pByte_tmp != 0)
					break;
			}	
			if (cZeroes > 16) {
				_cFileSize -= cZeroes;
				if (! Silent ())
					ErrorMsgBox ("The file '%s' is damaged: last %d byte(s) are zeroes.", _strMapFileName, cZeroes);
				ReportWarning ("The file '%s' is damaged: last %d byte(s) are zeroes.", _strMapFileName, cZeroes);
			}
		}
	}

	m_cOrdinal = 0;

	m_file_refs.clear ();
	m_file_refs.push_back (file_ref_t ());
	m_file_refs.back ().strFileName = _strMapFileName;
	m_file_refs.back ().bAttachment = true;

	m_pMap->cHeaderLen = 0;

	SetProgressStatus (0);

	size_t cOffset = sizeof (g_pMpsSignature);
	do {
		if (g_bStop)
			return lsInterrupted;

		const size_t cEndOffset_tmp = __min (cOffset + sizeof (CMpsRecordHdr), _cFileSize);
		const char * const pRecord_tmp = GetFilePart (cOffset, cEndOffset_tmp);
		if (pRecord_tmp == NULL)
			return lsFailed;
		const CMpsRecordHdr * const pHdr_tmp = reinterpret_cast<const CMpsRecordHdr *> (pRecord_tmp);

		const size_t cNextOffset = __min (cOffset + sizeof (CMpsRecordHdr) + pHdr_tmp->dwLen, _cFileSize);
		const char * const pRecord = GetFilePart (cOffset, cNextOffset);
		if (pRecord == NULL)
			return lsFailed;
		const CMpsRecordHdr * const pHdr = reinterpret_cast<const CMpsRecordHdr *> (pRecord);

		switch (pHdr->btType) {
			case 'D': { // MPS format version.
				const CMpsFormatVersion * const pVer = static_cast<const CMpsFormatVersion *> (pHdr);
				ReportText ("Format version: 0x%02X.", pVer->wdVersion);
				if (pVer->wdVersion < 0x67) {
					ReportWarning ("The file is created by an obsolete version of MapSource. MapSource 4.09 or later is required.");
					return lsFailed;
				} else if (pVer->wdVersion > 0x69)
					ReportWarning ("The file is created by a version of MapSource later than 5.4. Such format was not tested, and result of file loading may be unpredictable.");
				break;
			}
			case 'A': { // MapSource version.
				const CMpsMSVersion * const pVer = static_cast<const CMpsMSVersion *> (pHdr);
/*				if (pVer->btVersion < 0x99) {
					ReportWarning ("The file is created by an obsolete version of MapSource. MapSource 4.09 or later is required.");
					return lsFailed;
				} else if (pVer->btVersion > 0x9d)
					ReportWarning ("The file is created by a version of MapSource later than 4.13. Such format was not tested, and result of file loading may be unpredictable.");*/
				break;
			}

			case 'W': { // Waypoint.
				if (m_pMap->cHeaderLen == 0)
					m_pMap->cHeaderLen = cOffset;

				const CMpsWaypointHdr * const pWP = static_cast<const CMpsWaypointHdr *> (pHdr);
				ParseWaypoint (pWP, cOffset, cNextOffset);
				break;
			}

			case 'T': { // Track.
				if (m_pMap->cHeaderLen == 0)
					m_pMap->cHeaderLen = cOffset;

				const CMpsTrackHdr * const pTrack = static_cast<const CMpsTrackHdr *> (pHdr);
				ParseTrack (pTrack, cOffset, cNextOffset, _cFileSize, false);
				break;
			}

			case 'R':  // Route.
			case 'L':  // Map.
				if (m_pMap->cHeaderLen == 0)
					m_pMap->cHeaderLen = cOffset;
				break;

			case 'V': { // Tail mark and map set info.
				break;
			}
		}

		cOffset = cNextOffset;

		SetProgressStatus (::MulDiv (50, cOffset, _cFileSize));
	} while (cOffset < _cFileSize);

	SetProgressStatus (50);

	m_pMap->file_refs.swap (m_file_refs);

	// Suppress warning about empty Zoom etc.
	m_pMap->bNoHeader = true;

	return lsOK;
}
