//
// GPSMapEdit
// (c) Konstantin Galichsky (kg@geopainting.com), 2008
//
// MPS track writer.
//

# include "StdAfx.h"
# include "Tracking.h"
# include "Positioning.h"
# include "MpsFile.h"
# include "FileWriter.h"
# include "fast_floor.h"

class CMpsTrackWriter : public CTrackWriter {
  public:
	virtual void Start ();
	virtual void Stop ();
	virtual void StartNewTrack ();
	virtual void AddPointToTrack (const PositionInfo_t & _pi);

  private:
	void _FixTrackHdr ();

	size_t m_cTrackNumber;
	size_t m_cTrackOffset;
	size_t m_cPoints;
	size_t m_cPointsOffset;
};

DECLARE_TRACK_WRITER (CMpsTrackWriter, "mps", "Garmin MapSource file (*.mps)");

void CMpsTrackWriter::_FixTrackHdr () {
	assert (m_pFileWriter);
	const size_t cEnd = m_pFileWriter->GetPos ();

	// Save true size of the record.
	m_pFileWriter->SetPos (m_cTrackOffset);
	m_pFileWriter->WriteDWORD (cEnd - m_cTrackOffset - sizeof (CMpsRecordHdr));

	// Save true number of points.
	m_pFileWriter->SetPos (m_cPointsOffset);
	m_pFileWriter->WriteDWORD (m_cPoints);

	m_pFileWriter->SetPos (cEnd);
}

void CMpsTrackWriter::Start () {
	// Write the header.
	m_pFileWriter->WriteBin (g_pMpsHeader, c_cMpsHeaderSize);

	m_cTrackNumber = 0;
	m_cTrackOffset = 0;
	m_cPoints = 0;
	m_cPointsOffset = 0;
}

void CMpsTrackWriter::Stop () {
	// Update header of the last track.
	_FixTrackHdr ();

	// Save tail record.
	assert (m_pFileWriter);
	m_pFileWriter->WriteBin (g_pMpsTail, c_cMpsTailSize);

	m_pFileWriter->FlushBuffer ();
}

void CMpsTrackWriter::StartNewTrack () {
	if (m_cTrackNumber > 0 && m_cPoints == 0)
		return;

	if (m_cTrackNumber > 0)
		_FixTrackHdr ();

	// Prepare the name of the track.
	char strTrackName [32];
	::sprintf (strTrackName, "Track %03d", m_cTrackNumber);
	++ m_cTrackNumber;

	// Save 'size' field offset to fix it later.
	assert (m_pFileWriter);
	m_cTrackOffset = m_pFileWriter->GetPos ();
	m_pFileWriter->WriteDWORD (1000*sizeof (CMpsTrackPoint));

	m_pFileWriter->WriteBYTE ('T');   // Track

	m_pFileWriter->WriteBin (strTrackName, ::strlen (strTrackName) + 1);
	m_pFileWriter->WriteBYTE (0);     // btDisplayOnGPS = no
	m_pFileWriter->WriteDWORD (0);    // dwColor = black

	// Save 'points' field offset to fix it later.
	m_cPointsOffset = m_pFileWriter->GetPos ();
	m_pFileWriter->WriteDWORD (1000); // dwPoints = 1000 (in advance)
	m_cPoints = 0;
}

void CMpsTrackWriter::AddPointToTrack (const PositionInfo_t & _pi) {
	if ((_pi.dwDef & PositionInfo_t::defXY) == 0)
		return;

	++ m_cPoints;

	CMpsTrackPoint tp;
	::memset (& tp, 0, sizeof (tp));

	tp.lY = fast_floor_d (_pi.fY*(1UL << 31)/180.f + .5f);
	tp.lX = fast_floor_d (_pi.fX*(1UL << 31)/180.f + .5f);

	if (_pi.dwDef & PositionInfo_t::defAltitude) {
		tp.btHasAltitude = 1;
		tp.dblAltitude   = _pi.fAltitude;
	}

	if (_pi.dwDef & PositionInfo_t::defTime) {
		tp.btHasTime  = 1;
		// In seconds elapsed since 1 Jan 1970
		tp.dwDateTime = fast_floor_d ((_pi.dtTime - 25569.)*24.*60.*60. + .5);
	}

	if (_pi.dwDef & PositionInfo_t::defDepth) {
		tp.btHasDepth = 1;
		tp.dblDepth   = _pi.fDepth;
	}

	assert (m_pFileWriter);
	m_pFileWriter->WriteBin (& tp, sizeof (tp));
}
