//
// GPSMapEdit
// (c) Konstantin Galichsky (kg@geopainting.com), 2008
//
// Navitel skin parser.
//

# include "StdAfx.h"
# include "MapSkins.h"
# include "ParseHelpers.h"
# include "NavitelTypes.h"
# include "PNG.h"
# include "Log.h"
# include "Wrappers.h"
# include "Globals.h"

class CNSFile {
  public:
	CNSFile (const char * _strFileName);
	~CNSFile ();

	bool IsValid () const {return m_pFileData != NULL;}

	const BYTE * GetFile (const char * _strFileName, size_t & _cFileSize) const;

	struct CSubFile {
		size_t cFileSize;
		size_t cFileOffset;
	};

  private:
	const BYTE * m_pFileData;

	typedef std::map<string_t, CSubFile> files_t;
	files_t m_Files;
};

# ifdef _MSC_VER
#	pragma pack (push, 1)
# endif
typedef struct local_file_header {                 /* LOCAL */
	DWORD sig;
	BYTE version_needed_to_extract[2];
	WORD general_purpose_bit_flag;
	WORD compression_method;
	DWORD last_mod_dos_datetime;
	DWORD crc32;
	DWORD csize;
	DWORD ucsize;
	WORD filename_length;
	WORD extra_field_length;
} local_file_hdr;

typedef struct central_directory_file_header {     /* CENTRAL */
	DWORD sig;
	BYTE version_made_by[2];
	BYTE version_needed_to_extract[2];
	WORD general_purpose_bit_flag;
	WORD compression_method;
	DWORD last_mod_dos_datetime;
	DWORD crc32;
	DWORD csize;
	DWORD ucsize;
	WORD filename_length;
	WORD extra_field_length;
	WORD file_comment_length;
	WORD disk_number_start;
	WORD internal_file_attributes;
	DWORD external_file_attributes;
	DWORD relative_offset_local_header;
} cdir_file_hdr;

typedef struct end_central_dir_record {            /* END CENTRAL */
	DWORD sig;
	WORD number_this_disk;
	WORD num_disk_start_cdir;
	WORD num_entries_centrl_dir_ths_disk;
	WORD total_entries_central_dir;
	DWORD size_central_directory;
	DWORD offset_start_central_directory;
	WORD zipfile_comment_length;
} ecdir_rec;
# ifdef _MSC_VER
#	pragma pack (pop)
# endif

const BYTE local_file_header_sig             [4] = {0x50, 0x4b, 0x03, 0x04};
const BYTE central_directory_file_header_sig [4] = {0x50, 0x4b, 0x01, 0x02};
const BYTE end_central_dir_record_sig        [4] = {0x50, 0x4b, 0x05, 0x06};

static
int _FindEndCDRecord (const BYTE * _pData, size_t _nSize) {
	int offset = _nSize - sizeof (ecdir_rec);
	while (offset > 0) {
		if (::memcmp (end_central_dir_record_sig, _pData + offset, ARRAY_SIZE (end_central_dir_record_sig)) == 0)
			break;
		-- offset;
	}
	return offset;
}

CNSFile::CNSFile (const char * _strFileName) :
	m_pFileData (NULL)
{
	assert (_strFileName != NULL);

	CFileForRead hFile (_strFileName);
	if (! hFile.IsValid ())
		return;

	const size_t cFileSize = ::GetFileSize (hFile, NULL);
	CWinHandle hMapping = ::CreateFileMapping (hFile, NULL, PAGE_READONLY, 0, 0, NULL);
	if (! hMapping.IsValid ())
		return;
	m_pFileData = reinterpret_cast<const BYTE *> (::MapViewOfFile (hMapping, FILE_MAP_READ, 0, 0, 0));

	const BYTE * const pData = m_pFileData;

	const int endCDRecordOffset = _FindEndCDRecord (pData, cFileSize);
	if (endCDRecordOffset <= 0)
		return;
	const ecdir_rec * const pECDRec = reinterpret_cast<const ecdir_rec *> (pData + endCDRecordOffset);

	int nCDRecordOffset = pECDRec->offset_start_central_directory;
	int nCDRecord = pECDRec->total_entries_central_dir;
	while (nCDRecord > 0) {
		if (::memcmp (central_directory_file_header_sig, pData + nCDRecordOffset, ARRAY_SIZE (central_directory_file_header_sig)) != 0)
			break;
		const cdir_file_hdr * const pCDRec = reinterpret_cast<const cdir_file_hdr *> (pData + nCDRecordOffset);
		if (::memcmp (local_file_header_sig, pData + pCDRec->relative_offset_local_header, ARRAY_SIZE (local_file_header_sig)) != 0)
			break;
		const local_file_header * const pLFHeader = reinterpret_cast<const local_file_header *> (pData + pCDRec->relative_offset_local_header);

		const char * const strFileNameLocal = reinterpret_cast<const char *> (pData + pCDRec->relative_offset_local_header + sizeof (local_file_header));
		string_t strFileName (strFileNameLocal, pLFHeader->filename_length);
		strFileName.ToLower ();

		CSubFile file;
		file.cFileSize   = pLFHeader->csize;
		file.cFileOffset = pCDRec->relative_offset_local_header + sizeof (local_file_header) + pLFHeader->filename_length + pLFHeader->extra_field_length;
		m_Files [strFileName] = file;

		nCDRecordOffset += sizeof (cdir_file_hdr) + pCDRec->filename_length + pCDRec->file_comment_length + pCDRec->extra_field_length;
		-- nCDRecord;
	}
}

CNSFile::~CNSFile () {
	if (m_pFileData)
		::UnmapViewOfFile (m_pFileData);
	m_pFileData = NULL;
}

const BYTE * CNSFile::GetFile (const char * _strFileName, size_t & _cFileSize) const {
	_cFileSize = 0;
	if (! IsValid ())
		return NULL;

	const files_t::const_iterator i = m_Files.find (_strFileName);
	if (i == m_Files.end ())
		return NULL;
	const CSubFile & f = i->second;

	_cFileSize = f.cFileSize;
	return m_pFileData + f.cFileOffset;
}

//////////////////////////////////////////////////////////////////////////////

class CNavitelSkinParser {
  public:
	CNavitelSkinParser (const CNSFile & _nsf, const BYTE * _pFileData, size_t _cFileSize) :
		m_nsf (_nsf),
		m_cLine (1),
		m_bV11 (true),
		m_pSkin (NULL),
		m_cPatternsW (0), m_cPatternsH (0),
		m_cPOIsW (0), m_cPOIsH (0),
		m_hbmpPatterns (NULL),
		m_phdcPatterns (NULL),
		m_phdcMemPattern (NULL),
		m_phdcMem (NULL),
		m_pTD_Unknown (g_tsNavitel.GetUnknownTD ())
	{
		m_pStart = (const char *) _pFileData;
		m_pEnd = ((const char *) (m_pStart + _cFileSize));
	}

	~CNavitelSkinParser () {
		if (m_hbmpPatterns)
			::DeleteObject (m_hbmpPatterns);
	}

	bool Parse (CMapSkin * _pSkin);

	size_t GetLine () const {return m_cLine;}

  private:
	const CNSFile & m_nsf;

	const char * m_pStart;
	const char * m_pEnd;

	bool m_bV11;

	CMapSkin * m_pSkin;

	bool ParseColor        (const char * & _p, COLORREF & _col);
	bool ParseOutlineColor (const char * & _p, COLORREF & _col);
	bool ParsePenStyle     (const char * & _p, short & _nStyle);
	bool ParseBool         (const char * & _p, bool & _b);
	bool ParseFontSize     (const char * & _p);
	bool ParseRFontSize    (const char * & _p);

	const char * ParsePolygonDesc  (const char * _pLine);
	const char * ParsePolylineDesc (const char * _pLine);
	const char * ParsePointDesc    (const char * _pLine);
	const char * ParseImageList    (const char * _pLine, HBITMAP * _pBmp, size_t * _pcWidth, size_t * _pcHeight);

	size_t m_cLine;
	const char * FindNextLine (const char * _pBegin);

  public:
	struct kv_pair_t {
		const char * strKey;
		DWORD dwValue;
	};
  private:
	bool _Key2Value (const char * _pData, const char ** _ppEnd, const kv_pair_t * _pKVPairs, size_t _cSize, DWORD & _dwValue);

	const type_desc_t * const m_pTD_Unknown;

	size_t m_cPatternsW, m_cPatternsH;
	size_t m_cPOIsW, m_cPOIsH;

	HBITMAP m_hbmpPatterns;

	CDCHolder * m_phdcPatterns;
	CMemDC * m_phdcMemPattern;
	CMemDC * m_phdcMem;
};

bool CNavitelSkinParser::_Key2Value (const char * _pData, const char ** _ppEnd, const kv_pair_t * _pKVPairs, size_t _cSize, DWORD & _dwValue) {
	OmitSpaces (& _pData, m_pEnd);

	for (size_t c = 0; c < _cSize; ++ c) {
		const kv_pair_t & pair = _pKVPairs [c];

		const size_t cKeyLen = ::strlen (pair.strKey);
		if (::strncmp (_pData, pair.strKey, cKeyLen) == 0) {
			if (_ppEnd != NULL)
				* _ppEnd = _pData + cKeyLen;
			_dwValue = pair.dwValue;
			return true;
		}
	}

	return false;
}

const char * CNavitelSkinParser::FindNextLine (const char * _pBegin) {
	const char * p = _pBegin;
	do {
		p = ::FindNextLine (p, m_pEnd);
		++ m_cLine;

		// Omit spaces and TABs.
		while (p < m_pEnd && (* p == ' ' || * p == '\t'))
			++ p;
		if (* p == '\n' || * p == '\r')
			continue;
		if (* p != '#')
			return p;
	} while (p < m_pEnd);

	return m_pEnd;
}

////////////////////////////////////////////

const CNavitelSkinParser::kv_pair_t g_kvpColors [] = {
	{"black",     c_colBlack},
	{"white",     c_colWhite},
	{"ltgray",    c_colLtGray},
	{"gray",      c_colGray},
	{"dkgray",    c_colDkGray},
	{"red",       c_colRed},
	{"dkred",     c_colDkRed},
	{"green",     c_colGreen},
	{"dkgreen",   c_colDkGreen},
	{"blue",      c_colBlue},
	{"dkblue",    c_colDkBlue},
	{"yellow",    c_colYellow},
	{"dkyellow",  c_colDkYellow},
	{"cyan",      c_colCyan},
	{"dkcyan",    c_colDkCyan},
	{"magenta",   c_colMagenta},
	{"dkmagenta", c_colDkMagenta},
};

bool CNavitelSkinParser::ParseColor (const char * & _p, COLORREF & _col) {
	DWORD dwColor = 0;
	if (_Key2Value (_p, & _p, g_kvpColors, ARRAY_SIZE (g_kvpColors), dwColor)) {
		_col = dwColor;
		return true;
	}

	OmitSpaces (& _p, m_pEnd);

	unsigned long ulR;
	if (! ParseHex (_p, _p + 2, & _p, ulR))
		return false;
	if (ulR >= 256)
		return false;

	unsigned long ulG;
	if (! ParseHex (_p, _p + 2, & _p, ulG))
		return false;
	if (ulG >= 256)
		return false;

	unsigned long ulB;
	if (! ParseHex (_p, _p + 2, & _p, ulB))
		return false;
	if (ulB >= 256)
		return false;

	_col = RGB (ulR, ulG, ulB);
	return true;
}

bool CNavitelSkinParser::ParseOutlineColor (const char * & _p, COLORREF & _col) {
	if (m_bV11) {
		if (! ParseColor (_p, _col))
			return false;
	} else
		_col = c_colWhite;

	return true;
}

////////////////////////////////////////////

const CNavitelSkinParser::kv_pair_t g_kvpPenStyles [] = {
	{"none",  -1},
	{"solid", PS_SOLID},
	{"dash",  PS_DOT},
	{"dot",   PS_DOT},
	{"null",  PS_NULL},
};

bool CNavitelSkinParser::ParsePenStyle (const char * & _p, short & _nStyle) {
	DWORD dwPenStyle = 0;
	if (_Key2Value (_p, & _p, g_kvpPenStyles, ARRAY_SIZE (g_kvpPenStyles), dwPenStyle)) {
		_nStyle = dwPenStyle;
		return true;
	}

	return false;
}

////////////////////////////////////////////

const CNavitelSkinParser::kv_pair_t g_kvpBools [] = {
	{"true",  true},
	{"false", false},
};

bool CNavitelSkinParser::ParseBool (const char * & _p, bool & _b) {
	DWORD dwBool = 0;
	if (_Key2Value (_p, & _p, g_kvpBools, ARRAY_SIZE (g_kvpBools), dwBool)) {
		_b = dwBool != 0;
		return true;
	}

	return false;
}

////////////////////////////////////////////

const CNavitelSkinParser::kv_pair_t g_kvpRFontSizes [] = {
	{"small",  0},
	{"medium", 0},
	{"big",    0}, // NOTE: reserved
};

bool CNavitelSkinParser::ParseRFontSize (const char * & _p) {
	DWORD dwPtr = 0;
	if (_Key2Value (_p, & _p, g_kvpRFontSizes, ARRAY_SIZE (g_kvpRFontSizes), dwPtr)) {
		return true;
	}

	return false;
}

////////////////////////////////////////////

const CNavitelSkinParser::kv_pair_t g_kvpFontSizes [] = {
	{"point",  0},
	{"small",  0},
	{"medium", 0},
	{"big",    0},
};

bool CNavitelSkinParser::ParseFontSize (const char * & _p) {
	DWORD dwFontPtr = 0;
	if (_Key2Value (_p, & _p, g_kvpFontSizes, ARRAY_SIZE (g_kvpFontSizes), dwFontPtr))
		return true;

	return false;
}

////////////////////////////////////////////

const char * CNavitelSkinParser::ParsePolygonDesc (const char * _pLine) {
	const char * const pNextLine = FindNextLine (_pLine);

	const char * p = _pLine;

	// Type begin.
	unsigned long ulType;
	if (! ParseHex (p, m_pEnd, & p, ulType))
		return NULL;

	// Type end.
	unsigned long ulType2;
	if (! ParseHex (p, m_pEnd, & p, ulType2))
		return NULL;

	// Fill pattern index.
	OmitSpaces (& p, m_pEnd);		
	bool bPatternSign = false;
	if (* p == '-') {
		bPatternSign = true;
		++ p;
	}
	unsigned long ulPattern;
	if (! ParseDecU (p, m_pEnd, & p, ulPattern))
		return NULL;
	if (bPatternSign) {
		if (ulPattern != 2 && ulPattern != 1)
			return NULL;
		ulPattern = - (int) ulPattern;
	} else {
		if (ulPattern > m_cPatternsW/m_cPatternsH)
			return NULL;
	}

	// Background color.
	COLORREF colBrush;
	if (! ParseColor (p, colBrush))
		return NULL;

	// Label color.
	COLORREF colText;
	if (! ParseColor (p, colText))
		return NULL;

	// Label outline color.
	COLORREF colOutline;
	if (! ParseOutlineColor (p, colOutline))
		return NULL;

	// Apply.
	if (ulType2 == 0)
		ulType2 = ulType;
	for (USHORT usType = ulType; usType <= ulType2; ++ usType) {
		type_desc_t * const pTD = const_cast<type_desc_t *> (g_tsNavitel.GetTypeDesc (usType, rkPolygon));
		if (pTD == NULL || pTD == m_pTD_Unknown)
			continue;

		if (ulPattern == -1) {
			const HBRUSH hbrush = ::CreateSolidBrush (colBrush);
			pTD->phBrush = & m_pSkin->AddBrush (hbrush);
		} else if (ulPattern == -2) {
			pTD->phBrush = & g_hbrNull;
		} else if (ulPattern >= 0 && ulPattern < m_cPatternsW/m_cPatternsH) {
			::BitBlt (* m_phdcMemPattern, 0, 0, m_cPatternsH, m_cPatternsH, * m_phdcPatterns, m_cPatternsH*ulPattern, 0, SRCCOPY);

			const HBRUSH hbrColor = ::CreateSolidBrush (colBrush);
			CGdiObjectHolder hbrColor_Holder (hbrColor);
			const RECT rect = {0, 0, m_cPatternsH, m_cPatternsH};
			::FillRect (* m_phdcMem, & rect, hbrColor);

			::TransparentBlt (* m_phdcMem, 0, 0, m_cPatternsH, m_cPatternsH, * m_phdcMemPattern, 0, 0, m_cPatternsH, m_cPatternsH, c_colWhite);

			const HBRUSH hbr = ::CreatePatternBrush (m_phdcMem->GetBitmap ());
			pTD->phBrush = & m_pSkin->AddBrush (hbr);
		}

		pTD->colText = colText;

		pTD->DestroySecondaryAttributes ();
	}

	return pNextLine;
}

bool ParseWidth (
	const char * _pValue, const char * _pEnd, const char ** _ppEnd, 
	unsigned long & _ulWidth, unsigned long & _ulWidthInM
) {
	_ulWidthInM = 0;

	unsigned long ulWidth = 0;
	if (! ParseDecU (_pValue, _pEnd, _ppEnd, ulWidth))
		return false;

	// Parse unit now. Nothing - means pixel, [mM] - means meters
	bool bHasUnit = false;
	const char * p = * _ppEnd;

	// Omit spaces.
	while (p < _pEnd && * p == ' ')
		++ p;

	if (p < _pEnd && (* p == 'm' || * p == 'M')) {
		++ p;
		bHasUnit = true;
	}

	if (bHasUnit) {  
		* _ppEnd = p; 
		_ulWidth = 1; // ulWidth is in meters.
		_ulWidthInM = ulWidth;
	} else {
		_ulWidth = ulWidth;
	}

	bool bHasSlash = false;
	// Omit spaces.
	while (p < _pEnd && * p == ' ')
		++ p;

	if (p < _pEnd && (* p == '/')) {
		++ p;
		bHasSlash = true;
	}

	if (bHasSlash) {
		* _ppEnd = p;
		unsigned long ulWidthInM = 0;
		return ParseWidth (p, _pEnd, _ppEnd, ulWidthInM, _ulWidthInM);
	}

	return true;
}

const char * CNavitelSkinParser::ParsePolylineDesc (const char * _pLine) {
	const char * const pNextLine = FindNextLine (_pLine);

	const char * p = _pLine;

	// Type begin.
	unsigned long ulType;
	if (! ParseHex (p, m_pEnd, & p, ulType))
		return NULL;

	// Type end.
	unsigned long ulType2;
	if (! ParseHex (p, m_pEnd, & p, ulType2))
		return NULL;

	// Pen 1 style.
	short nPenStyle1;
	if (! ParsePenStyle (p, nPenStyle1))
		return NULL;

	// Pen 1 width.
	unsigned long ulNormalWidth, ulNormalWidthInM;
	if (! ParseWidth (p, m_pEnd, & p, ulNormalWidth, ulNormalWidthInM))
		return NULL;

	// Pen 1 width (hi-res).
	unsigned long ulHiResWidth, ulHiResWidthInM;
	if (! ParseWidth (p, m_pEnd, & p, ulHiResWidth, ulHiResWidthInM))
		return NULL;

	// Pen 1 color.
	COLORREF colPen1;
	if (! ParseColor (p, colPen1))
		return NULL;

	// Pen 2 style.
	short nPenStyle2;
	if (! ParsePenStyle (p, nPenStyle2))
		return NULL;

	// Pen 2 width.
	unsigned long ulNormalWidth2, ulNormalWidth2InM;
	if (! ParseWidth (p, m_pEnd, & p, ulNormalWidth2, ulNormalWidth2InM))
		return NULL;

	// Pen 2 width (hi-res).
	unsigned long ulHiResWidth2, ulHiResWidth2InM;
	if (! ParseWidth (p, m_pEnd, & p, ulHiResWidth2, ulHiResWidth2InM))
		return NULL;

	// Pen 2 color.
	COLORREF colPen2;
	if (! ParseColor (p, colPen2))
		return NULL;

	// Label font.
	if (! ParseRFontSize (p))
		return NULL;

	// Label color.
	COLORREF colText;
	if (! ParseColor (p, colText))
		return NULL;

	// Label outline color.
	COLORREF colOutline;
	if (! ParseOutlineColor (p, colOutline))
		return NULL;

	// Label v-center alignment.
	bool bVCenterAlignment;
	if (! ParseBool (p, bVCenterAlignment))
		return NULL;

	// Apply.
	if (ulType2 == 0)
		ulType2 = ulType;
	for (USHORT usType = ulType; usType <= ulType2; ++ usType) {
		type_desc_t * const pTD = const_cast<type_desc_t *> (g_tsNavitel.GetTypeDesc (usType, rkPolyline));
		if (pTD == NULL || pTD == m_pTD_Unknown)
			continue;

		if (nPenStyle1 == (short) -1)
			pTD->phPen = & g_hpenNull;
		else {
			const HPEN hpen = ::CreatePen (nPenStyle1, ulNormalWidth, colPen1);
			pTD->phPen = & m_pSkin->AddPen (hpen);
		}

		if (nPenStyle2 == (short) -1)
			pTD->phPen2 = NULL;
		else {
			const HPEN hpen2 = ::CreatePen (nPenStyle2, ulNormalWidth2, colPen2);
			pTD->phPen2 = & m_pSkin->AddPen (hpen2);
		}

		pTD->nWidthInM = __max (ulNormalWidthInM, ulNormalWidth2InM);

		pTD->colText = colText;

		pTD->DestroySecondaryAttributes ();
	}

	return pNextLine;
}

const char * CNavitelSkinParser::ParsePointDesc (const char * _pLine) {
	const char * const pNextLine = FindNextLine (_pLine);

	const char * p = _pLine;

	// Type begin.
	unsigned long ulType;
	if (! ParseHex (p, m_pEnd, & p, ulType))
		return NULL;

	// Type end.
	unsigned long ulType2;
	if (! ParseHex (p, m_pEnd, & p, ulType2))
		return NULL;

	// Icon index.
	OmitSpaces (& p, m_pEnd);
	bool bIconSign = false;
	if (* p == '-') {
		bIconSign = true;
		++ p;
	}
	unsigned long ulIconIdx;
	if (! ParseDecU (p, m_pEnd, & p, ulIconIdx))
		return NULL;
	if (bIconSign) {
		if (ulIconIdx != 1)
			return NULL;
		ulIconIdx = -1;
	} else {
		if (ulIconIdx >= m_cPOIsW/m_cPOIsH)
			return NULL;
	}

	// Label font.
	if (! ParseFontSize (p))
		return NULL;

	// Label color.
	COLORREF colText;
	if (! ParseColor (p, colText))
		return NULL;

	// Label outline color.
	COLORREF colOutline;
	if (! ParseOutlineColor (p, colOutline))
		return NULL;

	// Label alignment.
	bool bCenterAlignment;
	if (! ParseBool (p, bCenterAlignment))
		bCenterAlignment = false;

	// Apply.
	if (ulType2 == 0)
		ulType2 = ulType;
	for (USHORT usType = ulType; usType <= ulType2; ++ usType) {
		type_desc_t * const pTD = const_cast<type_desc_t *> (g_tsNavitel.GetTypeDesc (usType, rkPoint));
		if (pTD == NULL || pTD == m_pTD_Unknown)
			continue;

//		pTD->phCursor = NULL;
		if (ulIconIdx == -1) {
			pTD->cIconIdx = -1;
		} else {
			pTD->cIconIdx = ulIconIdx;
		}

		pTD->DestroySecondaryAttributes ();
	}

	return pNextLine;
}

const char * CNavitelSkinParser::ParseImageList (const char * _pLine, HBITMAP * _phBmp, size_t * _pcWidth, size_t * _pcHeight) {
	const char * const pNextLine = FindNextLine (_pLine);

	if (_phBmp == NULL)
		return pNextLine;

	// Parse image file name.
	const size_t cImageFileNameLen = GetLineLen (_pLine, ::FindNextLine (_pLine, m_pEnd));
	char * const strImageFileName = (char *) alloca (sizeof (char)*(cImageFileNameLen + 1));
	::strncpy (strImageFileName, _pLine, cImageFileNameLen);
	strImageFileName [cImageFileNameLen] = '\0';

	// Find subfile data inside .NS2
	size_t cFileSize;
	const BYTE * const pData = m_nsf.GetFile (strImageFileName, cFileSize);
	if (pData == NULL) {
		ReportWarning ("Sub-file '%s' is not found.", strImageFileName);
		return NULL;
	}

	// Load PNG from the data.
	* _phBmp = LoadPng (pData, cFileSize);
	if (* _phBmp == NULL)
		return NULL;

	BITMAP bmp;
	::GetObject (* _phBmp, sizeof (bmp), & bmp);

	if (_pcWidth)
		* _pcWidth  = bmp.bmWidth;
	if (_pcHeight)
		* _pcHeight = bmp.bmHeight;

	return pNextLine;
}

# pragma const_seg (".text")

const char c_strSkinSignature_1_0 [] = "Navitel Skin Version 1.0";
const char c_strSkinSignature_1_1 [] = "Navitel Skin Version 1.1";
const size_t c_cSkinSignatureLen_1_0 = ARRAY_SIZE (c_strSkinSignature_1_0) - 1;
const size_t c_cSkinSignatureLen_1_1 = ARRAY_SIZE (c_strSkinSignature_1_1) - 1;

const CNavitelSkinParser::kv_pair_t g_kvpKeys [] = {
	{"BackgroundColor", spcBgColor},
	{"TextColor",       spcTextColor},
	{"OutlineColor",    spcOutlineColor},
	{"HWSign1Text",     spcHWSign1Text},
	{"HWSign1Bg",       spcHWSign1Bg},
	{"HWSign2Text",     spcHWSign2Text},
	{"HWSign2Bg",       spcHWSign2Bg},
	{"HWSign3Text",     spcHWSign3Text},
	{"HWSign3Bg",       spcHWSign3Bg},
};

# pragma const_seg ()

bool CNavitelSkinParser::Parse (CMapSkin * _pSkin) {
	assert (_pSkin);
	m_pSkin = _pSkin;

	//
	// Parse header.
	//

	// Signature.
	const char * pLine = m_pStart;
	if (pLine + c_cSkinSignatureLen_1_1 >= m_pEnd)
		return false;
	if (::strncmp (pLine, c_strSkinSignature_1_0, c_cSkinSignatureLen_1_0) == 0)
		m_bV11 = false;
	else {
		if (::strncmp (pLine, c_strSkinSignature_1_1, c_cSkinSignatureLen_1_1) != 0)
			return false;
	}
	pLine = FindNextLine (pLine);

	// Global settings.
	m_pSkin->SetColor (spcBgColor,      RGB (0xFF, 0xFF, 0xDD));
	m_pSkin->SetColor (spcTextColor,    c_colBlack);
	m_pSkin->SetColor (spcOutlineColor, c_colWhite);

	m_pSkin->SetColor (spcHWSign1Text,  c_colBlack);
	m_pSkin->SetColor (spcHWSign1Bg,    c_colWhite);
	m_pSkin->SetColor (spcHWSign2Text,  c_colBlack);
	m_pSkin->SetColor (spcHWSign2Bg,    RGB (0xFF, 0xDD, 0x3E));
	m_pSkin->SetColor (spcHWSign3Text,  c_colBlack);
	m_pSkin->SetColor (spcHWSign3Bg,    RGB (0xF7, 0x59, 0x51));

	if (m_bV11) {
		do {
			if (::strncmp (pLine, "polygons", 8) == 0) {
				pLine += 8;
				break;
			}

			DWORD dwKey = 0;
			if (_Key2Value (pLine, & pLine, g_kvpKeys, ARRAY_SIZE (g_kvpKeys), dwKey)) {
				COLORREF col;
				if (ParseColor (pLine, col)) {
					if (dwKey < spcLast)
						m_pSkin->SetColor (static_cast<CMapColor> (dwKey), col);
				} else
					return false;

				unsigned long width;
				unsigned long width2;
				unsigned long width_in_m;
				if (ParseWidth (pLine, m_pEnd, & pLine, width, width_in_m)) {
					if (! ParseWidth (pLine, m_pEnd, & pLine, width2, width_in_m)) 
						width2 = width;
					if (dwKey < spcLast) {
//						widthSetting  [dwKey] = width;
//						widthSetting2 [dwKey] = width2;
					}
				}
			}

			if (::strncmp (pLine, "DlgBtnOldStyle ", 15) == 0) {
				const char * p = pLine + 15;
				bool bDlgBtnOldStyle = false;
				if (! ParseBool (p, bDlgBtnOldStyle))
					return false;
			}

			pLine = FindNextLine (pLine);
		} while (pLine < m_pEnd);
	}

	//
	// Parse polygons.
	//

	// Parse color for unknown polygons (in 'polygons' line).
	COLORREF colUnknownPolygon;
	if (! ParseColor (pLine, colUnknownPolygon))
		colUnknownPolygon = c_colWhite;
	pLine = FindNextLine (pLine);

	// Polygon patterns image list for normal resolution.
	pLine = ParseImageList (pLine, & m_hbmpPatterns, & m_cPatternsW, & m_cPatternsH);
	if (pLine == NULL)
		return false;

	// Polygon patterns image list for VGA resolution.
	pLine = ParseImageList (pLine, NULL, NULL, NULL);
	if (pLine == NULL)
		return false;

	CClientDC hdcScreen (NULL);

	CDCHolder hdcPatterns = ::CreateCompatibleDC (hdcScreen);
	m_phdcPatterns = & hdcPatterns;
	const HGDIOBJ hPrevBmp = ::SelectObject (hdcPatterns, m_hbmpPatterns);

	CMemDC hdcMemPattern (hdcScreen, m_cPatternsH, m_cPatternsH);
	CMemDC hdcMem        (hdcScreen, m_cPatternsH, m_cPatternsH);
	m_phdcMemPattern = & hdcMemPattern;
	m_phdcMem        = & hdcMem;

	do {
		pLine = ParsePolygonDesc (pLine);
		if (pLine == NULL)
			return false;
		if (::strncmp (pLine, "polylines", 9) == 0) {
			pLine += 9;
			break;
		}
	} while (pLine < m_pEnd);

	::SelectObject (hdcPatterns, hPrevBmp);

	//
	// Parse polylines.
	//

	// Parse color for unknown polylines (in 'polylines' line).
	COLORREF colUnknownPolyline;
	if (! ParseColor (pLine, colUnknownPolyline))
		colUnknownPolyline = c_colLtGray;
	pLine = FindNextLine (pLine);

	do {
		pLine = ParsePolylineDesc (pLine);
		if (pLine == NULL)
			return false;
		if (::strncmp (pLine, "points", 6) == 0)
			break;
	} while (pLine < m_pEnd);

	//
	// Parse POIs.
	//

	// Parse icon index for unknown types (in 'points' line).
	unsigned long ulUnknownIconIdx = -1;
	if (! ParseDecU (pLine + 6, m_pEnd, & pLine, ulUnknownIconIdx))
		ulUnknownIconIdx = -1;
	pLine = FindNextLine (pLine);

	// POIs image list for normal resolution.
	HBITMAP hbmpPOIs = NULL;
	pLine = ParseImageList (pLine, & hbmpPOIs, & m_cPOIsW, & m_cPOIsH);
	if (pLine == NULL)
		return false;
	m_pSkin->SetPOIs (hbmpPOIs, m_cPOIsW, m_cPOIsH);

	// POIs image list for VGA resolution.
	pLine = ParseImageList (pLine, NULL, NULL, NULL);
	if (pLine == NULL)
		return false;

	do {
		pLine = ParsePointDesc (pLine);
		if (pLine == NULL)
			return false;
	} while (pLine < m_pEnd);

	return true;
}

///////////////////////////////////////////////////////////

bool ParseSkin_NS2 (const char * _strFileName, const char * _strMode, CMapSkin * _pMapSkin) {
	const CNSFile nsf (_strFileName);
	if (! nsf.IsValid ())
		return false;

	char strSkinFile [MAX_PATH];
	::sprintf (strSkinFile, "%s.skin", _strMode);

	size_t cSize = 0;
	const BYTE * const pData = nsf.GetFile (strSkinFile, cSize);
	if (pData == NULL)
		ReportWarning ("Sub-file '%s' is not found in %s", strSkinFile, _strFileName);

	CNavitelSkinParser parser (nsf, pData, cSize);
	if (! parser.Parse (_pMapSkin)) {
		ReportWarning ("Syntax error at line %d of sub-file '%s'", parser.GetLine (), strSkinFile);
		return false;
	}

	return true;
}

DECLARE_MAP_SKIN (ParseSkin_NS2, "Navitel", "ns2", "Navitel skin (*.ns2)");
