//
// GPSMapEdit
// (c) Konstantin Galichsky (kg@geopainting.com), 2009
//
// Navitel Speedcams export (from bookmarks).
//

# include "StdAfx.h"
# include "Save.h"
# include "Map.h"
# include "Bookmark.h"
# include "Status.h"
# include "Log.h"
# include "SaveFileDlg.h"
# include "FileWriter.h"
# include "fast_floor.h"
# include "Globals.h"

class CNNSpeedcamsExport : public CMapSaver {
  public:
	virtual bool CanSave (const map_t & _map);
	virtual SaveStatus_t Save (HANDLE _hFile, const char * _strFileName, const map_t & _map);

  private:
	SaveStatus_t _Save (HANDLE _hFile, const map_t & _map);
};

DECLARE_BM_EXPORT (CNNSpeedcamsExport, "navitel-speedcams", "txt", "Navitel Speedcams (*.txt)");

bool CNNSpeedcamsExport::CanSave (const map_t & _map) {
	return true;//CheckBookmarks_Silent (_map);
}

SaveStatus_t CNNSpeedcamsExport::Save (HANDLE _hFile, const char * _strFileName, const map_t & _map) {
	if (_hFile != NULL) {
		return _Save (_hFile, _map);
	} else {
		//
		// Prepare default file name.
		//
		char strFileName [MAX_PATH + 1];
		::strncpy (strFileName, _strFileName, MAX_PATH);
		strFileName [MAX_PATH] = '\0';
		if (! SaveFileDlg ("txt", "Navitel Speedcams (*.txt)\0*.txt\0", strFileName))
			return ssCanceled;

		//
		// Open the file.
		//
		const CFileForWrite hFile (strFileName);
		if (! hFile.IsValid ())
			return Error ("Can't create file:\n %s", strFileName);

		return _Save (hFile, _map);
	}
}

SaveStatus_t CNNSpeedcamsExport::_Save (HANDLE _hFile, const map_t & _map) {
	//
	// Check if speedcams are presented.
	//
	bool bSpeedcams = false;
	const layer_t & point_layer = _map.layers [rkPoint];
	const size_t cPoints = point_layer.cRgns;
	const rgns_t & points = point_layer.rgns;
	rgns_t::const_iterator iRgn;
	for (iRgn = points.begin (); iRgn != points.end (); ++ iRgn) {
		const rgn_t & rgn = * iRgn;
		if (rgn.IsImmutable () || ! rgn.IsBookmark ())
			continue;

		const bookmark_t & bm = * rgn.pExtra->pBookmark;
		if (bm.strType == "nn_speedcam") {
			bSpeedcams = true;
			break;
		}
	}

	if (! bSpeedcams)
		return Error ("Nothing to export: bookmarks with type 'Navitel Speedcam' are not found.");

	SetProgressStatus (0);

	//
	// Write the file.
	//
	CFileWriter fw (_hFile);
	fw.SetFPrecision (6);
	fw.WriteS ("IDX,X,Y,TYPE,SPEED,DirType,Direction\r\n");

	size_t cPoint = 0;
	for (iRgn = points.begin (); iRgn != points.end (); ++ iRgn, ++ cPoint) {
		SetProgressStatus (fast_floor (100.f*cPoint/cPoints + .5f));

		const rgn_t & rgn = * iRgn;
		if (rgn.IsImmutable () || ! rgn.IsBookmark ())
			continue;

		const bookmark_t & bm = * rgn.pExtra->pBookmark;
		if (! (bm.strType == "nn_speedcam"))
			continue;

		bookmark_t::k2v_t::const_iterator ia = bm.attributes.find ("speed");
		string_t strSpeed;
		if (ia == bm.attributes.end ())
			strSpeed = "0";
		else
			strSpeed = ia->second;

		ia = bm.attributes.find ("directions");
		string_t strDirType;
		if (ia == bm.attributes.end ())
			strDirType = "0";
		else
			strDirType = ia->second;

		ia = bm.attributes.find ("comment");
		string_t strComment;
		if (ia != bm.attributes.end ())
			strComment = ia->second;

		fw.WriteU (cPoint);
		fw.WriteC (',');
		fw.WriteF (bm.ptPosFixed.x);
		fw.WriteC (',');
		fw.WriteF (bm.ptPosFixed.y);
		fw.WriteC (',');
		fw.WriteS (bm.strSubType);
		fw.WriteC (',');
		fw.WriteS (strSpeed);
		fw.WriteC (',');
		fw.WriteS (strDirType);
		fw.WriteC (',');
		fw.WriteU (bm.bHasHeading ? fast_floor (bm.fHeading + .5f) : 0);

		if (! strComment.empty ()) {
			fw.WriteC (' ');
			fw.WriteS (strComment);
		}

		fw.WriteS ("\r\n");
	}

	fw.FlushBuffer ();

	SetProgressStatus (100);

	if (fw.GetOverflow ())
		return Error ("File has been written partialy.\nThis may be caused by disk overflow.");

	return ssOK;
}