//
// GPSMapEdit
// (c) Konstantin Galichsky (kg@geopainting.com), 2009
//
// Navitel Speedcams import (to bookmarks).
//

# include "StdAfx.h"
# include "Resource.h"
# include "Map.h"
# include "Bookmark.h"
# include "Load.h"
# include "Log.h"
# include "Status.h"
# include "ParseHelpers.h"
# include "fast_floor.h"
# include "Wrappers.h"
# include "Globals.h"

static
const char * __fastcall _FindNextLine (const char * _pBegin, const char * _pEnd) {
	assert (_pBegin <= _pEnd);

	for (const char * p = _pBegin; p < _pEnd; ++ p) {
		if (* p == '\r')
			return * (p + 1) == '\n' ? p + 2 : p + 1;
		if (* p == '\n')
			return p + 1;
	}

	return _pEnd;
}

///////////////////////////////////////////////////////////////

class CNNSpeedcamsImport : public CMapLoader {
  public:
	// Override.
	bool IsValid (const char * _pData, size_t _cFileSize) const;
	virtual LoadStatus_t Load (const char * _pData, size_t _cFileSize, const char * _strFileName, map_t * _pMap);
};

DECLARE_MAP_IMPORT (CNNSpeedcamsImport, "navitel-speedcams", "txt", "Navitel Speedcams (*.txt) to Bookmarks");

bool CNNSpeedcamsImport::IsValid (const char * _pData, size_t _cFileSize) const {
	if (! HasPrefix (_pData, _pData + _cFileSize, "IDX", 3))
		return false;
	return true;
}

LoadStatus_t CNNSpeedcamsImport::Load (const char * _pData, size_t _cFileSize, const char * _strFileName, map_t * _pMap) {
	const char * const pEnd = _pData + _cFileSize;

	_pMap->file_refs.push_back (file_ref_t ());
	file_ref_t * const pFileRef = & _pMap->file_refs.back ();
	pFileRef->strFileName = _strFileName;
	pFileRef->bAttachment = false;

	size_t cLine = 0;

	// Parse lines.
	const char * pLine = _pData;
	do {
		const char * const pNextLine = _FindNextLine (pLine, pEnd);
		if (cLine != 0 && (pLine < pEnd && pLine [0] != '\n' && pLine [0] != '\r')) {
			size_t cField = 0;

			bookmark_t bm;
			bm.strType = "nn_speedcam";
			bm.ptPosFixed = point_t (0, 0);

			// Parse fields.
			const char * p = pLine;
			do {
				OmitSpaces (& p, pNextLine);
				const char * pField = p;

				// Search for delimiter.
				const char * pFieldEnd = pField;
				while (p < pNextLine) {
					if (* p == ',')
						break;
					if (* p != ' ')
						pFieldEnd = p + 1;
					++ p;
				}
				if (pFieldEnd > pEnd)
					pFieldEnd = pEnd;

				const char * pNumberEnd = NULL;
				float fVal = 0;
				if (! ParseReal (pField, pFieldEnd, & pNumberEnd, fVal))
					ReportWarning (pLine - _pData, "Failed to parse %d-th field in line #%d.", cField + 1, cLine + 1);
				else {
					switch (cField) {
							// IDX.
						case 0: break;

							// Position.
						case 1: bm.ptPosFixed.x = fVal; break;
						case 2: bm.ptPosFixed.y = fVal; break;
							
						case 3:
							// Type.
							bm.strSubType               .assign (pField, pFieldEnd);
							break;
							
						case 4:
							// Speed.
							bm.attributes ["speed"]     .assign (pField, pFieldEnd);
							break;

						case 5:
							// DirType.
							bm.attributes ["directions"].assign (pField, pFieldEnd);
							if (pField != pFieldEnd && pField [0] != '0')
								bm.bHasHeading = true;
							break;

						case 6:
							// Heading.
							bm.fHeading = fVal;
							if (bm.fHeading < 0)
								bm.fHeading += 360;
							if (bm.fHeading >= 360)
								bm.fHeading -= 360;

							// Comment.
							if (pNumberEnd < pFieldEnd) {
								OmitSpaces (& pNumberEnd, pFieldEnd);
								bm.attributes ["comment"].assign (pNumberEnd, GetLineLen (pNumberEnd, pFieldEnd)); break;
							}
							break;
					}
				}

				if (p >= pNextLine)
					break;

				// Omit the delimiter.
				if (* p == ',')
					++ p;
				++ cField;
			} while (p < pNextLine && cField <= 6);

			if (cField < 4) {
				ReportWarning (pLine - _pData, "Not enough fields in the line #%d. This line will be ignored.", cLine + 1);
			} else {
				// Create bookmark.
				rgn_t & rgn = _pMap->CreateRgn (rkPoint);
				rgn.SetType (_pMap->pTypeSet, 0, rkPoint);
				rgn.cBeginPos = pLine - _pData;
				rgn.cFileLine = cLine;
			//	rgn.cOrdinal  = m_cOrdinal ++;
				rgn.pFileRef  = pFileRef;

				rgn.origins.push_back (rgn_origin_t ());
				rgn_origin_t & origin0 = rgn.origins.back ();
				origin0.point       = bm.ptPosFixed;
				origin0.cLevel      = 0;
				origin0.cFileOffset = pLine - _pData;

				rgn.pExtra->pBookmark = new bookmark_t;
				* rgn.pExtra->pBookmark = bm;

				g_bHaveBookmarks = true;
			}
		}
		++ cLine;		
		pLine = pNextLine;
	} while (pLine < pEnd);

	return lsOK;
}