//
// GPSMapEdit
// (c) Konstantin Galichsky (kg@geopainting.com), 2009
//
// 'List of Notes' dialog.
//

# include "StdAfx.h"
# include "Resource.h"
# include "NotesList.h"
# include "NoteDrawing.h"
# include "Map.h"
# include "MapView.h"
# include "NewMap.h"
# include "UndoRedo.h"
# include "BackupAction.h"
# include "ShowRgnOnMap.h"
# include "Scale.h"
# include "Position.h"
# include "Selection.h"
# include "DeleteObject.h"
# include "CombinedAction.h"
# include "Dialog.h"
# include "ListView.h"
# include "ParseHelpers.h"
# include "PrintUnits.h"
# include "fast_floor.h"
# include "Globals.h"

enum {
	cndColor      = 0,
	cndWidth      = 1,
	cndProcessed  = 2,
	cndPosition   = 3,
};

class CNoteDrawings : public CDialog {
  public:
	CNoteDrawings () : CDialog (IDD_BOOKMARKS_LIST) {}

	static CNoteDrawings * g_pNoteDrawings;

	void Reset ();

	// Override.
	virtual bool OnInit (HWND _hwndFocus);
	virtual void OnShowModeless () {Reset ();}
	virtual void OnDestroy ();
	virtual void OnCommand (WORD _wCmd, WORD _wNotifyCode, HWND _hwndCtl);
	virtual bool OnNotify (int _idCtrl, NMHDR * _pInfo);

  private:
	CListView m_List;
	void _FillNoteDrawings ();
	rgn_t * _GetNoteDrawing (size_t _cItem = -1);
	void _UpdateButtons  (size_t _cItem = -1);
	void _UpdateTotal ();

	void _GetSubItemText (const rgn_t & _rgn, const note_drawing_t & _nd, size_t _cCol, char * _strBuf, size_t _cBufSize) const;

	typedef std::vector<rgn_t *> vecRgns_t;
	vecRgns_t m_vecRgns;

	struct LessNoteDrawings {
		LessNoteDrawings (const CNoteDrawings * _pNoteDrawings, bool _bAscending, size_t _cCol) :
			m_pNoteDrawings (_pNoteDrawings),
			m_bAscending (_bAscending),
			m_cCol (_cCol)
		{}

		bool operator () (const rgn_t * _pRgn1, const rgn_t * _pRgn2);

	  private:
		const CNoteDrawings * const m_pNoteDrawings;
		const bool m_bAscending;
		const size_t m_cCol;
	};
	friend struct LessNoteDrawings;
};

CNoteDrawings * CNoteDrawings::g_pNoteDrawings;

bool CNoteDrawings::LessNoteDrawings::operator () (const rgn_t * _pRgn1, const rgn_t * _pRgn2) {
	assert (_pRgn1 && _pRgn1->IsNote ());
	assert (_pRgn2 && _pRgn2->IsNote ());
	const note_drawing_t & nd1 = * _pRgn1->pExtra->pNoteDrawing;
	const note_drawing_t & nd2 = * _pRgn2->pExtra->pNoteDrawing;

	float fCmp = 0;
	switch (m_cCol) {
		case cndColor:     fCmp = nd1.col - nd2.col; break;
		case cndWidth:     fCmp = nd1.cWidth - nd2.cWidth; break;
		case cndProcessed: fCmp = int (nd1.bProcessed) - int (nd2.bProcessed); break;
		case cndPosition: {
			point_t pt1 (_pRgn1->rectBound.MeanX (), _pRgn1->rectBound.MeanY ());
			point_t pt2 (_pRgn2->rectBound.MeanX (), _pRgn2->rectBound.MeanY ());
			fCmp = pt1.y == pt2.y ? pt1.x - pt2.x : pt1.y - pt2.y;
			break;
		}

		default: {
			char strText1 [64], strText2 [64];
			m_pNoteDrawings->_GetSubItemText (* _pRgn1, nd1, m_cCol, strText1, ARRAY_SIZE (strText1));
			m_pNoteDrawings->_GetSubItemText (* _pRgn2, nd2, m_cCol, strText2, ARRAY_SIZE (strText2));
			fCmp = ::lstrcmp (strText1, strText2);
		}
	}

	return m_bAscending ? fCmp < 0 : fCmp > 0;
}

void CNoteDrawings::Reset () {
	_FillNoteDrawings ();
	_UpdateButtons ();
}

void CNoteDrawings::_FillNoteDrawings () {
	CCritSec::CLock lock (g_cs);

	CWaiter w;

	rgn_t * const pRgnSel = _GetNoteDrawing ();

	// Save selection status for each item.
	typedef std::set<rgn_t *> setRgns_t;
	setRgns_t setRgnsSelected;
	const size_t cItemsOld = ListView_GetItemCount (m_List);
	size_t cItem;
	for (cItem = 0; cItem < cItemsOld; ++ cItem) {
		if (! IsItemSelected (m_List, cItem))
			continue;

		rgn_t * const pRgn = _GetNoteDrawing (cItem);
		if (! pRgn || pRgn->IsImmutable ())
			continue;

		setRgnsSelected.insert (pRgn);
	}

	// Update notes list.
	m_vecRgns.clear ();

	layer_t & layer = g_map.layers [rkPolyline];
	rgns_t & rgns = layer.rgns;
	rgns_t::iterator iRgn;
	for (iRgn = rgns.begin (); iRgn != rgns.end (); ++ iRgn) {
		rgn_t & rgn = * iRgn;
		if (rgn.IsImmutable () || ! rgn.IsNote ())
			continue;

		m_vecRgns.push_back (& rgn);
	}

	// Sort items.
	if (m_List.GetSortingColumn () != -1)
		std::sort (m_vecRgns.begin (), m_vecRgns.end (), LessNoteDrawings (this, m_List.IsSortingAscending (), m_List.GetSortingColumn ()));

	// Fill the list view.
	const size_t cRgns = m_vecRgns.size ();
	ListView_SetItemCount (m_List, cRgns);

	// Restore selection.
	for (cItem = 0; cItem < cRgns; ++ cItem) {
		rgn_t * const pRgn = m_vecRgns [cItem];
		SelectItem (m_List, cItem, setRgnsSelected.find (pRgn) != setRgnsSelected.end ());

		if (pRgn == pRgnSel) {
			SetSelection           (m_List, cItem);
			ListView_EnsureVisible (m_List, cItem, FALSE);
		}
	}

	_UpdateTotal ();
}

rgn_t * CNoteDrawings::_GetNoteDrawing (size_t _cItem) {
	const size_t cSel = _cItem == -1 ? ListView_GetSelectionMark (m_List) : _cItem;
	if (cSel == -1)
		return NULL;

	assert (cSel < m_vecRgns.size ());
	if (cSel >= m_vecRgns.size ())
		return NULL;
	rgn_t * const pRgn = m_vecRgns [cSel];
	assert (pRgn);
	if (pRgn == NULL || pRgn->IsDeleted ())
		return NULL;

	return pRgn;
}

void CNoteDrawings::_UpdateButtons (size_t _cItem) {
	const size_t cSelItems = ListView_GetSelectedCount (m_List);
	EnableDlgItem (IDC_LOCATE, cSelItems == 1);
	EnableDlgItem (IDC_DELETE, cSelItems != 0);
	EnableDlgItem (IDC_EDIT,   cSelItems == 1);
}

void CNoteDrawings::_UpdateTotal () {
	const size_t cRgns = m_vecRgns.size ();
	SetDlgItemTextF (IDC_TOTAL, IDS_TOTAL, cRgns);
}

bool CNoteDrawings::OnInit (HWND _hwndFocus) {
	SetWindowText (m_hWnd, "List of Note Drawings");

	// Notes.
	m_List.Init (m_hWnd, IDC_LIST, "Notes", -1);
	m_List.SetExtStyle (LVS_EX_FULLROWSELECT | LVS_EX_HEADERDRAGDROP);

	AddColumn (m_List, cndColor,      "Color");
	AddColumn (m_List, cndWidth,      "Width");
	AddColumn (m_List, cndProcessed,  "Processed");
	AddColumn (m_List, cndPosition,   "Position");

	m_List.Sort ();

	_FillNoteDrawings ();
	SetSelection (m_List, 0);

	AutoSizeColumns (m_List);
	m_List.LoadHeaders ();

	// Buttons.
	_UpdateButtons ();

	return true;
}

void CNoteDrawings::OnDestroy () {
	m_List.SaveHeaders ();
	m_List.Destroy ();
}

void CNoteDrawings::OnCommand (WORD _wCmd, WORD _wNotifyCode, HWND _hwndCtl) {
	switch (_wCmd) {
		case IDC_EDIT: {
			const size_t cSelItems = ListView_GetSelectedCount (m_List);
			if (cSelItems <= 1) {
				rgn_t * const pRgn = _GetNoteDrawing ();
				if (pRgn) {
					if (ShowNoteProperties (m_hWnd, * pRgn))
						_FillNoteDrawings ();
				}
			}
			break;
		}

		case IDC_DELETE: {
			const size_t cSel = ListView_GetSelectionMark (m_List);

			const size_t cSelItems = ListView_GetSelectedCount (m_List);
			if (cSelItems <= 1) {
				rgn_t * const pRgn = _GetNoteDrawing ();
				if (pRgn)
					DeleteObject (pRgn);
			} else {
				const size_t cItems = ListView_GetItemCount (m_List);
				for (size_t cItem = 0; cItem < cItems; ++ cItem) {
					if (! IsItemSelected (m_List, cItem))
						continue;

					rgn_t * const pRgn = _GetNoteDrawing (cItem);
					if (! pRgn)
						continue;

					DeleteObject (pRgn);
				}
			}

			_FillNoteDrawings ();
			if (cSel != -1)
				SetSelection (m_List, cSel);

			_UpdateButtons ();

			Refresh ();

			return;
		}

		case IDC_LOCATE: {
			const size_t cSelItems = ListView_GetSelectedCount (m_List);
			if (cSelItems != 1)
				return;

			const rgn_t * const pRgn = _GetNoteDrawing ();
			if (pRgn)
				ShowRgnOnMap (* pRgn);

			return;
		}

		case IDC_REFRESH:
			Reset ();
			return;

		case IDCANCEL:
			Hide ();
			return;
	}
}

bool CNoteDrawings::OnNotify (int _idCtrl, NMHDR * _pInfo) {
	switch (_pInfo->code) {
		case NM_RCLICK:
			if (_idCtrl != IDC_LIST)
				break;
			m_List.PopupMenu_OnHeaderRClick (_pInfo);
			break;

		case NM_DBLCLK:
			if (_idCtrl != IDC_LIST)
				break;

			SendCommand (IDC_EDIT);
			break;

		case LVN_COLUMNCLICK:
			if (_idCtrl != IDC_LIST)
				break;

			m_List.Sort_OnColumnClick (_pInfo);
			_FillNoteDrawings ();

			break;

		case LVN_ITEMCHANGED: {
			if (_idCtrl != IDC_LIST)
				break;

			const NMLISTVIEW * const pInfo2 = reinterpret_cast<const NMLISTVIEW *> (_pInfo);
			if (pInfo2->uNewState & LVIS_SELECTED)
				_UpdateButtons (pInfo2->iItem);

			break;
		}

		case LVN_GETDISPINFO: {
			if (_idCtrl != IDC_LIST)
				break;

			const NMLVDISPINFO * const pInfo2 = reinterpret_cast<const NMLVDISPINFO *> (_pInfo);
			assert ((size_t) pInfo2->item.iItem < m_vecRgns.size ());
			if ((size_t) pInfo2->item.iItem >= m_vecRgns.size ())
				break;
			const rgn_t * const pRgn = m_vecRgns [pInfo2->item.iItem];
			if (pRgn == NULL)
				break;
			if (! pRgn->IsNote ())
				break;
			const note_drawing_t & nd = * pRgn->pExtra->pNoteDrawing;

			if (pInfo2->item.mask & LVIF_TEXT)
				_GetSubItemText (* pRgn, nd, pInfo2->item.iSubItem, pInfo2->item.pszText, pInfo2->item.cchTextMax);

			break;
		}
	}

	return false;
}

void CNoteDrawings::_GetSubItemText (const rgn_t & _rgn, const note_drawing_t & _nd, size_t _cCol, char * _strBuf, size_t _cBufSize) const {
	switch (_cCol) {
		default:
			assert (false);
			_strBuf [0] = '\0';
			break;

		case cndColor: ::sprintf (_strBuf, "%08x", _nd.col);    break;
		case cndWidth: ::sprintf (_strBuf, "%d",   _nd.cWidth); break;

		case cndProcessed:
			::_tcsncpy  (_strBuf, _nd.bProcessed ? "yes" : "no", _cBufSize);
			break;

		case cndPosition: {
			char strPos [32];
			PrintPosition (strPos, point_t (_rgn.rectBound.MeanX (), _rgn.rectBound.MeanY ()));
			::_tcsncpy  (_strBuf, strPos, _cBufSize);
			break;
		}
	}
}

static
void _ResetNoteDrawings () {
	if (CNoteDrawings::g_pNoteDrawings)
		CNoteDrawings::g_pNoteDrawings->Reset ();
}

INVOKE_AFTER_REMOVE_MAP (_ResetNoteDrawings);
INVOKE_ON_BACKUP_MAP    (_ResetNoteDrawings);
INVOKE_BEFORE_GC        (_ResetNoteDrawings);

void ShowNoteDrawings () {
	ShowModeless (CNoteDrawings::g_pNoteDrawings, g_hWnd);
}