//
// GPSMapEdit
// (c) Konstantin Galichsky (kg@geopainting.com), 2005-2008
//
// Map view overlapped paths.
//

# include "StdAfx.h"
# include "Paths.h"
# include "Cursors.h"
# include "Map.h"
# include "Position.h"
# include "Scale.h"
# include "DrawDirIndicator.h"
//# include "DrawHelpers.h"
# include "fast_floor.h"
# include "Wrappers.h"
# include "Globals.h"

static HFONT g_hfontPathArea;

void SetPathAreaLabelFont (const char * _strFont, int _nSize, bool _bBold, bool _bItalics) {
	if (g_hfontPathArea)
		::DeleteObject (g_hfontPathArea);

	LOGFONT lf = {
		_nSize, 0, 0, 0,
		_bBold ? FW_BOLD : FW_NORMAL, _bItalics ? TRUE : FALSE, FALSE, FALSE,
		DEFAULT_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS,
		DEFAULT_QUALITY,
		DEFAULT_PITCH | FF_DONTCARE
	};
	::strcpy (lf.lfFaceName, _strFont);
	g_hfontPathArea = ::CreateFontIndirect (& lf);
}

void DestroyPaths () {
	if (g_hfontPathArea)
		::DeleteObject (g_hfontPathArea);
	g_hfontPathArea = NULL;
}

///////////////////////////////////////////////////////

Paths_t g_Paths;

void DrawPaths (HDC _hDC, const rect_t & _rectVisible) {
	CCritSec::CLock lock (g_cs);

	if (g_Paths.empty ())
		return;

	// Get max points.
	size_t cMaxPoints = 0;
	Paths_t::const_iterator ip;
	for (ip = g_Paths.begin (); ip != g_Paths.end (); ++ ip) {
		const CPath_t & path = * ip;

		const size_t cPoints = path.points.size ();
		if (cMaxPoints < cPoints)
			cMaxPoints = cPoints;
	}

	if (cMaxPoints == 0)
		return;

	// Prepare points buffer.
	POINT * const pPoints = (POINT *) ::alloca (sizeof (POINT)*(cMaxPoints + 1));

	for (size_t cPass = 0; cPass < 2; ++ cPass) {
		const bool bFillArea = cPass == 0;

		for (ip = g_Paths.begin (); ip != g_Paths.end (); ++ ip) {
			const CPath_t & path = * ip;
			if (path.bHidden)
				continue;
			if (path.bFillArea != bFillArea)
				continue;
			if (! path.rectBound.Intersects (_rectVisible))
				continue;

			const points_t & points = path.points;
			const size_t cPoints = points.size ();
			if (cPoints == 0)
				continue;

			// Convert positions.
			for (size_t c = 0; c < cPoints; ++ c) {
				const point_t & point = points [c];

				assert (c < cMaxPoints);
				pPoints [c].x = XToScreen (point.x);
				pPoints [c].y = YToScreen (point.y);
			}

			// Draw area.
			if (path.bFillArea) {
				pPoints [cPoints] = pPoints [0];

				const HBRUSH hbr = ::CreateSolidBrush (path.colArea);
				const HBRUSH hbrPrev = (HBRUSH) ::SelectObject (_hDC, hbr);

				if (path.bTransparentArea)
					::SetROP2 (_hDC, R2_MASKPEN);

				::Polygon (_hDC, pPoints, cPoints);

				if (path.bTransparentArea)
					::SetROP2 (_hDC, R2_COPYPEN);

				::SelectObject (_hDC, hbrPrev);
				::DeleteObject (hbr);
			}

			// Draw vertices.
			if (path.bShowVertices) {
				const HPEN hpenVertices  = ::CreatePen (PS_SOLID, 1, path.col);
				const HBRUSH hbrVertices = ::CreateSolidBrush (path.col);

				const HPEN hpenPrev  = (HPEN)   ::SelectObject (_hDC, hpenVertices);
				const HBRUSH hbrPrev = (HBRUSH) ::SelectObject (_hDC, hbrVertices);

				for (size_t c = 0; c < cPoints; ++ c) {
					const point_t & point = points [c];
					if (_rectVisible.Contains (point))
						//DrawSquare (_hDC, pPoints [c].x, pPoints [c].y, path.cVertexSize);
						::Ellipse (
							_hDC,
							pPoints [c].x - path.cVertexSize,     pPoints [c].y - path.cVertexSize,
							pPoints [c].x + path.cVertexSize + 1, pPoints [c].y + path.cVertexSize + 1
						);
				}

				::SelectObject (_hDC, hpenPrev);
				::SelectObject (_hDC, hbrPrev);

				::DeleteObject (hpenVertices);
				::DeleteObject (hbrVertices);
			}

			const size_t cHSTo   = path.cHSTo   < cPoints ? path.cHSTo   : cPoints;
			const size_t cHSFrom = path.cHSFrom < cHSTo   ? path.cHSFrom : cHSTo;

			// Draw polyline except the highlighted segment.
			{
				const HPEN hpen = ::CreatePen (PS_SOLID, path.cWidth, path.col);
				const HPEN hpenPrev = (HPEN) ::SelectObject (_hDC, hpen);

				if (cHSFrom > 0)
					::Polyline (_hDC, pPoints, cHSFrom + 1);
				if (cHSTo < cPoints - 1)
					::Polyline (_hDC, pPoints + cHSTo, cPoints - cHSTo);

				if (path.bFillArea)
					::Polyline (_hDC, pPoints + cPoints - 1, 2);

				::SelectObject (_hDC, hpenPrev);
				::DeleteObject (hpen);
			}

			// Draw the highlighted segment.
			if (cHSFrom < cHSTo) {
				const HPEN hpen = ::CreatePen (PS_SOLID, path.cWidth, path.colHS);
				const HPEN hpenPrev = (HPEN) ::SelectObject (_hDC, hpen);

				::Polyline (_hDC, pPoints + cHSFrom, cHSTo - cHSFrom + 1);

				::SelectObject (_hDC, hpenPrev);
				::DeleteObject (hpen);
			}

			// Draw direction arrows.
			if (path.bShowDirection) {
				const RECT rectWnd = {0, 0, g_lWidth, g_lHeight};
				DrawDirIndicator (_hDC, pPoints, cPoints, & rectWnd);
			}

			// Draw label.
			if (path.bFillArea && ! path.strLabel.empty ()) {
				rect_t rect;
				rect_t::Intersection (rect, _rectVisible, path.rectBound);
				if (! rect.Empty ()) {
					const int x = XToScreen (rect.MeanX ());
					const int y = YToScreen (rect.MeanY ());

					::SetTextColor (_hDC, path.colLabel);
					::SetBkMode    (_hDC, TRANSPARENT);
					::SetTextAlign (_hDC, TA_CENTER | TA_BASELINE);
					::SelectObject (_hDC, g_hfontPathArea ? g_hfontPathArea : g_hFont);
					::TextOut (_hDC, x, y, path.strLabel.c_str (), path.strLabel.size ());
				}
			}
		}
	}
}

void DeletePath (CPath_t & _path) {
	CCritSec::CLock lock (g_cs);

	for (Paths_t::iterator ip = g_Paths.begin (); ip != g_Paths.end (); ++ ip) {
		CPath_t & path = * ip;
		if (& path == & _path) {
			g_Paths.erase (ip);
			return;
		}
	}
}

void GetPathInfotip (int _x, int _y, string_t & _strInfotip) {
	CCritSec::CLock lock (g_cs);

	if (g_Paths.empty ())
		return;

	// The precision of matching in degrees.
	const float c_fE = c_fE_in_pixels/g_kY;

	point_t pt;
	ScreenToPt (_x, _y, pt);

	const float fCosY = ::cos (pt.y*c_PI/180);

	// The rectangle will be used to match points.
	rect_t rectE;
	rectE.x0 = pt.x - c_fE/fCosY;
	rectE.x1 = pt.x + c_fE/fCosY;
	rectE.y0 = pt.y - c_fE;
	rectE.y1 = pt.y + c_fE;

	for (Paths_t::const_iterator ip = g_Paths.begin (); ip != g_Paths.end (); ++ ip) {
		const CPath_t & path = * ip;
		if (path.bHidden)
			continue;
		if (path.bFillArea)
			continue;
		if (path.bstrInfotip.Length () == 0 && path.bstrInfotipsForPoints.Length () == 0)
			continue;

		USES_CONVERSION;

		const size_t cPoints = path.points.size ();
		if (cPoints == 0)
			continue;
		else if (cPoints == 1) {
			if (PointContainsPt (path.points [0], pt, c_fE*2.5f, fCosY)) {
				if (path.bstrInfotipsForPoints) {
					if (! path.InfotipsForPoints.empty ()) {
						const size_t cOffset = path.InfotipsForPoints [0];
						if (cOffset < path.bstrInfotipsForPoints.Length ()) {
							_strInfotip = OLE2A (path.bstrInfotipsForPoints.m_str + path.InfotipsForPoints [0]);
							return;
						}
					}
				}
				_strInfotip = OLE2A (path.bstrInfotip);
				return;
			}
		} else {
			size_t cPoint = c_cUnknownPoint;
			point_t ptNearest;
			if (PolylineContainsPt (path.points, pt, c_fE*1.5f, fCosY, false, & ptNearest, & cPoint)) {
				if (path.bstrInfotipsForPoints) {
					cPoint = PointsContainPt (path.points, pt, c_fE*2.5f, fCosY);
					if (cPoint != -1 && cPoint < path.InfotipsForPoints.size ()) {
						const size_t cOffset = path.InfotipsForPoints [cPoint];
						if (cOffset < path.bstrInfotipsForPoints.Length ()) {
							_strInfotip = OLE2A (path.bstrInfotipsForPoints.m_str + path.InfotipsForPoints [cPoint]);
							return;
						}
					}
				}
				_strInfotip = OLE2A (path.bstrInfotip);
				return;
			}
		}
	}
}

void GetPathAreaInfotip (int _x, int _y, string_t & _strInfotip) {
	CCritSec::CLock lock (g_cs);

	if (g_Paths.empty ())
		return;

	// The precision of matching in degrees.
	const float c_fE = c_fE_in_pixels/g_kY;

	point_t pt;
	ScreenToPt (_x, _y, pt);

	const float fCosY = ::cos (pt.y*c_PI/180);

	for (Paths_t::const_iterator ip = g_Paths.begin (); ip != g_Paths.end (); ++ ip) {
		const CPath_t & path = * ip;
		if (path.bHidden)
			continue;
		if (! path.bFillArea)
			continue;
		if (path.bstrInfotip.Length () == 0)
			continue;
		if (path.points.size () < 2)
			continue;

		if (path.rectBound.Contains (pt) && PolygonContainsPt (path.points, pt) >= 0) {
			USES_CONVERSION;
			_strInfotip = OLE2A (path.bstrInfotip);
			return;
		}
	}
}
