//
// GPSMapEdit
// (c) Konstantin Galichsky (kg@geopainting.com), 2008-2009
// (c) Eugene Prosso (eprosso@navitel.su), 2009
//
// Verify polygons overlapping and self-intersection.
//

# include "StdAfx.h"
# include "PolygonIntersections.h"
# include "Map.h"
# include "VerifyMap.h"
# include "Types.h"
# include "PtAlgo.h"
# include "Status.h"
# include "fast_floor.h"
# include "Globals.h"

elem_refs_t * polygon_segment_t::m_pElemRefs = NULL;

void CPolygonIntersections::_CountPolygonVertices (const rgns_t & _polygons, size_t & _cPoints, size_t _cLevel) {
	_cPoints = 0;  

	for (rgns_t::const_iterator iPoly = _polygons.begin (); iPoly != _polygons.end (); ++ iPoly) { 
		const rgn_t & polygon = * iPoly;
		if (polygon.IsImmutable () || g_map.pTypeSet->pfIsBackground (polygon))
			continue;

		const size_t cElements = polygon.elements.size ();
		for (size_t c = 0; c < cElements; ++ c) {
			const rgn_element_t & elem = polygon.elements [c];
			if (elem.bDeleted)
				continue;

			const bool bLevel = elem.cLevel == _cLevel || (polygon.uchEndLevel != 0 && elem.cLevel <= _cLevel && _cLevel <= polygon.uchEndLevel);
			if (! bLevel)
				continue;

			_cPoints += elem.points.size ();
		}
	} 
} 

void CPolygonIntersections::_AddElementsToRefVector (
	const rgn_t & _polygon, const rgn_elements_t & _elements, size_t _cLevel,
	size_t & _cCurrElemRefIdx, size_t & _cCellsToAlloc, bool _bAllocOnly
) {
	const size_t cElements = _elements.size ();
	for (size_t c = 0; c < cElements; ++ c) {
		const rgn_element_t & elem = _elements [c];
		if (elem.bDeleted)
			continue;

		const bool bLevel = (elem.cLevel == _cLevel) || (_polygon.uchEndLevel != 0 && elem.cLevel <= _cLevel && _cLevel <= _polygon.uchEndLevel);
		if (! bLevel)
			continue;

		if (_bAllocOnly)
			++ _cCellsToAlloc;
		else { 
			const elem_ref_t elem_ref (_cCurrElemRefIdx, const_cast <rgn_t *> (& _polygon), c);
			m_elemRefs [_cCurrElemRefIdx ++] = elem_ref;
		}
	}
}

void CPolygonIntersections::_FormElemRefVector (const rgn_t & _polygon, const rgn_elements_t & _elements, bool _bAllocOnly, size_t _cLevel) { 
	size_t cToAlloc = 0;
	size_t elemRefIdx = 0;

	_AddElementsToRefVector (_polygon, _elements, _cLevel, elemRefIdx, cToAlloc, _bAllocOnly);

	if (_bAllocOnly)
		m_elemRefs.resize (cToAlloc);
} 

void CPolygonIntersections::_FormElemRefVector (const rgn_t & _polygon, const rgn_elements_t & _elements, size_t _cLevel) {
	_FormElemRefVector (_polygon, _elements, true,  _cLevel);
	_FormElemRefVector (_polygon, _elements, false, _cLevel);
} 

void CPolygonIntersections::_FormElemRefVector (const rgns_t & _polygons, bool _bAllocOnly, size_t _cLevel) { 
	size_t cToAlloc = 0;
	size_t elemRefIdx = 0;

	for (rgns_t::const_iterator iRgn = _polygons.begin (); iRgn != _polygons.end (); ++ iRgn) {
		const rgn_t & polygon = * iRgn;
		if (polygon.IsImmutable () || g_map.pTypeSet->pfIsBackground (polygon))
			continue;

		_AddElementsToRefVector (polygon, polygon.elements, _cLevel, elemRefIdx, cToAlloc, _bAllocOnly);
	}

	if (_bAllocOnly)
		m_elemRefs.resize (cToAlloc);
} 

void CPolygonIntersections::_FormElemRefVector (const rgns_t & _polygons, size_t _cLevel) {
	_FormElemRefVector (_polygons, true,  _cLevel);
	_FormElemRefVector (_polygons, false, _cLevel);
} 

void CPolygonIntersections::_AllocGridCells (float _fMapWidth, float _fMapHeight) {
	m_fCellHeight = _fMapHeight/m_lRows;
	m_fCellWidth  = _fMapWidth /m_lColumns;

	++ m_lRows;
	++ m_lColumns;

	m_ppCells = new grid_cell_t * [m_lRows + 1];
	
	for (long r = 0; r < m_lRows; ++ r) { 
		m_ppCells [r] = new grid_cell_t [m_lColumns];
		for (long c = 0; c < m_lColumns; ++ c) {
			grid_cell_t cell (m_rectBound.x0 + c*m_fCellWidth, m_rectBound.y0 + r*m_fCellHeight);
			m_ppCells [r] [c] = cell;
		}
	}
}

void CPolygonIntersections::_AllocDataStruct (intmap_t & _vecCounters, RGN_KIND _rgnKind) {
	int nMaxItemsPerCell = 0;
	intmap_t::const_iterator iIt;
	
	for (iIt = _vecCounters.begin (); iIt != _vecCounters.end (); ++ iIt)  
		if (iIt->first > nMaxItemsPerCell)
			nMaxItemsPerCell = iIt->first;

	if (_rgnKind == rkPoint) {
		m_cellSegmentVectors.resize (nMaxItemsPerCell + 1);
		for (int i = 1; i <= nMaxItemsPerCell; ++ i)
			m_cellSegmentVectors[i].resize (_vecCounters [i] * i);
	} else if (_rgnKind == rkPolygon) { 
		m_cellElemVectors.resize (nMaxItemsPerCell + 1);

		for (int i = 1; i <= nMaxItemsPerCell; ++ i)
			m_cellElemVectors [i].resize (_vecCounters [i]*i);
	}
} 

void CPolygonIntersections::_CoordToCellIndices (float _fX, float _fY, long & _lColumn, long & _lRow) { 
	_lColumn = (int) ((_fX - m_rectBound.x0)/m_fCellWidth);
	_lRow    = (int) ((_fY - m_rectBound.y0)/m_fCellHeight);
}

void CPolygonIntersections::_AddSegmentToCell (grid_cell_t & _cell, const polygon_segment_t & _segment, bool _bCountOnly) { 
	if (_bCountOnly) { 
		if (_cell.nSegments > 0)
			m_segmVecCounters [_cell.nSegments] --;
		_cell.nSegments ++;
		if (m_segmVecCounters.find (_cell.nSegments) == m_segmVecCounters.end ())
			m_segmVecCounters.insert (std::make_pair (_cell.nSegments, 1));
		else
			m_segmVecCounters [_cell.nSegments] ++;
	} else { 
		const size_t nSegments = _cell.nSegments;
		if (_cell.nSegmentIdx == -1) {
			_cell.nSegmentIdx = m_segmVecCounters [nSegments];
			m_segmVecCounters [nSegments] += nSegments;
		}

		m_cellSegmentVectors [nSegments] [_cell.nSegmentIdx ++] = _segment;
	}				
}

void CPolygonIntersections::_AddSegmentToCells (
	const size_t _eltRefIdx, const size_t _cIdx1, const size_t _cIdx2, bool _bCountOnly
) { 
	const polygon_segment_t segment (_eltRefIdx, _cIdx1, _cIdx2);
	long lStartRow = -1, lStartClmn = -1, lEndRow = -1, lEndClmn = -1;
	const rgn_element_t & elt = m_elemRefs [_eltRefIdx].Elem ();

	const point_t & pt1 = elt.points [_cIdx1];
	const point_t & pt2 = elt.points [_cIdx2];
	point_t ptStart, ptEnd;

	if (pt1.x < pt2.x) { 
		ptStart = pt1; 
		ptEnd = pt2;
	} else { 
		ptStart = pt2; 
		ptEnd = pt1;
	} 

	_CoordToCellIndices (ptStart.x, ptStart.y, lStartClmn, lStartRow);
	_CoordToCellIndices (ptEnd.x, ptEnd.y, lEndClmn, lEndRow);
	assert (
		lStartClmn >= 0 && lStartRow >= 0 && lEndClmn >= 0 && lEndRow >= 0 && 
		lStartClmn < m_lColumns && lStartRow < m_lRows && lEndClmn < m_lColumns && lEndRow < m_lRows
	);

	const grid_cell_t & start_cell = m_ppCells [lStartRow] [lStartClmn];
	const grid_cell_t & end_cell   = m_ppCells [lEndRow]   [lEndClmn];

	int nIncr = (lEndRow > lStartRow) ? 1 : -1;
	int nCrossingIncr = __max (0, nIncr);
	const float fXCrossing = 0, fYCrossing = 0;
	const float fSegmentHeight = ::fabs (ptEnd.y - ptStart.y), fSegmentLength = ptEnd.x - ptStart.x;
	const float fCtg = (fSegmentHeight != 0) ? fSegmentLength/fSegmentHeight : -1;

	long lCurColumn = lStartClmn, lNextColumn = lStartClmn;
	long cDummy;

	for (long r = lStartRow; r != lEndRow + nIncr; r += nIncr) { 
		if (lEndClmn > lStartClmn) { 
			if (r != lEndRow) { 
				const float fYCrossing = m_ppCells [r + nCrossingIncr][0].y0;
				const float fXCrossing = ptStart.x + fCtg*::fabs (fYCrossing - ptStart.y);

				_CoordToCellIndices (fXCrossing, 0, lNextColumn, cDummy);

				if (lNextColumn > lEndClmn)
					lNextColumn = lEndClmn;
				else if (fXCrossing == m_ppCells [0] [lNextColumn + 1].x0)
					lNextColumn ++;
			}
			else
				lNextColumn = lEndClmn;
		}

		for (long c = lCurColumn; c <= lNextColumn ; ++ c)
			_AddSegmentToCell (m_ppCells [r] [c], segment, _bCountOnly);

		lCurColumn = lNextColumn;

		if (lEndClmn > lStartClmn && fXCrossing == m_ppCells [0] [lNextColumn].x0)
			lCurColumn --;
	}

	const bool bAddPrevColumn = (ptStart.x == start_cell.x0 && lStartClmn > 0);
	const bool bAddNextColumn = (ptEnd.x == end_cell.x0 + m_fCellWidth && lEndClmn < m_lColumns - 1);
	const bool bAddPrevRow =
		(ptStart.y < ptEnd.y && ptStart.y == start_cell.y0 && lStartRow > 0)	||
		(ptStart.y > ptEnd.y && ptStart.y == start_cell.y0 + m_fCellHeight && lStartRow < m_lRows - 1);
	const bool bAddNextRow = 
		(ptEnd.y < ptStart.y && ptEnd.y == end_cell.y0 && lEndRow > 0)	||
		(ptEnd.y > ptStart.y && ptEnd.y == end_cell.y0 + m_fCellHeight && lEndRow < m_lRows - 1);

	if ((bAddPrevRow || bAddNextRow) && lStartRow == lEndRow)
		nIncr = (ptStart.y < ptEnd.y) ? 1 : -1;

	if (bAddPrevRow) {
		assert (lStartRow - nIncr <= m_lRows - 1);
		_AddSegmentToCell (m_ppCells [lStartRow - nIncr] [lStartClmn], segment, _bCountOnly);
		if (bAddPrevColumn)
			_AddSegmentToCell (m_ppCells [lStartRow - nIncr] [lStartClmn - 1], segment, _bCountOnly);
	} else if (bAddPrevColumn)
		_AddSegmentToCell (m_ppCells [lStartRow] [lStartClmn - 1], segment, _bCountOnly);

	if (bAddNextRow)  {
		assert (lEndRow + nIncr >= 0);
		_AddSegmentToCell (m_ppCells [lEndRow + nIncr] [lEndClmn], segment, _bCountOnly);
		if (bAddNextColumn)
			_AddSegmentToCell (m_ppCells [lEndRow + nIncr] [lEndClmn + 1], segment, _bCountOnly);
	}
	else if (bAddNextColumn)
		_AddSegmentToCell (m_ppCells [lEndRow] [lEndClmn + 1], segment, _bCountOnly);
} 

void CPolygonIntersections::_AddElemToCells (elem_ref_t & _elemRef, bool _bCountOnly) { 
	const rgn_element_t & elem = _elemRef.Elem ();
	long lMinClmn = -1, lMaxClmn = -1, lMinRow = -1, lMaxRow = -1;
	_CoordToCellIndices (elem.rectBound.x0, elem.rectBound.y0, lMinClmn, lMinRow);
	_CoordToCellIndices (elem.rectBound.x1, elem.rectBound.y1, lMaxClmn, lMaxRow);

	int nXIncr = (lMaxClmn > lMinClmn) ? 1 : -1;
	int nYIncr = (lMaxRow  > lMinRow)  ? 1 : -1;

	for (long r = lMinRow; (nYIncr > 0 && r <= lMaxRow) || (nYIncr < 0 && r >= lMaxRow); r += nYIncr) { 
		for (long c = lMinClmn; (nXIncr > 0 && c <= lMaxClmn) || (nXIncr < 0 && c >= lMaxClmn); c += nXIncr) { 
			if (_bCountOnly) { 
				if (m_elemVecCounters [m_ppCells [r][c].nElems] > 0)
					m_elemVecCounters [m_ppCells [r][c].nElems] --;
				m_ppCells [r][c].nElems ++;
				if (m_elemVecCounters.find (m_ppCells [r][c].nElems) == m_elemVecCounters.end ())
					m_elemVecCounters.insert (std::make_pair (m_ppCells [r][c].nElems, 1));
				else
					m_elemVecCounters [m_ppCells [r][c].nElems] ++;
			}
			else { 
				size_t nElems = m_ppCells [r][c].nElems;
				if (m_ppCells [r][c].nElemIdx == -1)	{
					m_ppCells [r][c].nElemIdx = m_elemVecCounters [nElems];
					m_elemVecCounters [nElems] += m_ppCells [r][c].nElems;
				}

				m_cellElemVectors [nElems] [m_ppCells [r][c].nElemIdx ++] = _elemRef;
			}				
		} 
	}
}

void CPolygonIntersections::_PopulateGrid (RGN_KIND _rgnKind, bool _bCountOnly) { 
	if (! _bCountOnly) { 
		for (intmap_t::iterator iSegmNum = m_segmVecCounters.begin (); iSegmNum != m_segmVecCounters.end (); ++ iSegmNum) 
			iSegmNum->second = 0;
		for (intmap_t::iterator iElemNum = m_elemVecCounters.begin (); iElemNum != m_elemVecCounters.end (); ++ iElemNum) 
			iElemNum->second = 0;
	}

	size_t cElemRefs = m_elemRefs.size ();

	for (size_t i = 0; i < cElemRefs; ++ i) {
		elem_ref_t & elRef = m_elemRefs [i];
		if (_rgnKind == rkPoint) {
			const size_t nPoints = elRef.Elem ().points.size ();
			if (nPoints > 1) { 
				const point_t ptPrev = elRef.Elem ().points [0];

				for (size_t j = 1 ; j < nPoints ; ++ j)
					_AddSegmentToCells (i, j - 1, j, _bCountOnly);

				_AddSegmentToCells (i, nPoints - 1, 0, _bCountOnly);
			}
		} else if (_rgnKind == rkPolygon) { 
			_AddElemToCells (elRef, _bCountOnly);
		}
	}

	if (! _bCountOnly) {
		for (long r = 0; r < m_lRows; ++ r) {
			for (long c = 0; c < m_lColumns; ++ c) {
				m_ppCells [r] [c].nElemIdx    -= m_ppCells [r] [c].nElems;
				m_ppCells [r] [c].nSegmentIdx -= m_ppCells [r] [c].nSegments;
			}
		}
	}
}

bool CPolygonIntersections::_AddIntersection (
	intersections_t & _intersections, rgn_element_t * _pElem, const intersection_t & _inters
) {
	intersections_t::iterator & iIntersSet = _intersections.find (_pElem);
	set_of_intersections_t intersSet;

	if (iIntersSet == _intersections.end ()) {
		set_of_intersections_t intersSet;
		_intersections.insert (std::make_pair (_pElem, intersSet));
		iIntersSet = _intersections.find (_pElem);
	}

	int nBefore = iIntersSet->second.size ();
	iIntersSet->second.insert (_inters);
	int nAfter = iIntersSet->second.size ();

	return (nAfter > nBefore);
}

bool CPolygonIntersections::_AddIntersection ( 
	intersections_t & _intersections,
	const size_t _cElemRefIdx1, const size_t _cElemRefIdx2,
	size_t _cPtIdx1, size_t _cPtIdx2, 
	const point_t & _pt,
	bool _bSelf, bool _bSuspected, bool _bPointOnSegment
) {
	rgn_t         * const pPolygon1 = m_elemRefs [_cElemRefIdx1].pPolygon;
	rgn_t         * const pPolygon2 = m_elemRefs [_cElemRefIdx2].pPolygon;
	rgn_element_t * const pElem1 = & m_elemRefs [_cElemRefIdx1].Elem ();	
	rgn_element_t * const pElem2 = & m_elemRefs [_cElemRefIdx2].Elem ();	
	bool bRet = false;

	if (_cElemRefIdx1 < _cElemRefIdx2)
		bRet = _AddIntersection (_intersections, pElem1, intersection_t (pPolygon2, pElem2, _cPtIdx1, _cPtIdx2, _pt, _bSelf, _bSuspected, _bPointOnSegment));
	else if (! _bSelf)
		bRet = _AddIntersection (_intersections, pElem2, intersection_t (pPolygon1, pElem1, _cPtIdx2, _cPtIdx1, _pt, _bSelf, _bSuspected, _bPointOnSegment));
	else { 
		assert (_cPtIdx1 != _cPtIdx2);
		bRet = _AddIntersection (_intersections, pElem2, intersection_t (pPolygon1, pElem1, __min (_cPtIdx2, _cPtIdx1), __max (_cPtIdx1, _cPtIdx2), _pt, _bSelf, _bSuspected, _bPointOnSegment));
	}

	return bRet;
}

bool CPolygonIntersections::_IntersectionExists (size_t _eltRefIdx1, size_t _eltRefIdx2, intersections_t & _intersections) { 	
	bool ret = false;
	rgn_element_t * const pElem1 = & m_elemRefs [_eltRefIdx1].Elem ();
	rgn_element_t * const pElem2 = & m_elemRefs [_eltRefIdx2].Elem ();
	intersections_t::iterator iIntersSet = _intersections.find (pElem1);//_eltRefIdx1);

	if (iIntersSet != _intersections.end ()) {
		for (
			set_of_intersections_t::iterator iInters = iIntersSet->second.begin ();
			! ret && iInters != iIntersSet->second.end ();
			++ iInters
		)
			ret = (iInters->pElem2 == pElem2);
	}
	
	return ret;
}

// ------------------------------------------------------------------------------------
// Self-intersection detection 
// ------------------------------------------------------------------------------------

size_t CPolygonIntersections::_PrevPtIdx (points_t & _points, int _nPtIdx)  { 
	return (_nPtIdx == 0) ? _points.size () - 1 : _nPtIdx - 1;
} 

size_t CPolygonIntersections::_NextPtIdx (points_t & _points, size_t _cPtIdx)  { 
	return (_cPtIdx == _points.size () - 1) ? 0 : _cPtIdx + 1;
} 

bool CPolygonIntersections::_IsRightOrientedAt (points_t & _points, int _nPtIdx, bool _bBackwardReverseDirection) {
	size_t cPrevPtIdx = _PrevPtIdx (_points, _nPtIdx);
	size_t cNextPtIdx = _NextPtIdx (_points, _nPtIdx);

	if (_bBackwardReverseDirection)
	    std::swap (cPrevPtIdx, cNextPtIdx);

	points_t points (3);
	points [0] = _points [cPrevPtIdx];
	points [1] = _points [_nPtIdx];
	points [2] = _points [cNextPtIdx];

	return IsRightOriented (points);
} 

bool CPolygonIntersections::_IsStraightDoublePassEndpoint (
	points_t _points, int _nForwardPtIdx, 
	int _nBackwardPtIdx, bool _bIsEnd, bool _bBackwardReverse
) { 
	// orientation of 3 points is measured around a double pass endpoints.
	const BYTE FORWARD_RIGHT    = 1;  // FORWARD are the points with smaller indices around the double pass endpoint
	const BYTE BACKWARD_RIGHT   = 2;  // BACKWARD are the points with larger indices --- // ----
	const BYTE SEQUENTIAL_RIGTH = 4;  // SEQUENTIAL points are the double pass endpoint 
								      // and 2 its neighboring points which are not part of the double pass

	const int arrStraightDoublePassEndpoint [] = { 
		FORWARD_RIGHT,  
		FORWARD_RIGHT | BACKWARD_RIGHT,
		FORWARD_RIGHT | BACKWARD_RIGHT | SEQUENTIAL_RIGTH,
		BACKWARD_RIGHT
	};

	const int arrReverseDoublePassEndpoint [] = { 
		FORWARD_RIGHT | SEQUENTIAL_RIGTH,  
		SEQUENTIAL_RIGTH,
		0,
		BACKWARD_RIGHT | SEQUENTIAL_RIGTH
	};

	int nDoublePassEndpointType = 0;
	
	if (_IsRightOrientedAt (_points, _nForwardPtIdx, false))
		nDoublePassEndpointType |= FORWARD_RIGHT;
	if (_IsRightOrientedAt (_points, _nBackwardPtIdx, _bBackwardReverse))
		nDoublePassEndpointType |= BACKWARD_RIGHT;

	points_t sequentialPts (3);

	sequentialPts [1] = _points [_nForwardPtIdx];

	sequentialPts [0] = _bIsEnd ? 
		_points [_bBackwardReverse ? _NextPtIdx (_points, _nBackwardPtIdx) : _PrevPtIdx (_points, _nBackwardPtIdx)] :
		_points [_PrevPtIdx (_points, _nForwardPtIdx)];
	sequentialPts [2] = _bIsEnd ? 
		_points [_NextPtIdx (_points, _nForwardPtIdx)] :
		_points [_bBackwardReverse ? _PrevPtIdx (_points, _nBackwardPtIdx) : _NextPtIdx (_points, _nBackwardPtIdx)]; 

	if (IsRightOriented (sequentialPts))
		nDoublePassEndpointType |= SEQUENTIAL_RIGTH;

	for (size_t c = 0; c < 4; ++ c) { 
		if (nDoublePassEndpointType == arrStraightDoublePassEndpoint [c])
			return true;
		if (nDoublePassEndpointType == arrReverseDoublePassEndpoint [c])
			return false;
	} 

	assert (false);
	return false;
} 

bool CPolygonIntersections::_FindDoublePass (points_t _points, size_list_map_t & _mpCoincidingPointLists, int _nStartPtIdx1, int & _nEndPtIdx1, int & _nEndPtIdx2) {
	bool bIsDoublePass = false;
	const int nPoints = _points.size ();
	int nIncr = 0;

	for (_nEndPtIdx1 = _nStartPtIdx1 + 1; _nEndPtIdx1 < nPoints; ++ _nEndPtIdx1) { 
		if (_mpCoincidingPointLists.find (_nEndPtIdx1) != _mpCoincidingPointLists.end ()) { 
			const int nMinMatchingIdx2 = _mpCoincidingPointLists [_nEndPtIdx1].empty () ? 
				-1 : _mpCoincidingPointLists [_nEndPtIdx1].front ();
						
			if (! bIsDoublePass && ::abs (nMinMatchingIdx2 - _nEndPtIdx2) == 1) { 
				bIsDoublePass = true;
				nIncr = (nMinMatchingIdx2 == _nEndPtIdx2 - 1) ? -1 : 1;
				_nEndPtIdx2 = nMinMatchingIdx2;
			} else if (bIsDoublePass) { 
				if (nMinMatchingIdx2 == _nEndPtIdx2 + nIncr) { 
					_mpCoincidingPointLists [_nEndPtIdx1].remove (nMinMatchingIdx2);
					_nEndPtIdx2 += nIncr;
				} else
					break;
			}
		} else
			break;
	}

	_nEndPtIdx1 --;

	return bIsDoublePass;
}

inline bool CPolygonIntersections::_PointBelongsToSegment (const point_t _ptSeg0, const point_t _ptSeg1, const point_t _pt) const { 
	bool bCheckX = ::fabs (_ptSeg1.x - _ptSeg0.x) > ::fabs (_ptSeg1.y - _ptSeg0.y); 
	point_t ptSeg0 = bCheckX ? ((_ptSeg0.x < _ptSeg1.x) ? _ptSeg0 : _ptSeg1) : ((_ptSeg0.y < _ptSeg1.y) ? _ptSeg0 : _ptSeg1);
	point_t ptSeg1 = (ptSeg0 == _ptSeg0) ? _ptSeg1 : _ptSeg0;
	bool bRet = bCheckX ? (ptSeg0.x < _pt.x && _pt.x < ptSeg1.x) : (ptSeg0.y < _pt.y && _pt.y < ptSeg1.y);

	if (bRet)
		bRet = bCheckX ? 
			(__min (ptSeg0.y, ptSeg1.y) <= _pt.y && _pt.y <= __max (ptSeg0.y, ptSeg1.y)) : 
			(__min (ptSeg0.x, ptSeg1.x) <= _pt.x && _pt.x <= __max (ptSeg0.x, ptSeg1.x));
	if (bRet) { 
		double fDX = ptSeg1.x - ptSeg0.x, fDY = ptSeg1.y - ptSeg0.y;
		double fdx = _pt.x - ptSeg0.x, fdy = _pt.y - ptSeg0.y;

		bRet = (fDX != 0 && fDY != 0) ? 
			(fDX*fdy == fDY*fdx) : (bCheckX ? (fdy == 0) : (fdx == 0));
	}

	return bRet;
} 

bool CPolygonIntersections::_TestPointOnSegmentIntersection (
	const intersection_t & inters,
	const points_t & _points1, const points_t & _points2, size_t _cIdx1, size_t _cIdx2
) const { 
	static point_t triangle1 [3];
	static point_t triangle2 [3];
	bool bPt2IsInters = (inters.point == _points2 [_cIdx2]);
	size_t cIdx1 = bPt2IsInters ? _cIdx1 : _cIdx2;
	size_t cIdx2 = bPt2IsInters ? _cIdx2 : _cIdx1;
	const points_t & points1 = bPt2IsInters ? _points1 : _points2;
	const points_t & points2 = bPt2IsInters ? _points2 : _points1;

	triangle2 [0] = triangle1 [0] = points1 [cIdx1];
	triangle2 [1] = triangle1 [1] = points1 [(cIdx1 < points1.size() - 1) ? cIdx1 + 1 : 0];
	triangle1 [2] = points2 [cIdx2 > 0 ? cIdx2 - 1 : points2.size () - 1]; // point before the one on segment
	triangle2 [2] = points2 [cIdx2 < points2.size () - 1 ? cIdx2 + 1 : 0]; // point after the one on segment

	double fArea1 = 0, fArea2 = 0;
	bool bRight1 = IsRightOriented (triangle1, 3, & fArea1), bRight2 = IsRightOriented (triangle2, 3, & fArea2);

	return (bRight1 != bRight2 && fArea1 != 0 && fArea2 != 0);
} 

void CPolygonIntersections::_AddSelfIntersectionsAtVertices (
	rgn_element_t * _pElem, 
	set_of_intersections_t & _setCoincidingPoints,
	bool _bStoreAllIntersections, 
	intersections_t & _intersections
) {
	bool bRet = false;
	bool bHasJitter = false;
	const size_t cCoincidingPoints = _setCoincidingPoints.size ();
	int nPoints = _pElem->points.size ();
	size_list_map_t mpCoincidingPointLists;
	size_list_map_t::iterator iPair, iPair2;
	rgn_t * pPolygon = NULL;

	for (set_of_intersections_t::iterator iInters = _setCoincidingPoints.begin (); iInters != _setCoincidingPoints.end (); ++ iInters) { 		
		intersection_t inters = * iInters;
		iPair = mpCoincidingPointLists.find (inters.cPtIdx1);

		if (iPair == mpCoincidingPointLists.end ()) { 
			size_list_t newLst;
			mpCoincidingPointLists.insert (std::make_pair (inters.cPtIdx1, newLst));
			iPair = mpCoincidingPointLists.find (inters.cPtIdx1);
		} 

		iPair->second.push_back (inters.cPtIdx2);
		pPolygon = inters.pPolygon2;
	}

	for (iPair = mpCoincidingPointLists.begin (); iPair != mpCoincidingPointLists.end (); ++ iPair)
		iPair->second.sort ();

	if (mpCoincidingPointLists.find (0) != mpCoincidingPointLists.end () && ! mpCoincidingPointLists [0].empty ()) { 
		size_list_map_t mpCoincidingPointLists2 (mpCoincidingPointLists);
		for (size_list_t::iterator iId = mpCoincidingPointLists [0].begin (); iId != mpCoincidingPointLists [0].end ()/* && ! bRet && ! bHasJitter*/; ++ iId) { 
			int nEndPtIdx1 = -1;
			int nStartPtIdx2 = * iId, nEndPtIdx2 = * iId;		
					
			const bool bIsDoublePass = _FindDoublePass (_pElem->points, mpCoincidingPointLists, 0, nEndPtIdx1, nEndPtIdx2);
			const int nIncr2 = (nEndPtIdx2 > nStartPtIdx2) ? 1 : -1;

			for (int n1 = 0, n2 = nStartPtIdx2; n1 <= nEndPtIdx1; ++ n1, n2 += nIncr2) { 
				const int nExtraIdx = nPoints + n1;
				if (mpCoincidingPointLists2.find (n2) == mpCoincidingPointLists2.end ()) { 
					size_list_t newLst;
					mpCoincidingPointLists2.insert (std::make_pair (n2, newLst));
				}

				mpCoincidingPointLists2 [n2].push_back (nExtraIdx);

				if (! bIsDoublePass)
					break;
			}
		}
		mpCoincidingPointLists = mpCoincidingPointLists2;
	} 

	bool bStop = false;
	for (iPair = mpCoincidingPointLists.begin (); iPair != mpCoincidingPointLists.end () && ! bStop; ++ iPair) { 
		const int nStartPtIdx1 = iPair->first;
		size_list_t & repeatingPointList = iPair->second;
		bool bFound = false;

		if (! repeatingPointList.empty ()) {
			for (size_list_t::iterator iId = repeatingPointList.begin (); iId != repeatingPointList.end () && ! bStop; ++ iId) { 
				int nEndPtIdx1 = -1;
				int nStartPtIdx2 = * iId, nEndPtIdx2 = * iId;		
				const bool bIsDoublePass = 
					_FindDoublePass (_pElem->points, mpCoincidingPointLists, nStartPtIdx1, nEndPtIdx1, nEndPtIdx2);
				const bool bReverse = (nEndPtIdx2 > nStartPtIdx2);
				bHasJitter = (::abs (nStartPtIdx1 - nStartPtIdx2) <= 2) || (::abs (nEndPtIdx1 - nEndPtIdx2) <= 2);
				if (nEndPtIdx2 >= nPoints)
					nEndPtIdx2 -= nPoints;
				if (nStartPtIdx2 >= nPoints)
					nStartPtIdx2 -= nPoints;
				
				if (bIsDoublePass && ! bHasJitter) { 
					const bool bStartOfDoublePassIsStraight = 
						_IsStraightDoublePassEndpoint (_pElem->points, nStartPtIdx1, nStartPtIdx2, false, bReverse);
					const bool bEndOfDoublePassIsStraight = 
						_IsStraightDoublePassEndpoint (_pElem->points, nEndPtIdx1, nEndPtIdx2, true, bReverse);

					bFound = (bStartOfDoublePassIsStraight != bEndOfDoublePassIsStraight);
				} 
				else
					bFound = PolygonsIntersectAtVertex (_pElem->points, _pElem->points, nStartPtIdx1, nStartPtIdx2, true);

				if (bFound || bHasJitter) { 
					intersection_t inters (pPolygon, _pElem, -1, -1, point_t (0, 0), true);
					if (bFound) { 
						inters.point = _pElem->points [nStartPtIdx1];
						inters.cPtIdx1 = nStartPtIdx1;
						inters.cPtIdx2 = bIsDoublePass ? nEndPtIdx1 : nStartPtIdx2;
					}  else if (bHasJitter) { 
						int nPointIdx = -1;
						if (::abs (nStartPtIdx1 - nStartPtIdx2) <= 2) { 
							nPointIdx =
								(::abs (nStartPtIdx1 - nStartPtIdx2) == 2) ? (nStartPtIdx1 + nStartPtIdx2)/2 : nStartPtIdx1; 
						} else { 
							nPointIdx =
								(::abs (nEndPtIdx1 - nEndPtIdx2) == 2) ? (nEndPtIdx1 + nEndPtIdx2)/2 : nEndPtIdx1; 
						} 

						if (nPointIdx == 0)	{ 
							inters.cPtIdx1 = 1;
							inters.cPtIdx2 = nPoints - 1;
						} else if (nPointIdx == nPoints - 1) { 
							inters.cPtIdx1 = 0;
							inters.cPtIdx2 = nPoints - 2;
						} else { 
							inters.cPtIdx1 = nPointIdx - 1;
							inters.cPtIdx2 = nPointIdx + 1;
						} 

						inters.point = _pElem->points [nPointIdx];
						if (_pElem->points [inters.cPtIdx1] == _pElem->points [inters.cPtIdx2]) { 
							if (! (inters.point == _pElem->points [inters.cPtIdx1]))
								inters.cPtIdx1 = nPointIdx;
							else
								inters.cPtIdx2 = nPointIdx;
						} 
						inters.bJitter = true;
						bFound = true;
					} 
					_AddIntersection	(_intersections, _pElem, inters);
				} 
			} 

			repeatingPointList.clear ();
		} 

		bStop = (bFound || bHasJitter) && ! _bStoreAllIntersections;
	} 
} 

bool CPolygonIntersections::_CheckAddSegmentIntersections (
	intersections_t & _intersections, intersections_t & _suspectedIntersections,
	const size_t _cElemRefIdx1, const size_t _cElemRefIdx2,
	const points_t & _points1, const points_t & _points2,
	const size_t _cPtIdx11, const size_t _cPtIdx12, const size_t _cPtIdx21, const size_t _cPtIdx22 
) { 
	bool bAdded = false;
	bool bSelf = (_cElemRefIdx1 == _cElemRefIdx2);

	const point_t & pt11 = _points1 [_cPtIdx11];
	const point_t & pt12 = _points1 [_cPtIdx12];
	const point_t & pt21 = _points2 [_cPtIdx21];
	const point_t & pt22 = _points2 [_cPtIdx22];

	point_t ptInters ((float) INT_MAX, (float) INT_MAX);

	const rgn_element_t & elem1 = m_elemRefs [_cElemRefIdx1].Elem ();
	const rgn_element_t & elem2 = m_elemRefs [_cElemRefIdx2].Elem ();

	if (SegmentIntersectsSegment (pt11, pt12, pt21, pt22, & ptInters))
		bAdded = _AddIntersection (_intersections, _cElemRefIdx1, _cElemRefIdx2, _cPtIdx11, _cPtIdx21, ptInters, bSelf);
	else if (pt11 == pt21 && (! bSelf || _cPtIdx11 != _cPtIdx21))
		bAdded = _AddIntersection (_suspectedIntersections, _cElemRefIdx1, _cElemRefIdx2, _cPtIdx11, _cPtIdx21, pt11, bSelf, true);
	else if (pt11 == pt22 && (! bSelf || _cPtIdx11 != _cPtIdx22))
		bAdded = _AddIntersection (_suspectedIntersections, _cElemRefIdx1, _cElemRefIdx2, _cPtIdx11, _cPtIdx22, pt11, bSelf, true);
	else if (pt12 == pt21 && (! bSelf || _cPtIdx12 != _cPtIdx21)) 
		bAdded = _AddIntersection (_suspectedIntersections, _cElemRefIdx1, _cElemRefIdx2, _cPtIdx12, _cPtIdx21, pt12, bSelf, true);
	else if (pt12 == pt22 && (! bSelf || _cPtIdx12 != _cPtIdx22))
		bAdded = _AddIntersection (_suspectedIntersections, _cElemRefIdx1, _cElemRefIdx2, _cPtIdx12, _cPtIdx22, pt12, bSelf, true);
	else if (_PointBelongsToSegment (pt11, pt12, pt21))
		bAdded = _AddIntersection (_suspectedIntersections, _cElemRefIdx1, _cElemRefIdx2, _cPtIdx11, _cPtIdx21, pt21, bSelf, true, true);
	else if (_PointBelongsToSegment (pt11, pt12, pt22))
		bAdded = _AddIntersection (_suspectedIntersections, _cElemRefIdx1, _cElemRefIdx2, _cPtIdx11, _cPtIdx22, pt22, bSelf, true, true);
	else if (_PointBelongsToSegment (pt21, pt22, pt11))
		bAdded = _AddIntersection (_suspectedIntersections, _cElemRefIdx2, _cElemRefIdx1, _cPtIdx21, _cPtIdx11, pt11, bSelf, true, true);
	else if (_PointBelongsToSegment (pt21, pt22, pt12))
		bAdded = _AddIntersection (_suspectedIntersections, _cElemRefIdx2, _cElemRefIdx1, _cPtIdx21, _cPtIdx12, pt12, bSelf, true, true);

	return bAdded;
} 

// ------------------------------------------------------------------------------------
// Search for segment intersections and suspected intersections (coinciding points)
// ------------------------------------------------------------------------------------

void CPolygonIntersections::_FindIntersectionsQuadratic (
	intersections_t & _intersections, intersections_t & _suspectedIntersections,
	bool _bCheckSelfIntersections, bool _bCheckOverlappingPolygons
) {
	size_t cElems = m_elemRefs.size ();

	for (size_t e1 = 0; e1 < cElems; ++ e1) { 
		const rgn_element_t & elem1 = m_elemRefs [e1].Elem ();
		size_t cPoints1 = elem1.points.size ();
		for (size_t c1 = 0; c1 < cPoints1; ++ c1) { 
			if (_bCheckSelfIntersections) { 
				for (size_t c2 = c1 + 1; c2 < cPoints1; ++ c2)
					_CheckAddSegmentIntersections (
						_intersections, _suspectedIntersections, e1, e1, elem1.points, elem1.points, 
						c1, (c1 < cPoints1 - 1) ? c1 + 1 : 0, c2, (c2 < cPoints1 - 1) ? c2 + 1 : 0
					);
			}
			if (_bCheckOverlappingPolygons) { 
				for (size_t e2 = e1 + 1; e2 < cElems; ++ e2) { 
					const rgn_element_t & elem2 = m_elemRefs [e2].Elem ();
					size_t cPoints2 = elem2.points.size ();
					for (size_t c2 = 0; c2 < cPoints2; ++ c2) { 
						_CheckAddSegmentIntersections (
							_intersections, _suspectedIntersections, e1, e2, elem1.points, elem2.points, 
							c1, (c1 < cPoints1 - 1) ? c1 + 1 : 0, c2, (c2 < cPoints2 - 1) ? c2 + 1 : 0
						);
					}
				} 
			} 
		} 
	} 
}

void CPolygonIntersections::_PreapreSegmentData () { 
	_PopulateGrid (rkPoint, true);
	_SetProgressStatus (15);
	_AllocDataStruct (m_segmVecCounters, rkPoint);
	_PopulateGrid (rkPoint, false);
} 

void CPolygonIntersections::_FindIntersectionsInGrid (
	intersections_t & _intersections, intersections_t & _suspectedIntersections,
	bool _bCheckSelfIntersections, bool _bCheckOverlappingPolygons
) { 
	_PreapreSegmentData ();
	_SetProgressStatus (20);

	int nProgressSpan = _bCheckOverlappingPolygons ? 37 : 77;
	size_t cIntersections = 0;
	for (long r = 0; r < m_lRows; ++ r) {
		_SetProgressStatus (20 + (r + 1)*nProgressSpan/m_lRows);
		for (long c = 0; c < m_lColumns; ++ c) {
			const int nSegments = m_ppCells [r][c].nSegments;
			if (nSegments < 2)
				continue;

			const polygon_segments_t & cellSegments = m_cellSegmentVectors [nSegments];
			size_t start_idx = m_ppCells [r][c].nSegmentIdx, end_idx = m_ppCells [r][c].nSegmentIdx + nSegments;
			for (size_t s1 = start_idx; s1 < end_idx; ++ s1) { 
				const polygon_segment_t * const pSgt1 = & cellSegments [s1];

				for (size_t s2 = s1 + 1; s2 < end_idx; ++ s2) {
					const polygon_segment_t * const pSgt2 = & cellSegments [s2];
					bool bSelf = false;
					
					if (pSgt2->cElemRefIdx == pSgt1->cElemRefIdx) { 
						if (! _bCheckSelfIntersections)
							continue;
						bSelf = true;
					} 
					else if (! _bCheckOverlappingPolygons)
						continue;

					point_t ptInters ((float) INT_MAX, (float) INT_MAX);

					bool bAdded = _CheckAddSegmentIntersections (
						_intersections, _suspectedIntersections, pSgt1->cElemRefIdx, pSgt2->cElemRefIdx,
						m_elemRefs [pSgt1->cElemRefIdx].Elem ().points, m_elemRefs [pSgt2->cElemRefIdx].Elem ().points,
						pSgt1->cPtIdx1, pSgt1->cPtIdx2, pSgt2->cPtIdx1, pSgt2->cPtIdx2
					);		

					if (bAdded)
						++ cIntersections;
				}
			}
		}
	}

	_FreeSegmentData ();
}

void CPolygonIntersections::_TestSuspectedIntersections (
	intersections_t & _suspIntersections, intersections_t & _intersections,
	bool _bCheckSelfIntersections, bool _bCheckOverlappingPolygons
) {
	for (intersections_t::iterator iIntersSet = _suspIntersections.begin (); iIntersSet != _suspIntersections.end (); ++ iIntersSet) {
		rgn_element_t * pElt1 = iIntersSet->first;
		set_of_intersections_t & intersSet = iIntersSet->second;
		set_of_intersections_t selfMatchesSet;
		intersection_t inters;

		for (set_of_intersections_t::iterator iInters = intersSet.begin (); iInters != intersSet.end (); ++ iInters) { 
			inters = * iInters;
			const rgn_element_t * pElt2 = inters.pElem2;
			
			if (inters.bPointOnSegment) {
				if (_TestPointOnSegmentIntersection (inters, pElt1->points, pElt2->points, inters.cPtIdx1, inters.cPtIdx2)) { 
					inters.bSuspected = false;
					_AddIntersection (_intersections, iIntersSet->first, inters);
				} 
			} else if (inters.bSelf) {
				selfMatchesSet.insert (inters);
			} else if (
				! pElt1->bHole && ! pElt2->bHole && // Holes are treated during containment check.
				PolygonsIntersectAtVertex (pElt1->points, pElt2->points, inters.cPtIdx1, inters.cPtIdx2, inters.bSelf)
			  ) {
				inters.bSuspected = false;
				_AddIntersection (_intersections, iIntersSet->first, inters);
			}
		}

		if (! selfMatchesSet.empty ())
			_AddSelfIntersectionsAtVertices (pElt1, selfMatchesSet, true, _intersections);
	} 
}


// To be used ONLY if it's certain that the polygons do not intersect!
int CPolygonIntersections::_PolygonElemContainsPolygonElem (const rgn_element_t & _extElem, const rgn_element_t & _intElem) {
	const size_t cIntElemPtsSize = _intElem.points.size ();
	int nContains = 0;
	for (size_t p = 0; p < cIntElemPtsSize && nContains == 0; ++ p)
		nContains = PolygonContainsPt (_extElem.points, _intElem.points [p]);

	// 0 means no difference found between polygons => they interlap
	// -1 means an "internal" point outside the external polygon found. 
	//		As there are no border intersections => polygons do not interlap.
	return ((nContains == 0 && cIntElemPtsSize == _extElem.points.size ()) || nContains == 1);
}

# pragma optimize ("", off) // Turn off optimizations (per function): the optimizer of VC6 generates incorrect code here.

void CPolygonIntersections::_TestReportContainedElement (
	intersections_t & _intersections, 
	const elem_ref_t * _pIntElemRef, const elem_chain_t & _containingChain
) {
	bool bAddIntersection = false;
	elem_ref_t * pIntersExtElemRef = NULL;

	if (_containingChain.first2Elems [1] != -1) { // Means more than 1 containing elem.
		std::multimap<float, elem_ref_t *> setContainingElemsByDiam;
		float fMinArea = FLT_MAX;
		//size_t cExtElemIdx = 0;
		const size_t cNumElts = _containingChain.restOfElems.size () + 2;

		for (size_t c = 0; c < cNumElts; ++ c) {
			const size_t elRefIdx = _containingChain.Get (c);
			elem_ref_t * const pExtElemRef = & m_elemRefs [elRefIdx];

			const float fRectArea = 
				pExtElemRef->Elem ().rectBound.Height ()*pExtElemRef->Elem ().rectBound.Width ();
			setContainingElemsByDiam.insert (std::make_pair (fRectArea, pExtElemRef));

			if (fMinArea > fRectArea)
				fMinArea = fRectArea;
		}
		
		std::pair<it_ref, it_ref> minAreaElems = setContainingElemsByDiam.equal_range (fMinArea);

		// XXX Add a more comprehensive check here! We need to make this check 
		// FOR EVERY POLYGON among the referenced SEPARATELY
		for (it_ref iRef = minAreaElems.first; iRef != minAreaElems.second; ++ iRef) {
			elem_ref_t * const pExtElemRef = iRef->second;
			if (! pExtElemRef->Elem ().bHole) { 
				 bAddIntersection = true;
	 		     pIntersExtElemRef = pExtElemRef;
				 break;
			} 
		}
	} else {
		bAddIntersection = true;
		pIntersExtElemRef = & m_elemRefs [_containingChain.first2Elems [0]];
	}

	if (bAddIntersection)  { 
		point_t ptInters (_pIntElemRef->Elem ().rectBound.MeanX (), _pIntElemRef->Elem ().rectBound.MeanY ());
		_AddIntersection (
			_intersections,
			& pIntersExtElemRef->Elem (),
			intersection_t (_pIntElemRef->pPolygon, & _pIntElemRef->Elem (), 0, 0, ptInters, false, false, false, false, true)
		);
	}
}

# pragma optimize ("", on) // Restore optimizations.

void CPolygonIntersections::_FindContainedPolygonsInGrid (
	intersections_t & _intersections, intersections_t & _existingIntersections
) {
	int nNumContainingChains = 0;

	// Data for multiple containment.
	const size_t cInitialSize = fast_floor (::sqrt ((double) m_lRows*m_lColumns));
	fvector<elem_ref_t *> vContainedElems  (cInitialSize); 
	containing_elem_chains_t containingChains (cInitialSize);

	long r = 0, c = 0, start_idx = -1, end_idx = -1, e1 = 0, e2 = 0; 

	elem_ref_t * pRef1 = NULL;
	elem_ref_t * pRef2 = NULL;
	elem_ref_t * pExtElemRef = NULL, * pIntElemRef = NULL;

	size_t containmentIdx = -1;

	size_t nElems = 0;

	for (r = 0; r < m_lRows; ++ r) {
		_SetProgressStatus (60 + (r + 1)*37/m_lRows);
		for (c = 0; c < m_lColumns; ++ c) {
			nElems = m_ppCells [r][c].nElems;
			if (nElems < 2)
				continue;

			start_idx = m_ppCells [r][c].nElemIdx;
			end_idx = m_ppCells [r][c].nElemIdx + nElems;
			for (e1 = start_idx; e1 < end_idx; ++ e1) { 
				pRef1 = & m_cellElemVectors [nElems] [e1];
				const rgn_element_t & elem1 = pRef1->Elem (); 
				const rgn_t * pPolygon1 = pRef1->pPolygon;

				for (e2 = e1 + 1; e2 < end_idx; ++ e2) {
					pRef2 = & m_cellElemVectors [nElems] [e2];
					const rgn_element_t & elem2 = pRef2->Elem (); 
					const rgn_t * pPolygon2 = pRef2->pPolygon;
					pExtElemRef = NULL;
					pIntElemRef = NULL;

					if (* pRef1 == * pRef2)
						assert (0);

					if (elem1.rectBound.Contains (elem2.rectBound)) { 
						pExtElemRef = pRef1;
						pIntElemRef = pRef2;
					} else if (elem2.rectBound.Contains (elem1.rectBound)) { 
						pExtElemRef = pRef2;
						pIntElemRef = pRef1;
					}

					if (pExtElemRef != NULL) {	// suspicion of a containment
						if (pExtElemRef->pPolygon == pIntElemRef->pPolygon) { // dont detect holes here!
							if (pIntElemRef->Elem ().bHole)
								continue;
							else if (pExtElemRef->cEltIdx == pIntElemRef->cEltIdx)
								assert (0);
							else 
								continue;
						}

						size_t cMinIdx = __min (pIntElemRef->cIdx, pExtElemRef->cIdx);
						size_t cMaxIdx = __max (pIntElemRef->cIdx, pExtElemRef->cIdx);

						if (
							m_elemRefs [cMinIdx].checkedElemRefChain.Find (cMaxIdx) != -1 ||
							_IntersectionExists (cMinIdx, cMaxIdx, _existingIntersections)
						) 
							continue;

						m_elemRefs [cMinIdx].checkedElemRefChain.Push (cMaxIdx);

						if (_PolygonElemContainsPolygonElem (pExtElemRef->Elem (), pIntElemRef->Elem ())) {
							containmentIdx = pIntElemRef->nContainedVecIdx;

							if (containmentIdx == -1) { 
								if (vContainedElems.size() < (size_t) nNumContainingChains + 1) { 
									vContainedElems .resize (nNumContainingChains*2);
									containingChains.resize (nNumContainingChains*2);
								} 

								vContainedElems [nNumContainingChains] = pIntElemRef;
								pIntElemRef->nContainedVecIdx = nNumContainingChains;
								containmentIdx = nNumContainingChains ++;
							}

							containingChains [containmentIdx].Push (pExtElemRef->cIdx);
						} 
					} 
				}
			}
		}
	}

	for (int i = 0; i < nNumContainingChains; ++ i)
		_TestReportContainedElement (_intersections, vContainedElems [i], containingChains [i]);

	vContainedElems.clear ();

	const size_t cContaining = vContainedElems.size ();
	for (size_t c2 = 0; c2 < cContaining; ++ c2) 
		containingChains [c2].restOfElems.clear ();
	containingChains.clear ();
}

void CPolygonIntersections::ReportIntersections (intersections_t & _intersections) { 						  
	for (
		intersections_t::iterator iIntersSet = _intersections.begin ();
		iIntersSet != _intersections.end (); ++ iIntersSet
	) { 
		const rgn_element_t * pElem1 = iIntersSet->first;
		const set_of_intersections_t & intersSet = iIntersSet->second;

		for (set_of_intersections_t::const_iterator iInters = intersSet.begin (); iInters != intersSet.end (); ++ iInters) { 
			const intersection_t inters = * iInters;
			const rgn_element_t * pElem2 = inters.pElem2;
			point_t ptInters = inters.point;
			rect_t rect (ptInters.x, ptInters.y, ptInters.x, ptInters.y);

			if (inters.bContained) { 
				assert (inters.pPolygon2 != NULL);
				rect.Extend (inters.pPolygon2->rectBound);
			} else { 
				size_t cPts1 = pElem1->points.size ();
				size_t cPts2 = pElem2->points.size ();
				const point_t & ptRect1       = inters.bSelf ?  
					pElem1->points [inters.cPtIdx1] : pElem1->points [inters.cPtIdx1 > 0 ? inters.cPtIdx1 - 1 : cPts1 - 1];
				const point_t & ptRect2       = inters.bSelf ?  
					pElem1->points [inters.cPtIdx2] : pElem1->points [inters.cPtIdx1 + 1 < cPts1 ? inters.cPtIdx1 + 1 : 0];
				const point_t & ptAdditional1 = pElem2->points [inters.cPtIdx2 > 0 ? inters.cPtIdx2 - 1 : cPts2 - 1];
				const point_t & ptAdditional2 = pElem2->points [inters.cPtIdx2 + 1 < cPts2 ? inters.cPtIdx2 + 1 : 0];


				if (ptInters.x == INT_MAX || ptInters.y == INT_MAX) { 
					ptInters.x = (ptRect1.x + ptRect2.x + ptAdditional1.x + ptAdditional2.x)/4;
					ptInters.y = (ptRect1.y + ptRect2.y + ptAdditional1.y + ptAdditional2.y)/4;
				} 

				rect.swap (rect_t (ptRect1.x, ptRect1.y, ptRect2.x, ptRect2.y));
				if (! inters.bSelf) { 
					rect.Extend (ptAdditional1);
					rect.Extend (ptAdditional2);
				} 
			} 

			if (inters.bJitter)
				AddMessageToVMList (
					"Polygon has a jitter", true,
					inters.pPolygon2, pElem2->cLevel, ptInters, rect
				);	
			else if (inters.bContained)
				AddMessageToVMList (
					"Interior areas of polygons overlap", true,
					inters.pPolygon2, pElem2->cLevel, ptInters, rect
				);
			else if (inters.bSelf)
				AddMessageToVMList (
					"Polygon contour intersects itself", true,
					inters.pPolygon2, pElem2->cLevel, ptInters, rect
				);
			//else if (inters.bPointOnSegment)	   // FOR DEBUG PURPOSES ONLY
			//	AddMessageToVMList (
			//		"POINT ON SEGMENT", true,
			//		inters.pPolygon2, pElem2->cLevel, ptInters, rect
			//	);
			else
				AddMessageToVMList (
					"Contour of polygon intersects with contour of another polygon", true,
					inters.pPolygon2, pElem2->cLevel, ptInters, rect
				);
		}
	}
} 

void CPolygonIntersections::_FreeContainmentData () {
	m_elemVecCounters.clear ();
	for (size_t i = 0; i < m_cellElemVectors.size (); ++ i)
		m_cellElemVectors [i].clear ();
	m_cellElemVectors.clear ();
}

void CPolygonIntersections::_FreeSegmentData () {
	m_segmVecCounters.clear ();
	for (size_t i = 0; i < m_cellSegmentVectors.size (); ++ i)
		m_cellSegmentVectors [i].clear ();
	m_cellSegmentVectors.clear ();
}

void CPolygonIntersections::_ClearGrid () { 
	for (long i = 0; i < m_lRows; ++ i)
		delete [] m_ppCells [i];
	delete [] m_ppCells;
} 

void CPolygonIntersections::_ClearElemRefVector () { 
  	m_elemRefs.clear ();
} 

void CPolygonIntersections::_SetProgressStatus (int _nProgress) { 
	if (m_bReportProgress)
		::SetProgressStatus (_nProgress);
} 

void CPolygonIntersections::_PrepareGrid (const rgns_t & _polygons, size_t _cLevel) { 
	size_t		 nPoints = 0;
	_CountPolygonVertices (_polygons, nPoints, _cLevel);
	const float  fMapWidth  = ::fabs (m_rectBound.x1 - m_rectBound.x0), 
				 fMapHeight = ::fabs (m_rectBound.y1 - m_rectBound.y0);
	const float  fCellArea = fMapHeight*fMapWidth/nPoints;		 
	const float  fCellSize = ::sqrtf (fCellArea);

	m_lRows = fast_floor (fMapHeight/fCellSize + 0.5f);
	m_lColumns = fast_floor (fMapWidth /fCellSize + 0.5f);
	m_ppCells = new grid_cell_t * [m_lRows + 1];

	polygon_segment_t::Init (m_elemRefs);

	_AllocGridCells (fMapWidth, fMapHeight);
} 

void CPolygonIntersections::_FindContourIntersections (
	bool _bCheckSelfIntersections,    bool _bCheckOverlappingPolygons, 
	intersections_t & _intersections, bool _bReportProgress
) { 
	intersections_t  suspectedIntersections;

	if (m_bUseGrid)
		_FindIntersectionsInGrid (_intersections, suspectedIntersections, _bCheckSelfIntersections, _bCheckOverlappingPolygons);
	else
		_FindIntersectionsQuadratic (_intersections, suspectedIntersections, _bCheckSelfIntersections, _bCheckOverlappingPolygons);
	_TestSuspectedIntersections (suspectedIntersections, _intersections, _bCheckSelfIntersections, _bCheckOverlappingPolygons);
} 

void CPolygonIntersections::FindPolygonContourIntersections (
	const rgns_t & _polygons, 
	bool _bCheckSelfIntersections, bool _bCheckMutualIntersections, 
	intersections_t & _intersections, bool _bReportProgress
) { 
	m_bReportProgress = _bReportProgress;
	const size_t cLevels = g_map.Levels.empty () ? 1 : g_map.Levels.size ();
	for (size_t cLevel = 0; cLevel < cLevels; ++ cLevel) { 
		_FormElemRefVector (_polygons, cLevel);
		if (m_bUseGrid)
			_PrepareGrid (_polygons, cLevel);
		_FindContourIntersections (_bCheckSelfIntersections, _bCheckMutualIntersections, _intersections, true);
		if (m_bUseGrid)
			_ClearGrid ();
		_ClearElemRefVector ();
	} 
} 

void CPolygonIntersections::FindPolygonContourIntersections (
	const rgn_t & _polygon, const rgn_elements_t & _elements, 
	bool _bCheckSelfIntersections, bool _bCheckMutualIntersections, 
	intersections_t & _intersections, bool _bReportProgress, size_t _cLevel
) { 
	m_bReportProgress = _bReportProgress;
	const size_t cLevels = (_cLevel == -1) ? 
		(g_map.Levels.empty () ? 1 : g_map.Levels.size ()) : 1;
	for (rgn_elements_t::const_iterator iElem = _elements.begin (); iElem != _elements.end (); ++ iElem)
		m_rectBound.Extend (iElem->rectBound);

	for (size_t cLevel = 0; cLevel < cLevels; ++ cLevel) { 
		_FormElemRefVector (_polygon, _elements, cLevel);
		if (m_bUseGrid) {
			rgns_t polygons (1);
			polygons.push_back (_polygon);
			_PrepareGrid (polygons, cLevel);
		}
		_FindContourIntersections (_bCheckSelfIntersections, _bCheckMutualIntersections, _intersections, _bReportProgress);
		if (m_bUseGrid)
			_ClearGrid ();
		_ClearElemRefVector ();
	} 
} 

void CPolygonIntersections::_FindInternalOverlappings (intersections_t & _existingIntersections, intersections_t & _overlappings) { 
	_PopulateGrid (rkPolygon, true);
	_AllocDataStruct (m_elemVecCounters, rkPolygon);
	_PopulateGrid (rkPolygon, false);
	_FindContainedPolygonsInGrid (_overlappings, _existingIntersections);	// provide intersections to avoid duplicates
	_FreeContainmentData ();
} 

void CPolygonIntersections::VerifyPolygonOverlapping (
	const rgns_t & _polygons, size_t _cLevel, bool _bCheckSelfIntersections, bool _bCheckMutualIntersections,
	intersections_t & _contourIntersections, intersections_t & _internalOverlappings
) {
	char strStatus [64];
	::sprintf (strStatus, "Verifying level#%d...", _cLevel);
	SetStatus (strStatus);

	m_rectBound.Extend (g_map.rectBound);

	_SetProgressStatus (0);
	_FormElemRefVector (_polygons, _cLevel);
	_SetProgressStatus (5);
	if (m_bUseGrid)
		_PrepareGrid (_polygons, _cLevel);
	_SetProgressStatus (10);

	_FindContourIntersections (_bCheckSelfIntersections, _bCheckMutualIntersections, _contourIntersections, true);

	if (_bCheckMutualIntersections)
		_FindInternalOverlappings (_contourIntersections, _internalOverlappings);

	_SetProgressStatus (99);
	if (m_bUseGrid)
		_ClearGrid ();
	_ClearElemRefVector ();
	m_elemRefs.clear ();

	_SetProgressStatus (100);
}