# ifndef __PolygonIntersections_H_
# define __PolygonIntersections_H_

# include <float.h>
# include "Map.h"

typedef std::map<int, size_t> intmap_t;

struct elem_chain_t { 
	size_t			first2Elems [2];  // Most of polygons contained in the polygon + hole
	fvector<size_t> restOfElems;      // Additional place to store other polygons (if there are more than 2).

	elem_chain_t () {
		first2Elems [0] = -1;
		first2Elems [1] = -1;
	}

	void Push (size_t _elemIdx) { 
		if (first2Elems [0] == -1)
			first2Elems [0] = _elemIdx;
		else if (first2Elems [1] == -1)
			first2Elems [1] = _elemIdx;
		else
			restOfElems.push_back (_elemIdx);
	} 

	size_t Find (size_t _elemIdx) const { 
		if (first2Elems [0] == _elemIdx)
			return 0;
		else if (first2Elems [1] == _elemIdx)
			return 1;
		else if (restOfElems.size () > 0) { 
			size_t nRestSize = restOfElems.size ();
			for (size_t c = 0; c < nRestSize; ++ c)
				if (_elemIdx == restOfElems [c])
					return c + 2;
		}

		return -1;
	}

	size_t Get (size_t _cIdx) const {
		if (_cIdx < 2)
			return first2Elems [_cIdx];
		else
			return restOfElems [_cIdx - 2];
	}
};

struct elem_ref_t { 
	size_t       cIdx;                // index in m_elemRefs	
	rgn_t      * pPolygon;            // owner polygon
	size_t       cEltIdx;             // index in pPolygon->elements vector
	int	         nContainedVecIdx;    // index in containingChains
	elem_chain_t checkedElemRefChain; // elements already tested for contain(ing/ed by) this

	elem_ref_t () :
		cIdx (-1), pPolygon (NULL), cEltIdx (-1), nContainedVecIdx (-1)
	{}
	elem_ref_t (size_t _cIdx, rgn_t * _pPolygon = NULL, size_t _cEltIdx = -1) :
		cIdx (_cIdx), pPolygon (_pPolygon), cEltIdx (_cEltIdx), nContainedVecIdx (-1)
	{}

	rgn_element_t & Elem () const {
		return pPolygon->elements [cEltIdx];
	}

	elem_ref_t & operator = (const elem_ref_t & other) { 
		cIdx = other.cIdx;
		pPolygon = other.pPolygon;
		cEltIdx = other.cEltIdx;

		return * this;
	} 

	bool operator == (const elem_ref_t & _other) const {return cIdx == _other.cIdx;}
};

typedef fvector<elem_ref_t> elem_refs_t;

typedef	std::multimap<float, elem_ref_t *>::iterator it_ref;

struct polygon_segment_t {
	size_t cElemRefIdx; // index of elem in rgn_t.elements
	size_t cPtIdx1;     // index of segment's first endpoint in elem.points
	size_t cPtIdx2;     // index of segment's second endpoint in elem.points
	static elem_refs_t * m_pElemRefs;

	polygon_segment_t (
		size_t _cElemRefIdx = -1,
		size_t _cPtIdx1 = -1, size_t _cPtIdx2 = -1
	) :
		cElemRefIdx (_cElemRefIdx), cPtIdx1 (_cPtIdx1), cPtIdx2 (_cPtIdx2)
	{} 
		static void Init (elem_refs_t & _elemRefs) {m_pElemRefs = & _elemRefs;}

	const rgn_element_t & Elem () const {
		return (* m_pElemRefs) [cElemRefIdx].Elem ();
	} 
	const point_t & Pt1 () const { 
		return (* m_pElemRefs) [cElemRefIdx].Elem ().points [cPtIdx1];
	} 
	const point_t & Pt2 () const { 
		return (* m_pElemRefs) [cElemRefIdx].Elem ().points [cPtIdx2];
	} 

	bool operator < (const polygon_segment_t & _other) const { 
		bool ret = false;
		if (cElemRefIdx == _other.cElemRefIdx) { 
			if (cPtIdx1 != _other.cPtIdx1)
				ret = cPtIdx1 < _other.cPtIdx1;
			else 
				ret = cPtIdx2 < _other.cPtIdx2;
		} else 
			ret = cElemRefIdx < _other.cElemRefIdx;

		return ret;
	}
};

struct intersection_t  {
	rgn_t         * pPolygon2;
	rgn_element_t * pElem2;
	//size_t cElt2RefIdx; // the bigger cIdx of the two elem_ref_t of intersecting elems in m_elemRefs
	size_t cPtIdx1;       // idx of pt of intersection in the elem_ref_t with smaller cIdx
	size_t cPtIdx2;       // idx of pt of intersection in the elem_ref_t with bigger cIdx
	point_t point;        // pt of intersection. used to center the screen around the intersection
	bool bSelf;           // self-intersection
	bool bSuspected;      // suspected intersection
	bool bPointOnSegment; // point of one belonging
	bool bJitter;         // the "intersection" is a jitter
	bool bContained;      // one polygon is contained by the other

	intersection_t () : 
		pPolygon2 (NULL), pElem2 (NULL), cPtIdx1 (-1), cPtIdx2 (-1), point (), 
		bSelf (false), bSuspected (false), bPointOnSegment (false), bJitter (false), bContained (false)
	{}
	intersection_t (const intersection_t & _other) : 
		pPolygon2 (_other.pPolygon2), 
		pElem2 (_other.pElem2),
		cPtIdx1 (_other.cPtIdx1),
		cPtIdx2 (_other.cPtIdx2),
		point (_other.point),
		bSelf (_other.bSelf),
		bSuspected (_other.bSuspected),
		bPointOnSegment (_other.bPointOnSegment),
		bJitter (_other.bJitter),
		bContained (_other.bContained)
	{}

	intersection_t (
		rgn_t * _pPolygon2, rgn_element_t * _pElem2, 
		const size_t _cPtIdx1, const size_t _cPtIdx2, point_t _pt, 
		bool _bSelf, bool _bSuspected = false, bool _bPointOnSegment = false, bool _bJitter = false, bool _bContained = false
	) :
		pPolygon2 (_pPolygon2), pElem2 (_pElem2), 
		cPtIdx1 (_cPtIdx1), cPtIdx2 (_cPtIdx2), point (_pt), 
		bSelf (_bSelf), bSuspected (_bSuspected), bPointOnSegment (_bPointOnSegment), bJitter (_bJitter), bContained (_bContained)
	{}

	bool operator < (const intersection_t & _other) const { 
		bool bRet = false;

		if (pElem2->cFileOffset != _other.pElem2->cFileOffset) {
			bRet = pElem2->cFileOffset < _other.pElem2->cFileOffset;
		} else if (pElem2->cFileOffset == _other.pElem2->cFileOffset && (bSelf && _other.bSelf || bSuspected || _other.bSuspected)) {
			const size_t cMin = __min (cPtIdx1, cPtIdx2), cMax = __max (cPtIdx1, cPtIdx2);
			const size_t cOtherMin = __min (_other.cPtIdx1, _other.cPtIdx2), cOtherMax = __max (_other.cPtIdx1, _other.cPtIdx2);

			bRet = 
				cMin < cOtherMin || 
				cMin == cOtherMin && cMax < cOtherMax ||
				cMin == cOtherMin && cMax == cOtherMax && (point.x < _other.point.x || point.x == _other.point.x &&  point.y < _other.point.y);
		} 

		return bRet;
	} 
};

typedef std::set<intersection_t> set_of_intersections_t;
typedef std::multimap<rgn_element_t *, set_of_intersections_t> intersections_t;

struct grid_cell_t {
	float x0, y0;               // upper-left coords of cell
	size_t nSegments, nElems;   // indices in m_cellSegmentVectors and m_cellElemVectors
	int nSegmentIdx, nElemIdx;  // indices in m_cellSegmentVectors[nSegments] and m_cellElemVectors[nElems]

	grid_cell_t () :
		x0 (0), y0 (0), nSegments (0), nElems (0), nSegmentIdx (-1), nElemIdx (-1)
	{} 

	grid_cell_t (float _x0, float _y0) : 
		x0 (_x0), y0 (_y0), nSegments (0), nElems (0), nSegmentIdx (-1), nElemIdx (-1)
	{} 
};

typedef fvector<polygon_segment_t> polygon_segments_t;
typedef fvector<polygon_segments_t> polygon_segment_vectors_t;
typedef fvector<elem_refs_t> elem_ref_vectors_t;

typedef fvector<elem_chain_t> containing_elem_chains_t;
typedef std::list<size_t> size_list_t;
typedef std::map<size_t, size_list_t> size_list_map_t;


class CPolygonIntersections { 
  public :
	CPolygonIntersections () :
		m_bUseGrid (true),
		m_fCellHeight (0), m_fCellWidth (0),
		m_ppCells (NULL),
		m_lRows (0), m_lColumns (0)
	{}

	bool PolygonIntersectsSelf (const rgn_t & _polygon);
	bool PolygonContourIntersectsSelf (const rgn_element_t & _elem);
	void VerifyPolygonOverlapping (
		const rgns_t & _polygons, size_t _cLevel, bool _bCheckSelfInters, bool _bCheckMutualInters,
		intersections_t & _contourIntersections, intersections_t & _internalOverlapings
	);
	void  FindPolygonContourIntersections (		// we have 2 methods, one taking list of polygon elements, 
		const rgn_t & _polygon, const rgn_elements_t & _elements, 
		bool _bCheckSelfIntersections, bool _bCheckMutualIntersections, 
		intersections_t & _intersections, bool _bReportProgress, size_t _cLevel = -1
	);
	void FindPolygonContourIntersections (								// the other one list of polygons, processing all elemnts in polygons
		const rgns_t & _polygons,
		bool _bCheckSelfIntersections, bool _bCheckMutualIntersections, 
		intersections_t & _intersections, bool _bReportProgress
	);

	void ReportIntersections (intersections_t & _intersections);

  private:
	bool   m_bUseGrid;
	rect_t m_rectBound;
	float  m_fCellHeight;
	float  m_fCellWidth;
	bool   m_bReportProgress;

	// Grid cells store the segments that cross them in m_cellSegmentVectors. 
	// A cell refers to its 1D vector via grid_cell_t.nSegments. The cell's segments' starting 
	// position in the vector m_cellSegmentVectors[cell.nSegments] is cell.nSegmentIdx
	// | segment | segment | segment |
	// | segment segment | segment segment | segment segment |
	// | segment segment segment | segment segment segment | segment segment segment |
	// ...
	polygon_segment_vectors_t  m_cellSegmentVectors; 
	elem_ref_vectors_t         m_cellElemVectors;	 // stores elements similar to m_cellSegmentVectors
	elem_refs_t                m_elemRefs;

	intmap_t				   m_segmVecCounters;
	intmap_t				   m_elemVecCounters;

	grid_cell_t ** m_ppCells;
	long m_lRows, m_lColumns;

	void _CountPolygonVertices (const rgns_t & _polygons, size_t & _cPoints, size_t _cLevel);
	void _PreapreSegmentData ();
	void _PrepareGrid (const rgns_t & _polygons, size_t _cLevel);
	void _AddElementsToRefVector (const rgn_t & _polygon, const rgn_elements_t & _elements, size_t _cLevel, size_t & _cCurrElemRefIdx, size_t & _cCellsToAlloc, bool _bAllocOnly);
	void _FormElemRefVector (const rgn_t & _polygon, const rgn_elements_t & _elements, bool _bAllocOnly, size_t _cLevel);
	void _FormElemRefVector (const rgn_t & _polygon, const rgn_elements_t & _elements, size_t _cLevel);
	void _FormElemRefVector (const rgns_t & _polygons, bool _bAllocOnly, size_t _cLevel);
	void _FormElemRefVector (const rgns_t & _polygons, size_t _cLevel);
	void _AllocGridCells (float _fMapWidth, float _fMapHeight);
	void _AllocDataStruct (intmap_t & _vecCounters, RGN_KIND _rgnKind);

	void _CoordToCellIndices (float _fX, float _fY, long & _lColumn, long & _lRow);
	void _AddSegmentToCell (grid_cell_t & _cell, const polygon_segment_t & _segment, bool _bCountOnly);
	void _AddSegmentToCells (
		const size_t _eltRefIdx, 
		const size_t _cIdx1, const size_t _cIdx2, 
		bool _bCountOnly
	);
	void _AddElemToCells (elem_ref_t & _elemRef, bool _bCountOnly);
	void _PopulateGrid (RGN_KIND _rgnKind, bool _bCountOnly);
	bool _AddIntersection (intersections_t & _intersections, rgn_element_t * _pElem, const intersection_t & _inters);
	bool _AddIntersection ( 
		intersections_t & _intersections,
		const size_t _cSgt1ElemRefIdx, const size_t _cSgt2ElemRefIdx,
		size_t _cPtIdx1, size_t _cPtIdx2, 
		const point_t & _pt,
		bool _bSelf, bool _bSuspected = false, bool _bPointOnSegment = false
	);
	bool _CheckAddSegmentIntersections (
		intersections_t & _intersections, intersections_t & _suspectedIntersections,
		const size_t _cElemRefIdx1, const size_t _cElemRefIdx2,
		const points_t & _points1, const points_t & _points2,
		const size_t _cPtIdx11, const size_t _cPtIdx12, const size_t _cPtIdx21, const size_t _cPtIdx22 
	);
	bool _IntersectionExists (size_t _eltRefIdx1, size_t _eltRefIdx2, intersections_t & _intersections);
	size_t _PrevPtIdx (points_t & _points, int _nPtIdx);
	size_t _NextPtIdx (points_t & _points, size_t _cPtIdx);
	bool _IsRightOrientedAt (points_t & _points, int _nPtIdx, bool _bBackwardReverseDirection);
	bool _IsStraightDoublePassEndpoint (points_t _points, int _nForwardPtIdx, int _nBackwardPtIdx, bool _bIsEnd, bool _bBackwardReverse);
	bool _FindDoublePass (points_t _points, size_list_map_t & _mpCoincidingPointLists, int _nStartPtIdx1, int & _nEndPtIdx1, int & _nEndPtIdx2);
	bool _PointBelongsToSegment (const point_t _ptSeg0, const point_t _ptSeg1, const point_t _pt) const;
	bool _TestPointOnSegmentIntersection (const intersection_t & _inters, const points_t & _points1, const points_t & _points2, size_t _cIdx1, size_t _cIdx2) const;

	void _AddSelfIntersectionsAtVertices (
		rgn_element_t * _pElem, 
		set_of_intersections_t & _setCoincidingPoints, bool _bStoreAllIntersections, 
		intersections_t & _intersections
	);
	void _FindIntersectionsQuadratic (
		intersections_t & _intersections, intersections_t & _suspectedIntersections,
		bool _bCheckSelfInters, bool _bCheckMutualInters
	);
	void _FindIntersectionsInGrid (
		intersections_t & _intersections, intersections_t & _suspectedIntersections,
		bool _bCheckSelfInters, bool _bCheckMutualInters
	);
	void _TestSuspectedIntersections (intersections_t & _suspInters, intersections_t & _intersections, bool _bCheckSelfInters, bool _bCheckOverlapping);
	int  _PolygonElemContainsPolygonElem (const rgn_element_t & _extElem, const rgn_element_t & _intElem);
	void _TestReportContainedElement (intersections_t & _intersections, const elem_ref_t * _pIntElemRef, const elem_chain_t & _containingChain);
	void _FindContainedPolygonsInGrid (intersections_t & _intersections, intersections_t & _existingIntersections);
	void _FindContourIntersections (
		//const rgns_t & _polygons, size_t _cLevel, 
		bool _bCheckSelfInters, bool _bCheckMutualInters, 
		intersections_t & _intersections, bool _bReportProgress
	);
	void _FindInternalOverlappings (intersections_t & _existingIntersections, intersections_t & _overlappings);
	void _FreeContainmentData ();
	void _FreeSegmentData ();
	void _ClearGrid ();
	void _ClearElemRefVector ();
	void _SetProgressStatus (int _nProgress);
}; 

# endif