//
// GPSMapEdit
// (c) Konstantin Galichsky (kg@geopainting.com), 2002-2009
//
// Positioning stuff.
//

# include "StdAfx.h"
# include "Positioning.h"
# include "Tracking.h"
# include "GpsPanel.h"
# include "ComPort.h"
# include "GpsProtocol.h"
# include "NMEA0183.h"
# include "GarminProtocol.h"
# include "GarminUSB.h"
# include "Map.h"
# include "Load.h"
# include "Settings.h"
# include "Globals.h"

static string_t g_strStatus;
static PositionInfo_t g_pi, g_piLastInTrack;
static CCritSec g_csPI;

void SetPositioningStatus (const string_t & _strStatus) {
	{
		CCritSec::CLock lock (g_csPI);
		g_strStatus = _strStatus;
	}
	UpdateGPSPanel ();
}

void GetPositioningStatus (string_t & _ret) {
	CCritSec::CLock lock (g_csPI);
	_ret = g_strStatus;
}

void PushPositionInfo (const PositionInfo_t & _pi) {
	bool bAddPointToTrack = false;
	{
		CCritSec::CLock lock (g_csPI);

		if (
			(g_piLastInTrack.dwDef & PositionInfo_t::defXY) != 0 &&
			(_pi.dwDef             & PositionInfo_t::defXY) != 0
		) {
			float fXPrec = 0, fYPrec = 0;
			if (_pi.dwDef & PositionInfo_t::defXYPrec) {
				fYPrec = 360.*_pi.fXYPrec/40e6;
				fXPrec = fYPrec/::cos (_pi.fY*c_PI/180.f);
			}

			if (
				::fabs (g_piLastInTrack.fX - _pi.fX) > fXPrec ||
				::fabs (g_piLastInTrack.fY - _pi.fY) > fYPrec
			)
				bAddPointToTrack = true;
		} else if (
			(g_piLastInTrack.dwDef & PositionInfo_t::defXY) == 0 &&
			(_pi.dwDef             & PositionInfo_t::defXY) != 0
		)
			// TODO: start new track
			bAddPointToTrack = true;

		g_pi = _pi;
		if (bAddPointToTrack)
			g_piLastInTrack = _pi;
	}
	UpdateGPSPanel ();

	if (bAddPointToTrack)
		AddPointToTrack (_pi);
}

void PushNoPositionInfo () {
	StartNewTrack ();
	{
		CCritSec::CLock lock (g_csPI);
		g_pi = PositionInfo_t ();
		g_piLastInTrack = g_pi;
	}
	UpdateGPSPanel ();
}

PositionInfo_t GetPositionInfo () {
	CCritSec::CLock lock (g_csPI);
	return g_pi;
}

/////////////////////////////////////////////

static bool g_bStopCom = true;
static std::set<HANDLE> g_hComThreads;
static HANDLE g_hStopEvent = NULL;
static HANDLE g_hReadyEvent = NULL;
static LONG g_lThreads;

static
DWORD WINAPI _ComThreadProc (LPVOID _lpParameter) {
	//
	// Open COM port.
	//
	const size_t cComPort = reinterpret_cast<size_t> (_lpParameter);
//	assert (1 <= cComPort && cComPort <= 19);
	char strComPort [12];
	if (1 <= cComPort && cComPort <= 9)
		::_snprintf (strComPort, ARRAY_SIZE (strComPort) - 1, "COM%d", cComPort);
	else if (10 <= cComPort && cComPort <= 199)
		::_snprintf (strComPort, ARRAY_SIZE (strComPort) - 1, "\\\\.\\COM%d", cComPort);
	else {
		assert (false);
		::InterlockedDecrement (& g_lThreads);
		return 10;
	}

	CComPort port (strComPort);
	if (! port.IsValid ()) {
		if (::InterlockedDecrement (& g_lThreads) == 0)
			SetPositioningStatus ("Unable to open any COM port.");
		return 1;
	}

	// Ignore modems, faxes, scanners etc.
	COMMPROP cp;
	::memset (& cp, 0, sizeof (cp));
	::GetCommProperties (port, & cp);
//	ATLTRACE ("COM%d 0x%X\n", cComPort, cp.dwProvSubType);
	if (cp.dwProvSubType != PST_RS232 && cp.dwProvSubType != PST_UNSPECIFIED) {
		::InterlockedDecrement (& g_lThreads);
		return 11;
	}

	//
	// Try protocols.
	//
	CNMEA0183 NMEA_protocol;
	CGarminProtocol Garmin_protocol;
	CGpsProtocol * const pProtocols [] = {& NMEA_protocol, & Garmin_protocol};
	const size_t cProtocols = ARRAY_SIZE (pProtocols);

	size_t cProtocolIndex = 0;
	while (! g_bStopCom) {
		assert (cProtocolIndex < cProtocols);
		CGpsProtocol * const pProtocol = pProtocols [cProtocolIndex];
		++ cProtocolIndex;
		cProtocolIndex %= cProtocols;

		const bool bExit = pProtocol->Run (& port, cComPort, g_hStopEvent, g_hReadyEvent);
		if (bExit)
			break;

		PushNoPositionInfo ();
	}

	::InterlockedDecrement (& g_lThreads);
	return 0;
}

///////////////////////////////////////////////


/*static
_GetAvailableComPorts () {
	const CDevInfoHandle hDevInfo = ::SetupDiGetClassDevs (
		& GUID_DEVINTERFACE_COMPORT,
		NULL, NULL, DIGCF_PRESENT | DIGCF_INTERFACEDEVICE
	);

	for (size_t cDev = 0; true; ++ cDev) {
		SP_DEVICE_INTERFACE_DATA InterfaceData;
		InterfaceData.cbSize = sizeof (InterfaceData);
		if (
			! ::SetupDiEnumDeviceInterfaces (
				hDevInfo,
				NULL,
				& GUID_DEVINTERFACE_COMPORT,
				cDev,
				& InterfaceData
			)
		) {
			if (::GetLastError () == ERROR_NO_MORE_ITEMS)
				break;
		}

		DWORD dwDetailDataSize = 0;
		::SetupDiGetDeviceInterfaceDetail (
			hDevInfo,
			& InterfaceData,
			NULL,
			0,
			& dwDetailDataSize,
			NULL
		);

		PSP_INTERFACE_DEVICE_DETAIL_DATA pDevDetailData = (PSP_INTERFACE_DEVICE_DETAIL_DATA) alloca (dwDetailDataSize);
		pDevDetailData->cbSize = sizeof (SP_INTERFACE_DEVICE_DETAIL_DATA);

		SP_DEVINFO_DATA DevInfoData = {sizeof (SP_DEVINFO_DATA)};
		::SetupDiGetDeviceInterfaceDetail (
			hDevInfo,
			& InterfaceData,
			pDevDetailData,
			dwDetailDataSize,
			NULL,
			& DevInfoData
		);

		char strBuf [1000];
		DWORD dwPropertyRegDataType = 0;
		DWORD dwRequiredSize = 0;
		// NOTE: not accessible under Win98!
//		::SetupDiGetCustomDeviceProperty (hDevInfo, & DevInfoData, "PortName", 0, & dwPropertyRegDataType, (BYTE *) strBuf, 1000, & dwRequiredSize);
		// Instead, call ::SetupDiOpenDeviceInterfaceRegKey (hDevInfo, & InterfaceData, 0, KEY_QUERY_VALUE);

//		ATLTRACE ("dev %s %s\n", strBuf, pDevDetailData->DevicePath);
	}
}*/

/////////////////////////////////////////////////////////////////////

static
DWORD WINAPI _PlayTrackThreadProc (LPVOID _lpParameter) {
	char strTrackFileName [MAX_PATH];
	GetSetting ("PlayTrack", strTrackFileName, MAX_PATH, NULL);
	if (strTrackFileName [0] == '\0') {
		::InterlockedDecrement (& g_lThreads);
		return 1;
	}

	const CFileTypeInfo * pSaverInfo = NULL;
	map_t mapTrack;
	const LoadStatus_t ls = CMapLoader::LoadFile (strTrackFileName, true, & mapTrack, & pSaverInfo);
	if (ls != lsOK) {
		SetPositioningStatus ("Failed to play track.");
		::InterlockedDecrement (& g_lThreads);
		return 2;
	}

	for (const rgn_t * pRgn = mapTrack.pFirstRgn; pRgn; pRgn = pRgn->pNextRgn) {
		if (! pRgn->IsTrack () || pRgn->elements.empty ())
			continue;

		const rgn_element_t & element = pRgn->elements [0];
		if (element.bDeleted)
			continue;

		char strStatus [128];
		::sprintf (strStatus, "Playing track '%s'.", pRgn->strLabel.c_str ());
		SetPositioningStatus (strStatus);

		const points_t & points = element.points;
		const size_t cPoints = points.size ();
		const trackpoints_t & trackpoints = element.pExtra->TrackPoints;
		const size_t cTrackPoints = trackpoints.size ();
		for (size_t c = 0; c < cPoints && c < cTrackPoints; ++ c) {
			const point_t & pt = points [c];
			const trackpoint_info_t & tpi = trackpoints [c];

			const float fCosY = ::cos (pt.y*c_PI/180);

			PositionInfo_t pi;
			pi.SetXY (pt.x, pt.y);
			if (tpi.wFlags & trackpoint_info_t::defAltitude)
				pi.SetAltitude (tpi.fAltitude);
			if (tpi.wFlags & trackpoint_info_t::defSpeed)
				pi.SetSpeed (tpi.fSpeed);
			else {
				if (
					c > 0 &&
					(tpi                .wFlags & trackpoint_info_t::defTime) != 0 &&
					(trackpoints [c - 1].wFlags & trackpoint_info_t::defTime) != 0
				) {
					const point_t & ptPrev = points [c - 1];
					const float fD = ::hypot (pt.y - ptPrev.y, (pt.x - ptPrev.x)*fCosY)*40e6/360;
					const float fDT = (tpi.dtTime - trackpoints [c - 1].dtTime)*60*60*24;
					const float fSpeed = fD/fDT*3.6f;
					pi.SetSpeed (fSpeed);
				}
			}
			if (tpi.wFlags & trackpoint_info_t::defHeading)
				pi.SetHeading (tpi.fHeading);
			else {
				if (c > 0) {
					const point_t & ptPrev = points [c - 1];
					float fHeading = 90 - ::atan2 (pt.y - ptPrev.y, (pt.x - ptPrev.x)*fCosY)*180/c_PI;
					if (fHeading < 0)
						fHeading += 360;
					pi.SetHeading (fHeading);
				}
			}
			pi.SetFix (PositionInfo_t::fixSimulation);
			PushPositionInfo (pi);

			if (g_bStopCom)
				break;
			::WaitForSingleObject (g_hStopEvent, 1000/GetSetting ("PlayTrackSpeed", DWORD (2)));
			if (g_bStopCom)
				break;
		}

		if (g_bStopCom)
			break;

		PushNoPositionInfo ();
	}

	SetPositioningStatus ("End of track reached.");

	::InterlockedDecrement (& g_lThreads);

	return 0;
}

/////////////////////////////////////////////////////////////////////

# include "Reg.h"
# include "ParseHelpers.h"

void StartGPSUnitCommunication () {
	StopGPSUnitCommunication ();

	if (g_hReadyEvent == NULL)
		g_hReadyEvent = ::CreateEvent (NULL, TRUE, FALSE, NULL);
	::ResetEvent (g_hReadyEvent);

	if (g_hStopEvent == NULL)
		g_hStopEvent = ::CreateEvent (NULL, TRUE, FALSE, NULL);
	::ResetEvent (g_hStopEvent);
	g_bStopCom = false;

	//
	// Play demo track, if any.
	//
	if (IsSettingDefined ("PlayTrack")) {
		::InterlockedIncrement (& g_lThreads);

		DWORD dwId = 0;
		g_hComThreads.insert (::CreateThread (NULL, 0, _PlayTrackThreadProc, 0, 0, & dwId));

		return;
	}

	//
	// Connect to Garmin GPSr over USB, if any.
	//

	const HANDLE hUsbThread = InitGarminUSB (g_hStopEvent);
	if (hUsbThread != NULL) {
		g_hComThreads.insert (hUsbThread);
		return;
	}

	//
	// Determine the list of availale COMM ports.
	//

	typedef std::set<size_t> setComPorts_t;
	setComPorts_t setComPorts;

	CRegistryKey hKey (HKEY_LOCAL_MACHINE, "Hardware\\DeviceMap\\SerialComm", KEY_QUERY_VALUE);
	if (hKey.IsValid ()) {
		for (size_t cIndex = 0; true; ++ cIndex) {
			char strValueName [MAX_PATH];
			DWORD dwValueNameLen = ARRAY_SIZE (strValueName);
			DWORD dwType = 0;
			char strValue [MAX_PATH];
			DWORD dwValueLen = sizeof (strValue);
			if (
				::RegEnumValue (
					hKey, cIndex,
					strValueName, & dwValueNameLen,
					NULL, & dwType,
					(BYTE *) strValue, & dwValueLen
				) != ERROR_SUCCESS
			)
				break;
			
			if (dwType != REG_SZ)
				continue;
			if (! HasPrefix (strValue, strValue + dwValueLen, "com", 3))
				continue;

			const int nPort = ::atoi (strValue + 3);
			if (nPort != 0)
				setComPorts.insert (nPort);
		}
	}

	if (setComPorts.empty ()) {
		for (size_t cPort = 1; cPort <= 19; ++ cPort)
			setComPorts.insert (cPort);
	}

	//
	// Run theads to scan the COMM ports.
	//

	g_lThreads = 0;
	for (setComPorts_t::const_iterator i = setComPorts.begin (); i != setComPorts.end (); ++ i) {
		const size_t cPort = * i;

		::InterlockedIncrement (& g_lThreads);

		DWORD dwId = 0;
		g_hComThreads.insert (::CreateThread (NULL, 0, _ComThreadProc, reinterpret_cast<LPVOID> (cPort), 0, & dwId));
	}
}

void StopGPSUnitCommunication () {
	g_bStopCom = true;
	if (g_hStopEvent)
		::SetEvent (g_hStopEvent);

	if (! g_hComThreads.empty ()) {
		const size_t cThreads = g_hComThreads.size ();
		HANDLE * const pThreads = (HANDLE *) alloca (sizeof (HANDLE)*cThreads);
		std::set<HANDLE>::iterator i;
		size_t c = 0;
		for (i = g_hComThreads.begin (); i != g_hComThreads.end (); ++ i)
			pThreads [c ++] = * i;

		::WaitForMultipleObjects (cThreads, pThreads, TRUE, 2000);

		for (i = g_hComThreads.begin (); i != g_hComThreads.end (); ++ i)
			::CloseHandle (* i);
		g_hComThreads.clear ();
	}
}

bool IsGPSUnitActive () {
	return ! g_bStopCom;
}
