# ifndef __PtFinder_H_
# define __PtFinder_H_

# include "PtAlgo.h"
# include "fast_floor.h"
# include "Globals.h"

template <typename _T>
class CPointFinder {
  public:
	CPointFinder () : m_fCosY (1), m_nW (1), m_nH (1) {}

	void Add (const point_t & _pt, const _T & _data);
	void Index ();
	const _T * Find (const point_t & _pt, float _fE) const;

  private:
	struct item_t {
		point_t pt;
		_T      data;
	};
	typedef std::list<item_t> items_t;
	items_t m_items;

	rect_t m_rectItems;
	float m_fCosY;
	int m_nW, m_nH;

	typedef std::vector<item_t> cell_t;
	typedef std::vector<cell_t> cells_t;
	cells_t m_cells;
};

template <typename _T>
void CPointFinder<_T>::Add (const point_t & _pt, const _T & _data) {
	m_items.push_back (item_t ());
	item_t & item = m_items.back ();
	item.pt   = _pt;
	item.data = _data;

	m_rectItems.Extend (_pt);
}

template <typename _T>
void CPointFinder<_T>::Index () {
	m_cells.clear ();

	const size_t cItems = m_items.size ();
	if (cItems == 0)
		return;

	const double fW = m_rectItems.Width  ();
	const double fH = m_rectItems.Height ();
	m_fCosY = ::cos (m_rectItems.MeanY ()*c_PI/180);

	// TODO: take aspect ratio into account.
	const int nN = __max (fast_floor (::sqrt ((float) cItems)), 1);
	m_nW = fW != 0 ? nN : 1;
	m_nH = fH != 0 ? nN : 1;

	m_cells.resize (m_nW*m_nH);
	for (items_t::const_iterator i = m_items.begin (); i != m_items.end (); ++ i) {
		const item_t & item = * i;
		const size_t cX = __min (fW != 0 ? fast_floor_d ((item.pt.x - m_rectItems.x0)*m_nW/fW) : 0, m_nW - 1);
		const size_t cY = __min (fH != 0 ? fast_floor_d ((item.pt.y - m_rectItems.y0)*m_nH/fH) : 0, m_nH - 1);

		assert (cX + cY*m_nW < m_cells.size ());
		cell_t & cell = m_cells [cX + cY*m_nW];
		if (cell.empty ())
			cell.reserve (2);
		cell.push_back (item);
	}

	m_items.clear ();
}

template <typename _T>
const _T * CPointFinder<_T>::Find (const point_t & _pt, float _fE) const {
	assert (_fE >= 0);

	if (m_cells.empty ())
		return NULL;

	const double fW = m_rectItems.Width  ();
	const double fH = m_rectItems.Height ();

	assert (m_nW == 1 || fW/m_nW >= _fE/m_fCosY);
	assert (m_nH == 1 || fH/m_nH >= _fE);

	const int nX = __min (fW != 0 ? fast_floor_d ((_pt.x - m_rectItems.x0)*m_nW/fW) : 0, m_nW - 1);
	const int nY = __min (fH != 0 ? fast_floor_d ((_pt.y - m_rectItems.y0)*m_nH/fH) : 0, m_nH - 1);

	const rect_t rectE (_pt.x - _fE/m_fCosY, _pt.y - _fE, _pt.x + _fE/m_fCosY, _pt.y + _fE);

	for (int nx = -1; nx <= 1; ++ nx) {
		const int x = nX + nx;
		if (x < 0 || x >= (int) m_nW)
			continue;
		for (int ny = -1; ny <= 1; ++ ny) {
			const int y = nY + ny;
			if (y < 0 || y >= (int) m_nH)
				continue;

			if (fW != 0 && fH != 0) {
				const rect_t rect (
					m_rectItems.x0 + x      *fW/m_nW, m_rectItems.y0 + y      *fH/m_nH,
					m_rectItems.x0 + (x + 1)*fW/m_nW, m_rectItems.y0 + (y + 1)*fH/m_nH
				);
				rect_t rectIntersection;
				rect_t::Intersection (rectIntersection, rect, rectE);
				if (rectIntersection.Empty ())
					continue;
			}

			assert (x + y*m_nW < (int) m_cells.size ());
			const cell_t & cell = m_cells [x + y*m_nW];

			const size_t cItems = cell.size ();
			for (size_t c = 0; c < cItems; ++ c) {
				const item_t & item = cell [c];
				if (PointContainsPt (item.pt, _pt, _fE, m_fCosY))
					return & item.data;
			}
		}
	}

	return NULL;
}

# endif // __PtFinder_H_