//
// GPSMapEdit
// (c) Konstantin Galichsky (kg@geopainting.com), 2009
//
// Tools to remove jitters in polygons.
//

# include "StdAfx.h"
# include "Resource.h"
# include "RemoveJitters.h"
# include "MapAction.h"
# include "Map.h"
# include "MapView.h"
# include "Status.h"
# include "Globals.h"

static
size_t _RemoveJitters (
	const point_t * _pPointsIn,  size_t _cPointsIn,
	point_t       * _pPointsOut, size_t _cPointsOut,
	bool _bClosed
) {
	// Output 0-th point.
	size_t c2 = 0;
	if (! (_bClosed && _pPointsIn [_cPointsIn - 1] == _pPointsIn [1]))
		if (c2 < _cPointsOut)
			_pPointsOut [c2 ++] = _pPointsIn [0];

	// Output rest of points.
	const size_t cPointsIn = _bClosed ? _cPointsIn : _cPointsIn - 1;
	for (size_t c = 1; c < cPointsIn; ++ c) {
		const point_t & ptPrev = _pPointsIn [c - 1];
		const point_t & ptNext = _pPointsIn [c + 1 < _cPointsIn ? c + 1 : 0];

		if (ptPrev == ptNext) {
			++ c;
			continue;
		}

		if (c2 < _cPointsOut)
			_pPointsOut [c2 ++] = _pPointsIn [c];
	}
	return c2;
}

bool RemoveJitters (points_t & _points, bool _bClosed) {
	bool bRemoved = false;

	const size_t cAttempts = _points.size ()/2;
	for (size_t c = 0; c < cAttempts; ++ c) {
		const size_t cPoints = _points.size ();
		points_t pointsNew (cPoints);
		const size_t cPointsNew = _RemoveJitters (& * _points.begin (), cPoints, & * pointsNew.begin (), cPoints, _bClosed);
		if (cPointsNew == cPoints)
			break;

		bRemoved = true;
		pointsNew.resize (cPointsNew);

		RemoveRepeatingPoints (pointsNew, _bClosed);

		_points.swap (pointsNew);
	}

	return bRemoved;
}

////////////////////////////////////////////////////////////////////

static
bool _DoRemoveJittersInPolygons (bool _bSilent) {
	CCritSec::CLock loc (g_cs);

	const size_t cTotalRgns = g_map.layers [rkPolygon].cRgns;
	if (cTotalRgns == 0)
		return false;
	size_t cRgnsProcessed = 0;

	SetStatus (GetStr (IDS_STATUS_REMOVING));
	SetProgressStatus (0);

	size_t cJitters = 0;

	const bool bHuge = g_map.cMaxPointsInElement > 30000;
	point_t * const pPoints = bHuge ? new point_t [g_map.cMaxPointsInElement] : reinterpret_cast<point_t *> (::alloca (sizeof (point_t)*g_map.cMaxPointsInElement));
	CArrayHolder<point_t> pPoints_Holder (bHuge ? pPoints : NULL);

	rgns_t & polygons = g_map.layers [rkPolygon].rgns;
	for (rgns_t::iterator iRgn = polygons.begin (); iRgn != polygons.end (); ++ iRgn) {
		rgn_t & polygon = * iRgn;
		if (polygon.IsImmutable () || polygon.IsNotMapObject ())
			continue;

		const size_t cElements = polygon.elements.size ();
		for (size_t c = 0; c < cElements; ++ c) {
			rgn_element_t & element = polygon.elements [c];
			const size_t cPoints = element.points.size ();
			assert (cPoints <= g_map.cMaxPointsInElement);

			if (RemoveJitters (element.points, true))
				++ cJitters;
		}

		++ cRgnsProcessed;
		SetProgressStatus (cRgnsProcessed*100/cTotalRgns);
	}

	polygons.sort ();

	SetStatusReady ();

	if (! _bSilent)
		InfoMsgBox ("Removed jitters in %d polygon(s).", cJitters);

	return true;
}

void RemoveJittersInPolygons () {
	PerformMapAction (_DoRemoveJittersInPolygons, "remove jitters", false);
	Refresh ();
}
