//
// GPSMapEdit
// (c) Konstantin Galichsky (kg@geopainting.com), 2009
// (c) Eugene Prosso        (eprosso@navitel.su), 2009
//
// Tools to remove self-intersections in polygons.
//

# include "StdAfx.h"
# include "Resource.h"
# include "RemoveSelfIntersections.h"
# include "PolygonIntersections.h"
# include "MapAction.h" 
# include "Map.h"
# include "MapView.h"
# include "Status.h"
# include "Distance.h"
# include "PtAlgo.h"
# include "fast_floor.h"
# include "Globals.h"

inline float _VecLength2 (const point_t & _pt) { 
	return _pt.x*_pt.x + _pt.y*_pt.y;
} 

struct pgnode_t;
typedef std::set<pgnode_t> pgnodes_t;

struct pgnode_t {
	point_t m_point;
	point_t pNeighbors [20];
	bool    pVisited [20];
	size_t  cNeighbors;

	static int s_cStatID;
	int cID;

	pgnode_t () { 
		Init (point_t (200, 200));
	} 

	pgnode_t (const point_t & _pt) { 
		Init (_pt);
	} 

	void Set (const point_t & _pt) { 
		m_point = _pt;
	} 

	size_t Find (const point_t & _pt) const { 
		point_t * pRet = NULL;
		for (size_t c = 0; c < cNeighbors; ++ c)
			if (pNeighbors [c] == _pt)
				return c;
		return -1;
	} 

	void Add (const point_t & pt)	{ 
		if (cNeighbors < 20)
			pNeighbors [cNeighbors ++] = pt;
		else
			assert (0);
	} 

	void Delete (size_t _cIdx) { 
		assert (_cIdx < cNeighbors);
		if (_cIdx != cNeighbors - 1) { 
			pNeighbors [_cIdx] = pNeighbors [cNeighbors - 1];
			pVisited    [_cIdx] = pVisited    [cNeighbors - 1];
		} 

		-- cNeighbors;
	} 

	void Delete (const point_t & _pt) { 
		Delete (Find (_pt));
	}

	bool operator == (const pgnode_t & _other) const { 
		return (m_point == _other.m_point);
	} 

	bool operator < (const pgnode_t & _other) const { 
		return (m_point.x != _other.m_point.x) ? 
			(m_point.x < _other.m_point.x) : (m_point.y < _other.m_point.y);
	} 

  private: 
	void Init (const point_t & _pt) {
		m_point = _pt;
		cID = s_cStatID ++ ;
		cNeighbors = 0;
		::memset (pNeighbors, 0, 20*sizeof (point_t));
		::memset (pVisited,   0, 20*sizeof (bool));
	}
};

int pgnode_t::s_cStatID = 0;

typedef std::set<pgnode_t> pgnodes_t;

struct insertion_t { 
	std::list<intersection_t> intersections;
	void Sort () { 
		intersections.sort ();
	} 
};

static size_t _NextPtIdx (points_t & _points, size_t _cIdx) { 
	return (_cIdx == 0) ? _points.size () - 1 : _cIdx - 1;
} 

static size_t _PrevPtIdx (points_t & _points, size_t _cIdx) { 
	return (_cIdx == _points.size () - 1) ? 0 : _cIdx + 1;
} 
typedef std::map<float, pgnode_t> segments_pgnodes_t;

static long lTicksGraphWalk4 = 0;

bool RemoveSelfIntersections (
	const point_t * _pPointsIn, size_t _cPointsIn,
	point_t       * _pointsOut, size_t & _cPointsOut, size_t _cMaxPointsOut,
	bool _bClosed
) { 
	// TODO - implement
	return false;
}

bool RemoveSelfIntersections (
	const point_t * _pPointsIn,  size_t _cPointsIn,
	points_t & _pointsOut,
	bool _bClosed
) { 
	size_t cPointsOut = 0;

	rgn_t polygon;
	rgn_element_t dummyElem; 

	pgnode_t::s_cStatID = 0;

	for (size_t c = 0; c < _cPointsIn; ++ c) {
		dummyElem.points.push_back (_pPointsIn [c]);
		dummyElem.rectBound.Extend (_pPointsIn [c]);
	} 

	dummyElem.cLevel = 0;

	polygon.elements.push_back (dummyElem);
	rgn_element_t & elem = polygon.elements [0];

	CPolygonIntersections intersProcessor;
	intersections_t intersections;

	intersProcessor.FindPolygonContourIntersections (polygon, polygon.elements, true, false, intersections, false, 1);

	const bool bHasIntersections = ! intersections.empty ();
	if (bHasIntersections) { 
		// Nodes in the same sequence as in _pPointsIn, plus nodes from intersection points in the sequence 
		// of encountering intersections while following _pPointsIn. 2 nodes per intersection are added.
		std::multimap<float, pgnode_t> mmSequentialNodes;	

		// Graph of nodes: one node per point, no matter how many times the polygon passes through it.
		pgnodes_t graph; 
		pgnode_t * const pInNodes = new pgnode_t [_cPointsIn];

		// Form mmSequentialNodes.
		float fDistFromStart = 0;
		size_t c;
		for (c = 0; c < _cPointsIn; ++ c) { 
			if (c > 0)
				fDistFromStart += Distance (_pPointsIn [c], _pPointsIn [c - 1]);
			pgnode_t & node = pInNodes [c];
			node.Set (_pPointsIn [c]);
			mmSequentialNodes.insert (std::make_pair (fDistFromStart, node));
		}

		delete [] pInNodes;

		// add nodes from intersections to mmSequentialNodes
		segments_pgnodes_t * const pNodesAtSegment = new segments_pgnodes_t [_cPointsIn];
		size_t cIntersections = 0;

		for (intersections_t::iterator iPair = intersections.begin (); iPair != intersections.end (); ++ iPair) { 
			set_of_intersections_t & intersSet = iPair->second;

			for (set_of_intersections_t::iterator iInters = intersSet.begin (); iInters != intersSet.end (); ++ iInters) { 
				const intersection_t & inters = * iInters;

				pgnode_t node (inters.point);														  
				pNodesAtSegment [inters.cPtIdx1].insert (std::make_pair (Distance (inters.point, _pPointsIn [inters.cPtIdx1]), node));
				pNodesAtSegment [inters.cPtIdx2].insert (std::make_pair (Distance (inters.point, _pPointsIn [inters.cPtIdx2]), node));

				++ cIntersections;
			}
		}

		_pointsOut.resize (_cPointsIn + 2*cIntersections);

		pgnode_t * pPrevNode = NULL;
		fDistFromStart = 0;
		for (c = 0; c < _cPointsIn; ++ c) { 
			if (c > 0)
				fDistFromStart += Distance (_pPointsIn [c], _pPointsIn [c - 1]);
			for (segments_pgnodes_t::iterator iND = pNodesAtSegment [c].begin (); iND != pNodesAtSegment [c].end (); ++ iND) { 
				const float fDistToSegmentStart = iND->first;
				pgnode_t & node                 = iND->second;

				mmSequentialNodes.insert (std::make_pair (fDistFromStart + fDistToSegmentStart, node));
			}
		}

		delete [] pNodesAtSegment;

		// Form graph.
		pgnodes_t::iterator iNodePrev  = graph.end ();
		pgnodes_t::iterator iNodeFirst = graph.end ();
		float fDistPrev = -1;
		size_t cNeighborIdx = -1;
		for (std::multimap<float, pgnode_t>::iterator iSN = mmSequentialNodes.begin (); iSN != mmSequentialNodes.end (); ++ iSN) { 
			const float fDist = iSN->first;
			pgnode_t & node   = iSN->second;

			if (fDist != fDistPrev) {
				const pgnodes_t::iterator iNode = graph.insert (node).first;
				pgnode_t & nodeInGraph = const_cast<pgnode_t &> (* iNode);

				if (iNodePrev != graph.end ()) {
					pgnode_t & nodePrevInGraph = const_cast<pgnode_t &> (* iNodePrev);

					const size_t cIdxOfPrev = nodeInGraph.Find (iNodePrev->m_point);
					if (cIdxOfPrev == -1) {
						nodeInGraph    .Add (nodePrevInGraph.m_point);
						nodePrevInGraph.Add (nodeInGraph    .m_point);
					} else { // jitter we have!
						nodeInGraph.Delete (cIdxOfPrev);
						if (nodePrevInGraph.cNeighbors == 1) // jitter into iNodePrev point with no previous edges
							graph.erase (iNodePrev);
						else							 // jitter into iNodePrev point with existing previous edges
							nodePrevInGraph.Delete (nodeInGraph.m_point);
					} 
				} else
					iNodeFirst = iNode;

				iNodePrev = iNode;
				fDistPrev = fDist;
			} 
		}

		pgnode_t & nodeFirstInGraph = const_cast<pgnode_t &> (* iNodeFirst);
		pgnode_t & nodePrevInGraph  = const_cast<pgnode_t &> (* iNodePrev);
		
		const size_t cIdxOfFirst = iNodePrev->Find (iNodeFirst->m_point);
		if (cIdxOfFirst == -1) {
			nodeFirstInGraph.Add (nodePrevInGraph .m_point);
			nodePrevInGraph. Add (nodeFirstInGraph.m_point);
		} else {
			nodePrevInGraph.Delete (cIdxOfFirst);
			if (nodeFirstInGraph.cNeighbors == 1) // jitter into iNodePrev point with no previous edges
				graph.erase (iNodeFirst);
			else							      // jitter into iNodePrev point with existing previous edges
				nodeFirstInGraph.Delete (nodePrevInGraph.m_point);
		}

		// Find external countour formed by graph edges.
		const pgnode_t & nodeFirst = * graph.begin ();

		pgnodes_t::iterator iNodeLeftMost = graph.begin ();
		for (pgnodes_t::iterator in = graph.begin (); in != graph.end (); ++ in)
			if (in->m_point.x < iNodeLeftMost->m_point.x)
				iNodeLeftMost = in;

		point_t pSeqTriangle [3];
		pgnode_t nodeDummyFirst (point_t (iNodeLeftMost->m_point.x, iNodeLeftMost->m_point.y - 1));
		pgnode_t & nodePrev = nodeDummyFirst;

		point_t  prev_pt (iNodeLeftMost->m_point.x, iNodeLeftMost->m_point.y - 1);
		bool bLoopStarted = false;
		pgnodes_t::iterator iNode = iNodeLeftMost;
		
		size_t cLoops = 0;
		while (! bLoopStarted || iNode != iNodeLeftMost) { 
			bLoopStarted = true;
			size_t cBestNextPointIdx = -1;
			double fQuasiCosMax = - 100, fQuasiCos = 0;

			pSeqTriangle [0] = prev_pt;
			pSeqTriangle [1] = iNode->m_point;

			for (size_t c = 0; c < iNode->cNeighbors; ++ c) { 
				if (! iNode->pVisited [c]) { 
					pSeqTriangle [2] = iNode->pNeighbors [c];
					point_t vecBack (pSeqTriangle [0].x - pSeqTriangle [1].x, pSeqTriangle [0].y - pSeqTriangle [1].y);
					point_t vecFwd  (pSeqTriangle [2].x - pSeqTriangle [1].x, pSeqTriangle [2].y - pSeqTriangle [1].y);
					fQuasiCos = ScalarProduct (vecBack, vecFwd)/::sqrt (_VecLength2 (vecBack)*_VecLength2 (vecFwd));
					if (IsRightOriented (pSeqTriangle, 3))
						fQuasiCos = - fQuasiCos - 2;				

					if (fQuasiCos > fQuasiCosMax) { 
						cBestNextPointIdx = c;
						fQuasiCosMax = fQuasiCos;
					}
				}
			}

			const pgnode_t & pgnode = * iNode;

			if (cBestNextPointIdx == -1) { 
				ATLTRACE ("Best next node not found. Num of neighbors: %d\n", iNode->cNeighbors);
				assert (false);
			} 

			const pgnode_t & node = * iNode;
			pgnode_t dummyNext (iNode->pNeighbors [cBestNextPointIdx]);
			const pgnodes_t::iterator iNodeNextInGraph = graph.find (dummyNext);

			const point_t & sought_p = iNode->m_point;
			pgnode_t & nodeNextInGraph = const_cast<pgnode_t &> (* iNodeNextInGraph);
			pgnode_t & nodeInGraph     = const_cast<pgnode_t &> (* iNode);

			const size_t cNodeCurrInNextIdx = iNodeNextInGraph->Find (iNode->m_point);
			assert (cNodeCurrInNextIdx != -1);
			nodeInGraph    .pVisited [cBestNextPointIdx]  = true;
			nodeNextInGraph.pVisited [cNodeCurrInNextIdx] = true;

			prev_pt = iNode->m_point;

			if (cPointsOut == _pointsOut.size ())
				_pointsOut.resize (cPointsOut*2);
			_pointsOut [cPointsOut ++] = iNode->m_point;
			iNode = iNodeNextInGraph;

			++ cLoops;
		} 
	}

	_pointsOut.resize (cPointsOut);
	return bHasIntersections;
}

////////////////////////////////////////////////////////////////////

static
bool _DoRemoveSelfIntersectionsInPolygons (bool _bSilent) {
	CCritSec::CLock loc (g_cs);

	const size_t cTotalRgns = g_map.layers [rkPolygon].cRgns;
	if (cTotalRgns == 0)
		return false;
	size_t cRgnsProcessed = 0;

	SetStatus (GetStr (IDS_STATUS_REMOVING));
	SetProgressStatus (0);

	size_t cSIs = 0;

	const bool bHuge = g_map.cMaxPointsInElement > 30000;
	static point_t * pPoints = NULL;
	if (pPoints == NULL)
		pPoints = bHuge ? new point_t [g_map.cMaxPointsInElement] : reinterpret_cast<point_t *> (::alloca (sizeof (point_t)*g_map.cMaxPointsInElement));

	rgns_t & polygons = g_map.layers [rkPolygon].rgns;
	for (rgns_t::iterator iRgn = polygons.begin (); iRgn != polygons.end (); ++ iRgn) {
		rgn_t & polygon = * iRgn;
		if (polygon.IsImmutable () || polygon.IsNotMapObject ())
			continue;

		const size_t cElements = polygon.elements.size ();
		for (size_t c = 0; c < cElements; ++ c) {
			rgn_element_t & element = polygon.elements [c];
			const size_t cPoints = element.points.size ();
			assert (cPoints <= g_map.cMaxPointsInElement);

			size_t cPointsAfter = 0;
			if (! RemoveSelfIntersections (& * element.points.begin (), cPoints, pPoints, cPointsAfter, g_map.cMaxPointsInElement, true))
				continue;

			assert (cPointsAfter < g_map.cMaxPointsInElement);
			element.points.resize (cPointsAfter);
			std::copy (pPoints, pPoints + cPointsAfter, element.points.begin ());

			++ cSIs;
		}

		++ cRgnsProcessed;
		SetProgressStatus (cRgnsProcessed*100/cTotalRgns);
	}

	polygons.sort ();

	SetStatusReady ();

	if (! _bSilent)
		InfoMsgBox ("%d self-intersection(s) are removed.", cSIs);

	return true;
}

void RemoveSelfIntersectionsInPolygons () {
	PerformMapAction (_DoRemoveSelfIntersectionsInPolygons, "remove self-intersections", false);
	Refresh ();
}

bool RemoveSmallLoops (points_t & _pointsIn, points_t & _pointsOut, float _fThresholdArea) { 
	const size_t cPointsIn = _pointsIn.size ();
	_pointsOut.resize (cPointsIn);

	const size_t cMaxLoopLength = __min (30, fast_floor (0.2f*cPointsIn));
	point_t loop [30];

	ATLTRACE ("\n\n\n\n   IN       RemoveSmallLoops   \n\n\n\n");

	size_t cNext = 1;
	size_t cPointsOut = 0;
	for (size_t cStart = 0; cStart < cPointsIn && cNext != -1 && cNext < cPointsIn; cStart = cNext) {
		const size_t cEnd = cStart + cMaxLoopLength;
		cNext = cStart + 1;

		size_t cMaxCoincPtIdx = -1;
		for (size_t c = cStart + 1; c < cEnd; ++ c) {
			const size_t cCheckPoint = (c >= cPointsIn) ? (c - cPointsIn) : c;
			if (_pointsIn [cCheckPoint] == _pointsIn [cStart]) {
				bool bLoopIsSmall = false;
				ATLTRACE ("cStart = %d, cEnd = %d, cCheckPoint = %d\n", cStart, cEnd, cCheckPoint);
				if (c < cPointsIn)
					bLoopIsSmall = GetArea (& _pointsIn [cStart], (c - cStart)) < _fThresholdArea;
				else {
					::memcpy (& loop [0], & _pointsIn [cStart], (cPointsIn - cStart)*sizeof (point_t));
					::memcpy (& loop [cPointsIn - cStart], & _pointsIn [cStart], (cCheckPoint + 1)*sizeof (point_t));
					bLoopIsSmall = GetArea (loop, (c - cStart)) < _fThresholdArea;
				}

				if (bLoopIsSmall)
					cMaxCoincPtIdx = c;
			}
		}

		if (cMaxCoincPtIdx != -1)
			cNext = cMaxCoincPtIdx + 1;

		_pointsOut [cPointsOut ++] = _pointsIn [cStart];
	} 

	if (cNext > cPointsIn) { 
		_pointsOut.erase (_pointsOut.begin (), _pointsOut.begin () + (cNext - 1 - cPointsIn));
		cPointsOut -= (cNext - 1 - cPointsIn);
	} 

	_pointsOut.resize (cPointsOut);

	return (cPointsOut < cPointsIn);
}
