//
// GPSMapEdit
// (c) Konstantin Galichsky (kg@geopainting.com), 2002-2009
//
// Objects of map.
//

# include "StdAfx.h"
# include "Map.h"
# include "Types.h"
# include "Log.h"
# include "Status.h"
# include "Globals.h"

void rgn_element_t::swap (rgn_element_t & _other) {
	points.swap (_other.points);
	rectBound.swap (_other.rectBound);
	std::swap (cLevel,        _other.cLevel);
	std::swap (cFileOffset,   _other.cFileOffset);
	std::swap (fLengthOrArea, _other.fLengthOrArea);
	std::swap (bHole,         _other.bHole);
	std::swap (bDeleted,      _other.bDeleted);
	pExtra.swap (_other.pExtra);
}

void rgn_origin_t::swap (rgn_origin_t & _other) {
	std::swap (point,         _other.point);
	std::swap (cLevel,        _other.cLevel);
	std::swap (cFileOffset,   _other.cFileOffset);
	std::swap (bDeleted,      _other.bDeleted);
}

bool rgn_element_t::IsHoleOf (const rgn_element_t & _Outer) const {
	if (
		_Outer.cLevel != cLevel ||
		_Outer.fLengthOrArea < 1.e-10f || _Outer.fLengthOrArea < fLengthOrArea ||
		! _Outer.rectBound.Contains (rectBound)
	)
		return false;

	assert (! _Outer.points.empty ());
	assert (! points.empty ());
	for (size_t cInnerPoint = 0; cInnerPoint < points.size (); ++ cInnerPoint) {
		const int iContains = PolygonContainsPt (_Outer.points, points [cInnerPoint]);
		if (iContains < 0)
			return false;
		else if (iContains > 0)
			return true;
	};

	return false;
}

void rgn_element_t::RemoveRepeatingPoints (bool _bClosed) {
	if (! pExtra || pExtra->PointExtras.empty ()) {
		::RemoveRepeatingPoints (points, _bClosed);
		return;
	}

	if (! AreThereRepeatingPoints (points, _bClosed))
		return;

	ReleaseRGNodes ();

	epoints_t epoints; 
	Points2EPoints (points, & pExtra->PointExtras, epoints);
	::RemoveRepeatingPoints (epoints);
	EPoints2Points (epoints, true, points, pExtra->PointExtras);

	AddRefRGNodes ();
}

////////////////////////////////////////////////////////////////

rgn_t::rgn_t () :
	pTypeDesc (NULL),
	dwType (0),
	Kind (rkPoint),
	uchEndLevel (0),
	ucLabelWidthBy8 (0),
	ucLabelHeight (0),
	fSize2 (0),
	fAveElementArea (0),
	cBeginPos (c_cUnknownPos),
	cEndPos   (c_cUnknownPos),
	cFileLine (c_cUnknownLine),
//	cOrdinal (0),
	pNextRgn (NULL),
	dwFlags (0),
	pFileRef (NULL),
	ucExtendedFlags (0),
	bTemporary (0)
{}

bool rgn_t::Contains (const point_t & _point, float _fE, size_t _cLevel) const {
	const float fCosY = ::cos (_point.y*c_PI/180.f);

	if (! elements.empty ()) {
		const bool bPolygon = IsPolygon ();
		for (rgn_elements_t::const_iterator i = elements.begin (); i != elements.end (); ++ i) {
			const rgn_element_t & elm = * i;
			if (elm.bDeleted)
				continue;
			if (_cLevel != -1 && elm.cLevel != _cLevel)
				continue;

			if (! elm.rectBound.Contains (_point))
				continue;

			if (bPolygon) {
				if (PolygonContainsPt (elm.points, _point) >= 0)
					return true;
			} else {
				if (PolylineContainsPt (elm.points, _point, _fE, fCosY))
					return true;
			}
		}
	} else if (! origins.empty ()) {
		for (rgn_origins_t::const_iterator i = origins.begin (); i != origins.end (); ++ i) {
			const rgn_origin_t & org = * i;
			if (org.bDeleted)
				continue;
			if (_cLevel != -1 && org.cLevel != _cLevel)
				continue;

			if (PointContainsPt (org.point, _point, _fE, fCosY))
				return true;
		}
	}

	return false;
}

bool rgn_t::IntersectsSegment (const point_t & _ptStart, const point_t & _ptEnd, size_t _cLevel) const {
	if (! elements.empty ()) {
		const float fCosY = ::cos (_ptStart.y*c_PI/180.f);
		const bool bPolygon = IsPolygon ();

		for (rgn_elements_t::const_iterator i = elements.begin (); i != elements.end (); ++ i) {
			const rgn_element_t & elm = * i;
			if (elm.bDeleted)
				continue;
			if (_cLevel != -1 && elm.cLevel != _cLevel)
				continue;

			if (SegmentIntersectsPolyline (_ptStart, _ptEnd, elm.points, bPolygon))
				return true;
		}
	} 

	return false;
}

rgn_element_t * rgn_t::GetNonDeletedElement (size_t _c, size_t _cLevel) {
	const size_t cElements = elements.size ();
	for (size_t cElement = 0; cElement < cElements; ++ cElement) {
		rgn_element_t & element = elements [cElement];
		if (element.bDeleted)
			continue;
		if (_cLevel != -1 && element.cLevel != _cLevel)
			continue;

		if (_c == 0)
			return & element;
		else
			-- _c;
	}

	return NULL;
}

const rgn_element_t * rgn_t::GetNonDeletedElement (size_t _c, size_t _cLevel) const {
	return const_cast<rgn_t *> (this)->GetNonDeletedElement (_c, _cLevel);
}

size_t rgn_t::GetElements () const {
	size_t cRet = 0;
	const size_t cElements = elements.size ();
	for (size_t c = 0; c < cElements; ++ c) {
		const rgn_element_t & element = elements [c];
		if (! element.bDeleted)
			++ cRet;
	}
	return cRet;
}

size_t rgn_t::GetOrigins () const {
	size_t cRet = 0;
	const size_t cOrigins = origins.size ();
	for (size_t c = 0; c < cOrigins; ++ c) {
		const rgn_origin_t & origin = origins [c];
		if (! origin.bDeleted)
			++ cRet;
	}
	return cRet;
}

size_t rgn_t::GetElementsInLevel (size_t _cLevel) const {
	size_t cRet = 0;
	const size_t cElements = elements.size ();
	for (size_t c = 0; c < cElements; ++ c) {
		const rgn_element_t & element = elements [c];
		if (element.bDeleted)
			continue;

		if (
			element.cLevel != _cLevel &&
			! (this->uchEndLevel != 0 && element.cLevel <= _cLevel && _cLevel <= this->uchEndLevel)
		)
			continue;

		++ cRet;
	}
	return cRet;
}

size_t rgn_t::GetOriginsInLevel (size_t _cLevel) const {
	size_t cRet = 0;
	const size_t cOrigins = origins.size ();
	for (size_t c = 0; c < cOrigins; ++ c) {
		const rgn_origin_t & origin = origins [c];
		if (origin.bDeleted)
			continue;

		if (
			origin.cLevel != _cLevel &&
			! (this->uchEndLevel != 0 && origin.cLevel <= _cLevel && _cLevel <= this->uchEndLevel)
		)
			continue;

		++ cRet;
	}
	return cRet;
}

void rgn_t::RefreshAttributes (map_t & _map, bool _bRefereshDrawOrder) {
	rectBound = rect_t ();

	const bool bIsPolygon = IsPolygon ();

	//
	// Explore elements.
	//
	const size_t cElements = elements.size ();
	if (cElements) {
		// Data used to count holes and average element area.
		float fArea = 0;
		size_t cNonHoleParts = 1;

		for (size_t cElement = 0; cElement < cElements; ++ cElement) {
			rgn_element_t & element = elements [cElement];
			if (element.bDeleted)
				continue;

			const points_t & points = element.points;
			const size_t cPoints = points.size ();

			if (cPoints > _map.cMaxPointsInElement)
				_map.cMaxPointsInElement = cPoints;

			// Determine the bounding rectangle.
			element.rectBound = rect_t ();
			for (size_t cPoint = 0; cPoint < cPoints; ++ cPoint) {
				const point_t & point = points [cPoint];

				element.rectBound.Extend (point);
			}
			rectBound.Extend (element.rectBound);

			if (IsPolyline ()) {
				// NOTE: For polylines only.
				element.fLengthOrArea = GetLength (points);
				element.bHole = false;
			} else if (bIsPolygon) {
				// NOTE: count element's area before hole detection.
				element.fLengthOrArea = GetArea (points);

				// NOTE: bHole is not updated here. Please use RefreshHoles() for this.
				if (element.bHole)
					fArea -= element.fLengthOrArea;
				else
					fArea += element.fLengthOrArea;
			}
		}

		fAveElementArea = bIsPolygon ? fArea/cNonHoleParts : 0;
	}
	if (bIsPolygon && _bRefereshDrawOrder) {
		//
		// Re-order polygons draw order.
		//

		// Search for iterator to this polygon.
		rgns_t & polygons = _map.layers [rkPolygon].rgns;
		rgns_t::iterator iThis;
		for (iThis = polygons.begin (); iThis != polygons.end (); ++ iThis) {
			if (& * iThis == this)
				break;
		}
		if (iThis != polygons.end ()) {
			bool bGreaterThenPrev = false;
			if (iThis != polygons.begin ()) {
				rgns_t::iterator iPrev = iThis;
				-- iPrev;
				bGreaterThenPrev = fAveElementArea > iPrev->fAveElementArea;
			}

			bool bLessThenNext = false;
			rgns_t::iterator iNext = iThis;
			++ iNext;
			if (iNext != polygons.end ())
				 bLessThenNext = fAveElementArea < iNext->fAveElementArea;

			if (bGreaterThenPrev || bLessThenNext) {
				// NOTE: the element should be shifted inside the list to restore sort order.

				// Search for the new position in the list.
				rgns_t::iterator iToBeNext;
				for (iToBeNext = polygons.begin (); iToBeNext != polygons.end (); ++ iToBeNext) {
					if (iToBeNext->fAveElementArea < fAveElementArea)
						break;
				}

				// Shift this polygon to the new position.
				polygons.splice (iToBeNext, polygons, iThis);
			}
		} else
			// NOTE: the polyline appears not in the list!
			assert (false);
	}

	//
	// Explore origins.
	//
	const size_t cOrigins = origins.size ();
	for (size_t cOrigin = 0; cOrigin < cOrigins; ++ cOrigin) {
		const rgn_origin_t & origin = origins [cOrigin];
		if (origin.bDeleted)
			continue;

		// Include origins in the bounding rect.
		rectBound.Extend (origin.point);
	}

	//
	// Explore entry points.
	//
	bool bHasEntryPoints = false;
	if (pExtra) {
		const size_t cEntryPoints = pExtra->entry_points.size ();
		for (size_t cEP = 0; cEP < cEntryPoints; ++ cEP) {
			const entry_point_t & ep = pExtra->entry_points [cEP];
			if (ep.bDeleted)
				continue;

			// Include origins in the bounding rect.
			rectBound.Extend (ep.pt);
			bHasEntryPoints = true;
		}
	}

	_map.rectBound.Extend (rectBound);

	fSize2 = 0;

	//
	// Prepare point for the label, if any.
	//
	if (! strLabel.empty () || bHasEntryPoints) {
		if (cElements)
			fSize2 = GetRgnSize (* this);

		const type_desc_t * const pTD = GetTypeDesc ();
		assert (pTD);
		if (pTD)
			pTD->pfPrepareLabel (* this);
	}
}

void rgn_t::ReorderHoles () {
	const size_t c_cVisitedInner = (size_t) (-1);
	const size_t cElements = elements.size ();

	// Array of indices of outer polygons.
	size_t cOuters = 0;
	size_t * const pcOuters = reinterpret_cast<size_t *> (alloca (cElements*sizeof (size_t)));

	// Array of indices of inner polygons.
	size_t cInners = 0;
	size_t * const pcInners = reinterpret_cast<size_t *> (alloca (cElements*sizeof (size_t)));

	for (size_t cElement = 0; cElement < cElements; ++ cElement) {
		const rgn_element_t & element = elements [cElement];
		if (element.bDeleted)
			continue;

		if (element.bHole)
			pcInners [cInners ++] = cElement;
		else
			pcOuters [cOuters ++] = cElement;
	}
	assert (cOuters <= cElements);
	assert (cInners <= cElements);
	assert (cOuters + cInners <= cElements);

	if (cInners == 0 || cOuters == 0)
		// Nothing to do!
		return;

	// Reorder in temporary vector.
	rgn_elements_t elements2 (cOuters + cInners);
	size_t cIndex = 0;

	for (size_t cOuterElement = 0; cOuterElement < cOuters; ++ cOuterElement) {
		assert (pcOuters [cOuterElement] < cElements);
		assert (cIndex < cOuters + cInners);
		elements2 [cIndex].swap (elements [pcOuters [cOuterElement]]);
		const rgn_element_t & elemOuter = elements2 [cIndex ++];

		for (size_t cInnerElement = 0; cInnerElement < cInners; ++ cInnerElement) {
			if (pcInners [cInnerElement] == c_cVisitedInner)
				// Already visited.
				continue;

			assert (pcInners [cInnerElement] < cElements);
			if (elements [pcInners [cInnerElement]].IsHoleOf (elemOuter)) {
				assert (cIndex < cOuters + cInners);
				elements2 [cIndex ++].swap (elements [pcInners [cInnerElement]]);

				// Mark to avoid second visiting.
				pcInners [cInnerElement] = c_cVisitedInner;
			}
		}
	}
	// NOTE: some inners may appear outside of any of outers.
	if (cIndex < cOuters + cInners) {
		// Copy them to the end of the list.
		for (size_t cInnerElement = 0; cInnerElement < cInners; ++ cInnerElement)
			if (pcInners [cInnerElement] != c_cVisitedInner) {
				assert (cIndex < cOuters + cInners);
				assert (pcInners [cInnerElement] < cElements);
				elements2 [cIndex ++].swap (elements [pcInners [cInnerElement]]);
			}
	}
	assert (cIndex == cOuters + cInners);

	elements.swap (elements2);
}

void rgn_t::SetType (const CTypeSet * _pTypeSet, DWORD _dwType, BYTE _btKind) {
	SetType (
		(_pTypeSet == NULL) ? NULL : _pTypeSet->GetTypeDesc (_dwType, _btKind),
		_dwType, _btKind
	);
}

const type_desc_t * rgn_t::GetTypeDesc () const {
	if (IsTrack () || IsRoute () || IsBookmarkOrNote ())
		return g_map.pTypeSet->GetUnknownTD ();

	if (pTypeDesc)
		return pTypeDesc;

	if (IsWaypoint ())
		return g_map.pTypeSet->GetUnknownTD ();
	else {
		assert (false);
		return NULL;
	}
}

bool rgn_t::operator < (const rgn_t & _other) const {
	return fAveElementArea > _other.fAveElementArea;
}

bool rgn_t::IsDeleted () const {
	return (dwFlags & fDeleted) || (pFileRef && pFileRef->bDeleted);
}

void rgn_t::Delete (map_t & _map, bool _bReleaseRGNodes) {
	if ((dwFlags & rgn_t::fDeleted) != 0)
		return;

	dwFlags |= rgn_t::fDeleted;
	++ _map.layers [Kind].cDeletedRgns;

	if (_bReleaseRGNodes)
		ReleaseRGNodes ();
}

void rgn_t::Undelete (map_t & _map, bool _bAddRefRGNodes) {
	if ((dwFlags & rgn_t::fDeleted) == 0)
		return;

	dwFlags &= ~rgn_t::fDeleted;
	-- _map.layers [Kind].cDeletedRgns;

	if (_bAddRefRGNodes)
		AddRefRGNodes ();
}

bool rgn_t::IsPolygon () const {
	const type_desc_t * const pTD = GetTypeDesc ();
	assert (pTD);
	if (! pTD)
		return false;

	return pTD->bUnknown ? (Kind == rkPolygon) : (pTD->Kind == rkPolygon);
}

bool rgn_t::IsPolyline () const {
	const type_desc_t * const pTD = GetTypeDesc ();
	assert (pTD);
	if (! pTD)
		return false;

	return pTD->bUnknown ? (Kind == rkPolyline) : (pTD->Kind == rkPolyline);
}

bool rgn_t::IsPoint () const {
	const type_desc_t * const pTD = GetTypeDesc ();
	assert (pTD);
	if (! pTD)
		return false;

	return pTD->bUnknown ? (Kind == rkPoint) : (pTD->Kind == rkPoint);
}

bool rgn_t::IsRoad () const {
	return IsMapObject () && IsPolyline () && g_map.pTypeSet->pfIsRoad (* this);
}

bool rgn_t::IsAttachment () const {
	return IsWTR () || (pFileRef && pFileRef->bAttachment);
}

bool rgn_t::IsVisible () const {
	return ! IsDeleted () && (! pFileRef || (pFileRef->bVisible && ! pFileRef->bDeleted));
}

void rgn_t::AddCommentLine (const char * _strLine) {
	AddCommentLine (NULL, _strLine);
}

void rgn_t::AddCommentLine (const char * _strKey, const char * _strValue) {
	const size_t cLen = strComment.length ();
	
	bool bNeedCRLF = false;
	if (cLen == 1)
		bNeedCRLF = true;
	else if (cLen > 1)
		bNeedCRLF = (strComment [cLen - 2] != '\r') || (strComment [cLen - 1] == '\n');

	if (bNeedCRLF)
		strComment.append ("\r\n");

	if (_strKey && _strKey [0]) {
		strComment.append (_strKey);
		strComment.append ("=");
	}
	strComment.append (_strValue);
}

void rgn_t::AddCommentLine (const char * _strKey, DWORD _dwValue) {
	char strValue [16];
	::sprintf (strValue, "%u", _dwValue);
	AddCommentLine (_strKey, strValue);
}

void rgn_t::AddCommentLine (const char * _strKey, long _lValue) {
	char strValue [16];
	::sprintf (strValue, "%d", _lValue);
	AddCommentLine (_strKey, strValue);
}

void rgn_element_t::ReleaseRGNodes () {
	if (! pExtra)
		return;

	const size_t cPoints = points.size ();

	point_extras_t & PointExtras = pExtra->PointExtras;
	const size_t cPointExtras = PointExtras.size ();
	for (size_t cPoint = 0; cPoint < cPointExtras; ++ cPoint) {
		rgnode_t * const pRGNode = PointExtras [cPoint].GetRGNode ();
		if (pRGNode) {
			-- pRGNode->btPolylines;

			if (cPoint == 0 || cPoint + 1 == cPoints)
				-- pRGNode->btCardinality;
			else
				pRGNode->btCardinality -= 2;
		}
	}
}

void rgn_element_t::AddRefRGNodes () {
	if (! pExtra)
		return;

	const size_t cPoints = points.size ();

	point_extras_t & PointExtras = pExtra->PointExtras;
	const size_t cPointExtras = PointExtras.size ();
	for (size_t cPoint = 0; cPoint < cPointExtras; ++ cPoint) {
		rgnode_t * const pRGNode = PointExtras [cPoint].GetRGNode ();
		if (pRGNode) {
			++ pRGNode->btPolylines;

			if (cPoint == 0 || cPoint + 1 == cPoints)
				++ pRGNode->btCardinality;
			else
				pRGNode->btCardinality += 2;
		}
	}
}

void rgn_t::ReleaseRGNodes () {
	if (! IsPolyline () || IsAttachment () || IsNotMapObject ())
		return;

	const size_t cElements = elements.size ();
	for (size_t cElement = 0; cElement < cElements; ++ cElement) {
		rgn_element_t & element = elements [cElement];
		if (element.bDeleted)
			continue;

		element.ReleaseRGNodes ();
	}
}

void rgn_t::AddRefRGNodes () {
	if (! IsPolyline () || IsImmutable () || IsNotMapObject ())
		return;

	const size_t cElements = elements.size ();
	for (size_t cElement = 0; cElement < cElements; ++ cElement) {
		rgn_element_t & element = elements [cElement];
		if (element.bDeleted)
			continue;

		element.AddRefRGNodes ();
	}
}

bool rgn_t::HasRGNodes () const {
	if (! IsPolyline () || IsImmutable ())
		return false;

	const size_t cElements = elements.size ();
	for (size_t cElement = 0; cElement < cElements; ++ cElement) {
		const rgn_element_t & element = elements [cElement];
		if (element.bDeleted || ! element.pExtra)
			continue;

		if (::HasRGNodes (element.pExtra->PointExtras))
			return true;
	}

	return false;
}

float GetRgnSize (const rgn_t & _rgn) {
	const size_t cElements = _rgn.elements.size ();
	if (! cElements)
		return 0;

	float fCosY;
	point_t p0, p1;
	bool b0Defined = false, b1Defined = false;

	float d2max = 0;

	// Find the most distant point for the first one.
	size_t cElement;
	for (cElement = 0; cElement < cElements; ++ cElement) {
		const rgn_element_t & element = _rgn.elements [cElement];
		if (element.bDeleted)
			continue;

		const points_t & points = element.points;
		const size_t cPoints = points.size ();
		if (! cPoints)
			continue;

		for (size_t cPoint = 0; cPoint < cPoints; ++ cPoint) {
			const point_t & point = points [cPoint];

			if (! b0Defined) {
				p0 = point;
				b0Defined = true;
				fCosY = ::cos (point.y*c_PI/180.f);
				continue;
			}

			const float d2 = (p0.x - point.x)*(p0.x - point.x)*fCosY*fCosY + (p0.y - point.y)*(p0.y - point.y);
			if (d2 > d2max) {
				d2max = d2;
				p1 = point;
				b1Defined = true;
			}
		}
	}

	if (! b1Defined)
		return 0;

	// Find the most distant point for the first most distant one.
	for (cElement = 0; cElement < cElements; ++ cElement) {
		const rgn_element_t & element = _rgn.elements [cElement];
		if (element.bDeleted)
			continue;

		const points_t & points = element.points;
		const size_t cPoints = points.size ();
		if (! cPoints)
			continue;

		for (size_t cPoint = 0; cPoint < cPoints; ++ cPoint) {
			const point_t & point = points [cPoint];

			const float d2 = (p1.x - point.x)*(p1.x - point.x)*fCosY*fCosY + (p1.y - point.y)*(p1.y - point.y);
			if (d2 > d2max)
				d2max = d2;
		}
	}

	return d2max;
}

///////////////////////////////////////////////

bool rgn_t::GetDirection (DWORD _dwStartNodeID, DWORD _dwEndNodeID, bool & _bForward) const {
	if (! HasRGNodes ())
		return false;

	const size_t cElements = elements.size ();
	for (size_t cElement = 0; cElement < cElements; ++ cElement) {
		const rgn_element_t & element = elements [cElement];
		if (element.bDeleted)
			continue;

		const point_extras_t & PointExtras = element.pExtra->PointExtras;
		if (PointExtras.empty ())
			continue;

		size_t cStartNode = c_cUnknownPoint;
		size_t cEndNode   = c_cUnknownPoint;
		const size_t cPoints = PointExtras.size ();
		for (size_t cPoint = 0; cPoint < cPoints; ++ cPoint) {
			const point_extra_t & point = PointExtras [cPoint];
			const rgnode_t * const pNode = point.GetRGNode ();
			if (pNode == NULL)
				continue;

			if (pNode->dwNodeID == _dwStartNodeID)
				cStartNode = cPoint;
			if (pNode->dwNodeID == _dwEndNodeID)
				cEndNode = cPoint;
		}

		if (cStartNode != c_cUnknownPoint && cEndNode != c_cUnknownPoint) {
			_bForward = cStartNode < cEndNode;
			return true;
		}
	}

	return false;
}

rgnode_t * rgn_element_t::GetNextNode (const rgnode_t * _pNode, bool _bForward) const {
	if (pExtra == NULL)
		return NULL;

	const size_t cPoints = pExtra->PointExtras.size ();
	rgnode_t * pPrevNode = NULL;
	for (size_t cPoint = 0; cPoint < cPoints; ++ cPoint) {
		const point_extra_t & point = pExtra->PointExtras [cPoint];
		rgnode_t * const pNode = const_cast<rgnode_t *> (point.GetRGNode ());
		if (pNode == NULL)
			continue;

		if (_bForward) {
			if (pPrevNode != NULL || _pNode == NULL)
				return pNode;
			if (pNode == _pNode)
				pPrevNode = pNode;
		} else {
			if (pNode == _pNode) {
//				assert (pPrevNode != NULL);
				return pPrevNode;
			}
			pPrevNode = pNode;
		}
	}

	if (pPrevNode != NULL && _pNode == NULL && ! _bForward)
		return pPrevNode;

	return NULL;
}

rgnode_t * rgn_t::GetNextNode (const rgnode_t * _pNode, bool _bForward) const {
	if (! HasRGNodes ())
		return NULL;

	const size_t cElements = elements.size ();
	for (size_t cElement = 0; cElement < cElements; ++ cElement) {
		const rgn_element_t & element = elements [cElement];
		if (element.bDeleted)
			continue;

		rgnode_t * const pNextNode = element.GetNextNode (_pNode, _bForward);
		if (pNextNode != NULL)
			return pNextNode;
	}
	return NULL;
}

size_t rgn_element_t::FindNode (const rgnode_t * _pNode) const {
	if (pExtra == NULL)
		return c_cUnknownPoint;

	const size_t cPoints = pExtra->PointExtras.size ();
	for (size_t cPoint = 0; cPoint < cPoints; ++ cPoint) {
		const point_extra_t & point = pExtra->PointExtras [cPoint];
		const rgnode_t * const pNode = point.GetRGNode ();
		if (pNode == NULL)
			continue;

		if (pNode == _pNode)
			return cPoint;
	}
	return c_cUnknownPoint;
}

bool rgn_element_t::HasNode (const rgnode_t * _pNode) const {
	return FindNode (_pNode) != c_cUnknownPoint;
}

bool rgn_t::HasNode (const rgnode_t * _pNode) const {
	if (! HasRGNodes ())
		return false;

	const size_t cElements = elements.size ();
	for (size_t cElement = 0; cElement < cElements; ++ cElement) {
		const rgn_element_t & element = elements [cElement];
		if (element.bDeleted)
			continue;

		if (element.HasNode (_pNode))
			return true;
	}
	return false;
}

bool rgn_t::HasValidRestriction (const rgrestriction_t & _r) const {
	const bool bUTurn = _r.pFromRoad == _r.pToRoad && _r.bFromDirection != _r.bToDirection;
	if (bUTurn) {
		if (! HasRGNodes ())
			return false;

		const size_t cElements = elements.size ();
		for (size_t cElement = 0; cElement < cElements; ++ cElement) {
			const rgn_element_t & element = elements [cElement];
			if (element.bDeleted)
				continue;

			const size_t cPoint = element.FindNode (_r.pNode);
			if (cPoint == c_cUnknownPoint)
				continue;

			if (cPoint == 0)
				return _r.bToDirection != 0;
			if (cPoint == element.points.size () - 1)
				return _r.bFromDirection != 0;

			return true;
		}
		return false;
	} else
		return HasNode (_r.pNode);
}

////////////////////////////////////////////////

void MakeRightOriented (rgn_t & _rgn) {
	for (size_t cElem = 0; cElem < _rgn.elements.size (); ++ cElem) {
		rgn_element_t & elem = _rgn.elements [cElem];
		if (elem.bDeleted)
			continue;

		if (! IsRightOriented (elem.points) == ! elem.bHole)
			std::reverse (elem.points.begin (), elem.points.end ());
	}
}

////////////////////////////////////////////////

# include "Bookmark.h"

rgn_extra_t::rgn_extra_t (const rgn_extra_t & _other) :
	pBookmark (NULL),
	pNoteDrawing (NULL)
{
	* this = _other;
}

rgn_extra_t & rgn_extra_t::operator = (const rgn_extra_t & _other) {
	if (this == & _other)
		return * this;

	this->pCity          = _other.pCity;
	this->strNumber      = _other.strNumber;
	this->strStreetDesc  = _other.strStreetDesc;
	this->pZip           = _other.pZip;

	this->strPhone       = _other.strPhone;
	this->strFax         = _other.strFax;
	this->strEmail       = _other.strEmail;
	this->strWebPage     = _other.strWebPage;
	this->strDescription = _other.strDescription;

	this->pHighway       = _other.pHighway;

	this->dwRoadID       = _other.dwRoadID;
	this->dwRouteParams  = _other.dwRouteParams;
	this->fCost          = _other.fCost;
	this->dwManeuverType = _other.dwManeuverType;
	this->bNoMerge       = _other.bNoMerge;

	this->entry_points   = _other.entry_points;
	this->lanes          = _other.lanes;
	this->strExtra       = _other.strExtra;
	this->wtr            = _other.wtr;

	if (this->pBookmark) {
		delete this->pBookmark;
		this->pBookmark = NULL;
	}
	if (_other.pBookmark) {
		assert (this->pBookmark == NULL);
		this->pBookmark = new bookmark_t;
		* this->pBookmark = * _other.pBookmark;
	}

	if (this->pNoteDrawing) {
		delete this->pNoteDrawing;
		this->pNoteDrawing = NULL;
	}
	if (_other.pNoteDrawing) {
		assert (this->pNoteDrawing == NULL);
		this->pNoteDrawing = new note_drawing_t;
		* this->pNoteDrawing = * _other.pNoteDrawing;
	}

	return * this;
}

rgn_extra_t::~rgn_extra_t () {
	delete this->pBookmark;
	delete this->pNoteDrawing;
}

static const float c_fSpeedLimitInKMH [8] = {5.f, 20.f, 40.f, 60.f, 80.f, 90.f, 110.f, 140.f};
static const float c_fSpeedLimitInMPH [8] = {3.f, 15.f, 25.f, 35.f, 50.f, 60.f,  70.f, 100.f};
static const float c_fRoadClassCoeff [8] = {2.0f, 1.4f, 1.2f, 1.0f, 0.8f, 0.7f, 0.6f, 0.5f};

float rgn_extra_t::GetSpeedLimitInKMH () const {
	return c_fSpeedLimitInKMH [fSpeedLimit];
}

float rgn_extra_t::GetSpeedLimitInMPH () const {
	return c_fSpeedLimitInMPH [fSpeedLimit];
}

float rgn_extra_t::GetRoadClassCoeff () const {
	return c_fRoadClassCoeff [fRouteClass];
}

# include "Arcs.h"

double rgn_extra_t::CalculateTime (double _fLength, bool _bInArcs) const {
	static const int c_nArcBits = 31;

	double fTime = _bInArcs ? Arcs2Meters (_fLength) : _fLength;

	fTime *= (3.6f/c_fSpeedLimitInKMH [fSpeedLimit]) * c_fRoadClassCoeff [fRouteClass];

	return fTime <= 1.f ? 1.f : fTime;
}

////////////////////////////////////////////////

lanes_t GetReverseLanes (const lanes_t & _lanes) {
	lanes_t rev;
	rev.bDefined        = _lanes.bDefined;
	rev.fForward        = _lanes.fReverse;
	rev.fForwardMixed   = _lanes.fReverseMixed;
	rev.nForwardExpLeft = _lanes.nReverseExpLeft;
	rev.fReverse        = _lanes.fForward;
	rev.fReverseMixed   = _lanes.fForwardMixed;
	rev.nReverseExpLeft = _lanes.nForwardExpLeft;

	return rev;
}

//////////////////////////////////////////////////

point_t rgn_extra_t::GetEntryPoint0 () const {
	const size_t c0 = GetEntryPointOrdinal0 ();
	if (c0 != -1)
		return entry_points [c0].pt;

	assert (false);
	return point_t (0, 0);
}

size_t rgn_extra_t::GetEntryPointOrdinal0 () const {
	const size_t cEntryPoints = entry_points.size ();
	for (size_t cEP = 0; cEP < cEntryPoints; ++ cEP) {
		const entry_point_t & ep = entry_points [cEP];
		if (! ep.bDeleted)
			return cEP;
	}

	return -1;
}

size_t rgn_extra_t::GetEntryPoints () const {
	size_t cRet = 0;
	const size_t cEntryPoints = entry_points.size ();
	for (size_t cEP = 0; cEP < cEntryPoints; ++ cEP) {
		const entry_point_t & ep = entry_points [cEP];
		if (! ep.bDeleted)
			++ cRet;
	}

	return cRet;
}

bool rgn_extra_t::AreEntryPointsDefined () const {
	if (entry_points.empty ())
		return false;
	const size_t cEntryPoints = entry_points.size ();
	for (size_t cEP = 0; cEP < cEntryPoints; ++ cEP) {
		const entry_point_t & ep = entry_points [cEP];
		if (! ep.bDeleted)
			return true;
	}
	return false;
}

void rgn_extra_t::AddEntryPoint (const point_t & _pt, const string_t & _strLabel) {
	entry_points.push_back (entry_point_t ());
	entry_point_t & ep = entry_points.back ();
	ep.pt       = _pt;
	ep.strLabel = _strLabel;
}

bool rgn_extra_t::ContainsEntryPoint (const point_t & _point, float _fE) const {
	const size_t cEntryPoints = entry_points.size ();
	const float fCosY = cEntryPoints != 0 ? ::cos (_point.y*c_PI/180.f) : 1;
	for (size_t cEP = 0; cEP < cEntryPoints; ++ cEP) {
		const entry_point_t & ep = entry_points [cEP];
		if (ep.bDeleted)
			continue;

		if (PointContainsPt (ep.pt, _point, _fE, fCosY))
			return true;
	}

	return false;
}

/////////////////////////////////////////////////////////////////

# include "MapView.h"

bool IsRgnKindVisible (const rgn_t & _rgn, const map_t & _map) {
	if (! _rgn.IsVisible ())
		return false;

	if (_rgn.IsBookmarkOrNote ()) {
		if (! g_bShowBookmarksAndNotes)
			return false;
	}

	if (_rgn.IsAttachment ()) {
		if (! g_bShowAttachments)
			return false;
		if (_rgn.pFileRef && ! _rgn.pFileRef->bVisible)
			return false;
	} else {
		if (! g_bShowMap)
			return false;
	}

	if (_map.pTypeSet->pfIsBackground (_rgn) && ! g_bShowCoverageArea)
		return false;

	return true;
}

# include "Levels.h"

bool IsRgnVisibleAtCurrentZoomLevel (const rgn_t & _rgn) {
	if (_rgn.IsNotMapObject ())
		return true;

	const size_t cElements = _rgn.elements.size ();
	for (size_t cElement = 0; cElement < cElements; ++ cElement) {
		const rgn_element_t & element = _rgn.elements [cElement];
		if (element.bDeleted)
			continue;

		if (LevelIsVisible (element.cLevel, _rgn.uchEndLevel))
			return true;
	}

	const size_t cOrigins = _rgn.origins.size ();
	for (size_t cOrigin = 0; cOrigin < cOrigins; ++ cOrigin) {
		const rgn_origin_t & origin = _rgn.origins [cOrigin];
		if (origin.bDeleted)
			continue;

		if (LevelIsVisible (origin.cLevel, _rgn.uchEndLevel))
			return true;
	}

	return false;
}