//
// GPSMapEdit
// (c) Konstantin Galichsky (kg@geopainting.com), 2005-2009
//
// Object properties dialog - 'Track points' page.
//

# include "StdAfx.h"
# include "Resource.h"
# include "RgnProperties.h"
# include "Map.h"
# include "ListView.h"
# include "PrintUnits.h"
# include "fast_floor.h"
# include "Settings.h"
# include "Globals.h"

enum {
	lvOrdinal     = 0,
	lvTime        = 1,
	lvDepth       = 2,
	lvTemperature = 3,
	lvAltitude    = 4,
	lvLegLength   = 5,
	lvLegTime     = 6,
	lvSpeed       = 7,
	lvCourse      = 8,
	lvPosition    = 9,
	lvDOP         = 10,
	lvHDOP        = 11,
	lvVDOP        = 12,
	lvSats        = 13,
	lvFix         = 14,

	lvMax
};

void CTrackPointsPage::_GetItemText (size_t _cPoint, size_t _cColumn, char * _strText, size_t _cTextMax) const {
	if (m_rgn.elements.empty ())
		return;
	const points_t & points = m_rgn.elements [0].points;
	const size_t cPoints = points.size ();
	const trackpoints_t & TrackPoints = m_rgn.elements [0].pExtra->TrackPoints;
	const size_t cExtraPoints = TrackPoints.size ();

	if (_cPoint >= cPoints) {
		::strcpy (_strText, "-");
		return;
	}
	const point_t & point = points [_cPoint];

	switch (_cColumn) {
		case lvOrdinal:
			::sprintf (_strText, "%d", _cPoint + 1);
			break; 

		case lvTime:
			if (_cPoint < cExtraPoints) {
				const trackpoint_info_t & ExtraPoint = TrackPoints [_cPoint];
				if (ExtraPoint.wFlags & trackpoint_info_t::defTime) {
					// Convert to local time.
					DATE dtLocalTime = ExtraPoint.dtTime - m_dtBias;

					// Check milliseconds (NOTE: ::VariantTimeToSystemTime() does not support this).
					int nMilliseconds = 0;
					if (ExtraPoint.wFlags & trackpoint_info_t::defTimeMilliseconds) {
						const float fSeconds = (dtLocalTime - fast_floor_d (dtLocalTime))*24*60*60;
						nMilliseconds = fast_floor ((fSeconds - fast_floor (fSeconds))*1000 + .5f);
						if (nMilliseconds <= 995)
							dtLocalTime -= nMilliseconds/24./60./60./1000;
						else
							nMilliseconds = 0;
					}
					SYSTEMTIME st;
					::VariantTimeToSystemTime (dtLocalTime, & st);

					char strDate [32];
					if (! ::GetDateFormat (LOCALE_USER_DEFAULT, DATE_SHORTDATE, & st, NULL, strDate, sizeof (strDate)/sizeof (strDate [0])))
						strDate [0] = '\0';
					char strTime [32];
					if (! ::GetTimeFormat (LOCALE_USER_DEFAULT, 0, & st, NULL, strTime, sizeof (strTime)/sizeof (strTime [0])))
						strTime [0] = '\0';

					if (ExtraPoint.wFlags & trackpoint_info_t::defTimeMilliseconds)
						::sprintf (_strText, "%s %s.%02d", strDate, strTime, (nMilliseconds + 5)/10);
					else
						::sprintf (_strText, "%s %s", strDate, strTime);
				}
			}
			break;

		case lvDepth:
			if (_cPoint < cExtraPoints) {
				const trackpoint_info_t & ExtraPoint = TrackPoints [_cPoint];
				if (ExtraPoint.wFlags & trackpoint_info_t::defDepth) {
					char strDepth [32];
					char strDepthUnits [16];
					PrintDepth (strDepth, strDepthUnits, ExtraPoint.fDepth);

					::sprintf (_strText, "%s %s", strDepth, strDepthUnits);
				}
			}
			break;

		case lvTemperature:
			if (_cPoint < cExtraPoints) {
				const trackpoint_info_t & ExtraPoint = TrackPoints [_cPoint];
				if (ExtraPoint.wFlags & trackpoint_info_t::defTemperature) {
					char strTemperature [32];
					char strTemperatureUnit [16];
					PrintTemperature (strTemperature, strTemperatureUnit, ExtraPoint.fTemperature);

					::sprintf (_strText, "%s%s", strTemperature, strTemperatureUnit);
				}
			}
			break;

		case lvAltitude:
			if (_cPoint < cExtraPoints) {
				const trackpoint_info_t & ExtraPoint = TrackPoints [_cPoint];
				if (ExtraPoint.wFlags & trackpoint_info_t::defAltitude) {
					char strAltitude [32];
					char strAltitudeUnit [16];
					PrintElevation (strAltitude, strAltitudeUnit, ExtraPoint.fAltitude, unMeters, false);

					::sprintf (_strText, "%s %s", strAltitude, strAltitudeUnit);
				}
			}
			break;

		case lvLegLength:
			if (_cPoint + 1 < cPoints) {
				const point_t & point1 = points [_cPoint + 1];
				const double dLegLength = ::hypot (point1.y - point.y, (point1.x - point.x)*m_fCosY)*40e6/360;

				char strLegLength [32];
				char strLegLengthUnits [16];
				PrintDistance (strLegLength, strLegLengthUnits, dLegLength);

				::sprintf (_strText, "%s %s", strLegLength, strLegLengthUnits);
			}
			break;

		case lvLegTime:
			if (_cPoint + 1 < cExtraPoints) {
				const trackpoint_info_t & ExtraPoint  = TrackPoints [_cPoint];
				const trackpoint_info_t & ExtraPoint1 = TrackPoints [_cPoint + 1];
				if (
					(ExtraPoint.wFlags  & trackpoint_info_t::defTime) &&
					(ExtraPoint1.wFlags & trackpoint_info_t::defTime)
				) {
					const double dLegTime = (ExtraPoint1.dtTime - ExtraPoint.dtTime)*24*60*60;
					const float fLegTime = ::fabs (dLegTime);

					if (ExtraPoint.wFlags & trackpoint_info_t::defTimeMilliseconds) {
						const int nSecond = fast_floor (fLegTime);
						const int nMinute = nSecond/60;
						const int nHour   = nMinute/60;

						::sprintf (
							_strText,
							"%d:%02d:%02.2f",
							nHour,
							nMinute  - nHour*60,
							fLegTime - nMinute*60
						);
					} else {
						const int nSecond = fast_floor (fLegTime + .5f);
						const int nMinute = nSecond/60;
						const int nHour   = nMinute/60;

						::sprintf (
							_strText,
							"%d:%02d:%02d",
							nHour,
							nMinute - nHour*60,
							nSecond - nMinute*60
						);
					}
				}
			}
			break;

		case lvSpeed:
			if (_cPoint + 1 < cPoints && _cPoint < cExtraPoints) {
				const point_t & point1 = points [_cPoint + 1];
				const double dLegLength = ::hypot (point1.y - point.y, (point1.x - point.x)*m_fCosY)*40e6/360;

				const trackpoint_info_t & ExtraPoint = TrackPoints [_cPoint];

				float fSpeed = -1;
				if (ExtraPoint.wFlags & trackpoint_info_t::defSpeed)
					fSpeed = ExtraPoint.fSpeed;
				else {
					double dLegTime = 0;
					if (_cPoint + 1 < cExtraPoints) {
						const trackpoint_info_t & ExtraPoint1 = TrackPoints [_cPoint + 1];
						if (
							(ExtraPoint.wFlags  & trackpoint_info_t::defTime) &&
							(ExtraPoint1.wFlags & trackpoint_info_t::defTime)
						)
							dLegTime = (ExtraPoint1.dtTime - ExtraPoint.dtTime)*24*60*60;
					}
					if (dLegTime > 0)
						fSpeed = 3.6*dLegLength/::fabs (dLegTime);
				}
				if (fSpeed >= 0) {
					char strSpeed [32];
					char strSpeedUnits [16];
					PrintSpeed (strSpeed, strSpeedUnits, fSpeed);

					::sprintf (_strText, "%s %s", strSpeed, strSpeedUnits);
				}
			}
			break;

		case lvCourse:
			if (_cPoint + 1 < cPoints) {
				const point_t & point1 = points [_cPoint + 1];
				if (point1 == point)
					break;

				double dCourse = 90 - ::atan2 (point1.y - point.y, (point1.x - point.x)*m_fCosY)*180/c_PI;
				if (dCourse < 0)
					dCourse += 360;

				PrintAzimuth (_strText, dCourse);
			}
			break;

		case lvPosition:
			PrintPosition (_strText, point);
			break;

		case lvDOP:
			if (_cPoint < cExtraPoints) {
				const trackpoint_info_t & ExtraPoint = TrackPoints [_cPoint];
				if (ExtraPoint.wFlags & trackpoint_info_t::defDOP)
					::sprintf (_strText, "%.1f", ExtraPoint.fDOP);
			}
			break;

		case lvHDOP:
			if (_cPoint < cExtraPoints) {
				const trackpoint_info_t & ExtraPoint = TrackPoints [_cPoint];
				if (ExtraPoint.wFlags & trackpoint_info_t::defHDOP)
					::sprintf (_strText, "%.1f", ExtraPoint.fHDOP);
			}
			break;

		case lvVDOP:
			if (_cPoint < cExtraPoints) {
				const trackpoint_info_t & ExtraPoint = TrackPoints [_cPoint];
				if (ExtraPoint.wFlags & trackpoint_info_t::defVDOP)
					::sprintf (_strText, "%.1f", ExtraPoint.fVDOP);
			}
			break;

		case lvSats:
			if (_cPoint < cExtraPoints) {
				const trackpoint_info_t & ExtraPoint = TrackPoints [_cPoint];
				if (ExtraPoint.wFlags & trackpoint_info_t::defSats)
					::sprintf (_strText, "%u", ExtraPoint.btSats);
			}
			break;

		case lvFix:
			if (_cPoint < cExtraPoints) {
				const trackpoint_info_t & ExtraPoint = TrackPoints [_cPoint];
				if (ExtraPoint.wFlags & trackpoint_info_t::defFix) {
					const char * strFix = NULL;
					switch (ExtraPoint.btFix) {
						default:
						case trackpoint_info_t::fixNone: strFix = "Invalid"; break;
						case trackpoint_info_t::fix2D:   strFix = "2D";      break;
						case trackpoint_info_t::fix3D:   strFix = "3D";      break;
						case trackpoint_info_t::fixDGPS: strFix = "DGPS";    break;
						case trackpoint_info_t::fixPPS:  strFix = "PPS";     break;
					}
					::strcpy (_strText, strFix);
				}
			}
			break;

		default:
			assert (false);
			break;
	}
}

void CTrackPointsPage::_FillPoints () {
	bool bEmptyColumn [lvMax];
	int nColumn;
	for (nColumn = 0; nColumn < lvMax; ++ nColumn)
		bEmptyColumn [nColumn] = true;

	if (m_rgn.elements.empty ())
		return;
	const points_t & points = m_rgn.elements [0].points;
	const size_t cPoints = points.size ();
	const trackpoints_t & TrackPoints = m_rgn.elements [0].pExtra->TrackPoints;
	const size_t cExtraPoints = TrackPoints.size ();

	for (size_t cPoint = 0; cPoint < cPoints; ++ cPoint) {
		const point_t & point = points [cPoint];

		bEmptyColumn [lvOrdinal] = false;
		bEmptyColumn [lvPosition] = false;

		double dLegLength = 0;
		if (cPoint + 1 < cPoints) {
			const point_t & point1 = points [cPoint + 1];

			bEmptyColumn [lvLegLength] = false;
			if (! (point1 == point))
				bEmptyColumn [lvCourse] = false;
		}

		if (cPoint < cExtraPoints) {
			const trackpoint_info_t & ExtraPoint = TrackPoints [cPoint];

			// Time.
			if (ExtraPoint.wFlags & trackpoint_info_t::defTime)
				bEmptyColumn [lvTime] = false;
			// Depth.
			if (ExtraPoint.wFlags & trackpoint_info_t::defDepth)
				bEmptyColumn [lvDepth] = false;
			// Temperature.
			if (ExtraPoint.wFlags & trackpoint_info_t::defTemperature)
				bEmptyColumn [lvTemperature] = false;
			// Altitude.
			if (ExtraPoint.wFlags & trackpoint_info_t::defAltitude)
				bEmptyColumn [lvAltitude] = false;

			double dLegTime = 0;
			if (cPoint + 1 < cExtraPoints) {
				const trackpoint_info_t & ExtraPoint1 = TrackPoints [cPoint + 1];

				// Leg time.
				if (
					(ExtraPoint.wFlags  & trackpoint_info_t::defTime) &&
					(ExtraPoint1.wFlags & trackpoint_info_t::defTime)
				) {
					dLegTime = ExtraPoint1.dtTime - ExtraPoint.dtTime;
					bEmptyColumn [lvLegTime] = false;
				}
			}

			// Speed.
			if (ExtraPoint.wFlags & trackpoint_info_t::defSpeed)
				bEmptyColumn [lvSpeed] = false;
			else {
				if (dLegTime > 0)
					bEmptyColumn [lvSpeed] = false;
			}

			// DOP.
			if (ExtraPoint.wFlags & trackpoint_info_t::defDOP)
				bEmptyColumn [lvDOP] = false;
			// HDOP.
			if (ExtraPoint.wFlags & trackpoint_info_t::defHDOP)
				bEmptyColumn [lvHDOP] = false;
			// VDOP.
			if (ExtraPoint.wFlags & trackpoint_info_t::defVDOP)
				bEmptyColumn [lvVDOP] = false;
			// Sats.
			if (ExtraPoint.wFlags & trackpoint_info_t::defSats)
				bEmptyColumn [lvSats] = false;
			// Fix.
			if (ExtraPoint.wFlags & trackpoint_info_t::defFix)
				bEmptyColumn [lvFix] = false;
		}
	}

	ListView_SetItemCount (m_hList, cPoints);

	if (m_cPoint != -1) {
		SetSelection (m_hList, m_cPoint);
		EnsureSelectionVisible (m_hList);
	}

	// Set the best width.
	AutoSizeColumns (m_hList);

	// Hide empty columns.
	for (nColumn = 0; nColumn < lvMax; ++ nColumn) {
		if (bEmptyColumn [nColumn])
			ListView_SetColumnWidth (m_hList, nColumn, 0);
	}
}

bool CTrackPointsPage::OnInit (HWND _hwndFocus) {
	// Get time zone bias + daylight saving time bias.
	TIME_ZONE_INFORMATION tzi;
	if (::GetTimeZoneInformation (& tzi) == TIME_ZONE_ID_DAYLIGHT)
		tzi.Bias += tzi.DaylightBias;
	m_dtBias = float (tzi.Bias)/(24*60);

	const size_t cPoints = m_rgn.elements.empty () ? 0 : m_rgn.elements [0].points.size ();
	m_fCosY = cPoints > 0 ? ::cos (c_PI*m_rgn.elements [0].points [0].y/180) : 1;

	// Points.
	m_hList.Init (m_hWnd, IDC_POINTS, "TrackPoints");
	m_hList.SetExtStyle (LVS_EX_FULLROWSELECT | LVS_EX_HEADERDRAGDROP);

	AddColumn (m_hList, lvOrdinal,     "#",           true);
	AddColumn (m_hList, lvTime,        "Time");
	AddColumn (m_hList, lvDepth,       "Depth",       true);
	AddColumn (m_hList, lvTemperature, "Temperature", true);
	AddColumn (m_hList, lvAltitude,    "Altitude",    true);
	AddColumn (m_hList, lvLegLength,   "Leg",         true);
	AddColumn (m_hList, lvLegTime,     "Leg Time",    true);
	AddColumn (m_hList, lvSpeed,       "Speed",       true);
	AddColumn (m_hList, lvCourse,      "Course",      true);
	AddColumn (m_hList, lvPosition,    "Position");
	AddColumn (m_hList, lvDOP,         "DOP",         true);
	AddColumn (m_hList, lvHDOP,        "HDOP",        true);
	AddColumn (m_hList, lvVDOP,        "VDOP",        true);
	AddColumn (m_hList, lvSats,        "Satellites",  true);
	AddColumn (m_hList, lvFix,         "Fix",         true);

	_FillPoints ();

//	AutoSizeColumns (m_hList);
//	m_hList.LoadHeaders ();

	return true;
}

void CTrackPointsPage::OnDestroy () {
//	m_hList.SaveHeaders ();
	m_hList.Destroy ();
}

bool CTrackPointsPage::OnNotify (int _idCtrl, NMHDR * _pInfo) {
	switch (_pInfo->code) {
		case NM_RCLICK:
			m_hList.PopupMenu_OnHeaderRClick (_pInfo);
			break;

		case LVN_GETDISPINFO: {
			NMLVDISPINFO * const pInfo2 = reinterpret_cast<NMLVDISPINFO *> (_pInfo);
			LVITEM & lvi = pInfo2->item;
			if (_idCtrl == IDC_POINTS) {
				if (lvi.mask & LVIF_TEXT)
					_GetItemText (lvi.iItem, lvi.iSubItem, lvi.pszText, lvi.cchTextMax);
			}

			return true;
		}
	}

	return false;
}