//
// GPSMapEdit
// (c) Konstantin Galichsky (kg@geopainting.com), 2008-2009
//
// Object properties dialog - 'Traffic Lanes' page.
//

# include "StdAfx.h"
# include "Resource.h"
# include "RgnProperties.h"
# include "Map.h"
# include "MapView.h"
# include "Types.h"
# include "SetRgnExtra.h"
# include "DrawHelpers.h"
# include "DrawDirIndicator.h"
# include "fast_floor.h"
# include "Globals.h"

//# include <float.h>

enum cr_link_type_t {ltIncoming, ltOutgoing};

struct cr_link_t {
	const bool bLanesDefined;
	const bool bReverse;
	const cr_link_type_t linkType;
	const int nLanes, nLanesMixed, nLanesExpLeft;

	const rgn_t * const pRgn;
	const rgn_element_t * const pElement;
	const size_t cPoint;

	const float fAngle;

	cr_link_t * pLinkPrev;
	cr_link_t * pLinkNext;
	float fDirection;

	int nLanesExpected, nLanesExpectedMixed;
	int nIncomingLinks, nOutgoingLinks;

	cr_link_t (
		const rgn_t * _pRgn, const rgn_element_t * _pElement, size_t _cPoint,
		cr_link_type_t _lt, const point_t & _ptCenter, const point_t & _ptEnd,
		bool _bReverse,
		float _fCosY
	) :
		bLanesDefined (_pRgn->pExtra->lanes.bDefined != 0),
		nLanes        (_bReverse ? _pRgn->pExtra->lanes.fReverse        : _pRgn->pExtra->lanes.fForward),
		nLanesMixed   (_bReverse ? _pRgn->pExtra->lanes.fReverseMixed   : _pRgn->pExtra->lanes.fForwardMixed),
		nLanesExpLeft (_bReverse ? _pRgn->pExtra->lanes.nReverseExpLeft : _pRgn->pExtra->lanes.nForwardExpLeft),
		pRgn (_pRgn),
		pElement (_pElement),
		cPoint (_cPoint),
		linkType (_lt),
		bReverse (_bReverse),
		fAngle (::atan2 (_ptEnd.y - _ptCenter.y, (_ptEnd.x - _ptCenter.x)*_fCosY)),
		pLinkPrev (NULL),
		pLinkNext (NULL),
		nIncomingLinks (0),
		nOutgoingLinks (0),
		nLanesExpected (1),
		nLanesExpectedMixed (0)
	{
		fDirection = _lt == ltOutgoing ? fAngle : ::atan2 (_ptCenter.y - _ptEnd.y, (_ptCenter.x - _ptEnd.x)*_fCosY);
	}

	cr_link_t & operator = (const cr_link_t & _other) {
		::memcpy (this, & _other, sizeof (cr_link_t));
		return * this;
	}
};

static
float _NormalizeAngle180 (float _fA) {
	if (_fA >= 180)
		_fA -= 360;
	if (_fA < -180)
		_fA += 360;
	assert (-180 <= _fA && _fA < 180);
	return _fA;
}

struct less_cr_link_t {
	less_cr_link_t (float _fA) : m_fA (_fA) {}

	bool operator () (const cr_link_t & _link1, const cr_link_t & _link2) {
		const float fA1 = _NormalizeAngle180 (_link1.fAngle - m_fA);
		const float fA2 = _NormalizeAngle180 (_link2.fAngle - m_fA);
		return fA1 < fA2;
	}

  private:
	const float m_fA;
};

struct crossroads_t {
	std::vector<cr_link_t> links;
	rgnode_t * const pNode;

	crossroads_t (map_t & _map, rgnode_t * _pNode);

	const cr_link_t * Find (const rgn_t * _pRgn, cr_link_type_t _lt) const {return _Find (_pRgn, _lt);}
	      cr_link_t * Find (const rgn_t * _pRgn, cr_link_type_t _lt)       {return const_cast<cr_link_t *> (_Find (_pRgn, _lt));}
	int CountExpectedLanes (cr_link_t * _pLink);

  private:
	const cr_link_t * _Find (const rgn_t * _pRgn, cr_link_type_t _lt) const;
};

crossroads_t::crossroads_t (map_t & _map, rgnode_t * _pNode) :
	pNode (_pNode)
{
	// Prepare list of links.
	const layer_t & polylines = _map.layers [rkPolyline];
	for (rgns_t::const_iterator iRgn = polylines.rgns.begin (); iRgn != polylines.rgns.end (); ++ iRgn) {
		const rgn_t & rgn = * iRgn;
		if (rgn.IsImmutable () || ! rgn.HasNode (_pNode))
			continue;

		const bool bOneWay = _map.pTypeSet->pfIsOneWayRoad (rgn);

		const size_t cElements = rgn.elements.size ();
		for (size_t cElement = 0; cElement < cElements; ++ cElement) {
			const rgn_element_t & element = rgn.elements [cElement];
			if (element.bDeleted)
				continue;
			if (! element.pExtra)
				continue;

			const size_t cPoint = element.FindNode (_pNode);
			if (cPoint == c_cUnknownPoint)
				continue;
			const point_t & pt = element.points [cPoint];

			const float fCosY = ::cos (pt.y*c_PI/180);

			// Scan for forward link.
			const size_t cPoints = element.points.size ();
			size_t c;
			for (c = cPoint + 1; c < cPoints; ++ c) {
				const rgnode_t * const pEndNode = element.pExtra->PointExtras [c].GetRGNode ();

				if (pEndNode) {
					const point_t & ptEnd = element.points [c];

					links.push_back     (cr_link_t (& rgn, & element, cPoint, ltOutgoing, pt, ptEnd, false, fCosY));
					if (! bOneWay)
						links.push_back (cr_link_t (& rgn, & element, cPoint, ltIncoming, pt, ptEnd, true,  fCosY));

					break;
				}
			}

			// Scan for backward link.
			for (c = cPoint - 1; c != (size_t) -1; -- c) {
				const rgnode_t * const pEndNode = element.pExtra->PointExtras [c].GetRGNode ();

				if (pEndNode) {
					const point_t & ptEnd = element.points [c];

					links.push_back     (cr_link_t (& rgn, & element, cPoint, ltIncoming, pt, ptEnd, false, fCosY));
					if (! bOneWay)
						links.push_back (cr_link_t (& rgn, & element, cPoint, ltOutgoing, pt, ptEnd, true,  fCosY));

					break;
				}
			}
		}
	}

	const size_t cLinks = links.size ();
	for (size_t c = 0; c < cLinks; ++ c) {
		cr_link_t & link = links [c];

		float fAngleDiffMin = 1000;
		cr_link_t * & pCouple = link.linkType == ltIncoming ? link.pLinkNext : link.pLinkPrev;
		cr_link_t * pSomeOther = NULL;

		for (size_t c2 = 0; c2 < cLinks; ++ c2) {
			cr_link_t & other = links [c2];
			if (link.linkType == other.linkType)
				continue;

			cr_link_t * const pFrom = link.linkType == ltIncoming ? & link : & other;
			cr_link_t * const pTo   = link.linkType == ltOutgoing ? & link : & other;

			// Exclude disabled turns.
			typedef rgrestrictions_t::iterator I;
			std::pair<I, I> bounds = _map.rgrestrictions.equal_range (_pNode);
			bool bRestricted = false;
			for (I iR = bounds.first; iR != bounds.second; ++ iR) {
				const rgrestriction_t & r = iR->second;

				if (
					r.pFromRoad == pFrom->pRgn &&
					r.pToRoad   == pTo  ->pRgn &&
					(r.bFromDirection == 0) == pFrom->bReverse &&
					(r.bToDirection   == 0) == pTo  ->bReverse
				) {
					bRestricted = true;
					break;
				}
			}
			if (bRestricted)
				continue;

			pSomeOther = & other;

			if (link.linkType == ltIncoming)
				++ link.nOutgoingLinks;
			else
				++ link.nIncomingLinks;

			// Check prev/next road.
			if (! (link.pRgn->strLabel == other.pRgn->strLabel) || link.pRgn->pExtra->dwRouteParams != other.pRgn->pExtra->dwRouteParams)
				continue;

			float fAngleDiff = ::fabsf (link.fDirection - other.fDirection);
			if (fAngleDiff > c_PI)
				fAngleDiff = ::fabsf (fAngleDiff - 2*c_PI);

			if (fAngleDiffMin > fAngleDiff) {
				fAngleDiffMin = fAngleDiff;
				pCouple = & other;
			}
		}

		if (link.nOutgoingLinks < 2 && link.nIncomingLinks < 2 && ! pCouple)
			pCouple = pSomeOther;
	}
}

const cr_link_t * crossroads_t::_Find (const rgn_t * _pRgn, cr_link_type_t _lt) const {
	for (size_t i = 0; i < links.size (); ++ i)
		if (links [i].linkType == _lt && links [i].pRgn == _pRgn)
			return & links [i];

	return NULL;
}

int crossroads_t::CountExpectedLanes (cr_link_t * _pLink) {
	_pLink->nLanesExpected      = 0;
	_pLink->nLanesExpectedMixed = 0;

	const size_t cLinks = links.size ();
	for (size_t i = 0; i < cLinks; ++ i) {
		const cr_link_t & link = links [i];

		if (& link == _pLink || ! link.bLanesDefined)
			continue;

		if (link.linkType == ltOutgoing)
			_pLink->nLanesExpected -= link.nLanes - link.nLanesMixed;
		else
			_pLink->nLanesExpected += link.nLanes;
	}

	if (_pLink->nIncomingLinks == 1 && _pLink->pLinkPrev && _pLink->pLinkPrev->pRgn == _pLink->pRgn)
		_pLink->nIncomingLinks = 0; // The only link is a U-turn.

	if (_pLink->nLanesExpected < 1) {
		_pLink->nLanesExpectedMixed = _pLink->nIncomingLinks == 1 ? 1 - _pLink->nLanesExpected : 0;
		_pLink->nLanesExpected = 1;
	}

	return _pLink->nLanesExpected;
}

///////////////////////////////////////////////////////////////////////////

static
void _AddComboIntItem (HWND _hCombo, int _nPos, int _nValue, int _nSelect) {
	char str [10];
	::_snprintf (str, ARRAY_SIZE (str), "%d", _nValue);

	::SendMessage (_hCombo, CB_INSERTSTRING, _nPos, (LPARAM) str);
	::SendMessage (_hCombo, CB_SETITEMDATA,  _nPos, _nValue);

	if (_nSelect == _nValue)
		::SendMessage (_hCombo, CB_SETCURSEL, _nPos, 0);
}

////////////////////////////////////////////////////////////////////////////

CRgnLanesPage::CRgnLanesPage (rgn_t & _rgn) :
	CRgnPPageBase (IDD_RGN_TRAFFIC_LANES, ppLanes, _rgn),
	m_pFirstCrossroads (NULL),
	m_pLastCrossroads (NULL)
{}

CRgnLanesPage::~CRgnLanesPage () {
	delete m_pFirstCrossroads;
	delete m_pLastCrossroads;
}

bool CRgnLanesPage::OnInit (HWND _hwndFocus) {
	const HWND hForwardLanes    = GetDlgItem (IDC_LANES_FORWARD_TOTAL);
	const HWND hForwardMixed    = GetDlgItem (IDC_LANES_FORWARD_TURNS);
	const HWND hForwardExpLeft  = GetDlgItem (IDC_LANES_FORWARD_EXP_LEFT);
//	const HWND hForwardExpRight = GetDlgItem (IDC_LANES_FORWARD_EXP_RIGHT);
	const HWND hReverseLanes    = GetDlgItem (IDC_LANES_REVERSE_TOTAL);
	const HWND hReverseMixed    = GetDlgItem (IDC_LANES_REVERSE_TURNS);
	const HWND hReverseExpLeft  = GetDlgItem (IDC_LANES_REVERSE_EXP_LEFT);
//	const HWND hReverseExpRight = GetDlgItem (IDC_LANES_REVERSE_EXP_RIGHT);

	const size_t cMaxLanes = 20;
	const bool bEnabled = m_rgn.pExtra && m_rgn.pExtra->lanes.bDefined != 0;
	int nReverseExpected = 0;
	int nForwardExpected = 0;

	m_lanes.dwLanes = 0;

	m_bForwardIsCont = false;
	m_bReverseIsCont = false;

	// First node crossroads.
	rgnode_t * const pFirstNode = m_rgn.GetNextNode (NULL, true);
	if (pFirstNode) {
		m_pFirstCrossroads = new crossroads_t (g_map, pFirstNode);

		cr_link_t * const pLink = m_pFirstCrossroads->Find (& m_rgn, ltOutgoing);
		assert (pLink);
		if (pLink) {
			m_pFirstCrossroads->CountExpectedLanes (pLink);

			if (bEnabled) {
				m_lanes.fForward        = pLink->nLanes;
				m_lanes.fForwardMixed   = pLink->nLanesMixed;
				m_lanes.nForwardExpLeft = pLink->nLanesExpLeft;
			} else {
				m_lanes.fForward        = pLink->nLanesExpected;
				m_lanes.fForwardMixed   = pLink->nLanesExpectedMixed;
				m_lanes.nForwardExpLeft = 0;
			}

			nForwardExpected   = pLink->nLanesExpected;
			m_nForwardIncoming = pLink->nIncomingLinks;
			m_bForwardIsCont   = pLink->pLinkPrev != NULL && pLink->pLinkPrev->pLinkNext == pLink && pLink->pRgn != pLink->pLinkPrev->pRgn;
		}
	} else {
		EnableDlgItem (IDC_TRAFFIC_LANES_ENABLED,  false);
		EnableDlgItem (IDC_LANES_FORWARD_TOTAL,    false);
		EnableDlgItem (IDC_LANES_FORWARD_TURNS,    false);
		EnableDlgItem (IDC_LANES_FORWARD_EXP_LEFT, false);
		EnableDlgItem (IDC_LANES_REVERSE_TOTAL,    false);
		EnableDlgItem (IDC_LANES_REVERSE_TURNS,    false);
		EnableDlgItem (IDC_LANES_REVERSE_EXP_LEFT, false);

		return true;
	}

	const bool bOneWay = g_map.pTypeSet->pfIsOneWayRoad (m_rgn);

	// Last node crossroads.
	rgnode_t * const pLastNode = m_rgn.GetNextNode (NULL, false);
	if (pLastNode) {
		m_pLastCrossroads = new crossroads_t (g_map, pLastNode);

		cr_link_t * const pLink = m_pLastCrossroads->Find (& m_rgn, ltOutgoing);
		if (pLink)
			m_pLastCrossroads->CountExpectedLanes (pLink);

		if (pLink && ! bOneWay) {
			if (bEnabled) {
				m_lanes.fReverse        = pLink->nLanes;
				m_lanes.fReverseMixed   = pLink->nLanesMixed;
				m_lanes.nReverseExpLeft = pLink->nLanesExpLeft;
			} else {
				m_lanes.fReverse        = pLink->nLanesExpected;
				m_lanes.fReverseMixed   = pLink->nLanesExpectedMixed;
				m_lanes.nReverseExpLeft = 0;
			}

			nReverseExpected   = pLink->nLanesExpected;
			m_nReverseIncoming = pLink->nIncomingLinks;
			m_bReverseIsCont   = pLink->pLinkPrev != NULL && pLink->pLinkPrev->pLinkNext == pLink && pLink->pRgn != pLink->pLinkPrev->pRgn;
		}
	}

	if (m_lanes.fReverse == 0)
		m_lanes.fReverse = 1;

	int n;
	for (n = 0; n < 9/*31*/; ++ n) {
		_AddComboIntItem (hForwardLanes, n, n + 1, m_lanes.fForward);
		_AddComboIntItem (hReverseLanes, n, n + 1, m_lanes.fReverse);
	}

	for (n = 0; n < 9/*16*/; ++ n) {
		_AddComboIntItem (hForwardMixed, n, n, m_lanes.fForwardMixed);
		_AddComboIntItem (hReverseMixed, n, n, m_lanes.fReverseMixed);
	}

	const int nMaxForwardLanes = __min (8, m_lanes.fForward - 1);
	const int nMaxReverseLanes = __min (8, m_lanes.fReverse - 1);

	for (n = -nMaxForwardLanes; n < 8; ++ n)
		_AddComboIntItem (hForwardExpLeft, n + nMaxForwardLanes, n, m_lanes.nForwardExpLeft);

	for (n = -nMaxReverseLanes; n < 8; ++ n)
		_AddComboIntItem (hReverseExpLeft, n + nMaxReverseLanes, n, m_lanes.nReverseExpLeft);

	m_lanes.fForward = nForwardExpected;
	m_lanes.fReverse = nReverseExpected;

	CheckDlgButton (IDC_TRAFFIC_LANES_ENABLED, bEnabled);

	_Update ();

	return true;
}

void CRgnLanesPage::OnActivatePage (const PSHNOTIFY * _pNotify) {
	const bool bIsOneWay = g_map.pTypeSet->pfIsOneWayRoad (m_rgn);

	EnableDlgItem (IDC_LANES_REVERSE_TOTAL,     ! bIsOneWay);
	EnableDlgItem (IDC_LANES_REVERSE_TURNS,     ! bIsOneWay);
	EnableDlgItem (IDC_LANES_REVERSE_EXP_LEFT,  ! bIsOneWay);
	EnableDlgItem (IDC_LANES_REVERSE_EXP_RIGHT, ! bIsOneWay);
	EnableDlgItem (IDC_STATIC_LANES_REVERSE1,   ! bIsOneWay);
	EnableDlgItem (IDC_STATIC_LANES_REVERSE2,   ! bIsOneWay);
	EnableDlgItem (IDC_STATIC_LANES_REVERSE3,   ! bIsOneWay);
	EnableDlgItem (IDC_STATIC_LANES_REVERSE4,   ! bIsOneWay);
	EnableDlgItem (IDC_STATIC_LANES_REVERSE5,   ! bIsOneWay);

	const HWND hFigureReverse = GetDlgItem (IDC_FIGURE_REVERSE);
	const DWORD dwOldStyle = ::GetWindowLong (hFigureReverse, GWL_STYLE);
	::SetWindowLong (hFigureReverse, GWL_STYLE, (bIsOneWay ? SS_GRAYFRAME : SS_BLACKFRAME) | (dwOldStyle & 0xFFFFFFF0));

	_Update ();

	CRgnPPageBase::OnActivatePage (_pNotify);
}

void CRgnLanesPage::_Update () {
	const bool bIsOneWay = g_map.pTypeSet->pfIsOneWayRoad (m_rgn);
	const bool bIsEnabled = IsDlgButtonChecked (IDC_TRAFFIC_LANES_ENABLED);
	const bool bIsReverseEnabled = bIsEnabled && ! bIsOneWay;

	EnableDlgItem (IDC_LANES_FORWARD_TOTAL,    bIsEnabled);
	EnableDlgItem (IDC_LANES_FORWARD_TURNS,    bIsEnabled && m_nForwardIncoming == 1);
	EnableDlgItem (IDC_LANES_FORWARD_EXP_LEFT, bIsEnabled && m_bForwardIsCont);
	EnableDlgItem (IDC_LANES_REVERSE_TOTAL,    bIsReverseEnabled);
	EnableDlgItem (IDC_LANES_REVERSE_TURNS,    bIsReverseEnabled && m_nReverseIncoming == 1);
	EnableDlgItem (IDC_LANES_REVERSE_EXP_LEFT, bIsReverseEnabled && m_bReverseIsCont);

	int nExpRight = 0;
	if (m_bForwardIsCont) {
		const int nLanes   = _GetSelectedData (IDC_LANES_FORWARD_TOTAL);
		const int nExpLeft = _GetSelectedData (IDC_LANES_FORWARD_EXP_LEFT);

		nExpRight = nLanes - m_lanes.fForward - nExpLeft;
	}

	SetDlgItemInt (IDC_LANES_FORWARD_EXP_RIGHT, nExpRight);

	if (! bIsOneWay) {
		nExpRight = 0;
		if (m_bReverseIsCont) {
			const int nLanes   = _GetSelectedData (IDC_LANES_REVERSE_TOTAL);
			const int nExpLeft = _GetSelectedData (IDC_LANES_REVERSE_EXP_LEFT);

			nExpRight = nLanes - m_lanes.fReverse - nExpLeft;
		}

		SetDlgItemInt (IDC_LANES_REVERSE_EXP_RIGHT, nExpRight);
	} else
		SetDlgItemInt (IDC_LANES_REVERSE_EXP_RIGHT, 0);

	::InvalidateRect (m_hWnd, NULL, TRUE);
	::UpdateWindow (m_hWnd);
}

void CRgnLanesPage::OnCommand (WORD _wCmd, WORD _wNotifyCode, HWND _hwndCtl) {
	switch (_wCmd) {
		case IDC_TRAFFIC_LANES_ENABLED:
			_Update ();
			PropSheetChanged ();
			break;

		case IDC_LANES_FORWARD_TOTAL:
		case IDC_LANES_FORWARD_TURNS:
		case IDC_LANES_FORWARD_EXP_LEFT:
		case IDC_LANES_FORWARD_EXP_RIGHT:
		case IDC_LANES_REVERSE_TOTAL:
		case IDC_LANES_REVERSE_TURNS:
		case IDC_LANES_REVERSE_EXP_LEFT:
		case IDC_LANES_REVERSE_EXP_RIGHT:
			if (_wNotifyCode == CBN_SELCHANGE) {
				_Update ();
				PropSheetChanged ();
			}
			break;
	}
}

int CRgnLanesPage::_GetSelectedData (UINT _uiId) const {
	const HWND hWnd = GetDlgItem (_uiId);
	const size_t cSel = ::SendMessage (hWnd, CB_GETCURSEL, 0, 0);
	if (cSel == CB_ERR)
		return 0;

	return ::SendMessage (hWnd, CB_GETITEMDATA, cSel, 0);
}

void CRgnLanesPage::OnApply (const PSHNOTIFY * _pPSHN) {
	if (IsDlgItemEnabled (IDC_TRAFFIC_LANES_ENABLED)) {
		lanes_t lanes = m_rgn.pExtra->lanes;
		lanes.bDefined        = IsDlgButtonChecked (IDC_TRAFFIC_LANES_ENABLED);
		lanes.fForward        = _GetSelectedData (IDC_LANES_FORWARD_TOTAL);
		lanes.fForwardMixed   = _GetSelectedData (IDC_LANES_FORWARD_TURNS);
		lanes.nForwardExpLeft = _GetSelectedData (IDC_LANES_FORWARD_EXP_LEFT);
		lanes.fReverse        = _GetSelectedData (IDC_LANES_REVERSE_TOTAL);
		lanes.fReverseMixed   = _GetSelectedData (IDC_LANES_REVERSE_TURNS);
		lanes.nReverseExpLeft = _GetSelectedData (IDC_LANES_REVERSE_EXP_LEFT);

		if (lanes.dwLanes != m_rgn.pExtra->lanes.dwLanes) {
			SetTrafficLanes (m_rgn, lanes.dwLanes);
			RefreshRequired ();
		}
	}

	OnApplyEnd (_pPSHN);
}

void CRgnLanesPage::OnPaint (const PAINTSTRUCT & _ps) const {
	const bool bIsEnabled = IsDlgButtonChecked (IDC_TRAFFIC_LANES_ENABLED);
	if (! bIsEnabled)
		return;

	_PaintFigure (_ps.hdc, true);

	const bool bIsOneWay = g_map.pTypeSet->pfIsOneWayRoad (m_rgn);
	if (! bIsOneWay)
		_PaintFigure (_ps.hdc, false);
}

void CRgnLanesPage::_PaintFigure (HDC _hDC, bool _bForward) const {
	const HWND hFigure = GetDlgItem (_bForward ? IDC_FIGURE_FORWARD : IDC_FIGURE_REVERSE);
	WINDOWPLACEMENT wp = {sizeof (WINDOWPLACEMENT)};
	::GetWindowPlacement (hFigure, & wp);

	const int x0 = wp.rcNormalPosition.left;
	const int x1 = wp.rcNormalPosition.right;
	const int y0 = wp.rcNormalPosition.top;
	const int y1 = wp.rcNormalPosition.bottom;
	const int width  = x1 - x0;
	const int height = y1 - y0;
	const int blank = 10;

	const int nLanesOutgoing = _GetSelectedData (_bForward ? IDC_LANES_FORWARD_TOTAL     : IDC_LANES_REVERSE_TOTAL);
	const int nExpLeft       = _GetSelectedData (_bForward ? IDC_LANES_FORWARD_EXP_LEFT  : IDC_LANES_REVERSE_EXP_LEFT);
	const int nExpRight      = GetDlgItemInt    (_bForward ? IDC_LANES_FORWARD_EXP_RIGHT : IDC_LANES_REVERSE_EXP_RIGHT);
	const int nLanesIncoming = nLanesOutgoing - nExpLeft - nExpRight;

	const float fLaneWidth = __min (30, float (width - 2*blank)/__max (nLanesOutgoing, nLanesIncoming));
	const int   nLaneWidth = fast_floor (fLaneWidth + 0.5f);

	const int y_outgoing = y0 + blank + nLaneWidth;
	const int y_incoming = y1 - 1;
	const int depth = (height - blank - nLaneWidth)/3;

	// Set clipping.
	const HRGN hClipRgn = ::CreateRectRgn (x0, y0, x1, y1);
	::SelectClipRgn (_hDC, hClipRgn);

	// Draw main road shape.
	const POINT pts [] = {
		{x0 + blank,                                                             y_outgoing + depth},
		{x0 + blank,                                                             y_outgoing},
		{x0 + blank + fast_floor (.5f + fLaneWidth*nLanesOutgoing),              y_outgoing},
		{x0 + blank + fast_floor (.5f + fLaneWidth*nLanesOutgoing),              y_outgoing + depth},

		{x0 + blank + fast_floor (.5f + fLaneWidth*(nExpLeft + nLanesIncoming)), y_incoming - depth},
		{x0 + blank + fast_floor (.5f + fLaneWidth*(nExpLeft + nLanesIncoming)), y_incoming},
		{x0 + blank + fast_floor (.5f + fLaneWidth*nExpLeft),                    y_incoming},
		{x0 + blank + fast_floor (.5f + fLaneWidth*nExpLeft),                    y_incoming - depth},
	};
	::SelectObject (_hDC, g_hbrWhite);
	::SelectObject (_hDC, g_hpenBlackThin);
	::Polygon      (_hDC, pts, ARRAY_SIZE (pts));

	// Draw outgoing lanes separators.
	::SelectObject (_hDC, g_hpenGrayThin);
	int nLane;
	for (nLane = 1; nLane < nLanesOutgoing; ++ nLane) {
		const int x_separator = x0 + blank + fast_floor (.5f + fLaneWidth*nLane);
		DrawLine (_hDC, x_separator, y_outgoing, x_separator, y_outgoing + depth);
	}
	// Draw incoming lanes separators.
	for (nLane = 1; nLane < nLanesIncoming; ++ nLane) {
		const int x_separator = x0 + blank + fast_floor (.5f + fLaneWidth*(nExpLeft + nLane));
		DrawLine (_hDC, x_separator, y_incoming, x_separator, y_incoming - depth);
	}

	crossroads_t cr = _bForward ? * m_pLastCrossroads : * m_pFirstCrossroads;
	cr_link_type_t lt = _bForward ? ltIncoming : ltOutgoing;

	// Get direction of the main road.
	const cr_link_t * const pLink = cr.Find (& m_rgn, lt);
	if (pLink && pLink->pElement) {
		const points_t & points = pLink->pElement->points;

		point_t pt0, pt1;
		bool bADefined = true;
		if (lt == ltIncoming) {
			pt1 = points [pLink->cPoint];
			bADefined = pLink->cPoint > 0;
			pt0 = bADefined ? points [pLink->cPoint - 1] : pt1;
		} else {
			pt0 = points [pLink->cPoint];
			bADefined = pLink->cPoint + 1 < points.size ();
			pt1 = bADefined ? points [pLink->cPoint + 1] : pt0;
		}

		if (bADefined) {
			const float fA = ::atan2 (pt1.y - pt0.y, (pt1.x - pt0.x)*::cos (pt0.y*c_PI/180));

			// Order by turn angles.
			std::sort (cr.links.begin (), cr.links.end (), less_cr_link_t (fA));

			// Draw outgoing maneuvers.
			::SelectObject (_hDC, g_hpenBlackThin);
			const int D = 2*nLaneWidth;
			nLane = nLanesOutgoing;
			const size_t cLinks = cr.links.size ();
			for (size_t cLink = 0; cLink < cLinks; ++ cLink) {
				const cr_link_t & link = cr.links [cLink];
				if (& link == pLink)
					continue;
				nLane -= link.nLanes - link.nLanesMixed;
				if (nLane < 0)
					break;

				const float fTurnAngle = _NormalizeAngle180 (fA - link.fAngle);
				const float fCos  = ::cos (fTurnAngle);
				const float fSin  = ::sin (fTurnAngle);
				const float fCos2 = ::cos (fTurnAngle/2);
				const float fSin2 = ::sin (fTurnAngle/2);

				for (int n = 0; n < link.nLanes; ++ n) {
					const int x = x0 + blank + fast_floor (.5f + fLaneWidth*(nLane + n + .5f));

					int Dh = D/6*(fTurnAngle < 0 ? n + 1 : link.nLanes - n);
					if (::fabsf (fTurnAngle) < 0.1f)
						Dh = 0;

					const POINT points [] = {
						{x,                                        y_outgoing},
						{x + fast_floor (.5f + fSin2*Dh),          y_outgoing - fast_floor (.5f + fCos2*Dh)},
						{x + fast_floor (.5f + fSin2*Dh + fSin*D), y_outgoing - fast_floor (.5f + fCos2*Dh + fCos*D)}
					};
					::Polyline       (_hDC, points, ARRAY_SIZE (points));
					DrawDirIndicator (_hDC, points, ARRAY_SIZE (points), & wp.rcNormalPosition, true);
				}
			}
		}
	}

	// Remove clippig.
	::SelectClipRgn (_hDC, NULL);
	::DeleteObject (hClipRgn);
}
