//
// GPSMapEdit
// (c) Konstantin Galichsky (kg@geopainting.com), 2008-2009
//
// Rotate object.
//

# include "StdAfx.h"
# include "RotateObject.h"
# include "UndoRedo.h"
# include "RefreshHoles.h"
# include "Label.h"
# include "Globals.h"

class CRotateObjectAction : public CEditAction {
  public:
	CRotateObjectAction (rgn_t * _pRgn, size_t _cElementOrOrigin, float _fTurnAngle, const point_t & _pt0) :
		m_pRgn (_pRgn),
		m_cElementOrOrigin (_cElementOrOrigin),
		m_fTurnAngle (_fTurnAngle),
		m_pt0 (_pt0),
		m_bWasModified ((_pRgn->dwFlags & rgn_t::fModified) != 0)
	{}

	virtual void GetName (string_t & _str) const;
	virtual cu_t CanUndo () const {return cuYes;}
	virtual bool Do ();
	virtual void Undo ();
	virtual void Redo ();

  private:
	const float   m_fTurnAngle;
	const point_t & m_pt0;
	rgn_t * const m_pRgn;
	const size_t  m_cElementOrOrigin;
	const bool    m_bWasModified;
	polypoints_t  m_pointsSaved;

	void Rotate (float _fTurnAngle);
};

void CRotateObjectAction::GetName (string_t & _str) const {
	CCritSec::CLock lock (g_cs);

	assert (m_pRgn);
	if (m_cElementOrOrigin == -1)
		_str.append ("rotate [");
	else
		_str.append ("rotate element of [");
	GetHintText (* m_pRgn, _str);
	_str.append ("]");
}

void CRotateObjectAction::Rotate (float _fTurnAngle) {
	CCritSec::CLock lock (g_cs);

	assert (m_pRgn);

	if (m_pRgn->IsWTR () || m_pRgn->IsPoint ())
		return;

	//
	// Back-up points.
	//

	const size_t cElements = m_pRgn->elements.size ();
	for (size_t cElement = 0; cElement < cElements; ++ cElement) {
		rgn_element_t & element = m_pRgn->elements [cElement];
		if (element.bDeleted)
			continue;

		m_pointsSaved.push_back (points_t ());
		points_t & points = m_pointsSaved.back ();
		points = element.points;
	}

	//
	// Rotate points.
	//

	const float fCosTA = ::cos (_fTurnAngle);
	const float fSinTA = ::sin (_fTurnAngle);

	if (m_cElementOrOrigin == -1) {
		const float fCosY  = ::cos (m_pt0.y*c_PI/180.f);

		for (size_t cElement = 0; cElement < cElements; ++ cElement) {
			rgn_element_t & element = m_pRgn->elements [cElement];
			if (element.bDeleted)
				continue;

			points_t & points = element.points;
			const size_t cPoints = points.size ();
			for (size_t cPoint = 0; cPoint < cPoints; ++ cPoint) {
				point_t & pt = points [cPoint];

				point_t ptRotated;
				ptRotated.x = m_pt0.x + ( (pt.x - m_pt0.x)*fCosY*fCosTA + (pt.y - m_pt0.y)*fSinTA)/fCosY;
				ptRotated.y = m_pt0.y + (-(pt.x - m_pt0.x)*fCosY*fSinTA + (pt.y - m_pt0.y)*fCosTA);

				pt = ptRotated;
			}
		}
	} else {
		assert (m_cElementOrOrigin < m_pRgn->elements.size ());
		rgn_element_t & elem = m_pRgn->elements [m_cElementOrOrigin];

		const float fCosY  = ::cos (m_pt0.y*c_PI/180.f);

		points_t & points = elem.points;
		const size_t cPoints = points.size ();
		for (size_t cPoint = 0; cPoint < cPoints; ++ cPoint) {
			point_t & pt = points [cPoint];

			point_t ptRotated;
			ptRotated.x = m_pt0.x + ( (pt.x - m_pt0.x)*fCosY*fCosTA + (pt.y - m_pt0.y)*fSinTA)/fCosY;
			ptRotated.y = m_pt0.y + (-(pt.x - m_pt0.x)*fCosY*fSinTA + (pt.y - m_pt0.y)*fCosTA);

			pt = ptRotated;
		}
	}

	m_pRgn->RefreshAttributes (g_map, false);
}

bool CRotateObjectAction::Do () {
	Rotate (m_fTurnAngle);
	m_pRgn->dwFlags |= rgn_t::fModified;
	return true;
}

void CRotateObjectAction::Undo () {
	// Restore points.
	if (! m_pointsSaved.empty ()) {
		polypoints_t::iterator i = m_pointsSaved.begin ();
		const size_t cElements = m_pRgn->elements.size ();
		for (size_t cElement = 0; cElement < cElements; ++ cElement) {
			rgn_element_t & element = m_pRgn->elements [cElement];
			if (element.bDeleted)
				continue;

			if (i == m_pointsSaved.end ())
				break;
			element.points = * i;
			++ i;
		}
	}

	m_pRgn->RefreshAttributes (g_map, false);

	if (! m_bWasModified)
		m_pRgn->dwFlags &= ~rgn_t::fModified;
}

void CRotateObjectAction::Redo () {
	CRotateObjectAction::Do ();
}

///////////////////////////////////////////

void RotateObject (rgn_t * _pRgn, size_t _cElementOrOrigin, float _fTurnAngle, const point_t & _pt0) {
	PerformEditAction (new CRotateObjectAction (_pRgn, _cElementOrOrigin, _fTurnAngle, _pt0));

	if (_pRgn->IsPolygon () && _cElementOrOrigin != -1)
		RefreshHoles (_pRgn, true);
}
