//
// GPSMapEdit
// (c) Konstantin Galichsky (kg@geopainting.com), 2002-2009
//
// 'File|Save' and 'File|Save As...' commands.
//

# include "StdAfx.h"
# include "Resource.h"
# include "Save.h"
# include "UndoRedo.h"
# include "Map.h"
# include "Log.h"
# include "Menu.h"
# include "Status.h"
# include "RecentFiles.h"
# include "File.h"
# include "fast_floor.h"
# include "Globals.h"

bool CanSave () {
	return g_bModified && g_pSaverInfo && (g_pSaverInfo->m_dwFlags & ftiImportExport) == 0 && g_strFileName [0];
}

bool RequestSave () {
	if (g_bModified) {
		switch (::MessageBox (g_hWnd, "The map is modified.\nDo you want to save it?", "Save", MB_YESNOCANCEL | MB_ICONQUESTION)) {
			case IDYES:
				if (g_pSaverInfo && (g_pSaverInfo->m_dwFlags & ftiImportExport) == 0 && g_strFileName [0])
					SaveMap ();
				else
					SaveMapAs ();
				return true;
			case IDNO:     return true;
			case IDCANCEL: return false;
		}
	}
	return true;
}

void SaveMapAs () {
	CFileTypeInfo::ShowSaveFile ();
}

void SaveMap () { 
	SaveMap (g_pSaverInfo, g_strFileName, false);
}

bool SaveMap (const CFileTypeInfo * _pSaverInfo, const char * _strFileName, bool _bSilent) {
	assert (_pSaverInfo);

	CMapSaver * const pSaver = _pSaverInfo->m_pfCreateSaver ();
	CObjectHolder<CMapSaver> pSaver_Holder (pSaver);
	pSaver->SetSilent (_bSilent);

	//
	// Prepare map to export.
	//

	SetStatus (GetStr (IDS_STATUS_PREPARE_SAVING));

	// Prepare map copy, if required.
	const bool bCopyMap = (_pSaverInfo->m_dwFlags & ftiSaveCopyOfMap) != 0;
	map_t * pMap;
	if (bCopyMap) {
		pMap = new map_t;

		CCritSec::CLock lock (g_cs);
		* pMap = g_map;
	} else
		pMap = & g_map;
	CObjectHolder<map_t> pMap_Holder (bCopyMap ? pMap : NULL);

	CCritSec::CLock lock (g_cs, ! bCopyMap);

	// Check if map is savable.
	SetProgressStatus (0);
	if (! pSaver->PrepareToSave (* pMap)) {
		SetStatusReady ();
		return false;
	}

	// Rename old file to .BAK.
	bool bHaveBakFile = false;
	char strBakFile [MAX_PATH + 1] = "";
	if (DoesFileExist (_strFileName)) {
		CreateUniqueFileName (_strFileName, "BAK", strBakFile);
		::MoveFile (_strFileName, strBakFile);
		bHaveBakFile = true;
	}

	CWaiter w;

	const DWORD dwStartTime = ::GetTickCount ();

	//
	// Open the file.
	//

	CFileForWrite hFile (_strFileName);
	if (! hFile.IsValid ()) {
		if (_bSilent) {
			ReportWarning ("Can't open file\n%s.", _strFileName);
			ShowMessageLog ();
		} else
			ErrorMsgBox ("Can't open file\n%s.", _strFileName);
		SetStatusReady ();
		return false;
	}

	//
	// Save the map.
	//

	ClearLog ();
	ReportText ("Saving map to file '%s'...", _strFileName);

	SetStatus (GetStr (IDS_STATUS_SAVING));
	SetProgressStatus (0);
	const SaveStatus_t ss = ExecuteSaveFileSafely (pSaver, hFile, _strFileName, * pMap);

	//
	// Report about results.
	//

	if (ss != ssOK) {
		if (! _bSilent)
			ErrorMsgBox ("Saving has failed.");
		SetStatusReady ();

		::DeleteFile (_strFileName);

		ReportText ("");
		ReportText ("ERROR: Saving has failed.");
		ShowMessageLog ();

		return false;
	}

	if (GetWarningsCount ())
		ReportText ("%d warning(s).", GetWarningsCount ());

	ReportText ("Export time %0.2f s", .001f*(::GetTickCount () - dwStartTime));

	// Delete BAK file.
	if (ss == ssOK && bHaveBakFile && strBakFile [0] != '\0')
		::DeleteFile (strBakFile);

	//
	// Update GUI elements.
	//

	if (! _bSilent) {
		CFileTypeInfo * const pLoaderInfo = CFileTypeInfo::FindLoaderInfoByClass (_pSaverInfo->m_strClass);
		if (pLoaderInfo != NULL)
			AddRecentFile (_strFileName);
		else
			::SHAddToRecentDocs (SHARD_PATH, _strFileName);
	}

	ResetUndoRedo ();
	g_bModified = false;

	g_pSaverInfo = _pSaverInfo;
	::strncpy (g_strFileName, _strFileName, MAX_PATH);
	UpdateTitle ();

	UpdateMenu ();

	SetStatusReady ();

	return true;
}

///////////////////////////////////////////////////////////////

bool CheckVectorData_Silent (const map_t & _map) {
	CCritSec::CLock lock (g_cs);

	for (size_t cKind = 0; cKind < rkMax; ++ cKind) {
		const layer_t & layer = _map.layers [cKind];
		const rgns_t & rgns = layer.rgns;
		for (rgns_t::const_iterator iRgn = rgns.begin (); iRgn != rgns.end (); ++ iRgn) {
			const rgn_t & rgn = * iRgn;
			if (rgn.IsImmutable () || rgn.IsNotMapObject ())
				continue;

			return true;
		}
	}

	return false;
}

bool CheckVectorData (const map_t & _map) {
	CCritSec::CLock lock (g_cs);

	const bool bThereAreVectorData = CheckVectorData_Silent (_map);
	if (! bThereAreVectorData) {
		ErrorMsgBox (
			"There are no vector data to export.\n"
			"\n"
			"NOTE: raster maps should be vectorized before loading to GPS receiver,\n"
			"and tracks and waypoints should be converted to map objects."
		);
		return false;
	}

	if (g_map.bNoHeader || g_map.Levels.empty ()) {
		ErrorMsgBox (
			"The map has no header data or zoom levels.\n"
			"Please check settings in 'File | Map Properties'."
		);
		return false;
	}

	return true;
}

bool CheckBookmarks_Silent (const map_t & _map) {
	CCritSec::CLock lock (g_cs);

	const layer_t & layer = _map.layers [rkPoint];
	const rgns_t & rgns = layer.rgns;
	for (rgns_t::const_iterator iRgn = rgns.begin (); iRgn != rgns.end (); ++ iRgn) {
		const rgn_t & rgn = * iRgn;
		if (rgn.IsImmutable () || ! rgn.IsBookmark ())
			continue;

		return true;
	}

	return false;
}

bool CheckBookmarks (const map_t & _map) {
	CCritSec::CLock lock (g_cs);

	const bool bThereAreBookmarks = CheckBookmarks_Silent (_map);
	if (! bThereAreBookmarks) {
		ErrorMsgBox ("There are no bookmarks to export.\n");
		return false;
	}

	return true;
}

///////////////////////////////////////////////////////////////

void CMapSaver::Warning (const char * _strMessage) {
	if (Silent ())
		ReportWarning (_strMessage);
	else
		WarningMsgBox (_strMessage);
}

void CMapSaver::Warning (const char * _strMessage, int _n1) {
	if (Silent ())
		ReportWarning (_strMessage, _n1);
	else
		WarningMsgBox (_strMessage, _n1);
}

SaveStatus_t CMapSaver::Error (const char * _strMessage) {
	if (Silent ()) {
		ReportWarning (_strMessage);
		return ssFailed;
	} else {
		ErrorMsgBox (_strMessage);
		return ssCanceled;
	}
}

SaveStatus_t CMapSaver::Error (const char * _strMessage, const char * _str1) {
	if (Silent ()) {
		ReportWarning (_strMessage, _str1);
		return ssFailed;
	} else {
		ErrorMsgBox (_strMessage, _str1);
		return ssCanceled;
	}
}

