//
// GPSMapEdit
// (c) Konstantin Galichsky (kg@geopainting.com), 2002-2009
// (c) Philipp Peshin (phil@mobilae.ru), 2003
// (c) Alexander Litvinov (lan@ac-sw.com), 2006
//
// Object selection handling.
//

# include "StdAfx.h"
# include "Resource.h"
# include "Selection.h"
# include "Position.h"
# include "Scale.h"
# include "MapView.h"
# include "UndoRedo.h"
# include "Menu.h"
# include "Load.h"
# include "Levels.h"
# include "Cursors.h"
# include "TypesDraw.h"
# include "DrawNumbers.h"
# include "Colors.h"
# include "fast_floor.h"
# include "Globals.h"

//////////////////////////////////////////////////////////////

enum {
	cnSelPolylineContour = 0,
	cnSelPolylineFill1   = 1,
	cnSelPolylineFill2   = 2,
	cnSelNodeContour     = 3,
	cnSelNodeFill        = 4,
	cnSelNodeAddedFill   = 5,
	cnSelNodeStickFill   = 6,
	cnSelPointCircle     = 7,

	cnSelMax
};

static CColors * g_pSelColors;

static const COLORREF g_colSels [cnSelMax] = {
	c_colMagenta,
	c_colWhite,
	c_colBlue,
	c_colBlack,
	c_colYellow,
	RGB (255, 128, 64),
	c_colRed,
	c_colDkRed,
};

void InitSelection () {
	assert (g_pSelColors == NULL);
	g_pSelColors = new CColors;
	g_pSelColors->LoadFromSetting ("SelColors", g_colSels, ARRAY_SIZE (g_colSels));
}

void DestroySelection () {
	delete g_pSelColors;
	g_pSelColors = NULL;
}

bool AdjustColorsForSelection (HWND _hWnd) {
	assert (g_pSelColors != NULL);

	const size_t cColors = g_pSelColors->GetColors ();
	const char ** const strColorNames = (const char **) alloca (sizeof (char *)*cColors);

	fstring str [cnSelMax];
	assert (cColors == ARRAY_SIZE (str));
	str [0].LoadFromResource (IDS_SEL_POLYLINE_OUTER);
    str [1].LoadFromResource (IDS_SEL_POLYLINE_INNER_1);
    str [2].LoadFromResource (IDS_SEL_POLYLINE_INNER_2);
    str [3].LoadFromResource (IDS_SEL_NODE_OUTER);
    str [4].LoadFromResource (IDS_SEL_NODE_INNER);
	str [5].LoadFromResource (IDS_SEL_NODE_INNER_ADDED);
	str [6].LoadFromResource (IDS_SEL_NODE_INNER_STICK);
	str [7].LoadFromResource (IDS_SEL_POINT_CIRCLE);

	for (size_t c = 0; c < __min (cColors, ARRAY_SIZE (str)); ++ c)
		strColorNames [c] = str [c].c_str ();

	if (g_pSelColors->Adjust (_hWnd, g_colSels, strColorNames, cColors)) {
		g_pSelColors->SaveToSetting ("SelColors");
		return true;
	}
	return false;
}

//////////////////////////////////////////////////////////////

bool IsSelMultipleOn (DWORD _dwKeys) {
	if (g_bMSelByShift)
		return (_dwKeys & MK_SHIFT)   != 0;
	else
		return (_dwKeys & MK_CONTROL) != 0;
}

bool IsSelElementOn (DWORD _dwKeys) {
	if (g_bMSelByShift)
		return (_dwKeys & MK_CONTROL) != 0;
	else
		return (_dwKeys & MK_SHIFT)   != 0;
}

//////////////////////////////////////////////////////////////

static
void _CleanSelection () {
	CCritSec::CLock lock (g_cs);

	size_t cItems = g_Selection.size ();
	size_t c;
	for (c = 0; c < cItems; ++ c) {
		const find_info_t & fi = g_Selection [c];

		bool bDelete = false;
		if (fi.pLayer) {
			const rgn_t & rgn = * fi.iRgn;
			if (rgn.IsDeleted ())
				bDelete = true;
			if (fi.cElementOrOrigin != -1 && fi.cElementOrOrigin < rgn.elements.size () && rgn.elements [fi.cElementOrOrigin].bDeleted)
				bDelete = true;
			if (fi.cElementOrOrigin != -1 && fi.cElementOrOrigin < rgn.origins .size () && rgn.origins  [fi.cElementOrOrigin].bDeleted)
				bDelete = true;
		} else if (fi.pRenderer) {
			if (fi.pRenderer->bDeleted)
				bDelete = true;
		}

		if (bDelete) {
			g_Selection.erase (g_Selection.begin () + c);
			-- c;
			-- cItems;
			assert (cItems == g_Selection.size ());
		}
	}

	size_t cSels = g_SelPoints.size ();
	for (c = 0; c < cSels; ++ c) {
		const sel_point_t & sp = g_SelPoints [c];
		if (sp.pRgn && sp.pRgn->IsDeleted ()) {
			g_SelPoints.erase (g_SelPoints.begin () + c);
			-- c;
			-- cSels;
			assert (cSels == g_SelPoints.size ());
		}
	}
}

INVOKE_BEFORE_GC (_CleanSelection);

//////////////////////////////////////////////////////////////

find_info_m_t g_Selection;

find_info_t FindRgn (map_t & _map, int _x, int _y, const rgn_t * _pExceptRgn, bool _bEnableStickedNodes) {
	if (g_bLoading)
		return find_info_t ();

	CCritSec::CLock lock (g_cs);

	point_t p;
	p.x = g_map.rectBound.x0 + g_X + _x/g_kX;
	p.y = g_map.rectBound.y0 + g_Y + (g_lHeight - _y)/g_kY;

	return _map.Find (p, _pExceptRgn, _bEnableStickedNodes);
}

find_info_t FindRgn (int _x, int _y, const rgn_t * _pExceptRgn, bool _bEnableStickedNodes) {
	return FindRgn (g_map, _x, _y, _pExceptRgn, _bEnableStickedNodes);
}

bool StickPointToNodeOrGrid (size_t _cLevelIdx, int & _x, int & _y, point_t & _pointStick, const rgn_t * _pExceptRgn) {
	if (! g_bEnableSticking) {
		ScreenToPt (_x, _y, _pointStick);
		g_map.pTypeSet->pfRoundPoint (g_map, _cLevelIdx, _pointStick);
		_x = XToScreen (_pointStick.x);
		_y = YToScreen (_pointStick.y);

		return false;
	}

	const find_info_t fiStickTo = FindRgn (_x, _y, _pExceptRgn, true);

    bool bRes = (fiStickTo.pLayer != NULL);
    if (bRes) {
        const rgn_t & rgn = * fiStickTo.iRgn;
		if (rgn.pExtra && fiStickTo.cEntryPoint != -1) {
			// All entry points stick exactly.
			bRes = true;
			_pointStick = fiStickTo.ptNearest;
		} else if (rgn.IsPolygon () || rgn.IsPolyline ()) {
		    // The precision of matching in degrees.
		    const float c_fE = c_fE_in_pixels/g_kY;

			if (fiStickTo.cElementOrOrigin < rgn.elements.size ()) {
				const points_t & pts = rgn.elements [fiStickTo.cElementOrOrigin].points;
				const size_t cIndex = PointsContainPt (pts, fiStickTo.ptNearest, c_fE, g_map.fCosY);
				bRes = (cIndex != -1);
	            if (bRes)
		            _pointStick = pts [cIndex];
			}
        } else {
            // All POI/points stick exactly.
            bRes = true;
            _pointStick = fiStickTo.ptNearest;
        }
    }
    if (! bRes)
        ScreenToPt (_x, _y, _pointStick);

    g_map.pTypeSet->pfRoundPoint (g_map, _cLevelIdx, _pointStick);
	_x = XToScreen (_pointStick.x);
	_y = YToScreen (_pointStick.y);

	return bRes;
}

///////////////////////////////////////////////////////////////////////

static
find_info_m_t::iterator _FindSel (const find_info_t & _SelItem) {
	return std::find (g_Selection.begin (), g_Selection.end (), _SelItem);
}

bool IsInSelection (const find_info_t & _SelItem) {
	// TODO: check elements.
	const find_info_m_t::iterator i = _FindSel (_SelItem);
	return i != g_Selection.end ();
}

bool SelectObjectAt (int _x, int _y, bool _bAdd, bool _bElementOnly, bool _bEntryPointEnabled, bool _bRight, bool _bEnableRedraw) {
	if (_bRight && ! _bElementOnly) {
		//
		// NOTE: On right click to selection, do not change existing selection.
		//

		point_t pt;
		ScreenToPt (_x, _y, pt);

	    // The precision of matching in degrees.
	    const float c_fE = c_fE_in_pixels/g_kY;

		const size_t cLevelIdx = GetZoomLevel (g_fScale);

		const size_t cSels = g_Selection.size ();
		for (size_t c = 0; c < cSels; ++ c) {
			const find_info_t & fi = g_Selection [c];
			if (fi.pLayer) {
				const rgn_t & rgn = * fi.iRgn;
				if (rgn.Contains (pt, c_fE, cLevelIdx) && fi.cEntryPoint == -1) {
					if (rgn.pExtra && rgn.pExtra->ContainsEntryPoint (pt, c_fE))
						continue;
					return true;
				}
			} else if (fi.pRenderer) {
				if (fi.pRenderer->rect.Contains (pt))
					return true;
			}
		}
	}

	find_info_t SelItem = FindRgn (_x, _y);
	if (SelItem.pLayer && ! _bElementOnly)
		SelItem.cElementOrOrigin = -1;
	if (SelItem.pLayer && ! _bEntryPointEnabled)
		SelItem.cEntryPoint      = -1;

	bool bRedraw = false;
	bool bInsideSelection = false;

	if (SelItem.empty ()) {
		if (! _bAdd) {
			// Reset the selection.
			g_Selection.clear ();

			bRedraw = true;
		} else {
			// Do not modify the selection.
		}
	} else {
		const find_info_m_t::iterator i = _FindSel (SelItem);
		if (i != g_Selection.end ()) {
			// NOTE: The item is already selected.
			if (_bAdd) {
				// Exclude it from the selection.
				// TODO: handle whole & elements as well.
				g_Selection.erase (i);
				bRedraw = true;
			} else {
				// Do not modify the selection but prepare to drag it.
				bInsideSelection = true;

				// NOTE: re-selection may be at different point.
				i->cPoint = SelItem.cPoint;
			}
		} else {
			if (! _bAdd)
				// Reset the selection.
				g_Selection.clear ();

			// Include to the selection.
			// TODO: handle whole & elements as well.
			g_Selection.push_back (SelItem);

			bRedraw = true;
			bInsideSelection = true;
		}
	}

	// Redraw map with the new selection.
	if (bRedraw && _bEnableRedraw) {
		Redraw ();
		UpdateMenu ();
	}

	return bInsideSelection;
}

void SelectObjectsInsideRect (const rect_t & _rectSelection, bool _bAdd, bool _bElementOnly) {
	if (! _bAdd)
		g_Selection.clear ();

	find_info_t SelItem;
	for (int cKind = rkMax - 1; cKind >= 0; -- cKind) {
		layer_t & layer = g_map.layers [cKind];
		rgns_t & rgns = layer.rgns;
		for (rgns_t::iterator iRgn = rgns.begin (); iRgn != rgns.end (); ++ iRgn) {
			rgn_t & rgn = * iRgn;

			if (! IsRgnKindVisible (rgn, g_map))
				continue;

			if (! rgn.rectBound.Intersects (_rectSelection)) 
				continue;

			bool bAddRgn = false;

			// Match origins.
			for (size_t cOrigin = 0; cOrigin < rgn.origins.size (); ++ cOrigin) {
				const rgn_origin_t & origin = rgn.origins [cOrigin];
				if (origin.bDeleted)
					continue;

				if (! rgn.IsAttachment () && rgn.IsMapObject () && ! LevelIsVisible (origin.cLevel, rgn.uchEndLevel))
					continue;

				if (_rectSelection.Contains (origin.point)) {
					if (_bElementOnly) {
						SelItem.pLayer           = & layer;
						SelItem.iRgn             = iRgn;
						SelItem.cElementOrOrigin = cOrigin;

						g_Selection.push_back (SelItem);

						continue;
					} else
						bAddRgn = true;
				}
			}
			
			// Match elements.
			for (size_t cElement = 0; cElement < rgn.elements.size (); ++ cElement) {
				const rgn_element_t & element = rgn.elements [cElement]; 
				if (element.bDeleted)
					continue;

				const points_t & points = element.points;
				if (points.empty ())
					continue;

				if (! element.rectBound.Intersects (_rectSelection))
					continue;

				if (! rgn.IsAttachment () && rgn.IsMapObject () && ! LevelIsVisible (element.cLevel, rgn.uchEndLevel))
					continue;

				if (RectIntersectsPolyline (_rectSelection, points, rgn.IsPolygon ())) {
					if (_bElementOnly) {
						SelItem.pLayer           = & layer;
						SelItem.iRgn             = iRgn;
						SelItem.cElementOrOrigin = cElement;

						if (! IsInSelection (SelItem))
							g_Selection.push_back (SelItem);
					} else
						bAddRgn = true;
				}
			}

			if (bAddRgn) {
				SelItem.pLayer           = & layer;
				SelItem.iRgn             = iRgn;
				SelItem.cElementOrOrigin = -1;

				if (! IsInSelection (SelItem))
					g_Selection.push_back (SelItem);
			}
		}
	}
}

void SelectAllObjects (bool _bObjects, bool _bBookmarks, bool _bNotes, bool _bTracks, bool _bWaypoints, bool _bRoutes, bool _bRasters) {
	ClearSelPoints ();
	g_Selection.clear ();

	find_info_t SelItem;
	for (size_t cKind = 0; cKind < rkMax; ++ cKind) {
		layer_t & layer = g_map.layers [cKind];
		rgns_t & rgns = layer.rgns;
		for (rgns_t::iterator iRgn = rgns.begin (); iRgn != rgns.end (); ++ iRgn) {
			rgn_t & rgn = * iRgn;

			if (rgn.IsDeleted ())
				continue;

			bool bAdd = false;
			if (_bObjects   && rgn.IsMapObject () && ! rgn.IsAttachment ())
				bAdd = true;
			if (_bBookmarks && rgn.IsBookmark  () && ! rgn.IsAttachment ())
				bAdd = true;
			if (_bNotes     && rgn.IsNote      () && ! rgn.IsAttachment ())
				bAdd = true;
			if (_bTracks    && rgn.IsTrack ())
				bAdd = true;
			if (_bWaypoints && rgn.IsWaypoint ())
				bAdd = true;
			if (_bRoutes    && rgn.IsRoute ())
				bAdd = true;

			if (bAdd) {
				SelItem.pRenderer        = NULL;
				SelItem.pLayer           = & layer;
				SelItem.iRgn             = iRgn;
				SelItem.cElementOrOrigin = -1;

				g_Selection.push_back (SelItem);
			}
		}
	}

	if (_bRasters) {
		for (renderers_t::iterator iRenderer = g_map.renderers.begin (); iRenderer != g_map.renderers.end (); ++ iRenderer) {
			renderer_base_t * const pRenderer = & iRenderer->data ();
			if (pRenderer->bDeleted || ! pRenderer->bVisible)
				continue;
			
			SelItem.pRenderer        = pRenderer;
			SelItem.pLayer           = NULL;
			SelItem.cElementOrOrigin = -1;

			g_Selection.push_back (SelItem);
		}
	}

	// Redraw map with selection.
	Redraw ();
	UpdateMenu ();
}

void UnselectObjects () {
	if (g_Selection.empty ())
		return;

	g_Selection.clear ();

	// Redraw map without selection.
	Redraw ();
	UpdateMenu ();
}

////////////////////////////////////////////////////

bool IsInvertSelectionDisabled () {
	const size_t cSelItems = g_Selection.size ();
	for (size_t cSelItem = 0; cSelItem < cSelItems; ++ cSelItem) {
		const find_info_t & SelItem = g_Selection [cSelItem];

		// Check if we have raster image selection.
		if (SelItem.pLayer == NULL)
			return true;

		// Check if we have element selection.
		if (SelItem.cElementOrOrigin != -1)
			return true;
	}

	return false;
}

void InvertSelection () {
	// Clear temporary flag at all rgns.
	g_map.ClearTemporaryFlags ();

	// Set temp flag for selected rgns.
	if (! g_Selection.empty ()) {
		for (find_info_m_t::iterator selIt = g_Selection.begin (); selIt != g_Selection.end (); ++ selIt) {
			find_info_t & info = * selIt;
	        rgn_t & rgn = * info.iRgn;

			rgn.bTemporary = true;
		}
	}

	// Create new selection.
	g_Selection.clear ();

    find_info_t SelItem;
	for (size_t cKind = 0; cKind < rkMax; ++ cKind) {
		layer_t & layer = g_map.layers [cKind];
		rgns_t & rgns = layer.rgns;
		for (rgns_t::iterator iRgn = rgns.begin (); iRgn != rgns.end (); ++ iRgn) {
			rgn_t & rgn = * iRgn;

			if (rgn.IsDeleted ())
				continue;

			if (rgn.bTemporary == false) {
				SelItem.pLayer           = & layer;
				SelItem.iRgn             = iRgn;
				SelItem.cElementOrOrigin = -1;

				g_Selection.push_back (SelItem);
				
			}
			rgn.bTemporary = false;
		}
	}

	// Redraw map without selection.
	Redraw ();
	UpdateMenu ();
}

////////////////////////////////////////////////////

void SelectObjectsInsidePolygon (const rgn_t & _rgnSelection, bool _bAdd) {
	assert (_rgnSelection.GetKind () == rkPolygon);
	if (_rgnSelection.GetKind () != rkPolygon)
		return;
	if (_rgnSelection.GetElements () == 0)
		return;

	if (! _bAdd)
		g_Selection.clear ();

	const float c_fE = c_fE_in_pixels/g_kY;

	find_info_t SelItem;
	for (int cKind = rkMax - 1; cKind >= 0; -- cKind) {
		layer_t & layer = g_map.layers [cKind];
		rgns_t & rgns = layer.rgns;
		for (rgns_t::iterator iRgn = rgns.begin (); iRgn != rgns.end (); ++ iRgn) {
			rgn_t & rgn = * iRgn;

			if (& rgn == & _rgnSelection)
				continue;

			if (rgn.IsImmutable ())
				continue;
			if (! IsRgnKindVisible (rgn, g_map))
				continue;

			if (! _rgnSelection.rectBound.Intersects (rgn.rectBound))
				continue;

			bool bSelectRgn = false;
			if (rgn.GetKind () == rkPoint) {
				for (rgn_origins_t::const_iterator orgIt = rgn.origins.begin(); orgIt != rgn.origins.end (); ++ orgIt) {
					const rgn_origin_t & org = * orgIt;
					bSelectRgn = _rgnSelection.Contains (org.point, c_fE);
					if (bSelectRgn)
						break;
				}
			} else {
				assert (rgn.GetKind () == rkPolyline || rgn.GetKind () == rkPolygon);

				for (rgn_elements_t::const_iterator elmIt = rgn.elements.begin (); elmIt != rgn.elements.end (); ++ elmIt) {
					const rgn_element_t & elm = * elmIt;
					if (elm.points.empty ())
						continue;

					const point_t * pPrevPoint = rgn.GetKind () == rkPolygon ? & elm.points.back () : NULL;
					for (points_t::const_iterator ptIt = elm.points.begin (); ptIt != elm.points.end (); ++ ptIt) {
						const point_t & pt = * ptIt;
						bSelectRgn = _rgnSelection.Contains (pt, c_fE);
						
						if (! bSelectRgn && pPrevPoint != NULL)
							bSelectRgn = _rgnSelection.IntersectsSegment (* pPrevPoint, pt);
						if (bSelectRgn)
							break;
						
						pPrevPoint = & pt;
					}

					// Test if elm contains _rgnSelection.
					if (rgn.GetKind () == rkPolygon) {
						for (
							rgn_elements_t::const_iterator elmIt = _rgnSelection.elements.begin ();
							! bSelectRgn && elmIt != _rgnSelection.elements.end (); ++ elmIt
						) {
							if (! elmIt->bDeleted && ! elmIt->points.empty ())
								bSelectRgn = PolygonContainsPt (elm.points, elmIt->points [0]) >= 0;
						}
					}
					//bSelectRgn = bSelectRgn || PolygonContainsPt (elm.points, ptTest) >= 0;

					if (bSelectRgn)
						break;
				}
			}

			if (bSelectRgn) {
				SelItem.pLayer           = & layer;
				SelItem.iRgn             = iRgn;
				SelItem.cElementOrOrigin = -1;

				if (! IsInSelection (SelItem))
					g_Selection.push_back (SelItem);
			}
        }
    }
}

//////////////////////////////////////////

static int g_iXOffset;
static int g_iYOffset;
static float g_fTurnAngle;

void GetSelItemCentroid (const find_info_t & _SelItem, point_t & _ptCentroid) {
	const rgn_t * const pSelectedRgn = & * _SelItem.iRgn;
	assert (pSelectedRgn != NULL);

	if (_SelItem.cElementOrOrigin == -1 || pSelectedRgn->IsPoint ()) {
		if (pSelectedRgn->IsPolygon ()) {
			double M = 0, x = 0, y = 0;
			const size_t cElements = pSelectedRgn->elements.size ();
			for (size_t c = 0; c < cElements; ++ c) {
				const rgn_element_t & elem = pSelectedRgn->elements [c];
				if (elem.bDeleted)
					continue;

				point_t ptc;
				GetCentroid (elem.points, ptc);

				const float m = elem.bHole ? -elem.fLengthOrArea : elem.fLengthOrArea;
				M += m;

				x += m*ptc.x;
				y += m*ptc.y;
			}
			if (::fabs (M) > 1e-20) {
				_ptCentroid.x = x/M;
				_ptCentroid.y = y/M;
			}
			if (! pSelectedRgn->rectBound.Contains (_ptCentroid))
				_ptCentroid = point_t (pSelectedRgn->rectBound.MeanX (), pSelectedRgn->rectBound.MeanY ());	
		} else
			_ptCentroid = point_t (pSelectedRgn->rectBound.MeanX (), pSelectedRgn->rectBound.MeanY ());
	} else {
		const rgn_element_t & elem = pSelectedRgn->elements [_SelItem.cElementOrOrigin];
		if (pSelectedRgn->IsPolygon ())
			GetCentroid (elem.points, _ptCentroid);
		else
			_ptCentroid = point_t (elem.rectBound.MeanX (), elem.rectBound.MeanY ());
	}
}

static
void _DrawSelection (HDC _hDC, const rect_t & _rectVisible, const find_info_t & _SelItem) {
	//
	// Raster images.
	//

	if (g_bShowAttachments && _SelItem.pRenderer) {
		const renderer_base_t & renderer = * _SelItem.pRenderer;
		if (renderer.bDeleted || ! renderer.bVisible)
			return;
		if (! _rectVisible.Intersects (renderer.rect))
			return;

		const long lX = XToScreen (renderer.rect.x0) + g_iXOffset;
		const long lY = YToScreen (renderer.rect.y1) + g_iYOffset;
		const long lW = fast_floor (renderer.rect.Width  ()*g_kX + .5f);
		const long lH = fast_floor (renderer.rect.Height ()*g_kY + .5f);

		::SetBkMode    (_hDC, TRANSPARENT);
		::SetBkColor   (_hDC, c_colWhite);
		::SetTextColor (_hDC, c_colBlack); // NOTE: it really has undocumented influence!
		::SelectObject (_hDC, g_bTransparentPolygons ? g_hbrSelectionTransp : g_hbrSelection);
		::PatBlt       (_hDC, lX, lY, lW, lH, PATINVERT);
		
		return;
	}

	//
	// Objects.
	//

	if (! _SelItem.pLayer)
		return;
	const rgn_t * const pSelectedRgn = & * _SelItem.iRgn;
	if (! pSelectedRgn)
		return;

	CCritSec::CLock lock (g_cs);

	if (! IsRgnKindVisible (* pSelectedRgn, g_map))
		return;

	if (! _rectVisible.Intersects (pSelectedRgn->rectBound))
		return;

	const type_desc_t * const pTD = pSelectedRgn->GetTypeDesc ();
	if (! pTD)
		return;

	const bool bHuge = g_map.cMaxPointsInElement > 30000;
	// NOTE: reserve +1 additional point to store copy of first point to be able to complete polygon's perimeter.
	POINT * const pts = bHuge ? new POINT [g_map.cMaxPointsInElement + 1] : reinterpret_cast<POINT *> (alloca (sizeof (POINT)*(g_map.cMaxPointsInElement + 1)));
	CArrayHolder<POINT> pts_Holder (bHuge ? pts : NULL);

	const int iXOffset = pSelectedRgn->IsAttachment () ? 0 : g_iXOffset;
	const int iYOffset = pSelectedRgn->IsAttachment () ? 0 : g_iYOffset;

	point_t pt0;
	GetSelItemCentroid (_SelItem, pt0);
	const float fCosY  = ::cos (pt0.y*c_PI/180.f);
	const float fCosTA = ::cos (g_fTurnAngle);
	const float fSinTA = ::sin (g_fTurnAngle);

	bool bVisibleEP = false;

	const rgn_elements_t & elements = pSelectedRgn->elements;
	for (size_t cElement = 0; cElement < elements.size (); ++ cElement) {	
		const rgn_element_t & element = elements [cElement];
		if (element.bDeleted)
			continue;

		const points_t & points = element.points;
		const size_t cPoints = points.size ();

		if (_SelItem.cElementOrOrigin != -1 && _SelItem.cElementOrOrigin != cElement)
			continue;

		if (pSelectedRgn->IsMapObject () && ! LevelIsVisible (element.cLevel, pSelectedRgn->uchEndLevel))
			continue;
		bVisibleEP = true;

		if (! _rectVisible.Intersects (element.rectBound))
			continue;

		// Calculate screen coordinates.
		size_t cPoint;
		for (cPoint = 0; cPoint < cPoints; ++ cPoint) {
			const point_t & pt = points [cPoint];
			point_t ptRotated;
			if (g_fTurnAngle != 0) {
				ptRotated.x = pt0.x + ( (pt.x - pt0.x)*fCosY*fCosTA + (pt.y - pt0.y)*fSinTA)/fCosY;
				ptRotated.y = pt0.y + (-(pt.x - pt0.x)*fCosY*fSinTA + (pt.y - pt0.y)*fCosTA);
			} else
				ptRotated = pt;
			pts [cPoint].x = XToScreen (ptRotated.x) + iXOffset;
			pts [cPoint].y = YToScreen (ptRotated.y) + iYOffset;
		}
		pts [cPoint].x = pts [0].x;
		pts [cPoint].y = pts [0].y;

		if (pSelectedRgn->IsPolygon ()) {
			// NOTE: XORing gray solid brush has no effect.
			bool bGrayBrush = false;
			if (pTD->phBrush) {
				LOGBRUSH lb;
				::GetObject (* pTD->phBrush, sizeof (lb), & lb);
				bGrayBrush = lb.lbColor == c_colGray && lb.lbStyle == BS_SOLID;
			}

			// Fill polygon with selection pattern.
			::SelectObject (_hDC, g_hpenNull);
			::SelectObject (_hDC, g_bTransparentPolygons ? g_hbrSelectionTransp : g_hbrSelection);
			::SetBkMode    (_hDC, TRANSPARENT);
			::SetBkColor   (_hDC, c_colWhite);
			::SetTextColor (_hDC, c_colBlack); // NOTE: it really has undocumented influence!
			::SetROP2      (_hDC, bGrayBrush ? R2_MERGEPEN : R2_XORPEN);
			::Polygon      (_hDC, pts, cPoints);

			// Outline the perimeter.
			::SelectObject (_hDC, g_hpenBlackThin);
			::SetROP2      (_hDC, R2_COPYPEN);
			::Polyline     (_hDC, pts, cPoints + 1);
		} else {
			// Draw polyline.
			::SelectObject (_hDC, g_pSelColors->GetPen (cnSelPolylineContour, PS_SOLID, 3));
			::Polyline     (_hDC, pts, cPoints);

			::SetBkMode    (_hDC, OPAQUE);
			::SetBkColor   (_hDC, g_pSelColors->GetColor (cnSelPolylineFill2));
			::SetROP2      (_hDC, R2_COPYPEN);
			::SelectObject (_hDC, g_pSelColors->GetPen (cnSelPolylineFill1, PS_DOT));
			::Polyline     (_hDC, pts, cPoints);

			DrawNumbersSelected (_hDC, * pSelectedRgn, element, pts, cPoints);
		}
	}

	//
	// POI.
	//

	const rgn_origins_t & origins = pSelectedRgn->origins; 
	if (! origins.empty ()) {
		for (size_t cOrigin = 0; cOrigin < origins.size (); ++ cOrigin) {
			const rgn_origin_t & origin = origins [cOrigin];
			if (origin.bDeleted)
				continue;

			if (_SelItem.cElementOrOrigin != -1 && _SelItem.cElementOrOrigin != cOrigin)
				continue;
			bVisibleEP = true;

			if (pSelectedRgn->IsMapObject () && ! LevelIsVisible (origin.cLevel, pSelectedRgn->uchEndLevel))
				continue;

			if (! _rectVisible.Contains (origin.point))
				continue;

			// Calculate screen coordinates for POI and label.
			const int orig_x = XToScreen (origin.point.x) + iXOffset;
			const int orig_y = YToScreen (origin.point.y) + iYOffset;

			::SelectObject (_hDC, g_pSelColors->GetPen (cnSelPointCircle, PS_SOLID, 1));
			::SelectObject (_hDC, g_hbrNull);
			::SetBkMode    (_hDC, OPAQUE);
			::SetBkColor   (_hDC, c_colWhite);
			::SetROP2      (_hDC, R2_COPYPEN);
			::SetTextColor (_hDC, c_colBlack); // NOTE: it really has undocumented influence!
			
			::Ellipse (_hDC, orig_x - 8, orig_y - 8, orig_x + 9, orig_y + 9);

			::SetPixel (_hDC, orig_x, orig_y, c_colBlack);
		}
	}

	//
	// Entry point.
	//

	if (bVisibleEP && _SelItem.cEntryPoint != -1 && pSelectedRgn->pExtra && _SelItem.cEntryPoint < pSelectedRgn->pExtra->entry_points.size ()) {
		const entry_point_t & ep = pSelectedRgn->pExtra->entry_points [_SelItem.cEntryPoint];
		if (! ep.bDeleted && _rectVisible.Contains (ep.pt)) {
			const int x = XToScreen (ep.pt.x) + iXOffset;
			const int y = YToScreen (ep.pt.y) + iYOffset;

			::SelectObject (_hDC, g_hpenBlackThin);
			::SelectObject (_hDC, g_hbrNull);
			::SetBkMode    (_hDC, TRANSPARENT);
			::SetBkColor   (_hDC, c_colWhite);
			::SetROP2      (_hDC, R2_COPYPEN);
			::SetTextColor (_hDC, c_colBlack); // NOTE: it really has undocumented influence!
			
			::Ellipse (_hDC, x - 8, y - 8, x + 9, y + 9);
			::SelectObject (_hDC, g_hpenRedDot);
			::Ellipse (_hDC, x - 8, y - 8, x + 9, y + 9);
		}
	}
}

void DrawSelection (HDC _hDC) {
	DrawCursorsSelection (_hDC);

	if (! g_Selection.empty ()) {
		rect_t rectVisible;
		GetVisibleRect (rectVisible);

		for (size_t cSelItem = 0; cSelItem < g_Selection.size (); ++ cSelItem)
			_DrawSelection (_hDC, rectVisible, g_Selection [cSelItem]);
	}

	DrawSelPoints (_hDC);
}

void SetSelectionOffset (int _iXOffset, int _iYOffset) {
	g_iXOffset = _iXOffset;
	g_iYOffset = _iYOffset;
}

void SetSelectionTurn (float _fAngle) {
	g_fTurnAngle = _fAngle;
}

////////////////////////////////////////////////////////////////////

bool sel_point_t::IsValid () const {
	return pRgn != NULL && ! pRgn->IsImmutable () && cElement < pRgn->elements.size () && ! pRgn->elements [cElement].bDeleted && cPoint < pRgn->elements [cElement].points.size ();
}

sel_points_t g_SelPoints;

void ClearSelPoints () {
	g_SelPoints.clear ();
}

void SelectPoint (rgn_t * _pRgn, size_t _cElement, size_t _cPoint, bool _bAdd) {
	sel_point_t sp;
	sp.pRgn     = _pRgn;
	sp.cElement = _cElement;
	sp.cPoint   = _cPoint;
	sp.bSticked = false;

	// Find existing selected point.
	size_t cExisting = -1;
	const size_t cSels = g_SelPoints.size ();
	for (size_t c = 0; c < cSels; ++ c) {
		if (g_SelPoints [c] == sp) {
			cExisting = c;
			break;
		}
	}

	if (cExisting == -1) {
		if (! _bAdd)
			ClearSelPoints ();

		g_SelPoints.push_back (sp);
		std::sort (g_SelPoints.begin (), g_SelPoints.end ());
	} else {
		if (_bAdd)
			g_SelPoints.erase (g_SelPoints.begin () + cExisting);
	}
}

void UnselectPoint (int _x, int _y) {
	const size_t cIdx = FindSelPoint (_x, _y);
	if (cIdx == -1)
		return;

	g_SelPoints.erase (g_SelPoints.begin () + cIdx);
}

void SelectPointsInsideRect (const rect_t & _rectSelection, bool _bAdd) {
	if (! _bAdd)
		g_SelPoints.clear ();

	for (int cKind = rkMax - 1; cKind >= 0; -- cKind) {
		layer_t & layer = g_map.layers [cKind];
		rgns_t & rgns = layer.rgns;
		for (rgns_t::iterator iRgn = rgns.begin (); iRgn != rgns.end (); ++ iRgn) {
			rgn_t & rgn = * iRgn;
			if (rgn.IsImmutable () || rgn.IsPoint ())
				continue;
			if (! IsRgnKindVisible (rgn, g_map))
				continue;

			if (! rgn.rectBound.Intersects (_rectSelection)) 
				continue;

			// Match elements.
			for (size_t cElement = 0; cElement < rgn.elements.size (); ++ cElement) {
				const rgn_element_t & element = rgn.elements [cElement]; 
				if (element.bDeleted)
					continue;

				if (! element.rectBound.Intersects (_rectSelection))
					continue;
				if (! LevelIsVisible (element.cLevel, rgn.uchEndLevel))
					continue;

				const points_t & points = element.points;
				const size_t cPoints = points.size ();
				for (size_t c = 0; c < cPoints; ++ c) {
					if (_rectSelection.Contains (points [c]))
						SelectPoint (& rgn, cElement, c, true);
				}
			}
		}
	}
}

size_t FindSelPoint (int _x, int _y) {
	const size_t cSels = g_SelPoints.size ();
	for (size_t c = 0; c < cSels; ++ c) {
		const sel_point_t & sp = g_SelPoints [c];
		if (! sp.IsValid ())
			continue;

		point_t pt;
		sp.GetPoint (pt);

		const long lX = XToScreen (pt.x);
		const long lY = YToScreen (pt.y);
		if (::labs (lX - _x) < c_nNodeBoxSize && ::labs (lY - _y) < c_nNodeBoxSize)
			return c;
	}

	return -1;
}

static int g_nXOffsetSP;
static int g_nYOffsetSP;

void SetSelPointsOffset (int _nXOffset, int _nYOffset) {
	g_nXOffsetSP = _nXOffset;
	g_nYOffsetSP = _nYOffset;
}

# include "DrawHelpers.h"

void DrawSelPoints (HDC _hDC) {
	const size_t cSels = g_SelPoints.size ();
	if (cSels == 0)
		return;

	bool * const pbBusy = (bool *) alloca (sizeof (bool)*g_lWidth);
	::memset (pbBusy, 0, sizeof (bool)*g_lWidth);

	::SetROP2 (_hDC, R2_COPYPEN);

	for (size_t c = 0; c < cSels; ++ c) {
		const sel_point_t & sp = g_SelPoints [c];
		if (! sp.IsValid ())
			continue;

		point_t point;
		sp.GetPoint (point);
		const long lX = XToScreen (point.x) + g_nXOffsetSP;
		if (lX < 0 || lX >= g_lWidth)
			continue;
		const long lY = YToScreen (point.y) + g_nYOffsetSP;
		if (lY < 0 || lY >= g_lHeight)
			continue;

		if (sp.bSticked) {
			DrawStickPoint (_hDC, sp.nStickedX, sp.nStickedY, true);
			continue;
		}

		::SelectObject (_hDC, g_pSelColors->GetPen   (cnSelNodeContour));
		::SelectObject (_hDC, g_pSelColors->GetBrush (cnSelNodeFill));

		int nNodeBoxSize = c_nNodeBoxSize;

		rgnode_t * const pRGNode = GetRGNode (* sp.pRgn, sp.cElement, sp.cPoint);
		if (pRGNode) {
			++ nNodeBoxSize;
		} else {
			// Search for collisions.
			if (pbBusy [lX]) {
				for (size_t c2 = 0; c2 < cSels; ++ c2) {
					const sel_point_t & sp2 = g_SelPoints [c2];
					if (! sp2.IsValid ())
						continue;

					point_t point2;
					sp2.GetPoint (point2);
					const long lY2 = YToScreen (point2.y) + g_nYOffsetSP;

					if (lY2 == lY) {
						nNodeBoxSize += 2;
						break;
					}
				}
			}
			pbBusy [lX] = true;
		}

		DrawDiamond (_hDC, lX, lY, nNodeBoxSize);
	}
}

////////////////////////////////////////////////////////////

void DrawSelPointAdded (HDC _hDC, int _x, int _y) {
	::SetROP2      (_hDC, R2_COPYPEN);
	::SelectObject (_hDC, g_pSelColors->GetPen   (cnSelNodeContour));
	::SelectObject (_hDC, g_pSelColors->GetBrush (cnSelNodeAddedFill));

	DrawDiamond    (_hDC, _x, _y, c_nNodeBoxSize);
}

void DrawStickPoint (HDC _hDC, int _x, int _y, bool _bStick) {
	if (_bStick) {
		::SetROP2      (_hDC, R2_COPYPEN);
		::SelectObject (_hDC, g_pSelColors->GetPen (cnSelNodeContour));
		::SelectObject (_hDC, g_pSelColors->GetBrush (cnSelNodeStickFill));
		DrawSquare     (_hDC, _x, _y, c_nNodeBoxSize);
	} else {
		::SetROP2      (_hDC, R2_XORPEN);
		::SelectObject (_hDC, g_hpenGrayThin);
		DrawCross      (_hDC, _x, _y);
	}
}
