//
// GPSMapEdit
// (c) Konstantin Galichsky (kg@geopainting.com), 2009
//
// Set entry point of object.
//

# include "StdAfx.h"
# include "SetEntryPoint.h"
# include "UndoRedo.h"
# include "Label.h"
# include "Globals.h"

class CCreateEntryPointAction : public CEditAction {
  public:
	CCreateEntryPointAction (rgn_t * _pRgn, const point_t & _pt, const string_t & _strLabel) :
		m_pRgn (_pRgn),
		m_cOrdinal (-1),
		m_pt (_pt),
		m_strLabel (_strLabel),
		m_bWasModified ((_pRgn->dwFlags & rgn_t::fModified) != 0)
	{}

	virtual void GetName (string_t & _str) const;
	virtual cu_t CanUndo () const {return cuYes;}
	virtual bool Do ();
	virtual void Undo ();
	virtual void Redo ();

  private:
	rgn_t * const m_pRgn;
	const bool    m_bWasModified;

	size_t        m_cOrdinal;
	point_t       m_pt;
	string_t      m_strLabel;

	void _Swap ();
};

void CCreateEntryPointAction::GetName (string_t & _str) const {
	CCritSec::CLock lock (g_cs);

	assert (m_pRgn);
	_str.append ("add entry point for [");
	GetHintText (* m_pRgn, _str);
	_str.append ("]");
}

bool CCreateEntryPointAction::Do () {
	m_cOrdinal = m_pRgn->pExtra->entry_points.size ();
	m_pRgn->pExtra->AddEntryPoint (m_pt, m_strLabel);

	m_pRgn->RefreshAttributes (g_map, false);
	m_pRgn->dwFlags |= rgn_t::fModified;
	return true;
}

void CCreateEntryPointAction::Undo () {
	if (m_cOrdinal < m_pRgn->pExtra->entry_points.size ())
		m_pRgn->pExtra->entry_points [m_cOrdinal].bDeleted = true;

	m_pRgn->RefreshAttributes (g_map, false);
	if (! m_bWasModified)
		m_pRgn->dwFlags &= ~rgn_t::fModified;
}

void CCreateEntryPointAction::Redo () {
	if (m_cOrdinal < m_pRgn->pExtra->entry_points.size ())
		m_pRgn->pExtra->entry_points [m_cOrdinal].bDeleted = false;
	m_pRgn->RefreshAttributes (g_map, false);
	m_pRgn->dwFlags |= rgn_t::fModified;
}

//////////////////////////////////////////////////////

class CSetEntryPointAction : public CEditAction {
  public:
	CSetEntryPointAction (rgn_t * _pRgn, size_t _cOrdinal, const point_t & _pt, const string_t & _strLabel, bool _bSet) :
		m_pRgn (_pRgn),
		m_cOrdinal (_cOrdinal),
		m_bSet (_bSet),
		m_bDeleted (! _bSet),
		m_pt (_pt),
		m_strLabel (_strLabel),
		m_bWasModified ((_pRgn->dwFlags & rgn_t::fModified) != 0)
	{}

	virtual void GetName (string_t & _str) const;
	virtual cu_t CanUndo () const {return cuYes;}
	virtual bool Do ();
	virtual void Undo ();
	virtual void Redo ();

  private:
	rgn_t * const m_pRgn;
	const size_t  m_cOrdinal;
	const bool    m_bSet;
	const bool    m_bWasModified;

	bool          m_bDeleted;
	point_t       m_pt;
	string_t      m_strLabel;

	void _Swap ();
};

void CSetEntryPointAction::GetName (string_t & _str) const {
	CCritSec::CLock lock (g_cs);

	assert (m_pRgn);
	_str.append (m_bSet ? "change entry point for [" : "remove entry point for [");
	GetHintText (* m_pRgn, _str);
	_str.append ("]");
}

void CSetEntryPointAction::_Swap () {
	entry_points_t & eps = m_pRgn->pExtra->entry_points;

	assert (m_cOrdinal < eps.size ());
	if (m_cOrdinal < eps.size ()) {
		entry_point_t & ep = m_pRgn->pExtra->entry_points [m_cOrdinal];

		std::swap (m_pt,       ep.pt);
		std::swap (m_strLabel, ep.strLabel);
		std::swap (m_bDeleted, ep.bDeleted);

		if (! (m_pt == ep.pt) || m_bDeleted != ep.bDeleted)
			m_pRgn->RefreshAttributes (g_map, false);
	}
}

bool CSetEntryPointAction::Do () {
	_Swap ();
	m_pRgn->dwFlags |= rgn_t::fModified;
	return true;
}

void CSetEntryPointAction::Undo () {
	_Swap ();
	if (! m_bWasModified)
		m_pRgn->dwFlags &= ~rgn_t::fModified;
}

void CSetEntryPointAction::Redo () {
	CSetEntryPointAction::Do ();
}

///////////////////////////////////////////

void CreateEntryPoint (rgn_t * _pRgn, const point_t & _pt, const string_t & _strLabel) {
	PerformEditAction (new CCreateEntryPointAction (_pRgn, _pt, _strLabel));
}

void SetEntryPoint (rgn_t * _pRgn, size_t _cOrdinal, const point_t & _pt) {
	const entry_points_t & eps = _pRgn->pExtra->entry_points;
	assert (_cOrdinal < eps.size ());
	if (_cOrdinal < eps.size ()) {
		const entry_point_t & ep = _pRgn->pExtra->entry_points [_cOrdinal];
		PerformEditAction (new CSetEntryPointAction (_pRgn, _cOrdinal, _pt, ep.strLabel, true));
	}
}

void SetEntryPointLabel (rgn_t * _pRgn, size_t _cOrdinal, const string_t & _strLabel) {
	const entry_points_t & eps = _pRgn->pExtra->entry_points;
	assert (_cOrdinal < eps.size ());
	if (_cOrdinal < eps.size ()) {
		const entry_point_t & ep = _pRgn->pExtra->entry_points [_cOrdinal];
		PerformEditAction (new CSetEntryPointAction (_pRgn, _cOrdinal, ep.pt, _strLabel, true));
	}
}

void RemoveEntryPoint (rgn_t * _pRgn, size_t _cOrdinal) {
	PerformEditAction (new CSetEntryPointAction (_pRgn, _cOrdinal, point_t (0, 0), NULL, false));
}
