//
// GPSMapEdit
// (c) Konstantin Galichsky (kg@geopainting.com), 2008-2009
//
// Set 3D-level info.
//

# include "StdAfx.h"
# include "SetHLevel.h"
# include "UndoRedo.h"
# include "CombinedAction.h"
# include "Label.h"
# include "NodesExtra.h"

class CSetHLevelAction : public CEditAction {
  public:
	CSetHLevelAction (rgn_t * _pRgn, size_t _cElement, size_t _cPoint, int _nHLevel, bool _bHasHLevel) :
		m_pRgn (_pRgn),
		m_cElement (_cElement),
		m_cPoint (_cPoint),
		m_bHasHLevelOld (false),
		m_bHasHLevelNew (_bHasHLevel),
		m_nHLevelNew (_nHLevel),
		m_bWasModified ((_pRgn->dwFlags & rgn_t::fModified) != 0)
	{
		const rgn_element_t & element = _pRgn->elements [_cElement];
		if (element.pExtra && _cPoint < element.pExtra->PointExtras.size ()) {
			const point_extra_t & pe = element.pExtra->PointExtras [_cPoint];

			m_bHasHLevelOld = pe.HasHLevel ();
			m_nHLevelOld    = pe.GetHLevel ();
		}
	}

	virtual void GetName (string_t & _str) const;
	virtual cu_t CanUndo () const {return cuYes;}
	virtual bool Do ();
	virtual void Undo ();
	virtual void Redo ();

  private:
	rgn_t * const m_pRgn;
	size_t  const m_cElement;
	size_t  const m_cPoint;
	bool          m_bHasHLevelOld;
	int           m_nHLevelOld;
	bool    const m_bHasHLevelNew;
	int     const m_nHLevelNew;
	bool    const m_bWasModified;
};

void CSetHLevelAction::GetName (string_t & _str) const {
	_str.append ("set 3D-level [");
	GetHintText (* m_pRgn, _str);
	_str.append ("]");
}

bool CSetHLevelAction::Do () {
	CCritSec::CLock lock (g_cs);

	assert (m_pRgn);
	assert (m_cElement < m_pRgn->elements.size ());
	rgn_element_t & element = m_pRgn->elements [m_cElement];

	assert (m_cPoint < element.points.size ());
	element.pExtra->PointExtras.resize (__max (m_cPoint + 1, element.points.size ()));
	element.pExtra->PointExtras [m_cPoint].SetHLevel (m_nHLevelNew, m_bHasHLevelNew);

	m_pRgn->dwFlags |= rgn_t::fModified;

	return true;
}

void CSetHLevelAction::Undo () {
	CCritSec::CLock lock (g_cs);

	assert (m_pRgn);
	assert (m_cElement < m_pRgn->elements.size ());
	rgn_element_t & element = m_pRgn->elements [m_cElement];

	assert (m_cPoint < element.points.size ());
	element.pExtra->PointExtras [m_cPoint].SetHLevel (m_nHLevelOld, m_bHasHLevelOld);

//	m_pRgn->RefreshAttributes (g_map);
	if (! m_bWasModified)
		m_pRgn->dwFlags &= ~rgn_t::fModified;
}

void CSetHLevelAction::Redo () {
	CSetHLevelAction::Do ();
}

//////////////////////////////////////

static
void _SetHLevel (rgn_t * _pRgn, size_t _cElement, size_t _cPoint, int _nHLevel, bool _bHasHLevel) {
	assert (_pRgn);
	assert (_cElement < _pRgn->elements.size ());
	assert (_cPoint < _pRgn->elements [_cElement].points.size ());

	PerformEditAction (new CSetHLevelAction (_pRgn, _cElement, _cPoint, _nHLevel, _bHasHLevel));
}

void SetHLevel (rgn_t * _pRgn, size_t _cElement, size_t _cPoint, int _nHLevel, bool _bHasHLevel) {
	assert (_pRgn);
	assert (_cElement < _pRgn->elements.size ());
	rgn_element_t & element = _pRgn->elements [_cElement];

	assert (_cPoint < element.points.size ());
	if (element.pExtra && _cPoint < element.pExtra->PointExtras.size ()) {
		rgnode_t * const pRGNode = element.pExtra->PointExtras [_cPoint].GetRGNode ();
		if (pRGNode) {
			const int nDepth = BeginCombinedAction ();

			rgns_t & polylines = g_map.layers [rkPolyline].rgns;
			for (rgns_t::iterator iRgn = polylines.begin (); iRgn != polylines.end (); ++ iRgn) {
				rgn_t & rgn = * iRgn;
				if (rgn.IsImmutable () || rgn.IsNotMapObject ())
					continue;

				const size_t cElements = rgn.elements.size ();
				for (size_t cElement = 0; cElement < cElements; ++ cElement) {
					rgn_element_t & element1 = rgn.elements [cElement];
					if (element1.bDeleted || element1.pExtra == NULL)
						continue;

					point_extras_t & PointExtras = element1.pExtra->PointExtras;
					const size_t cPointExtras = PointExtras.size ();
					for (size_t c = 0; c < cPointExtras; ++ c) {
						if (PointExtras [c].GetRGNode () != pRGNode)
							continue;

						_SetHLevel (& rgn, cElement, c, _nHLevel, _bHasHLevel);
					}
				}
			}

			EndCombinedAction (nDepth, "set 3D-level");

			return;
		}
	}

	_SetHLevel (_pRgn, _cElement, _cPoint, _nHLevel, _bHasHLevel);
}