//
// GPSMapEdit
// (c) Konstantin Galichsky (kg@geopainting.com), 2002-2009
//
// ESRI shapefile import wizard.
//

# include "StdAfx.h"
# include "Map.h"
# include "Types.h"
# include "Load.h"
# include "Log.h"
# include "Status.h"
# include "ShpFile.h"
# include "DbfFile.h"
# include "Transliteration.h"
# include "Transform.h"
# include "ShpWizard.h"
# include "ShpExtraLoader.h"
# include "fast_floor.h"
# include "ParseHelpers.h"

class CShpLoader : public CMapLoader {
  public:
	// Override.
	virtual bool IsValid (const char * _pData, size_t _cFileSize) const;
	virtual LoadStatus_t Load (const char * _pData, size_t _cFileSize, const char * _strFileName, map_t * _pMap);

  private:
	map_t * m_pMap;
	const char * m_pStart, * m_pEnd;

	CShpWizard * m_pWizard;
	CShpExtraLoader m_cl;

	file_refs_t m_file_refs;

	// Object counter.
	size_t m_cOrdinal;
	rgn_t & CreateRgn (const CShpRecordHdr * _pStart, const char * _strEnd, RGN_KIND _Kind, DWORD _dwType, const char * _strLabel);

	// Index of level to fill with imported data.
	size_t m_cFirstLevel;
	// Indecis of rest levels.
	std::vector<size_t> m_LevelsToCopy;
	void CopyOrigins  (rgn_t & _rgn);
	void CopyElements (rgn_t & _rgn, size_t _cElementsPerLevel);

	rgn_t * ParsePoint      (const CShpPoint      * _pPoint,      const char * _pEnd, RGN_KIND _Kind, DWORD _dwType, const char * _strLabel);
	rgn_t * ParseMultiPoint (const CShpMultiPoint * _pMultiPoint, const char * _pEnd, RGN_KIND _Kind, DWORD _dwType, const char * _strLabel);
	rgn_t * ParsePolyline   (const CShpPolyline   * _pPolyline,   const char * _pEnd,                 DWORD _dwType, const char * _strLabel);
	rgn_t * ParsePolygon    (const CShpPolygon    * _pPolygon,    const char * _pEnd,                 DWORD _dwType, const char * _strLabel);
	rgn_t * ParseMultiPatch (const CShpMultiPatch * _pMultiPatch, const char * _pEnd,                 DWORD _dwType, const char * _strLabel);
};

DECLARE_MAP_IMPORT (CShpLoader, "esri-shapefile", "shp", "ESRI shape (*.shp)");

bool CShpLoader::IsValid (const char * _pData, size_t _cFileSize) const {
	assert (sizeof (CShpHdr) == 100);
	if (_cFileSize < sizeof (CShpHdr))
		return false;
	return reinterpret_cast<const CShpHdr *> (_pData)->dwSignature == c_dwShpSignature;
}

rgn_t & CShpLoader::CreateRgn (const CShpRecordHdr * _pStart, const char * _pEnd, RGN_KIND _Kind, DWORD _dwType, const char * _strLabel) {
	rgn_t & rgn = m_pMap->CreateRgn (_Kind);

	rgn.pFileRef  = & m_file_refs.back ();
	rgn.cBeginPos = reinterpret_cast<const char *> (_pStart) - m_pStart;
	rgn.cEndPos   = _pEnd - m_pStart;
	rgn.cFileLine = -1;
//	rgn.cOrdinal  = m_cOrdinal ++;

	rgn.strLabel  = _strLabel;

	rgn.SetType (m_pMap->pTypeSet, _dwType, _Kind);

	return rgn;
}

void CShpLoader::CopyOrigins (rgn_t & _rgn) {
	for (size_t cOrigin = 1; cOrigin < m_LevelsToCopy.size () + 1; ++ cOrigin) {
		assert (cOrigin < _rgn.origins.size ());
		_rgn.origins [cOrigin] = _rgn.origins [0];
		_rgn.origins [cOrigin].cLevel = m_LevelsToCopy [cOrigin - 1];
	}
}

void CShpLoader::CopyElements (rgn_t & _rgn, size_t _cElementsPerLevel) {
	const size_t cLevelsToCopy = m_LevelsToCopy.size ();
	for (size_t cLevelIndex = 0; cLevelIndex < cLevelsToCopy; ++ cLevelIndex) {
		for (size_t cElement0 = 0; cElement0 < _cElementsPerLevel; ++ cElement0) {
			assert (cElement0 + (cLevelIndex + 1)*_cElementsPerLevel < _rgn.elements.size ());
			rgn_element_t & element = _rgn.elements [cElement0 + (cLevelIndex + 1)*_cElementsPerLevel];

			assert (cElement0 < _rgn.elements.size ());
			element = _rgn.elements [cElement0];

			assert (cLevelIndex < m_LevelsToCopy.size ());
			element.cLevel = m_LevelsToCopy [cLevelIndex];
		}
	}
}

rgn_t * CShpLoader::ParsePoint (const CShpPoint * _pPoint, const char * _pEnd, RGN_KIND _Kind, DWORD _dwType, const char * _strLabel) {
	rgn_t & rgn = CreateRgn (_pPoint, _pEnd, _Kind, _dwType, _strLabel);

	assert (m_pWizard->GetLevelsToImport () > 0);
	rgn.origins.resize (m_pWizard->GetLevelsToImport ());
	rgn_origin_t & origin = rgn.origins [0];

	origin.cLevel      = m_cFirstLevel;
	origin.cFileOffset = rgn.cBeginPos;

	origin.point.x     = _pPoint->point.dX;
	origin.point.y     = _pPoint->point.dY;

	CopyOrigins (rgn);

	return & rgn;
}

rgn_t * CShpLoader::ParseMultiPoint (const CShpMultiPoint * _pMultiPoint, const char * _pEnd, RGN_KIND _Kind, DWORD _dwType, const char * _strLabel) {
	// TODO: test it!
	for (size_t cPoint = 0; cPoint < _pMultiPoint->dwPoints; ++ cPoint) {
		rgn_t & rgn = CreateRgn (_pMultiPoint, _pEnd, _Kind, _dwType, _strLabel);

		assert (m_pWizard->GetLevelsToImport () > 0);
		rgn.origins.resize (m_pWizard->GetLevelsToImport ());
		rgn_origin_t & origin = rgn.origins [0];

		origin.cLevel      = m_cFirstLevel;
		origin.cFileOffset = rgn.cBeginPos;

		origin.point.x     = _pMultiPoint->points [cPoint].dX;
		origin.point.y     = _pMultiPoint->points [cPoint].dY;

		CopyOrigins (rgn);

		if (cPoint + 1 == _pMultiPoint->dwPoints)
			return & rgn;
	}

	return NULL;
}

rgn_t * CShpLoader::ParsePolyline (const CShpPolyline * _pPolyline, const char * _pEnd, DWORD _dwType, const char * _strLabel) {
	rgn_t & rgn = CreateRgn (_pPolyline, _pEnd, rkPolyline, _dwType, _strLabel);

	size_t cPointIndex = 0;
	const CShpPointData * const pPoints = reinterpret_cast<const CShpPointData *> (_pPolyline->dwaParts + _pPolyline->dwNumParts);

	rgn_elements_t & elements = rgn.elements;
	const size_t cElements = _pPolyline->dwNumParts;
	assert (m_pWizard->GetLevelsToImport () > 0);
	elements.resize (cElements*m_pWizard->GetLevelsToImport ());
	for (size_t cElement = 0; cElement < cElements; ++ cElement) {
		rgn_element_t & element = elements [cElement];

		element.cLevel = m_cFirstLevel;
		element.cFileOffset = rgn.cBeginPos;

		const size_t cPoints = cElement < cElements - 1 ?
			_pPolyline->dwaParts [cElement + 1] - _pPolyline->dwaParts [cElement] :
			_pPolyline->dwNumPoints             - _pPolyline->dwaParts [cElement];
		points_t & points = element.points;
		points.resize (cPoints);
		for (size_t cPoint = 0; cPoint < cPoints; ++ cPoint, ++ cPointIndex) {
			point_t & point = points [cPoint];
			point.x = static_cast<float> (pPoints [cPointIndex].dX);
			point.y = static_cast<float> (pPoints [cPointIndex].dY);
		}
	}
	assert (cPointIndex == _pPolyline->dwNumPoints);

	CopyElements (rgn, cElements);

	return & rgn;
}

rgn_t * CShpLoader::ParsePolygon (const CShpPolygon * const _pPolygon, const char * _pEnd, DWORD _dwType, const char * _strLabel) {
	rgn_t & rgn = CreateRgn (_pPolygon, _pEnd, rkPolygon, _dwType, _strLabel);

	size_t cPointIndex = 0;
	const CShpPointData * const pPoints = reinterpret_cast<const CShpPointData *> (_pPolygon->dwaParts + _pPolygon->dwNumParts);

	rgn_elements_t & elements = rgn.elements;
	const size_t cElements = _pPolygon->dwNumParts;
	assert (m_pWizard->GetLevelsToImport () > 0);
	elements.resize (cElements*m_pWizard->GetLevelsToImport ());
	for (size_t cElement = 0; cElement < cElements; ++ cElement) {
		rgn_element_t & element = elements [cElement];

		element.cLevel = m_cFirstLevel;
		element.cFileOffset = rgn.cBeginPos;

		size_t cPoints = cElement < cElements - 1 ?
			_pPolygon->dwaParts [cElement + 1] - _pPolygon->dwaParts [cElement]:
			_pPolygon->dwNumPoints             - _pPolygon->dwaParts [cElement];
		if (cPoints == 0)
			continue;
		// NOTE: last point is always equal to first. Omit it.
		const bool bOmitLastPoint =
			pPoints [cPoints - 1].dX == pPoints [0].dX &&
			pPoints [cPoints - 1].dY == pPoints [0].dY;
		if (bOmitLastPoint)
			-- cPoints;

		points_t & points = element.points;
		points.resize (cPoints);
		for (size_t cPoint = 0; cPoint < cPoints; ++ cPoint, ++ cPointIndex) {
			point_t & point = points [cPoint];
			point.x = static_cast<float> (pPoints [cPointIndex].dX);
			point.y = static_cast<float> (pPoints [cPointIndex].dY);
		}
		if (bOmitLastPoint)
			++ cPointIndex;

		// Detect holes.
		element.bHole = cElements > 1 && ! IsRightOriented (points);
	}
	assert (cPointIndex == _pPolygon->dwNumPoints);

	CopyElements (rgn, cElements);

	rgn.ReorderHoles ();

	return & rgn;
}

rgn_t * CShpLoader::ParseMultiPatch (const CShpMultiPatch * const _pMultiPatch, const char * _pEnd, DWORD _dwType, const char * _strLabel) {
	rgn_t & rgn = CreateRgn (_pMultiPatch, _pEnd, rkPolygon, _dwType, _strLabel);

	const DWORD * const pPartTypes = reinterpret_cast<const DWORD *> (_pMultiPatch->dwaParts + _pMultiPatch->dwNumParts);

	size_t cPointIndex = 0;
	const CShpPointData * const pPoints = reinterpret_cast<const CShpPointData *> (_pMultiPatch->dwaParts + 2*_pMultiPatch->dwNumParts);

	rgn_elements_t & elements = rgn.elements;
	const size_t cElements = _pMultiPatch->dwNumParts;
	assert (m_pWizard->GetLevelsToImport () > 0);
	elements.resize (cElements*m_pWizard->GetLevelsToImport ());
	for (size_t cElement = 0; cElement < cElements; ++ cElement) {
		rgn_element_t & element = elements [cElement];

		element.cLevel = m_cFirstLevel;
		element.cFileOffset = rgn.cBeginPos;

		size_t cPoints = cElement < cElements - 1 ?
			_pMultiPatch->dwaParts [cElement + 1] - _pMultiPatch->dwaParts [cElement]:
			_pMultiPatch->dwNumPoints             - _pMultiPatch->dwaParts [cElement];
		if (cPoints == 0)
			continue;
		// NOTE: last point is always equal to first. Omit it.
		const bool bOmitLastPoint =
			pPoints [cPoints - 1].dX == pPoints [0].dX &&
			pPoints [cPoints - 1].dY == pPoints [0].dY;
		if (bOmitLastPoint)
			-- cPoints;

		points_t & points = element.points;
		points.resize (cPoints);
		for (size_t cPoint = 0; cPoint < cPoints; ++ cPoint, ++ cPointIndex) {
			point_t & point = points [cPoint];
			point.x = static_cast<float> (pPoints [cPointIndex].dX);
			point.y = static_cast<float> (pPoints [cPointIndex].dY);
		}
		if (bOmitLastPoint)
			++ cPointIndex;

		// Detect holes.
		if (cElements == 1)
			element.bHole = false;
		else {
			switch (pPartTypes [cElement]) {
				case shpMpTriangleStrip:
					element.bHole = false;

					// NOTE: triangle strip seems like a self-intersecting polygon.
					ReportWarning (
						reinterpret_cast<const char *> (_pMultiPatch) - m_pStart,
						"Triangle strip in multi-patch is not supported. It will be ignored."
					);
					element.bDeleted = true;

					break;

				case shpMpTriangleFan:
					element.bHole = false;
					break;

				case shpMpOuterRing:
					element.bHole = false;
					break;
				case shpMpInnerRing:
					element.bHole = true;
					break;

				case shpMpFirstRing:
					element.bHole = ! IsRightOriented (points);
					break;

				case shpMpRing:          
					if (pPartTypes [0] == shpMpFirstRing)
						element.bHole = ! IsRightOriented (points);
					else
						element.bHole = false;
					break;

				default:
					element.bHole = false;

					ReportWarning (
						reinterpret_cast<const char *> (_pMultiPatch) - m_pStart,
						"Unknown type of multi-patch part: %d. It will be ignored.",
						pPartTypes [cElement]
					);
					element.bDeleted = true;

					break;
			}
		}
	}
	assert (cPointIndex == _pMultiPatch->dwNumPoints);

	CopyElements (rgn, cElements);

	rgn.ReorderHoles ();

	return & rgn;
}

LoadStatus_t CShpLoader::Load (const char * _pData, size_t _cFileSize, const char * _strMapFileName, map_t * _pMap) {
	m_pMap = _pMap;
	m_pStart = _pData;
	m_pEnd = _pData + _cFileSize;

	m_pWizard = NULL;

	m_cFirstLevel = 0;
	m_LevelsToCopy.clear ();

	m_cOrdinal = 0;

	m_file_refs.clear ();
	m_file_refs.push_back (file_ref_t ());
	m_file_refs.back ().strFileName = _strMapFileName;
	m_file_refs.back ().bAttachment = false;

	SetProgressStatus (0);

	//
	// Prepare access to the associated .dbf file.
	//

	char strDbfName [MAX_PATH + 1];
	::strncpy (strDbfName, _strMapFileName, MAX_PATH);
	::PathRenameExtension (strDbfName, ".dbf");

	CDbfFile dbf (strDbfName);
	if (dbf.CantOpen ()) {
		ReportWarning ("Can't open or read DBF file: '%s'.", strDbfName);
		return lsFailed;
	}
	const size_t cRecords = dbf.GetRecords ();

	//
	// Parse .shp header.
	//

	const CShpHdr * const pHdr = reinterpret_cast<const CShpHdr *> (_pData);
	m_pMap->cHeaderLen = sizeof (CShpHdr);

	//
	// Show wizard.
	//

	RGN_KIND Kind = rkPoint;
	switch (pHdr->dwType) {
		default:
		case shpNull:
			ReportWarning ("Null type shapefiles are not supported.");
			return lsFailed;

		case shpPoint:
		case shpPointZ:
		case shpPointM:
		case shpMultiPoint:
		case shpMultiPointZ:
		case shpMultiPointM:
			Kind = rkPoint;
			break;

		case shpPolyline:
		case shpPolylineZ:
		case shpPolylineM:
			Kind = rkPolyline;
			break;

		case shpPolygon:
		case shpPolygonZ:
		case shpPolygonM:
		case shpMultiPatch:
			Kind = rkPolygon;
			break;
	}
/*
	switch (pHdr->dwType) {
		default:             strShpKind = "Unknown";     break;
		case shpNull:        strShpKind = "Null";        break;
		case shpPoint:       strShpKind = "Point";       break;
		case shpPointZ:      strShpKind = "PointZ";      break;
		case shpPointM:      strShpKind = "PointM";      break;
		case shpMultiPoint:  strShpKind = "MultiPoint";  break;
		case shpMultiPointZ: strShpKind = "MultiPointZ"; break;
		case shpMultiPointM: strShpKind = "MultiPointM"; break;
		case shpPolyline:    strShpKind = "Polyline";    break;
		case shpPolylineZ:   strShpKind = "PolylineZ";   break;
		case shpPolylineM:   strShpKind = "PolylineM";   break;
		case shpPolygon:     strShpKind = "Polygon";     break;
		case shpPolygonZ:    strShpKind = "PolygonZ";    break;
		case shpPolygonM:    strShpKind = "PolygonM";    break;
		case shpMultiPatch:  strShpKind = "MultiPatch";  break;
	}
*/
	CShpWizard wizard (_strMapFileName, dbf, Kind, false, rect_t (pHdr->box.dX0, pHdr->box.dY0, pHdr->box.dX1, pHdr->box.dY1));
	const LoadStatus_t ls = wizard.Show ();
	if (ls != lsOK)
		return ls;

	m_pWizard = & wizard;

	// Prepare access to kind & type.
	m_pMap->pTypeSet = wizard.GetTypeSet ();

	const bool bFromTypeField = wizard.GetFromTypeField ();
	const size_t cTypeFieldIndex = wizard.GetTypeField ();
	const size_t cTypeFieldLen = cTypeFieldIndex != -1 ? dbf.GetFieldLength (cTypeFieldIndex) : 0;
	const DWORD dwTypeFixed = wizard.GetType ();

	// Prepare access to field with labels.
	const size_t cLabelFieldIndex = wizard.GetLabelField ();
	const size_t cLabelFieldLen = cLabelFieldIndex != -1 ? dbf.GetFieldLength (cLabelFieldIndex) : 0;
	char * const strLabelBuf = reinterpret_cast<char *> (::alloca ((cLabelFieldLen + 1)*sizeof (char)));

	// Prepare access to extra fields.
	m_cl.Bind (& wizard, & dbf, m_pMap);

	const CTransliterator * const pTransliterator = wizard.GetTransliterator ();

	// Prepare access to levels.
	if (m_pWizard->GetLevelsToImport () == 0) {
		ReportWarning ("No levels to import selected.");
		return lsFailed;
	}

	for (size_t cLevel = 0; cLevel < wizard.GetLevels ().size (); ++ cLevel)
		if (wizard.GetLevels () [cLevel]) {
			m_cFirstLevel = cLevel;
			for (++ cLevel; cLevel < wizard.GetLevels ().size (); ++ cLevel) {
				if (wizard.GetLevels () [cLevel])
					m_LevelsToCopy.push_back (cLevel);
			}
			break;
		}
	assert (m_LevelsToCopy.size () + 1 == wizard.GetLevelsToImport ());

	map_levels_t Levels;
	{
		CCritSec::CLock lock (g_cs);
		Levels = g_map.Levels;
	}

	if (Levels.size () == 0) {
		const size_t cBits = m_pMap->pTypeSet->cMaxBits;
		m_pMap->Levels.resize (2);
		m_pMap->Levels [0].SetBasicScale (cBits);
		m_pMap->Levels [0].cZoom = 0;
		m_pMap->Levels [1].SetBasicScale (cBits - 1);
		m_pMap->Levels [1].cZoom = 1;
	} else {
		m_pMap->Levels.swap (Levels);
	}
	m_pMap->bNoHeader = false;

	//
	// Parse shapes.
	//

	const char * pRecord = reinterpret_cast<const char *> (pHdr + 1);
	size_t cRecordCounter = 0;
	do {
		if (g_bStop)
			return lsInterrupted;

		DWORD dwType;
		string_t strLabel;
		CShpExtraLoader::extra_t extra;
		if (cRecordCounter < cRecords) {
			// Omit deleted records in .dbf.
			while (cRecordCounter < cRecords && ! dbf.RecordExists (cRecordCounter))
				++ cRecordCounter;

			//
			// Get type of object from .dbf.
			//
			if (bFromTypeField && cTypeFieldIndex != -1) {
				const char * const pType = dbf.GetRawField (cRecordCounter, cTypeFieldIndex);

				unsigned long ulType = 0;
				const char * pEnd = NULL;
				if (pType [0] == '0' && pType [1] == 'x')
					ParseHex (pType + 2, pType + cTypeFieldLen, & pEnd, ulType);
				else
					ParseDecU (pType, pType + cTypeFieldLen, & pEnd, ulType);

				dwType = ulType;
			} else
				dwType = dwTypeFixed;

			//
			// Get label of object from .dbf.
			//
			if (cLabelFieldIndex != -1) {
				const size_t cLen = dbf.GetFieldS (cRecordCounter, cLabelFieldIndex, strLabelBuf);
				
				if (wizard.GetOEM ())
					::OemToCharBuff (strLabelBuf, strLabelBuf, cLen);

				if (pTransliterator)
					pTransliterator->Do (strLabelBuf, strLabelBuf + cLen, strLabel);
				else
					strLabel.assign (strLabelBuf, cLen);
			}

			//
			// Get extraattributes from .dbf.
			//
			m_cl.ReadExtra (cRecordCounter, extra);

			++ cRecordCounter;
		} else {
			ReportWarning ("Not enough records in DBF file.");

			if (bFromTypeField)
				dwType = 0;
			else
				dwType = dwTypeFixed;
		}

		// Get next record in .shp.
		const CShpRecordHdr * const pRecHdr = reinterpret_cast<const CShpRecordHdr *> (pRecord);
		const char * const pNextRecord = pRecord + 8 + Big2LowEndian (pRecHdr->dwLen)*sizeof (WORD);

		if (pRecHdr->dwType != shpNull && pRecHdr->dwType != pHdr->dwType) {
			ReportWarning (pRecord - m_pStart, "Type of shape is not equal to type of whole shapefile. It will be ignored");
		} else {
			rgn_t * pRgn = NULL;
			switch (pRecHdr->dwType) {
				case shpNull: break;

				case shpPoint:
				case shpPointM:
				case shpPointZ:      pRgn = ParsePoint      (static_cast<const CShpPoint *>      (pRecHdr), pNextRecord, Kind, dwType, strLabel.c_str ()); break;

				case shpMultiPoint:
				case shpMultiPointM:
				case shpMultiPointZ: pRgn = ParseMultiPoint (static_cast<const CShpMultiPoint *> (pRecHdr), pNextRecord, Kind, dwType, strLabel.c_str ()); break;

				case shpPolyline:
				case shpPolylineM:
				case shpPolylineZ:   pRgn = ParsePolyline   (static_cast<const CShpPolyline *>   (pRecHdr), pNextRecord, dwType, strLabel.c_str ()); break;

				case shpPolygon:
				case shpPolygonM:
				case shpPolygonZ:    pRgn = ParsePolygon    (static_cast<const CShpPolygon *>    (pRecHdr), pNextRecord, dwType, strLabel.c_str ()); break;

				case shpMultiPatch:	 pRgn = ParseMultiPatch (static_cast<const CShpMultiPatch *> (pRecHdr), pNextRecord, dwType, strLabel.c_str ()); break;

				default:
					ReportWarning (reinterpret_cast<const char *> (pRecHdr) - m_pStart, "Unknown shape type: %d.", pRecHdr->dwType);
					return lsFailed;
			}

			if (pRgn)
				m_cl.SetRgnExtraAttributes (* pRgn, extra);
		}

		pRecord = pNextRecord;

		SetProgressStatus (fast_floor (50.f*(pRecord - _pData)/_cFileSize + .5f));
	} while (pRecord < m_pEnd);

	// Convert units to degrees.
	if (wizard.GetUnitTransformation ()) {
		m_pMap->TransformPoints (* wizard.GetUnitTransformation ());
		if (g_bStop)
			return lsInterrupted;
	}

	// Change datum.
	if (wizard.GetDatumTransformation ()) {
		m_pMap->TransformPoints (* wizard.GetDatumTransformation ());
		if (g_bStop)
			return lsInterrupted;
	}

	m_pMap->file_refs.swap (m_file_refs);

	SetProgressStatus (50);

	return lsOK;
}
