//
// GPSMapEdit
// (c) Konstantin Galichsky (kg@geopainting.com), 2002-2009
//
// ESRI shapefile import wizard.
//

# include "StdAfx.h"
# include "Resource.h"
# include "ShpWizard.h"
# include "Options.h"
# include "Datum.h"
# include "ShpFile.h"
# include "DatabaseTable.h"
# include "Transliteration.h"
# include "TransverseMercator.h"
# include "Mercator.h"
# include "S42.h"
# include "NewZealandMapGrid.h"
# include "Types.h"
# include "TypeLocalization.h"
# include "Map.h"
# include "MapSkins.h"
# include "ParseHelpers.h"
# include "EditHelpers.h"
# include "Settings.h"
# include "ListView.h"
# include "PropertyPages.h"
# include "Static.h"
# include "fast_floor.h"
# include "Globals.h"

CShpWizard::CShpWizard (
	const char * _strFileName, const CDatabaseTable & _dbt,
	RGN_KIND _Kind, bool _bEnableSelectKind,
	const rect_t & _box
) :
	m_strFileName (_strFileName),
	m_pDbt (& _dbt),
	m_pTypeSet (NULL),
	m_Kind (_Kind),
	m_bEnableSelectKind (_bEnableSelectKind),
	m_box (_box),
	m_bFromTypeField (false),
	m_dwType (0),
	m_cTypeField (-1),
	m_cLabelField (-1),
	m_bOEM (false),
	m_pTransliterator (NULL),
	m_bInitLatLon (false),
	m_bInitTransverseMercator (false),
	m_fCentralMeridian (0),
	m_fFE (0),
	m_fFN (0),
	m_bEnableCoordinates (true),
	m_pDatumTransformation (NULL),
	m_cLevelsToImport (0),
	m_cTypeIndex (-1),
	m_hType (NULL),
	m_hTypeField (NULL),
	m_hTypeFieldStatic (NULL)
{
	for (size_t c = 0; c < iaLast; ++ c)
		m_cFields [c] = -1;
}

static int g_iSortColumn = -1;
static bool g_bAccent = true;

static
INT_PTR CALLBACK _CompareItems (LPARAM _lParam1, LPARAM _lParam2, LPARAM _lData) {
	const HWND hTypes = reinterpret_cast<HWND> (_lData);

	char str1 [64], str2 [64];
	ListView_GetItemText (hTypes, _lParam1, g_iSortColumn, str1, 64); 
	ListView_GetItemText (hTypes, _lParam2, g_iSortColumn, str2, 64);

	if (g_iSortColumn == 4 || g_iSortColumn == 5) {
		// Compare as integers.
		return g_bAccent ? ParseDecU (str1, str1 + 64) - ParseDecU (str2, str2 + 64) : ParseDecU (str2, str2 + 64) - ParseDecU (str1, str1 + 64);
	} else
		// Compare as strings.
		return g_bAccent ? ::strcmp (str1, str2) : ::strcmp (str2, str1);
}

static
void _FillTypes (HWND _hListView, RGN_KIND _Kind, const CTypeSet * _pTypeSet) {
	assert (_pTypeSet);

	// Update icons.
	assert (_Kind < rkMax);
	ListView_SetImageList (_hListView, _pTypeSet->GetTypeIcons (), LVSIL_SMALL);

	ListView_DeleteAllItems (_hListView);

	// For each type.
	// NOTE: omit 0-th type ("unknown").
	size_t cItem = 0;
	CTypeSet::types_t::const_iterator it = _pTypeSet->Types ().begin ();
	for (++ it; it != _pTypeSet->Types ().end (); ++ it) {
		const type_desc_t & td = * it;
		if (static_cast<RGN_KIND> (td.Kind) != _Kind)
			continue;

		// Code.
		InsertItem  (_hListView, cItem, td.Type, "0x%04x", td.Type);
		SetItemIcon (_hListView, cItem, td.cImageListIndex);

		// Description.
		SetSubItemText (_hListView, cItem, 1, "%s", td.strTypeName);

		// Localized description.
		if (_pTypeSet->bHasLocalizedTypes && td.strTypeNameLocalized)
			SetSubItemText (_hListView, cItem, 2, "%s", td.strTypeNameLocalized);

		++ cItem;
	}

	AutoSizeColumns (_hListView);
}

void CShpWizard::_PrepareDBFView (HWND _hDbfFields) {
	ListView_SetExtendedListViewStyleEx (_hDbfFields, LVS_EX_GRIDLINES, LVS_EX_GRIDLINES);

	// Prepare columns.
	const size_t cFields = m_pDbt->GetFields ();
	for (size_t cField = 0; cField < cFields; ++ cField) {
		const char * const strFieldName = m_pDbt->GetFieldName (cField);

		// Prepare text.
		char strText [64];
		if (m_bOEM) {
			const size_t cLen = ::_tcslen (strFieldName);
			::OemToCharBuff (strFieldName, strText, cLen);
			strText [cLen] = '\0';
		} else
			::_tcscpy (strText, strFieldName);

		AddColumn (_hDbfFields, cField, strText);
	}
}

void CShpWizard::_FillDBFView (HWND _hDbfFields) {
	ListView_DeleteAllItems (_hDbfFields);

	const size_t cRecords = m_pDbt->GetRecords ();
	const size_t cFields  = m_pDbt->GetFields  ();

	for (size_t cField = 0; cField < cFields; ++ cField) {
		// Check if the field contains plain text.
		size_t cRecordsShown = 0;
		for (size_t cRecord = 0; cRecord < cRecords; ++ cRecord) {
			if (! m_pDbt->RecordExists (cRecord))
				continue;

			if (cRecordsShown > 100)
				break;
			++ cRecordsShown;

			const size_t cMaxLen = m_pDbt->GetFieldMaxLength (cRecord, cField);
			char * const strValue = reinterpret_cast<char *> (alloca (cMaxLen + 1));
			const size_t cLen = m_pDbt->GetFieldS (cRecord, cField, strValue);
			const char * const pEnd = strValue + cLen;

			// Omit leading spaces.
			const char * p = strValue;
			while (p < pEnd && * p == ' ')
				++ p;

			if (p >= pEnd) {
				// NOTE: per-record line should be inserted anyway. 
				if (cField == 0)
					InsertItem (_hDbfFields, cRecord, 0, "");
				continue;
			}

			// Prepare text.
			char strText [64];
			const size_t cValueLen = pEnd - p < sizeof (strText) - 1 ? pEnd - p : sizeof (strText) - 1;
			if (m_bOEM)
				::OemToCharBuff (p, strText, cValueLen);
			else
				::memcpy (strText, p, cValueLen);
			strText [cValueLen] = '\0';

			// Visualize the record's field.
			if (cField == 0)
				InsertItem     (_hDbfFields, cRecord, 0,      "%s", strText);
			else
				SetSubItemText (_hDbfFields, cRecord, cField, "%s", strText);
		}
	}
}

INT_PTR CALLBACK CShpWizard::PageTypeProc (HWND _hWnd, UINT _uMsg, WPARAM _wParam, LPARAM _lParam) {
	CShpWizard * const pThis = reinterpret_cast<CShpWizard *> (::GetWindowLong (_hWnd, GWL_USERDATA));

	switch (_uMsg) {
		case WM_INITDIALOG: {
			CenterPropertyPages (_hWnd);

			//
			// Prepare access to wizard's data.
			//

			const PROPSHEETPAGE * const pPage = reinterpret_cast<PROPSHEETPAGE *> (_lParam);
			CShpWizard * const pThis = reinterpret_cast<CShpWizard *> (pPage->lParam);
			::SetWindowLong (_hWnd, GWL_USERDATA, pPage->lParam);

			assert (pThis);

			//
			// Prepare bitmap.
			//

			SetStaticPicture (_hWnd, IDC_BITMAP, IDR_ShpWizard);

			//
			// Prepare TypeSet.
			//

			const HWND hTypeSet = ::GetDlgItem (_hWnd, IDC_SHP_TYPESET);

			if (g_map.pTypeSet != & g_tsNull) {
				::SendMessage (hTypeSet, CB_INSERTSTRING, 0, reinterpret_cast<LPARAM> (g_map.pTypeSet->strDisplayName));
				::SendMessage (hTypeSet, CB_SETITEMDATA,  0, reinterpret_cast<LPARAM> (g_map.pTypeSet));
				::SendMessage (hTypeSet, CB_SETCURSEL,    0, 0);
				::EnableWindow (hTypeSet, FALSE);

				pThis->m_pTypeSet = g_map.pTypeSet;
			} else {
				const CTypeSet * const pDefaultTypeSet = GetDefaultTypeSet ();
				if (g_map.empty ()) {
					g_map.pTypeSet = pDefaultTypeSet;
					UpdateMapSkin ();
				}

				size_t cTypeSet = 0;
				for (const CTypeSet * pTypeSet = CTypeSet::s_pFirst; pTypeSet; pTypeSet = pTypeSet->m_pNext) {
					if (pTypeSet == & g_tsNull)
						continue;

					::SendMessage (hTypeSet, CB_INSERTSTRING, cTypeSet, reinterpret_cast<LPARAM> (pTypeSet->strDisplayName));
					::SendMessage (hTypeSet, CB_SETITEMDATA,  cTypeSet, reinterpret_cast<LPARAM> (pTypeSet));

					if (pTypeSet == pDefaultTypeSet)
						::SendMessage (hTypeSet, CB_SETCURSEL, cTypeSet, 0);

					++ cTypeSet;
				}

				pThis->m_pTypeSet = pDefaultTypeSet;
			}

			//
			// Prepare 'Source' field.
			//

			const HWND hSrc = ::GetDlgItem (_hWnd, IDC_SHP_TYPE_SOURCE);

			::SetWindowLong (hSrc, GWL_EXSTYLE, WS_EX_CONTROLPARENT | ::GetWindowLong (hSrc, GWL_EXSTYLE));
			TabCtrl_SetExtendedStyle (hSrc, TCS_EX_FLATSEPARATORS);

			TCITEM tci;
			tci.mask       = TCIF_TEXT;
			tci.pszText    = "From list";
			tci.cchTextMax = ::strlen (tci.pszText);
			TabCtrl_InsertItem (hSrc, 0, & tci);
			tci.pszText    = "From field";
			tci.cchTextMax = ::strlen (tci.pszText);
			TabCtrl_InsertItem (hSrc, 1, & tci);

			TabCtrl_SetCurSel (hSrc, 0);

			RECT rectSrcDisplay;
			::GetClientRect (hSrc, & rectSrcDisplay);
			TabCtrl_AdjustRect (hSrc, FALSE, & rectSrcDisplay);
			::InflateRect (& rectSrcDisplay, -3, -3);

			//
			// Prepare 'Type' page.
			//

			pThis->m_hType = ::CreateWindowEx (
				WS_EX_CLIENTEDGE, WC_LISTVIEW, NULL, WS_CHILD | WS_TABSTOP | WS_VISIBLE | LVS_REPORT | LVS_SHOWSELALWAYS | LVS_SINGLESEL | LVS_SHAREIMAGELISTS,
				rectSrcDisplay.left, rectSrcDisplay.top, rectSrcDisplay.right - rectSrcDisplay.left, rectSrcDisplay.bottom - rectSrcDisplay.top,
				hSrc, reinterpret_cast<HMENU> (IDC_SHP_TYPE), g_hInst, 0
			);

			ListView_SetExtendedListViewStyleEx (pThis->m_hType, LVS_EX_FULLROWSELECT, LVS_EX_FULLROWSELECT);

			AddColumn (pThis->m_hType, 0, "Code");
			AddColumn (pThis->m_hType, 1, "Description");
			if (! g_strTypeLocalizationName.empty ())
				AddColumn (pThis->m_hType, 2, g_strTypeLocalizationName.c_str ());

			AutoSizeColumns (pThis->m_hType);

			//
			// Prepare 'Type field' page.
			//
			
			assert (pThis->m_pDbt);
			const CDatabaseTable & dbt = * pThis->m_pDbt;
			const size_t cRecords = dbt.GetRecords ();
			const size_t cFields  = dbt.GetFields  ();

			const size_t c_cTypeFieldStaticH = 16*HIWORD (::GetDialogBaseUnits ())/8;

			pThis->m_hTypeField = ::CreateWindowEx (
				WS_EX_CLIENTEDGE, WC_LISTVIEW, NULL, WS_CHILD | WS_TABSTOP | LVS_REPORT | LVS_SHOWSELALWAYS | LVS_SHAREIMAGELISTS,
				rectSrcDisplay.left, rectSrcDisplay.top, rectSrcDisplay.right - rectSrcDisplay.left, rectSrcDisplay.bottom - rectSrcDisplay.top - c_cTypeFieldStaticH,
				hSrc, reinterpret_cast<HMENU> (IDC_SHP_TYPE_FIELD), g_hInst, 0
			);

			pThis->m_hTypeFieldStatic = ::CreateWindowEx (
				0, _T ("Static"),
				_T ("NOTE: The type field should contain hexadecimal codes of types prefixed with '0x' or decimal ones."),
				WS_CHILD | WS_TABSTOP,
				rectSrcDisplay.left, rectSrcDisplay.bottom - c_cTypeFieldStaticH, rectSrcDisplay.right - rectSrcDisplay.left, c_cTypeFieldStaticH,
				hSrc, reinterpret_cast<HMENU> (-1), g_hInst, 0
			);
			const HFONT hFont = reinterpret_cast<HFONT> (::SendMessage (_hWnd, WM_GETFONT, 0, 0));
			::SendMessage (pThis->m_hTypeFieldStatic, WM_SETFONT, reinterpret_cast<WPARAM> (hFont), 0);

			pThis->m_bOEM = GetSetting ("ShpImport_OEM", false);
			pThis->_PrepareDBFView (pThis->m_hTypeField);
			pThis->_FillDBFView    (pThis->m_hTypeField);
			AutoSizeColumns (pThis->m_hTypeField);

			//
			// Prepare 'Kind' field.
			//
//			::SetDlgItemText (_hWnd, IDC_SHP_KIND, g_strKinds [pThis->m_Kind]);
			const HWND hKind = ::GetDlgItem (_hWnd, IDC_SHP_KIND);
			for (size_t cKind = 0; cKind < rkMax; ++ cKind) {
				::SendMessage (hKind, CB_INSERTSTRING, cKind, reinterpret_cast<LPARAM> (g_strKinds [cKind]));
				::SendMessage (hKind, CB_SETITEMDATA,  cKind, cKind);
				if (cKind == static_cast<size_t> (pThis->m_Kind))
					::SendMessage (hKind, CB_SETCURSEL, cKind, 0);
			}
			::EnableWindow (hKind, pThis->m_bEnableSelectKind);

			pThis->m_bFromTypeField = false;
			pThis->m_dwType         = 0;
			pThis->m_cTypeIndex     = -1;
			_FillTypes (pThis->m_hType, pThis->m_Kind, pThis->m_pTypeSet);

			return FALSE;
		}

		case WM_COMMAND:
			switch (LOWORD (_wParam)) {
				case IDC_SHP_KIND: {
					if (HIWORD (_wParam) == CBN_SELCHANGE) {
						//
						// Kind changed.
						//

						const HWND hKind = ::GetDlgItem (_hWnd, IDC_SHP_KIND);

						const size_t cSelIndex = ::SendMessage (hKind, CB_GETCURSEL, 0, 0);
						pThis->m_Kind = static_cast<RGN_KIND> (::SendMessage (hKind, CB_GETITEMDATA, cSelIndex, 0));

						pThis->m_dwType     = 0;
						pThis->m_cTypeIndex = -1;
						_FillTypes (pThis->m_hType, pThis->m_Kind, pThis->m_pTypeSet);

						if (! pThis->IsTypeDefined ())
							PropSheet_SetWizButtons (::GetParent (_hWnd), 0);

						return TRUE;
					}
					break;
				}
				case IDC_SHP_TYPESET: {
					if (HIWORD (_wParam) == CBN_SELCHANGE) {
						//
						// TypeSet changed.
						//

						const HWND hTypeSet = ::GetDlgItem (_hWnd, IDC_SHP_TYPESET);

						const size_t cSelIndex = ::SendMessage (hTypeSet, CB_GETCURSEL, 0, 0);
						pThis->m_pTypeSet = reinterpret_cast<const CTypeSet *> (::SendMessage (hTypeSet, CB_GETITEMDATA, cSelIndex, 0));

						pThis->m_dwType     = 0;
						pThis->m_cTypeIndex = -1;
						_FillTypes (pThis->m_hType, pThis->m_Kind, pThis->m_pTypeSet);

						if (! pThis->IsTypeDefined ())
							PropSheet_SetWizButtons (::GetParent (_hWnd), 0);

						return TRUE;
					}
					break;
				}
			}
			break;

	    case WM_NOTIFY: {
			const NMHDR * const pInfo = reinterpret_cast<NMHDR *> (_lParam);
			switch (pInfo->code) {
				case PSN_SETACTIVE: {
					if (! pThis->IsTypeDefined ())
						PropSheet_SetWizButtons (::GetParent (_hWnd), 0);
					else {
						PropSheet_SetWizButtons (::GetParent (_hWnd), PSWIZB_NEXT);

						// Restore selection.
						ListView_SetSelectionMark (pThis->m_hType, pThis->m_cTypeIndex);
					}

					break;
				}

				case PSN_QUERYCANCEL:
					pThis->m_bCanceled = true;
					break;

				case TCN_SELCHANGE: {
					//
					// Tab selected.
					//

					pThis->m_bFromTypeField = TabCtrl_GetCurSel (::GetDlgItem (_hWnd, IDC_SHP_TYPE_SOURCE)) != 0;
					::ShowWindow (pThis->m_hType,            pThis->m_bFromTypeField ? SW_HIDE : SW_SHOW);
					::ShowWindow (pThis->m_hTypeField,       pThis->m_bFromTypeField ? SW_SHOW : SW_HIDE);
					::ShowWindow (pThis->m_hTypeFieldStatic, pThis->m_bFromTypeField ? SW_SHOW : SW_HIDE);

					break;
				}

				case LVN_ITEMCHANGED: {
					if (pInfo->idFrom == IDC_SHP_TYPE) {
						//
						// Type changed.
						//

						const NMLISTVIEW * const pInfo2 = reinterpret_cast<const NMLISTVIEW *> (_lParam);
						if (
							(pInfo2->uOldState & LVIS_SELECTED) == 0 &&
							(pInfo2->uNewState & LVIS_SELECTED) != 0
						) {
							pThis->m_cTypeIndex = pInfo2->iItem;
							pThis->m_dwType     = pInfo2->lParam;

							PropSheet_SetWizButtons (::GetParent (_hWnd), PSWIZB_NEXT);
						}
					}

					break;
				}

				case LVN_COLUMNCLICK: {
					const NMLISTVIEW * const pInfo2 = reinterpret_cast<const NMLISTVIEW *> (_lParam);
					switch (pInfo->idFrom) {
						case IDC_SHP_TYPE: {
							//
							// Re-sort types.
							//

							if (g_iSortColumn != pInfo2->iSubItem)
								g_iSortColumn = pInfo2->iSubItem;
							else
								g_bAccent = ! g_bAccent;
							
							ListView_SortItemsEx (pInfo->hwndFrom, _CompareItems, reinterpret_cast<LPARAM> (pInfo->hwndFrom));

							break;
						}
						case IDC_SHP_TYPE_FIELD: {
							//
							// Another DBF field selected.
							//

							pThis->m_cTypeField = pInfo2->iSubItem;
							SelectColumn (pThis->m_hTypeField, pThis->m_cTypeField);

							PropSheet_SetWizButtons (::GetParent (_hWnd), PSWIZB_NEXT);

							break;
						}
					}

					break;
				}

				case NM_CLICK:
				case LVN_ITEMACTIVATE: {
					if (pInfo->idFrom == IDC_SHP_TYPE_FIELD) {
						//
						// Another DBF field selected.
						//
						const NMITEMACTIVATE * const pInfo2 = reinterpret_cast<const NMITEMACTIVATE *> (_lParam);

						pThis->m_cTypeField = pInfo2->iSubItem;
						SelectColumn (pThis->m_hTypeField, pThis->m_cTypeField);

						PropSheet_SetWizButtons (::GetParent (_hWnd), PSWIZB_NEXT);
					}

					break;
				}

            }
			break;
        }

		case WM_DESTROY:
			DeleteStaticPicture (_hWnd, IDC_BITMAP);
			break;
	}

	return FALSE;
}

//////////////////////////////////////////////////////

INT_PTR CALLBACK CShpWizard::PageLabelsProc (HWND _hWnd, UINT _uMsg, WPARAM _wParam, LPARAM _lParam) {
	CShpWizard * const pThis = reinterpret_cast<CShpWizard *> (::GetWindowLong (_hWnd, GWL_USERDATA));

	switch (_uMsg) {
		case WM_INITDIALOG: {
			//
			// Prepare access to wizard's data.
			//

			const PROPSHEETPAGE * const pPage = reinterpret_cast<PROPSHEETPAGE *> (_lParam);
			CShpWizard * const pThis = reinterpret_cast<CShpWizard *> (pPage->lParam);
			::SetWindowLong (_hWnd, GWL_USERDATA, pPage->lParam);

			//
			// Prepare bitmap.
			//

			SetStaticPicture (_hWnd, IDC_BITMAP, IDR_ShpWizard);

			//
			// Prepare DBF fields list view.
			//

			assert (pThis);
			assert (pThis->m_pDbt);
			const CDatabaseTable & dbt = * pThis->m_pDbt;
			const size_t cRecords = dbt.GetRecords ();
			const size_t cFields = dbt.GetFields ();

			::CheckDlgButton (_hWnd, IDC_ENABLE_LABELS_IMPORT, BST_CHECKED);

			pThis->m_bOEM = GetSetting ("ShpImport_OEM", false);
			::CheckDlgButton (_hWnd, IDC_SHP_LABELS_OEM, pThis->m_bOEM ? BST_CHECKED : BST_UNCHECKED);

			const HWND hDbfFields = ::GetDlgItem (_hWnd, IDC_DBF_FIELDS);
			pThis->_PrepareDBFView (hDbfFields);
			pThis->_FillDBFView    (hDbfFields);
			AutoSizeColumns (hDbfFields);

			// Check if a field contains plain text.
			for (size_t cField = 0; cField < cFields; ++ cField) {
				size_t cStrings = 0;
				size_t cDigits = 0;

				size_t cRecordsShown = 0;
				for (size_t cRecord = 0; cRecord < cRecords; ++ cRecord) {
					if (! dbt.RecordExists (cRecord))
						continue;

					if (cRecordsShown > 100)
						break;
					++ cRecordsShown;

					const size_t cMaxLen = dbt.GetFieldMaxLength (cRecord, cField);
					char * const strValue = reinterpret_cast<char *> (alloca (cMaxLen + 1));
					const size_t cLen = dbt.GetFieldS (cRecord, cField, strValue);
					const char * const pEnd = strValue + cLen;

					// Omit leading spaces.
					const char * p = strValue;
					while (p < pEnd && * p == ' ')
						++ p;
					if (p >= pEnd)
						continue;

					// Count statistics.
					while (p < pEnd) {
						if ('0' <= * p && * p <= '9')
							++ cDigits;
						else if  ((BYTE) * p >= 'A')
							++ cStrings;
						++ p;
					}
				}
				if (cStrings > 0 && cStrings >= cDigits) {
					pThis->m_cLabelField = cField;
					break;
				}
			}

			// Highlight the auto-selected column.
			if (pThis->m_cLabelField != -1)
				SelectColumn (hDbfFields, pThis->m_cLabelField);

			//
			// Prepare 'Transliterate' field.
			//

			const HWND hTransliteration = ::GetDlgItem (_hWnd, IDC_TRANSLITERATION);

			char strTransliteration [64];
			GetSetting ("ShpImport_Transliteration", strTransliteration, sizeof (strTransliteration), "");

			::SendMessage (hTransliteration, CB_INSERTSTRING, 0, (LPARAM) "Don't transliterate");
			::SendMessage (hTransliteration, CB_SETCURSEL, 0, 0);
			if (strTransliteration [0] == '\0')
				::SendMessage (hTransliteration, CB_SETCURSEL, 0, 0);
			pThis->m_pTransliterator = NULL;

			g_Transliteration.Load ();
			for (size_t cTrl = 0; cTrl < g_Transliteration.GetNumOfTransliterators (); ++ cTrl) {
				const CTransliterator * const pTransliterator = g_Transliteration.GetTransliterator (cTrl);

				::SendMessage (hTransliteration, CB_INSERTSTRING, cTrl + 1, (LPARAM) pTransliterator->GetName ().c_str ());

				if (pTransliterator->GetName () == strTransliteration) {
					::SendMessage (hTransliteration, CB_SETCURSEL, cTrl + 1, 0);
					pThis->m_pTransliterator = pTransliterator;
				}
			}

			return TRUE;
		}

		case WM_COMMAND:
			switch (LOWORD (_wParam)) {
				case IDC_EXTRA:
					DlgBox (IDD_IMPORT_SHP_EXTRA, _hWnd, ExtraProc, reinterpret_cast<LPARAM> (pThis));
					return TRUE;

				case IDC_ENABLE_LABELS_IMPORT:
					//
					// Labels import is enabled or disabled.
					//
					if (::IsDlgButtonChecked (_hWnd, IDC_ENABLE_LABELS_IMPORT) == BST_UNCHECKED) {
						::EnableWindow (::GetDlgItem (_hWnd, IDC_DBF_FIELDS), FALSE);
						::EnableWindow (::GetDlgItem (_hWnd, IDC_TRANSLITERATION), FALSE);
						::EnableWindow (::GetDlgItem (_hWnd, IDC_SHP_LABELS_OEM), FALSE);

						pThis->m_cLabelField = -1;
						SelectColumn (::GetDlgItem (_hWnd, IDC_DBF_FIELDS), -1);
					} else {
						::EnableWindow (::GetDlgItem (_hWnd, IDC_DBF_FIELDS), TRUE);
						::EnableWindow (::GetDlgItem (_hWnd, IDC_TRANSLITERATION), TRUE);
						::EnableWindow (::GetDlgItem (_hWnd, IDC_SHP_LABELS_OEM), TRUE);
					}

					return TRUE;

				case IDC_SHP_LABELS_OEM: {
					//
					// Convert from OEM or not.
					//
					pThis->m_bOEM = ::IsDlgButtonChecked (_hWnd, IDC_SHP_LABELS_OEM) == BST_CHECKED;
					SaveSetting ("ShpImport_OEM", pThis->m_bOEM);

					assert (pThis);
					assert (pThis->m_pDbt);
					const CDatabaseTable & dbt = * pThis->m_pDbt;

					const HWND hDbfFields = ::GetDlgItem (_hWnd, IDC_DBF_FIELDS);
					pThis->_FillDBFView (hDbfFields);

					if (pThis->m_cLabelField != -1)
						SelectColumn (::GetDlgItem (_hWnd, IDC_DBF_FIELDS), pThis->m_cLabelField);
				}

				case IDC_TRANSLITERATION: {
					switch (HIWORD (_wParam)) {
						case CBN_SELCHANGE: {
							//
							// Transliteration is selected.
							//

							const HWND hTransliteration = ::GetDlgItem (_hWnd, IDC_TRANSLITERATION);

							const size_t cSelIndex = ::SendMessage (hTransliteration, CB_GETCURSEL, 0, 0);
							if (cSelIndex == 0) {
								pThis->m_pTransliterator = NULL;
								SaveSetting ("ShpImport_Transliteration", "");
							} else {
								pThis->m_pTransliterator = g_Transliteration.GetTransliterator (cSelIndex - 1);
								SaveSetting ("ShpImport_Transliteration", pThis->m_pTransliterator->GetName ().c_str ());
							}

							return TRUE;
						}
					}
					break;
				}
			}
			break;

	    case WM_NOTIFY: {
			const NMHDR * const pInfo = reinterpret_cast<const NMHDR *> (_lParam);
            switch (pInfo->code) {
				case PSN_SETACTIVE:
					PropSheet_SetWizButtons (::GetParent (_hWnd), PSWIZB_BACK | PSWIZB_NEXT);
					break;

				case PSN_QUERYCANCEL:
					pThis->m_bCanceled = true;
					break;

				case LVN_COLUMNCLICK: {
					//
					// Another DBF field selected.
					//
					const NMLISTVIEW * const pInfo2 = reinterpret_cast<const NMLISTVIEW *> (_lParam);

					pThis->m_cLabelField = pInfo2->iSubItem;
					SelectColumn (::GetDlgItem (_hWnd, IDC_DBF_FIELDS), pThis->m_cLabelField);

					break;
				}

				case NM_CLICK:
				case LVN_ITEMACTIVATE: {
					//
					// Another DBF field selected.
					//
					const NMITEMACTIVATE * const pInfo2 = reinterpret_cast<const NMITEMACTIVATE *> (_lParam);

					pThis->m_cLabelField = pInfo2->iSubItem;
					SelectColumn (::GetDlgItem (_hWnd, IDC_DBF_FIELDS), pThis->m_cLabelField);

					break;
				}
			}
			break;
		}

		case WM_DESTROY:
			DeleteStaticPicture (_hWnd, IDC_BITMAP);
			break;
	}

	return FALSE;
}

static
bool _IsReal (HWND _hWnd, UINT _uID) {
	char strText [32];
	::GetDlgItemText (_hWnd, _uID, strText, sizeof (strText) - 1);
	strText [sizeof (strText) - 1] = '\0';
	const size_t cLen = ::strlen (strText);

	const char * pEnd = NULL;
	float f = 0;
	if (! ParseReal (strText, strText + sizeof (strText), & pEnd, f))
		return false;
	return pEnd >= strText + cLen;
}

enum cm_t {
	cmZone = 0,
	cmAuto = 1,
	cmCustom = 2
};

enum cu_t {
	cuDegrees  = 0,
	cuUTM_N    = 1,
	cuUTM_S    = 2,
	cuGK       = 3,
	cuSK42     = 4,
	cuBNG      = 5,
	cuIrishG   = 6,
	cuSwiss    = 7,
	cuNZMG     = 8,
	cuRT90     = 9,
	cuDHDN     = 10,
	cuIG1      = 11,
	cuIG2      = 12,
	cuIrishTM  = 13,
	cuMercator = 14,
	cuNZTM2000 = 15,
	cuDutchRD  = 16,
};

struct CUnits {
	const char * strName;
	cu_t cu;
};

const CUnits g_units [] = {
	{"Latitude/Longitude (deg)",       cuDegrees},
	{"UTM North hemisphere (m)",       cuUTM_N},
	{"UTM South hemisphere (m)",       cuUTM_S},
	{"British National grid",          cuBNG},
	{"Dutch RD grid",                  cuDutchRD},
	{"Gauss-Kruger (m)",               cuGK},
	{"Germany DHDN (Potsdam)",         cuDHDN},
	{"Irish Grid",                     cuIrishG},
	{"Irish Transverse Mercator",      cuIrishTM},
	{"New Zealand Map Grid",           cuNZMG},
	{"New Zealand TM 2000",            cuNZTM2000},
//	{"Mercator",                       cuMercator},
	{"Monte Mario / Italy zone 1",     cuIG1},
	{"Monte Mario / Italy zone 2",     cuIG2},
	{"Russian SK-42",                  cuSK42},
	{"Swedish RT90 2.5 gon W",         cuRT90},
	{"Swiss Coordinate System",        cuSwiss},
};

const size_t c_cUnits = sizeof (g_units)/sizeof (g_units [0]);

static
void _SetDlgItem (HWND _hWnd, int _nIdDlgItem, const char * _strText, BOOL _bEnable) {
	::SetDlgItemText (_hWnd, _nIdDlgItem, _strText);
	::EnableWindow (::GetDlgItem (_hWnd, _nIdDlgItem), _bEnable);
}

static
void _UpdateCMForNationalGrid (HWND _hWnd, const char * _strFE, const char * _strFN) {
	::EnableWindow (::GetDlgItem (_hWnd, IDC_SHP_X_ZONE),     FALSE);
	::EnableWindow (::GetDlgItem (_hWnd, IDC_SHP_ZONE),       FALSE);
	::EnableWindow (::GetDlgItem (_hWnd, IDC_SHP_X_AUTO),     FALSE);
	::EnableWindow (::GetDlgItem (_hWnd, IDC_SHP_X_CUSTOM),   FALSE);
	::EnableWindow (::GetDlgItem (_hWnd, IDC_SHP_OFFSET_X),   FALSE);

	_SetDlgItem (_hWnd, IDC_SHP_FE, _strFE, FALSE);
	_SetDlgItem (_hWnd, IDC_SHP_FN, _strFN, FALSE);
}

void CShpWizard::UpdateCentralMeridian (HWND _hWnd, int _cu, const CShpWizard * _pThis) {
	switch (_cu) {
		case cuDegrees:
			::EnableWindow (::GetDlgItem (_hWnd, IDC_SHP_X_ZONE),     FALSE);
			::EnableWindow (::GetDlgItem (_hWnd, IDC_SHP_ZONE),       FALSE);
			::EnableWindow (::GetDlgItem (_hWnd, IDC_SHP_X_AUTO),     FALSE);
			::EnableWindow (::GetDlgItem (_hWnd, IDC_SHP_X_CUSTOM),   FALSE);
			::EnableWindow (::GetDlgItem (_hWnd, IDC_SHP_OFFSET_X),   FALSE);
			::EnableWindow (::GetDlgItem (_hWnd, IDC_SHP_FE),         FALSE);
			::EnableWindow (::GetDlgItem (_hWnd, IDC_SHP_FN),         FALSE);

//			if (_pThis->m_pShpHdr->box.dY1 - _pThis->m_pShpHdr->box.dY0 > 180)
//				WarningMsgBox ("This selection of units may cause unpredictable results!");
			break;

		case cuUTM_S:
		case cuUTM_N:
			::EnableWindow (::GetDlgItem (_hWnd, IDC_SHP_X_ZONE),     TRUE);
			::EnableWindow (::GetDlgItem (_hWnd, IDC_SHP_ZONE),       TRUE);
			::EnableWindow (::GetDlgItem (_hWnd, IDC_SHP_X_AUTO),     FALSE);
			::EnableWindow (::GetDlgItem (_hWnd, IDC_SHP_X_CUSTOM),   FALSE);
			::EnableWindow (::GetDlgItem (_hWnd, IDC_SHP_OFFSET_X),   FALSE);
			::CheckRadioButton (_hWnd, IDC_SHP_X_ZONE, IDC_SHP_X_CUSTOM, IDC_SHP_X_ZONE);

			_SetDlgItem (_hWnd, IDC_SHP_FE, "500000.0",               FALSE);
			_SetDlgItem (_hWnd, IDC_SHP_FN, "0.0",                    FALSE);
			break;

		case cuMercator:
		case cuGK: {
			::EnableWindow (::GetDlgItem (_hWnd, IDC_SHP_X_ZONE),     FALSE);
			::EnableWindow (::GetDlgItem (_hWnd, IDC_SHP_ZONE),       FALSE);
			::EnableWindow (::GetDlgItem (_hWnd, IDC_SHP_X_AUTO),     FALSE);
			::EnableWindow (::GetDlgItem (_hWnd, IDC_SHP_X_CUSTOM),   TRUE);
			::EnableWindow (::GetDlgItem (_hWnd, IDC_SHP_OFFSET_X),   TRUE);
			::CheckRadioButton (_hWnd, IDC_SHP_X_ZONE, IDC_SHP_X_CUSTOM, IDC_SHP_X_CUSTOM);

			char strBuf [32];
			if (_pThis->m_bInitTransverseMercator) {
				::sprintf (strBuf, "%0.5f", _pThis->m_fCentralMeridian);
				::SetDlgItemText (_hWnd, IDC_SHP_OFFSET_X, strBuf);

				::sprintf (strBuf, "%0.1f", _pThis->m_fFE);
				_SetDlgItem (_hWnd, IDC_SHP_FE, strBuf, TRUE);

				::sprintf (strBuf, "%0.1f", _pThis->m_fFN);
				_SetDlgItem (_hWnd, IDC_SHP_FN, strBuf, TRUE);
			} else {
				GetSetting ("ShpImport_OriginX", strBuf, sizeof (strBuf) - 1, "0.00000");
				::SetDlgItemText (_hWnd, IDC_SHP_OFFSET_X, strBuf);

				GetSetting ("ShpImport_FE", strBuf, sizeof (strBuf) - 1, "500000.0");
				_SetDlgItem (_hWnd, IDC_SHP_FE, strBuf, TRUE);

				GetSetting ("ShpImport_FN", strBuf, sizeof (strBuf) - 1, "0.0");
				_SetDlgItem (_hWnd, IDC_SHP_FN, strBuf, TRUE);
			}

			break;
		}

		case cuSK42:
			::EnableWindow (::GetDlgItem (_hWnd, IDC_SHP_X_ZONE),     FALSE);
			::EnableWindow (::GetDlgItem (_hWnd, IDC_SHP_ZONE),       FALSE);
			::EnableWindow (::GetDlgItem (_hWnd, IDC_SHP_X_AUTO),     TRUE);
			::EnableWindow (::GetDlgItem (_hWnd, IDC_SHP_X_CUSTOM),   FALSE);
			::EnableWindow (::GetDlgItem (_hWnd, IDC_SHP_OFFSET_X),   FALSE);
			::CheckRadioButton (_hWnd, IDC_SHP_X_ZONE, IDC_SHP_X_CUSTOM, IDC_SHP_X_AUTO);

			_SetDlgItem (_hWnd, IDC_SHP_FE, "500000.0",               FALSE);
			_SetDlgItem (_hWnd, IDC_SHP_FN, "0.0",                    FALSE);

//			if ((_pThis->m_pShpHdr->box.dX0 + _pThis->m_pShpHdr->box.dX1)*.5 < 1000000)
//				WarningMsgBox ("This selection of units may cause unpredictable results!");
			break;

		case cuDHDN:
			::EnableWindow (::GetDlgItem (_hWnd, IDC_SHP_X_ZONE),     FALSE);
			::EnableWindow (::GetDlgItem (_hWnd, IDC_SHP_ZONE),       FALSE);
			::EnableWindow (::GetDlgItem (_hWnd, IDC_SHP_X_AUTO),     TRUE);
			::EnableWindow (::GetDlgItem (_hWnd, IDC_SHP_X_CUSTOM),   FALSE);
			::EnableWindow (::GetDlgItem (_hWnd, IDC_SHP_OFFSET_X),   FALSE);
			::CheckRadioButton (_hWnd, IDC_SHP_X_ZONE, IDC_SHP_X_CUSTOM, IDC_SHP_X_AUTO);

			_SetDlgItem (_hWnd, IDC_SHP_FE, "500000.0",               FALSE);
			_SetDlgItem (_hWnd, IDC_SHP_FN, "0.0",                    FALSE);

			break;

		case cuBNG:      _UpdateCMForNationalGrid (_hWnd,  "400000.0",  "-100000.0"); break;
		case cuDutchRD:  _UpdateCMForNationalGrid (_hWnd,  "155000.0", "-5316592.0"); break;
		case cuIrishG:   _UpdateCMForNationalGrid (_hWnd,  "200000.0",  "-250000.0"); break;
		case cuIrishTM:  _UpdateCMForNationalGrid (_hWnd,  "600000.0",  "-750000.0"); break;
		case cuSwiss:    _UpdateCMForNationalGrid (_hWnd,  "600000.0",   "200000.0"); break;
		case cuRT90:     _UpdateCMForNationalGrid (_hWnd, "1500000.0",        "0.0"); break;
		case cuNZMG:     _UpdateCMForNationalGrid (_hWnd, "2510000.0",  "6023150.0"); break;
		case cuNZTM2000: _UpdateCMForNationalGrid (_hWnd, "1600000.0", "10000000.0"); break;
		case cuIG1:      _UpdateCMForNationalGrid (_hWnd, "1500000.0",        "0.0"); break;
		case cuIG2:      _UpdateCMForNationalGrid (_hWnd, "2520000.0",        "0.0"); break;

		default:
			assert (false);
	}
}

static
bool _ValidateUnits (HWND _hWnd) {
	const bool bValid_OffsetX = _IsReal (_hWnd, IDC_SHP_OFFSET_X);
	const bool bValid_FE      = _IsReal (_hWnd, IDC_SHP_FE);
	const bool bValid_FN      = _IsReal (_hWnd, IDC_SHP_FN);

	if (bValid_OffsetX && bValid_FE && bValid_FN) {
		PropSheet_SetWizButtons (::GetParent (_hWnd), PSWIZB_BACK | PSWIZB_NEXT);
		return true;
	} else {
		PropSheet_SetWizButtons (::GetParent (_hWnd), 0);
		return false;
	}
}

static
bool _TestTransformation (
	point_t & _p1, point_t & _p2, point_t & _p3, point_t & _p4,
	CTransformationBase * _pUnitTransformation,
	CTransformationBase * _pDatumTransformation
) {
	__try {
		if (_pUnitTransformation) {
			_pUnitTransformation->ApplyTo (_p1);
			_pUnitTransformation->ApplyTo (_p2);
			_pUnitTransformation->ApplyTo (_p3);
			_pUnitTransformation->ApplyTo (_p4);
		}

		if (_pDatumTransformation) {
			_pDatumTransformation->ApplyTo (_p1);
			_pDatumTransformation->ApplyTo (_p2);
			_pDatumTransformation->ApplyTo (_p3);
			_pDatumTransformation->ApplyTo (_p4);
		}

		return true;
	} __except (EXCEPTION_EXECUTE_HANDLER) {
		return false;
	}
}

void CShpWizard::UpdateTargetCoordinates (HWND _hWnd, CShpWizard * _pThis) {
	// Central meridian.
	cm_t cm = cmAuto;
	if (::IsDlgButtonChecked (_hWnd, IDC_SHP_X_ZONE))
		cm = cmZone;
	else if (::IsDlgButtonChecked (_hWnd, IDC_SHP_X_AUTO))
		cm = cmAuto;
	else if (::IsDlgButtonChecked (_hWnd, IDC_SHP_X_CUSTOM))
		cm = cmCustom;
	SaveSetting ("ShpImport_CM", DWORD (cm));

	// Zone.
	const HWND hZone = ::GetDlgItem (_hWnd, IDC_SHP_ZONE);
	const size_t cZone_SelIndex = ::SendMessage (hZone, CB_GETCURSEL, 0, 0);
	if (cZone_SelIndex == -1)
		return;
	const size_t cZone = ::SendMessage (hZone, CB_GETITEMDATA, cZone_SelIndex, 0);
	assert (1 <= cZone && cZone <= 60);
	SaveSetting ("ShpImport_Zone", DWORD (cZone));

	// Central meridian.
	char strText [32];
	::GetDlgItemText (_hWnd, IDC_SHP_OFFSET_X, strText, sizeof (strText) - 1);
	strText [31] = '\0';
	const float fCustomX = ParseReal (strText, strText + ::strlen (strText));
	::sprintf (strText, "%0.5f", fCustomX);
	SaveSetting ("ShpImport_OriginX", strText);

	const float fAveX = (_pThis->m_box.x0 + _pThis->m_box.x1)*.5;

	// False Easting.
	::GetDlgItemText (_hWnd, IDC_SHP_FE, strText, sizeof (strText) - 1);
	strText [31] = '\0';
	float fFE = ParseReal (strText, strText + ::strlen (strText));
	::sprintf (strText, "%0.1f", fFE);
	SaveSetting ("ShpImport_FE", strText);

	// False Northing.
	::GetDlgItemText (_hWnd, IDC_SHP_FN, strText, sizeof (strText) - 1);
	strText [31] = '\0';
	const float fFN = ParseReal (strText, strText + ::strlen (strText));
	::sprintf (strText, "%0.1f", fFN);
	SaveSetting ("ShpImport_FN", strText);

	// Datum.
	const HWND hDatum = ::GetDlgItem (_hWnd, IDC_SHP_DATUM);
	const size_t cDatum_SelIndex = ::SendMessage (hDatum, CB_GETCURSEL, 0, 0);
	if (cDatum_SelIndex == -1)
		return;
	const size_t cDatum = ::SendMessage (hDatum, CB_GETITEMDATA, cDatum_SelIndex, 0);
	assert (cDatum < c_cDatums);
	if (cDatum < c_cDatums) {
		_pThis->m_pDatumTransformation = g_Datums [cDatum].ToWGS84 ();
		SaveSetting ("ShpImport_Datum", (DWORD) cDatum);
	}

	const CEllipsoid & Ellipsoid = g_Datums [cDatum].pEllipsoid ? * g_Datums [cDatum].pEllipsoid : c_eWGS84;

	// Units.
	const HWND hUnits = ::GetDlgItem (_hWnd, IDC_SHP_UNITS);
	const size_t cUnits_SelIndex = ::SendMessage (hUnits, CB_GETCURSEL, 0, 0);
	const int cu = ::SendMessage (hUnits, CB_GETITEMDATA, cUnits_SelIndex, 0);
	SaveSetting ("ShpImport_Units", (DWORD) cu);

	//
	// Create the unit transformation object.
	//
	switch (cu) {
		case cuDegrees:  _pThis->m_pUnitTransformation = std::auto_ptr<CTransformationBase> (NULL); break;
		case cuUTM_N:    _pThis->m_pUnitTransformation = std::auto_ptr<CTransformationBase> (new CUTM_North (Ellipsoid, cZone)); break;
		case cuUTM_S:    _pThis->m_pUnitTransformation = std::auto_ptr<CTransformationBase> (new CUTM_South (Ellipsoid, cZone)); break;
		case cuGK:       _pThis->m_pUnitTransformation = std::auto_ptr<CTransformationBase> (new CTransverseMercatorToLL (Ellipsoid, 1, fFE, fFN, fCustomX, 0)); break;
//		case cuMercator: _pThis->m_pUnitTransformation = std::auto_ptr<CTransformationBase> (new CMercatorToLL          (Ellipsoid.a, Ellipsoid.e, 1, fFE, fFN, fCustomX)); break;
		case cuSK42:     _pThis->m_pUnitTransformation = std::auto_ptr<CTransformationBase> (new CS42Grid (fAveX)); break;
		case cuDHDN:     _pThis->m_pUnitTransformation = std::auto_ptr<CTransformationBase> (new CDHDNGrid (fAveX)); break;
		case cuBNG:      _pThis->m_pUnitTransformation = std::auto_ptr<CTransformationBase> (new CBritishNationalGrid   ()); break;
		case cuDutchRD:  _pThis->m_pUnitTransformation = std::auto_ptr<CTransformationBase> (new CDutchRDGrid           ()); break;
		case cuIrishG:   _pThis->m_pUnitTransformation = std::auto_ptr<CTransformationBase> (new CIrishGrid             ()); break;
		case cuIrishTM:  _pThis->m_pUnitTransformation = std::auto_ptr<CTransformationBase> (new CIrishTM               ()); break;
		case cuSwiss:    _pThis->m_pUnitTransformation = std::auto_ptr<CTransformationBase> (new CSwissCoordinateSystem ()); break;
		case cuRT90:     _pThis->m_pUnitTransformation = std::auto_ptr<CTransformationBase> (new CSwedishRT90_25GonW    ()); break;
		case cuNZMG:     _pThis->m_pUnitTransformation = std::auto_ptr<CTransformationBase> (new CNewZealandMapGrid     ()); break;
		case cuNZTM2000: _pThis->m_pUnitTransformation = std::auto_ptr<CTransformationBase> (new CNZTM2000              ()); break;
		case cuIG1:      _pThis->m_pUnitTransformation = std::auto_ptr<CTransformationBase> (new CItalyGrid1            ()); break;
		case cuIG2:      _pThis->m_pUnitTransformation = std::auto_ptr<CTransformationBase> (new CItalyGrid2            ()); break;
		default: assert (false);
	}

	if (_pThis->m_box.Invalid ()) {
		::SetDlgItemText (_hWnd, IDC_SHP_X0_OUT, "-");
		::SetDlgItemText (_hWnd, IDC_SHP_X1_OUT, "-");
		::SetDlgItemText (_hWnd, IDC_SHP_Y0_OUT, "-");
		::SetDlgItemText (_hWnd, IDC_SHP_Y1_OUT, "-");
		return;
	} else {
		//
		// Test the transformations.
		//
		point_t p1 (_pThis->m_box.x0, _pThis->m_box.y0);
		point_t p2 (_pThis->m_box.x1, _pThis->m_box.y0);
		point_t p3 (_pThis->m_box.x1, _pThis->m_box.y1);
		point_t p4 (_pThis->m_box.x0, _pThis->m_box.y1);

		if (! _TestTransformation (p1, p2, p3, p4, _pThis->m_pUnitTransformation.get (), _pThis->m_pDatumTransformation)) {
			::SetDlgItemText (_hWnd, IDC_SHP_X0_OUT, "(error)");
			::SetDlgItemText (_hWnd, IDC_SHP_X1_OUT, "(error)");
			::SetDlgItemText (_hWnd, IDC_SHP_Y0_OUT, "(error)");
			::SetDlgItemText (_hWnd, IDC_SHP_Y1_OUT, "(error)");
		} else {
			const double fMinX = __min (__min (p1.x, p2.x), __min (p3.x, p4.x));
			const double fMaxX = __max (__max (p1.x, p2.x), __max (p3.x, p4.x));
			const double fMinY = __min (__min (p1.y, p2.y), __min (p3.y, p4.y));
			const double fMaxY = __max (__max (p1.y, p2.y), __max (p3.y, p4.y));

			char strBuf [64];
			::sprintf (strBuf, "%0.5f", fMinX);
			::SetDlgItemText (_hWnd, IDC_SHP_X0_OUT, strBuf);
			::sprintf (strBuf, "%0.5f", fMaxX);
			::SetDlgItemText (_hWnd, IDC_SHP_X1_OUT, strBuf);
			::sprintf (strBuf, "%0.5f", fMinY);
			::SetDlgItemText (_hWnd, IDC_SHP_Y0_OUT, strBuf);
			::sprintf (strBuf, "%0.5f", fMaxY);
			::SetDlgItemText (_hWnd, IDC_SHP_Y1_OUT, strBuf);
		}
	}
}

INT_PTR CALLBACK CShpWizard::PageUnitsProc (HWND _hWnd, UINT _uMsg, WPARAM _wParam, LPARAM _lParam) {
	CShpWizard * const pThis = reinterpret_cast<CShpWizard *> (::GetWindowLong (_hWnd, GWL_USERDATA));

	switch (_uMsg) {
		case WM_INITDIALOG: {
			//
			// Prepare access to wizard's data.
			//

			const PROPSHEETPAGE * const pPage = reinterpret_cast<PROPSHEETPAGE *> (_lParam);
			CShpWizard * const pThis = reinterpret_cast<CShpWizard *> (pPage->lParam);
			::SetWindowLong (_hWnd, GWL_USERDATA, pPage->lParam);

			//
			// Prepare bitmap.
			//

			SetStaticPicture (_hWnd, IDC_BITMAP, IDR_ShpWizard);

			char strBuf [64];

			//
			// Input bounding rect.
			//
			if (pThis->m_box.Invalid ()) {
				::SetDlgItemText (_hWnd, IDC_SHP_X0, "-");
				::SetDlgItemText (_hWnd, IDC_SHP_X1, "-");
				::SetDlgItemText (_hWnd, IDC_SHP_Y0, "-");
				::SetDlgItemText (_hWnd, IDC_SHP_Y1, "-");
			} else {
				::sprintf (strBuf, "%0.5f", pThis->m_box.x0);
				::SetDlgItemText (_hWnd, IDC_SHP_X0, strBuf);
				::sprintf (strBuf, "%0.5f", pThis->m_box.x1);
				::SetDlgItemText (_hWnd, IDC_SHP_X1, strBuf);
				::sprintf (strBuf, "%0.5f", pThis->m_box.y0);
				::SetDlgItemText (_hWnd, IDC_SHP_Y0, strBuf);
				::sprintf (strBuf, "%0.5f", pThis->m_box.y1);
				::SetDlgItemText (_hWnd, IDC_SHP_Y1, strBuf);
			}

			//
			// Coordinate System.
			//
			const bool bIsInM = pThis->m_box.y1 - pThis->m_box.y0 > 180 || pThis->m_box.y1 > 180;
			int cunits = bIsInM ? GetSetting ("ShpImport_Units", DWORD (cuUTM_N)) : cuDegrees;
			if (cunits == cuSK42 && (pThis->m_box.x0 + pThis->m_box.x1)*.5 < 1000000)
				cunits = cuGK;

			if (pThis->m_bInitLatLon)
				cunits = cuDegrees;
			else if (pThis->m_bInitTransverseMercator)
				cunits = cuGK;

			const HWND hUnits = ::GetDlgItem (_hWnd, IDC_SHP_UNITS);
			for (size_t cu = 0; cu < c_cUnits; ++ cu) {
				::SendMessage (hUnits, CB_INSERTSTRING, cu, (LPARAM) g_units [cu].strName);
				::SendMessage (hUnits, CB_SETITEMDATA,  cu, g_units [cu].cu);
				if (g_units [cu].cu == cunits)
					::SendMessage (hUnits, CB_SETCURSEL, cu, 0);
			}

			//
			// Central meridian.
			//

			// Zone.
			size_t cZone = GetSetting ("ShpImport_Zone", (DWORD) 1);
			if (cZone == 0)
				cZone = 1;
			if (cZone > 60)
				cZone = 60;

			const HWND hZone = ::GetDlgItem (_hWnd, IDC_SHP_ZONE);
			for (size_t z = 1; z <= 60; ++ z) {
				char strZone [10];
				::sprintf (strZone, "%d", z);

				::SendMessage (hZone, CB_INSERTSTRING, z - 1, reinterpret_cast<LPARAM> (strZone));
				::SendMessage (hZone, CB_SETITEMDATA,  z - 1, z);
				if (z == cZone)
					::SendMessage (hZone, CB_SETCURSEL, z - 1, 0);
			}

			// Custom.
			GetSetting ("ShpImport_OriginX", strBuf, sizeof (strBuf) - 1, "0.00000");
			::SetDlgItemText (_hWnd, IDC_SHP_OFFSET_X, strBuf);

			// FE & FN.
			UpdateCentralMeridian (_hWnd, cunits, pThis);

			//
			// Datum.
			//
			const HWND hDatum = ::GetDlgItem (_hWnd, IDC_SHP_DATUM);

			size_t d = 0;
			for (size_t cDatum = 0; cDatum < c_cDatums; ++ cDatum) {
				if (g_Datums [cDatum].strName == NULL)
					continue;

				::SendMessage (hDatum, CB_INSERTSTRING, d, reinterpret_cast<LPARAM> (g_Datums [cDatum].strName));
				::SendMessage (hDatum, CB_SETITEMDATA,  d, cDatum);
				++ d;
			}

			DWORD dwDatum = GetSetting ("ShpImport_Datum", (DWORD) c_cDatums - 1);
			if (dwDatum > c_cDatums)
				dwDatum = c_cDatums - 1;
			::SendMessage (hDatum, CB_SETCURSEL, dwDatum, 0);

			//
			// Output bounding rect.
			//
			UpdateTargetCoordinates (_hWnd, pThis);

			return TRUE;
		}

		case WM_COMMAND: {
			const UINT uId = LOWORD (_wParam);
			switch (uId) {
				case IDC_SHP_UNITS: {
					switch (HIWORD (_wParam)) {
						case CBN_SELCHANGE: {
							//
							// Coordinate system is selected.
							//
							const HWND hUnits = ::GetDlgItem (_hWnd, IDC_SHP_UNITS);

							const size_t cSelIndex = ::SendMessage (hUnits, CB_GETCURSEL, 0, 0);
							const int cu = ::SendMessage (hUnits, CB_GETITEMDATA, cSelIndex, 0);

							UpdateCentralMeridian (_hWnd, cu, pThis);

							UpdateTargetCoordinates (_hWnd, pThis);

							return TRUE;
						}
					}
					break;
				}
				case IDC_SHP_DATUM: {
					switch (HIWORD (_wParam)) {
						case CBN_SELCHANGE: {
							UpdateTargetCoordinates (_hWnd, pThis);
							return TRUE;
						}
					}
					break;
				}

				case IDC_SHP_X_ZONE:
					::EnableWindow (::GetDlgItem (_hWnd, IDC_SHP_ZONE),     TRUE);
					::EnableWindow (::GetDlgItem (_hWnd, IDC_SHP_OFFSET_X), FALSE);
					::CheckRadioButton (_hWnd, IDC_SHP_X_ZONE, IDC_SHP_X_CUSTOM, IDC_SHP_X_ZONE);
					UpdateTargetCoordinates (_hWnd, pThis);
					break;

				case IDC_SHP_X_AUTO:
					::EnableWindow (::GetDlgItem (_hWnd, IDC_SHP_ZONE),     FALSE);
					::EnableWindow (::GetDlgItem (_hWnd, IDC_SHP_OFFSET_X), FALSE);
					::CheckRadioButton (_hWnd, IDC_SHP_X_ZONE, IDC_SHP_X_CUSTOM, IDC_SHP_X_AUTO);
					UpdateTargetCoordinates (_hWnd, pThis);
					break;

				case IDC_SHP_X_CUSTOM:
					::EnableWindow (::GetDlgItem (_hWnd, IDC_SHP_ZONE),     FALSE);
					::EnableWindow (::GetDlgItem (_hWnd, IDC_SHP_OFFSET_X), TRUE);
					::CheckRadioButton (_hWnd, IDC_SHP_X_ZONE, IDC_SHP_X_CUSTOM, IDC_SHP_X_CUSTOM);
					UpdateTargetCoordinates (_hWnd, pThis);
					break;

				case IDC_SHP_ZONE:
					UpdateTargetCoordinates (_hWnd, pThis);
					break;

				case IDC_SHP_OFFSET_X:
				case IDC_SHP_FE:
				case IDC_SHP_FN: {
					switch (HIWORD (_wParam)) {
						case EN_CHANGE:
							if (_ValidateUnits (_hWnd))
								UpdateTargetCoordinates (_hWnd, pThis);	
							break;

//						case EN_KILLFOCUS:
//							UpdateTargetCoordinates (_hWnd, pThis);
//							break;
					}
					break;
				}
			}
			break;
		}

	    case WM_NOTIFY: {
			const NMHDR * const pInfo = reinterpret_cast<NMHDR *> (_lParam);
			switch (pInfo->code) {
				case PSN_SETACTIVE:
					PropSheet_SetWizButtons (::GetParent (_hWnd), PSWIZB_BACK | PSWIZB_NEXT);
					break;

				case PSN_QUERYCANCEL:
					pThis->m_bCanceled = true;
					break;
			}

			break;
        }

		case WM_DESTROY:
			DeleteStaticPicture (_hWnd, IDC_BITMAP);
			break;
	}

	return FALSE;
}

INT_PTR CALLBACK CShpWizard::PageLevelsProc (HWND _hWnd, UINT _uMsg, WPARAM _wParam, LPARAM _lParam) {
	CShpWizard * const pThis = reinterpret_cast<CShpWizard *> (::GetWindowLong (_hWnd, GWL_USERDATA));

	switch (_uMsg) {
		case WM_INITDIALOG: {
			//
			// Prepare access to wizard's data.
			//

			const PROPSHEETPAGE * const pPage = reinterpret_cast<PROPSHEETPAGE *> (_lParam);
			CShpWizard * const pThis = reinterpret_cast<CShpWizard *> (pPage->lParam);
			::SetWindowLong (_hWnd, GWL_USERDATA, pPage->lParam);

			//
			// Prepare bitmap.
			//

			SetStaticPicture (_hWnd, IDC_BITMAP, IDR_ShpWizard);

			//
			// Prepare levels list.
			//

			const HWND hLevels = ::GetDlgItem (_hWnd, IDC_SHP_LEVELS);

			ListView_SetExtendedListViewStyleEx (hLevels, LVS_EX_CHECKBOXES, LVS_EX_CHECKBOXES);
			AddColumn (hLevels, 0, "Level");

			size_t cLevels;
			{
				CCritSec::CLock lock (g_cs);
				// NOTE: the last level is always empty.
				cLevels = g_map.Levels.empty () ? 0 : g_map.Levels.size () - 1;
			}
			if (cLevels == 0)
				cLevels = 1;

			pThis->m_Levels.resize (cLevels);
			for (size_t cLevel = 0; cLevel < cLevels; ++ cLevel) {
				pThis->m_Levels [cLevel] = false;

				InsertItem (hLevels, cLevel, 0, "Level%d", cLevel);
				ListView_SetCheckState (hLevels, cLevel, TRUE);
			}

			AutoSizeColumns (hLevels);
			
			return TRUE;
		}

	    case WM_NOTIFY: {
			const NMHDR * const pInfo = reinterpret_cast<NMHDR *> (_lParam);
			switch (pInfo->code) {
				case PSN_SETACTIVE:
					PropSheet_SetWizButtons (::GetParent (_hWnd), PSWIZB_BACK | PSWIZB_FINISH);
					break;

				case PSN_QUERYCANCEL:
					pThis->m_bCanceled = true;
					break;

				case PSN_WIZNEXT:
				case PSN_WIZFINISH: {
					const HWND hLevels = ::GetDlgItem (_hWnd, IDC_SHP_LEVELS);
					pThis->m_cLevelsToImport = 0;
					for (size_t cLevel = 0; cLevel < pThis->m_Levels.size (); ++ cLevel) {
						pThis->m_Levels [cLevel] = ListView_GetCheckState (hLevels, cLevel) != FALSE;
						if (pThis->m_Levels [cLevel])
							++ pThis->m_cLevelsToImport;
					}

					break;
				}

			}
			break;
		}

		case WM_DESTROY:
			DeleteStaticPicture (_hWnd, IDC_BITMAP);
			break;
	}

	return FALSE;
}

LoadStatus_t CShpWizard::Show () {
	// Prepare title.
	string_t strTitle = "Import ";
	strTitle.append (m_strFileName);

	// Prepare pages.
	const size_t c_cPages = 4;
	PROPSHEETPAGE pshp [c_cPages];
	::memset (pshp, 0, sizeof (pshp));
	size_t cPage;
	for (cPage = 0; cPage < c_cPages; ++ cPage) {
		pshp [cPage].dwSize    = sizeof (PROPSHEETPAGE);
		pshp [cPage].dwFlags   = PSP_DEFAULT | PSP_USETITLE;
		pshp [cPage].hInstance = g_hResInst;
		pshp [cPage].lParam    = reinterpret_cast<LPARAM> (this);
		pshp [cPage].pszTitle  = strTitle.c_str ();
	}
	cPage = 0;

	pshp [cPage].pszTemplate = MAKEINTRESOURCE (IDD_IMPORT_SHP_TYPE);
	pshp [cPage].pfnDlgProc  = PageTypeProc;
	++ cPage;

	pshp [cPage].pszTemplate = MAKEINTRESOURCE (IDD_IMPORT_SHP_LABELS);
	pshp [cPage].pfnDlgProc  = PageLabelsProc;
	++ cPage;

	if (m_bEnableCoordinates) {
		pshp [cPage].pszTemplate = MAKEINTRESOURCE (IDD_IMPORT_SHP_UNITS);
		pshp [cPage].pfnDlgProc  = PageUnitsProc;
		++ cPage;
	}

	pshp [cPage].pszTemplate = MAKEINTRESOURCE (IDD_IMPORT_SHP_LEVELS);
	pshp [cPage].pfnDlgProc  = PageLevelsProc;
	++ cPage;

	// Prepare header.
	PROPSHEETHEADER psh;
	::memset (& psh, 0, sizeof (psh));
	psh.dwSize = sizeof (psh);
	psh.dwFlags = PSH_WIZARD | PSH_PROPTITLE | PSH_PROPSHEETPAGE;
	psh.hwndParent = g_hWnd;
	psh.hInstance = g_hResInst;
	psh.nPages = cPage;
	psh.nStartPage = 0;
	psh.ppsp = pshp;

	m_bCanceled = false;

	// Show the wizard.
	if (::PropertySheet (& psh) <= 0)
		return m_bCanceled ? lsCanceled : lsFailed;

	return lsOK;
}
