//
// GPSMapEdit
// (c) Konstantin Galichsky (kg@geopainting.com), 2008
//
// Split multi-element routable roads.
//

# include "StdAfx.h"
# include "Globals.h"
# include "UndoRedo.h"
# include "MapView.h"
# include "SplitRoutingRgns.h"

class CSplitRoutingRgns : public CEditAction {
  public:
	CSplitRoutingRgns () {}

	virtual void GetName (string_t & _str) const {_str.append ("Split multi-element routable roads");}
	virtual cu_t CanUndo () const {return cuBackup;}
	virtual bool Do ();
	virtual void Undo () {}
	virtual void Redo () {}
};

bool CSplitRoutingRgns::Do () {
	CCritSec::CLock lock (g_cs);

	return SplitRoutingRgns (g_map);
}

bool SplitRoutingRgns (map_t &_map) {
	rgns_t & rgns = _map.layers [rkPolyline].rgns;
	typedef std::list<rgn_t *> rgn_list_t;
	rgn_list_t suspiciousRgns;

	// Prepare list of rgns to process.
	for (rgns_t::iterator iRgn = rgns.begin (); iRgn != rgns.end (); ++ iRgn) {
		rgn_t & rgn = * iRgn;

		if (rgn.IsImmutable ())
			continue;

		if (! rgn.HasRGNodes ())
			continue;

		if (rgn.GetElements () < 2)
			continue;

		suspiciousRgns.push_back (& rgn);
	}

	// Do split.
	for (rgn_list_t::iterator ipRgn = suspiciousRgns.begin (); ipRgn != suspiciousRgns.end (); ++ ipRgn) {
		rgn_t & rgn = ** ipRgn;
		bool bFirst = true;

		for (size_t i = 0; i < rgn.elements.size (); ++ i) {
			rgn_element_t & element = rgn.elements [i];

			if (element.bDeleted)
				continue;

			if (bFirst) {
				bFirst = false;
				continue;
			}

			rgn_t & newRgn = _map.CreateRgn (rkPolyline);

			newRgn.elements.push_back (element);

			rgn_element_t & newElem = newRgn.elements.back ();

			// Fix restrictions.
			if (newElem.pExtra) {
				point_extras_t & extras = newElem.pExtra->PointExtras;

				for (size_t j = 0; j < extras.size (); ++ j) {
					rgnode_t * pNode = extras [j].GetRGNode ();

					if (pNode) {
						std::pair<rgrestrictions_t::iterator, rgrestrictions_t::iterator> bounds =
							_map.rgrestrictions.equal_range (pNode);

						for (rgrestrictions_t::iterator iRestr = bounds.first; iRestr != bounds.second; ++ iRestr) {
							rgrestriction_t & restr = iRestr->second;

							// Let's hope that this node doesn't belong to previous polylines of the rgn.
							if (restr.pFromRoad == & rgn)
								restr.pFromRoad = & newRgn;

							if (restr.pToRoad == & rgn)
								restr.pToRoad = & newRgn;
						}
					}
				}
			}

			element.bDeleted = true;

			// Type.
			newRgn.SetType (rgn.GetTypeDesc (), rgn.GetType (), rgn.GetKind ());
			// Label.
			newRgn.strLabel    = rgn.strLabel;
			// Comment.
			newRgn.strComment  = rgn.strComment;
			// End level.
			newRgn.uchEndLevel = rgn.uchEndLevel;
			// Extra.
			newRgn.pExtra      = rgn.pExtra;

			newRgn.RefreshAttributes (_map, false);
		}
	}

	_map.RefreshDrawOrder ();

	return true;
}

void SplitRoutingRgns () {
	PerformEditAction (new CSplitRoutingRgns ());
	Refresh ();
}
