//
// GPSMapEdit
// (c) Konstantin Galichsky (kg@geopainting.com), 2002-2009
// (c) Philipp Peshin (phil@mobilae.ru), 2003
//
// 'Create' tool.
//

# include "StdAfx.h"
# include "Resource.h"
# include "ToolCreate.h"
# include "Tool.h"
# include "Types.h"
# include "Position.h"
# include "Scale.h"
# include "Levels.h"
# include "MapView.h"
# include "Selection.h"
# include "UndoRedo.h"
# include "CreateObject.h"
# include "ChangeType.h"
# include "Rename.h"
# include "CombinedAction.h"
# include "RgnProperties.h"
# include "Vectorize.h"
# include "Bookmark.h"
# include "RemoveSelfIntersections.h"
# include "RemoveJitters.h"
# include "DrawNodes.h"
# include "DrawHelpers.h"
# include "Menu.h"
# include "Load.h"
# include "Distance.h"
# include "Course.h"
# include "Status.h"
# include "PrintUnits.h"
# include "Generalize.h"
# include "fast_floor.h"
# include "Options.h"
# include "Globals.h"
# include "Version.h"

static
bool _GetRectPoint (
	const point_t & _pt0, const point_t & _pt1, const point_t & _pt2, float _fCosY,
	point_t & _ptR
) {
	const float dx = (_pt1.x - _pt0.x)*_fCosY;
	const float dy = _pt1.y - _pt0.y;

	const float D2 = dx*dx + dy*dy;
	if (D2 < 1e-12) {
		_ptR = _pt0;
		return false;
	}

	const float dxa = (_pt2.x - _pt0.x)*_fCosY;
	const float dya = _pt2.y - _pt0.y;
	const float t = (dxa*dx + dya*dy)/D2;

	_ptR.x = _pt0.x + dx*t/_fCosY;
	_ptR.y = _pt0.y + dy*t;

	return true;
}

///////////////////////////////////////////////////////////////

class CCreateTool : public CTool {
  public:
	CCreateTool () : CTool (toolCreate), m_pLastNote (NULL) {}

	virtual bool Init ();

	virtual void OnActivate ();
	virtual void OnDeactivate ();
	virtual void OnRemoveMap ();

	virtual bool OnLDown (int _x, int _y, DWORD _dwKeys);
	virtual bool OnLUp   (int _x, int _y, DWORD _dwKeys);
	virtual void OnMove  (int _x, int _y, DWORD _dwKeys);
	virtual bool OnRUp   (int _x, int _y, DWORD _dwKeys);

	virtual void OnPaint (HDC _hDC, const RECT & _rect) const;

	virtual bool OnCtxMenu (int _x, int _y);
	virtual bool OnEsc ();
	virtual bool OnBack ();
	virtual bool OnEnter ();
	virtual bool OnIns ();
	virtual bool OnCommand (WORD _wID);
	virtual void OnSwitchSubTool ();

  public:
	points_t m_points;
	point_t m_pointLast;
	POINT m_ptLast;
	bool m_bStick;
	bool m_bAfterLDown;

	bool m_bLargeShift;
	bool m_bIsInVectorizeMode;
	POINT m_ptLastVPoint;

	std::vector<size_t> m_node_points;

	points_t m_pointsFinal;
	void _ReBuildFinalPoints (bool _bRectify, bool _bVectorize);
	bool _ReBuildFinalPoints_Circle (bool _bDisk, const point_t & _pt0, const point_t & _pt1);

	HCURSOR m_hcurCreatePoint;
	HCURSOR m_hcurCreatePolyline;
	HCURSOR m_hcurCreatePolylineV;
	HCURSOR m_hcurCreatePolygon;
	HCURSOR m_hcurCreatePolygonV;
	HCURSOR m_hcurCreateStrip;
	HCURSOR m_hcurCreateRect;
	HCURSOR m_hcurCreateDisk;
	HCURSOR m_hcurCreateCircle;
	HCURSOR m_hcurCreateBookmark;
	HCURSOR m_hcurCreateNote;

	void _SetCreateCursor (DWORD _dwKeys = 0);

	void _DoCreate (bool _bRectify);
	void _Create (RGN_KIND _Kind, points_t & _points);

	void _CreateBookmark (const point_t & _point);
	void _CreateNote ();

	void _AddPoint (int _x, int _y, bool _bRectify, bool _bVectorize);

	void _DrawPolyline (HDC _hDC) const;

	void _UpdateStatus ();

	static TC_KIND _GetNextCreateKind (TC_KIND _ck);

	bool _IsInRectifyMode   (DWORD _dwKeys) const;
	bool _IsInVectorizeMode (DWORD _dwKeys) const;

	rgn_t * m_pLastNote;
	DWORD m_dwLastNoteCreated;
};

DECLARE_TOOL (CCreateTool);

bool CCreateTool::_IsInRectifyMode   (DWORD _dwKeys) const {
	return (_dwKeys & MK_SHIFT) != 0;
}

bool CCreateTool::_IsInVectorizeMode (DWORD _dwKeys) const {
	const bool bCtrl = (_dwKeys & MK_CONTROL) != 0;
	const TC_KIND Kind = GetCreateKind ();
	if (Kind == tcPolylineV || Kind == tcPolygonV)
		return ! bCtrl;
	else
		return bCtrl;
}

bool CCreateTool::Init () {
	m_bStick      = false;
	m_bAfterLDown = false;
	m_bLargeShift        = false;
	m_bIsInVectorizeMode = false;

	m_hcurCreatePoint     = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_CursorCreatePoint));
	m_hcurCreatePolyline  = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_CursorCreatePolyline));
	m_hcurCreatePolylineV = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_CursorCreatePolylineV));
	m_hcurCreatePolygon   = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_CursorCreatePolygon));
	m_hcurCreatePolygonV  = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_CursorCreatePolygonV));
	m_hcurCreateStrip     = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_CursorCreateStrip));
	m_hcurCreateRect      = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_CursorCreateRect));
	m_hcurCreateDisk      = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_CursorCreateDisk));
	m_hcurCreateCircle    = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_CursorCreateCircle));
	m_hcurCreateBookmark  = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_CursorCreateBookmark));
	m_hcurCreateNote      = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_CursorCreateNote));

	m_pLastNote = NULL;

	return true;
}

void CCreateTool::_SetCreateCursor (DWORD _dwKeys) {
	switch (GetCreateKind ()) {
		case tcPoint:     CTool::SetPermanentCursor (m_hcurCreatePoint);     break;
		case tcPolyline:
		case tcPolylineV: CTool::SetPermanentCursor (_IsInVectorizeMode (_dwKeys) ? m_hcurCreatePolylineV : m_hcurCreatePolyline); break;
		case tcPolygon:
		case tcPolygonV:  CTool::SetPermanentCursor (_IsInVectorizeMode (_dwKeys) ? m_hcurCreatePolygonV  : m_hcurCreatePolygon);  break;
		case tcStrip:     CTool::SetPermanentCursor (m_hcurCreateStrip);     break;
		case tcRect:      CTool::SetPermanentCursor (m_hcurCreateRect);      break;
		case tcDisk:      CTool::SetPermanentCursor (m_hcurCreateDisk);      break;
		case tcCircle:    CTool::SetPermanentCursor (m_hcurCreateCircle);    break;
		case tcBookmark:  CTool::SetPermanentCursor (m_hcurCreateBookmark);  break;
		case tcNote:      CTool::SetPermanentCursor (m_hcurCreateNote);      break;
		default:
			assert (false);
	}
}

void CCreateTool::OnSwitchSubTool () {
	m_pLastNote = NULL;

	hidden_menu_items_t hmi;
	GetHiddenMenuItems (false, hmi);

	size_t cLoops = 0;

	TC_KIND CreateKind = GetCreateKind ();
	while (true) {
		CreateKind = _GetNextCreateKind (CreateKind);
		if (hmi.find (ID_TOOLS_CREATE_0 + CreateKind) == hmi.end ())
			break;

		// Protection from infinite loop.
		++ cLoops;
		if (cLoops > 20) {
			CTool::SetCurrentTool (toolDefault);
			return;
		}
	}
	SetCreateKind (CreateKind);
	UpdateMenu ();

	_SetCreateCursor ();
}

TC_KIND CCreateTool::_GetNextCreateKind (TC_KIND _ck) {
	switch (_ck) {
		case tcPoint:     return tcPolyline;  break;
		case tcPolyline:  return tcPolygon;   break;
		case tcPolygon:   return tcPolylineV; break;
		case tcPolylineV: return tcPolygonV;  break;
		case tcPolygonV:  return tcStrip;     break;
		case tcStrip:     return tcRect;      break;
		case tcRect:      return tcDisk;      break;
		case tcDisk:      return tcCircle;    break;
		case tcCircle:    return tcBookmark;  break;
# if defined (GME_AS) || defined (_DEBUG)
		case tcBookmark:  return tcNote;      break;
# else
		case tcBookmark:  return tcPoint;     break;
# endif
		case tcNote:      return tcPoint;     break;
		default:
			assert (false);
			return tcPolyline;
	}
}

RGN_KIND TCKind2RgnKind (TC_KIND _Kind) {
	switch (_Kind) {
		case tcPoint:     return rkPoint;
		case tcPolyline:  return rkPolyline;
		case tcPolylineV: return rkPolyline;
		case tcPolygon:   return rkPolygon;
		case tcPolygonV:  return rkPolygon;
		case tcStrip:     return rkPolygon;
		case tcRect:      return rkPolygon;
		case tcDisk:      return rkPolygon;
		case tcCircle:    return rkPolyline;
		case tcBookmark:  return rkPoint;
		case tcNote:      return rkPolyline;
		default:          assert (false); return rkMax;
	}
}

void CCreateTool::OnActivate () {
	_SetCreateCursor ();

	m_bStick = false;

	{
		hidden_menu_items_t hmi;
		GetHiddenMenuItems (false, hmi);

		if (hmi.find (ID_TOOLS_CREATE_0 + GetCreateKind ()) != hmi.end ())
			OnSwitchSubTool ();
	}

	m_pLastNote = NULL;
}

void CCreateTool::OnDeactivate () {
	SetStatusInfo (NULL);

	ResetVectorizer ();

	m_pLastNote = NULL;
}

void CCreateTool::OnRemoveMap () {
	m_points.clear ();
	m_node_points.clear ();
	m_pointsFinal.clear ();

	m_bStick = false;

	SetStatusInfo (NULL);

	ResetVectorizer ();

	m_pLastNote = NULL;
}

bool CCreateTool::OnLDown (int _x, int _y, DWORD _dwKeys) {
	m_bAfterLDown = true;
	return true;
}

bool CCreateTool::OnLUp (int _x, int _y, DWORD _dwKeys) {
	if (! m_bAfterLDown)
		return false;
	m_bAfterLDown = false;

	_SetCreateCursor (_dwKeys);

	_AddPoint (_x, _y, _IsInRectifyMode (_dwKeys), _IsInVectorizeMode (_dwKeys));

	const TC_KIND Kind = GetCreateKind ();
	if (Kind == tcNote) {
		_CreateNote ();
		assert (m_points.empty ());
		assert (m_pointsFinal.empty ());
	}

	return true;
}

void CCreateTool::_AddPoint (int _x, int _y, bool _bRectify, bool _bVectorize) {
	const TC_KIND Kind = GetCreateKind ();

	const size_t cLevel = GetZoomLevel (g_fScale);
	if (cLevel != 0 && m_points.empty ()) {
		if (GetSetting_NotifyCreatingNon0Level () && Kind != tcBookmark && Kind != tcNote)
			WarningMsgBox ("Please be aware of non-0 level is used to create the new object.");
	}
	if (cLevel != -1 && cLevel < g_map.Levels.size () && Kind != tcNote)
		// NOTE: this may change _x & _y.
		m_bStick = StickPointToNodeOrGrid (cLevel, _x, _y, m_pointLast);
	else {
		ScreenToPt (_x, _y, m_pointLast);
		m_bStick = false;
	}
	m_ptLast.x = _x;
	m_ptLast.y = _y;

	if (Kind == tcPoint) {
		points_t points (1);
		points [0] = m_pointLast;

		_Create (rkPoint, points);
	} else if (Kind == tcBookmark) {
		_CreateBookmark (m_pointLast);
	} else {
		if (m_points.empty () || ! (m_points.back () == m_pointLast)) {
			m_node_points.push_back (m_points.size ());
			m_points.push_back (m_pointLast);
			_ReBuildFinalPoints (_bRectify, _bVectorize);

			const TC_KIND Kind = GetCreateKind ();
			if ((_bRectify || _bVectorize) && (Kind == tcPolyline || Kind == tcPolygon || Kind == tcPolylineV || Kind == tcPolygonV))
				m_points = m_pointsFinal;
		}

		Redraw ();
		_UpdateStatus ();

		if (Kind == tcRect && m_points.size () == 3)
			_DoCreate (false);

		if ((Kind == tcDisk || Kind == tcCircle) && m_points.size () == 2)
			_DoCreate (false);
	}
}

void CCreateTool::OnMove (int _x, int _y, DWORD _dwKeys) {
	_SetCreateCursor (_dwKeys);

	const int iPrevLastX = m_ptLast.x;
	const int iPrevLastY = m_ptLast.y;

	const TC_KIND Kind = GetCreateKind ();

	const size_t cLevel = GetZoomLevel (g_fScale);
	if (cLevel != -1 && cLevel < g_map.Levels.size () && Kind != tcNote)
		// NOTE: this may change _x & _y.
		m_bStick = StickPointToNodeOrGrid (cLevel, _x, _y, m_pointLast);
	else {
		ScreenToPt (_x, _y, m_pointLast);
		m_bStick = false;
	}
	m_ptLast.x = _x;
	m_ptLast.y = _y;

	if (Kind == tcNote && m_bAfterLDown)
		_AddPoint (_x, _y, false, false);

	// Enable vectorization only after some threshold shift after CTRL pressing.
	const bool bIsInVectorizeMode = _IsInVectorizeMode (_dwKeys);
	if (bIsInVectorizeMode && ! m_bIsInVectorizeMode) {
		m_bLargeShift = false;
		m_ptLastVPoint.x = _x;
		m_ptLastVPoint.y = _y;
	}
	m_bIsInVectorizeMode = bIsInVectorizeMode;
	if (
		bIsInVectorizeMode == true && 
		m_bLargeShift == false &&
		(
			::labs (m_ptLastVPoint.x - _x) > ::GetSystemMetrics (SM_CXDOUBLECLK) ||
			::labs (m_ptLastVPoint.y - _y) > ::GetSystemMetrics (SM_CYDOUBLECLK)
		)
	)
		m_bLargeShift = true;

	_ReBuildFinalPoints (_IsInRectifyMode (_dwKeys), bIsInVectorizeMode && m_bLargeShift);

	if (! m_points.empty ())
		Redraw ();
	else if (iPrevLastX != m_ptLast.x || iPrevLastY != m_ptLast.y) {
		// Redraw stick mark only.
		RECT rect;

		if (iPrevLastX < m_ptLast.x) {
			rect.left  = iPrevLastX;
			rect.right = m_ptLast.x;
		} else {
			rect.right = iPrevLastX;
			rect.left  = m_ptLast.x;
		}

		if (iPrevLastY < m_ptLast.y) {
			rect.top = iPrevLastY;
			rect.bottom = m_ptLast.y;
		} else {
			rect.bottom = iPrevLastY;
			rect.top = m_ptLast.y;
		}
		::InflateRect (& rect, c_iCrossSize + 1, c_iCrossSize + 1);

		Redraw (& rect);
	}

	_UpdateStatus ();
}

void CCreateTool::OnPaint (HDC _hDC, const RECT & _rect) const {
	// Draw polyline of object to create.
	_DrawPolyline (_hDC);

	// Draw cross for stick point, if any.
	DrawStickPoint (_hDC, m_ptLast.x, m_ptLast.y, m_bStick);

 	if (! m_pointsFinal.empty () && g_bEnableSticking)
		HighlightNearestDrawnNodes (_hDC, m_ptLast.x, m_ptLast.y, 40, NULL);
}

void CCreateTool::_DrawPolyline (HDC _hDC) const {
	if (m_pointsFinal.empty ())
		return;

	const TC_KIND tcKind = GetCreateKind ();

	//
	// Draw polyline of object to create.
	//

	const size_t cPoints = m_pointsFinal.size ();
	POINT * const pPoints = reinterpret_cast<POINT *> (::alloca (sizeof (POINT)*cPoints));
	for (size_t cPoint = 0; cPoint < cPoints; ++ cPoint) {
		const point_t & point = m_pointsFinal [cPoint];
		pPoints [cPoint].x = XToScreen (point.x);
		pPoints [cPoint].y = YToScreen (point.y);
	}

	const HPEN hpenPolyline = ::CreatePen (PS_SOLID, GetNewObjectWidth (), GetNewObjectColor ());
	const HPEN hpen = (HPEN) ::SelectObject (_hDC, hpenPolyline);

	::SetROP2 (_hDC, R2_COPYPEN);

	// Show exact circle, if any.
	if (tcKind == tcDisk || tcKind == tcCircle) {
		if (cPoints >= 2) {
			const int x0 = XToScreen (m_points [0].x);
			const int y0 = YToScreen (m_points [0].y);
			const int x1 = XToScreen (m_pointLast.x);
			const int y1 = YToScreen (m_pointLast.y);

			const float fXc = (x0 + x1)*.5f;
			const float fYc = (y0 + y1)*.5f;
			const float fR = ::hypot (x1 - x0, y1 - y0)/2;
			::Arc (
				_hDC,
				fast_floor (fXc - fR +  .5f), fast_floor (fYc - fR +  .5f),
				fast_floor (fXc + fR + 1.5f), fast_floor (fYc + fR + 1.5f),
				x0, y0, x0, y0
			);
		}
	}

	::Polyline (_hDC, pPoints, cPoints);

	// Complete polygon, if any.
	if (tcKind == tcPolygon || tcKind == tcPolygonV || tcKind == tcStrip || tcKind == tcRect || tcKind == tcDisk) {
		::MoveToEx (_hDC, pPoints [cPoints - 1].x, pPoints [cPoints - 1].y, NULL);
		::LineTo   (_hDC, pPoints [0]          .x, pPoints [0]          .y);
	}

	::SelectObject (_hDC, hpen);
	::DeleteObject (hpenPolyline);
}

bool CCreateTool::OnRUp (int _x, int _y, DWORD _dwKeys) {
	return CCreateTool::OnCtxMenu (_x, _y);
}

bool CCreateTool::OnEsc () {
	if (m_points.empty ())
		return false;

	POINT p;
	::GetCursorPos (& p);
	::ScreenToClient (g_hMapWnd, & p);
	CCreateTool::OnCtxMenu (p.x, p.y);
	return true;
}

bool CCreateTool::OnCtxMenu (int _x, int _y) {
	if (g_bLoading)
		return false;

	const TC_KIND Kind = GetCreateKind ();
	if (Kind == tcPoint || Kind == tcBookmark || Kind == tcNote)
		return false;

	const size_t cPoints = m_points.size ();
	const bool bCanEnd = Kind == tcPolyline || Kind == tcPolylineV ? cPoints >= 2 : cPoints >= 3;
	const bool bCanEndRect = (Kind == tcPolygon || Kind == tcPolygonV) && cPoints >= 4; 

	const HMENU hPopupMenu = LoadMenu (IDM_POPUP);
	CMenuHolder hPopupMenu_Holder (hPopupMenu);
	HMENU hmenu = ::GetSubMenu (hPopupMenu, smpCreate);

	::EnableMenuItem (hmenu, ID_CREATE_REMOVE_LAST_NODE, (cPoints > 0 ? MF_ENABLED : MF_GRAYED) | MF_BYCOMMAND);
	::EnableMenuItem (hmenu, ID_CREATE_END,              (bCanEnd     ? MF_ENABLED : MF_GRAYED) | MF_BYCOMMAND);
	::EnableMenuItem (hmenu, ID_CREATE_END_RECT,         (bCanEndRect ? MF_ENABLED : MF_GRAYED) | MF_BYCOMMAND);
	::EnableMenuItem (hmenu, ID_CREATE_CONTINUE,         (cPoints > 0 ? MF_ENABLED : MF_GRAYED) | MF_BYCOMMAND);
//	::EnableMenuItem (hmenu, ID_CREATE_CANCEL,           (cPoints > 0 ? MF_ENABLED : MF_GRAYED) | MF_BYCOMMAND);

	if (Kind != tcPolygon)
		::DeleteMenu (hmenu, ID_CREATE_END_RECT, MF_BYCOMMAND);

	if (hmenu) {
		CTool::PrepareCtxMenu (hmenu);
		PopupMenu (hmenu);
		return true;
	} else
		return false;
}

bool CCreateTool::OnBack () {
	return CCreateTool::OnCommand (ID_CREATE_REMOVE_LAST_NODE);
}

bool CCreateTool::OnEnter () {
	return CCreateTool::OnCommand (ID_CREATE_END);
}

bool CCreateTool::OnIns () {
	_AddPoint (CTool::s_lX, CTool::s_lY, false, false);
	return true;
}

bool CCreateTool::OnCommand (WORD _wID) {
	switch (_wID) {
		case ID_CREATE_ADD_NODE:
			_AddPoint (CTool::s_lX, CTool::s_lY, false, false);
			return true;
			
		case ID_CREATE_REMOVE_LAST_NODE: {
			const TC_KIND Kind = GetCreateKind ();
			if (
				Kind == tcPolyline || Kind == tcPolylineV ||
				Kind == tcPolygon  || Kind == tcPolygonV  ||
				Kind == tcStrip    || Kind == tcRect
			) {
				if (! m_node_points.empty ()) {
					const size_t cLastIdx = m_node_points.back ();
					m_node_points.pop_back ();
					if (! m_points.empty () && cLastIdx < m_points.size ())
						m_points.resize (cLastIdx);
				}
				_ReBuildFinalPoints (false, false);
			} else if (Kind == tcDisk || Kind == tcCircle) {
				m_points.clear ();
				m_node_points.clear ();
				m_pointsFinal.clear ();
			}
			Redraw ();
			_UpdateStatus ();
			return true;
		}

		case ID_CREATE_END:
			m_pLastNote = NULL;
			_DoCreate (false);
			return true;

		case ID_CREATE_END_RECT:
			_DoCreate (true);
			return true;

		case ID_CREATE_CONTINUE:
			return true;

		case ID_CREATE_CANCEL:
			m_points.clear ();
			m_node_points.clear ();
			m_pointsFinal.clear ();
			Redraw ();
			_UpdateStatus ();
			return true;
	}

	return false;
}

void CCreateTool::_Create (RGN_KIND _Kind, points_t & _points) {
	g_Selection.clear ();

	const int nDepth = BeginCombinedAction ();

	rgn_t & rgn = CreateObject (_points, _Kind, GetZoomLevel (g_fScale));

	m_points.clear ();
	m_node_points.clear ();
	m_pointsFinal.clear ();
	Refresh ();

	// Show 'Change Type' dialog.
	const bool bOK = ShowChangeTypeForSelection (g_hWnd, _Kind);
	if (bOK)
		// Edit label.
		OnEditLabelStart ();

	string_t strAction = "create ";
	if (_Kind < rkMax)
		strAction.append (g_strKinds [_Kind]);
	EndCombinedAction (nDepth, strAction.c_str ());
	if (! bOK && _Kind == rkPoint) {
		UndoEditAction ();
		Refresh ();
	}
}

void CCreateTool::_CreateBookmark (const point_t & _point) {
	g_Selection.clear ();

	const int nDepth = BeginCombinedAction ();

	rgn_t & rgn = CreateBookmark (_point);

	m_points.clear ();
	m_node_points.clear ();
	m_pointsFinal.clear ();
	Refresh ();

	// Show 'Properties' dialog.
	bool bDeleteApplied = false;
	const bool bOK = ShowBookmarkProperties (g_hWnd, rgn, true, bDeleteApplied);

	EndCombinedAction (nDepth, "create bookmark");
	if (! bOK)
		UndoEditAction ();

	Refresh ();
}

void CCreateTool::_CreateNote () {
	const DWORD dwTime = ::GetTickCount ();
	if (m_pLastNote && ! m_pLastNote->IsDeleted () && dwTime - m_dwLastNoteCreated <= 3000) {
		// Append element to the existing note drawing.
		m_pLastNote->elements.push_back (rgn_element_t ());
		rgn_element_t & element = m_pLastNote->elements.back ();
		element.points.swap (m_points);
		element.cLevel = 0;

		m_pLastNote->RefreshAttributes (g_map, false);
	} else {
		// Create new note drawing.
		m_pLastNote = & CreateNoteDrawing (m_points);
		g_Selection.clear ();
	}
	m_dwLastNoteCreated = ::GetTickCount ();

	m_points.clear ();
	m_node_points.clear ();
	m_pointsFinal.clear ();
	Refresh ();
}

void CCreateTool::_DoCreate (bool _bRectify) {
	size_t cPoints = m_points.size ();
	const TC_KIND Kind = GetCreateKind ();
	if (
		(Kind == tcPolyline  && cPoints               < 2) ||
		(Kind == tcPolylineV && cPoints               < 2) ||
		(Kind == tcPolygon   && cPoints               < 3) ||
		(Kind == tcPolygonV  && cPoints               < 3) ||
		(Kind == tcStrip     && m_pointsFinal.size () < 3) ||
		(Kind == tcRect      && m_pointsFinal.size () < 2) ||
		(Kind == tcDisk      && m_pointsFinal.size () < 2) ||
		(Kind == tcCircle    && m_pointsFinal.size () < 2) ||
		(Kind == tcNote      && cPoints               < 2)
	)
		return;

	if (Kind == tcPolyline || Kind == tcPolygon || Kind == tcPolylineV || Kind == tcPolygonV) {
		// Rectify, if requested.
		if ((Kind == tcPolygon || Kind == tcPolygonV) && _bRectify && cPoints >= 4) {
			const float fCosY = ::cos (m_points [cPoints - 1].y*c_PI/180);
			point_t pt;
			_GetRectPoint (m_points [cPoints - 1], m_points [cPoints - 2], m_points [0], fCosY, pt);

			m_points [cPoints - 1] = pt;
		}

		const RGN_KIND RgnKind = TCKind2RgnKind (Kind);

		// Filter repeating points.
		RemoveRepeatingPoints (m_points, RgnKind == rkPolygon);
		cPoints = m_points.size ();

		// Hide last 'rubber' segment.
		m_pointsFinal = m_points;
		Redraw ();

		points_t points;
		if (RgnKind == rkPolygon) {
			const bool bRemoved = RemoveJitters (m_points, true);

			if (RemoveSelfIntersections (& * m_points.begin (), m_points.size (), points, true))
				points.swap (m_points);

			const float fCosY = m_points.empty () ? 1 : ::cos (m_points [0].y*c_PI/180);
			const float fPixelArea = (g_fScale*360/40e6)*(g_fScale/fCosY*360/40e6);
			while (RemoveSmallLoops (m_points, points, 10*fPixelArea))
				points.swap (m_points);
		}

		points.swap (m_points);
		_Create (RgnKind, points);

		assert (m_points.empty ());
		m_pointsFinal.clear ();
	} else if (Kind == tcStrip || Kind == tcRect) {
		points_t points;
		points.swap (m_pointsFinal);

		_Create (rkPolygon, points);

		assert (m_pointsFinal.empty ());
		m_points.clear ();
		m_node_points.clear ();
	} else if (Kind == tcDisk || Kind == tcCircle) {
		if (m_pointsFinal.size () > 3) {
			_Create (TCKind2RgnKind (Kind), m_pointsFinal);

			assert (m_pointsFinal.empty ());
			m_points.clear ();
			m_node_points.clear ();
		}
	} else if (Kind == tcNote) {
		_CreateNote ();
		assert (m_points.empty ());
		assert (m_pointsFinal.empty ());
	}
}

bool CCreateTool::_ReBuildFinalPoints_Circle (bool _bDisk, const point_t & _pt0, const point_t & _pt1) {
	const float fXc = (_pt0.x + _pt1.x)/2;
	const float fYc = (_pt0.y + _pt1.y)/2;
	const float fCosYc = ::cos (fYc*c_PI/180);
	const float fRy = ::sqrt (Distance2 (_pt0, _pt1, fCosYc*fCosYc))/2;
	const float fRx = fRy/fCosYc;

	const size_t cLevelIdx = GetZoomLevel (g_fScale);
	const float fLatGridInM = g_map.pTypeSet->pfGetLatGrid (g_map, cLevelIdx);
	const float fGrid = fLatGridInM != 0 ? fLatGridInM*360/40e6 : 360.f/(1 << 24);

//	const int Rx = fast_floor (fRx/fGrid + .5f);
//	const int Ry = fast_floor (fRy/fGrid + .5f);

	points_t pointsQ;

	const int N = fast_floor (::sqrt (fRx/fGrid) + .5);
	if (N == 0)
		return false;
	const float a_step = c_PI/(2*N);
	for (float a = 0; a <= c_PI/2 + a_step/2; a += a_step) {
//		const int x = fast_floor (fRx/fGrid*::sin (a) + .5f);
//		const int y = fast_floor (fRy/fGrid*::cos (a) + .5f);
//		pointsQ.push_back (point_t (x*fGrid, y*fGrid));
		pointsQ.push_back (point_t (fRx*::sin (a), fRy*::cos (a)));
	}
	const size_t cPointsQ = pointsQ.size ();
	if (cPointsQ == 0)
		return false;

	// Fill m_pointsFinal.
	m_pointsFinal.resize (4*cPointsQ + (_bDisk ? 0 : 1));
	for (size_t c = 0; c < cPointsQ; ++ c) {
		m_pointsFinal [                 c].x = fXc + pointsQ [c].x;
		m_pointsFinal [                 c].y = fYc + pointsQ [c].y;
		m_pointsFinal [2*cPointsQ - 1 - c].x = fXc + pointsQ [c].x;
		m_pointsFinal [2*cPointsQ - 1 - c].y = fYc - pointsQ [c].y;
		m_pointsFinal [2*cPointsQ     + c].x = fXc - pointsQ [c].x;
		m_pointsFinal [2*cPointsQ     + c].y = fYc - pointsQ [c].y;
		m_pointsFinal [4*cPointsQ - 1 - c].x = fXc - pointsQ [c].x;
		m_pointsFinal [4*cPointsQ - 1 - c].y = fYc + pointsQ [c].y;
	}
	if (! _bDisk)
		m_pointsFinal [cPointsQ*4] = m_pointsFinal [0];

	g_map.pTypeSet->pfRoundPoints (g_map, cLevelIdx, m_pointsFinal);

	// Remove repeating points.
	RemoveRepeatingPoints (m_pointsFinal, _bDisk);
	const size_t cPointsF = m_pointsFinal.size ();
	if (cPointsF < 4) {
		m_pointsFinal.clear ();
		return false;
	}
	if (_bDisk && cPointsF > 0 && m_pointsFinal [cPointsF - 1] == m_pointsFinal [0])
		m_pointsFinal.erase (m_pointsFinal.begin () + cPointsF - 1);

	// Make (most probable) roundabouts for right-side driving.
	std::reverse (m_pointsFinal.begin (), m_pointsFinal.end ());

	return true;
}

void CCreateTool::_ReBuildFinalPoints (bool _bRectify, bool _bVectorize) {
	const size_t cPoints = m_points.size ();
	const bool bAddP = cPoints > 0 && ! (m_points [cPoints - 1] == m_pointLast);
	const size_t cPointsP = cPoints + (bAddP ? 1 : 0);

	const TC_KIND Kind = GetCreateKind ();
	if (Kind == tcPoint || Kind == tcBookmark) {
		m_pointsFinal.clear ();
	} else if (Kind == tcPolyline || Kind == tcPolygon || Kind == tcPolylineV || Kind == tcPolygonV || Kind == tcNote) {
		m_pointsFinal = m_points;
		if (bAddP)
			m_pointsFinal.push_back (m_pointLast);
		assert (cPointsP == m_pointsFinal.size ());

		if (! _bVectorize && _bRectify && cPointsP >= 3) {
			const float fCosY = ::cos (m_pointsFinal [cPointsP - 1].y*c_PI/180);
			point_t pt;
			_GetRectPoint (m_pointsFinal [cPointsP - 3], m_pointsFinal [cPointsP - 2], m_pointsFinal [cPointsP - 1], fCosY, pt);

			m_pointsFinal [cPointsP - 2] = pt;
		}

		if (! _bRectify && _bVectorize && cPoints >= 1) {
			const point_t & ptPrev = m_points [cPoints - 1] == m_pointLast && cPoints >= 2 ? m_points [cPoints - 2] : m_points [cPoints - 1];

			points_t points;
			Vectorize (ptPrev, m_pointLast, points, false);

			if (! points.empty ()) {
				if (! m_pointsFinal.empty () && m_pointsFinal [m_pointsFinal.size () - 1] == m_pointLast)
					m_pointsFinal.pop_back ();
				m_pointsFinal.insert (m_pointsFinal.end (), points.begin (), points.end ());
				m_pointsFinal.push_back (m_pointLast);
			}
		}
	} else if (Kind == tcStrip || Kind == tcRect) {
		if (cPointsP < 3) {
			m_pointsFinal = m_points;
			if (bAddP)
				m_pointsFinal.push_back (m_pointLast);
		} else {
			// Copy all the points except first one.
			m_pointsFinal.clear ();
			if (! m_points.empty ())
				m_pointsFinal.assign (m_points.begin () + 1, m_points.end ());
			if (bAddP)
				m_pointsFinal.push_back (m_pointLast);

			const float fCosY = ::cos (c_PI*m_points [0].y/180);

			// Calculate width of the strip.
			const point_t pointsBase [3] = {m_points [0], m_points [1], m_points.size () >= 3 ? m_points [2] : m_pointLast};
			point_t ptNearest;
			float fH = ::sqrt (DistanceToSegment2 (pointsBase [1], pointsBase [2], pointsBase [0], fCosY, ptNearest));
			if (! IsRightOriented (pointsBase, ARRAY_SIZE (pointsBase)))
				fH = -fH;

			// Prepare second side of the strip.
			const size_t cPointsF = m_pointsFinal.size ();
			point_t * const pPoints2 = reinterpret_cast<point_t *> (alloca (sizeof (point_t)*(cPointsF)));
			const size_t cPoints2 = ShiftPolyline (& * m_pointsFinal.begin (), cPointsF, pPoints2, cPointsF, fCosY, fH);
			std::reverse (pPoints2, pPoints2 + cPoints2);

			// Join the sides of the strip.
			m_pointsFinal.insert (m_pointsFinal.end (), pPoints2, pPoints2 + cPoints2);

			points_t points;
			if (RemoveSelfIntersections (& * m_pointsFinal.begin (), m_pointsFinal.size (), points, true))
				points.swap (m_pointsFinal);
		}
	} else if (Kind == tcDisk || Kind == tcCircle) {
		if (cPointsP >= 2)
			_ReBuildFinalPoints_Circle (Kind == tcDisk, m_points [0], m_pointLast);
	} else if (Kind == tcNote) {
		m_pointsFinal = m_points;
		if (bAddP)
			m_pointsFinal.push_back (m_pointLast);
	}
}

void CCreateTool::_UpdateStatus () {
	const size_t cPoints = m_points.size ();

	const TC_KIND Kind = GetCreateKind ();
	if (Kind == tcPolyline || Kind == tcPolylineV || Kind == tcNote) {
		float fLength = 0;
		for (size_t c = 0; c + 1 < cPoints; ++ c) {
			const point_t & p0 = m_points [c];
			const point_t & p1 = m_points [c + 1];

			fLength += Distance (p0, p1);
		}

		float fLast = 0;
		float fCourse = 0;
		if (cPoints > 0) {
			fLast = Distance (m_points [cPoints - 1], m_pointLast);
			fLength += fLast;

			fCourse = Course (m_points [cPoints - 1], m_pointLast);
		}

		char strLength [16], strLengthUnits [16];
		PrintDistance (strLength, strLengthUnits, fLength);

		char strLast [16], strLastUnits [16];
		PrintDistance (strLast, strLastUnits, fLast);

		char strCourse [16];
		PrintAzimuth (strCourse, fCourse);

		char strOut [256];
		::sprintf (strOut, "Length: %s %s  Last segment: %s %s %s", strLength, strLengthUnits, strLast, strLastUnits, strCourse);
		SetStatusInfo (strOut);
	} else if (Kind == tcPolygon || Kind == tcPolygonV) {
		m_points.push_back (m_pointLast);
		const double fArea = GetArea (m_points)*40e6*40e6/360/360*::cos (m_pointLast.y/180*c_PI);
		m_points.pop_back ();

		char strArea [16], strAreaUnits [16];
		PrintArea (strArea, strAreaUnits, fArea);

		char strOut [256];
		::sprintf (strOut, "Area: %s %s", strArea, strAreaUnits);
		SetStatusInfo (strOut);
	} else if (Kind == tcStrip || Kind == tcRect) {
		float fWidth = 0, fLength = 0;
		if (cPoints == 1) {
			fWidth = Distance (m_pointLast, m_points [0]);
		} else if (cPoints > 1) {
			const float fCosY = ::cos (c_PI*m_points [0].y/180);

			// Calculate width of the strip.
			const point_t pointsBase [3] = {m_points [0], m_points [1], cPoints >= 3 ? m_points [2] : m_pointLast};
			point_t ptNearest;
			fWidth = ::sqrt (DistanceToSegment2 (pointsBase [1], pointsBase [2], pointsBase [0], fCosY, ptNearest));
			fWidth *= 40e6f/360;

			// Calculate length of the strip.
			for (size_t c = 1; c + 1 < cPoints; ++ c) {
				const point_t & p0 = m_points [c];
				const point_t & p1 = m_points [c + 1];

				fLength += Distance (p0, p1);
			}
			if (cPoints > 0)
				fLength += Distance (m_points [cPoints - 1], m_pointLast);
		}

		char strWidth [16], strWidthUnits [16];
		PrintDistance (strWidth, strWidthUnits, ::fabs (fWidth));

		char strLength [16], strLengthUnits [16];
		PrintDistance (strLength, strLengthUnits, fLength);

		char strOut [256];
		::sprintf (strOut, "Width: %s %s  Length: %s %s", strWidth, strWidthUnits, strLength, strLengthUnits);
		SetStatusInfo (strOut);
	} else if (Kind == tcDisk || Kind == tcCircle) {
		if (cPoints > 0) {
			const float fR = Distance (m_points [0], m_pointLast)/2;

			char strR [16], strDistUnits [16];
			PrintDistance (strR, strDistUnits, fR);

			char strOut [256];
			::sprintf (strOut, "Radius: %s %s", strR, strDistUnits);
			SetStatusInfo (strOut);
		}
	}
}

////////////////////////////////////////

static TC_KIND g_Kind = tcPolyline;

void SetCreateKind (TC_KIND _Kind) {
	if (g_Kind == _Kind)
		return;
	g_Kind = _Kind;

	CTool * const pToolCreate = CTool::GetCurrentTool ();
	if (pToolCreate && pToolCreate->m_ToolId == toolCreate) {
		pToolCreate->OnDeactivate ();
		pToolCreate->OnActivate ();
	}
	Redraw ();
}

TC_KIND GetCreateKind () {return g_Kind;}
