//
// GPSMapEdit
// (c) Konstantin Galichsky (kg@geopainting.com), 2008-2009
//
// 'Distance' tool.
//

# include "StdAfx.h"
# include "Resource.h"
# include "Tool.h"
# include "Map.h"
# include "MapView.h"
# include "Load.h"
# include "Menu.h"
# include "Position.h"
# include "Status.h"
# include "PrintUnits.h"
# include "Distance.h"
# include "Course.h"
# include "DrawHelpers.h"
# include "fast_floor.h"
# include "Globals.h"

class CDistanceTool : public CTool {
  public:
	CDistanceTool () : CTool (toolDistance), m_hcurDistance (NULL) {}

	virtual bool Init ();

	virtual void OnActivate ();
	virtual void OnDeactivate ();
	virtual void OnRemoveMap ();

	virtual bool OnLDown   (int _x, int _y, DWORD _dwKeys);
	virtual bool OnLUp     (int _x, int _y, DWORD _dwKeys);
	virtual void OnMove    (int _x, int _y, DWORD _dwKeys);
	virtual bool OnRUp     (int _x, int _y, DWORD _dwKeys);

	virtual void OnPaint (HDC _hDC, const RECT & _rect) const;

	virtual bool OnEsc ();
	virtual bool OnBack ();
	virtual bool OnEnter ();
	virtual bool OnIns ();
	virtual bool OnCtxMenu (int _x, int _y);
	virtual bool OnCommand (WORD _wID);

  private:
	HCURSOR m_hcurDistance;

	RECT m_rectPrev;

	point_t m_pointLast;
	POINT m_ptLast;
	bool m_bAfterLDown;
	bool m_bEnd;

	points_t m_points;

	void _AddPoint (int _x, int _y);
	void _ShowStatusInfo ();
};

DECLARE_TOOL (CDistanceTool);

static HCURSOR g_hcurDistance;

bool CDistanceTool::Init () {
	m_bAfterLDown = false;
	m_bEnd = false;

	m_hcurDistance = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_CursorDistance));

	::GetClientRect (g_hMapWnd, & m_rectPrev);

	return true;
}

void CDistanceTool::OnActivate () {
	SetPermanentCursor (m_hcurDistance);
	_ShowStatusInfo ();
}

void CDistanceTool::OnDeactivate () {
	SetStatusInfo (NULL);
}

void CDistanceTool::OnRemoveMap () {
	CDistanceTool::OnCommand (ID_DISTANCE_CLEAR);
}

bool CDistanceTool::OnLDown (int _x, int _y, DWORD _dwKeys) {
	m_bAfterLDown = true;
	return true;
}

bool CDistanceTool::OnLUp (int _x, int _y, DWORD _dwKeys) {
	if (! m_bAfterLDown)
		return false;
	m_bAfterLDown = false;

	_AddPoint (_x, _y);
	return true;
}

void CDistanceTool::OnMove (int _x, int _y, DWORD _dwKeys) {
	SetPermanentCursor (m_hcurDistance);

	ScreenToPt (_x, _y, m_pointLast);
	m_ptLast.x = _x;
	m_ptLast.y = _y;

	if (m_points.empty ())
		return;
	const point_t pointLast = m_points [m_points.size () - 1];
	const int x = XToScreen (pointLast.x);
	const int y = YToScreen (pointLast.y);

	if (x != m_ptLast.x || y != m_ptLast.y) {
		RECT rect;
		rect.left   = __min (x,  m_ptLast.x);
		rect.right  = __max (x,  m_ptLast.x);
		rect.top    = __min (y, m_ptLast.y);
		rect.bottom = __max (y, m_ptLast.y);

		RECT rect2;
		::UnionRect (& rect2, & rect, & m_rectPrev);
		::InflateRect (& rect2, c_iCrossSize + 2, c_iCrossSize + 2);

		Redraw (& rect2);

		m_rectPrev = rect;
	}

	_ShowStatusInfo ();
}

bool CDistanceTool::OnRUp (int _x, int _y, DWORD _dwKeys) {
	return CDistanceTool::OnCtxMenu (_x, _y);
}

bool CDistanceTool::OnCtxMenu (int _x, int _y) {
	if (g_bLoading)
		return false;

	const size_t cPoints = m_points.size ();

	const HMENU hPopupMenu = LoadMenu (IDM_POPUP);
	CMenuHolder hPopupMenu_Holder (hPopupMenu);

	const HMENU hMenu = ::GetSubMenu (hPopupMenu, smpDistance);
	assert (hMenu);

	::EnableMenuItem (hMenu, ID_DISTANCE_REMOVE_LAST_SEG, (! m_bEnd && cPoints > 0 ? MF_ENABLED : MF_GRAYED) | MF_BYCOMMAND);
	::EnableMenuItem (hMenu, ID_DISTANCE_END,             (! m_bEnd && cPoints > 1 ? MF_ENABLED : MF_GRAYED) | MF_BYCOMMAND);
	::EnableMenuItem (hMenu, ID_DISTANCE_CONTINUE,        (! m_bEnd && cPoints > 0 ? MF_ENABLED : MF_GRAYED) | MF_BYCOMMAND);
	::EnableMenuItem (hMenu, ID_DISTANCE_CLEAR,           (            cPoints > 0 ? MF_ENABLED : MF_GRAYED) | MF_BYCOMMAND);

	if (hMenu) {
		CTool::PrepareCtxMenu (hMenu);
		PopupMenu (hMenu);
		return true;
	} else
		return false;
}

bool CDistanceTool::OnBack () {
	return CDistanceTool::OnCommand (ID_DISTANCE_REMOVE_LAST_SEG);
}

bool CDistanceTool::OnEnter () {
	return CDistanceTool::OnCommand (ID_DISTANCE_END);
}

bool CDistanceTool::OnIns () {
	return CDistanceTool::OnCommand (ID_DISTANCE_ADD);
}

bool CDistanceTool::OnEsc () {
	return CDistanceTool::OnCommand (ID_DISTANCE_CLEAR);
}

bool CDistanceTool::OnCommand (WORD _wID) {
	switch (_wID) {
		case ID_DISTANCE_ADD:
			_AddPoint (CTool::s_lX, CTool::s_lY);
			return true;

		case ID_DISTANCE_REMOVE_LAST_SEG:
			if (! m_bEnd && ! m_points.empty ()) {
				m_points.pop_back ();

				_ShowStatusInfo ();
				Refresh ();
			}
			return true;

		case ID_DISTANCE_END:
			if (m_points.size () > 1) {
				m_bEnd = true;
				_ShowStatusInfo ();
				Refresh ();
			}
			break;

		case ID_DISTANCE_CONTINUE:
			break;

		case ID_DISTANCE_CLEAR:
			m_points.clear ();

			_ShowStatusInfo ();
			Refresh ();
			return true;
	}

	return false;
}

void CDistanceTool::_AddPoint (int _x, int _y) {
	if (m_bEnd)
		m_points.clear ();
	m_bEnd = false;

	ScreenToPt (_x, _y, m_pointLast);
	m_ptLast.x = _x;
	m_ptLast.y = _y;

	m_points.push_back (m_pointLast);

	_ShowStatusInfo ();

	Refresh ();
}

void CDistanceTool::OnPaint (HDC _hDC, const RECT & _rect) const {
	if (m_points.empty ())
		return;

	const size_t cPoints = m_points.size ();
	POINT * const pPoints = reinterpret_cast<POINT *> (::alloca (sizeof (POINT)*(cPoints + 1)));
	for (size_t cPoint = 0; cPoint < cPoints; ++ cPoint) {
		const point_t & point = m_points [cPoint];
		pPoints [cPoint].x = XToScreen (point.x);
		pPoints [cPoint].y = YToScreen (point.y);
	}

	pPoints [cPoints].x = m_ptLast.x;
	pPoints [cPoints].y = m_ptLast.y;

	::SelectObject (_hDC, g_hpenWhiteThick);
	::SetROP2  (_hDC, R2_XORPEN);
	::Polyline (_hDC, pPoints, m_bEnd ? cPoints : cPoints + 1);
}

void CDistanceTool::_ShowStatusInfo () {
	const size_t cPoints = m_points.size ();
	if (cPoints == 0) {
		SetStatusInfo (NULL);
		return;
	}

	// Calculate distance.
	double fDistance = 0;
	for (size_t c = 0; c + 1 < cPoints; ++ c) {
		const point_t & p0 = m_points [c];
		const point_t & p1 = m_points [c + 1];
		fDistance += Distance (p0, p1);
	}
	if (! m_bEnd && cPoints != 0)
		fDistance += Distance (m_points [cPoints - 1], m_pointLast);

	char strDistance [16], strDistUnits [16];
	PrintDistance (strDistance, strDistUnits, fDistance);

	// Calculate area.
	if (! m_bEnd)
		m_points.push_back (m_pointLast);
	const double fArea = GetArea (m_points)*40e6*40e6/360/360*::cos (m_pointLast.y/180*c_PI);
	if (! m_bEnd)
		m_points.pop_back ();

	char strArea [16], strAreaUnits [16];
	PrintArea (strArea, strAreaUnits, fArea);

	// Calculate course.
	const float fCourse = Course (m_points [0], m_bEnd ? m_points [cPoints - 1] : m_pointLast);
	char strCourse [16];
	PrintAzimuth (strCourse, fCourse);

	// Print report.
	char strOut [256];
	::sprintf (strOut, "Distance: %s %s  Area: %s %s  Course: %s (from the beginning point to cursor)", strDistance, strDistUnits, strArea, strAreaUnits, strCourse);
	SetStatusInfo (strOut);
}
