//
// GPSMapEdit
// (c) Konstantin Galichsky (kg@geopainting.com), 2002-2009
// (c) Philipp Peshin (phil@mobilae.ru), 2003
//
// 'Edit nodes' tool.
//

# include "StdAfx.h"
# include "Resource.h"
# include "ToolEditNodes.h"
# include "Tool.h"
# include "CombinedAction.h"
# include "AddPoint.h"
# include "DeletePoint.h"
# include "MovePoint.h"
# include "EditPoint.h"
# include "SetRGNode.h"
# include "SplitPolyline.h"
# include "ReversePoints.h"
# include "ClosePolyline.h"
# include "GeneralizeRgn.h"
# include "DeleteElement.h"
# include "RectifyPolygon.h"
# include "RefreshHoles.h"
# include "EditElement.h"
# include "NodeProperties.h"
# include "RgnProperties.h"
# include "View3D.h"
# include "Types.h"
# include "MapView.h"
# include "Load.h"
# include "WebMap.h"
# include "Selection.h"
# include "Position.h"
# include "DrawNodes.h"
# include "DrawHelpers.h"
# include "Menu.h"
# include "Label.h"
# include "Scale.h"
# include "Levels.h"
# include "Colors.h"
# include "Options.h"
# include "Globals.h"

enum {
	cnNode          = 0,
	cnNodeFirst     = 1,
	cnNodeCollision = 2,

	cnNodeMax
};

static CColors * g_pNodeColors;

static const COLORREF g_colNodes [cnNodeMax] = {
	c_colBlue,
	c_colGray,
	c_colRed,
};

static
void _InitColorsForNodes () {
	assert (g_pNodeColors == NULL);
	g_pNodeColors = new CColors;
	g_pNodeColors->LoadFromSetting ("NodeColors", g_colNodes, ARRAY_SIZE (g_colNodes));
}

static
void _DestroyColorsForNodes () {
	delete g_pNodeColors;
	g_pNodeColors = NULL;
}

bool AdjustColorsForNodes (HWND _hWnd) {
	assert (g_pNodeColors != NULL);

	const size_t cColors = g_pNodeColors->GetColors ();
	const char ** const strColorNames = (const char **) alloca (sizeof (char *)*cColors);

	fstring str [cnNodeMax];
	assert (cColors == ARRAY_SIZE (str));
    str [0].LoadFromResource (IDS_COL_NODE);
	str [1].LoadFromResource (IDS_COL_NODE_FIRST);
	str [2].LoadFromResource (IDS_COL_NODE_COLLISION);

	for (size_t c = 0; c < __min (cColors, ARRAY_SIZE (str)); ++ c)
		strColorNames [c] = str [c].c_str ();

	if (g_pNodeColors->Adjust (_hWnd, g_colNodes, strColorNames, cColors)) {
		g_pNodeColors->SaveToSetting ("NodeColors");
		return true;
	}
	return false;
}

//////////////////////////////////////////////////////////////////

class CNodesTool : public CTool {
  public:
	CNodesTool () :
		CTool (toolEditNodes),
		m_bStick (false),
		m_bNewPointMode (false),
		m_bFrameSelectionMode (false)
	{}

	// Override.
	virtual bool Init ();
	virtual void Destroy ();

	virtual void OnActivate ();
	virtual void OnDeactivate ();

	virtual bool OnLDown   (int _x, int _y, DWORD _dwKeys);
	virtual bool OnLUp     (int _x, int _y, DWORD _dwKeys);
	virtual void OnMove    (int _x, int _y, DWORD _dwKeys);
	virtual bool OnRUp     (int _x, int _y, DWORD _dwKeys);
	virtual bool OnLDblClk (int _x, int _y, DWORD _dwKeys);

	virtual bool OnEsc ();
	virtual bool OnDel ();
	virtual bool OnIns ();
	virtual bool OnArrow (int _dx, int _dy);

	virtual void OnPaint (HDC _hDC, const RECT & _rect) const;

	virtual bool OnCtxMenu (int _x, int _y);
	virtual bool OnCommand (WORD _wID);

  private:
	HCURSOR m_hcurMovePointMode;
	HCURSOR m_hcurMovePointDrag;
	HCURSOR m_hcurMovePoint;
	HCURSOR m_hcurInsertPoint;

	HPEN m_hpenRubber;
	HPEN m_hpenGroupSelection;

	bool m_bStick;
	point_t m_pointStick;

	bool m_bShowNewPoint;

	bool m_bNewPointMode;
	bool m_bFrameSelectionMode;

	// Found point/segment.
	rgn_elements_t::iterator m_iElement;
	size_t m_cPointIndex;
	point_t m_pointNew;
	bool _FindPoint   (int _x, int _y);
	bool _FindSegment (int _x, int _y, bool _bEnableNodes = false);

	void _DrawCrosses (HDC _hDC, const RECT & _rect) const;
	void _DrawRubber  (HDC _hDC) const;

	void _SelectPoint (int _x, int _y, bool _bMultiple, bool _bRight);

	void _MoveSelPoints (long _lXOffset, long _lYOffset);

	void _AddPoint    (rgn_t & _rgnSel, size_t _cElement, size_t _cPoint, point_t & _pointNew, bool _bClearSelPoints = true);
	void _InsertPoint (rgn_t & _rgnSel);

	rgn_t * _GetSelRgn () const {
		return g_Selection.size () == 1 && g_Selection [0].pLayer ? & * g_Selection [0].iRgn : NULL;
	}
};

DECLARE_TOOL (CNodesTool);

bool CNodesTool::Init () {
	_InitColorsForNodes ();

	m_bStick        = false;
	m_bShowNewPoint = false;

	m_hpenRubber         = ::CreatePen (PS_SOLID, 1, c_colGray);
	m_hpenGroupSelection = ::CreatePen (PS_SOLID, 1, c_colGray);

	m_hcurMovePointMode = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_CursorMovePointMode));
	m_hcurMovePoint     = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_CursorMovePoint));
	m_hcurMovePointDrag = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_CursorMovePointDrag));
	m_hcurInsertPoint   = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_CursorInsertPoint));

	return true;
}

void CNodesTool::Destroy () {
	if (m_hpenRubber)
		::DeleteObject (m_hpenRubber);
	if (m_hpenGroupSelection)
		::DeleteObject (m_hpenGroupSelection);

	_DestroyColorsForNodes ();
}

void CNodesTool::OnActivate () {
	SetPermanentCursor (m_hcurMovePointMode);

	m_bStick        = false;
	m_bShowNewPoint = false;

	SetSelPointsOffset (0, 0);
}

void CNodesTool::OnDeactivate () {
	ClearSelPoints ();
}

bool CNodesTool::_FindPoint (int _x, int _y) {
	if (g_Selection.size () != 1 || ! g_Selection [0].pLayer)
		return false;
	rgn_t & rgn = * g_Selection [0].iRgn;
	if (rgn.IsDeleted ())
		return false;

	bool bRet = false;

	for (rgn_elements_t::iterator i = rgn.elements.begin (); i != rgn.elements.end (); ++ i) {
		rgn_element_t & element = * i;
		if (element.bDeleted)
			continue;
		if (! LevelIsVisible (element.cLevel, rgn.uchEndLevel))
			continue;

		int min_d2 = -1;

		const points_t & points = element.points;
		const size_t cPoints = points.size ();
		for (size_t cPoint = 0; cPoint < cPoints; ++ cPoint) {
			const point_t & point = points [cPoint];

			const long lX = XToScreen (point.x);
			const long lY = YToScreen (point.y);
			if (::labs (_x - lX) <= c_fE_in_pixels && ::labs (_y - lY) <= c_fE_in_pixels) {
				const int d2 = (_x - lX)*(_x - lX) + (_y - lY)*(_y - lY);
				if (min_d2 == -1 || min_d2 > d2) {
					min_d2        = d2;
					m_iElement    = i;
					m_cPointIndex = cPoint;
					bRet          = true;
				}
			}
		}
	}

	return bRet;
}

bool CNodesTool::_FindSegment (int _x, int _y, bool _bEnableNodes) {
	if (g_Selection.size () != 1 || ! g_Selection [0].pLayer)
		return false;
	rgn_t & rgn = * g_Selection [0].iRgn;
	if (rgn.IsDeleted ())
		return false;

	const size_t cLevelIdx = GetZoomLevel (g_fScale);

	const float c_fE = c_fE_in_pixels/g_kY;
    point_t ptCurrent;
    ScreenToPt (_x, _y, ptCurrent);

	const float fCosY = ::cos (rgn.rectBound.MeanY ()*c_PI/180);
	const bool bIsPolygon = rgn.IsPolygon ();

	for (rgn_elements_t::iterator i = rgn.elements.begin (); i != rgn.elements.end (); ++ i) {
		rgn_element_t & element = * i;
		if (element.bDeleted)
			continue;
		if (! LevelIsVisible (element.cLevel, rgn.uchEndLevel))
			continue;

		const points_t & points = element.points;
        if (PolylineContainsPt (element.points, ptCurrent, c_fE, fCosY, bIsPolygon, & m_pointNew, & m_cPointIndex)) {
			if (_bEnableNodes == false) {
				// NOTE: The current point should be distant enough from the ends of the found segment.
				if (Distance2 (element.points [m_cPointIndex], ptCurrent, fCosY*fCosY) < 2*c_fE*c_fE)
					continue;
				const size_t cPointNext = m_cPointIndex + 1 < element.points.size () ? m_cPointIndex + 1 : 0;
				if (Distance2 (element.points [cPointNext],    ptCurrent, fCosY*fCosY) < 2*c_fE*c_fE)
					continue;
			}

			g_map.pTypeSet->pfRoundPoint (g_map, cLevelIdx, m_pointNew);
			m_iElement = i;
			return true;
        }
	}

	return false;
}

void CNodesTool::_SelectPoint (int _x, int _y, bool _bMultiple, bool _bRight) {
	const size_t cSelPoint = FindSelPoint (_x, _y);
	const bool bInsideSelected = cSelPoint != -1;
	if (bInsideSelected && _bRight && ! _bMultiple)
		// Nothing to do since the point is already selected.
		return;

	bool bPoint = bInsideSelected ? true : _FindPoint (_x, _y);

	bool bRedraw = false;

	if (! bPoint) {
		const rgn_t * const pSelRgn0 = _GetSelRgn ();
		const bool bAtSide = pSelRgn0 && pSelRgn0->IsPolygon () && _FindSegment (_x, _y);
		if (! bAtSide)
			SelectObjectAt (_x, _y, false, false, false, _bRight, false);

		if (! _bMultiple)
			ClearSelPoints ();

		const rgn_t * const pSelRgn = _GetSelRgn ();
		if (pSelRgn) {
			if (pSelRgn->IsImmutable ())
				// NOTE: don't edit waypoints and tracks.
				g_Selection.clear ();
			else
				// NOTE: Try again after object selection to select point on single click.
				bPoint = _FindPoint (_x, _y);
		} else {
			// NOTE: clicking out of any object.
		}
		bRedraw = true;
	}

	if (bPoint) {
		if (! _bRight) {
			SetPermanentCursor (m_hcurMovePointDrag);
			StartDND (_x, _y);
		}

		// Select/unselect point.
		if (cSelPoint != -1) {
			assert (cSelPoint < g_SelPoints.size ());
			const sel_point_t & sp = g_SelPoints [cSelPoint];
			SelectPoint (sp.pRgn, sp.cElement, sp.cPoint, _bMultiple);
		} else {
			rgn_t * const pSelRgn = _GetSelRgn ();
			if (pSelRgn) {
				if (pSelRgn->IsMutable ()) {
					SelectPoint (pSelRgn, m_iElement - pSelRgn->elements.begin (), m_cPointIndex, _bMultiple);
					bRedraw = true;
				}
			} else {
				if (_bMultiple)
					UnselectPoint (_x, _y);
			}
		}
	}

	if (bRedraw) {
		Redraw ();
		UpdateMenu ();
	}
}

bool CNodesTool::OnLDown (int _x, int _y, DWORD _dwKeys) {
	_SelectPoint (_x, _y, IsSelMultipleOn (_dwKeys), false);

	m_bFrameSelectionMode = false;
	if (FindSelPoint (_x, _y) == -1) {
		m_bNewPointMode = _FindSegment (_x, _y);
		if (! m_bNewPointMode)
			m_bFrameSelectionMode = true;
	}

	SetPermanentCursor (m_bFrameSelectionMode ? g_hArrowCursor : m_hcurMovePointDrag);
	StartDND (_x, _y);

	return true;
}

static
void _RedrawNewPoint (const point_t & _pointPrev, const point_t & _pointNext) {
	const int x0 = XToScreen (_pointPrev.x);
	const int y0 = YToScreen (_pointPrev.y);

	const int x1 = XToScreen (_pointNext.x);
	const int y1 = YToScreen (_pointNext.y);
		
	RECT rect;

	if (x0 < x1) {
		rect.left  = x0;
		rect.right = x1;
	} else {
		rect.right = x0;
		rect.left  = x1;
	}

	if (y0 < y1) {
		rect.top    = y0;
		rect.bottom = y1;
	} else {
		rect.bottom = y0;
		rect.top    = y1;
	}
	::InflateRect (& rect, c_iCrossSize + 1, c_iCrossSize + 1);

	Redraw (& rect);
}

bool CNodesTool::OnLUp (int _x, int _y, DWORD _dwKeys) {
	SetPermanentCursor (m_hcurMovePointMode);

	const bool bNewPointMode = m_bNewPointMode;
	if (m_bNewPointMode)
		m_bNewPointMode = false;

	if (! CTool::s_bDND)
		return false;
	StopDND ();
	SetSelPointsOffset (0, 0);

	//
	// Complete selection by frame.
	//
    if (m_bFrameSelectionMode) {
		m_bFrameSelectionMode = false;

        if (
			// TODO: allow shift over double-click area?
			CTool::s_lX_DND_Start != CTool::s_lX_DND ||
            CTool::s_lY_DND_Start != CTool::s_lY_DND
		) {
			// Prepare the selection rectangle.
	        rect_t rectSelection;

            point_t pt1;
            ScreenToPt (CTool::s_lX_DND_Start, CTool::s_lY_DND_Start, pt1);

			point_t pt2;
            ScreenToPt (CTool::s_lX_DND,       CTool::s_lY_DND,       pt2);

            if (pt1.x < pt2.x) {
                rectSelection.x0 = pt1.x;
                rectSelection.x1 = pt2.x;
            } else {
                rectSelection.x0 = pt2.x;
                rectSelection.x1 = pt1.x;
            }

            if (pt1.y < pt2.y) {
                rectSelection.y0 = pt1.y;
                rectSelection.y1 = pt2.y;
            } else {
                rectSelection.y0 = pt2.y;
                rectSelection.y1 = pt1.y;
            }

            SelectPointsInsideRect (rectSelection, IsSelMultipleOn (_dwKeys));
        }

		Redraw ();
		UpdateMenu ();

		return true;
    }

	//
	// Check if DND is real.
	//
	const long lXOffset = CTool::s_lX_DND - XToScreen (CTool::s_ptDND_Start.x);
	const long lYOffset = CTool::s_lY_DND - YToScreen (CTool::s_ptDND_Start.y);
	if (lXOffset == 0 && lYOffset == 0) {
		Redraw ();
		return true;
	}
    if (
		::labs (lXOffset) <= 2*::GetSystemMetrics (SM_CXDOUBLECLK) &&
		::labs (lYOffset) <= 2*::GetSystemMetrics (SM_CYDOUBLECLK) &&
		::GetTickCount () - CTool::s_dwTime_DND_Start <= ::GetDoubleClickTime ()
	) {
		Redraw ();
		return true;
	}

	if (bNewPointMode) {
		//
		// Complete new point creation.
		//
		rgn_t * const pSelRgn = _GetSelRgn ();
		if (pSelRgn) {
			const size_t cElement = m_iElement - pSelRgn->elements.begin ();
			_AddPoint (* pSelRgn, cElement, m_cPointIndex, m_pointNew, ! IsSelMultipleOn (_dwKeys));
		}
	} else {
		//
		// Apply changes (moving).
		//
		_MoveSelPoints (lXOffset, lYOffset);
	}

	return true;
}

void CNodesTool::_MoveSelPoints (long _lXOffset, long _lYOffset) {
	typedef std::set<rgnode_t *> setNodes_t;
	setNodes_t setNodesMoved;

	const int nDepth = BeginCombinedAction ();

	const size_t cSelPoints = g_SelPoints.size ();
	for (size_t c = 0; c < cSelPoints; ++ c) {
		sel_point_t & sp = g_SelPoints [c];
		if (! sp.IsValid ())
			continue;
		const rgn_element_t & element = sp.pRgn->elements [sp.cElement];
		const point_t & pt = element.points [sp.cPoint];

		// Avoid multiple moving of connected nodes.
		rgnode_t * const pRGNode = GetRGNode (* sp.pRgn, sp.cElement, sp.cPoint);
		if (pRGNode) {
			if (setNodesMoved.find (pRGNode) != setNodesMoved.end ())
				continue;
			setNodesMoved.insert (pRGNode);
		}

		sp.bSticked = false;

		int x = XToScreen (pt.x) + _lXOffset;
		int y = YToScreen (pt.y) + _lYOffset;
		point_t pointNew;
		if (element.cLevel < g_map.Levels.size ()) 
			// NOTE: this may change x & y.
			StickPointToNodeOrGrid (element.cLevel, x, y, pointNew, sp.pRgn);
		else
			ScreenToPt (x, y, pointNew);

		MovePoint (sp.pRgn, sp.cElement, sp.cPoint, pointNew);
	}

	EndCombinedAction (nDepth, GetStr (IDS_MOVE_NODES));

	Refresh ();
}

void CNodesTool::OnMove (int _x, int _y, DWORD _dwKeys) {
	if (! CTool::s_bDND) {
		if (FindSelPoint (_x, _y) != -1)
			SetPermanentCursor (m_hcurMovePoint);
		else {
			const rgn_t * const pSelRgn = _GetSelRgn ();
			if (pSelRgn && pSelRgn->IsMutable ()) {
				if (_FindPoint (_x, _y))
					SetPermanentCursor (m_hcurMovePoint);
				else {
					// NOTE: _FindSegment() affects m_pointNew.
					const point_t pointPrevNew = m_pointNew;
					if (_FindSegment (_x, _y)) {
						SetPermanentCursor (m_hcurInsertPoint);

						// Show "insert point" feedback.
						m_bShowNewPoint = true;
						m_bStick        = false;
						_RedrawNewPoint (pointPrevNew, m_pointNew);

						return;
					} else
						SetPermanentCursor (m_hcurMovePointMode);
				}
			} else
				SetPermanentCursor (m_hcurMovePointMode);
		}

		// Erase "insert point" feedback, if need.
		if (m_bShowNewPoint) {
			m_bShowNewPoint = false;
			_RedrawNewPoint (m_pointNew, m_pointNew);
		}
	} else {
		SetPermanentCursor (m_bFrameSelectionMode ? g_hArrowCursor : m_hcurMovePointDrag);

		if (! m_bFrameSelectionMode) {
			if (m_bNewPointMode) {
				const size_t cLevel = GetZoomLevel (g_fScale);
				if (cLevel != -1 && cLevel < g_map.Levels.size ())
					// NOTE: this may change _x & _y.
					m_bStick = StickPointToNodeOrGrid (cLevel, _x, _y, m_pointStick, _GetSelRgn ());
				else {
					ScreenToPt (_x, _y, m_pointStick);
					m_bStick = false;
				}

				m_pointNew = m_pointStick;
			} else {
				const long lXOffset = _x - XToScreen (CTool::s_ptDND_Start.x);
				const long lYOffset = _y - YToScreen (CTool::s_ptDND_Start.y);

				const size_t cSelPoints = g_SelPoints.size ();
				for (size_t c = 0; c < cSelPoints; ++ c) {
					sel_point_t & sp = g_SelPoints [c];
					if (! sp.IsValid ())
						continue;
					const rgn_element_t & element = sp.pRgn->elements [sp.cElement];
					const point_t & pt = element.points [sp.cPoint];

					sp.bSticked  = false;
					sp.nStickedX = XToScreen (pt.x) + lXOffset;
					sp.nStickedY = YToScreen (pt.y) + lYOffset;

					point_t pointNew;
					if (element.cLevel < g_map.Levels.size ()) 
						sp.bSticked = StickPointToNodeOrGrid (element.cLevel, sp.nStickedX, sp.nStickedY, pointNew, sp.pRgn);
				}

				SetSelPointsOffset (lXOffset, lYOffset);
			}
		}

		Redraw ();
    }
}

bool CNodesTool::OnArrow (int _dx, int _dy) {
	const size_t cSelPoints = g_SelPoints.size ();
	if (cSelPoints == 0)
		return false;

	_MoveSelPoints (_dx, _dy);
	return true;
}

void CNodesTool::OnPaint (HDC _hDC, const RECT & _rect) const {
	_DrawCrosses (_hDC, _rect);

 	if (! CTool::s_bDND)
		return;

	if (! m_bFrameSelectionMode) {
		_DrawRubber (_hDC);

		if (g_bEnableSticking) {
			const rgn_t * const pRgnSel0 = _GetSelRgn ();
			HighlightNearestDrawnNodes (_hDC, CTool::s_lX_DND, CTool::s_lY_DND, 40, pRgnSel0);
		}
	} else {
		// Draw selection frame.
		::SelectObject (_hDC, m_hpenGroupSelection);
		::SetROP2      (_hDC, R2_COPYPEN);

		POINT ptSelRect [5] = {
			{CTool::s_lX_DND_Start, CTool::s_lY_DND_Start},
			{CTool::s_lX_DND_Start, CTool::s_lY_DND},
			{CTool::s_lX_DND,       CTool::s_lY_DND},
			{CTool::s_lX_DND,       CTool::s_lY_DND_Start},
			{CTool::s_lX_DND_Start, CTool::s_lY_DND_Start}
		};
		::Polyline (_hDC, ptSelRect, 5);
	}
}

struct CCrossData {
	long lY;
	bool bFirst;
	bool bCollision;
	bool bSel;
	CCrossData (long _lY, bool _bFirst, bool _bSel) : lY (_lY), bFirst (_bFirst), bCollision (false), bSel (_bSel) {}
};

void CNodesTool::_DrawCrosses (HDC _hDC, const RECT & _rect) const {
	if (g_Selection.size () != 1 || ! g_Selection [0].pLayer || g_Selection [0].iRgn->IsImmutable ())
		return;
	const rgn_t & rgn = * g_Selection [0].iRgn;
	if (rgn.IsDeleted ())
		return;

	//
	// Prepare the list of unique pairs (X, Y) as map: X -> (Y, bFirst).
	//

	typedef std::multimap<long, CCrossData> crosses_t;
	crosses_t crosses;

	// Mark selected points to suppress drawing their crosses.
	const size_t cSelPoints = g_SelPoints.size ();
	for (size_t c = 0; c < cSelPoints; ++ c) {
		const sel_point_t & sp = g_SelPoints [c];
		if (! sp.IsValid ())
			continue;

		point_t point;
		sp.GetPoint (point);

		const long lX = XToScreen (point.x);
		if (lX < _rect.left || lX >= _rect.right)
			continue;
		const long lY = YToScreen (point.y);
		if (lY < _rect.top || lY >= _rect.bottom)
			continue;

		crosses.insert (crosses_t::value_type (lX, CCrossData (lY, false, true)));
	}

	// Resolve collisions of crosses.
	for (rgn_elements_t::const_iterator i = rgn.elements.begin (); i != rgn.elements.end (); ++ i) {
		const rgn_element_t & element = * i;
		if (element.bDeleted)
			continue;
		if (! LevelIsVisible (element.cLevel, rgn.uchEndLevel))
			continue;

		const points_t & points = element.points;
		const size_t cPoints = points.size ();
		for (size_t cPoint = 0; cPoint < cPoints; ++ cPoint) {
			const point_t & point = points [cPoint];

			const long lX = XToScreen (point.x);
			if (lX < _rect.left || lX >= _rect.right)
				continue;
			const long lY = YToScreen (point.y);
			if (lY < _rect.top || lY >= _rect.bottom)
				continue;

			const bool bFirst = cPoint == 0;

			const std::pair<crosses_t::iterator, crosses_t::iterator> cr = crosses.equal_range (lX);
			if (cr.first == cr.second)
				crosses.insert (crosses_t::value_type (lX, CCrossData (lY, bFirst, false)));
			else {
				bool bInsert = true;
				for (crosses_t::iterator ic = cr.first; ic != cr.second; ++ ic) {
					// Make sure that (X, Y) is unique pair.
					if (ic->second.lY == lY) {
						if (bFirst)
							// NOTE: first point has higher priority to draw.
							ic->second.bFirst = true;
						ic->second.bCollision = true;

						bInsert = false;
						break;
					}
				}

				if (bInsert)
					crosses.insert (cr.first, crosses_t::value_type (lX, CCrossData (lY, bFirst, false)));
			}
		}
	}

	//
	// Draw crosses.
	//

	const DWORD dwNodeShape = GetNodeShape ();

	const HPEN hpenNode          = ::CreatePen (PS_SOLID, 1, 0xFFFFFF & ~g_pNodeColors->GetColor (cnNode));
	const HPEN hpenNodeFirst     = ::CreatePen (PS_SOLID, 1, 0xFFFFFF & ~g_pNodeColors->GetColor (cnNodeFirst));
	const HPEN hpenNodeCollision = ::CreatePen (PS_SOLID, 1, 0xFFFFFF & ~g_pNodeColors->GetColor (cnNodeCollision));

	HPEN hpenPrev = NULL;
	for (crosses_t::iterator ic = crosses.begin (); ic != crosses.end (); ++ ic) {
		if (ic->second.bSel)
			continue;
		const long lX         = ic->first;
		const long lY         = ic->second.lY;
		const bool bFirst     = ic->second.bFirst;
		const bool bCollision = ic->second.bCollision;

		if (dwNodeShape != 1) {
			::SetROP2 (_hDC, R2_XORPEN);

			// Highlight the first point by different color.
			const HPEN hpen = bFirst ? hpenNodeFirst : (bCollision ? hpenNodeCollision : hpenNode);
			if (hpen != hpenPrev) {
				::SelectObject (_hDC, hpen);
				hpenPrev = hpen;
			}

			DrawCross (_hDC, lX, lY);
		} else {
			::SetROP2 (_hDC, R2_COPYPEN);

			// Highlight the first point by different color.
			const HPEN hpen = g_pNodeColors->GetPen (bFirst ? cnNodeFirst : (bCollision ? cnNodeCollision : cnNode));
			if (hpen != hpenPrev) {
				::SelectObject (_hDC, hpen);
				hpenPrev = hpen;
			}

			const int c_nNodeSize1 = 3;
			const int c_nNodeSize2 = 1;
			DrawCrossSerif (_hDC, lX, lY, c_nNodeSize1, c_nNodeSize2);
		}
	}

	::SelectObject (_hDC, g_hpenBlackThin);
	::DeleteObject (hpenNode);
	::DeleteObject (hpenNodeFirst);
	::DeleteObject (hpenNodeCollision);

	//
	// Draw stick point or new point.
	//
	if (! CTool::s_bDND && ! m_bShowNewPoint)
		return;

	if (m_bShowNewPoint) {
		const long x = XToScreen (m_pointNew.x);
		const long y = YToScreen (m_pointNew.y);

		DrawStickPoint (_hDC, x, y, m_bStick);
	}
}

void CNodesTool::_DrawRubber (HDC _hDC) const {
	::SelectObject (_hDC, m_hpenRubber);
	::SetROP2      (_hDC, R2_XORPEN);

	if (m_bNewPointMode) {
		rgn_t * const pSelRgn = _GetSelRgn ();
		if (pSelRgn) {
			const points_t & points = m_iElement->points;
			const size_t cPointNext = m_cPointIndex + 1 < points.size () ? m_cPointIndex + 1 : 0;
			const point_t & pt0 = points [m_cPointIndex < points.size () ? m_cPointIndex : 0];
			const point_t & pt1 = points [cPointNext];

			const long lX = XToScreen (m_pointNew.x);
			const long lY = YToScreen (m_pointNew.y);

			::MoveToEx (_hDC, XToScreen (pt0.x), YToScreen (pt0.y), NULL);
			::LineTo   (_hDC, lX,                lY);
			::LineTo   (_hDC, XToScreen (pt1.x), YToScreen (pt1.y));

			DrawSelPointAdded (_hDC, lX, lY);
		}
		return;
	}

	const size_t cSelPoints = g_SelPoints.size ();
	if (cSelPoints == 0)
		return;

	const long lXOffset = CTool::s_lX_DND - XToScreen (CTool::s_ptDND_Start.x);
	const long lYOffset = CTool::s_lY_DND - YToScreen (CTool::s_ptDND_Start.y);
	if (lXOffset == 0 && lYOffset == 0)
		return;

	rgn_t * pSelRgnPrev = NULL;
	size_t cSelElementPrev = -1;
	size_t cSelPointPrev = -1;
	long lSelXPrev = 0;
	long lSelYPrev = 0;
	bool bFirstPointIsSelected = false;
	for (size_t c = 0; c < cSelPoints; ++ c) {
		const sel_point_t & sp = g_SelPoints [c];
		if (! sp.IsValid ())
			continue;
		const rgn_element_t & element = sp.pRgn->elements [sp.cElement];
		const size_t cPoints = element.points.size ();

		point_t point;
		sp.GetPoint (point);
		const long lX = XToScreen (point.x) + lXOffset;
		const long lY = YToScreen (point.y) + lYOffset;

		const bool bPolygon = sp.pRgn->IsPolygon ();
		const bool bFirstInPolygon = bPolygon && sp.cPoint == 0;
		const bool bLastInPolygon  = bPolygon && sp.cPoint == cPoints - 1;

		bool bPrevContinue = false;
		if (sp.pRgn == pSelRgnPrev && sp.cElement == cSelElementPrev) {
			bPrevContinue = sp.cPoint == cSelPointPrev + 1;
		} else {
			// NOTE: start enum for selected points of next object.
			bFirstPointIsSelected = sp.cPoint == 0;
		}

		if (bPrevContinue) {
			// Rubber line for two adjacent selected points.
			::MoveToEx (_hDC, lSelXPrev, lSelYPrev, NULL);
			::LineTo   (_hDC, lX, lY);
		} else {
			// Rubber line from previous non-selected point to current selected one.
			size_t cPrevPoint = sp.cPoint - 1;

			if (bFirstInPolygon) {
				cPrevPoint = cPoints - 1;
				// Search for the last selected point in polygon.
				for (size_t c2 = c + 1; c2 < cSelPoints; ++ c2) {
					const sel_point_t & sp2 = g_SelPoints [c2];
					if (! sp2.IsValid ())
						continue;

					if (sp2.pRgn != sp.pRgn || sp2.cElement != sp.cElement)
						break;

					if (sp2.cPoint == cPoints - 1) {
						cPrevPoint = -1;
						break;
					}
				}
			}

			if (cPrevPoint < cPoints) {
				const point_t & point_p = element.points [cPrevPoint];
				const long lX_p = XToScreen (point_p.x);
				const long lY_p = YToScreen (point_p.y);

				::MoveToEx (_hDC, lX_p, lY_p, NULL);
				::LineTo   (_hDC, lX,   lY);
			}
		}

		pSelRgnPrev     = sp.pRgn;
		cSelElementPrev = sp.cElement;
		cSelPointPrev   = sp.cPoint;
		lSelXPrev       = lX;
		lSelYPrev       = lY;

		bool bNextContinue = false;
		if (c + 1 < cSelPoints) {
			const sel_point_t & spNext = g_SelPoints [c + 1];
			if (spNext.IsValid ())
				bNextContinue = spNext.pRgn == sp.pRgn && spNext.cElement == sp.cElement && spNext.cPoint == sp.cPoint + 1;
		}
		if (! bNextContinue) {
			// Rubber line from current selected point to next non-selected one.
			size_t cNextPoint = sp.cPoint + 1;
			if (bLastInPolygon && ! bFirstPointIsSelected)
				cNextPoint = 0;
			if (cNextPoint < cPoints) {
				const point_t & point_n = element.points [cNextPoint];
				const long lX_n = XToScreen (point_n.x);
				const long lY_n = YToScreen (point_n.y);

				::MoveToEx (_hDC, lX,   lY, NULL);
				::LineTo   (_hDC, lX_n, lY_n);
			}
		}

		if (bLastInPolygon && bFirstPointIsSelected) {
			const point_t & point_0 = element.points [0];
			const long lX_0 = XToScreen (point_0.x) + lXOffset;
			const long lY_0 = YToScreen (point_0.y) + lYOffset;

			::MoveToEx (_hDC, lX,   lY, NULL);
			::LineTo   (_hDC, lX_0, lY_0);
		}
	}
}

bool CNodesTool::OnRUp (int _x, int _y, DWORD _dwKeys) {
	_SelectPoint (_x, _y, IsSelMultipleOn (_dwKeys), true);
	return CNodesTool::OnCtxMenu (_x, _y);
}

bool CNodesTool::OnEsc () {
	const size_t cSelPoints = g_SelPoints.size ();
	if (cSelPoints == 0)
		return false;

	ClearSelPoints ();
	Redraw ();
	return true;
}

bool CNodesTool::OnDel () {
	return CNodesTool::OnCommand (ID_EDIT_DELETE_NODE);
}

enum {ID_EDIT_INSERT_NODE = 10000};

bool CNodesTool::OnIns () {
	return CNodesTool::OnCommand (ID_EDIT_INSERT_NODE);
}

bool CNodesTool::OnCtxMenu (int _x, int _y) {
	if (g_bLoading)
		return false;

	const HMENU hPopupMenu = LoadMenu (IDM_POPUP);
	CMenuHolder hPopupMenu_Holder (hPopupMenu);

	HMENU hMenu = NULL;

	const size_t cLevel = GetZoomLevel (g_fScale);
	const bool bCanGeneralize = cLevel != -1 && cLevel < g_map.Levels.size ();

	const size_t cSelPoints = g_SelPoints.size ();
	if (cSelPoints != 0) {
		hMenu = ::GetSubMenu (hPopupMenu, smpEditNode);
		::SetMenuDefaultItem (hMenu, ID_NODE_PROPERTIES, MF_BYCOMMAND);
		assert (hMenu);

		bool bIsPolyline    = false;
		bool bCanDelete     = false;
		bool bCanSplit      = false;
		bool bCanConnect    = false;
		bool bCanDisconnect = false;
		bool bCanViewIn3D   = false;
		for (size_t c = 0; c < cSelPoints; ++ c) {
			const sel_point_t & sp = g_SelPoints [c];
			if (! sp.IsValid () || sp.pRgn->IsImmutable ())
				continue;

			const size_t cPoints = sp.pRgn->elements [sp.cElement].points.size ();
			if (sp.pRgn->IsPolygon () ? cPoints >= 4 : cPoints >= 3)
				bCanDelete     = true;
			if (cSelPoints == 1 && sp.pRgn->IsPolyline () && 0 < sp.cPoint && sp.cPoint < cPoints - 1)
				bCanSplit      = true;
			if (CanConnectNodes    (sp.pRgn, sp.cElement, sp.cPoint))
				bCanConnect    = true;
			if (CanDisconnectNodes (sp.pRgn, sp.cElement, sp.cPoint))
				bCanDisconnect = true;
			if (sp.pRgn->IsPolyline ())
				bIsPolyline    = true;
			if (CanViewIn3D (* sp.pRgn))
				bCanViewIn3D = true;
		}

		::EnableMenuItem (hMenu, ID_EDIT_DELETE_NODE,      (cSelPoints != 0 && bCanDelete     ? MF_ENABLED : MF_GRAYED) | MF_BYCOMMAND);
		::EnableMenuItem (hMenu, ID_EDIT_DELETE_NODES,     (cSelPoints != 0 && bCanDelete     ? MF_ENABLED : MF_GRAYED) | MF_BYCOMMAND);
		::EnableMenuItem (hMenu, ID_EDIT_CONNECT_NODES,    (cSelPoints != 0 && bCanConnect    ? MF_ENABLED : MF_GRAYED) | MF_BYCOMMAND);
		::EnableMenuItem (hMenu, ID_EDIT_DISCONNECT_NODES, (cSelPoints != 0 && bCanDisconnect ? MF_ENABLED : MF_GRAYED) | MF_BYCOMMAND);
		::EnableMenuItem (hMenu, ID_EDIT_SPLIT_POLYLINE,   (cSelPoints == 1 && bCanSplit      ? MF_ENABLED : MF_GRAYED) | MF_BYCOMMAND);
		::EnableMenuItem (hMenu, ID_EDIT_GENERALIZE,       (bCanGeneralize                    ? MF_ENABLED : MF_GRAYED) | MF_BYCOMMAND);

		if (cSelPoints == 1)
			::DeleteMenu (hMenu, ID_EDIT_DELETE_NODES,     MF_BYCOMMAND);
		else {
			::DeleteMenu (hMenu, ID_EDIT_DELETE_NODE,      MF_BYCOMMAND);
			::DeleteMenu (hMenu, ID_EDIT_SPLIT_POLYLINE,   MF_BYCOMMAND);
			::DeleteMenu (hMenu, ID_EDIT_REVERSE_NODES,    MF_BYCOMMAND);
			::DeleteMenu (hMenu, ID_EDIT_GENERALIZE,       MF_BYCOMMAND);
			::DeleteMenu (hMenu, ID_NODE_PROPERTIES,       MF_BYCOMMAND);
		}

		if (! bIsPolyline) {
			::DeleteMenu (hMenu, ID_EDIT_CONNECT_NODES,    MF_BYCOMMAND);
			::DeleteMenu (hMenu, ID_EDIT_DISCONNECT_NODES, MF_BYCOMMAND);
			::DeleteMenu (hMenu, ID_EDIT_SPLIT_POLYLINE,   MF_BYCOMMAND);
			::DeleteMenu (hMenu, ID_EDIT_REVERSE_NODES,    MF_BYCOMMAND);
		}

		if (! bCanViewIn3D)
			::DeleteMenu (hMenu, ID_VIEW_IN_3D,            MF_BYCOMMAND);

		DeleteDuplicatedMenuSeparators (hMenu);
	} else {
		rgn_t * const pSelRgn = _GetSelRgn ();
		if (pSelRgn) {
			if (pSelRgn->IsImmutable ())
				// NOTE: waypoints and tracks are not editable.
				return false;
			const rgn_t & rgn = * pSelRgn;
			const bool bIsPolyline = rgn.IsPolyline ();
			const bool bIsPolygon  = rgn.IsPolygon ();

			if (_FindSegment (_x, _y)) {
				hMenu = ::GetSubMenu (hPopupMenu, smpEditLine);
				::SetMenuDefaultItem (hMenu, ID_LINE_PROPERTIES, MF_BYCOMMAND);
				assert (hMenu);

				const points_t & points = m_iElement->points;
				const size_t cPoints = points.size ();

				const bool bCanAdd       = true;
				const bool bCanClose     = IsClosePolylineAppliable ();
				const bool bCanReverse   = bIsPolyline;
				const bool bRefreshHoles = bIsPolygon && pSelRgn->elements.size () > 1;
				const bool bCanRectify   = bIsPolygon;

				::EnableMenuItem (hMenu, ID_EDIT_ADD_NODE,        (bCanAdd        ? MF_ENABLED : MF_GRAYED) | MF_BYCOMMAND);
				::EnableMenuItem (hMenu, ID_EDIT_REVERSE_NODES,   (bCanReverse    ? MF_ENABLED : MF_GRAYED) | MF_BYCOMMAND);
				::EnableMenuItem (hMenu, ID_EDIT_GENERALIZE,      (bCanGeneralize ? MF_ENABLED : MF_GRAYED) | MF_BYCOMMAND);
				::EnableMenuItem (hMenu, ID_EDIT_REFRESH_HOLES,   (bRefreshHoles  ? MF_ENABLED : MF_GRAYED) | MF_BYCOMMAND);
				::EnableMenuItem (hMenu, ID_EDIT_CLOSE_POLYLINE,  (bCanClose      ? MF_ENABLED : MF_GRAYED) | MF_BYCOMMAND);
				::EnableMenuItem (hMenu, ID_EDIT_POLYGON_RECTIFY, (bCanRectify    ? MF_ENABLED : MF_GRAYED) | MF_BYCOMMAND);

				if (! bIsPolyline) {
					::DeleteMenu (hMenu, ID_EDIT_REVERSE_NODES,   MF_BYCOMMAND);
					::DeleteMenu (hMenu, ID_EDIT_CLOSE_POLYLINE,  MF_BYCOMMAND);
				}
				if (! bIsPolygon) {
					::DeleteMenu (hMenu, ID_EDIT_REFRESH_HOLES,   MF_BYCOMMAND);
					::DeleteMenu (hMenu, ID_EDIT_POLYGON_RECTIFY, MF_BYCOMMAND);
				}
				if (! CanViewIn3D (rgn))
					::DeleteMenu (hMenu, ID_VIEW_IN_3D,           MF_BYCOMMAND);

				DeleteDuplicatedMenuSeparators (hMenu);
			} else if (bIsPolygon) {
				hMenu = ::GetSubMenu (hPopupMenu, smpEditPolygon);
				::SetMenuDefaultItem (hMenu, ID_RGN_PROPERTIES, MF_BYCOMMAND);

				const bool bRefreshHoles = bIsPolygon && pSelRgn->elements.size () > 1;
				::EnableMenuItem (hMenu, ID_EDIT_REFRESH_HOLES,   (bRefreshHoles  ? MF_ENABLED : MF_GRAYED) | MF_BYCOMMAND);
				::EnableMenuItem (hMenu, ID_EDIT_GENERALIZE,      (bCanGeneralize ? MF_ENABLED : MF_GRAYED) | MF_BYCOMMAND);
			}
		}
	}

	if (hMenu == NULL && g_bShowGoogleMaps) {
		hMenu = ::GetSubMenu (hPopupMenu, smpGoogle);
		::EnableMenuItem (hMenu, ID_GM_REFRESH_TILE, (CanRefreshWebMapTile () ? MF_ENABLED : MF_GRAYED) | MF_BYCOMMAND);
	}

	if (hMenu) {
		CTool::PrepareCtxMenu (hMenu);
		PopupMenu (hMenu);
		return true;
	} else
		return false;
}

bool CNodesTool::OnCommand (WORD _wID) {
	switch (_wID) {
		case ID_EDIT_ADD_NODE: {
			if (! _FindSegment (CTool::s_lX, CTool::s_lY))
				return false;
			rgn_t * const pSelRgn = _GetSelRgn ();
			if (pSelRgn == NULL || pSelRgn->IsPoint () || pSelRgn->IsImmutable ())
				return false;
			_AddPoint (* pSelRgn, m_iElement - pSelRgn->elements.begin (), m_cPointIndex, m_pointNew);
			return true;
		}

		case ID_EDIT_INSERT_NODE: {
			rgn_t * const pSelRgn = _GetSelRgn ();
			if (pSelRgn == NULL || pSelRgn->IsPoint () || pSelRgn->IsImmutable ())
				return false;
			_InsertPoint (* pSelRgn);
			return true;
		}

		case ID_EDIT_DELETE_NODES:
		case ID_EDIT_DELETE_NODE: {
			const size_t cSelPoints = g_SelPoints.size ();
			if (cSelPoints == 0)
				return false;

			const int nDepth = BeginCombinedAction ();
			for (int n = cSelPoints - 1; n >= 0; -- n) {
				const sel_point_t & sp = g_SelPoints [n];
				if (! sp.IsValid () || sp.pRgn->IsImmutable ())
					continue;

				if (sp.cElement < sp.pRgn->elements.size ()) {
					const size_t cPoints = sp.pRgn->elements [sp.cElement].points.size ();
					const size_t cMinPoints = sp.pRgn->IsPolygon () ? 3 : 2;
					if (cPoints <= cMinPoints) {
						DeleteElement (sp.pRgn, sp.cElement);
						continue;
					}
				}

				DeletePoint (sp.pRgn, sp.cElement, sp.cPoint);
			}
			EndCombinedAction (nDepth, "delete nodes");

			ClearSelPoints ();
			Refresh ();

			return true;
		}

		case ID_EDIT_CONNECT_NODES: {
			const size_t cSelPoints = g_SelPoints.size ();
			if (cSelPoints == 0)
				return false;

			for (size_t c = 0; c < cSelPoints; ++ c) {
				const sel_point_t & sp = g_SelPoints [c];
				if (! sp.IsValid () || sp.pRgn->IsImmutable ())
					continue;

				if (! CanConnectNodes (sp.pRgn, sp.cElement, sp.cPoint))
					continue;
				ConnectNodes (sp.pRgn, sp.cElement, sp.cPoint);
			}

			return true;
		}

		case ID_EDIT_DISCONNECT_NODES: {
			const size_t cSelPoints = g_SelPoints.size ();
			if (cSelPoints == 0)
				return false;

			for (size_t c = 0; c < cSelPoints; ++ c) {
				const sel_point_t & sp = g_SelPoints [c];
				if (! sp.IsValid () || sp.pRgn->IsImmutable ())
					continue;

				if (! CanDisconnectNodes (sp.pRgn, sp.cElement, sp.cPoint))
					continue;
				DisconnectNodes (sp.pRgn, sp.cElement, sp.cPoint);
			}

			return true;
		}

		case ID_EDIT_SPLIT_POLYLINE: {
			if (g_SelPoints.size () != 1)
				return false;
			const sel_point_t & sp = g_SelPoints [0];
			if (! sp.IsValid ())
				return false;

			if (! sp.pRgn->IsPolyline () || sp.pRgn->IsImmutable ())
				return false;
			if (sp.cPoint == 0 || sp.cPoint + 1 >= sp.pRgn->elements [sp.cElement].points.size ())
				return false;
			SplitPolyline (sp.pRgn, sp.cElement, sp.cPoint);

			ClearSelPoints ();
			Refresh ();
			return true;
		}

		case ID_EDIT_REVERSE_NODES: {
			if (! _FindSegment (CTool::s_lX, CTool::s_lY, true))
				return false;
			rgn_t * const pSelRgn = _GetSelRgn ();
			if (pSelRgn == NULL || ! pSelRgn->IsPolyline () || pSelRgn->IsImmutable ())
				return false;
			ReversePoints (pSelRgn, m_iElement - pSelRgn->elements.begin ());
			ClearSelPoints ();
			Refresh ();
			return true;
		}

		case ID_EDIT_CLOSE_POLYLINE: {
			if (! _FindSegment (CTool::s_lX, CTool::s_lY))
				return false;
			rgn_t * const pSelRgn = _GetSelRgn ();
			if (pSelRgn == NULL || ! IsClosePolylineAppliable () || pSelRgn->IsImmutable ())
				return false;
			ClosePolyline (pSelRgn, m_iElement - pSelRgn->elements.begin ());
			Refresh ();
			return true;
		}

		case ID_EDIT_POLYGON_RECTIFY: {
			rgn_t * const pSelRgn = _GetSelRgn ();
			if (pSelRgn == NULL || ! pSelRgn->IsPolygon ())
				return false;
			RectifyPolygon (* pSelRgn);
			Refresh ();
			return true;
		}

		case ID_EDIT_REFRESH_HOLES: {
			rgn_t * const pSelRgn = _GetSelRgn ();
			if (pSelRgn == NULL || ! pSelRgn->IsPolygon () || pSelRgn->elements.size () <= 1)
				return false;
			RefreshHoles (pSelRgn);
			Refresh ();
			return true;
		}

		case ID_EDIT_GENERALIZE: {
			rgn_t * const pSelRgn = _GetSelRgn ();
			if (pSelRgn == NULL || pSelRgn->IsPoint () || pSelRgn->IsImmutable ())
				return false;
			const size_t cLevel = GetZoomLevel (g_fScale);
			if (cLevel == -1 || cLevel >= g_map.Levels.size ())
				return false;
			GeneralizeRgnInLevel (* pSelRgn, cLevel);
			ClearSelPoints ();
			Refresh ();
			return true;
		}

		case ID_VIEW_IN_3D: {
			if (g_Selection.size () != 1 || ! g_Selection [0].pLayer || ! CanViewIn3D (* g_Selection [0].iRgn))
				return false;

			point_t pt0;
			ScreenToPt (CTool::s_lX, CTool::s_lY, pt0);

			ViewIn3D (pt0);
			return true;
		}

		case ID_NODE_PROPERTIES: {
			if (g_SelPoints.size () != 1)
				return false;
			const sel_point_t & sp = g_SelPoints [0];
			if (! sp.IsValid ())
				return false;
			ShowNodeProperties (g_hWnd, * sp.pRgn, sp.cElement, sp.cPoint);
			return true;
		}

		case ID_LINE_PROPERTIES: {
			if (! _FindSegment (CTool::s_lX, CTool::s_lY))
				return false;

			rgn_t * const pSelRgn = _GetSelRgn ();
			if (pSelRgn == NULL || pSelRgn->IsPoint () || pSelRgn->IsImmutable ())
				return false;

			const size_t cLevel = GetZoomLevel (g_fScale);
			if (cLevel >= g_map.Levels.size ())
				return false;

			if (EditElementPoints (g_hWnd, pSelRgn, m_iElement - pSelRgn->elements.begin ())) {
				ClearSelPoints ();
				Redraw ();
			}
			return true;
		}

		case ID_GM_REFRESH_TILE: RefreshWebMapTile (CTool::s_lX, CTool::s_lY); return true;
		case ID_GM_SHIFT:        ShiftWebMap (g_hWnd); return true;
		case ID_GM_RESET_SHIFT:  ResetWebMapShift ();  return true;
	}

	return false;
}

bool CNodesTool::OnLDblClk (int _x, int _y, DWORD _dwKeys) {
	if (g_Selection.empty ())
		return false;

	const find_info_t & Sel0 = g_Selection [0];
	if (! Sel0.pLayer || Sel0.iRgn->IsImmutable ())
		// NOTE: waypoints and tracks are not editable.
		return false;

	if (_FindPoint (_x, _y)) {
		CNodesTool::OnCommand (ID_NODE_PROPERTIES);
	} else if (_FindSegment (_x, _y)) {
		CNodesTool::OnCommand (ID_LINE_PROPERTIES);
	} else {
		if (g_Selection.size () == 1 && g_Selection [0].pLayer)
			ShowRgnProperties (g_hWnd, g_Selection [0]);
	}

	return true;
}

void CNodesTool::_AddPoint (rgn_t & _rgnSel, size_t _cElement, size_t _cPoint, point_t & _pointNew, bool _bClearSelPoints) {
	if (_bClearSelPoints)
		ClearSelPoints ();

	AddPoint (& _rgnSel, _cElement, _cPoint, _pointNew);

	if (! _bClearSelPoints) {
		// Re-number previously selected points.
		const size_t cSelPoints = g_SelPoints.size ();
		for (size_t c = 0; c < cSelPoints; ++ c) {
			sel_point_t & sp = g_SelPoints [c];
			if (! sp.IsValid ())
				continue;

			if (sp.pRgn == & _rgnSel && sp.cElement == _cElement && sp.cPoint > _cPoint)
				++ sp.cPoint;
		}
	}

	// Select the new point.
	SelectPoint (& _rgnSel, _cElement, _cPoint + 1, true);
	Redraw ();
}

void CNodesTool::_InsertPoint (rgn_t & _rgnSel) {
	const bool bPolygon = _rgnSel.IsPolygon ();

	const float fCosY = ::cos (_rgnSel.rectBound.MeanY ()*c_PI/180);

	point_t pt;
	const size_t cLevel = GetZoomLevel (g_fScale);
	if (cLevel != -1 && cLevel < g_map.Levels.size ()) {
		int x = CTool::s_lX;
		int y = CTool::s_lY;
		// NOTE: this may change x & y.
		StickPointToNodeOrGrid (cLevel, x, y, pt, & _rgnSel);
	} else
		ScreenToPt (CTool::s_lX, CTool::s_lY, m_pointStick);

	// Find nearest segment.
	float fD2Min = -1;
	size_t cNearestElement = -1;
	size_t cNearestPoint = -2;
	const size_t cElements = _rgnSel.elements.size ();
	for (size_t cElement = 0; cElement < cElements; ++ cElement) {
		rgn_element_t & elem = _rgnSel.elements [cElement];
		if (elem.bDeleted)
			continue;

		const size_t cPoints = elem.points.size ();
		const size_t cPointsE = cPoints + (bPolygon ? 1 : 0);
		for (size_t c = 0; c + 1 < cPointsE; ++ c) {
			const point_t & pt0 = elem.points [c];
			const point_t & pt1 = elem.points [c + 1 < cPoints ? c + 1 : 0];

			point_t ptNearest;
			const float fD2 = DistanceToSegment2 (pt0, pt1, pt, fCosY, ptNearest);
			if (! bPolygon && c == 0           && ptNearest == pt0)
				continue;
			if (! bPolygon && c + 2 == cPoints && ptNearest == pt1)
				continue;
			if (fD2Min > fD2 || fD2Min == -1) {
				fD2Min = fD2;
				cNearestElement = cElement;
				cNearestPoint = c;
			}
		}

		if (! bPolygon && cPoints != 0) {
			const float fD2_0 = Distance2 (elem.points [0],           pt, fCosY*fCosY);
			const float fD2_L = Distance2 (elem.points [cPoints - 1], pt, fCosY*fCosY);

			if (fD2Min > fD2_0 || fD2Min == -1) {
				fD2Min = fD2_0;
				cNearestElement = cElement;
				cNearestPoint = -1;
			} else if (fD2Min > fD2_L || fD2Min == -1) {
				fD2Min = fD2_L;
				cNearestElement = cElement;
				cNearestPoint = cPoints - 1;
			}
		}
	}

	if (cNearestElement != -1 && cNearestPoint != -2)
		_AddPoint (_rgnSel, cNearestElement, cNearestPoint, pt);
}
