//
// GPSMapEdit
// (c) Konstantin Galichsky (kg@geopainting.com), 2008-2009
//
// 'Rotate' tool.
//

# include "StdAfx.h"
# include "Resource.h"
# include "Tool.h"
# include "Selection.h"
# include "MapView.h"
# include "Load.h"
# include "Menu.h"
# include "RgnProperties.h"
# include "RotateObject.h"
# include "FlipObject.h"
# include "WebMap.h"
# include "Position.h"
# include "Scale.h"
# include "Status.h"
# include "Paste.h"
# include "fast_floor.h"
# include "Globals.h"

class CRotateTool : public CTool {
  public:
	CRotateTool () : CTool (toolRotate) {}

	virtual bool Init ();

	virtual void OnActivate ();
	virtual void OnDeactivate ();

	virtual bool OnLDown   (int _x, int _y, DWORD _dwKeys);
	virtual bool OnLUp     (int _x, int _y, DWORD _dwKeys);
	virtual void OnMove    (int _x, int _y, DWORD _dwKeys);
	virtual bool OnRUp     (int _x, int _y, DWORD _dwKeys);
	virtual bool OnLDblClk (int _x, int _y, DWORD _dwKeys);

	virtual bool OnCtxMenu (int _x, int _y);

	virtual bool OnCommand (WORD _wID);

  private:
	static float _GetTurnAngle (int _x, int _y);

	HCURSOR m_hcurRotateMode;
	HCURSOR m_hcurRotateDrag;
};

DECLARE_TOOL (CRotateTool);

bool CRotateTool::Init () {
	m_hcurRotateMode = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_CursorRotateMode));
	m_hcurRotateDrag = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_CursorRotate));

	return true;
}

void CRotateTool::OnActivate () {
	SetPermanentCursor (m_hcurRotateMode);
}

void CRotateTool::OnDeactivate () {
	SetSelectionTurn (0);
	SetStatusInfo (NULL);
}

bool CRotateTool::OnLDown (int _x, int _y, DWORD _dwKeys) {
	if (SelectObjectAt (_x, _y, false, IsSelElementOn (_dwKeys), false, false))
		SetPermanentCursor (m_hcurRotateDrag);

	// Make sure only 1 object or element is selected.
	const size_t cSelItems = g_Selection.size ();
	if (cSelItems > 1) {
		g_Selection.resize (1);
		Redraw ();
	}

	StartDND (_x, _y);

	return true;
}

bool CRotateTool::OnLUp (int _x, int _y, DWORD _dwKeys) {
	SetPermanentCursor (m_hcurRotateMode);

	if (! CTool::s_bDND)
		return false;
	StopDND ();

	SetSelectionTurn (0);
	SetStatusInfo (NULL);

	if (g_Selection.empty ())
		return true;
	const find_info_t & SelItem0 = g_Selection [0];
	if (! SelItem0.pLayer)
		return true;

	const float fTA = _GetTurnAngle (_x, _y);

	const rgn_t * const pSelectedRgn = & * SelItem0.iRgn;
	if (! pSelectedRgn || pSelectedRgn->IsImmutable () || pSelectedRgn->IsPoint ())
		return true;

	if (fTA != 0) {
		point_t pt0;
		GetSelItemCentroid (SelItem0, pt0);

		RotateObject (& * SelItem0.iRgn, SelItem0.cElementOrOrigin, fTA, pt0);
		Refresh ();
	} else
		Redraw ();

	return true;
}

float CRotateTool::_GetTurnAngle (int _x, int _y) {
	if (g_Selection.empty ())
		return 0;
	const find_info_t & SelItem0 = g_Selection [0];
	if (! SelItem0.pLayer)
		return 0;
	const rgn_t * const pSelectedRgn = & * SelItem0.iRgn;
	if (! pSelectedRgn || pSelectedRgn->IsImmutable () || pSelectedRgn->IsPoint ())
		return 0;

	point_t pt0;
	GetSelItemCentroid (SelItem0, pt0);
	const float fCosY = ::cos (pt0.y*c_PI/180.f);

	point_t ptStart;
	ScreenToPt (CTool::s_lX_DND_Start, CTool::s_lY_DND_Start, ptStart);
	point_t ptMove;
	ScreenToPt (_x,                    _y,                    ptMove);

	const float fAngleStart = ::atan2 (ptStart.y - pt0.y, (ptStart.x - pt0.x)*fCosY);
	const float fAngleMove  = ::atan2 (ptMove .y - pt0.y, (ptMove .x - pt0.x)*fCosY);

	return fAngleStart - fAngleMove;
}

void CRotateTool::OnMove (int _x, int _y, DWORD _dwKeys) {
	if (! s_bDND) {
		SetPermanentCursor (m_hcurRotateMode);
		SetStatusInfo (NULL);
		return;
	}

	const size_t cSelItems = g_Selection.size ();
	if (cSelItems == 0) {
		SetPermanentCursor (m_hcurRotateMode);
		return;
	}

	SetPermanentCursor (m_hcurRotateDrag);

	const float fTA = _GetTurnAngle (_x, _y);
	SetSelectionTurn (fTA);
	Redraw ();

	char strTurn [20];
	::sprintf (strTurn, "%.1f%s", ::fabs (fTA)*180/c_PI, g_strDegreeSign);
	SetStatusInfo (strTurn);
}

bool CRotateTool::OnRUp (int _x, int _y, DWORD _dwKeys) {
	SelectObjectAt (_x, _y, false, IsSelElementOn (_dwKeys), false, true);

	return CRotateTool::OnCtxMenu (_x, _y);
}

bool CRotateTool::OnCtxMenu (int _x, int _y) {
	if (g_bLoading)
		return false;

	const HMENU hPopupMenu = LoadMenu (IDM_POPUP);
	CMenuHolder hPopupMenu_Holder (hPopupMenu);

	HMENU hMenu = NULL;

	if (! g_Selection.empty ()) {
		size_t cObjects     = 0;
		size_t cPolygons    = 0;
		size_t cPolylines   = 0;

		const size_t cSelItems = g_Selection.size ();
		for (size_t cSelItem = 0; cSelItem < cSelItems; ++ cSelItem) {
			const find_info_t & SelItem = g_Selection [cSelItem];
			if (SelItem.pLayer && SelItem.iRgn->IsMutable ()) {
				++ cObjects;
				if (SelItem.cElementOrOrigin == -1) {
					if (SelItem.iRgn->IsPolyline ())
						++ cPolylines;
					if (SelItem.iRgn->IsPolygon ())
						++ cPolygons;
				}
			}
		}

		if (cPolygons + cPolylines > 0) {
			hMenu = ::GetSubMenu (hPopupMenu, smpRotate);
			assert (hMenu);
			::SetMenuDefaultItem (hMenu, ID_RGN_PROPERTIES, MF_BYCOMMAND);

			const UINT uiFlags = (g_Selection.size () == 1 && g_Selection [0].cElementOrOrigin == -1) ? MF_BYCOMMAND | MF_ENABLED : MF_BYCOMMAND | MF_GRAYED;
			::EnableMenuItem (hMenu, ID_RGN_PROPERTIES,  uiFlags);
		}
	}

	if (hMenu == NULL && g_bShowGoogleMaps) {
		hMenu = ::GetSubMenu (hPopupMenu, smpGoogle);
		::EnableMenuItem (hMenu, ID_GM_REFRESH_TILE, (CanRefreshWebMapTile () ? MF_ENABLED : MF_GRAYED) | MF_BYCOMMAND);
	}

	if (hMenu) {
		::EnableMenuItem (hMenu, ID_VIEW_BEST_FIT, MF_BYCOMMAND | (! g_Selection.empty () ? MF_ENABLED : MF_GRAYED));

		CTool::PrepareCtxMenu (hMenu);
		PopupMenu (hMenu);
		return true;
	} else
		return false;
}

bool CRotateTool::OnCommand (WORD _wID) {
	switch (_wID) {
		case ID_EDIT_FLIP_VERTICAL:
		case ID_EDIT_FLIP_HORIZONTAL: {
			if (g_Selection.empty ())
				break;
			const find_info_t & SelItem0 = g_Selection [0];
			if (SelItem0.pLayer == NULL)
				break;
			const rgn_t * const pSelectedRgn = & * SelItem0.iRgn;
			if (! pSelectedRgn || pSelectedRgn->IsImmutable () || pSelectedRgn->IsPoint ())
				break;

			FlipObject (& * SelItem0.iRgn, SelItem0.cElementOrOrigin, _wID == ID_EDIT_FLIP_VERTICAL);
			Refresh ();

			return true;
		}

		case ID_GM_REFRESH_TILE: RefreshWebMapTile (CTool::s_lX, CTool::s_lY); return true;
		case ID_GM_SHIFT:        ShiftWebMap (g_hWnd); return true;
		case ID_GM_RESET_SHIFT:  ResetWebMapShift ();  return true;
	}

	return false;
}

bool CRotateTool::OnLDblClk (int _x, int _y, DWORD _dwKeys) {
	SelectObjectAt (_x, _y, false, IsSelElementOn (_dwKeys), false, false);
	if (g_Selection.size () == 1) {
		if (g_Selection [0].pLayer) {
			ShowRgnProperties (g_hWnd, g_Selection [0]);
			return true;
		}
	}

	return false;
}
