//
// GPSMapEdit
// (c) Konstantin Galichsky (kg@geopainting.com), 2008-2009
//
// Track video window.
//

# include "StdAfx.h"
# include "Resource.h"
# include "TrackVideo.h"
# include "TrackVideoBookmark.h"
# include "Bookmark.h"
# include "BookmarkTypes.h"
//# include "ImageProcessing.h"
# include "MapView.h"
# include "Map.h"
# include "NewMap.h"
# include "UndoRedo.h"
# include "CombinedAction.h"
# include "BackupAction.h"
# include "Load.h"
# include "Position.h"
# include "Scale.h"
# include "PosCursor.h"
# include "Status.h"
# include "PrintUnits.h"
# include "Dialog.h"
# include "Menu.h"
# include "Icon.h"
# include "DIBitmap.h"
# include "Settings.h"
# include "Window.h"
# include "ListView.h"
# include "Threads.h"
# include "fast_floor.h"
# include "Wrappers.h"
# include "Globals.h"

class CRoadSignsDetector : public CDialog {
  public:
	CRoadSignsDetector (size_t _cChannels) :
		CDialog (IDD_RS_DETECTOR),
		m_cChannels (_cChannels)
	{}

	// Override.
	virtual bool OnInit (HWND _hwndFocus) {
		m_hChannels = GetDlgItem (IDC_CHANNELS);
		ListView_SetExtendedListViewStyleEx (m_hChannels, LVS_EX_FULLROWSELECT | LVS_EX_CHECKBOXES, LVS_EX_FULLROWSELECT | LVS_EX_CHECKBOXES);
		AddColumn (m_hChannels, 0, "#");
		
		m_dwChannels = GetSetting ("RSD_Channels", (DWORD) (1 << 2));

		for (size_t c = 0; c < m_cChannels; ++ c) {
			InsertItem (m_hChannels, c, c, "Channel #%d", c + 1);
			ListView_SetCheckState (m_hChannels, c, m_dwChannels & (1 << c));
		}
		AutoSizeColumns (m_hChannels);
		return true;
	}
	
	virtual void OnCommand (WORD _wCmd, WORD _wNotifyCode, HWND _hwndCtl) {
		switch (_wCmd) {
			case IDCANCEL:
				EndDialog (0);
				break;

			case IDC_START: {
				m_dwChannels = 0;
				for (size_t c = 0; c < m_cChannels; ++ c) {
					const bool bChecked = ListView_GetCheckState (m_hChannels, c) != FALSE;
					if (bChecked)
						m_dwChannels |= (1 << c);
				}
				SaveSetting ("RSD_Channels", m_dwChannels);
				
				EndDialog (1);
				break;
			}
		}
	}

	DWORD GetChannelsToScan () const {return m_dwChannels;}

  private:
	HWND m_hChannels;  
	const size_t m_cChannels;
	DWORD m_dwChannels;	
};

////////////////////////////////////////////////////////////////

class CTrackVideo;

class CTrackVideoChannel : public CDialog {
  public:
	CTrackVideoChannel () :
		CDialog (IDD_TRACK_VIDEO_CHANNEL),
		m_pTrackVideo (NULL),
		m_cChannel (0)
	{}

	void Init (CTrackVideo * _pTrackVideo, size_t _cChannel) {
		m_pTrackVideo = _pTrackVideo;
		m_cChannel = _cChannel;
	}

	// Override.
	virtual bool OnInit (HWND _hwndFocus) {return true;}
	virtual void OnSize (DWORD _dwResizeFlag, WORD _wWidth,	WORD _wHeight);
	virtual void OnMove (int _x, int _y);
	virtual void OnClose () {Hide ();}
	virtual	void OnPaint (const PAINTSTRUCT & _ps) const;

  private:
	CTrackVideo * m_pTrackVideo;
	size_t m_cChannel;

	CMemDC m_MemDC;
};

void CTrackVideoChannel::OnSize (DWORD _dwResizeFlag, WORD _wWidth, WORD _wHeight) {
	if (_dwResizeFlag != SIZE_RESTORED)
		return;

	// Save position.
	RECT rect;
	::GetWindowRect (m_hWnd, & rect);

	char strWidth [32], strHeight [32];
	::sprintf (strWidth,  "TrackVideoChannel%d_Width",  m_cChannel);
	::sprintf (strHeight, "TrackVideoChannel%d_Height", m_cChannel);
	SaveSetting (strWidth,  (DWORD) (rect.right - rect.left));
	SaveSetting (strHeight, (DWORD) (rect.bottom - rect.top));

	// Resize memory DC buffer.
	RECT rectClient;
	::GetClientRect (m_hWnd, & rectClient);
	const int nW = rectClient.right - rectClient.left;
	const int nH = rectClient.bottom - rectClient.top;

	m_MemDC.Resize (CClientDC (NULL), nW, nH);

	// Redraw.
	::InvalidateRect (m_hWnd, NULL, FALSE);
	::UpdateWindow (m_hWnd);
}

void CTrackVideoChannel::OnMove (int _x, int _y) {
	RECT rect;
	::GetWindowRect (m_hWnd, & rect);
	if (rect.left < 0)
		rect.left = 0;
	if (rect.top < 0)
		rect.top = 0;

	char strX [32], strY [32];
	::sprintf (strX, "TrackVideoChannel%d_X", m_cChannel);
	::sprintf (strY, "TrackVideoChannel%d_Y", m_cChannel);
	SaveSetting (strX, (DWORD) rect.left);
	SaveSetting (strY, (DWORD) rect.top);
}

////////////////////////////////////////////////////

const UINT c_uiTimerRender = 10;

class CTrackVideo : public CDialog {
  public:
	CTrackVideo () :
		CDialog (IDD_TRACK_VIDEO),
		m_pRenderer (NULL),
		m_pBookmarks (NULL),
		m_nSpeed (0),
		m_hView (NULL),
		m_hToolbarWnd (NULL),
		m_hProgress (NULL),
		m_hChannel (NULL),
		m_hInfo (NULL),
		m_lDW (0), m_lDH (0),
		m_lViewW (0), m_lViewH (0),
		m_nFPS (1),
		m_bCursorShow (false),
		m_cChannel (0),
		m_bAllChannels (false),
		m_pRgn (NULL),
		m_uiTimerID (0),
		m_lTimerMessagesMissed (0),
		m_lTimerMessagesInQueue (0),
		m_bProcessing (false),
		m_bStop (false),
		m_hProcessingThread (NULL),
		m_dwChannelsToScan (0)
	{}

	static CTrackVideo * g_pTrackVideo;

	void GoFrame (const rgn_t & _rgn, CTrackVideoRendererBase * _pRenderer, size_t _cTrackPointVideoData0, size_t _cTrackPointVideoData1);
	void GoFrame (const rgn_t & _rgn, CTrackVideoRendererBase * _pRenderer, DWORD _dwProgressPrec);

	void Reset () {
		StopProcessing ();

		m_pRgn = NULL;
		m_strTrackFileName.clear ();

		m_pRenderer  = NULL;
		m_pBookmarks = NULL;
		_UpdateTab ();

		_SetSpeed (0);
		_UpdateTimerRender ();
		_Update ();

		ShowVideoPosCursor (false);
		m_bCursorShow = false;
	}

	void Clean () {
		if (m_pRgn && m_pRgn->IsDeleted ())
			Reset ();
		else {
			if (m_pBookmarks)
				m_pBookmarks->Update ();
		}
	}

	// Override.
	virtual bool OnInit (HWND _hwndFocus);
	virtual void OnClose () {StopProcessing ();}
	virtual void OnCommand (WORD _wCmd, WORD _wNotifyCode, HWND _hwndCtl);
	virtual void OnHScroll (WORD _wRequest, WORD _wPos, HWND _hwndCtl);
	virtual void OnSize (DWORD _dwResizeFlag, WORD _wWidth, WORD _wHeight);
	virtual void OnMove (int _x, int _y);
	virtual bool OnNotify (int _idCtrl, NMHDR * _pInfo);
	virtual INT_PTR OnMessage (UINT _uMsg, WPARAM _wParam, LPARAM _lParam);

	void OnMouseMove_View   (int _x, int _y);
	void OnLButtonDown_View (int _x, int _y);
	void OnLButtonUp_View   (int _x, int _y);
	void OnMouseWheel_View (int _nDelta);
	bool OnSetCursor_View ();

	void TogglePlay ();

	void PaintFrame (const PAINTSTRUCT & _ps) const;

  private:
	friend class CTrackVideoChannel;

	HWND m_hToolbarWnd;
	HWND m_hProgress;
	HWND m_hInfo;
	HWND m_hChannel;
	HWND m_hView;
	long m_lDW, m_lDH;
	long m_lViewW, m_lViewH;

	point_t m_ptCursor;
	float   m_fCursorHeading;
	bool    m_bCursorShow;

	const rgn_t * m_pRgn;
	string_t m_strTrackFileName;

	size_t m_cChannel;
	bool m_bAllChannels;

	TIME_ZONE_INFORMATION m_tzi;

	CMemDC m_MemDC;

	CTrackVideoRendererBase * m_pRenderer;
	int m_nFPS;
	int m_nSpeed;
	void _SetSpeed (int _nSpeed);
	void _PauseAndUpdate ();

	void _UpdateTimerRender (UINT _uiLost = 0);
	void _OnTimerRender (size_t _cTimes);
	static void CALLBACK _TimeProc (UINT _uTimerID, UINT _uMsg, DWORD_PTR _dwUser, DWORD_PTR _dw1, DWORD_PTR _dw2);
	UINT m_uiTimerID;
	long m_lTimerMessagesInQueue;
	long m_lTimerMessagesMissed;

	void _Update ();
	void _UpdateToolBar ();
	void _UpdateInfo ();
	void _UpdateTab ();
	void _UpdateFrame ();

	bool _IsTButtonChecked (UINT _uiCmd)        {return ::SendMessage (m_hToolbarWnd, TB_ISBUTTONCHECKED, _uiCmd, 0) != 0;}
	void _CheckTBButton    (UINT _uiCmd, bool _bCheck) {::SendMessage (m_hToolbarWnd, TB_CHECKBUTTON,     _uiCmd, MAKELONG (_bCheck ? TRUE : FALSE, 0));}
	void _EnableTBButton   (UINT _uiCmd, bool _bCheck) {::SendMessage (m_hToolbarWnd, TB_ENABLEBUTTON,    _uiCmd, MAKELONG (_bCheck ? TRUE : FALSE, 0));}

	void _AutoScrollMap ();

	std::vector<CTrackVideoChannel *> m_vecUndockedWindows;
	void _DestroyUndockedWindows ();
	bool m_bDocked;
	void _Dock ();
	void _Undock ();

	tv_bookmarks_t * m_pBookmarks;
	void DrawPresenceOfBookmarks (HDC _hDC, int _nW, int _nH) const;
	rgn_t * _AddBookmark ();

	bool m_bProcessing;
	bool m_bStop;
	HANDLE m_hProcessingThread;
	typedef void (CTrackVideo::* pfProcess_t) ();
	pfProcess_t m_pfProcess;
	void _ProcessInParallel (const char * _strStatus, pfProcess_t _pfProcess);
	static DWORD WINAPI _BookmarkProcessProc (LPVOID _pParams);
	void StopProcessing ();

	void _DoBookmarkVoices ();
	void _DoBookmarkRoadSigns ();
	
	DWORD m_dwChannelsToScan;
};

CTrackVideo * CTrackVideo::g_pTrackVideo;

///////////////////////////////////////////////////////////////////

static WNDPROC g_wpViewProc = NULL;

static
LRESULT CALLBACK _TrackVideoProc (HWND _hWnd, UINT _uMsg, WPARAM _wParam, LPARAM _lParam) {
	if (CTrackVideo::g_pTrackVideo) {
		switch (_uMsg) {
			case WM_GETDLGCODE:
				return DLGC_WANTALLKEYS | DLGC_WANTARROWS | DLGC_WANTCHARS;

			case WM_RBUTTONDOWN:
			case WM_RBUTTONUP:
				::SetFocus (_hWnd);
				// NOTE: suppress context menu of the main window.
				return FALSE;

			case WM_LBUTTONDOWN:
				::SetFocus (_hWnd);
				CTrackVideo::g_pTrackVideo->OnLButtonDown_View (GET_X_LPARAM (_lParam), GET_Y_LPARAM (_lParam));
				return FALSE;
			case WM_LBUTTONUP:
				CTrackVideo::g_pTrackVideo->OnLButtonUp_View   (GET_X_LPARAM (_lParam), GET_Y_LPARAM (_lParam));
				return FALSE;
			case WM_MOUSEMOVE:
				CTrackVideo::g_pTrackVideo->OnMouseMove_View   (GET_X_LPARAM (_lParam), GET_Y_LPARAM (_lParam));
				return FALSE;

			case WM_MOUSEWHEEL: {
				const int nKeys = GET_KEYSTATE_WPARAM (_wParam);
				const short sDelta = GET_WHEEL_DELTA_WPARAM (_wParam);
				CTrackVideo::g_pTrackVideo->OnMouseWheel_View (sDelta/WHEEL_DELTA);
				return FALSE;
			}

			case WM_SETCURSOR:
				if (CTrackVideo::g_pTrackVideo->OnSetCursor_View ())
					return TRUE;
				break;

			case WM_KEYDOWN:
				switch (_wParam) {
					case VK_DOWN:  CTrackVideo::g_pTrackVideo->SendCommand (ID_VIDEO_STEP_BW); break;
					case VK_LEFT:  CTrackVideo::g_pTrackVideo->SendCommand (ID_VIDEO_1S_BW);   break;
					case VK_UP:    CTrackVideo::g_pTrackVideo->SendCommand (ID_VIDEO_STEP_FW); break;
					case VK_RIGHT: CTrackVideo::g_pTrackVideo->SendCommand (ID_VIDEO_1S_FW);   break;
					case VK_SPACE: CTrackVideo::g_pTrackVideo->TogglePlay (); break;
				} 
 				return FALSE;

			case WM_PAINT:
				PAINTSTRUCT ps;
				::BeginPaint (_hWnd, & ps);
				CTrackVideo::g_pTrackVideo->PaintFrame (ps);
				::EndPaint (_hWnd, & ps);
				return FALSE;
		}
	}
	return ::CallWindowProc (g_wpViewProc, _hWnd, _uMsg, _wParam, _lParam);
}

///////////////////////////////////////////////////////////////////

void CTrackVideo::_DestroyUndockedWindows () {
	const size_t cOldUndockedWindows = m_vecUndockedWindows.size ();
	for (size_t c = 0; c < cOldUndockedWindows; ++ c) {
		CTrackVideoChannel * pTVC = m_vecUndockedWindows [c];
		DestroyModeless (pTVC);
	}
	m_vecUndockedWindows.clear ();
}

void CTrackVideo::_Dock () {
	_DestroyUndockedWindows ();
	_UpdateFrame ();
}

void CTrackVideo::_Undock () {
	if (m_pRenderer == NULL)
		return;

	if (! m_vecUndockedWindows.empty ())
		_DestroyUndockedWindows ();

	RECT rectThis;
	::GetWindowRect (m_hWnd, & rectThis);

	const size_t cChannels = m_pRenderer->GetChannels ();
	for (size_t c = 0; c < cChannels; ++ c) {
		CTrackVideoChannel * pTVC = NULL;
		ShowModeless (pTVC, g_hMapWnd);
		m_vecUndockedWindows.push_back (pTVC);

		pTVC->Init (this, c);

		char strTitle [64];
		::sprintf (strTitle, "Channel #%d", c + 1);
		::SetWindowText (* pTVC, strTitle);

		const size_t cXDef = rectThis.left + (c + 1)*20;
		const size_t cYDef = rectThis.top  + (c + 1)*20 + 70;

		RECT rect = {0, 0, m_pRenderer->GetWidth (c), m_pRenderer->GetHeight (c)};
		::AdjustWindowRectEx (& rect, ::GetWindowLong (* pTVC, GWL_STYLE), FALSE, ::GetWindowLong (* pTVC, GWL_EXSTYLE));
		const size_t cWidthDef  = rect.right - rect.left;
		const size_t cHeightDef = rect.bottom - rect.top;

		char strX [32], strY [32], strWidth [32], strHeight [32];
		::sprintf (strX,      "TrackVideoChannel%d_X",      c);
		::sprintf (strY,      "TrackVideoChannel%d_Y",      c);
		::sprintf (strWidth,  "TrackVideoChannel%d_Width",  c);
		::sprintf (strHeight, "TrackVideoChannel%d_Height", c);
		const DWORD dwX      = GetSetting (strX,      (DWORD) cXDef);
		const DWORD dwY      = GetSetting (strY,      (DWORD) cYDef);
		const DWORD dwWidth  = GetSetting (strWidth,  (DWORD) cWidthDef);
		const DWORD dwHeight = GetSetting (strHeight, (DWORD) cHeightDef);

		::SetWindowPos (* pTVC, NULL, dwX, dwY, dwWidth, dwHeight, SWP_NOZORDER);
		EnsureVisibleOnScreen (* pTVC);
	}

	_UpdateFrame ();
}

void CTrackVideo::_SetSpeed (int _nSpeed) {
	m_nSpeed = _nSpeed;

	if (m_pRenderer)
		m_pRenderer->SetSpeed (m_nSpeed);
}

void CTrackVideo::GoFrame (const rgn_t & _rgn, CTrackVideoRendererBase * _pRenderer, DWORD _dwProgressPrec) {
	GoFrame (_rgn, _pRenderer, 0, 0);
	m_pRenderer->SetProgressPrec (_dwProgressPrec);
	_Update ();
}

void CTrackVideo::GoFrame (
	const rgn_t & _rgn, CTrackVideoRendererBase * _pRenderer,
	size_t _cTrackPointVideoData0, size_t _cTrackPointVideoData1
) {
	m_pRgn = & _rgn;

	ShowVideoPosCursor (true);
	m_bCursorShow = true;

	// Prepare _pRenderer.
	m_pRenderer = _pRenderer;
	_UpdateTab ();
	if (m_pRenderer) {
		m_pRenderer->Mute (_IsTButtonChecked (ID_VIDEO_MUTE));

		m_nFPS = m_pRenderer->GetFPS ();
		m_pRenderer->GoFrame (_cTrackPointVideoData0, _cTrackPointVideoData1);

		m_pBookmarks = m_pRenderer->GetBookmarks ();
		m_pBookmarks->Update ();
	}

	// Load bookmarks.
	assert (m_pRgn->pFileRef);
	if (m_pRgn->pFileRef)
		m_strTrackFileName = m_pRgn->pFileRef->strFileName;
	else
		m_strTrackFileName.clear ();

	// Prepare docking.
	if (! m_bDocked)
		_Undock ();

	// Pause playing.
	_SetSpeed (0);
	_UpdateTimerRender ();
	_Update ();
}

void CTrackVideo::_Update () {
	if (m_pRgn && m_pRgn->IsDeleted ()) {
		Reset ();
		return;
	}

	_UpdateToolBar ();
	_UpdateInfo ();
	_UpdateFrame ();
}

void CTrackVideo::_UpdateToolBar () {
	const bool bAtStart = m_pRenderer != NULL && m_pRenderer->IsAtStart ();
	const bool bAtEnd   = m_pRenderer != NULL && m_pRenderer->IsAtEnd ();

	const bool bProcessingVoices    = _IsTButtonChecked (ID_VIDEO_BOOKMARK_VOICES);
	const bool bProcessingRoadSigns = _IsTButtonChecked (ID_VIDEO_BOOKMARK_SIGNS);

	if (m_pBookmarks && m_pRenderer)
		m_pBookmarks->SetPositionPrec (m_pRenderer->GetProgressPrec ());

	_EnableTBButton (ID_VIDEO_PLAY,            ! m_bProcessing && m_pRenderer != NULL && ! bAtEnd);
	_EnableTBButton (ID_VIDEO_PAUSE,           ! m_bProcessing && m_pRenderer != NULL);
	_EnableTBButton (ID_VIDEO_FAST_BW,         ! m_bProcessing && m_pRenderer != NULL && ! bAtStart);
	_EnableTBButton (ID_VIDEO_FAST_FW,         ! m_bProcessing && m_pRenderer != NULL && ! bAtEnd);
	_EnableTBButton (ID_VIDEO_1S_BW,           ! m_bProcessing && m_pRenderer != NULL && m_nSpeed == 0 && ! bAtStart);
	_EnableTBButton (ID_VIDEO_1S_FW,           ! m_bProcessing && m_pRenderer != NULL && m_nSpeed == 0 && ! bAtEnd);
	_EnableTBButton (ID_VIDEO_STEP_BW,         ! m_bProcessing && m_pRenderer != NULL && m_nSpeed == 0 && ! bAtStart);
	_EnableTBButton (ID_VIDEO_STEP_FW,         ! m_bProcessing && m_pRenderer != NULL && m_nSpeed == 0 && ! bAtEnd);
	_EnableTBButton (ID_VIDEO_RESIZE,          ! m_bProcessing && m_pRenderer != NULL);
	_EnableTBButton (ID_VIDEO_AUTOSCROLL_MAP,  ! m_bProcessing && m_pRenderer != NULL);
	_EnableTBButton (ID_VIDEO_MUTE,            ! m_bProcessing && m_pRenderer != NULL);
	_EnableTBButton (ID_VIDEO_BOOKMARK_ADD,    ! m_bProcessing && m_pRenderer != NULL && m_nSpeed == 0);
	_EnableTBButton (ID_VIDEO_BOOKMARK,        ! m_bProcessing && m_pRenderer != NULL && m_nSpeed == 0 && m_pBookmarks && m_pBookmarks->GetBookmarksHere () != 0);
	_EnableTBButton (ID_VIDEO_BOOKMARK_PREV,   ! m_bProcessing && m_pRenderer != NULL && m_nSpeed == 0 && m_pBookmarks && m_pBookmarks->HasPrevPosition ());
	_EnableTBButton (ID_VIDEO_BOOKMARK_NEXT,   ! m_bProcessing && m_pRenderer != NULL && m_nSpeed == 0 && m_pBookmarks && m_pBookmarks->HasNextPosition ());
	_EnableTBButton (ID_VIDEO_BOOKMARK_VOICES, ! bProcessingRoadSigns && m_pRenderer != NULL);
	_EnableTBButton (ID_VIDEO_BOOKMARK_SIGNS,  ! bProcessingVoices    && m_pRenderer != NULL);

	_CheckTBButton  (ID_VIDEO_PLAY,     m_nSpeed == 1);
	_CheckTBButton  (ID_VIDEO_PAUSE,    m_nSpeed == 0);
	_CheckTBButton  (ID_VIDEO_FAST_BW,  m_nSpeed  < 0);
	_CheckTBButton  (ID_VIDEO_FAST_FW,  m_nSpeed >= 2);

	EnableDlgItem   (IDC_PROGRESS,             ! m_bProcessing && m_pRenderer != NULL);
	EnableDlgItem   (IDC_CHANNEL,              ! m_bProcessing && m_bDocked);
	EnableDlgItem   (IDC_INFO,                 ! m_bProcessing);
}

void CTrackVideo::_UpdateInfo () {
	if (m_pRenderer == NULL) {
		SetDlgItemText (IDC_INFO, "-");
		::SendMessage (m_hProgress, TBM_SETPOS, TRUE, 0);
		if (m_pBookmarks)
			m_pBookmarks->Clear ();
		return;
	}

	// Update Progress.
	const double fProgress = m_pRenderer->GetProgress ();
	::SendMessage (m_hProgress, TBM_SETPOS, TRUE, fast_floor_d (fProgress*10000 + .5f));

	// Update bookmarks.
	if (m_pBookmarks)
		m_pBookmarks->SetPositionPrec (m_pRenderer->GetProgressPrec ());

	// Update Info.
	char strInfo [128] = "";
	if (m_nSpeed != 0)
		::sprintf (strInfo, "[x%d] ", ::labs (m_nSpeed));

	point_t pt;
	trackpoint_info_t tpi;
	const bool bPtDefined = m_pRenderer->GetVehicleInfo (pt, tpi);

	if ((tpi.wFlags & trackpoint_info_t::defTime) != 0) {
		// Convert to local time.
		DATE dtLocalTime = tpi.dtTime - float (m_tzi.Bias)/(24*60);

		// Check milliseconds (NOTE: ::VariantTimeToSystemTime() does not support this).
		int nMilliseconds = 0;
		if (tpi.wFlags & trackpoint_info_t::defTimeMilliseconds) {
			const float fSeconds = (dtLocalTime - fast_floor_d (dtLocalTime))*24*60*60;
			nMilliseconds = fast_floor ((fSeconds - fast_floor (fSeconds))*1000 + .5f);
			dtLocalTime -= nMilliseconds/24./60./60./1000;
		}
		SYSTEMTIME st;
		::VariantTimeToSystemTime (dtLocalTime, & st);

		char strDate [32];
		if (! ::GetDateFormat (LOCALE_USER_DEFAULT, DATE_SHORTDATE, & st, NULL, strDate, sizeof (strDate)/sizeof (strDate [0])))
			strDate [0] = '\0';
		char strTime [32];
		if (! ::GetTimeFormat (LOCALE_USER_DEFAULT, 0, & st, NULL, strTime, sizeof (strTime)/sizeof (strTime [0])))
			strTime [0] = '\0';

		::strcat (strInfo, strDate);
		::strcat (strInfo, " ");
		::strcat (strInfo, strTime);
		if (tpi.wFlags & trackpoint_info_t::defTimeMilliseconds) {
			char strMS [16];
			::sprintf (strMS, ".%02d", (nMilliseconds + 5)/10);
			::strcat (strInfo, strMS);
		}
		::strcat (strInfo, "  ");
	}

	// Position.
	if (bPtDefined) {
		char strPosition [32];
		PrintPosition (strPosition, pt);

		::strcat (strInfo, strPosition);
		::strcat (strInfo, "  ");
	}

	// Speed.
	if ((tpi.wFlags & trackpoint_info_t::defSpeed) != 0) {
		char strSpeed [16], strSpeedUnit [8];
		PrintSpeed (strSpeed, strSpeedUnit, tpi.fSpeed);
		::strcat (strInfo, strSpeed);
		::strcat (strInfo, " ");
		::strcat (strInfo, strSpeedUnit);
	}

	// Vehicle ID.
	string_t strVehicleId;
	m_pRenderer->GetVehicleID (strVehicleId);
	if (! strVehicleId.empty ()) {
		::strcat (strInfo, "  [VID:");
		::strcat (strInfo, strVehicleId.c_str ());
		::strcat (strInfo, "]");
	}

	SetDlgItemText (IDC_INFO, strInfo);

	// Update position on map.
	if (bPtDefined) {
		if (! (m_ptCursor == pt)) {
			m_ptCursor = pt;
			m_fCursorHeading = (tpi.wFlags & trackpoint_info_t::defHeading) != 0 ? tpi.fHeading : 0;
			SetVideoPosCursor (pt.x, pt.y, m_fCursorHeading);

			ShowVideoPosCursor (true);
			m_bCursorShow = true;

			Redraw ();

			_AutoScrollMap ();
		}
	} else {
		if (m_bCursorShow) {
			ShowVideoPosCursor (false);
			m_bCursorShow = false;
			Redraw ();
		}
	}
}		

void CTrackVideo::_UpdateFrame () {
	::InvalidateRect (m_hView, NULL, FALSE);
	::UpdateWindow (m_hView);

	if (! m_bDocked) {
		const size_t cUndockedWindows = m_vecUndockedWindows.size ();
		for (size_t c = 0; c < cUndockedWindows; ++ c) {
			const HWND hwnd = * m_vecUndockedWindows [c];
			::InvalidateRect (hwnd, NULL, FALSE);
			::UpdateWindow (hwnd);
		}
	}
}

void CTrackVideo::_UpdateTab () {
	TabCtrl_DeleteAllItems (m_hChannel);

	TCITEM tci;
	tci.mask       = TCIF_TEXT;
	tci.pszText    = "All";
	tci.cchTextMax = ::strlen (tci.pszText);
	TabCtrl_InsertItem (m_hChannel, 0, & tci);

	const size_t cChannels = m_pRenderer ? m_pRenderer->GetChannels () : 1;
	for (size_t cChannel = 0; cChannel < cChannels; ++ cChannel) {
		char str [16];
		::sprintf (str, "%d", cChannel + 1);

		tci.pszText    = str;
		tci.cchTextMax = ::strlen (tci.pszText);
		TabCtrl_InsertItem (m_hChannel, cChannel + 1, & tci);
	}
	TabCtrl_SetCurSel (m_hChannel, m_bAllChannels ? 0 : 1 + m_cChannel);
}

bool CTrackVideo::OnInit (HWND _hwndFocus) {
	if (::GetTimeZoneInformation (& m_tzi) == TIME_ZONE_ID_DAYLIGHT)
		m_tzi.Bias += m_tzi.DaylightBias;

	// Toolbar.
	const TBBUTTON btns [] = {
		{0,  ID_VIDEO_FAST_BW,         TBSTATE_ENABLED, TBSTYLE_CHECK,  0, 0},
		{0,  -1,                       TBSTATE_ENABLED, TBSTYLE_SEP,    0, 0},
		{1,  ID_VIDEO_PLAY,            TBSTATE_ENABLED, TBSTYLE_CHECK,  0, 0},
		{2,  ID_VIDEO_PAUSE,           TBSTATE_ENABLED, TBSTYLE_CHECK,  0, 0},
		{0,  -1,                       TBSTATE_ENABLED, TBSTYLE_SEP,    0, 0},
		{3,  ID_VIDEO_1S_BW,           TBSTATE_ENABLED, TBSTYLE_BUTTON, 0, 0},
		{4,  ID_VIDEO_STEP_BW,         TBSTATE_ENABLED, TBSTYLE_BUTTON, 0, 0},
		{0,  -1,                       TBSTATE_ENABLED, TBSTYLE_SEP,    0, 0},
		{5,  ID_VIDEO_STEP_FW,         TBSTATE_ENABLED, TBSTYLE_BUTTON, 0, 0},
		{6,  ID_VIDEO_1S_FW,           TBSTATE_ENABLED, TBSTYLE_BUTTON, 0, 0},
		{0,  -1,                       TBSTATE_ENABLED, TBSTYLE_SEP,    0, 0},
		{7,  ID_VIDEO_FAST_FW,         TBSTATE_ENABLED, TBSTYLE_CHECK,  0, 0},
		{0,  -1,                       TBSTATE_ENABLED, TBSTYLE_SEP,    0, 0},
		{8,  ID_VIDEO_RESIZE,          TBSTATE_ENABLED, TBSTYLE_BUTTON, 0, 0},
		{9,  ID_VIDEO_DOCKED,          TBSTATE_ENABLED, TBSTYLE_CHECK,  0, 0},
		{10, ID_VIDEO_AUTOSCROLL_MAP,  TBSTATE_ENABLED, TBSTYLE_CHECK,  0, 0},
		{11, ID_VIDEO_MUTE,            TBSTATE_ENABLED, TBSTYLE_CHECK,  0, 0},
		{0,  -1,                       TBSTATE_ENABLED, TBSTYLE_SEP,    0, 0},
		{12, ID_VIDEO_BOOKMARK_ADD,    TBSTATE_ENABLED, TBSTYLE_BUTTON, 0, 0},
		{13, ID_VIDEO_BOOKMARK,        TBSTATE_ENABLED, TBSTYLE_BUTTON, 0, 0},
		{0,  -1,                       TBSTATE_ENABLED, TBSTYLE_SEP,    0, 0},
		{14, ID_VIDEO_BOOKMARK_PREV,   TBSTATE_ENABLED, TBSTYLE_BUTTON, 0, 0},
		{15, ID_VIDEO_BOOKMARK_NEXT,   TBSTATE_ENABLED, TBSTYLE_BUTTON, 0, 0},
		{0,  -1,                       TBSTATE_ENABLED, TBSTYLE_SEP,    0, 0},
		{16, ID_VIDEO_BOOKMARK_VOICES, TBSTATE_ENABLED, TBSTYLE_CHECK,  0, 0},
		{17, ID_VIDEO_BOOKMARK_SIGNS,  TBSTATE_ENABLED, TBSTYLE_CHECK,  0, 0},
	};
	assert (m_hToolbarWnd == NULL);
	m_hToolbarWnd = ::CreateToolbarEx (
		m_hWnd, WS_CHILD | WS_VISIBLE | WS_CLIPSIBLINGS | CCS_TOP | TBSTYLE_FLAT, 100,
		7, g_hInst, IDT_TRACK_VIDEO,
		btns, ARRAY_SIZE (btns), 0, 0, 16, 16, sizeof (TBBUTTON)
	);

	const bool bScroll = GetSetting ("TrackVideoAutoScroll", true);
	_CheckTBButton (ID_VIDEO_AUTOSCROLL_MAP, bScroll);

	m_bDocked = GetSetting ("TrackVideoDocked", true);
	_CheckTBButton (ID_VIDEO_DOCKED, m_bDocked);

	const bool bMute = GetSetting ("TrackVideoMute", false);
	_CheckTBButton (ID_VIDEO_MUTE, bMute);

	// Progress.
	m_hProgress = GetDlgItem (IDC_PROGRESS);
	::SendMessage (m_hProgress, TBM_SETRANGE, FALSE, MAKELONG (0, 10000));

	// Info.
	m_hInfo = GetDlgItem (IDC_INFO);

	// Channel.
	m_hChannel = GetDlgItem (IDC_CHANNEL);
	::SetWindowLong (m_hChannel, GWL_STYLE,   WS_CLIPSIBLINGS     | ::GetWindowLong (m_hChannel, GWL_STYLE));
	::SetWindowLong (m_hChannel, GWL_EXSTYLE, WS_EX_CONTROLPARENT | ::GetWindowLong (m_hChannel, GWL_EXSTYLE));
	TabCtrl_SetExtendedStyle (m_hChannel, TCS_EX_FLATSEPARATORS);
	_UpdateTab ();

	// View.
	m_hView = GetDlgItem (IDC_VIEW);
	// Window subclassing.
	g_wpViewProc = reinterpret_cast<WNDPROC> (::GetWindowLong (m_hView, GWL_WNDPROC));
	::SetWindowLong (m_hView, GWL_WNDPROC, reinterpret_cast<long> (_TrackVideoProc));
	::SetFocus (m_hView);

	RECT rect;
	::GetClientRect (m_hWnd, & rect);
	RECT rectView;
	::GetWindowRect (m_hView, & rectView);
	m_lDW = (rect.right - rect.left) - (rectView.right - rectView.left);
	m_lDH = (rect.bottom - rect.top) - (rectView.bottom - rectView.top);

	// Restore size and position of the window.
	int x = GetSetting ("TrackVideoX", (DWORD) 200);
	int y = GetSetting ("TrackVideoY", (DWORD) 100);
	POINT pt = {x, y};
	const HMONITOR hMonitor = MonitorFromPoint (pt, MONITOR_DEFAULTTONULL);
	if (hMonitor == NULL) {
		x = 0;
		y = 0;
	}
	::SetWindowPos (
		m_hWnd, NULL,
		x, y, GetSetting ("TrackVideoW", (DWORD) 300), GetSetting ("TrackVideoH", (DWORD) 300),
		SWP_NOZORDER
	);
	EnsureVisibleOnScreen (m_hWnd);

	_UpdateTimerRender ();
	return false;
}

rgn_t * CTrackVideo::_AddBookmark () {
	point_t pt;
	trackpoint_info_t tpi;
	if (! m_pRenderer->GetVehicleInfo (pt, tpi))
		return NULL;

	if (m_pBookmarks == NULL)
		return NULL;

	rgn_t & rgn = CreateBookmark (pt);
	bookmark_t * const pBookmark = rgn.pExtra->pBookmark;
	pBookmark->strFileName = m_strTrackFileName;
	pBookmark->dwFilePos   = m_pRenderer->GetProgressPrec ();
	pBookmark->dt          = tpi.dtTime;
	pBookmark->bHasHeading = true;
	pBookmark->fHeading    = tpi.fHeading;

	m_pBookmarks->AddHere (rgn);

	return & rgn;
}

void CTrackVideo::OnCommand (WORD _wCmd, WORD _wNotifyCode, HWND _hwndCtl) {
	switch (_wCmd) {
		case ID_VIDEO_PLAY:
			_SetSpeed (1);
			_Update ();
			_UpdateTimerRender ();
			::SetFocus (m_hView);
			break;
		case ID_VIDEO_PAUSE:
			_PauseAndUpdate ();
			::SetFocus (m_hView);
			break;

		case ID_VIDEO_STEP_FW:
			if (m_pRenderer)
				m_pRenderer->GoForward (1);
			_PauseAndUpdate ();
			::SetFocus (m_hView);
			break;
		case ID_VIDEO_STEP_BW:
			if (m_pRenderer)
				m_pRenderer->GoBackward (1);
			_PauseAndUpdate ();
			::SetFocus (m_hView);
			break;

		case ID_VIDEO_1S_FW:
			if (m_pRenderer)
				m_pRenderer->GoForward (m_nFPS);
			_PauseAndUpdate ();
			::SetFocus (m_hView);
			break;
		case ID_VIDEO_1S_BW:
			if (m_pRenderer)
				m_pRenderer->GoBackward (m_nFPS);
			_PauseAndUpdate ();
			::SetFocus (m_hView);
			break;

		case ID_VIDEO_FAST_FW:
			if (m_nSpeed <= 0)
				_SetSpeed (2);
			else
				_SetSpeed (m_nSpeed*2);
			_Update ();
			_UpdateTimerRender ();
			::SetFocus (m_hView);
			break;
		case ID_VIDEO_FAST_BW:
			if (m_nSpeed >= 0)
				_SetSpeed (-2);
			else
				_SetSpeed (m_nSpeed*2);
			_Update ();
			_UpdateTimerRender ();
			::SetFocus (m_hView);
			break;

		case ID_VIDEO_RESIZE: {
			if (m_pRenderer == NULL)
				break;

			if (m_bDocked) {
				int wr = 0;
				int hr = 0;
				if (m_bAllChannels) {
					const size_t cChannels = m_pRenderer->GetChannels ();
					const size_t cRows = __max (1, fast_floor (::sqrt ((float) cChannels) + .5f));
					const size_t cCols = __max (1, cChannels/cRows);
					assert (cCols*cRows >= cChannels);

					for (size_t cCol = 0; cCol < cCols; ++ cCol) {
						size_t cColWidthMax = 0;
						for (size_t cRow = 0; cRow < cRows; ++ cRow) {
							const size_t cChannel = cCol + cRow*cCols;
							if (cChannel >= cChannels)
								continue;
							const size_t w = m_pRenderer->GetWidth (cChannel);
							if (cColWidthMax < w)
								cColWidthMax = w;
						}
						wr += cColWidthMax;
					}
					for (size_t cRow = 0; cRow < cRows; ++ cRow) {
						size_t cRowHeightMax = 0;
						for (size_t cCol = 0; cCol < cCols; ++ cCol) {
							const size_t cChannel = cCol + cRow*cCols;
							if (cChannel >= cChannels)
								continue;
							const size_t h = m_pRenderer->GetHeight (cChannel);
							if (cRowHeightMax < h)
								cRowHeightMax = h;
						}
						hr += cRowHeightMax;
					}
				} else {
					wr = m_pRenderer->GetWidth  (m_cChannel);
					hr = m_pRenderer->GetHeight (m_cChannel);
				}

				int w = wr + 2*m_lDW;
				int h = hr + m_lDH + 30;
				for (size_t cAttempt = 0; cAttempt < 3; ++ cAttempt) {
					::SetWindowPos (m_hWnd, NULL, 0, 0, w, h, SWP_NOMOVE | SWP_NOZORDER);
					RECT rectView;
					::GetClientRect (m_hView, & rectView);
					const int wc = rectView.right - rectView.left;
					const int hc = rectView.bottom - rectView.top;
					if (wc == wr && hc == hr)
						break;
					if (wc != wr)
						w -= wc - wr;
					if (hc != hr)
						h -= hc - hr;
				}
			} else {
				const size_t cChannels = m_pRenderer->GetChannels ();
				const size_t cUndockedWindows = m_vecUndockedWindows.size ();
				for (size_t c = 0; c < cUndockedWindows && c < cChannels; ++ c) {
					CTrackVideoChannel * const pTVC = m_vecUndockedWindows [c];

					RECT rect = {0, 0, m_pRenderer->GetWidth (c), m_pRenderer->GetHeight (c)};
					::AdjustWindowRectEx (& rect, ::GetWindowLong (* pTVC, GWL_STYLE), FALSE, ::GetWindowLong (* pTVC, GWL_EXSTYLE));

					::SetWindowPos (* pTVC, NULL, 0, 0, rect.right - rect.left, rect.bottom - rect.top, SWP_NOZORDER | SWP_NOMOVE);
				}
			}

			::SetFocus (m_hView);
			break;
		}

		case ID_VIDEO_DOCKED: {
			m_bDocked = _IsTButtonChecked (ID_VIDEO_DOCKED);
			SaveSetting ("TrackVideoDocked", m_bDocked);
			if (m_bDocked) {
				_Dock ();
				::SetFocus (m_hView);
			} else
				_Undock ();

			break;
		}

		case ID_VIDEO_AUTOSCROLL_MAP:
			_AutoScrollMap ();
			::SetFocus (m_hView);
			break;

		case ID_VIDEO_MUTE: {
			const bool bMute = _IsTButtonChecked (ID_VIDEO_MUTE);
			SaveSetting ("TrackVideoMute", bMute);
			if (m_pRenderer)
				m_pRenderer->Mute (bMute);
			break;
		}

		case ID_VIDEO_BOOKMARK_ADD: {
			if (m_pRenderer && m_pBookmarks) {
				const int nDepth = BeginCombinedAction ();

				rgn_t * const pBookmarkRgn = _AddBookmark ();
				if (pBookmarkRgn == NULL) {
					ErrorMsgBox ("No position defined.");
					break;
				}
				const size_t cNewOrdinal = m_pBookmarks->GetBookmarksHere () - 1;

				bool bDeleteApplied = false;
				const bool bOK = ShowBookmarkProperties (m_hWnd, * pBookmarkRgn, true, bDeleteApplied);
				EndCombinedAction (nDepth, "create bookmark");
				if (! bOK)
					UndoEditAction ();

				Refresh ();
				m_pBookmarks->Update ();
				_Update ();
			}
			break;
		}

		case ID_VIDEO_BOOKMARK: {
			rgn_t * const pBookmarkRgn = m_pBookmarks->GetBookmarkHere (0);
			if (pBookmarkRgn == NULL || pBookmarkRgn->IsDeleted ())
				break;
			bool bDeleteApplied = false;
			const bool bOK = ShowBookmarkProperties (m_hWnd, * pBookmarkRgn, false, bDeleteApplied);
			if (bDeleteApplied)
				m_pBookmarks->Update ();
			if (bOK || bDeleteApplied)
				_Update ();
			break;
		}

		case ID_VIDEO_BOOKMARK_PREV:
			if (m_pBookmarks && m_pRenderer) {
				m_pRenderer->SetProgressPrec (m_pBookmarks->FindPrevPosition ());
				_Update ();
			}
			break;
		case ID_VIDEO_BOOKMARK_NEXT:
			if (m_pBookmarks && m_pRenderer) {
				m_pRenderer->SetProgressPrec (m_pBookmarks->FindNextPosition ());
				_Update ();
			}
			break;

		case ID_VIDEO_BOOKMARK_VOICES: {
			const bool bStop = ! _IsTButtonChecked (ID_VIDEO_BOOKMARK_VOICES);
			if (bStop)
				StopProcessing ();
			else {
				_ProcessInParallel ("Search for voices...", & CTrackVideo::_DoBookmarkVoices);
				_CheckTBButton (ID_VIDEO_BOOKMARK_VOICES, false);
			}
			break;
		}

		case ID_VIDEO_BOOKMARK_SIGNS: {
			const bool bStop = ! _IsTButtonChecked (ID_VIDEO_BOOKMARK_SIGNS);
			if (bStop)
				StopProcessing ();
			else {
				CRoadSignsDetector rsd (m_pRenderer ? m_pRenderer->GetChannels () : 0);
				if (rsd.ShowModal (m_hWnd) == 1) {
					m_dwChannelsToScan = rsd.GetChannelsToScan ();
					_ProcessInParallel ("Search for road signs...", & CTrackVideo::_DoBookmarkRoadSigns);
					_CheckTBButton (ID_VIDEO_BOOKMARK_SIGNS, false);
				}
			}
			break;
		}

		case IDCANCEL:
			if (m_uiTimerID != 0) {
				::timeKillEvent (m_uiTimerID);
				m_uiTimerID = 0;
			}

			Hide ();
			::SetFocus (g_hMapWnd);

			_DestroyUndockedWindows ();

			ShowVideoPosCursor (false);
			m_bCursorShow = false;
			Redraw ();
			break;
	}
}

void CTrackVideo::OnHScroll (WORD _wRequest, WORD _wPos, HWND _hwndCtl) {
	::SetFocus (m_hView);

	if (_hwndCtl == m_hProgress) {
		const long lProgress = _wRequest == SB_THUMBPOSITION || _wRequest == SB_THUMBTRACK ? _wPos : ::SendMessage (m_hProgress, TBM_GETPOS, 0, 0);
		if (m_pRenderer) {
			m_pRenderer->SetProgress (lProgress/10000.);
			_Update ();
		}
	}
}

void CTrackVideo::OnSize (DWORD _dwResizeFlag, WORD _wWidth, WORD _wHeight) {
	if (_dwResizeFlag != SIZE_RESTORED)
		return;

	RECT rect;
	::GetWindowRect (m_hWnd, & rect);
	SaveSetting ("TrackVideoW", (DWORD) (rect.right - rect.left));
	SaveSetting ("TrackVideoH", (DWORD) (rect.bottom - rect.top));

	RECT rectClient;
	::GetClientRect (m_hWnd, & rectClient);
	const int nW = rectClient.right - rectClient.left;
	const int nH = rectClient.bottom - rectClient.top;

	// NOTE: toolbar set size & position automaticaly.
	::SetWindowPos (m_hToolbarWnd, NULL, 0, 0, 0, 0, SWP_NOZORDER | SWP_DEFERERASE);

	const int w = __max (nW - m_lDW, 100);

	// Progress.
	RECT rectProgress;
	::GetWindowRect (m_hProgress, & rectProgress);
	const int nProgressH = rectProgress.bottom - rectProgress.top;
	::SetWindowPos (m_hProgress, NULL, 0, 0, w, nProgressH, SWP_NOMOVE | SWP_NOZORDER);

	// Info.
	RECT rectInfo;
	::GetWindowRect (m_hInfo, & rectInfo);
	const int nInfoH = rectInfo.bottom - rectInfo.top;
	::SetWindowPos (m_hInfo, NULL, 0, 0, w, nInfoH, SWP_NOMOVE | SWP_NOZORDER);

	// Channel.
	RECT rectChannel;
	::GetWindowRect (m_hChannel, & rectChannel);
	const int nChannelH = rectChannel.bottom - rectChannel.top;
	::SetWindowPos (m_hChannel, NULL, 0, 0, w, nChannelH, SWP_NOMOVE | SWP_NOZORDER);

	// View.
	::SetWindowPos (m_hView, NULL, 0, 0, w, __max (nH - m_lDH, 100), SWP_NOMOVE | SWP_NOZORDER);
	RECT rectView;
	::GetClientRect (m_hView, & rectView);
	m_lViewW = rectView.right - rectView.left;
	m_lViewH = rectView.bottom - rectView.top;

	m_MemDC.Resize (CClientDC (NULL), m_lViewW, m_lViewH);

	_UpdateFrame ();
}

void CTrackVideo::OnMove (int _x, int _y) {
	RECT rect;
	::GetWindowRect (m_hWnd, & rect);
	POINT p = {rect.left, rect.top};
//	::ScreenToClient (g_hMapWnd, & p);
	if (p.x < 0)
		p.x = 0;
	if (p.y < 0)
		p.y = 0;

	SaveSetting ("TrackVideoX", DWORD (p.x));
	SaveSetting ("TrackVideoY", DWORD (p.y));
}

void CTrackVideoChannel::OnPaint (const PAINTSTRUCT & _ps) const {
	if (m_pTrackVideo) {
		RECT rect;
		::GetClientRect (m_hWnd, & rect);
		int nW = rect.right - rect.left;
		int nH = rect.bottom - rect.top;

		::FillRect (m_MemDC, & rect, ::GetSysColorBrush (COLOR_WINDOW));

		if (m_pTrackVideo && m_pTrackVideo->m_pRenderer)
			m_pTrackVideo->m_pRenderer->Render (m_MemDC, 0, 0, nW, nH, m_cChannel);

		m_pTrackVideo->DrawPresenceOfBookmarks (m_MemDC, nW, nH);

		::BitBlt (_ps.hdc, 0, 0, nW, nH, m_MemDC, 0, 0, SRCCOPY);
	}
}

void CTrackVideo::PaintFrame (const PAINTSTRUCT & _ps) const {
	RECT rectView;
	::GetClientRect (m_hView, & rectView);
	int nW = rectView.right - rectView.left;
	int nH = rectView.bottom - rectView.top;
	::FillRect (m_bDocked ? m_MemDC : _ps.hdc, & rectView, ::GetSysColorBrush (COLOR_WINDOW));

	if (! m_bDocked)
		return;

	if (m_pRenderer)
		if (m_bAllChannels) {
			const size_t cChannels = m_pRenderer->GetChannels ();
			const size_t cRows = __max (1, fast_floor (::sqrt ((float) cChannels) + .5f));
			const size_t cCols = __max (1, cChannels/cRows);
			assert (cCols*cRows >= cChannels);

			size_t * const pcColX = reinterpret_cast<size_t *> (alloca (sizeof (size_t)*(cCols + 1)));
			::memset (pcColX, 0, sizeof (size_t)*(cCols + 1));
			size_t * const pcRowY = reinterpret_cast<size_t *> (alloca (sizeof (size_t)*(cRows + 1)));
			::memset (pcRowY, 0, sizeof (size_t)*(cRows + 1));

			size_t cCol;
			for (cCol = 0; cCol < cCols; ++ cCol) {
				size_t cColWidthMax = 0;
				for (size_t cRow = 0; cRow < cRows; ++ cRow) {
					const size_t cChannel = cCol + cRow*cCols;
					if (cChannel >= cChannels)
						continue;
					const size_t w = m_pRenderer->GetWidth (cChannel);
					if (cColWidthMax < w)
						cColWidthMax = w;
				}
				pcColX [cCol + 1] = cColWidthMax + pcColX [cCol];
			}
			for (size_t cRow = 0; cRow < cRows; ++ cRow) {
				size_t cRowHeightMax = 0;
				for (size_t cCol = 0; cCol < cCols; ++ cCol) {
					const size_t cChannel = cCol + cRow*cCols;
					if (cChannel >= cChannels)
						continue;
					const size_t h = m_pRenderer->GetHeight (cChannel);
					if (cRowHeightMax < h)
						cRowHeightMax = h;
				}
				pcRowY [cRow + 1] = cRowHeightMax + pcRowY [cRow];
			}

			const size_t cTotalW = pcColX [cCols];
			const size_t cTotalH = pcRowY [cRows];
			if (cTotalW != 0 && cTotalH != 0) {
				for (cCol = 0; cCol < cCols; ++ cCol) {
					for (size_t cRow = 0; cRow < cRows; ++ cRow) {
						const size_t cChannel = cCol + cRow*cCols;
						if (cChannel >= cChannels)
							continue;

						m_pRenderer->Render (
							m_MemDC,
							pcColX [cCol]*nW/pcColX [cCols],
							pcRowY [cRow]*nH/pcRowY [cRows],
							(pcColX [cCol + 1] - pcColX [cCol])*nW/cTotalW,
							(pcRowY [cRow + 1] - pcRowY [cRow])*nH/cTotalH,
							cChannel
						);
					}
				}
			}
		} else
			m_pRenderer->Render (m_MemDC, 0, 0, nW, nH, m_cChannel);

	DrawPresenceOfBookmarks (m_MemDC, nW, nH);

	::BitBlt (_ps.hdc, 0, 0, nW, nH, m_MemDC, 0, 0, SRCCOPY);
}

void CTrackVideo::DrawPresenceOfBookmarks (HDC _hDC, int _nW, int _nH) const {
	const size_t cBookmarks = m_pBookmarks ? m_pBookmarks->GetBookmarksHere () : 0;
	if (cBookmarks != 0) {
		::SelectObject (_hDC, g_hbrNull);
		::SelectObject (_hDC, g_hpenRedThick);
		::Rectangle (_hDC, 1, 1, _nW - 1, _nH - 1);

		int x = 6;
		for (size_t c = 0; c < cBookmarks; ++ c) {
			rgn_t * const pBookmarkRgn = m_pBookmarks->GetBookmarkHere (c);
			if (pBookmarkRgn == NULL || pBookmarkRgn->IsDeleted ())
				continue;
			bookmark_t * const pBookmark = pBookmarkRgn->pExtra->pBookmark;
			CCPtr<CIcon> pIcon = GetBookmarkIcon (pBookmark->strType, pBookmark->strSubType);
			if (pIcon) {
				const size_t w = pIcon->Width ();
				const size_t h = pIcon->Height ();
				pIcon->Draw (_hDC, x, _nH - 6 - h, w, h);
				x += w + 3;
			}

			if (pBookmark->bImagePosDefined) {
				::SelectObject (_hDC, g_hbrNull);
				::SelectObject (_hDC, g_hpenRedThin);

				if (m_bAllChannels) {
					// TODO:
				} else {
					if (m_cChannel == pBookmark->cChannel) {
						const int nWidth  = m_pRenderer->GetWidth  (m_cChannel);
						const int nHeight = m_pRenderer->GetHeight (m_cChannel);

						if (nWidth > 0 && nHeight > 0) { 
							const int ws = __min (nWidth *_nW/nWidth, nWidth *_nH/nHeight);
							const int hs = __min (nHeight*_nW/nWidth, nHeight*_nH/nHeight);

							const int x0 = (_nW - ws)/2;
							const int y0 = (_nH - hs)/2;

							if (pBookmark->cPoints == 1) {
								const int rx = __min (pBookmark->points [1].x*_nW/nWidth, pBookmark->points [1].x*_nH/nHeight);
								const int ry = __min (pBookmark->points [1].y*_nW/nWidth, pBookmark->points [1].y*_nH/nHeight);
								int x = pBookmark->points [0].x;
								int y = pBookmark->points [0].y;

								::Ellipse (
									_hDC,
									x0 + x*ws/nWidth - rx,     y0 + y*hs/nHeight - ry,
									x0 + x*ws/nWidth + rx + 1, y0 + y*hs/nHeight + ry + 1
								);
							} else if (2 <= pBookmark->cPoints && pBookmark->cPoints <= bookmark_t::MAX_POINTS) {
								POINT points [bookmark_t::MAX_POINTS + 1];
								for (size_t c = 0; c < bookmark_t::MAX_POINTS && c < pBookmark->cPoints; ++ c) {
									points [c].x = x0 + pBookmark->points [c].x*ws/nWidth;
									points [c].y = y0 + pBookmark->points [c].y*hs/nHeight;
								}
								points [pBookmark->cPoints] = points [0];

								::Polyline (_hDC, points, pBookmark->cPoints + 1);
							}
						}
					}
				}
			}
		}
	}
}

bool CTrackVideo::OnNotify (int _idCtrl, NMHDR * _pInfo) {
	switch (_pInfo->code) {
		case NM_RCLICK:
			// Suppress context menu of the main window.
			SetMsgResult (TRUE);
			return true;

		case TCN_SELCHANGE: {
			// Tab selected.
			m_cChannel = TabCtrl_GetCurSel (m_hChannel);
			m_bAllChannels = m_cChannel == 0;
			if (m_bAllChannels)
				m_cChannel = -1;
			else
				-- m_cChannel;
				
			_UpdateFrame ();
			::SetFocus (m_hView);
			break;
		}
	}

	return false;
}

enum {MY_WM_TIMER = WM_USER + 10};

INT_PTR CTrackVideo::OnMessage (UINT _uMsg, WPARAM _wParam, LPARAM _lParam) {
	switch (_uMsg) {
/*		case WM_TIMER:
			switch (_wParam) {
				case c_uiTimerRender: _OnTimerRender (); break;
			}
			break;*/
		case MY_WM_TIMER:
			_OnTimerRender (_wParam + 1);
			::InterlockedDecrement (& m_lTimerMessagesInQueue);
			break;

		case WM_ACTIVATE:
			if (LOWORD (_wParam) == WA_ACTIVE || LOWORD (_wParam) == WA_CLICKACTIVE)
				::SetFocus (m_hView);
			break;

		case WM_NEXTDLGCTL:
			::SetFocus (m_hView);
			return FALSE;

		case WM_RBUTTONDOWN:
		case WM_RBUTTONUP:
			// NOTE: suppress context menu of the main window.
			return TRUE;

		case WM_SETCURSOR:
			if (m_bProcessing) {
				::SetCursor (g_hWaitCursor);
				return TRUE;
			}
			break;
	}

	return CDialog::OnMessage (_uMsg, _wParam, _lParam);
}

void CTrackVideo::_PauseAndUpdate () {
	_SetSpeed (0);
	_UpdateTimerRender ();
	_Update ();
}

void CTrackVideo::TogglePlay () {
	SendCommand (m_nSpeed == 0 ? ID_VIDEO_PLAY : ID_VIDEO_PAUSE);
}

void CTrackVideo::OnLButtonDown_View (int _x, int _y) {}
void CTrackVideo::OnMouseMove_View (int _x, int _y) {}
void CTrackVideo::OnLButtonUp_View (int _x, int _y) {}

void CTrackVideo::OnMouseWheel_View (int _nDelta) {
	for (int i = 0; i < ::labs (_nDelta); ++ i)
		SendCommand (_nDelta < 0 ? ID_VIDEO_STEP_FW : ID_VIDEO_STEP_BW);
}

bool CTrackVideo::OnSetCursor_View () {
	if (m_bProcessing) {
		::SetCursor (g_hWaitCursor);
		return true;
	}
	return false;
}

void CTrackVideo::_UpdateTimerRender (UINT _uiLost) {
	if (m_uiTimerID != 0) {
		::timeKillEvent (m_uiTimerID);
		m_uiTimerID = 0;
	}

	if (m_nSpeed != 0) {
		UINT uiPeriod = 1000/m_nFPS;
		if (_uiLost != 0) {
			if (uiPeriod > _uiLost + 5)
				uiPeriod -= _uiLost;
			else if (uiPeriod < _uiLost)
				uiPeriod = 5;
			assert (uiPeriod >= 5);
		}
		m_uiTimerID = ::timeSetEvent (uiPeriod, 0, _TimeProc, (DWORD) this, TIME_ONESHOT | TIME_CALLBACK_FUNCTION);
	}
}

void CALLBACK CTrackVideo::_TimeProc (UINT _uiTimerID, UINT _uMsg, DWORD_PTR _dwUser, DWORD_PTR _dw1, DWORD_PTR _dw2) {
	CTrackVideo * const pThis = reinterpret_cast<CTrackVideo *> (_dwUser);
	if (_uiTimerID == pThis->m_uiTimerID) {
		if (pThis->m_lTimerMessagesInQueue != 0)
			++ pThis->m_lTimerMessagesMissed;
		else {
			::PostMessage (pThis->m_hWnd, MY_WM_TIMER, pThis->m_lTimerMessagesMissed, 0);
			pThis->m_lTimerMessagesMissed = 0;
			::InterlockedIncrement (& pThis->m_lTimerMessagesInQueue);
		}
	}
}

void CTrackVideo::_OnTimerRender (size_t _cTimes) {
	::timeBeginPeriod (1);
	const DWORD dwStartTime = ::timeGetTime ();
	::timeEndPeriod (1);

	if (::IsWindowVisible (m_hWnd)) {
		if (m_pRenderer) {
			if (m_nSpeed > 0)
				m_pRenderer->GoForward  (m_nSpeed*_cTimes);
			else
				m_pRenderer->GoBackward (-m_nSpeed*_cTimes);

			if (m_pRenderer->IsAtStart ())
				_SetSpeed (0);

			if (m_pRenderer->IsAtEnd ()) {
# if 1
				// Stop playing.
				_SetSpeed (0);
# else
				// Loop playing.
				m_pRenderer->SetProgress (0);
				_SetSpeed (4);
# endif
			}
		}

		_Update ();
	}

	::timeBeginPeriod (1);
	const DWORD dwLostTime = ::timeGetTime () - dwStartTime;
	::timeEndPeriod (1);

	_UpdateTimerRender (dwLostTime);
}

void CTrackVideo::_AutoScrollMap () {
	const bool bScroll = _IsTButtonChecked (ID_VIDEO_AUTOSCROLL_MAP);
	SaveSetting ("TrackVideoAutoScroll", bScroll);

	if (bScroll && m_bCursorShow) {
		bool bValid = false;
		rect_t rectVisibleCenter;
		{
			CCritSec::CLock lock (g_cs);

			bValid = ! g_map.rectBound.Invalid ();
			if (bValid) {
				rect_t rectVisible;
				GetVisibleRect (rectVisible);

				rectVisibleCenter.x0 = rectVisible.x0 + rectVisible.Width  ()*.15f;
				rectVisibleCenter.x1 = rectVisible.x0 + rectVisible.Width  ()*.85f;
				rectVisibleCenter.y0 = rectVisible.y0 + rectVisible.Height ()*.15f;
				rectVisibleCenter.y1 = rectVisible.y0 + rectVisible.Height ()*.85f;
			}
		}
		if (bValid) {
			if (! rectVisibleCenter.Contains (m_ptCursor)) {
				SetToCenter (m_ptCursor);
				Refresh ();
			}
		}
	}
}

static
DWORD _EHFilter (void * _ei) {
# ifdef _DEBUG
	return EXCEPTION_CONTINUE_SEARCH;
# else
	if (::IsDebuggerPresent ())
		return EXCEPTION_CONTINUE_SEARCH;

	ErrorMsgBox ("Severe error: exception handled.");

	return EXCEPTION_EXECUTE_HANDLER;
# endif // _DEBUG
}

DWORD WINAPI CTrackVideo::_BookmarkProcessProc (LPVOID _pParams) {
	CTrackVideo * const pThis = reinterpret_cast<CTrackVideo *> (_pParams);
	__try {
		const pfProcess_t pfProcess = pThis->m_pfProcess;
		(pThis->*pfProcess) ();
	} __except (_EHFilter (_exception_info ())) {return 1;}

	return 0;
}

void CTrackVideo::StopProcessing () {
	if (m_bProcessing) {
		m_bStop = true;

		if (m_hProcessingThread)
			::WaitForSingleObject (m_hProcessingThread, 5000);
	}
}

void CTrackVideo::_ProcessInParallel (const char * _strStatus, pfProcess_t _pfProcess) {
	if (! m_pRenderer || ! m_pBookmarks)
		return;

	const size_t cBookmarks0 = m_pBookmarks->GetBookmarks ();

	const DWORD dwProgressPrec_Saved = m_pRenderer->GetProgressPrec ();
	m_pRenderer->SetProgress (0);

	m_pRenderer->Mute (true);

	SetStatus (_strStatus);
	SetProgressStatus (0);

	m_pfProcess = _pfProcess;
	m_bStop = false;
	m_bProcessing = true;
	_UpdateToolBar ();

	{
		CWaiter waiter;

		DWORD dwThreadId = 0;
		CWinHandle hThread = ::CreateThread (NULL, 0, & _BookmarkProcessProc, this, 0, & dwThreadId);
		m_hProcessingThread = hThread;

		WaitThreadForStop (hThread);
		m_hProcessingThread = NULL;
	}
	const bool bTerminatedByClose = m_pRenderer == NULL;

	m_bProcessing = false;
	_UpdateToolBar ();

	if (m_pRenderer) {
		// Restore mute.
		m_pRenderer->Mute (_IsTButtonChecked (ID_VIDEO_MUTE));

		m_pRenderer->SetProgressPrec (dwProgressPrec_Saved);
	}
	m_pBookmarks->SetPositionPrec (dwProgressPrec_Saved);

	_Update ();
	SetStatusReady ();

	if (! bTerminatedByClose)
		InfoMsgBox ("%d bookmark(s) added.", m_pBookmarks->GetBookmarks () - cBookmarks0);

	Refresh ();
}

void CTrackVideo::_DoBookmarkVoices () {
	SetStatusInfo ("");

	if (! m_pRenderer || ! m_pBookmarks)
		return;

	std::vector<std::pair<float, DWORD> > vecLevels;
	double fNetLevel = 0;
	while (! m_pRenderer->IsAtEnd ()) {
		if (m_bStop)
			return;

		const DWORD dwProgressPrec = m_pRenderer->GetProgressPrec ();
		const float fLevel = m_pRenderer->GetSoundLevel ();
		fNetLevel += fLevel;
		vecLevels.push_back (std::make_pair (fLevel, dwProgressPrec));

		const float fProgress = m_pRenderer->GetProgress ();
		SetProgressStatus (fast_floor (fProgress*100 + .5f));
		m_pRenderer->GoForward (1, true);
	}
	const size_t cFrames = vecLevels.size ();
	const double fAveLevel = fNetLevel/cFrames;

	const size_t c_cFPS = m_pRenderer->GetFPS ();

	bool bVoicePresented = false;
	size_t cSilentFramesAfterVoice = 0;
	for (size_t c = 0; c < cFrames; ++ c) {
		if (m_bStop)
			return;

		const float fLevel = vecLevels [c].first;
		const bool bSilence = fLevel < 10*fAveLevel;
		if (! bSilence)
			cSilentFramesAfterVoice = 0;

		if (! bVoicePresented && ! bSilence) {
			bVoicePresented = true;

			const DWORD dwPositionPrec = vecLevels [c].second;
			m_pRenderer->SetProgressPrec (dwPositionPrec);
			m_pBookmarks->SetPositionPrec (dwPositionPrec);
			if (m_pBookmarks->GetBookmarksHere () == 0) {
				rgn_t * const pBookmarkRgn = _AddBookmark ();
				if (pBookmarkRgn != NULL) {
					bookmark_t * const pBookmark = pBookmarkRgn->pExtra->pBookmark;
					pBookmark->strType    = "voice";
					pBookmark->strSubType = "0";
					pBookmark->FixPos (* pBookmarkRgn);
				}
			}
		} else if (bVoicePresented && bSilence) {
			++ cSilentFramesAfterVoice;
			if (cSilentFramesAfterVoice >= c_cFPS*5) // Wait for 5 s of silence.
				bVoicePresented = false;
		}
	}
}

void CTrackVideo::_DoBookmarkRoadSigns () {
	SetStatusInfo ("");
	
	if (! m_pRenderer || m_dwChannelsToScan == 0)
		return;

	const size_t cChannels = m_pRenderer->GetChannels ();
	
	// Allocate bitmaps.
	CDIBitmap ** const ppBitmaps = reinterpret_cast<CDIBitmap **> (alloca (sizeof (CDIBitmap *)*cChannels));
	size_t cChannel;
	for (cChannel = 0; cChannel < cChannels; ++ cChannel) {
		const size_t cWidth  = m_pRenderer->GetWidth  (cChannel);
		const size_t cHeight = m_pRenderer->GetHeight (cChannel);

		ppBitmaps [cChannel] = new CDIBitmap (cWidth, cHeight);
	}

	size_t cFrames = 0;
	size_t cRoadSigns = 0;
	const DWORD dwStart = ::GetTickCount ();

	while (! m_pRenderer->IsAtEnd ()) {
		if (m_bStop)
			return;

		const DWORD dwProgressPrec = m_pRenderer->GetProgressPrec ();

		point_t ptDummy;
		trackpoint_info_t tpi;
		m_pRenderer->GetVehicleInfo (ptDummy, tpi);

		for (cChannel = 0; cChannel < cChannels; ++ cChannel) {
			if (m_bStop)
				return;
				
			if ((m_dwChannelsToScan & (1 << cChannel)) == 0)
				continue;

			CDIBitmap * const pBitmap = ppBitmaps [cChannel];
			m_pRenderer->CaptureDIBitmap (cChannel, pBitmap->GetBits (), pBitmap->GetBitmapInfo ());

			bookmark_t bookmark;
			const bool bRoadSignDetected = false;/*DetectRoadSign (
				pBitmap->GetBits (), pBitmap->Width (), pBitmap->Height (),
				tpi, bookmark
			);*/
			bookmark.cChannel = cChannel;
			
			++ cFrames;
			if (bRoadSignDetected)
				++ cRoadSigns;
				
			// Show scan progress.
			const DWORD dwElapsed = ::GetTickCount () - dwStart;
			const float fSPerFrame = float (dwElapsed)/cFrames/1000;
			char strStatus [64];
			::sprintf (strStatus, "%d frame(s) scanned, %d sign(s) detected. %0.1f s/frame", cFrames, cRoadSigns, fSPerFrame);
			SetStatusInfo (strStatus);
			
			if (! bRoadSignDetected)
				continue;

			// Save bookmark, if it is not duplicated.
			m_pBookmarks->SetPositionPrec (dwProgressPrec);
			if (m_pBookmarks->GetBookmarksHere () == 0) {
				rgn_t * const pBookmarkRgn = _AddBookmark ();
				if (pBookmarkRgn != NULL) {
					bookmark_t * const pBookmark = pBookmarkRgn->pExtra->pBookmark;

					pBookmark->strType          = "comment";
					pBookmark->strSubType       = "info";

					pBookmark->bImagePosDefined = true;
					pBookmark->cChannel         = bookmark.cChannel;
					pBookmark->cPoints          = bookmark.cPoints;
					::memcpy (pBookmark->points, bookmark.points, bookmark_t::MAX_POINTS*sizeof (POINT));

					pBookmark->attributes ["comment"] = "<road sign data>";

					pBookmark->FixPos (* pBookmarkRgn);
				}
			}
		}
		
		const float fProgress = m_pRenderer->GetProgress ();
		SetProgressStatus (fast_floor (fProgress*100 + .5f));
		
		m_pRenderer->GoForward (cChannels, false);
	}

	// Free bitmaps.
	for (cChannel = 0; cChannel < cChannels; ++ cChannel)
		delete ppBitmaps [cChannel];
}

//////////////////////////////////////////////////////////////////////////

void ShowTrackVideo (const rgn_t & _rgn, DWORD _dwProgressPrec) {
	if (! _rgn.IsTrack () || _rgn.pFileRef == NULL || _rgn.pFileRef->pTrackVideo == NULL)
		return;

	ShowModeless (CTrackVideo::g_pTrackVideo, g_hMapWnd);
	if (CTrackVideo::g_pTrackVideo)
		CTrackVideo::g_pTrackVideo->GoFrame (_rgn, & _rgn.pFileRef->pTrackVideo.data (), _dwProgressPrec);
}

void ShowTrackVideo (const rgn_t & _rgn, const point_t & _pt) {
	if (! _rgn.IsTrack () || _rgn.pFileRef == NULL || _rgn.pFileRef->pTrackVideo == NULL)
		return;
	const rgn_element_t & element = _rgn.elements [0];
	if (element.pExtra == NULL || element.pExtra->PointExtras.size ())
		return;

	const float c_fE = c_fE_in_pixels/g_kY;
	const float fCosY = g_map.fCosY;

	const points_t & points = element.points;
	const size_t cPoints = points.size ();

	//
	// Find all the nearest points.
	//

	std::vector<size_t> NearestPoints;
	bool bNear = false;
	for (size_t c = 0; c + 1 < cPoints; ++ c) {
		point_t ptNearest;
		const float fD2 = DistanceToSegment2 (points [c], points [c + 1], _pt, fCosY, ptNearest);
		if (fD2 > c_fE*c_fE) {
			bNear = false;
			continue;
		}

		if (bNear)
			continue;

		NearestPoints.push_back (c);
		bNear = true;
	}

	const size_t cNearestPoints = NearestPoints.size ();
	if (cNearestPoints == 0)
		return;

	size_t cPoint = -1;
	if (cNearestPoints == 1) {
		cPoint = NearestPoints [0];
	} else {
		assert (NearestPoints.size () >= 2);

		//
		// Show menu.
		//

		CPopupMenu hMenu;
		const int c_nCmdId0 = 100;

		for (size_t c = 0; c < cNearestPoints; ++ c) {
			const size_t cPoint = NearestPoints [c];
			const point_t & pt0 = points [cPoint != 0          ? cPoint - 1 : 0];
			const point_t & pt1 = points [cPoint + 1 < cPoints ? cPoint + 1 : cPoint];

			float fHeading = 90 - ::atan2 (pt1.y - pt0.y, (pt1.x - pt0.x)*fCosY)*180/c_PI;
			if (fHeading < 0)
				fHeading += 360;
			char strHeading [16];
			PrintAzimuth (strHeading, fHeading);

			char strItem [64];
			::sprintf (strItem, "%s (#%d)", strHeading, cPoint);

			hMenu.AppendItem (c_nCmdId0 + c, strItem);
		}

		const long lX = XToScreen (_pt.x);
		const long lY = YToScreen (_pt.y);
		RECT rect = {lX, lY, lX + 1, lY + 1};
		::MapWindowPoints (g_hMapWnd, HWND_DESKTOP, (LPPOINT) & rect, 2);

		const int nCmdId = hMenu.ShowAsPopup (g_hWnd, rect);
		if (nCmdId == 0)
			// NOTE: menu is cancelled.
			return;
		const int nIdx = nCmdId - c_nCmdId0;
		if (0 <= nIdx && nIdx < (int) NearestPoints.size ())
			cPoint = NearestPoints [nIdx];
		else
			return;
	}

	//
	// Show video box.
	//

	if (cPoint >= element.pExtra->TrackPoints.size ())
		return;
	const trackpoint_info_t & tpi = element.pExtra->TrackPoints [cPoint];
	if ((tpi.wFlags & trackpoint_info_t::defVideo) == 0)
		return;

	ShowModeless (CTrackVideo::g_pTrackVideo, g_hMapWnd);
	if (CTrackVideo::g_pTrackVideo)
		CTrackVideo::g_pTrackVideo->GoFrame (_rgn, & _rgn.pFileRef->pTrackVideo.data (), tpi.cVideoData0, tpi.cVideoData1);
}

void ResetTrackVideo () {
	if (CTrackVideo::g_pTrackVideo)
		CTrackVideo::g_pTrackVideo->Reset ();
}

INVOKE_BEFORE_REMOVE_MAP (ResetTrackVideo);
INVOKE_ON_BACKUP_MAP     (ResetTrackVideo);

void CleanTrackVideo () {
	if (CTrackVideo::g_pTrackVideo)
		CTrackVideo::g_pTrackVideo->Clean ();
}

INVOKE_BEFORE_GC (CleanTrackVideo);