//
// GPSMapEdit
// (c) Konstantin Galichsky (kg@geopainting.com), 2002-2009
//
// Positioning stuff.
//

# include "StdAfx.h"
# include "Tracking.h"
# include "Positioning.h"
///# include "UBXProtocol.h"
# include "Map.h"
# include "MapView.h"
# include "NewMap.h"
# include "Types.h"
# include "Position.h"
# include "FileWriter.h"
# include "Settings.h"
# include "Wrappers.h"
# include "Globals.h"

# define WRITE_TRKS_TO_MEMORY

CTrackWriter::~CTrackWriter () {
	if (m_pFileWriter) {
		m_pFileWriter->FlushBuffer ();
		delete m_pFileWriter;
	}

	if (m_hTrackFile)
		::CloseHandle (m_hTrackFile);
}

void CTrackWriter::Init (HANDLE _hTrackFile) {
	assert (m_pFileWriter == NULL);

	m_hTrackFile = _hTrackFile;
	m_pFileWriter = new CFileWriter (m_hTrackFile, 1024);
}

/////////////////////////////////////////////

CTrackWriterTypeInfo::CTrackWriterTypeInfo (
	pfCreateTrackWriter_t _pfCreateTrackWriter,
	const char * _strExt, const char * _strDisplayName
) :
	m_pfCreateTrackWriter (_pfCreateTrackWriter),
	m_strExt (_strExt),
	m_strDisplayName (_strDisplayName)
{
	// Insert to the head of the list.
	m_pNext = CTrackWriterTypeInfo::s_pFirst;
	CTrackWriterTypeInfo::s_pFirst = this;
}

CTrackWriterTypeInfo * CTrackWriterTypeInfo::s_pFirst;

CTrackWriterTypeInfo * CTrackWriterTypeInfo::GetCurrent () {
	char strExt [16];
	GetSetting ("TrackWriter", strExt, ARRAY_SIZE (strExt), "mps");

	for (CTrackWriterTypeInfo * p = CTrackWriterTypeInfo::s_pFirst; p != NULL; p = p->m_pNext) {
		if (::strcmp (p->m_strExt, strExt) == 0)
			return p;
	}

	return CTrackWriterTypeInfo::s_pFirst;
}

/////////////////////////////////////////////

void GetTracksFolder (char _strTracksFolder [MAX_PATH]) {
	// Get folder with mapedit.exe.
	char strProcessName [MAX_PATH + 1];
	::GetModuleFileName (NULL, strProcessName, MAX_PATH);
	char * const pPathEnd = ::strrchr (strProcessName, '\\');
	if (pPathEnd)
		* pPathEnd = '\0';

	::sprintf (_strTracksFolder, "%s\\Tracks", strProcessName);
	// Make sure that 'Tracks' folder exists.
	if (::GetFileAttributes (_strTracksFolder) != FILE_ATTRIBUTE_DIRECTORY)
		::CreateDirectory (_strTracksFolder, NULL);
}

/////////////////////////////////////////////

static CCritSec g_csTrackWriter;
static CTrackWriter * g_pTrackWriter;

# if defined (WRITE_TRKS_TO_MEMORY)
static CCritSec g_csTracks;
static std::vector<points_t> g_Tracks;
# else
static file_ref_t * g_pFileRef = NULL;
static rgn_t * g_pTrackRgn = NULL;
# endif

void StartTracking () {
	if (g_pTrackWriter)
		StopTracking ();

///	CUBXProtocol::m_bMayOpenBinFile = true;

	CTrackWriterTypeInfo * const pTWTI = CTrackWriterTypeInfo::GetCurrent ();
	if (pTWTI == NULL) {
		ErrorMsgBox ("Failed to write tracks.");
		return;
	}

	char strTracksFolder [MAX_PATH];
	GetTracksFolder (strTracksFolder);

	// Get current time.
	SYSTEMTIME lt;
	::GetLocalTime (& lt);

	// Prepare name of the track file.
	char strFileName [MAX_PATH + 1];
	::sprintf (strFileName, "%s\\%d-%02d-%02d_%02d-%02d-%02d.%s", strTracksFolder, lt.wYear, lt.wMonth, lt.wDay, lt.wHour, lt.wMinute, lt.wSecond, pTWTI->m_strExt);

	CFileForWrite hTrackFile (strFileName);
	if (! hTrackFile.IsValid ()) {
		ErrorMsgBox ("Can't create file to write tracks:\n %s", strFileName);
		return;
	}

	// Start writing.
	{
		CCritSec::CLock lock (g_csTrackWriter);

		g_pTrackWriter = pTWTI->m_pfCreateTrackWriter ();
		g_pTrackWriter->Init (hTrackFile.Detach ());
		g_pTrackWriter->Start ();
	}

# if ! defined (WRITE_TRKS_TO_MEMORY)
	// Prepare to hot visualization.
	g_map.file_refs.push_back (file_ref_t ());
	g_pFileRef = & g_map.file_refs.back ();

	g_pFileRef->bAttachment = true;
	g_pFileRef->bDeleted    = false;
	g_pFileRef->bVisible    = true;
	g_pFileRef->strFileName = strFileName;
# endif
	
	StartNewTrack ();
}

void StopTracking () {
///	if (CUBXProtocol::m_pBinFileWriter != NULL)
///		CUBXProtocol::CloseBinFile ();

	CCritSec::CLock lock (g_csTrackWriter);

	if (g_pTrackWriter == NULL)
		return;
	g_pTrackWriter->Stop ();

	delete g_pTrackWriter;
	g_pTrackWriter = NULL;
}

static
void _ResetTracking () {
# if ! defined (WRITE_TRKS_TO_MEMORY)
	g_pFileRef  = NULL;
	g_pTrackRgn = NULL;
# endif
}

INVOKE_AFTER_REMOVE_MAP (_ResetTracking);

void StartNewTrack () {
	{
		CCritSec::CLock lock (g_csTrackWriter);

		if (g_pTrackWriter == NULL)
			return;
		g_pTrackWriter->StartNewTrack ();
	}

# if defined (WRITE_TRKS_TO_MEMORY)
	{
		CCritSec::CLock lock (g_csTracks);

		g_Tracks.push_back (points_t ());
	}
# else
	if (g_pFileRef) {
		g_pTrackRgn = & g_map.CreateRgn (rkPolyline);
		g_pTrackRgn->SetTrackType ();
		g_pTrackRgn->pFileRef = g_pFileRef;
		//g_pTrackRgn->strLabel = strTrackName;

		g_pTrackRgn->elements.push_back (rgn_element_t ());
		rgn_element_t & element = g_pTrackRgn->elements.back ();
		element.cLevel   = 0;
		element.bDeleted = false;
	}
# endif
}

void AddPointToTrack (const PositionInfo_t & _pi) {
	{
		CCritSec::CLock lock (g_csTrackWriter);

		if (g_pTrackWriter == NULL)
			return;
		g_pTrackWriter->AddPointToTrack (_pi);
	}

	bool bRefresh = false;

	point_t pt (_pi.fX, _pi.fY);

# if defined (WRITE_TRKS_TO_MEMORY)
	{
		CCritSec::CLock lock (g_csTracks);

		const size_t cTracks = g_Tracks.size ();
		if (cTracks != 0) {
			points_t & track = g_Tracks [cTracks - 1];
			track.push_back (pt);

			bRefresh = true;
		}
	}
# else
	// Visualize new point.
	if (g_pTrackRgn) {
		rgn_element_t & element = g_pTrackRgn->elements [0];
		element.points.push_back (pt);

		g_pTrackRgn->RefreshAttributes (g_map);

		bRefresh = true;
	}
# endif

	rect_t rect (pt.x, pt.y, pt.x, pt.y);
	ExtendMapBoundingRect (rect, false);

	if (bRefresh)
		// TODO: invoke it from GUI thread.
		DelayedRefresh (500);
}

//////////////////////////////////////////////////////

# include "Position.h"
# include "Options.h"

void DrawTracks (HDC _hDC, const rect_t & _rectVisible) {
# if defined (WRITE_TRKS_TO_MEMORY)
	const HPEN hpenTrack = ::CreatePen (PS_SOLID, GetDefaultTrackWidth (), GetDefaultTrackColor ());
	::SelectObject (_hDC, hpenTrack);

	{
		CCritSec::CLock lock (g_csTracks);

		const size_t cTracks = g_Tracks.size ();
		for (size_t cTrack = 0; cTrack < cTracks; ++ cTrack) {
			const points_t & track = g_Tracks [cTrack];

			const size_t cPoints = track.size ();
			if (cPoints > 0) { 
				std::vector<POINT> pts (cPoints);
				for (size_t c = 0; c < cPoints; ++ c) {
					const point_t & pt = track [c];
					pts [c].x = XToScreen (pt.x);
					pts [c].y = YToScreen (pt.y);
				}

				::Polyline (_hDC, & * pts.begin (), cPoints);
			}
		}
	}

	::SelectObject (_hDC, g_hpenBlackThin);
	::DeleteObject (hpenTrack);
# endif
}