//
// GPSMapEdit
// (c) Konstantin Galichsky (kg@geopainting.com), 2002-2009
//
// Transverse Mercator projection.
//

# include "StdAfx.h"
# include "TransverseMercator.h"
# include "Ellipsoid.h"
# include "PtAlgo.h"
# include "fast_floor.h"
# include "Globals.h"

//
// Translate from TM grid (meters) to latitude/longitude degrees.
// The formula is given from http://posc.org/Epicentre.2_2/DataModel/ExamplesofUsage/eu_cs34h.html
//
void TM2LL (const CEllipsoid & _e, double _k0, double _FE, double _FN, double _fLon0, double _M0, double _x, double _y, double & _fLon, double & _fLat, bool _bSouthOriented) {
	const double NS = _bSouthOriented ? -(_y - _FN) : (_y - _FN);
	const double EW = _bSouthOriented ? -(_x - _FE) : (_x - _FE);

	const double M1 = _M0 + NS/_k0;

	const double mu1 = M1/(_e.a*(1 - _e.e2*(1./4 + _e.e2*(3./64 + 5./256*_e.e2))));
	const double fSin2Mu1 = ::sin (2*mu1);
	const double fCos2Mu1 = ::cos (2*mu1);

	const double e1_2 = _e.e1*_e.e1;
	const double e1_3 = e1_2*_e.e1;
	const double fi1 = mu1 +
		_e.e1*fSin2Mu1*(
			( 3./2        - 27./32*e1_2) + 
			(21./16*_e.e1 - 55./32*e1_3)*2*fCos2Mu1 +
			(              151./96*e1_2)*(3 - 4*fSin2Mu1*fSin2Mu1) +
			(            1097./512*e1_3)*4*fCos2Mu1*(1 - 2*fSin2Mu1*fSin2Mu1)
		);
	const double fSinFi1 = ::sin (fi1);
	const double fCosFi1 = ::cos (fi1);
	const double fTanFi1 = fSinFi1/fCosFi1;

	const double fEFi      = 1 - _e.e2*fSinFi1*fSinFi1;
	const double fSqrtEFi2 = ::sqrt (fEFi);

	const double D  = EW*fSqrtEFi2/(_e.a*_k0);
	const double D2 = D*D;
	const double T1 = fTanFi1*fTanFi1;
	const double C1 = _e.ee2*fCosFi1*fCosFi1;

	const double fLat = fi1 - fEFi*fTanFi1/(1 - _e.e2)*D2*(
		1./2 + D2*(
			- (5 + 3*T1 + C1*(10 - 4*C1) - 9*_e.ee2)/24 +
			(61 + T1*(90 + 45*T1) + C1*(298 - 3*C1) - 252*_e.ee2)*D2/720
		)
	);

	const double fLon = D*(
		1 + D2*(
			- (1 + 2*T1 + C1)/6 +
			(5 - C1*(2 + 3*C1) + T1*(28 + 24*T1) + 8*_e.ee2)*D2/120
		)
	)/fCosFi1;

	_fLat = fLat*(180/c_PId);
	_fLon = fLon*(180/c_PId) + _fLon0;
}

static
double GetM0 (const CEllipsoid & _e, double _y0) {
	const double fi = _y0*c_PId/180;
	return _e.a*(
		(1 -   _e.e2/4 -  3*_e.e2*_e.e2/64  -  5*_e.e2*_e.e2*_e.e2/256)*fi -
		(    3*_e.e2/8 +  3*_e.e2*_e.e2/32  + 45*_e.e2*_e.e2*_e.e2/1024)*::sin (2*fi) +
		(                15*_e.e2*_e.e2/256 + 45*_e.e2*_e.e2*_e.e2/1024)*::sin (4*fi) -
		(                                     35*_e.e2*_e.e2*_e.e2/3072)*::sin (6*fi)
	);
}

//
// Translate from latitude/longitude degrees to TM grid (meters).
// The formula is given from http://posc.org/Epicentre.2_2/DataModel/ExamplesofUsage/eu_cs34h.html
//
void LL2TM (const CEllipsoid & _e, double _k0, double _FE, double _FN, double _fLon0, double _M0, double _fLon, double _fLat, double & _x, double & _y) {
	// (x is lambda, y is phi)
	const double Yr  = _fLat *c_PId/180;
	const double Xr  = _fLon *c_PId/180;
	const double X0r = _fLon0*c_PId/180;

	const double fCosY = ::cos (Yr);
	const double fSinY = ::sin (Yr);
	const double fTanY = fSinY/fCosY;
	const double T = fTanY*fTanY;

	const double C = (_e.e2/(1 - _e.e2))*fCosY*fCosY;
	const double A = (Xr - X0r)*fCosY;
	const double A2 = A*A;
	const double A4 = A2*A2;

	const double M = GetM0 (_e, _fLat);

	const double nu = _e.a/::sqrt (1 - _e.e2*fSinY*fSinY);

	_x = _FE + _k0*nu*(A + (1 - T + C)*A2*A/6 + (5 - 18*T + T*T + 72*C - 58*_e.ee2)*A4*A/120);

	_y = _FN + _k0*(
		M - _M0 + nu*fTanY*(
			A2/2 +
			(5  -    T       +   9*C + 4*C*C     )*A4/24 +
			(61 - 58*T + T*T + 600*C - 330*_e.ee2)*A4*A2/720
		)
	);
}

////////////////////////////////////////////////////

CTransverseMercatorToLL::CTransverseMercatorToLL (const CEllipsoid & _e, double _k0, double _FE, double _FN, double _fLon0, double _fLat0, bool _bSouthOriented) :
	m_Ellipsoid (_e),
	m_k0 (_k0),
	m_FE (_FE),
	m_FN (_FN),
	m_fLon0 (_fLon0),
	m_M0 (GetM0 (_e, _fLat0)),
	m_bSouthOriented (_bSouthOriented)
{}

void CTransverseMercatorToLL::ApplyTo (point_t & _point) const {
	double fLat, fLon;
	TM2LL (m_Ellipsoid, m_k0, m_FE, m_FN, m_fLon0, m_M0, _point.x, _point.y, fLon, fLat, m_bSouthOriented);
	_point.x = fLon;
	_point.y = fLat;
}

void CTransverseMercatorToLL::Apply (double _x, double _y, double & _fLon, double & _fLat) const {
	TM2LL (m_Ellipsoid, m_k0, m_FE, m_FN, m_fLon0, m_M0, _x, _y, _fLon, _fLat, m_bSouthOriented);
}

////////////////////////////////////////////////////

CLLToTransverseMercator::CLLToTransverseMercator (const CEllipsoid & _e, double _k0, double _FE, double _FN, double _fLon0, double _fLat0) :
	m_Ellipsoid (_e),
	m_k0 (_k0),
	m_FE (_FE),
	m_FN (_FN),
	m_fLon0 (_fLon0),
	m_M0 (GetM0 (_e, _fLat0))
{}

void CLLToTransverseMercator::ApplyTo (point_t & _point) const {
	double x, y;
	LL2TM (m_Ellipsoid, m_k0, m_FE, m_FN, m_fLon0, m_M0, _point.x, _point.y, x, y);
	_point.x = x;
	_point.y = y;
}

void CLLToTransverseMercator::Apply (double _fLon, double _fLat, double & _x, double & _y) const {
	LL2TM (m_Ellipsoid, m_k0, m_FE, m_FN, m_fLon0, m_M0, _fLon, _fLat, _x, _y);
}

//////////////////////////////////

CUTM_North::CUTM_North (const CEllipsoid & _e, size_t _cZone) :
	CTransverseMercatorToLL (_e, 0.9996, 500000, 0, _cZone*6. - 3 - 180, 0)
{
	assert (1 <= _cZone && _cZone <= 60);
}

CLL2UTM_North::CLL2UTM_North (const CEllipsoid & _e, size_t _cZone) :
	CLLToTransverseMercator (_e, 0.9996, 500000, 0, _cZone*6. - 3 - 180, 0)
{
	assert (1 <= _cZone && _cZone <= 60);
}

CUTM_South::CUTM_South (const CEllipsoid & _e, size_t _cZone) :
	CTransverseMercatorToLL (_e, 0.9996, 500000, 10000000, _cZone*6. - 3 - 180, 0)
{
	assert (1 <= _cZone && _cZone <= 60);
}

CLL2UTM_South::CLL2UTM_South (const CEllipsoid & _e, size_t _cZone) :
	CLLToTransverseMercator (_e, 0.9996, 500000, 10000000, _cZone*6. - 3 - 180, 0)
{
	assert (1 <= _cZone && _cZone <= 60);
}

////////////////////////////////////

CBritishNationalGrid::CBritishNationalGrid     () : CTransverseMercatorToLL (c_eAiry1830,      0.9996012717, 400000,  -100000,  -2,           49) {}
// http://www.osi.ie/gps/overview/irishgrid.asp
CIrishGrid::CIrishGrid                         () : CTransverseMercatorToLL (c_eAiryMod,       1.000035,     200000,   250000,  -8,           53.5) {}
// http://www.irishscientist.ie/2001/contents.asp?contentxml=01p66.xml&contentxsl=IS01pages.xsl
CIrishTM::CIrishTM                             () : CTransverseMercatorToLL (c_eGRS80,         0.999820,     600000,   750000,  -8,           53.5) {}
CSwissCoordinateSystem::CSwissCoordinateSystem () : CTransverseMercatorToLL (c_eBessel1841,    1,            600000,   200000, 7 + 26./60 + 22.5/3600, 46 + 57./60 + 8.66/3600) {}
CSwedishRT90_25GonW::CSwedishRT90_25GonW       () : CTransverseMercatorToLL (c_eBessel1841,    1,           1500000,        0, 15.808277778,  0) {}
CItalyGrid1::CItalyGrid1                       () : CTransverseMercatorToLL (c_eInternational, 0.9996,      1500000,        0,   9,           0) {}
CItalyGrid2::CItalyGrid2                       () : CTransverseMercatorToLL (c_eInternational, 0.9996,      2520000,        0,  15,           0) {}
// http://www.linz.govt.nz/geodetic/datums-projections-heights/projections/new-zealand-transverse-mercator-2000/index.aspx
CNZTM2000::CNZTM2000                           () : CTransverseMercatorToLL (c_eGRS80,         0.9996,      1600000, 10000000, 173,           0) {}
CDutchRDGrid::CDutchRDGrid                     () : CTransverseMercatorToLL (c_eBessel1841,    0.9999079,    155000, -5316592, 5 + 23.258/60, 0) {}

CDHDNGrid::CDHDNGrid (float _fAveX) :
	CTransverseMercatorToLL (c_eBessel1841, 1, 500000 + fast_floor (_fAveX/1000000)*1000000, 0, fast_floor (_fAveX/1000000)*3, 0)
{}

//////////////////////////////////

# ifdef _DEBUG

# include "S42.h"
# include "Datum.h"

void Test_TM2LL () {
	// OSGB 1936 / British National Grid
	CBritishNationalGrid bng;

	// S42
	const CEllipsoid eKrasovsky1940 ("", 6378245., 1/298.3);
	CTransverseMercatorToLL sk42 (eKrasovsky1940, 1, 500000, 0, -3, 0);

	////////////////////////////

	point_t p (577274.989f, 69740.497f);
	bng.ApplyTo (p);

	assert (::fabs (p.x -  0.5f) < 1e-5);
	assert (::fabs (p.y - 50.5f) < 1e-5);

	////////////////////////////

// 6+15/60.+17.46/60./60., 53+25/60.+27.66/60./60.
// O(300 km, 200 km) 16000 43000
	CIrishGrid ig;
	point_t pIG (316000, 243000);
	ig.ApplyTo (pIG);

	assert (::fabs (pIG.x - -(6+15/60.+17.46/60./60.)) < 6e-5);
	assert (::fabs (pIG.y - (53+25/60.+27.66/60./60.)) < 1e-5);

	////////////////////////////

	point_t p1 (700000, 7000000);
	sk42.ApplyTo (p1);

	CS42Grid gk (200000);
	point_t p2 (700000, 7000000);
	gk.ApplyTo (p2);

	assert (::fabs (p1.x - p2.x) < 1e-5);
	assert (::fabs (p1.y - p2.y) < 1e-5);

	CLLToTransverseMercator sk42r (eKrasovsky1940, 1, 500000, 0, -3, 0);
	sk42r.ApplyTo (p1);

	assert (::fabs (p1.x -  700000) < 1);
	assert (::fabs (p1.y - 7000000) < 1);

	////////////////////////////

	point_t pDHDN (4417702.5, 5697747);
	CDHDNGrid dhdn (pDHDN.x);
	dhdn.ApplyTo (pDHDN);
	CMolodenskyTransformation (c_eBessel1841, c_eWGS84, 606, 23, 413).ApplyTo (pDHDN);
	assert (::fabs (pDHDN.x - 10.81570) < 2e-5);
	assert (::fabs (pDHDN.y - 51.40844) < 4e-5);

	////////////////////////////
/*
	const DWORD dwStartTime1 = ::GetTickCount ();
	size_t c;
	for (c = 0; c < 1000000; ++ c) {
		point_t p (700000, 7000000);
		sk42.ApplyTo (p);
	}
	const DWORD dwEndTime1 = ::GetTickCount ();

	const DWORD dwStartTime2 = ::GetTickCount ();
	for (c = 0; c < 1000000; ++ c) {
		point_t p (700000, 7000000);
		gk.ApplyTo (p);
	}
	const DWORD dwEndTime2 = ::GetTickCount ();

	ErrorMsgBox (
		"SK42=%.2f\n"
		"GK  =%.2f\n",
		(dwEndTime1 - dwStartTime1)/1000.,
		(dwEndTime2 - dwStartTime2)/1000.
	);*/

	///////////////////////////////////////

/*	http://www.lm.se/geodesi/refsys/eng/refsys-eng.htm

Latitude, longitude and height above the ellipsoid GRS 80 in SWEREF 99:
 = 58 00' 00.000000"    = 17 00' 00.000000"   h  = 30.000 
transformation to latitude, longitude and height above the ellipsoid Bessel 1841 in RT 90:
 = 58 00' 01.213296"    = 17 00' 11.683659"   h  = -5.3970 
and projected to plane grid coordinates in RT 90 2.5 gon V 0:-15:
northing  = 6431274.6309 m   easting  = 1570650.2449 m 

*/
	point_t pt (1570650.2449f, 6431274.6309f);
	CSwedishRT90_25GonW ().ApplyTo (pt);

	const CTransformationBase * pRT90 = NULL;
	FindOziDatumToWGS84 ("RT 90", pRT90);
	pRT90->ApplyTo (pt);

	assert (::fabs (pt.x - 17.0f) < 1e-5);
	assert (::fabs (pt.y - 58.0f) < 1e-5);
}

# endif // _DEBUG
