//
// GPSMapEdit
// (c) Konstantin Galichsky (kg@geopainting.com), 2009
//
// List of turn restrictions.
//

# include "StdAfx.h"
# include "Resource.h"
# include "TurnRestrictions.h"
# include "TimePeriod.h"
# include "Position.h"
# include "MapView.h"
# include "NodeProperties.h"
# include "BackupAction.h"
# include "NewMap.h"
# include "Status.h"
# include "Window.h"
# include "Dialog.h"
# include "ListView.h"
# include "Settings.h"
# include "Globals.h"

enum {
	trNode     = 0,
	trFromRoad = 1,
	trToRoad   = 2,
	trVehicles = 3,
	trTime     = 4
};

namespace {
	struct LessNode {
		LessNode (bool _bAscending) : m_bAscending (_bAscending) {}

		bool operator () (const rgrestriction_t & _tr1, const rgrestriction_t & _tr2) {
			const DWORD dw1 = _tr1.pNode ? _tr1.pNode->dwNodeID : 0;
			const DWORD dw2 = _tr2.pNode ? _tr2.pNode->dwNodeID : 0;
			const int iCmp = dw1 - dw2;
			return m_bAscending ? iCmp < 0 : iCmp > 0;
		}

	  private:
		const bool m_bAscending;
	};

	struct LessFromRoad {
		LessFromRoad (bool _bAscending) : m_bAscending (_bAscending) {}

		bool operator () (const rgrestriction_t & _tr1, const rgrestriction_t & _tr2) {
			const DWORD dw1 = _tr1.pFromRoad && _tr1.pFromRoad->pExtra ? _tr1.pFromRoad->pExtra->dwRoadID : 0;
			const DWORD dw2 = _tr2.pFromRoad && _tr2.pFromRoad->pExtra ? _tr2.pFromRoad->pExtra->dwRoadID : 0;
			const int iCmp = dw1 - dw2;
			return m_bAscending ? iCmp < 0 : iCmp > 0;
		}

	  private:
		const bool m_bAscending;
	};

	struct LessToRoad {
		LessToRoad (bool _bAscending) : m_bAscending (_bAscending) {}

		bool operator () (const rgrestriction_t & _tr1, const rgrestriction_t & _tr2) {
			const DWORD dw1 = _tr1.pToRoad && _tr1.pToRoad->pExtra ? _tr1.pToRoad->pExtra->dwRoadID : 0;
			const DWORD dw2 = _tr2.pToRoad && _tr2.pToRoad->pExtra ? _tr2.pToRoad->pExtra->dwRoadID : 0;
			const int iCmp = dw1 - dw2;
			return m_bAscending ? iCmp < 0 : iCmp > 0;
		}

	  private:
		const bool m_bAscending;
	};

	struct LessVehicles {
		LessVehicles (bool _bAscending) : m_bAscending (_bAscending) {}

		bool operator () (const rgrestriction_t & _tr1, const rgrestriction_t & _tr2) {
			const int iCmp = _tr1.btVehicleKinds - _tr2.btVehicleKinds;
			return m_bAscending ? iCmp < 0 : iCmp > 0;
		}

	  private:
		const bool m_bAscending;
	};

	struct LessTime {
		LessTime (bool _bAscending) : m_bAscending (_bAscending) {}

		bool operator () (const rgrestriction_t & _tr1, const rgrestriction_t & _tr2) {
			const int nSize1 = _tr1.TimePeriods.size ();
			const int nSize2 = _tr2.TimePeriods.size ();
			const int iCmpSize = nSize1 - nSize2;
			if (iCmpSize != 0)
				return m_bAscending ? iCmpSize < 0 : iCmpSize > 0;

			for (size_t cPeriod = 0; cPeriod < nSize1; ++ cPeriod) {
				const time_period_t & tp1 = _tr1.TimePeriods [cPeriod];
				const time_period_t & tp2 = _tr2.TimePeriods [cPeriod];

				const DWORD dw1 = (DWORD (tp1.btFromHour) << 24) + (DWORD (tp1.btFromMin) << 16) + (DWORD (tp1.btToHour) << 8) + tp1.btToMin;
				const DWORD dw2 = (DWORD (tp2.btFromHour) << 24) + (DWORD (tp2.btFromMin) << 16) + (DWORD (tp2.btToHour) << 8) + tp2.btToMin;
				const int iCmpTime = int (dw1) - int (dw2);
				if (iCmpTime != 0)
					return m_bAscending ? iCmpTime < 0 : iCmpTime > 0;

				const int iCmpDW = int (tp1.btDaysOfWeek) - int (tp2.btDaysOfWeek);
				return m_bAscending ? iCmpDW < 0 : iCmpDW > 0;
			}

			return false;
		}

	  private:
		const bool m_bAscending;
	};
}

/////////////////////////////////////////////////////

class CTurnRestrictions : public CDialog {
  public:
	CTurnRestrictions () : CDialog (IDD_TURN_RESTRICTIONS) {}

	static CTurnRestrictions * g_pTurnRestrictions;

	// Override.
	virtual bool OnInit (HWND _hwndFocus);
	virtual void OnDestroy ();
	virtual void OnCommand (WORD _wCmd, WORD _wNotifyCode, HWND _hwndCtl);
	virtual bool OnNotify (int _idCtrl, NMHDR * _pInfo);
	virtual void OnMove (int _x, int _y);

	void Reset ();

  private:
	CListView m_hList;
	typedef std::vector<rgrestriction_t> trs_t;
	trs_t m_TRs;

	void _GetItemText (const rgrestriction_t & _restr, size_t _cColumn, char * _strBuf, size_t _cSize) const;
	void _Sort ();
	void _LocateSel (bool _bShowProps = false);
	void _UpdateButtons ();
};

CTurnRestrictions * CTurnRestrictions::g_pTurnRestrictions;

void CTurnRestrictions::_GetItemText (const rgrestriction_t & _restr, size_t _cColumn, char * _strBuf, size_t _cSize) const {
	if (! _restr.IsValid (g_map.pTypeSet)) {
		::strncpy (_strBuf, "(invalid)", _cSize);
		return;
	}

	::strncpy (_strBuf, "-", _cSize);

	switch (_cColumn) {
		case trNode:     if (_restr.pNode) ::sprintf (_strBuf, "%d", _restr.pNode->dwNodeID); break;

		case trFromRoad:
			if (_restr.pFromRoad && _restr.pFromRoad->pExtra) {
				::sprintf (_strBuf, "%d", _restr.pFromRoad->pExtra->dwRoadID);
				::strcat (_strBuf, _restr.bFromDirection == 0 ? "<-" : "->");
			}
			break;

		case trToRoad:
			if (_restr.pToRoad   && _restr.pToRoad->pExtra) {
				::sprintf (_strBuf, "%d", _restr.pToRoad->pExtra->dwRoadID);
				::strcat (_strBuf, _restr.bToDirection == 0 ? "<-" : "->");
			}
			break;

		case trVehicles:
			if (_restr.AppliedToAllVehicles ())
				::strncpy (_strBuf, "All", _cSize);
			else
				::sprintf (
					_strBuf, "%d%d%d%d %d%d%d%d",
					_restr.bNoEmergency,
					_restr.bNoDelivery,
					_restr.bNoCar,
					_restr.bNoBus,
					_restr.bNoTaxi,
					_restr.bNoPedestrian,
					_restr.bNoBicycle,
					_restr.bNoTruck
				);
			break;

		case trTime:
			if (_restr.AppliedToAllTheTime ())
				::strncpy (_strBuf, "All", _cSize);
			else {
				string_t strTime;
				DisplayTimePeriods (_restr.TimePeriods, strTime, false);
				::strncpy (_strBuf, strTime.c_str (), _cSize);
			}
			break;
	}
}

void CTurnRestrictions::_Sort () {
	switch (m_hList.GetSortingColumn ()) {
		case -1:
			break;

		case trNode:     std::sort (m_TRs.begin (), m_TRs.end (), LessNode     (m_hList.IsSortingAscending ())); break;
		case trFromRoad: std::sort (m_TRs.begin (), m_TRs.end (), LessFromRoad (m_hList.IsSortingAscending ())); break;
		case trToRoad:   std::sort (m_TRs.begin (), m_TRs.end (), LessToRoad   (m_hList.IsSortingAscending ())); break;
		case trVehicles: std::sort (m_TRs.begin (), m_TRs.end (), LessVehicles (m_hList.IsSortingAscending ())); break;
		case trTime:     std::sort (m_TRs.begin (), m_TRs.end (), LessTime     (m_hList.IsSortingAscending ())); break;

		default:
			assert (false);
			break;
	}

	// NOTE: this does not sort virtual list view indeed, but it shows sorting arrows in the header.
	m_hList.Sort ();

	const size_t cItems = m_TRs.size ();
	ListView_SetItemCount (m_hList, cItems);
	SetDlgItemTextF (IDC_TOTAL, IDS_TOTAL, cItems);
}

void CTurnRestrictions::Reset () {
	m_TRs.clear ();

	SetStatus (GetStr (IDS_STATUS_BUILDING_LIST));

	{
		CWaiter w;
		CCritSec::CLock lock (g_cs);

		m_TRs.reserve (g_map.rgrestrictions.size ());

		for (rgrestrictions_t::iterator r = g_map.rgrestrictions.begin (); r != g_map.rgrestrictions.end (); ++ r) {
			const rgrestriction_t & restriction = r->second;
			if (! restriction.IsValid (g_map.pTypeSet))
				continue;
			m_TRs.push_back (restriction);
		}
	}

	_Sort ();

	SetStatusReady ();
}

void CTurnRestrictions::_UpdateButtons () {
	const size_t cSel = ListView_GetSelectionMark (m_hList);
	EnableDlgItem (IDC_LOCATE, cSel != -1);
}

void CTurnRestrictions::_LocateSel (bool _bShowProps) {
	const size_t cSel = ListView_GetSelectionMark (m_hList);
	if (cSel == -1 || cSel >= m_TRs.size ())
		return;
	const rgnode_t * const pRGNodeSel = m_TRs [cSel].pNode;
	if (pRGNodeSel == NULL)
		return;

	CWaiter w;
	CCritSec::CLock lock (g_cs);

	layer_t & polylines = g_map.layers [rkPolyline];
	rgns_t & rgns = polylines.rgns;
	for (rgns_t::iterator iRgn = rgns.begin (); iRgn != rgns.end (); ++ iRgn) {
		rgn_t & rgn = * iRgn;
		if (rgn.IsImmutable () || ! rgn.HasRGNodes ())
			continue;

		const size_t cElements = rgn.elements.size ();
		for (size_t cElement = 0; cElement < cElements; ++ cElement) {
			const rgn_element_t & element = rgn.elements [cElement];
			if (element.bDeleted || ! element.pExtra)
				continue;

			const point_extras_t & PointExtras = element.pExtra->PointExtras;
			const size_t cPoints = PointExtras.size ();
			for (size_t c = 0; c < cPoints; ++ c) {
				const rgnode_t * const pRGNode = PointExtras [c].GetRGNode ();
				if (pRGNode == pRGNodeSel && c < element.points.size ()) {
					const point_t & point = element.points [c];

					SetToCenter (point);
					Refresh ();

					if (_bShowProps)
						ShowNodeProperties (m_hWnd, rgn, cElement, c);

					return;
				}
			}
		}
	}
}

bool CTurnRestrictions::OnInit (HWND _hwndFocus) {
	// Turn restrictions.
	m_hList.Init (m_hWnd, IDC_LIST, "TurnRestrictions");
	m_hList.SetExtStyle (LVS_EX_FULLROWSELECT | LVS_EX_HEADERDRAGDROP);

	AddColumn (m_hList, trNode,     "Node",      true);
	AddColumn (m_hList, trFromRoad, "From road", true);
	AddColumn (m_hList, trToRoad,   "To road",   true);
	AddColumn (m_hList, trVehicles, "Vehicles");
	AddColumn (m_hList, trTime,     "Time");

	AutoSizeColumns (m_hList);
	m_hList.LoadHeaders ();

	Reset ();

	// Locate.
	_UpdateButtons ();

	// Restore position.
	if (IsSettingDefined ("TurnRestrictions_X")) {
		const DWORD dwX = GetSetting ("TurnRestrictions_X", (DWORD) 0);
		const DWORD dwY = GetSetting ("TurnRestrictions_Y", (DWORD) 0);
		::SetWindowPos (m_hWnd, NULL, dwX, dwY, 0, 0, SWP_NOZORDER | SWP_NOSIZE);
		EnsureVisibleOnScreen (m_hWnd);
	}

	return true;
}

void CTurnRestrictions::OnDestroy () {
	m_hList.SaveHeaders ();
	m_hList.Destroy ();
}

void CTurnRestrictions::OnCommand (WORD _wCmd, WORD _wNotifyCode, HWND _hwndCtl) {
	switch (_wCmd) {
		case IDCANCEL:
			Hide ();
			break;

		case IDC_LOCATE:
			_LocateSel ();
			break;

		case IDC_REFRESH:
			Reset ();
			break;
	}
}

bool CTurnRestrictions::OnNotify (int _idCtrl, NMHDR * _pInfo) {
	switch (_pInfo->code) {
		case NM_RCLICK:
			m_hList.PopupMenu_OnHeaderRClick (_pInfo);
			break;

		case LVN_GETDISPINFO: {
			NMLVDISPINFO * const pInfo2 = reinterpret_cast<NMLVDISPINFO *> (_pInfo);
			LVITEM & lvi = pInfo2->item;
			if (_idCtrl == IDC_LIST) {
				if ((size_t) lvi.iItem >= m_TRs.size ())
					break;
				const rgrestriction_t & restr = m_TRs [lvi.iItem];

				if (lvi.mask & LVIF_TEXT)
					_GetItemText (restr, lvi.iSubItem, lvi.pszText, lvi.cchTextMax);
			}

			return true;
		}

		case LVN_COLUMNCLICK: {
			if (_idCtrl == IDC_LIST) {
				// Save selection.
				const size_t cSel = ListView_GetSelectionMark (m_hList);
				rgrestriction_t restrSel;
				const bool bRestoreSel = cSel < m_TRs.size ();
				if (bRestoreSel)
					restrSel = m_TRs [cSel];

				m_hList.Sort_OnColumnClick (_pInfo);
				_Sort ();

				// Restore selection.
				if (bRestoreSel) {
					const size_t cTRs = m_TRs.size ();
					for (size_t c = 0; c < cTRs; ++ c) {
						const rgrestriction_t & restr = m_TRs [c];
						if (
							restr.pNode          == restrSel.pNode          &&
							restr.pFromRoad      == restrSel.pFromRoad      &&
							restr.pToRoad        == restrSel.pToRoad        &&
							restr.bFromDirection == restrSel.bFromDirection &&
							restr.bToDirection   == restrSel.bToDirection
						) {
							SetSelection (m_hList, c);
							EnsureSelectionVisible (m_hList);
							break;
						}
					}
				}
			}

			break;
		}

		case LVN_ITEMCHANGED: {
			const NMLISTVIEW * const pInfo2 = reinterpret_cast<const NMLISTVIEW *> (_pInfo);
			if (pInfo2->uNewState & LVIS_SELECTED)
				_UpdateButtons ();

			break;
		}

		case NM_RETURN:
		case NM_DBLCLK:
			_LocateSel (true);
			break;
	}

	return false;
}

void CTurnRestrictions::OnMove (int _x, int _y) {
	RECT rect;
	::GetWindowRect (m_hWnd, & rect);
	if (rect.left < 0)
		rect.left = 0;
	if (rect.top < 0)
		rect.top = 0;

	SaveSetting ("TurnRestrictions_X", (DWORD) rect.left);
	SaveSetting ("TurnRestrictions_Y", (DWORD) rect.top);
}

//////////////////////////////////

static
void _ResetTurnRestrictions () {
	if (CTurnRestrictions::g_pTurnRestrictions)
		CTurnRestrictions::g_pTurnRestrictions->Reset ();
}

INVOKE_AFTER_REMOVE_MAP (_ResetTurnRestrictions);
INVOKE_ON_BACKUP_MAP    (_ResetTurnRestrictions);
INVOKE_BEFORE_GC        (_ResetTurnRestrictions);

void ShowTurnRestrictions () {
	ShowModeless (CTurnRestrictions::g_pTurnRestrictions, g_hWnd);
}