//
// GPSMapEdit
// (c) Konstantin Galichsky (kg@geopainting.com), 2002-2009
//
// Stuff for 'Edit|Undo' and 'Edit|Redo' commands.
//

# include "StdAfx.h"
# include "UndoRedo.h"
# include "Menu.h"
# include "MapView.h"
# include "Position.h"
# include "BackupAction.h"
# include "CombinedAction.h"
# include "SaveState.h"
# include "Globals.h"

CInvokeBeforeGC * CInvokeBeforeGC::s_pFirst;

CInvokeBeforeGC::CInvokeBeforeGC (pvfv_t _pf) :
	m_pf (_pf)
{
	// Insert to the head of the list.
	m_pNext = CInvokeBeforeGC::s_pFirst;
	CInvokeBeforeGC::s_pFirst = this;
}

/////////////////////////////////////////////////

bool g_bModified = false;

actions_t g_Undo;
actions_t g_Redo;

bool CanUndo () {return ! g_Undo.empty ();}
bool CanRedo () {return ! g_Redo.empty ();}

void ResetUndoRedo (bool _bCollectGarbage) {
	actions_t::iterator i;
	for (i = g_Undo.begin (); i != g_Undo.end (); ++ i)
		delete * i;
	g_Undo.clear ();

	for (i = g_Redo.begin (); i != g_Redo.end (); ++ i)
		delete * i;
	g_Redo.clear ();

	//
	// Remove deleted objects.
	//
	if (_bCollectGarbage) {
		CCritSec::CLock lock (g_cs);

		// Just before collecting garbage, let all subscribers to remove links to deleted objects.
		for (const CInvokeBeforeGC * pI = CInvokeBeforeGC::s_pFirst; pI != NULL; pI = pI->m_pNext)
			pI->m_pf ();

		// Collect garbage in turn restrictions.
		rgrestrictions_t::iterator rNext;
		for (rgrestrictions_t::iterator r = g_map.rgrestrictions.begin (); r != g_map.rgrestrictions.end (); r = rNext) {
			rgrestriction_t & restriction = r->second;

			rNext = r;
			++ rNext;

			if (
				restriction.pFromRoad == NULL || restriction.pFromRoad->IsDeleted () ||
				restriction.pToRoad   == NULL || restriction.pToRoad  ->IsDeleted ()
			)
				g_map.rgrestrictions.erase (r);
		}

		// Collect garbage in objects.
		rgn_t * pPrevRgn = NULL;
		for (rgn_t * pRgn = g_map.pFirstRgn; pRgn; pRgn = pRgn->pNextRgn) {
			if (pRgn->IsDeleted ()) {
				rgn_t * const pNextRgn = pRgn->pNextRgn;
				if (pPrevRgn)
					pPrevRgn->pNextRgn = pNextRgn;
				else
					g_map.pFirstRgn = pNextRgn;
			} else {
				pPrevRgn = pRgn;
			}
		}
		g_map.pLastRgn = pPrevRgn;

		for (size_t cKind = 0; cKind < rkMax; ++ cKind) {
			layer_t & layer = g_map.layers [cKind];
			rgns_t & rgns = layer.rgns;

			for (rgns_t::iterator iRgn = rgns.begin (); iRgn != rgns.end ();) {
				rgn_t & rgn = * iRgn;
				if (rgn.IsDeleted ())
					iRgn = layer.rgns.erase (iRgn);
				else
					++ iRgn;
			}
			layer.cDeletedRgns = 0;
		}

		// Collect garbage in attached files.
		for (file_refs_t::iterator iFR = g_map.file_refs.begin (); iFR != g_map.file_refs.end ();) {
			file_ref_t & fr = * iFR;
			if (fr.bDeleted)
				iFR = g_map.file_refs.erase (iFR);
			else
				++ iFR;
		}

		// Collect garbage in renderers.
		for (renderers_t::iterator ri = g_map.renderers.begin (); ri != g_map.renderers.end ();) {
			renderer_base_t * const pRenderer = & ri->data ();
			if (pRenderer->bDeleted)
				ri = g_map.renderers.erase (ri);
			else
				++ ri;
		}
	}
}

bool PerformEditAction (CEditAction * _pAction) {
	assert (_pAction);

	CEditAction::cu_t cu = _pAction->CanUndo ();
	if (cu == CEditAction::cuBackup) {
		// Estimate free memory.
		bool bHaveMemory = false;
		{
			CCritSec::CLock lock (g_cs);
			const size_t cMapOccupied = g_map.occupied ();

			MEMORYSTATUS ms = {sizeof (ms)};
			::GlobalMemoryStatus (& ms);
			const size_t cFree = ms.dwAvailPhys;

			bHaveMemory = cFree > cMapOccupied;
		}
		if (bHaveMemory) {
			_pAction = new CBackupAction (_pAction);
		} else
			cu = CEditAction::cuNo;
	}
	if (cu == CEditAction::cuNo && ! g_Undo.empty ()) {
		string_t strName;
		_pAction->GetName (strName);

		if (
			YesNoMsgBox (
				"Please be aware that the menu item 'Edit | Undo' will not be able to undone any actions made before this one:\n"
				" '%s'.\n"
				"\n"
				"Do you want to continue?\n"
				"(Press YES to apply the action or NO to cancel it.)",
				strName.c_str ()
			) == false
		) {
			delete _pAction;
			return false;
		}
	}

	if (! _pAction->Do ()) {
		string_t strName;
		_pAction->GetName (strName);

		ErrorMsgBox ("The action failed:\n %s.", strName.c_str ());

		delete _pAction;
		return false;
	}

	const bool bWasModified = g_bModified;
	g_bModified = true;

	if (cu == CEditAction::cuYes || cu == CEditAction::cuBackup) {
		// NOTE: action may modify selection in Do().
		if (cu == CEditAction::cuYes && ! IsInCombinedAction ())
			_pAction = new CSaveStateWrapperAction (_pAction, g_Selection, g_SelPoints, bWasModified);

		// Reset Redo stack.
		for (actions_t::iterator i = g_Redo.begin (); i != g_Redo.end (); ++ i)
			delete * i;
		g_Redo.clear ();

		g_Undo.push_back (_pAction);
	} else {
		ResetUndoRedo ();
		delete _pAction;
	}

	if (! IsInCombinedAction ())
		UpdateMenu ();

	return true;
}

void UndoEditAction () {
	if (g_Undo.empty ())
		return;

	CEditAction * const pAction = g_Undo.back ();
	g_Undo.pop_back ();

	pAction->Undo ();

	g_Redo.push_back (pAction);

	// TODO: ? restore the flag ?
//	g_bModified = true;
}

void RedoEditAction () {
	if (g_Redo.empty ())
		return;

	CEditAction * const pAction = g_Redo.back ();
	g_Redo.pop_back ();

	pAction->Redo ();

	g_Undo.push_back (pAction);

	g_bModified = true;
}

