//
// GPSMapEdit
// (c) Konstantin Galichsky (kg@geopainting.com), 2007-2009
//
// Raster map mosaic downloaded from WWW.
//

# include "StdAfx.h"
# include "Resource.h"
# include "WebMap.h"
# include "MapView.h"
# include "EditHelpers.h"
# include "PrintUnits.h"
# include "ParseHelpers.h"
# include "PtAlgo.h"
# include "Position.h"
# include "Settings.h"
# include "Str.h"
# include "Dialog.h"
# include "ComboBox.h"
# include "EditText.h"
# include "Wrappers.h"
# include "fast_floor.h"
# include "Globals.h"

# include "WebMap.inl"

bool g_bUseGMOnline;
bool g_bUseGMCache;
bool g_bWriteGMCache;
bool g_bWriteEarthSlicerNames;
DWORD g_dwGMVersion;

void GetSetting_GoogleMapsCache (char _strPath [MAX_PATH]) {
	GetSetting ("GoogleMapsCache", _strPath, MAX_PATH, "c:\\GoogleMaps");
}

float GetWebMapShiftDXInM () {
	const int dx = GetSetting ("GMShiftDX", (DWORD) 0);
	return dx*0.1f;
}

float GetWebMapShiftDYInM () {
	const int dy = GetSetting ("GMShiftDY", (DWORD) 0);
	return dy*0.1f;
}

class CShiftWebMap : public CDialog {
  public:
	CShiftWebMap () : CDialog (IDD_EDIT_SHIFT) {}

	static CShiftWebMap * s_pShiftWebMap;

	void Update () {
		_LoadList ();
		_FillList ();
	}

	// Override.
	virtual bool OnInit (HWND _hwndFocus);
	virtual void OnCommand (WORD _wCmd, WORD _wNotifyCode, HWND _hwndCtl);
	virtual bool OnNotify (int _idCtrl, NMHDR * _pInfo);

  private:
	HWND m_hList;
	void _FillList (size_t _cSel = -1);
	void _OnSelChange ();

	struct item_t {
		float fDX, fDY;
		point_t pt;
		string_t strName;

		bool operator == (const item_t & _other) const {return strName == _other.strName;}
		bool operator <  (const item_t & _other) const {return strName <  _other.strName;}
	};
	typedef std::vector<item_t> items_t;
	items_t m_items;

	void _InitListNames ();
	void _LoadList ();
	void _SaveList ();
	void _UpdateList (const string_t & _strNameToSel);

	void _Apply ();

	void _EnableButtons (size_t _cItem);
};

CShiftWebMap * CShiftWebMap::s_pShiftWebMap;

void CShiftWebMap::_EnableButtons (size_t _cItem) {
	EnableDlgItem (IDC_RENAME, _cItem != 0);
	EnableDlgItem (IDC_DELETE, _cItem != 0);
}

void CShiftWebMap::_FillList (size_t _cSel) {
	EmptyComboBox (m_hList);

	const size_t cItems = m_items.size ();
	for (size_t cItem = 0; cItem < cItems; ++ cItem) {
		const item_t & item = m_items [cItem];
		AddComboBox (m_hList, cItem, cItem, item.strName.c_str ());
	}

	if (_cSel == -1)
		_cSel = GetSetting ("GMShiftIdx", (DWORD) 0);
	if (_cSel >= cItems)
		_cSel = 0;

	SelectComboBox (m_hList, _cSel);
	_OnSelChange ();
}

void CShiftWebMap::_OnSelChange () {
	int nItem = 0;
	if (GetComboBoxItemData (m_hWnd, IDC_LIST, nItem) && (size_t) nItem < m_items.size ()) {
		SaveSetting ("GMShiftIdx", (DWORD) nItem);

		const item_t & item = m_items [nItem];
		SetDlgItemTextF (IDC_DX, "%0.1f", item.fDX);
		SetDlgItemTextF (IDC_DY, "%0.1f", item.fDY);

		_Apply ();

		_EnableButtons (nItem);
	}
}

void CShiftWebMap::_InitListNames () {
	typedef std::map<string_t, size_t> mapNames2Count_t;
	mapNames2Count_t mapNames2Count;

	const size_t cItems = m_items.size ();
	for (size_t c = 0; c < cItems; ++ c) {
		item_t & item = m_items [c];

		char strName [MAX_PATH];
		::strcpy (strName, item.strName.c_str ());

		const mapNames2Count_t::iterator i = mapNames2Count.find (strName);
		if (i == mapNames2Count.end ()) {
			mapNames2Count [strName] = 1;
		} else {
			size_t & cCount = i->second;
			::sprintf (strName + ::strlen (strName), " (%d)", cCount);
			++ cCount;
		}

		item.strName = strName;
	}
}

void CShiftWebMap::_LoadList () {
	m_items.clear ();

	const bool bDefault = GetSetting ("GMShiftIdx", (DWORD) 0) == 0;

	const size_t cItems = GetSetting ("GMShifts", (DWORD) 0);
	if (cItems == 0) {
		item_t item;
		item.strName = "Default";
		item.pt      = point_t (0, 0);
		item.fDX     = GetWebMapShiftDXInM ();
		item.fDY     = GetWebMapShiftDYInM ();

		m_items.push_back (item);

		return;
	}

	m_items.reserve (cItems);
	for (size_t c = 0; c < cItems; ++ c) {
		char strKey [32];
		::sprintf (strKey, "GMShift.%d", c);
		
		char strValue [256 + MAX_PATH];
		GetSetting (strKey, strValue, ARRAY_SIZE (strValue), "");

		item_t item;

		const char * p = strValue;
		const char * const pEnd = strValue + ::strlen (strValue);

		// DX.
		if (! ParseReal (p, pEnd, & p, item.fDX))
			continue;
		if (p < pEnd && * p != '*')
			continue;
		++ p;

		// DY.
		if (! ParseReal (p, pEnd, & p, item.fDY))
			continue;
		if (p < pEnd && * p != '*')
			continue;
		++ p;

		// X.
		if (! ParseReal (p, pEnd, & p, item.pt.x))
			continue;
		if (p < pEnd && * p != '*')
			continue;
		++ p;

		// Y.
		if (! ParseReal (p, pEnd, & p, item.pt.y))
			continue;
		if (p < pEnd && * p != '*')
			continue;
		++ p;

		// Name.
		item.strName.assign (p, pEnd);

		if (bDefault && c == 0) {
			// NOTE: this is required for 'reset shift'.
			item.fDX = GetWebMapShiftDXInM ();
			item.fDY = GetWebMapShiftDYInM ();
		}

		m_items.push_back (item);
	}

	_InitListNames ();
}

void CShiftWebMap::_SaveList () {
	const size_t cItems = m_items.size ();
	SaveSetting ("GMShifts", (DWORD) cItems);
	for (size_t c = 0; c < cItems; ++ c) {
		const item_t & item = m_items [c];

		char strKey [32];
		::sprintf (strKey, "GMShift.%d", c);
		char strValue [256 + MAX_PATH];
		::sprintf (strValue, "%0.1f*%0.1f*%0.1f*%0.1f*%s", item.fDX, item.fDY, item.pt.x, item.pt.y, item.strName.c_str ());

		SaveSetting (strKey, strValue);
	}
}

void CShiftWebMap::_UpdateList (const string_t & _strNameToSel) {
	_InitListNames ();

	// NOTE: don't sort 0-th default item.
	std::sort (m_items.begin () + 1, m_items.end ());

	_SaveList ();
	_FillList ();

	const size_t cIdx = ::SendMessage (m_hList, CB_FINDSTRINGEXACT, 0, (LPARAM) _strNameToSel.c_str ());
	SelectComboBox (m_hList, cIdx);
	_OnSelChange ();
}

bool CShiftWebMap::OnInit (HWND _hwndFocus) {
	const float fDX = GetWebMapShiftDXInM (); 
	const float fDY = GetWebMapShiftDYInM ();

	if (g_DistanceUnit == unMetric) {
		SetDlgItemTextF (IDC_DX, "%.1f", fDX);
		SetDlgItemTextF (IDC_DY, "%.1f", fDY);
	} else {
		SetDlgItemTextF (IDC_DX, "%.0f", fDX*c_feet_per_m);
		SetDlgItemTextF (IDC_DY, "%.0f", fDY*c_feet_per_m);

		SetDlgItemText (IDC_UNIT_X, "ft");
		SetDlgItemText (IDC_UNIT_Y, "ft");
	}

	// List.
	m_hList = GetDlgItem (IDC_LIST);
	_LoadList ();
	_FillList ();

	// Apply.
	EnableDlgItem (ID_APPLY, false);

	return true;
}

void CShiftWebMap::OnCommand (WORD _wCmd, WORD _wNotifyCode, HWND _hwndCtl) {
	switch (_wCmd) {
		case IDC_DX:
		case IDC_DY:
			if (_wNotifyCode == EN_CHANGE) {
				const bool bOK =
					ValidateAsFloat (GetDlgItem (IDC_DX)) &&
					ValidateAsFloat (GetDlgItem (IDC_DY));
				EnableDlgItem (ID_APPLY, bOK);
			}
			break;

		case IDC_SAVE: {
			string_t strName = "New Item";
			if (! EditText (m_hWnd, strName, "Name", false))
				break;

			rect_t rectVisible;
			GetVisibleRect (rectVisible);

			item_t item;
			item.strName = strName;
			item.pt.x    = rectVisible.MeanX ();
			item.pt.y    = rectVisible.MeanY ();
			item.fDX     = GetWebMapShiftDXInM ();
			item.fDY     = GetWebMapShiftDYInM ();
			m_items.push_back (item);

			_UpdateList (strName);

			break;
		}

		case IDC_RENAME: {
			int nItem = 0;
			if (GetComboBoxItemData (m_hWnd, IDC_LIST, nItem) && (size_t) nItem < m_items.size ()) {
				if (nItem == 0)
					break;

				if (! EditText (m_hWnd, m_items [nItem].strName, "Name", false))
					break;

				// NOTE: save string copy by value since m_items[] will be sorted.
				const string_t strName = m_items [nItem].strName;
				_UpdateList (strName);
			}	
			break;
		}

		case IDC_DELETE: {
			int nItem = 0;
			if (GetComboBoxItemData (m_hWnd, IDC_LIST, nItem) && (size_t) nItem < m_items.size ()) {
				if (nItem == 0)
					break;

				m_items.erase (m_items.begin () + nItem);
				_SaveList ();

				_FillList (0);
			}	
			break;
		}

		case IDC_LIST:
			if (_wNotifyCode == CBN_SELCHANGE)
				_OnSelChange ();
			break;

		case ID_APPLY:
			_Apply ();
			break;

		case IDOK:
		case IDCANCEL:
			Hide ();
			break;
	}
}

bool CShiftWebMap::OnNotify (int _idCtrl, NMHDR * _pInfo) {
	if (_pInfo->code == UDN_DELTAPOS && (_idCtrl == IDC_SPIN_DX || _idCtrl == IDC_SPIN_DY)) {
		const NMUPDOWN * const pUD = reinterpret_cast<const NMUPDOWN *> (_pInfo);
		const bool bUp = pUD->iDelta > 0;
		const UINT uiID = _idCtrl == IDC_SPIN_DX ? IDC_DX : IDC_DY;

		float f = GetTextAsFloat (GetDlgItem (uiID));
		if (bUp)
			f -= 0.5f;
		else
			f += 0.5f;

		SetDlgItemTextF (uiID, "%.1f", f);

		_Apply ();
	}
	return false;
}

void CShiftWebMap::_Apply () {
	float fDX = GetTextAsFloat (GetDlgItem (IDC_DX));
	float fDY = GetTextAsFloat (GetDlgItem (IDC_DY));
	if (g_DistanceUnit != unMetric) {
		fDX /= c_feet_per_m;
		fDY /= c_feet_per_m;
	}
	SaveSetting ("GMShiftDX", (DWORD) fast_floor (fDX*10 + .5f));
	SaveSetting ("GMShiftDY", (DWORD) fast_floor (fDY*10 + .5f));

	int nItem = 0;
	if (GetComboBoxItemData (m_hWnd, IDC_LIST, nItem) && (size_t) nItem < m_items.size ()) {
		item_t & item = m_items [nItem];
		item.fDX = fDX;
		item.fDY = fDY;

		_SaveList ();
	}

	Refresh ();
}

/////////////////////////////////////////////////////////////

void ShiftWebMap (HWND _hWnd) {
	ShowModeless (CShiftWebMap::s_pShiftWebMap, g_hWnd);

	if (CShiftWebMap::s_pShiftWebMap)
		CShiftWebMap::s_pShiftWebMap->Update ();
}

void ResetWebMapShift () {
	SaveSetting ("GMShiftDX", (DWORD) 0);
	SaveSetting ("GMShiftDY", (DWORD) 0);
	SaveSetting ("GMShiftIdx", (DWORD) 0);

	if (CShiftWebMap::s_pShiftWebMap)
		CShiftWebMap::s_pShiftWebMap->Update ();

	Refresh ();
}
